  #include <fcntl.h>
  #include <unistd.h>

#include "pxll.h"

static int lookup_field (int tag, int label);

static
inline
pxll_int
get_typecode (object * ob)
{
  if (IMMEDIATE(ob)) {
    if (IS_INTEGER(ob)) {
      return TC_INT;
    } else {
      return (pxll_int)ob & 0xff;
    }
  } else {
    return (pxll_int)*((pxll_int *)ob) & 0xff;
  }
}

// for pvcase/nvcase
static
inline
pxll_int
get_case (object * ob)
{
  if (is_immediate (ob)) {
    if (is_int (ob)) {
      return TC_INT;
    } else {
      return (pxll_int) ob;
    }
  } else {
    return (pxll_int)*((pxll_int *)ob) & 0xff;
  }
}

// for pvcase/nvcase
static
inline
pxll_int
get_case_noint (object * ob)
{
  if (is_immediate (ob)) {
    return (pxll_int) ob;
  } else {
    return (pxll_int) * ((pxll_int*) ob) & 0xff;
  }
}

// for pvcase/nvcase
static
inline
pxll_int
get_case_imm (object * ob)
{
  return (pxll_int)ob;
}

static
inline
pxll_int
get_case_tup (object * ob)
{
  return (pxll_int)*((pxll_int *)ob) & 0xff;
}

static
inline
pxll_int
get_imm_payload (object * ob)
{
  return ((pxll_int) ob) >> 8;
}

static
pxll_int
get_tuple_size (object * ob)
{
  header * h = (header *) ob;
  return (*h)>>8;
}

static
void
indent (int n)
{
  while (n--) {
    fprintf (stdout, "  ");
  }
}

void print_string (object * ob, int quoted);
void print_list (pxll_pair * l);

// this is kinda lame, it's part pretty-printer, part not.
static
object *
dump_object (object * ob, int depth)
{
  // indent (depth);
  if (depth > 100) {
    fprintf (stdout , "...");
    return (object *) PXLL_UNDEFINED;
  }
  if (!ob) {
    fprintf (stdout, "<null>");
  } else if (is_int (ob)) {
    // integer
    fprintf (stdout, "%zd", unbox (ob));
  } else {
    int tc = is_immediate (ob);
    switch (tc) {
    case TC_CHAR:
      if ((pxll_int)ob>>8 == 257) {
	// deliberately out-of-range character
	fprintf (stdout, "#\\eof");
      } else {
	char ch = ((char)((pxll_int)ob>>8));
	switch (ch) {
	case '\000': fprintf (stdout, "#\\nul"); break;
	case ' '   : fprintf (stdout, "#\\space"); break;
	case '\n'  : fprintf (stdout, "#\\newline"); break;
	case '\r'  : fprintf (stdout, "#\\return"); break;
	case '\t'  : fprintf (stdout, "#\\tab"); break;
	default    : fprintf (stdout, "#\\%c", ch);
	}
      }
      break;
    case TC_BOOL:
      fprintf (stdout, ((pxll_int)ob >> 8) & 0xff ? "#t" : "#f");
      break;
    case TC_NIL:
      fprintf (stdout, "()");
      break;
    case TC_UNDEFINED:
      fprintf (stdout, "#u");
      break;
    case TC_EMPTY_VECTOR:
      fprintf (stdout, "#()");
      break;
    case 0: {
      // structure
      header h = (header) (ob[0]);
      int tc = h & 0xff;
      switch (tc) {
      case TC_SAVE: {
	// XXX fix me - now holds saved registers
        pxll_save * s = (pxll_save* ) ob;
        fprintf (stdout, "<save pc=%p\n", s->pc);
        dump_object ((object *) s->lenv, depth+1); fprintf (stdout, "\n");
        dump_object ((object *) s->next, depth+1); fprintf (stdout, ">");
      }
        break;
      case TC_CLOSURE: {
        pxll_closure * c = (pxll_closure *) ob;
        //fprintf (stdout, "<closure pc=%p\n", c->pc);
        //dump_object ((object *) c->lenv, depth+1); fprintf (stdout, ">\n");
	fprintf (stdout, "<closure pc=%p lenv=%p>", c->pc, c->lenv);
      }
        break;
      case TC_TUPLE: {
        pxll_tuple * t = (pxll_tuple *) ob;
        pxll_int n = get_tuple_size (ob);
        int i;
	fprintf (stdout, "<tuple\n");
        for (i=0; i < n-1; i++) {
          dump_object ((object *) t->val[i], depth + 1); fprintf (stdout, "\n");
        }
        dump_object ((object *) t->next, depth + 1);
        fprintf (stdout, ">");
      }
	break;
      case TC_VECTOR: {
        pxll_vector * t = (pxll_vector *) ob;
        pxll_int n = get_tuple_size (ob);
        int i;
	fprintf (stdout, "#(");
        for (i=0; i < n; i++) {
          dump_object ((object *) t->val[i], depth+1);
	  if (i < n-1) {
	    fprintf (stdout, " ");
	  }
        }
        fprintf (stdout, ")");
      }
	break;
      case TC_VEC16: {
        pxll_vec16 * t = (pxll_vec16 *) ob;
        pxll_int n = t->len;
        int i;
	fprintf (stdout, "#16(");
        for (i=0; i < n; i++) {
	  fprintf (stdout, "%d", t->data[i]);
	  if (i < n-1) {
	    fprintf (stdout, " ");
	  }
        }
        fprintf (stdout, ")");
      }
	break;
      case TC_PAIR:
	print_list ((pxll_pair *) ob);
        break;
      case TC_STRING:
	print_string (ob, 1);
	break;
      case TC_BUFFER: {
	pxll_int n = get_tuple_size (ob);
	fprintf (stdout, "<buffer %" PRIuPTR " words %" PRIuPTR " bytes>", n, n * (sizeof(pxll_int)));
	break;
      }
      case TC_SYMBOL:
	print_string (ob[1], 0);
	break;
      default: {
        pxll_vector * t = (pxll_vector *) ob;
        pxll_int n = get_tuple_size (ob);
        int i;
	fprintf (stdout, "{u%d ", (tc - TC_USEROBJ)>>2);
        for (i=0; i < n; i++) {
          dump_object ((object *) t->val[i], depth+1);
	  if (i < n-1) {
	    fprintf (stdout, " ");
	  }
        }
        fprintf (stdout, "}");
      }
      }
    }
      break;
    case TC_USERIMM:
      // a user immediate unit-type...
      fprintf (stdout, "<u%" PRIuPTR ">", (((pxll_int)ob)>>8));
    }
  }
  return (object *) PXLL_UNDEFINED;
}

// for gdb...
void
DO (object * x)
{
  dump_object (x, 0);
  fprintf (stdout, "\n");
  fflush (stdout);
}

// for debugging
void
stack_depth_indent (object * k)
{
  while (k != PXLL_NIL) {
    k = k[1];
    fprintf (stderr, "  ");
  }
}

void
print_string (object * ob, int quoted)
{
  pxll_string * s = (pxll_string *) ob;
  char * ps = s->data;
  int i;
  //fprintf (stderr, "<printing string of len=%d (tuple-len=%d)>\n", s->len, get_tuple_size (ob));
  if (quoted) {
    fputc ('"', stdout);
  }
  for (i=0; i < (s->len); i++, ps++) {
    if (*ps == '"') {
      fputc ('\\', stdout);
      fputc ('"', stdout);
    } else {
      if (isprint(*ps)) {
	fputc (*ps, stdout);
      } else {
	fprintf (stdout, "\\0x%02x", *ps);
      }
    }
    if (i > 50) {
      fprintf (stdout, "...");
      break;
    }
  }
  if (quoted) {
    fputc ('"', stdout);
  }
}

void
print_list (pxll_pair * l)
{
  fprintf (stdout, "(");
  while (1) {
    object * car = l->car;
    object * cdr = l->cdr;
    dump_object (car, 0);
    if (cdr == PXLL_NIL) {
      fprintf (stdout, ")");
      break;
    } else if (!is_immediate (cdr) && GET_TYPECODE (*cdr) == TC_PAIR) {
      fprintf (stdout, " ");
      l = (pxll_pair *) cdr;
    } else {
      fprintf (stdout, " . ");
      dump_object (cdr, 0);
      fprintf (stdout, ")");
      break;
    }
  }
}

int
read_header (FILE * file)
{
  int depth = 0;
  // tiny lisp 'skipper' (as opposed to 'reader')
  do {
    char ch = fgetc (file);
    switch (ch) {
    case '(':
      depth++;
      break;
    case ')':
      depth--;
      break;
    case '"':
      while (fgetc (file) != '"') {
        // empty body
      }
      break;
    default:
      break;
    }
  } while (depth);
  // read terminating newline
  fgetc (file);
  return 0;
}

#ifndef NO_RANGE_CHECK
// used to check array references.  some day we might try to teach
//   the compiler when/how to skip doing this...
static
void
inline
range_check (unsigned int length, unsigned int index)
{
  if (index >= length) {
    fprintf (stderr, "array/string reference out of range: %d[%d]\n", length, index);
    abort();
  }
}
#else
static
void
inline
range_check (unsigned int length, unsigned int index)
{
}
#endif

pxll_int verbose_gc = 1;
pxll_int clear_fromspace = 0;
pxll_int clear_tospace = 0;

pxll_int vm (int argc, char * argv[]);

#include "rdtsc.h"

unsigned long long gc_ticks = 0;

static
void
clear_space (object * p, pxll_int n)
{
  while (n--) {
    *p++ = PXLL_NIL;
  }
}

int
main (int argc, char * argv[])
{
  heap0 = malloc (sizeof (object) * heap_size);
  heap1 = malloc (sizeof (object) * heap_size);
  if (!heap0 || !heap1) {
    fprintf (stderr, "unable to allocate heap\n");
    return -1;
  } else {
    unsigned long long t0, t1;
    pxll_int result;
    if (clear_tospace) {
      clear_space (heap0, heap_size);
    }
    t0 = rdtsc();
    result = vm (argc, argv);
    t1 = rdtsc();
    dump_object ((object *) result, 0);
    fprintf (stdout, "\n");
    fprintf (stderr, "{total ticks: %lld gc ticks: %lld}\n", t1 - t0, gc_ticks);
    return (int) result;
  }
}

// CONSTRUCTED LITERALS //
pxll_string constructed_0 = {STRING_HEADER(8), 8, "system: " };
pxll_string constructed_1 = {STRING_HEADER(1), 1, "\n" };
pxll_string constructed_2 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_3 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_4 = {STRING_HEADER(6), 6, ".c -o " };
pxll_string constructed_5 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_6 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_7 = {STRING_HEADER(2), 2, "-O" };
pxll_string constructed_8 = {STRING_HEADER(0), 0, "" };
pxll_string constructed_9 = {STRING_HEADER(6), 6, "CFLAGS" };
pxll_string constructed_10 = {STRING_HEADER(2), 2, "CC" };
pxll_string constructed_11 = {STRING_HEADER(13), 13, "compiling...\n" };
pxll_string constructed_12 = {STRING_HEADER(6), 6, "done.\n" };
pxll_string constructed_13 = {STRING_HEADER(2), 2, "}}" };
pxll_string constructed_14 = {STRING_HEADER(10), 10, "  } break;" };
pxll_string constructed_15 = {STRING_HEADER(10), 10, "     case " };
pxll_string constructed_16 = {STRING_HEADER(9), 9, ": return " };
pxll_string constructed_17 = {STRING_HEADER(8), 8, "; break;" };
pxll_string constructed_18 = {STRING_HEADER(18), 18, "  switch (label) {" };
pxll_string constructed_19 = {STRING_HEADER(7), 7, "  case " };
pxll_string constructed_20 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_21 = {STRING_HEADER(16), 16, "{ switch (tag) {" };
pxll_string constructed_22 = {STRING_HEADER(44), 44, "static int lookup_field (int tag, int label)" };
pxll_string constructed_23 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_24 = {STRING_HEADER(25), 25, "return (pxll_int) result;" };
pxll_string constructed_25 = {STRING_HEADER(8), 8, "Lreturn:" };
pxll_string constructed_26 = {STRING_HEADER(37), 37, "abcdefghijklmnopqrstuvwxyz_0123456789" };
pxll_string constructed_27 = {STRING_HEADER(5), 5, "_bang" };
pxll_string constructed_28 = {STRING_HEADER(6), 6, "_splat" };
pxll_string constructed_29 = {STRING_HEADER(9), 9, "_question" };
pxll_string constructed_30 = {STRING_HEADER(1), 1, "_" };
pxll_string constructed_31 = {STRING_HEADER(5), 5, "_plus" };
pxll_string constructed_32 = {STRING_HEADER(8), 8, "_percent" };
pxll_int constructed_33[] = {UPTR(33,21),UOHEAD(3,1),9474,UPTR0(32),UITAG(0),UOHEAD(3,1),11010,UPTR0(31),UPTR(33,1),UOHEAD(3,1),11522,UPTR0(30),UPTR(33,5),UOHEAD(3,1),16130,UPTR0(29),UPTR(33,9),UOHEAD(3,1),10754,UPTR0(28),UPTR(33,13),UOHEAD(3,1),8450,UPTR0(27),UPTR(33,17)};
pxll_string constructed_34 = {STRING_HEADER(5), 5, "minus" };
pxll_string constructed_35 = {STRING_HEADER(1), 1, "_" };
pxll_int constructed_36[] = {TC_NIL};
pxll_int constructed_37[] = {TC_NIL};
pxll_string constructed_38 = {STRING_HEADER(1), 1, "_" };
pxll_string constructed_39 = {STRING_HEADER(20), 20, "(pxll_int)PXLL_FALSE" };
pxll_string constructed_40 = {STRING_HEADER(5), 5, "false" };
// symbol false
pxll_int constructed_41[] = {UPTR(41,1), SYMBOL_HEADER, UPTR0(40), INTCON(0)};
pxll_string constructed_42 = {STRING_HEADER(19), 19, "(pxll_int)PXLL_TRUE" };
pxll_string constructed_43 = {STRING_HEADER(4), 4, "true" };
// symbol true
pxll_int constructed_44[] = {UPTR(44,1), SYMBOL_HEADER, UPTR0(43), INTCON(1)};
pxll_string constructed_45 = {STRING_HEADER(4), 4, "bool" };
// symbol bool
pxll_int constructed_46[] = {UPTR(46,1), SYMBOL_HEADER, UPTR0(45), INTCON(2)};
pxll_string constructed_47 = {STRING_HEADER(6), 6, "TC_NIL" };
pxll_string constructed_48 = {STRING_HEADER(3), 3, "nil" };
// symbol nil
pxll_int constructed_49[] = {UPTR(49,1), SYMBOL_HEADER, UPTR0(48), INTCON(3)};
pxll_string constructed_50 = {STRING_HEADER(4), 4, "list" };
// symbol list
pxll_int constructed_51[] = {UPTR(51,1), SYMBOL_HEADER, UPTR0(50), INTCON(4)};
pxll_string constructed_52 = {STRING_HEADER(6), 6, "UITAG(" };
pxll_string constructed_53 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_54 = {STRING_HEADER(9), 9, "TC_SYMBOL" };
pxll_string constructed_55 = {STRING_HEADER(1), 1, "t" };
// symbol t
pxll_int constructed_56[] = {UPTR(56,1), SYMBOL_HEADER, UPTR0(55), INTCON(5)};
pxll_string constructed_57 = {STRING_HEADER(6), 6, "symbol" };
// symbol symbol
pxll_int constructed_58[] = {UPTR(58,1), SYMBOL_HEADER, UPTR0(57), INTCON(6)};
pxll_string constructed_59 = {STRING_HEADER(7), 7, "TC_PAIR" };
pxll_string constructed_60 = {STRING_HEADER(4), 4, "cons" };
// symbol cons
pxll_int constructed_61[] = {UPTR(61,1), SYMBOL_HEADER, UPTR0(60), INTCON(7)};
pxll_string constructed_62 = {STRING_HEADER(6), 6, "UOTAG(" };
pxll_string constructed_63 = {STRING_HEADER(1), 1, ")" };
pxll_int constructed_64[] = {TC_NIL};
pxll_string constructed_65 = {STRING_HEADER(3), 3, "..." };
// symbol ...
pxll_int constructed_66[] = {UPTR(66,1), SYMBOL_HEADER, UPTR0(65), INTCON(8)};
pxll_string constructed_67 = {STRING_HEADER(79), 79, "Why, sometimes I've believed as many as six impossible things before breakfast." };
pxll_string constructed_68 = {STRING_HEADER(29), 29, "\n***\nRuntime Error, halting: " };
pxll_string constructed_69 = {STRING_HEADER(4), 4, "FUN_" };
pxll_string constructed_70 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_71 = {STRING_HEADER(4), 4, " = r" };
pxll_string constructed_72 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_73 = {STRING_HEADER(12), 12, "PXLL_RETURN(" };
pxll_string constructed_74 = {STRING_HEADER(2), 2, ");" };
pxll_string constructed_75 = {STRING_HEADER(14), 14, " = (object *) " };
pxll_string constructed_76 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_77 = {STRING_HEADER(8), 8, "// dead " };
pxll_string constructed_78 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_79 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_80 = {STRING_HEADER(26), 26, " = (object*) &constructed_" };
pxll_string constructed_81 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_82 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_83 = {STRING_HEADER(26), 26, " = (object *) constructed_" };
pxll_string constructed_84 = {STRING_HEADER(4), 4, "[0];" };
pxll_string constructed_85 = {STRING_HEADER(6), 6, "string" };
// symbol string
pxll_int constructed_86[] = {UPTR(86,1), SYMBOL_HEADER, UPTR0(85), INTCON(9)};
pxll_string constructed_87 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_88 = {STRING_HEADER(8), 8, "} else {" };
pxll_string constructed_89 = {STRING_HEADER(17), 17, "if PXLL_IS_TRUE(r" };
pxll_string constructed_90 = {STRING_HEADER(3), 3, ") {" };
pxll_string constructed_91 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_92 = {STRING_HEADER(8), 8, "} else {" };
pxll_string constructed_93 = {STRING_HEADER(16), 16, "if PXLL_IS_TRUE(" };
pxll_string constructed_94 = {STRING_HEADER(3), 3, ") {" };
pxll_string constructed_95 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_96 = {STRING_HEADER(5), 5, "arrow" };
// symbol arrow
pxll_int constructed_97[] = {UPTR(97,1), SYMBOL_HEADER, UPTR0(96), INTCON(10)};
pxll_string constructed_98 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_99 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_100 = {STRING_HEADER(3), 3, " = " };
pxll_string constructed_101 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_102 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_103 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_104 = {STRING_HEADER(8), 8, "[1] = &&" };
pxll_string constructed_105 = {STRING_HEADER(3), 3, "; r" };
pxll_string constructed_106 = {STRING_HEADER(11), 11, "[2] = lenv;" };
pxll_string constructed_107 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_108 = {STRING_HEADER(28), 28, " = allocate (TC_CLOSURE, 2);" };
pxll_string constructed_109 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_110 = {STRING_HEADER(15), 15, "check_heap (0);" };
pxll_string constructed_111 = {STRING_HEADER(49), 49, "stack_depth_indent(k); fprintf (stderr, \">> [%d] " };
pxll_string constructed_112 = {STRING_HEADER(15), 15, "\\n\", __LINE__);" };
pxll_string constructed_113 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_114 = {STRING_HEADER(5), 5, "goto " };
pxll_string constructed_115 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_116 = {STRING_HEADER(7), 7, "// def " };
pxll_string constructed_117 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_118 = {STRING_HEADER(3), 3, " = " };
pxll_string constructed_119 = {STRING_HEADER(4), 4, "top[" };
pxll_string constructed_120 = {STRING_HEADER(2), 2, "];" };
pxll_string constructed_121 = {STRING_HEADER(9), 9, "((object*" };
pxll_string constructed_122 = {STRING_HEADER(8), 8, ") lenv) " };
pxll_string constructed_123 = {STRING_HEADER(1), 1, "[" };
pxll_string constructed_124 = {STRING_HEADER(2), 2, "];" };
pxll_string constructed_125 = {STRING_HEADER(3), 3, "[1]" };
pxll_string constructed_126 = {STRING_HEADER(1), 1, "*" };
pxll_string constructed_127 = {STRING_HEADER(4), 4, "top[" };
pxll_string constructed_128 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_129 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_130 = {STRING_HEADER(9), 9, "((object*" };
pxll_string constructed_131 = {STRING_HEADER(8), 8, ") lenv) " };
pxll_string constructed_132 = {STRING_HEADER(1), 1, "[" };
pxll_string constructed_133 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_134 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_135 = {STRING_HEADER(3), 3, "[1]" };
pxll_string constructed_136 = {STRING_HEADER(1), 1, "*" };
pxll_string constructed_137 = {STRING_HEADER(7), 7, "top = r" };
pxll_string constructed_138 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_139 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_140 = {STRING_HEADER(23), 23, " = allocate (TC_TUPLE, " };
pxll_string constructed_141 = {STRING_HEADER(2), 2, ");" };
pxll_string constructed_142 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_143 = {STRING_HEADER(12), 12, " = (object*)" };
pxll_string constructed_144 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_145 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_146 = {STRING_HEADER(13), 13, " = allocate (" };
pxll_string constructed_147 = {STRING_HEADER(2), 2, ", " };
pxll_string constructed_148 = {STRING_HEADER(2), 2, ");" };
pxll_string constructed_149 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_150 = {STRING_HEADER(6), 6, "UITAG(" };
pxll_string constructed_151 = {STRING_HEADER(6), 6, "UOTAG(" };
pxll_string constructed_152 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_153 = {STRING_HEADER(1), 1, "[" };
pxll_string constructed_154 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_155 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_156 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_157 = {STRING_HEADER(10), 10, " = result;" };
pxll_string constructed_158 = {STRING_HEADER(24), 24, "; lenv = k[2]; k = k[1];" };
pxll_string constructed_159 = {STRING_HEADER(2), 2, "; " };
pxll_string constructed_160 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_161 = {STRING_HEADER(5), 5, " = k[" };
pxll_string constructed_162 = {STRING_HEADER(1), 1, "]" };
pxll_string constructed_163 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_164 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_165 = {STRING_HEADER(7), 7, "[1] = r" };
pxll_string constructed_166 = {STRING_HEADER(13), 13, "[2]; lenv = r" };
pxll_string constructed_167 = {STRING_HEADER(2), 2, "; " };
pxll_string constructed_168 = {STRING_HEADER(8), 8, "lenv = r" };
pxll_string constructed_169 = {STRING_HEADER(5), 5, "[2]; " };
pxll_string constructed_170 = {STRING_HEADER(7), 7, "goto *r" };
pxll_string constructed_171 = {STRING_HEADER(4), 4, "[1];" };
pxll_string constructed_172 = {STRING_HEADER(5), 5, "goto " };
pxll_string constructed_173 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_174 = {STRING_HEADER(32), 32, "t[1] = k; t[2] = lenv; t[3] = &&" };
pxll_string constructed_175 = {STRING_HEADER(2), 2, "; " };
pxll_string constructed_176 = {STRING_HEADER(8), 8, "; k = t;" };
pxll_string constructed_177 = {STRING_HEADER(2), 2, "; " };
pxll_string constructed_178 = {STRING_HEADER(2), 2, "t[" };
pxll_string constructed_179 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_180 = {STRING_HEADER(23), 23, "t = allocate (TC_SAVE, " };
pxll_string constructed_181 = {STRING_HEADER(2), 2, ");" };
pxll_string constructed_182 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_183 = {STRING_HEADER(7), 7, "[1] = r" };
pxll_string constructed_184 = {STRING_HEADER(13), 13, "[2]; lenv = r" };
pxll_string constructed_185 = {STRING_HEADER(2), 2, "; " };
pxll_string constructed_186 = {STRING_HEADER(8), 8, "lenv = r" };
pxll_string constructed_187 = {STRING_HEADER(5), 5, "[2]; " };
pxll_string constructed_188 = {STRING_HEADER(7), 7, "goto *r" };
pxll_string constructed_189 = {STRING_HEADER(4), 4, "[1];" };
pxll_string constructed_190 = {STRING_HEADER(5), 5, "goto " };
pxll_string constructed_191 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_192 = {STRING_HEADER(5), 5, "goto " };
pxll_string constructed_193 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_194 = {STRING_HEADER(5), 5, "lenv[" };
pxll_string constructed_195 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_196 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_197 = {STRING_HEADER(16), 16, "lenv = ((object " };
pxll_string constructed_198 = {STRING_HEADER(6), 6, ")lenv)" };
pxll_string constructed_199 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_200 = {STRING_HEADER(3), 3, "[1]" };
pxll_string constructed_201 = {STRING_HEADER(1), 1, "*" };
pxll_string constructed_202 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_203 = {STRING_HEADER(20), 20, "[1] = lenv; lenv = r" };
pxll_string constructed_204 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_205 = {STRING_HEADER(15), 15, "lenv = lenv[1];" };
pxll_string constructed_206 = {STRING_HEADER(5), 5, "k = r" };
pxll_string constructed_207 = {STRING_HEADER(11), 11, "; // %putcc" };
pxll_string constructed_208 = {STRING_HEADER(6), 6, "primop" };
pxll_string constructed_209 = {STRING_HEADER(6), 6, "%putcc" };
// symbol %putcc
pxll_int constructed_210[] = {UPTR(210,1), SYMBOL_HEADER, UPTR0(209), INTCON(11)};
pxll_string constructed_211 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_212 = {STRING_HEADER(15), 15, " = k; // %getcc" };
pxll_string constructed_213 = {STRING_HEADER(6), 6, "%getcc" };
// symbol %getcc
pxll_int constructed_214[] = {UPTR(214,1), SYMBOL_HEADER, UPTR0(213), INTCON(12)};
pxll_string constructed_215 = {STRING_HEADER(3), 3, " = " };
pxll_string constructed_216 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_217 = {STRING_HEADER(3), 3, "(((" };
pxll_string constructed_218 = {STRING_HEADER(15), 15, "*)((pxll_int*)r" };
pxll_string constructed_219 = {STRING_HEADER(5), 5, ")+1)[" };
pxll_string constructed_220 = {STRING_HEADER(2), 2, "])" };
pxll_string constructed_221 = {STRING_HEADER(2), 2, "((" };
pxll_string constructed_222 = {STRING_HEADER(7), 7, ")unbox(" };
pxll_string constructed_223 = {STRING_HEADER(2), 2, "))" };
pxll_string constructed_224 = {STRING_HEADER(37), 37, "lookup-cast: can't cast from int to: " };
pxll_string constructed_225 = {STRING_HEADER(3), 3, "int" };
// symbol int
pxll_int constructed_226[] = {UPTR(226,1), SYMBOL_HEADER, UPTR0(225), INTCON(13)};
pxll_string constructed_227 = {STRING_HEADER(43), 43, "lookup-cast: unable to cast between types: " };
pxll_string constructed_228 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_229 = {STRING_HEADER(5), 5, "%cset" };
// symbol %cset
pxll_int constructed_230[] = {UPTR(230,1), SYMBOL_HEADER, UPTR0(229), INTCON(14)};
pxll_string constructed_231 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_232 = {STRING_HEADER(3), 3, " = " };
pxll_string constructed_233 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_234 = {STRING_HEADER(3), 3, "(((" };
pxll_string constructed_235 = {STRING_HEADER(15), 15, "*)((pxll_int*)r" };
pxll_string constructed_236 = {STRING_HEADER(5), 5, ")+1)[" };
pxll_string constructed_237 = {STRING_HEADER(2), 2, "])" };
pxll_string constructed_238 = {STRING_HEADER(5), 5, "%cget" };
// symbol %cget
pxll_int constructed_239[] = {UPTR(239,1), SYMBOL_HEADER, UPTR0(238), INTCON(15)};
pxll_string constructed_240 = {STRING_HEADER(24), 24, "PXLL_UNDEFINED; result=r" };
pxll_string constructed_241 = {STRING_HEADER(15), 15, "; goto Lreturn;" };
pxll_string constructed_242 = {STRING_HEADER(5), 5, "%exit" };
// symbol %exit
pxll_int constructed_243[] = {UPTR(243,1), SYMBOL_HEADER, UPTR0(242), INTCON(16)};
pxll_string constructed_244 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_245 = {STRING_HEADER(48), 48, " = alloc_no_clear (TC_BUFFER, HOW_MANY (sizeof (" };
pxll_string constructed_246 = {STRING_HEADER(11), 11, ") * unbox(r" };
pxll_string constructed_247 = {STRING_HEADER(21), 21, "), sizeof (object)));" };
pxll_string constructed_248 = {STRING_HEADER(24), 24, "%callocate: dead target?" };
pxll_string constructed_249 = {STRING_HEADER(10), 10, "%callocate" };
// symbol %callocate
pxll_int constructed_250[] = {UPTR(250,1), SYMBOL_HEADER, UPTR0(249), INTCON(17)};
pxll_string constructed_251 = {STRING_HEADER(13), 13, "ensure_heap (" };
pxll_string constructed_252 = {STRING_HEADER(9), 9, ", unbox(r" };
pxll_string constructed_253 = {STRING_HEADER(3), 3, "));" };
pxll_string constructed_254 = {STRING_HEADER(12), 12, "%ensure-heap" };
// symbol %ensure-heap
pxll_int constructed_255[] = {UPTR(255,1), SYMBOL_HEADER, UPTR0(254), INTCON(18)};
pxll_string constructed_256 = {STRING_HEADER(16), 16, "((pxll_vector*)r" };
pxll_string constructed_257 = {STRING_HEADER(7), 7, ")->val[" };
pxll_string constructed_258 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_259 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_260 = {STRING_HEADER(16), 16, "((pxll_vector*)r" };
pxll_string constructed_261 = {STRING_HEADER(36), 36, ")->val[lookup_field((GET_TYPECODE(*r" };
pxll_string constructed_262 = {STRING_HEADER(17), 17, ")-TC_USEROBJ)>>2," };
pxll_string constructed_263 = {STRING_HEADER(6), 6, ")] = r" };
pxll_string constructed_264 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_265 = {STRING_HEADER(11), 11, "%record-set" };
// symbol %record-set
pxll_int constructed_266[] = {UPTR(266,1), SYMBOL_HEADER, UPTR0(265), INTCON(19)};
pxll_string constructed_267 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_268 = {STRING_HEADER(19), 19, " = ((pxll_vector*)r" };
pxll_string constructed_269 = {STRING_HEADER(7), 7, ")->val[" };
pxll_string constructed_270 = {STRING_HEADER(2), 2, "];" };
pxll_string constructed_271 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_272 = {STRING_HEADER(19), 19, " = ((pxll_vector*)r" };
pxll_string constructed_273 = {STRING_HEADER(36), 36, ")->val[lookup_field((GET_TYPECODE(*r" };
pxll_string constructed_274 = {STRING_HEADER(17), 17, ")-TC_USEROBJ)>>2," };
pxll_string constructed_275 = {STRING_HEADER(3), 3, ")];" };
pxll_string constructed_276 = {STRING_HEADER(11), 11, "%record-get" };
// symbol %record-get
pxll_int constructed_277[] = {UPTR(277,1), SYMBOL_HEADER, UPTR0(276), INTCON(20)};
pxll_string constructed_278 = {STRING_HEADER(16), 16, "((pxll_vector*)r" };
pxll_string constructed_279 = {STRING_HEADER(14), 14, ")->val[unbox(r" };
pxll_string constructed_280 = {STRING_HEADER(6), 6, ")] = r" };
pxll_string constructed_281 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_282 = {STRING_HEADER(41), 41, "range_check (GET_TUPLE_LENGTH(*(object*)r" };
pxll_string constructed_283 = {STRING_HEADER(10), 10, "), unbox(r" };
pxll_string constructed_284 = {STRING_HEADER(3), 3, "));" };
pxll_string constructed_285 = {STRING_HEADER(10), 10, "%array-set" };
// symbol %array-set
pxll_int constructed_286[] = {UPTR(286,1), SYMBOL_HEADER, UPTR0(285), INTCON(21)};
pxll_string constructed_287 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_288 = {STRING_HEADER(19), 19, " = ((pxll_vector*)r" };
pxll_string constructed_289 = {STRING_HEADER(14), 14, ")->val[unbox(r" };
pxll_string constructed_290 = {STRING_HEADER(3), 3, ")];" };
pxll_string constructed_291 = {STRING_HEADER(41), 41, "range_check (GET_TUPLE_LENGTH(*(object*)r" };
pxll_string constructed_292 = {STRING_HEADER(10), 10, "), unbox(r" };
pxll_string constructed_293 = {STRING_HEADER(3), 3, "));" };
pxll_string constructed_294 = {STRING_HEADER(10), 10, "%array-ref" };
// symbol %array-ref
pxll_int constructed_295[] = {UPTR(295,1), SYMBOL_HEADER, UPTR0(294), INTCON(22)};
pxll_string constructed_296 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_297 = {STRING_HEADER(3), 3, "  r" };
pxll_string constructed_298 = {STRING_HEADER(5), 5, " = t;" };
pxll_string constructed_299 = {STRING_HEADER(21), 21, "  for (i=0; i<unbox(r" };
pxll_string constructed_300 = {STRING_HEADER(20), 20, "); i++) { t[i+1] = r" };
pxll_string constructed_301 = {STRING_HEADER(3), 3, "; }" };
pxll_string constructed_302 = {STRING_HEADER(40), 40, "  t = alloc_no_clear (TC_VECTOR, unbox(r" };
pxll_string constructed_303 = {STRING_HEADER(3), 3, "));" };
pxll_string constructed_304 = {STRING_HEADER(11), 11, "if (unbox(r" };
pxll_string constructed_305 = {STRING_HEADER(11), 11, ") == 0) { r" };
pxll_string constructed_306 = {STRING_HEADER(39), 39, " = (object *) TC_EMPTY_VECTOR; } else {" };
pxll_string constructed_307 = {STRING_HEADER(12), 12, "%make-vector" };
// symbol %make-vector
pxll_int constructed_308[] = {UPTR(308,1), SYMBOL_HEADER, UPTR0(307), INTCON(23)};
pxll_string constructed_309 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_310 = {STRING_HEADER(13), 13, " = UOBJ_GET(r" };
pxll_string constructed_311 = {STRING_HEADER(1), 1, "," };
pxll_string constructed_312 = {STRING_HEADER(2), 2, ");" };
pxll_string constructed_313 = {STRING_HEADER(6), 6, "%nvget" };
// symbol %nvget
pxll_int constructed_314[] = {UPTR(314,1), SYMBOL_HEADER, UPTR0(313), INTCON(24)};
pxll_string constructed_315 = {STRING_HEADER(29), 29, "emit-primop: no such datatype" };
pxll_string constructed_316 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_317 = {STRING_HEADER(12), 12, " = (object*)" };
pxll_string constructed_318 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_319 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_320 = {STRING_HEADER(5), 5, " = t;" };
pxll_string constructed_321 = {STRING_HEADER(2), 2, "t[" };
pxll_string constructed_322 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_323 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_324 = {STRING_HEADER(20), 20, "t = alloc_no_clear (" };
pxll_string constructed_325 = {STRING_HEADER(1), 1, "," };
pxll_string constructed_326 = {STRING_HEADER(2), 2, ");" };
pxll_string constructed_327 = {STRING_HEADER(6), 6, "%dtcon" };
// symbol %dtcon
pxll_int constructed_328[] = {UPTR(328,1), SYMBOL_HEADER, UPTR0(327), INTCON(25)};
pxll_string constructed_329 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_330 = {STRING_HEADER(4), 4, " = r" };
pxll_string constructed_331 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_332 = {STRING_HEADER(1), 1, "r" };
pxll_string constructed_333 = {STRING_HEADER(4), 4, " = r" };
pxll_string constructed_334 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_335 = {STRING_HEADER(7), 7, "fatbar_" };
pxll_string constructed_336 = {STRING_HEADER(8), 8, "_over: ;" };
pxll_string constructed_337 = {STRING_HEADER(7), 7, "fatbar_" };
pxll_string constructed_338 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_339 = {STRING_HEADER(12), 12, "goto fatbar_" };
pxll_string constructed_340 = {STRING_HEADER(6), 6, "_over;" };
pxll_string constructed_341 = {STRING_HEADER(12), 12, "goto fatbar_" };
pxll_string constructed_342 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_343 = {STRING_HEADER(16), 16, "lenv = ((object " };
pxll_string constructed_344 = {STRING_HEADER(6), 6, ")lenv)" };
pxll_string constructed_345 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_346 = {STRING_HEADER(3), 3, "[1]" };
pxll_string constructed_347 = {STRING_HEADER(1), 1, "*" };
pxll_string constructed_348 = {STRING_HEADER(11), 11, "emit-nvcase" };
pxll_string constructed_349 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_350 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_351 = {STRING_HEADER(10), 10, "default: {" };
pxll_string constructed_352 = {STRING_HEADER(8), 8, "} break;" };
pxll_string constructed_353 = {STRING_HEADER(10), 10, "default: {" };
pxll_string constructed_354 = {STRING_HEADER(6), 6, "case (" };
pxll_string constructed_355 = {STRING_HEADER(4), 4, "): {" };
pxll_string constructed_356 = {STRING_HEADER(8), 8, "switch (" };
pxll_string constructed_357 = {STRING_HEADER(3), 3, " (r" };
pxll_string constructed_358 = {STRING_HEADER(4), 4, ")) {" };
pxll_string constructed_359 = {STRING_HEADER(8), 8, "get_case" };
pxll_string constructed_360 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_361 = {STRING_HEADER(8), 8, "} break;" };
pxll_string constructed_362 = {STRING_HEADER(10), 10, "default: {" };
pxll_string constructed_363 = {STRING_HEADER(6), 6, "case (" };
pxll_string constructed_364 = {STRING_HEADER(4), 4, "): {" };
pxll_string constructed_365 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_366 = {STRING_HEADER(6), 6, "UITAG(" };
pxll_string constructed_367 = {STRING_HEADER(6), 6, "UOTAG(" };
pxll_string constructed_368 = {STRING_HEADER(32), 32, "variant constructor never called" };
pxll_string constructed_369 = {STRING_HEADER(25), 25, "switch (get_case_noint (r" };
pxll_string constructed_370 = {STRING_HEADER(4), 4, ")) {" };
pxll_string constructed_371 = {STRING_HEADER(22), 22, "pvcase else-clause NYI" };
pxll_string constructed_372 = {STRING_HEADER(2), 2, "}}" };
pxll_string constructed_373 = {STRING_HEADER(7), 7, "  case " };
pxll_string constructed_374 = {STRING_HEADER(3), 3, ": r" };
pxll_string constructed_375 = {STRING_HEADER(9), 9, " = heap0[" };
pxll_string constructed_376 = {STRING_HEADER(2), 2, "];" };
pxll_string constructed_377 = {STRING_HEADER(14), 14, "  switch (n) {" };
pxll_string constructed_378 = {STRING_HEADER(26), 26, "void gc_regs_out (int n) {" };
pxll_string constructed_379 = {STRING_HEADER(2), 2, "}}" };
pxll_string constructed_380 = {STRING_HEADER(7), 7, "  case " };
pxll_string constructed_381 = {STRING_HEADER(8), 8, ": heap1[" };
pxll_string constructed_382 = {STRING_HEADER(5), 5, "] = r" };
pxll_string constructed_383 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_384 = {STRING_HEADER(14), 14, "  switch (n) {" };
pxll_string constructed_385 = {STRING_HEADER(25), 25, "void gc_regs_in (int n) {" };
pxll_string constructed_386 = {STRING_HEADER(19), 19, "register object * r" };
pxll_string constructed_387 = {STRING_HEADER(1), 1, ";" };
pxll_string constructed_388 = {STRING_HEADER(37), 37, "pxll_int pxll_internal_symbols[] = {(" };
pxll_string constructed_389 = {STRING_HEADER(16), 16, "<<8)|TC_VECTOR, " };
pxll_string constructed_390 = {STRING_HEADER(2), 2, "};" };
pxll_string constructed_391 = {STRING_HEADER(2), 2, ", " };
pxll_string constructed_392 = {STRING_HEADER(5), 5, "UPTR(" };
pxll_string constructed_393 = {STRING_HEADER(3), 3, ",1)" };
pxll_int constructed_394[] = {TC_NIL};
pxll_string constructed_395 = {STRING_HEADER(24), 24, "pxll_string constructed_" };
pxll_string constructed_396 = {STRING_HEADER(18), 18, " = {STRING_HEADER(" };
pxll_string constructed_397 = {STRING_HEADER(3), 3, "), " };
pxll_string constructed_398 = {STRING_HEADER(3), 3, ", \"" };
pxll_string constructed_399 = {STRING_HEADER(4), 4, "\" };" };
pxll_int constructed_400[] = {TC_NIL};
pxll_string constructed_401 = {STRING_HEADER(2), 2, "\\\"" };
pxll_string constructed_402 = {STRING_HEADER(2), 2, "\\\\" };
pxll_string constructed_403 = {STRING_HEADER(2), 2, "\\t" };
pxll_string constructed_404 = {STRING_HEADER(2), 2, "\\n" };
pxll_string constructed_405 = {STRING_HEADER(2), 2, "\\r" };
pxll_string constructed_406 = {STRING_HEADER(1), 1, "0" };
pxll_string constructed_407 = {STRING_HEADER(2), 2, "00" };
pxll_string constructed_408 = {STRING_HEADER(3), 3, "000" };
pxll_string constructed_409 = {STRING_HEADER(30), 30, "unable to oct-encode character" };
pxll_int constructed_410[] = {TC_NIL};
pxll_string constructed_411 = {STRING_HEADER(21), 21, "pxll_int constructed_" };
pxll_string constructed_412 = {STRING_HEADER(11), 11, "[] = {UPTR(" };
pxll_string constructed_413 = {STRING_HEADER(26), 26, ",1), SYMBOL_HEADER, UPTR0(" };
pxll_string constructed_414 = {STRING_HEADER(10), 10, "), INTCON(" };
pxll_string constructed_415 = {STRING_HEADER(3), 3, ")};" };
pxll_string constructed_416 = {STRING_HEADER(10), 10, "// symbol " };
pxll_string constructed_417 = {STRING_HEADER(21), 21, "pxll_int constructed_" };
pxll_string constructed_418 = {STRING_HEADER(6), 6, "[] = {" };
pxll_string constructed_419 = {STRING_HEADER(2), 2, "};" };
pxll_string constructed_420 = {STRING_HEADER(1), 1, "," };
pxll_int constructed_421[] = {TC_NIL};
pxll_string constructed_422 = {STRING_HEADER(5), 5, "UPTR(" };
pxll_string constructed_423 = {STRING_HEADER(1), 1, "," };
pxll_string constructed_424 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_425 = {STRING_HEADER(16), 16, "no such datatype" };
pxll_string constructed_426 = {STRING_HEADER(5), 5, "UPTR(" };
pxll_string constructed_427 = {STRING_HEADER(1), 1, "," };
pxll_string constructed_428 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_429 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_430 = {STRING_HEADER(14), 14, "<<8)|TC_VECTOR" };
pxll_string constructed_431 = {STRING_HEADER(5), 5, "UPTR(" };
pxll_string constructed_432 = {STRING_HEADER(3), 3, ",1)" };
pxll_string constructed_433 = {STRING_HEADER(15), 15, "unknown symbol?" };
pxll_string constructed_434 = {STRING_HEADER(6), 6, "UPTR0(" };
pxll_string constructed_435 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_436 = {STRING_HEADER(29), 29, "emit-constructed: lost string" };
pxll_string constructed_437 = {STRING_HEADER(6), 6, "TC_NIL" };
pxll_string constructed_438 = {STRING_HEADER(6), 6, "UITAG(" };
pxll_string constructed_439 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_440 = {STRING_HEADER(11), 11, "CONS_HEADER" };
pxll_string constructed_441 = {STRING_HEADER(7), 7, "UOHEAD(" };
pxll_string constructed_442 = {STRING_HEADER(1), 1, "," };
pxll_string constructed_443 = {STRING_HEADER(1), 1, ")" };
pxll_int constructed_444[] = {TC_NIL};
pxll_string constructed_445 = {STRING_HEADER(39), 39, "template strings not found in header.c?" };
pxll_int constructed_446[] = {TC_NIL};
pxll_string constructed_447 = {STRING_HEADER(8), 8, "header.c" };
pxll_string constructed_448 = {STRING_HEADER(10), 10, "#include <" };
pxll_string constructed_449 = {STRING_HEADER(1), 1, ">" };
pxll_string constructed_450 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_451 = {STRING_HEADER(16), 16, "\n-- C output --\n" };
pxll_string constructed_452 = {STRING_HEADER(2), 2, "  " };
pxll_string constructed_453 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_454 = {STRING_HEADER(1), 1, "\n" };
pxll_string constructed_455 = {STRING_HEADER(22), 22, "\n-- variant labels --\n" };
pxll_string constructed_456 = {STRING_HEADER(2), 2, "  " };
pxll_string constructed_457 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_458 = {STRING_HEADER(1), 1, "\n" };
pxll_string constructed_459 = {STRING_HEADER(15), 15, "\n-- symbols --\n" };
pxll_string constructed_460 = {STRING_HEADER(15), 15, "\n-- records --\n" };
pxll_string constructed_461 = {STRING_HEADER(14), 14, "\n-- labels --\n" };
pxll_string constructed_462 = {STRING_HEADER(17), 17, "\n-- datatypes --\n" };
pxll_string constructed_463 = {STRING_HEADER(11), 11, "\n-- RTL --\n" };
pxll_string constructed_464 = {STRING_HEADER(1), 1, "\n" };
pxll_string constructed_465 = {STRING_HEADER(2), 2, "  " };
pxll_int constructed_466[] = {TC_NIL};
pxll_int constructed_467[] = {TC_NIL};
pxll_int constructed_468[] = {TC_NIL};
pxll_string constructed_469 = {STRING_HEADER(16), 16, "unbound variable" };
pxll_string constructed_470 = {STRING_HEADER(15), 15, "empty sequence?" };
pxll_string constructed_471 = {STRING_HEADER(6), 6, "vector" };
// symbol vector
pxll_int constructed_472[] = {UPTR(472,1), SYMBOL_HEADER, UPTR0(471), INTCON(26)};
pxll_string constructed_473 = {STRING_HEADER(11), 11, "constructor" };
// symbol constructor
pxll_int constructed_474[] = {UPTR(474,1), SYMBOL_HEADER, UPTR0(473), INTCON(27)};
pxll_string constructed_475 = {STRING_HEADER(13), 13, "c-conditional" };
pxll_string constructed_476 = {STRING_HEADER(3), 3, "pop" };
pxll_int constructed_477[] = {TC_NIL};
pxll_string constructed_478 = {STRING_HEADER(28), 28, "c-call function in register?" };
pxll_string constructed_479 = {STRING_HEADER(20), 20, "c-call: no function?" };
pxll_string constructed_480 = {STRING_HEADER(6), 6, "buffer" };
// symbol buffer
pxll_int constructed_481[] = {UPTR(481,1), SYMBOL_HEADER, UPTR0(480), INTCON(28)};
pxll_string constructed_482 = {STRING_HEADER(5), 5, "%rset" };
// symbol %rset
pxll_int constructed_483[] = {UPTR(483,1), SYMBOL_HEADER, UPTR0(482), INTCON(29)};
pxll_string constructed_484 = {STRING_HEADER(8), 8, "%raccess" };
// symbol %raccess
pxll_int constructed_485[] = {UPTR(485,1), SYMBOL_HEADER, UPTR0(484), INTCON(30)};
pxll_int constructed_486[] = {TC_NIL};
pxll_string constructed_487 = {STRING_HEADER(6), 6, "%rmake" };
// symbol %rmake
pxll_int constructed_488[] = {UPTR(488,1), SYMBOL_HEADER, UPTR0(487), INTCON(31)};
pxll_string constructed_489 = {STRING_HEADER(18), 18, "malformed %rextend" };
pxll_string constructed_490 = {STRING_HEADER(8), 8, "%rextend" };
// symbol %rextend
pxll_int constructed_491[] = {UPTR(491,1), SYMBOL_HEADER, UPTR0(490), INTCON(32)};
pxll_string constructed_492 = {STRING_HEADER(23), 23, "c-record-extension: NYI" };
pxll_string constructed_493 = {STRING_HEADER(22), 22, "unknown variant label?" };
pxll_string constructed_494 = {STRING_HEADER(16), 16, "bad %vcon params" };
pxll_string constructed_495 = {STRING_HEADER(5), 5, "%vcon" };
// symbol %vcon
pxll_int constructed_496[] = {UPTR(496,1), SYMBOL_HEADER, UPTR0(495), INTCON(33)};
pxll_string constructed_497 = {STRING_HEADER(7), 7, "%fatbar" };
// symbol %fatbar
pxll_int constructed_498[] = {UPTR(498,1), SYMBOL_HEADER, UPTR0(497), INTCON(34)};
pxll_string constructed_499 = {STRING_HEADER(21), 21, "%fail without fatbar?" };
pxll_string constructed_500 = {STRING_HEADER(5), 5, "%fail" };
// symbol %fail
pxll_int constructed_501[] = {UPTR(501,1), SYMBOL_HEADER, UPTR0(500), INTCON(35)};
pxll_string constructed_502 = {STRING_HEADER(12), 12, "%match-error" };
// symbol %match-error
pxll_int constructed_503[] = {UPTR(503,1), SYMBOL_HEADER, UPTR0(502), INTCON(36)};
pxll_string constructed_504 = {STRING_HEADER(16), 16, "no such datatype" };
pxll_string constructed_505 = {STRING_HEADER(16), 16, "incomplete match" };
pxll_int constructed_506[] = {TC_NIL};
pxll_string constructed_507 = {STRING_HEADER(13), 13, "add-free-regs" };
pxll_string constructed_508 = {STRING_HEADER(3), 3, "top" };
// symbol top
pxll_int constructed_509[] = {UPTR(509,1), SYMBOL_HEADER, UPTR0(508), INTCON(37)};
pxll_int constructed_510[] = {UPTR(510,1),CONS_HEADER,UPTR(509,1),TC_NIL};
pxll_string constructed_511 = {STRING_HEADER(7), 7, "cps...\n" };
pxll_string constructed_512 = {STRING_HEADER(20), 20, "\n-- after typing --\n" };
pxll_int constructed_513[] = {TC_NIL};
pxll_string constructed_514 = {STRING_HEADER(28), 28, "apply-tenv: unbound variable" };
pxll_string constructed_515 = {STRING_HEADER(3), 3, "raw" };
// symbol raw
pxll_int constructed_516[] = {UPTR(516,1), SYMBOL_HEADER, UPTR0(515), INTCON(38)};
pxll_string constructed_517 = {STRING_HEADER(26), 26, "strange constructor scheme" };
pxll_string constructed_518 = {STRING_HEADER(16), 16, "no such datatype" };
pxll_string constructed_519 = {STRING_HEADER(28), 28, "wrong number of args to cexp" };
pxll_string constructed_520 = {STRING_HEADER(20), 20, "malformed arrow type" };
pxll_string constructed_521 = {STRING_HEADER(21), 21, "malformed conditional" };
pxll_string constructed_522 = {STRING_HEADER(15), 15, "empty sequence?" };
pxll_int constructed_523[] = {TC_NIL};
pxll_int constructed_524[] = {TC_NIL};
pxll_string constructed_525 = {STRING_HEADER(11), 11, "empty call?" };
pxll_int constructed_526[] = {TC_NIL};
pxll_int constructed_527[] = {TC_NIL};
pxll_int constructed_528[] = {TC_NIL};
pxll_int constructed_529[] = {TC_NIL};
pxll_int constructed_530[] = {TC_NIL};
pxll_int constructed_531[] = {TC_NIL};
pxll_int constructed_532[] = {TC_NIL};
pxll_int constructed_533[] = {TC_NIL};
pxll_int constructed_534[] = {TC_NIL};
pxll_string constructed_535 = {STRING_HEADER(31), 31, "wrong number of args to primapp" };
pxll_string constructed_536 = {STRING_HEADER(15), 15, "type-of-primapp" };
pxll_string constructed_537 = {STRING_HEADER(12), 12, "continuation" };
// symbol continuation
pxll_int constructed_538[] = {UPTR(538,1), SYMBOL_HEADER, UPTR0(537), INTCON(39)};
pxll_int constructed_539[] = {TC_NIL};
pxll_string constructed_540 = {STRING_HEADER(4), 4, "rsum" };
// symbol rsum
pxll_int constructed_541[] = {UPTR(541,1), SYMBOL_HEADER, UPTR0(540), INTCON(40)};
pxll_string constructed_542 = {STRING_HEADER(3), 3, "pre" };
// symbol pre
pxll_int constructed_543[] = {UPTR(543,1), SYMBOL_HEADER, UPTR0(542), INTCON(41)};
pxll_string constructed_544 = {STRING_HEADER(7), 7, "product" };
// symbol product
pxll_int constructed_545[] = {UPTR(545,1), SYMBOL_HEADER, UPTR0(544), INTCON(42)};
pxll_int constructed_546[] = {TC_NIL};
pxll_string constructed_547 = {STRING_HEADER(32), 32, "lookup-primapp: no such datatype" };
pxll_string constructed_548 = {STRING_HEADER(24), 24, "bad parameters to primop" };
pxll_int constructed_549[] = {TC_NIL};
pxll_int constructed_550[] = {TC_NIL};
pxll_string constructed_551 = {STRING_HEADER(32), 32, "lookup-primapp: no such datatype" };
pxll_string constructed_552 = {STRING_HEADER(8), 8, "rproduct" };
// symbol rproduct
pxll_int constructed_553[] = {UPTR(553,1), SYMBOL_HEADER, UPTR0(552), INTCON(43)};
pxll_int constructed_554[] = {TC_NIL};
pxll_int constructed_555[] = {TC_NIL};
pxll_string constructed_556 = {STRING_HEADER(8), 8, "rdefault" };
// symbol rdefault
pxll_int constructed_557[] = {UPTR(557,1), SYMBOL_HEADER, UPTR0(556), INTCON(44)};
pxll_int constructed_558[] = {TC_NIL};
pxll_string constructed_559 = {STRING_HEADER(3), 3, "abs" };
// symbol abs
pxll_int constructed_560[] = {UPTR(560,1), SYMBOL_HEADER, UPTR0(559), INTCON(45)};
pxll_int constructed_561[] = {TC_NIL};
pxll_int constructed_562[] = {TC_NIL};
pxll_int constructed_563[] = {TC_NIL};
pxll_string constructed_564 = {STRING_HEADER(14), 14, "lookup-primapp" };
pxll_string constructed_565 = {STRING_HEADER(16), 16, "malformed pvcase" };
pxll_string constructed_566 = {STRING_HEADER(16), 16, "no such datatype" };
pxll_string constructed_567 = {STRING_HEADER(10), 10, "typing NYI" };
pxll_int constructed_568[] = {TC_NIL};
pxll_string constructed_569 = {STRING_HEADER(23), 23, "instantiate-type-scheme" };
pxll_string constructed_570 = {STRING_HEADER(10), 10, "typing...\n" };
pxll_int constructed_571[] = {TC_NIL};
pxll_string constructed_572 = {STRING_HEADER(3), 3, "pop" };
pxll_int constructed_573[] = {TC_NIL};
pxll_int constructed_574[] = {TC_NIL};
pxll_int constructed_575[] = {TC_NIL};
pxll_int constructed_576[] = {TC_NIL};
pxll_int constructed_577[] = {TC_NIL};
pxll_int constructed_578[] = {TC_NIL};
pxll_string constructed_579 = {STRING_HEADER(31), 31, "strongly-connected components:\n" };
pxll_string constructed_580 = {STRING_HEADER(20), 20, "after second round:\n" };
pxll_int constructed_581[] = {TC_NIL};
pxll_int constructed_582[] = {TC_NIL};
pxll_string constructed_583 = {STRING_HEADER(7), 7, "lambda_" };
pxll_string constructed_584 = {STRING_HEADER(6), 6, "lambda" };
// symbol lambda
pxll_int constructed_585[] = {UPTR(585,1), SYMBOL_HEADER, UPTR0(584), INTCON(46)};
pxll_int constructed_586[] = {TC_NIL};
pxll_int constructed_587[] = {TC_NIL};
pxll_int constructed_588[] = {TC_NIL};
pxll_int constructed_589[] = {TC_NIL};
pxll_string constructed_590 = {STRING_HEADER(2), 2, ".c" };
pxll_string constructed_591 = {STRING_HEADER(9), 9, "find-base" };
pxll_string constructed_592 = {STRING_HEADER(1), 1, "." };
pxll_string constructed_593 = {STRING_HEADER(3), 3, "scm" };
pxll_string constructed_594 = {STRING_HEADER(2), 2, "if" };
// symbol if
pxll_int constructed_595[] = {UPTR(595,1), SYMBOL_HEADER, UPTR0(594), INTCON(47)};
pxll_string constructed_596 = {STRING_HEADER(4), 4, "set!" };
// symbol set!
pxll_int constructed_597[] = {UPTR(597,1), SYMBOL_HEADER, UPTR0(596), INTCON(48)};
pxll_string constructed_598 = {STRING_HEADER(5), 5, "begin" };
// symbol begin
pxll_int constructed_599[] = {UPTR(599,1), SYMBOL_HEADER, UPTR0(598), INTCON(49)};
pxll_string constructed_600 = {STRING_HEADER(8), 8, "function" };
// symbol function
pxll_int constructed_601[] = {UPTR(601,1), SYMBOL_HEADER, UPTR0(600), INTCON(50)};
pxll_string constructed_602 = {STRING_HEADER(5), 5, "vcase" };
// symbol vcase
pxll_int constructed_603[] = {UPTR(603,1), SYMBOL_HEADER, UPTR0(602), INTCON(51)};
pxll_string constructed_604 = {STRING_HEADER(9), 9, "let-splat" };
// symbol let-splat
pxll_int constructed_605[] = {UPTR(605,1), SYMBOL_HEADER, UPTR0(604), INTCON(52)};
pxll_string constructed_606 = {STRING_HEADER(5), 5, "match" };
// symbol match
pxll_int constructed_607[] = {UPTR(607,1), SYMBOL_HEADER, UPTR0(606), INTCON(53)};
pxll_string constructed_608 = {STRING_HEADER(8), 8, "cinclude" };
// symbol cinclude
pxll_int constructed_609[] = {UPTR(609,1), SYMBOL_HEADER, UPTR0(608), INTCON(54)};
pxll_string constructed_610 = {STRING_HEADER(9), 9, "cverbatim" };
// symbol cverbatim
pxll_int constructed_611[] = {UPTR(611,1), SYMBOL_HEADER, UPTR0(610), INTCON(55)};
pxll_string constructed_612 = {STRING_HEADER(7), 7, "%nvcase" };
// symbol %nvcase
pxll_int constructed_613[] = {UPTR(613,1), SYMBOL_HEADER, UPTR0(612), INTCON(56)};
pxll_string constructed_614 = {STRING_HEADER(6), 6, "%%cexp" };
// symbol %%cexp
pxll_int constructed_615[] = {UPTR(615,1), SYMBOL_HEADER, UPTR0(614), INTCON(57)};
pxll_string constructed_616 = {STRING_HEADER(16), 16, "malformed %%cexp" };
pxll_string constructed_617 = {STRING_HEADER(21), 21, "malformed <cverbatim>" };
pxll_string constructed_618 = {STRING_HEADER(20), 20, "malformed <cinclude>" };
pxll_int constructed_619[] = {TC_NIL};
pxll_int constructed_620[] = {TC_NIL};
pxll_string constructed_621 = {STRING_HEADER(3), 3, "let" };
// symbol let
pxll_int constructed_622[] = {UPTR(622,1), SYMBOL_HEADER, UPTR0(621), INTCON(58)};
pxll_string constructed_623 = {STRING_HEADER(4), 4, "with" };
// symbol with
pxll_int constructed_624[] = {UPTR(624,1), SYMBOL_HEADER, UPTR0(623), INTCON(59)};
pxll_string constructed_625 = {STRING_HEADER(26), 26, "malformed match expression" };
pxll_string constructed_626 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_627 = {STRING_HEADER(3), 3, "arg" };
pxll_string constructed_628 = {STRING_HEADER(1), 1, ":" };
pxll_int constructed_629[] = {TC_NIL};
pxll_int constructed_630[] = {TC_NIL};
pxll_int constructed_631[] = {TC_NIL};
pxll_int constructed_632[] = {TC_NIL};
pxll_int constructed_633[] = {TC_NIL};
pxll_string constructed_634 = {STRING_HEADER(4), 4, "else" };
// symbol else
pxll_int constructed_635[] = {UPTR(635,1), SYMBOL_HEADER, UPTR0(634), INTCON(60)};
pxll_string constructed_636 = {STRING_HEADER(10), 10, "split-alts" };
pxll_int constructed_637[] = {TC_NIL};
pxll_string constructed_638 = {STRING_HEADER(1), 1, "_" };
// symbol _
pxll_int constructed_639[] = {UPTR(639,1), SYMBOL_HEADER, UPTR0(638), INTCON(61)};
pxll_int constructed_640[] = {TC_NIL};
pxll_int constructed_641[] = {TC_NIL};
pxll_string constructed_642 = {STRING_HEADER(12), 12, "expand-vcase" };
pxll_string constructed_643 = {STRING_HEADER(18), 18, "malformed FUNCTION" };
pxll_string constructed_644 = {STRING_HEADER(16), 16, "malformed LAMBDA" };
pxll_string constructed_645 = {STRING_HEADER(30), 30, "malformed binding in LET-SPLAT" };
pxll_string constructed_646 = {STRING_HEADER(19), 19, "malformed LET-SPLAT" };
pxll_string constructed_647 = {STRING_HEADER(11), 11, "empty BEGIN" };
pxll_string constructed_648 = {STRING_HEADER(14), 14, "malformed set!" };
pxll_string constructed_649 = {STRING_HEADER(14), 14, "malformed <if>" };
pxll_int constructed_650[] = {TC_NIL};
pxll_string constructed_651 = {STRING_HEADER(1), 1, "o" };
// symbol o
pxll_int constructed_652[] = {UPTR(652,1), SYMBOL_HEADER, UPTR0(651), INTCON(62)};
pxll_string constructed_653 = {STRING_HEADER(7), 7, "%method" };
// symbol %method
pxll_int constructed_654[] = {UPTR(654,1), SYMBOL_HEADER, UPTR0(653), INTCON(63)};
pxll_int constructed_655[] = {TC_NIL};
pxll_int constructed_656[] = {TC_NIL};
pxll_int constructed_657[] = {TC_NIL};
pxll_string constructed_658 = {STRING_HEADER(20), 20, "malformed definition" };
pxll_string constructed_659 = {STRING_HEADER(2), 2, "->" };
// symbol ->
pxll_int constructed_660[] = {UPTR(660,1), SYMBOL_HEADER, UPTR0(659), INTCON(64)};
pxll_string constructed_661 = {STRING_HEADER(18), 18, "malformed <define>" };
pxll_string constructed_662 = {STRING_HEADER(6), 6, "define" };
// symbol define
pxll_int constructed_663[] = {UPTR(663,1), SYMBOL_HEADER, UPTR0(662), INTCON(65)};
pxll_int constructed_664[] = {TC_NIL};
pxll_string constructed_665 = {STRING_HEADER(2), 2, ")\n" };
pxll_string constructed_666 = {STRING_HEADER(2), 2, "\n\t" };
pxll_string constructed_667 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_668 = {STRING_HEADER(3), 3, ">\n\t" };
pxll_string constructed_669 = {STRING_HEADER(8), 8, "(macro <" };
pxll_string constructed_670 = {STRING_HEADER(28), 28, "no matching clause for macro" };
pxll_int constructed_671[] = {TC_NIL};
pxll_string constructed_672 = {STRING_HEADER(27), 27, "malformed macro definition:" };
pxll_string constructed_673 = {STRING_HEADER(27), 27, "malformed macro definition:" };
pxll_string constructed_674 = {STRING_HEADER(8), 8, "defmacro" };
// symbol defmacro
pxll_int constructed_675[] = {UPTR(675,1), SYMBOL_HEADER, UPTR0(674), INTCON(66)};
pxll_string constructed_676 = {STRING_HEADER(25), 25, "malformed alt in datatype" };
pxll_string constructed_677 = {STRING_HEADER(23), 23, "no such alt in datatype" };
pxll_string constructed_678 = {STRING_HEADER(23), 23, "no such alt in datatype" };
pxll_string constructed_679 = {STRING_HEADER(18), 18, "malformed datatype" };
pxll_string constructed_680 = {STRING_HEADER(8), 8, "datatype" };
// symbol datatype
pxll_int constructed_681[] = {UPTR(681,1), SYMBOL_HEADER, UPTR0(680), INTCON(67)};
pxll_int constructed_682[] = {TC_NIL};
pxll_int constructed_683[] = {TC_NIL};
pxll_string constructed_684 = {STRING_HEADER(3), 3, "fix" };
// symbol fix
pxll_int constructed_685[] = {UPTR(685,1), SYMBOL_HEADER, UPTR0(684), INTCON(68)};
pxll_int constructed_686[] = {TC_NIL};
pxll_string constructed_687 = {STRING_HEADER(39), 39, "unexpected s-expression in transformer:" };
pxll_string constructed_688 = {STRING_HEADER(2), 2, "-O" };
pxll_string constructed_689 = {STRING_HEADER(2), 2, "-f" };
pxll_string constructed_690 = {STRING_HEADER(2), 2, "-t" };
pxll_string constructed_691 = {STRING_HEADER(2), 2, "-v" };
pxll_string constructed_692 = {STRING_HEADER(2), 2, "-c" };
pxll_string constructed_693 = {STRING_HEADER(16), 16, "self/selfmac.scm" };
pxll_int constructed_694[] = {TC_NIL};
pxll_int constructed_695[] = {TC_NIL};
pxll_string constructed_696 = {STRING_HEADER(0), 0, "" };
pxll_int constructed_697[] = {TC_NIL};
pxll_int constructed_698[] = {TC_NIL};
pxll_int constructed_699[] = {TC_NIL};
pxll_int constructed_700[] = {TC_NIL};
pxll_string constructed_701 = {STRING_HEADER(31), 31, "Usage: compile <irken-src-file>" };
pxll_string constructed_702 = {STRING_HEADER(3), 3, "-I." };
pxll_string constructed_703 = {STRING_HEADER(3), 3, "gcc" };
pxll_string constructed_704 = {STRING_HEADER(28), 28, "// REGISTER_DECLARATIONS //\n" };
pxll_string constructed_705 = {STRING_HEADER(27), 27, "// CONSTRUCTED LITERALS //\n" };
pxll_string constructed_706 = {STRING_HEADER(95), 95, "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~ " };
pxll_string constructed_707 = {STRING_HEADER(7), 7, "uint8_t" };
// symbol uint8_t
pxll_int constructed_708[] = {UPTR(708,1), SYMBOL_HEADER, UPTR0(707), INTCON(69)};
pxll_string constructed_709 = {STRING_HEADER(8), 8, "uint16_t" };
// symbol uint16_t
pxll_int constructed_710[] = {UPTR(710,1), SYMBOL_HEADER, UPTR0(709), INTCON(70)};
pxll_string constructed_711 = {STRING_HEADER(8), 8, "uint32_t" };
// symbol uint32_t
pxll_int constructed_712[] = {UPTR(712,1), SYMBOL_HEADER, UPTR0(711), INTCON(71)};
pxll_string constructed_713 = {STRING_HEADER(8), 8, "uint64_t" };
// symbol uint64_t
pxll_int constructed_714[] = {UPTR(714,1), SYMBOL_HEADER, UPTR0(713), INTCON(72)};
pxll_string constructed_715 = {STRING_HEADER(6), 6, "int8_t" };
// symbol int8_t
pxll_int constructed_716[] = {UPTR(716,1), SYMBOL_HEADER, UPTR0(715), INTCON(73)};
pxll_string constructed_717 = {STRING_HEADER(7), 7, "int16_t" };
// symbol int16_t
pxll_int constructed_718[] = {UPTR(718,1), SYMBOL_HEADER, UPTR0(717), INTCON(74)};
pxll_string constructed_719 = {STRING_HEADER(7), 7, "int32_t" };
// symbol int32_t
pxll_int constructed_720[] = {UPTR(720,1), SYMBOL_HEADER, UPTR0(719), INTCON(75)};
pxll_string constructed_721 = {STRING_HEADER(7), 7, "int64_t" };
// symbol int64_t
pxll_int constructed_722[] = {UPTR(722,1), SYMBOL_HEADER, UPTR0(721), INTCON(76)};
pxll_int constructed_723[] = {UPTR(723,22),CONS_HEADER,UPTR(722,1),TC_NIL,CONS_HEADER,UPTR(720,1),UPTR(723,1),CONS_HEADER,UPTR(718,1),UPTR(723,4),CONS_HEADER,UPTR(716,1),UPTR(723,7),CONS_HEADER,UPTR(714,1),UPTR(723,10),CONS_HEADER,UPTR(712,1),UPTR(723,13),CONS_HEADER,UPTR(710,1),UPTR(723,16),CONS_HEADER,UPTR(708,1),UPTR(723,19)};
pxll_string constructed_724 = {STRING_HEADER(1), 1, "L" };
pxll_int constructed_725[] = {TC_NIL};
pxll_string constructed_726 = {STRING_HEADER(26), 26, "key already present in map" };
pxll_int constructed_727[] = {TC_NIL};
pxll_int constructed_728[] = {TC_NIL};
pxll_int constructed_729[] = {TC_NIL};
pxll_int constructed_730[] = {TC_NIL};
pxll_string constructed_731 = {STRING_HEADER(1), 1, "?" };
// symbol ?
pxll_int constructed_732[] = {UPTR(732,1), SYMBOL_HEADER, UPTR0(731), INTCON(77)};
pxll_string constructed_733 = {STRING_HEADER(4), 4, "char" };
// symbol char
pxll_int constructed_734[] = {UPTR(734,1), SYMBOL_HEADER, UPTR0(733), INTCON(78)};
pxll_string constructed_735 = {STRING_HEADER(9), 9, "undefined" };
// symbol undefined
pxll_int constructed_736[] = {UPTR(736,1), SYMBOL_HEADER, UPTR0(735), INTCON(79)};
pxll_int constructed_737[] = {TC_NIL};
pxll_int constructed_738[] = {TC_NIL};
pxll_int constructed_739[] = {TC_NIL};
pxll_int constructed_740[] = {TC_NIL};
pxll_int constructed_741[] = {TC_NIL};
pxll_int constructed_742[] = {TC_NIL};
pxll_int constructed_743[] = {UPTR(743,10),CONS_HEADER,3330,TC_NIL,CONS_HEADER,2562,UPTR(743,1),CONS_HEADER,2306,UPTR(743,4),CONS_HEADER,8194,UPTR(743,7)};
pxll_string constructed_744 = {STRING_HEADER(10), 10, "0123456789" };
pxll_string constructed_745 = {STRING_HEADER(52), 52, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ" };
pxll_string constructed_746 = {STRING_HEADER(7), 7, "()[]{}:" };
pxll_int constructed_747[] = {UPTR(747,10),CONS_HEADER,3330,TC_NIL,CONS_HEADER,2562,UPTR(747,1),CONS_HEADER,2306,UPTR(747,4),CONS_HEADER,8194,UPTR(747,7)};
pxll_int constructed_748[] = {UPTR(748,29),UOHEAD(3,1),14082,15,UITAG(0),UOHEAD(3,1),13826,13,UPTR(748,1),UOHEAD(3,1),13570,11,UPTR(748,5),UOHEAD(3,1),13314,9,UPTR(748,9),UOHEAD(3,1),13058,7,UPTR(748,13),UOHEAD(3,1),12802,5,UPTR(748,17),UOHEAD(3,1),12546,3,UPTR(748,21),UOHEAD(3,1),12290,1,UPTR(748,25)};
pxll_int constructed_749[] = {UPTR(749,37),UOHEAD(3,1),14594,19,UITAG(0),UOHEAD(3,1),14338,17,UPTR(749,1),UOHEAD(3,1),14082,15,UPTR(749,5),UOHEAD(3,1),13826,13,UPTR(749,9),UOHEAD(3,1),13570,11,UPTR(749,13),UOHEAD(3,1),13314,9,UPTR(749,17),UOHEAD(3,1),13058,7,UPTR(749,21),UOHEAD(3,1),12802,5,UPTR(749,25),UOHEAD(3,1),12546,3,UPTR(749,29),UOHEAD(3,1),12290,1,UPTR(749,33)};
pxll_int constructed_750[] = {UPTR(750,85),UOHEAD(3,1),17922,31,UITAG(0),UOHEAD(3,1),17666,29,UPTR(750,1),UOHEAD(3,1),17410,27,UPTR(750,5),UOHEAD(3,1),17154,25,UPTR(750,9),UOHEAD(3,1),16898,23,UPTR(750,13),UOHEAD(3,1),16642,21,UPTR(750,17),UOHEAD(3,1),26114,31,UPTR(750,21),UOHEAD(3,1),25858,29,UPTR(750,25),UOHEAD(3,1),25602,27,UPTR(750,29),UOHEAD(3,1),25346,25,UPTR(750,33),UOHEAD(3,1),25090,23,UPTR(750,37),UOHEAD(3,1),24834,21,UPTR(750,41),UOHEAD(3,1),14594,19,UPTR(750,45),UOHEAD(3,1),14338,17,UPTR(750,49),UOHEAD(3,1),14082,15,UPTR(750,53),UOHEAD(3,1),13826,13,UPTR(750,57),UOHEAD(3,1),13570,11,UPTR(750,61),UOHEAD(3,1),13314,9,UPTR(750,65),UOHEAD(3,1),13058,7,UPTR(750,69),UOHEAD(3,1),12802,5,UPTR(750,73),UOHEAD(3,1),12546,3,UPTR(750,77),UOHEAD(3,1),12290,1,UPTR(750,81)};
pxll_string constructed_751 = {STRING_HEADER(0), 0, "" };
pxll_int constructed_752[] = {UPTR(752,1),(16<<8)|TC_VECTOR,12290,12546,12802,13058,13314,13570,13826,14082,14338,14594,24834,25090,25346,25602,25858,26114};
pxll_string constructed_753 = {STRING_HEADER(0), 0, "" };
pxll_string constructed_754 = {STRING_HEADER(30), 30, "malformed cexp template string" };
pxll_string constructed_755 = {STRING_HEADER(1), 1, "%" };
pxll_string constructed_756 = {STRING_HEADER(0), 0, "" };
pxll_string constructed_757 = {STRING_HEADER(30), 30, "malformed cexp template string" };
pxll_string constructed_758 = {STRING_HEADER(9), 9, "(object*)" };
pxll_string constructed_759 = {STRING_HEADER(3), 3, "ptr" };
// symbol ptr
pxll_int constructed_760[] = {UPTR(760,1), SYMBOL_HEADER, UPTR0(759), INTCON(80)};
pxll_string constructed_761 = {STRING_HEADER(9), 9, "(object*)" };
pxll_string constructed_762 = {STRING_HEADER(7), 7, "cstring" };
// symbol cstring
pxll_int constructed_763[] = {UPTR(763,1), SYMBOL_HEADER, UPTR0(762), INTCON(81)};
pxll_string constructed_764 = {STRING_HEADER(10), 10, "PXLL_TEST(" };
pxll_string constructed_765 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_766 = {STRING_HEADER(14), 14, "box((pxll_int)" };
pxll_string constructed_767 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_768 = {STRING_HEADER(14), 14, "box((pxll_int)" };
pxll_string constructed_769 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_770 = {STRING_HEADER(7), 7, "struct " };
pxll_string constructed_771 = {STRING_HEADER(6), 6, "struct" };
// symbol struct
pxll_int constructed_772[] = {UPTR(772,1), SYMBOL_HEADER, UPTR0(771), INTCON(82)};
pxll_string constructed_773 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_774 = {STRING_HEADER(2), 2, "*)" };
pxll_string constructed_775 = {STRING_HEADER(15), 15, "malformed ctype" };
pxll_string constructed_776 = {STRING_HEADER(16), 16, "((pxll_string*)(" };
pxll_string constructed_777 = {STRING_HEADER(2), 2, "))" };
pxll_string constructed_778 = {STRING_HEADER(25), 25, "unknown raw type in %cexp" };
pxll_string constructed_779 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_780 = {STRING_HEADER(16), 16, "(((pxll_vector*)" };
pxll_string constructed_781 = {STRING_HEADER(5), 5, ")+1))" };
pxll_string constructed_782 = {STRING_HEADER(7), 7, "(char*)" };
pxll_string constructed_783 = {STRING_HEADER(16), 16, "((pxll_string*)(" };
pxll_string constructed_784 = {STRING_HEADER(8), 8, "))->data" };
pxll_string constructed_785 = {STRING_HEADER(6), 6, "unbox(" };
pxll_string constructed_786 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_787 = {STRING_HEADER(6), 6, "unbox(" };
pxll_string constructed_788 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_789 = {STRING_HEADER(8), 8, "wrap-in:" };
pxll_string constructed_790 = {STRING_HEADER(27), 27, "expected immediate literal " };
pxll_string constructed_791 = {STRING_HEADER(17), 17, "inline: bad arity" };
pxll_int constructed_792[] = {TC_NIL};
pxll_string constructed_793 = {STRING_HEADER(2), 2, "_i" };
pxll_int constructed_794[] = {TC_NIL};
pxll_int constructed_795[] = {TC_NIL};
pxll_int constructed_796[] = {TC_NIL};
pxll_int constructed_797[] = {TC_NIL};
pxll_string constructed_798 = {STRING_HEADER(2), 2, "_i" };
pxll_int constructed_799[] = {TC_NIL};
pxll_string constructed_800 = {STRING_HEADER(43), 43, "inline-application - inlining non-function?" };
pxll_int constructed_801[] = {TC_NIL};
pxll_int constructed_802[] = {TC_NIL};
pxll_int constructed_803[] = {TC_NIL};
pxll_int constructed_804[] = {TC_NIL};
pxll_int constructed_805[] = {TC_NIL};
pxll_int constructed_806[] = {TC_NIL};
pxll_int constructed_807[] = {TC_NIL};
pxll_string constructed_808 = {STRING_HEADER(33), 33, "find-escaping-funs: failed lookup" };
pxll_int constructed_809[] = {TC_NIL};
pxll_int constructed_810[] = {TC_NIL};
pxll_int constructed_811[] = {TC_NIL};
pxll_string constructed_812 = {STRING_HEADER(11), 11, "simpleopt: " };
pxll_string constructed_813 = {STRING_HEADER(1), 1, "\n" };
pxll_int constructed_814[] = {TC_NIL};
pxll_int constructed_815[] = {TC_NIL};
pxll_string constructed_816 = {STRING_HEADER(3), 3, "moo" };
// symbol moo
pxll_int constructed_817[] = {UPTR(817,1), SYMBOL_HEADER, UPTR0(816), INTCON(83)};
pxll_string constructed_818 = {STRING_HEADER(6), 6, "rlabel" };
// symbol rlabel
pxll_int constructed_819[] = {UPTR(819,1), SYMBOL_HEADER, UPTR0(818), INTCON(84)};
pxll_string constructed_820 = {STRING_HEADER(10), 10, "type error" };
pxll_string constructed_821 = {STRING_HEADER(14), 14, "\nType Error:\n\t" };
pxll_string constructed_822 = {STRING_HEADER(2), 2, "\n\t" };
pxll_string constructed_823 = {STRING_HEADER(1), 1, "\n" };
pxll_string constructed_824 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_825 = {STRING_HEADER(5), 5, "- ret" };
pxll_string constructed_826 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_827 = {STRING_HEADER(4), 4, "tail" };
pxll_string constructed_828 = {STRING_HEADER(6), 6, "trcall" };
pxll_string constructed_829 = {STRING_HEADER(3), 3, "lit" };
pxll_string constructed_830 = {STRING_HEADER(6), 6, "litcon" };
pxll_string constructed_831 = {STRING_HEADER(4), 4, "cexp" };
pxll_string constructed_832 = {STRING_HEADER(4), 4, "test" };
pxll_string constructed_833 = {STRING_HEADER(3), 3, "jmp" };
pxll_string constructed_834 = {STRING_HEADER(5), 5, "close" };
pxll_string constructed_835 = {STRING_HEADER(3), 3, "ref" };
pxll_string constructed_836 = {STRING_HEADER(3), 3, "set" };
pxll_string constructed_837 = {STRING_HEADER(4), 4, "stor" };
pxll_string constructed_838 = {STRING_HEADER(6), 6, "invoke" };
pxll_string constructed_839 = {STRING_HEADER(3), 3, "env" };
pxll_string constructed_840 = {STRING_HEADER(5), 5, "alloc" };
pxll_string constructed_841 = {STRING_HEADER(4), 4, "push" };
pxll_string constructed_842 = {STRING_HEADER(3), 3, "pop" };
pxll_string constructed_843 = {STRING_HEADER(6), 6, "primop" };
pxll_string constructed_844 = {STRING_HEADER(4), 4, "move" };
pxll_string constructed_845 = {STRING_HEADER(6), 6, "fatbar" };
pxll_string constructed_846 = {STRING_HEADER(4), 4, "fail" };
pxll_string constructed_847 = {STRING_HEADER(8), 8, "testcexp" };
pxll_string constructed_848 = {STRING_HEADER(6), 6, "nvcase" };
pxll_string constructed_849 = {STRING_HEADER(6), 6, "pvcase" };
pxll_string constructed_850 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_851 = {STRING_HEADER(1), 1, "-" };
pxll_string constructed_852 = {STRING_HEADER(2), 2, "- " };
pxll_string constructed_853 = {STRING_HEADER(8), 8, "k/target" };
pxll_string constructed_854 = {STRING_HEADER(6), 6, "k/free" };
pxll_int constructed_855[] = {TC_NIL};
pxll_int constructed_856[] = {TC_NIL};
pxll_string constructed_857 = {STRING_HEADER(1), 1, "_" };
pxll_string constructed_858 = {STRING_HEADER(22), 22, "unhandled literal type" };
pxll_int constructed_859[] = {TC_NIL};
pxll_int constructed_860[] = {TC_NIL};
pxll_int constructed_861[] = {TC_NIL};
pxll_string constructed_862 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_863 = {STRING_HEADER(9), 9, "let_subst" };
// symbol let_subst
pxll_int constructed_864[] = {UPTR(864,1), SYMBOL_HEADER, UPTR0(863), INTCON(85)};
pxll_string constructed_865 = {STRING_HEADER(6), 6, "letrec" };
// symbol letrec
pxll_int constructed_866[] = {UPTR(866,1), SYMBOL_HEADER, UPTR0(865), INTCON(86)};
pxll_string constructed_867 = {STRING_HEADER(7), 7, "literal" };
// symbol literal
pxll_int constructed_868[] = {UPTR(868,1), SYMBOL_HEADER, UPTR0(867), INTCON(87)};
pxll_string constructed_869 = {STRING_HEADER(5), 5, "quote" };
// symbol quote
pxll_int constructed_870[] = {UPTR(870,1), SYMBOL_HEADER, UPTR0(869), INTCON(88)};
pxll_string constructed_871 = {STRING_HEADER(28), 28, "null primapp missing params?" };
pxll_string constructed_872 = {STRING_HEADER(14), 14, "syntax error: " };
pxll_string constructed_873 = {STRING_HEADER(16), 16, "syntax error 2: " };
pxll_int constructed_874[] = {TC_NIL};
pxll_int constructed_875[] = {TC_NIL};
pxll_int constructed_876[] = {TC_NIL};
pxll_int constructed_877[] = {TC_NIL};
pxll_int constructed_878[] = {TC_NIL};
pxll_int constructed_879[] = {TC_NIL};
pxll_string constructed_880 = {STRING_HEADER(15), 15, "unpack-bindings" };
pxll_int constructed_881[] = {TC_NIL};
pxll_string constructed_882 = {STRING_HEADER(16), 16, "malformed formal" };
pxll_string constructed_883 = {STRING_HEADER(7), 7, "varref " };
pxll_string constructed_884 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_885 = {STRING_HEADER(7), 7, "varset " };
pxll_string constructed_886 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_887 = {STRING_HEADER(8), 8, "literal " };
pxll_string constructed_888 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_889 = {STRING_HEADER(5), 5, "cexp " };
pxll_string constructed_890 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_891 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_892 = {STRING_HEADER(11), 11, "sequence : " };
pxll_string constructed_893 = {STRING_HEADER(14), 14, "conditional : " };
pxll_string constructed_894 = {STRING_HEADER(7), 7, "call : " };
pxll_string constructed_895 = {STRING_HEADER(9), 9, "function " };
pxll_string constructed_896 = {STRING_HEADER(2), 2, " (" };
pxll_string constructed_897 = {STRING_HEADER(4), 4, ") : " };
pxll_string constructed_898 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_899 = {STRING_HEADER(5), 5, "fix (" };
pxll_string constructed_900 = {STRING_HEADER(4), 4, ") : " };
pxll_string constructed_901 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_902 = {STRING_HEADER(7), 7, "nvcase " };
pxll_string constructed_903 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_904 = {STRING_HEADER(3), 3, ") (" };
pxll_string constructed_905 = {STRING_HEADER(4), 4, ") : " };
pxll_string constructed_906 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_907 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_908 = {STRING_HEADER(6), 6, "subst " };
pxll_string constructed_909 = {STRING_HEADER(2), 2, "->" };
pxll_string constructed_910 = {STRING_HEADER(8), 8, "primapp " };
pxll_string constructed_911 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_912 = {STRING_HEADER(3), 3, " : " };
pxll_string constructed_913 = {STRING_HEADER(5), 5, "let (" };
pxll_string constructed_914 = {STRING_HEADER(4), 4, ") : " };
pxll_string constructed_915 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_916 = {STRING_HEADER(1), 1, " " };
pxll_int constructed_917[] = {TC_NIL};
pxll_string constructed_918 = {STRING_HEADER(2), 2, "#u" };
pxll_string constructed_919 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_920 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_921 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_922 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_923 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_924 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_925 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_926 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_927 = {STRING_HEADER(2), 2, "#(" };
pxll_string constructed_928 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_929 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_930 = {STRING_HEADER(20), 20, "unpack-fix: no body?" };
pxll_string constructed_931 = {STRING_HEADER(12), 12, "varref->name" };
pxll_string constructed_932 = {STRING_HEADER(4), 4, "  )\n" };
pxll_string constructed_933 = {STRING_HEADER(4), 4, "  (:" };
pxll_string constructed_934 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_935 = {STRING_HEADER(2), 2, ")\n" };
pxll_string constructed_936 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_937 = {STRING_HEADER(10), 10, "(datatype " };
pxll_int constructed_938[] = {TC_NIL};
pxll_int constructed_939[] = {TC_NIL};
pxll_int constructed_940[] = {TC_NIL};
pxll_int constructed_941[] = {TC_NIL};
pxll_int constructed_942[] = {TC_NIL};
pxll_string constructed_943 = {STRING_HEADER(1), 1, "." };
// symbol .
pxll_int constructed_944[] = {UPTR(944,1), SYMBOL_HEADER, UPTR0(943), INTCON(89)};
pxll_string constructed_945 = {STRING_HEADER(17), 17, "malformed pattern" };
pxll_int constructed_946[] = {TC_NIL};
pxll_string constructed_947 = {STRING_HEADER(25), 25, "not a constructor pattern" };
pxll_string constructed_948 = {STRING_HEADER(25), 25, "not a constructor pattern" };
pxll_string constructed_949 = {STRING_HEADER(25), 25, "not a constructor pattern" };
pxll_string constructed_950 = {STRING_HEADER(15), 15, "stack underflow" };
pxll_string constructed_951 = {STRING_HEADER(15), 15, "stack underflow" };
pxll_int constructed_952[] = {TC_NIL};
pxll_string constructed_953 = {STRING_HEADER(21), 21, "not a literal pattern" };
pxll_string constructed_954 = {STRING_HEADER(20), 20, "not a record pattern" };
pxll_string constructed_955 = {STRING_HEADER(1), 1, "_" };
pxll_string constructed_956 = {STRING_HEADER(40), 40, "record pattern with different label sigs" };
pxll_string constructed_957 = {STRING_HEADER(6), 6, "record" };
// symbol record
pxll_int constructed_958[] = {UPTR(958,1), SYMBOL_HEADER, UPTR0(957), INTCON(90)};
pxll_string constructed_959 = {STRING_HEADER(33), 33, "arity mismatch in variant pattern" };
pxll_int constructed_960[] = {TC_NIL};
pxll_int constructed_961[] = {TC_NIL};
pxll_string constructed_962 = {STRING_HEADER(39), 39, "more than one datatype in pattern match" };
pxll_string constructed_963 = {STRING_HEADER(8), 8, "variable" };
// symbol variable
pxll_int constructed_964[] = {UPTR(964,1), SYMBOL_HEADER, UPTR0(963), INTCON(91)};
pxll_string constructed_965 = {STRING_HEADER(8), 8, "string=?" };
// symbol string=?
pxll_int constructed_966[] = {UPTR(966,1), SYMBOL_HEADER, UPTR0(965), INTCON(92)};
pxll_string constructed_967 = {STRING_HEADER(3), 3, "eq?" };
// symbol eq?
pxll_int constructed_968[] = {UPTR(968,1), SYMBOL_HEADER, UPTR0(967), INTCON(93)};
pxll_string constructed_969 = {STRING_HEADER(19), 19, "empty pattern list?" };
pxll_string constructed_970 = {STRING_HEADER(29), 29, "remove-first-pat: empty pats?" };
pxll_string constructed_971 = {STRING_HEADER(1), 1, "m" };
pxll_int constructed_972[] = {TC_NIL};
pxll_int constructed_973[] = {TC_NIL};
pxll_string constructed_974 = {STRING_HEADER(14), 14, "bad record sig" };
pxll_string constructed_975 = {STRING_HEADER(26), 26, "unable to get record sig: " };
pxll_string constructed_976 = {STRING_HEADER(1), 1, "\n" };
pxll_int constructed_977[] = {TC_NIL};
pxll_string constructed_978 = {STRING_HEADER(8), 8, "bad type" };
pxll_string constructed_979 = {STRING_HEADER(15), 15, "malformed type?" };
pxll_int constructed_980[] = {TC_NIL};
pxll_string constructed_981 = {STRING_HEADER(14), 14, "bad arrow type" };
pxll_string constructed_982 = {STRING_HEADER(19), 19, "malformed predicate" };
pxll_string constructed_983 = {STRING_HEADER(1), 1, "t" };
pxll_string constructed_984 = {STRING_HEADER(2), 2, "->" };
pxll_string constructed_985 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_986 = {STRING_HEADER(3), 3, ")->" };
pxll_string constructed_987 = {STRING_HEADER(2), 2, ", " };
pxll_string constructed_988 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_989 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_990 = {STRING_HEADER(2), 2, ", " };
pxll_string constructed_991 = {STRING_HEADER(18), 18, "expected sexp:list" };
pxll_string constructed_992 = {STRING_HEADER(13), 13, "expected list" };
pxll_int constructed_993[] = {TC_NIL};
pxll_int constructed_994[] = {TC_NIL};
pxll_int constructed_995[] = {TC_NIL};
pxll_string constructed_996 = {STRING_HEADER(36), 36, "unexpected args to ellipsis-sub-envs" };
pxll_int constructed_997[] = {TC_NIL};
pxll_int constructed_998[] = {TC_NIL};
pxll_string constructed_999 = {STRING_HEADER(2), 2, "()" };
pxll_string constructed_1000 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_1001 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_1002 = {STRING_HEADER(2), 2, "  " };
pxll_string constructed_1003 = {STRING_HEADER(1), 1, "(" };
pxll_string constructed_1004 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_1005 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_1006 = {STRING_HEADER(1), 1, "\"" };
pxll_string constructed_1007 = {STRING_HEADER(1), 1, "\"" };
pxll_string constructed_1008 = {STRING_HEADER(2), 2, "#\\" };
pxll_string constructed_1009 = {STRING_HEADER(2), 2, "#t" };
pxll_string constructed_1010 = {STRING_HEADER(2), 2, "#f" };
pxll_string constructed_1011 = {STRING_HEADER(2), 2, "#u" };
pxll_string constructed_1012 = {STRING_HEADER(2), 2, "#(" };
pxll_string constructed_1013 = {STRING_HEADER(1), 1, ")" };
pxll_string constructed_1014 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_1015 = {STRING_HEADER(1), 1, "{" };
pxll_string constructed_1016 = {STRING_HEADER(1), 1, "}" };
pxll_string constructed_1017 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_1018 = {STRING_HEADER(1), 1, ":" };
pxll_string constructed_1019 = {STRING_HEADER(0), 0, "" };
pxll_string constructed_1020 = {STRING_HEADER(1), 1, "." };
pxll_string constructed_1021 = {STRING_HEADER(3), 3, "..." };
pxll_string constructed_1022 = {STRING_HEADER(1), 1, "=" };
pxll_string constructed_1023 = {STRING_HEADER(9), 9, "sexp->int" };
pxll_string constructed_1024 = {STRING_HEADER(12), 12, "sexp->symbol" };
pxll_int constructed_1025[] = {TC_NIL};
pxll_string constructed_1026 = {STRING_HEADER(7), 7, "include" };
// symbol include
pxll_int constructed_1027[] = {UPTR(1027,1), SYMBOL_HEADER, UPTR0(1026), INTCON(94)};
pxll_int constructed_1028[] = {TC_NIL};
pxll_string constructed_1029 = {STRING_HEADER(13), 13, "bad hex digit" };
pxll_string constructed_1030 = {STRING_HEADER(12), 12, "colon syntax" };
pxll_string constructed_1031 = {STRING_HEADER(30), 30, "expected closing ]/} character" };
pxll_string constructed_1032 = {STRING_HEADER(22), 22, "unexpected close-paren" };
pxll_string constructed_1033 = {STRING_HEADER(3), 3, "nul" };
pxll_string constructed_1034 = {STRING_HEADER(3), 3, "eof" };
pxll_string constructed_1035 = {STRING_HEADER(3), 3, "tab" };
pxll_string constructed_1036 = {STRING_HEADER(6), 6, "return" };
pxll_string constructed_1037 = {STRING_HEADER(5), 5, "space" };
pxll_string constructed_1038 = {STRING_HEADER(7), 7, "newline" };
pxll_string constructed_1039 = {STRING_HEADER(22), 22, "bad character constant" };
pxll_string constructed_1040 = {STRING_HEADER(12), 12, "syntax error" };
pxll_string constructed_1041 = {STRING_HEADER(15), 15, "expected symbol" };
pxll_string constructed_1042 = {STRING_HEADER(36), 36, "no dots allowed in constructor names" };
pxll_string constructed_1043 = {STRING_HEADER(5), 5, "comma" };
// symbol comma
pxll_int constructed_1044[] = {UPTR(1044,1), SYMBOL_HEADER, UPTR0(1043), INTCON(95)};
pxll_int constructed_1045[] = {TC_NIL};
pxll_string constructed_1046 = {STRING_HEADER(30), 30, "bad backslash escape in string" };
pxll_int constructed_1047[] = {TC_NIL};
pxll_string constructed_1048 = {STRING_HEADER(30), 30, "expected '=' in record literal" };
pxll_int constructed_1049[] = {TC_NIL};
pxll_string constructed_1050 = {STRING_HEADER(22), 22, "unexpected end of file" };
pxll_string constructed_1051 = {STRING_HEADER(18), 18, "bad decimal digit?" };
pxll_int constructed_1052[] = {TC_NIL};
pxll_string constructed_1053 = {STRING_HEADER(22), 22, "unexpected end-of-file" };
pxll_string constructed_1054 = {STRING_HEADER(13), 13, "reading file " };
pxll_int constructed_1055[] = {TC_NIL};
pxll_int constructed_1056[] = {TC_NIL};
pxll_string constructed_1057 = {STRING_HEADER(0), 0, "" };
pxll_string constructed_1058 = {STRING_HEADER(13), 13, "open() failed" };
pxll_int constructed_1059[] = {TC_NIL};
pxll_int constructed_1060[] = {TC_NIL};
pxll_int constructed_1061[] = {TC_NIL};
pxll_int constructed_1062[] = {TC_NIL};
pxll_int constructed_1063[] = {TC_NIL};
pxll_int constructed_1064[] = {TC_NIL};
pxll_int constructed_1065[] = {UPTR(1065,1),(0<<8)|TC_VECTOR};
pxll_int constructed_1066[] = {TC_NIL};
pxll_string constructed_1067 = {STRING_HEADER(29), 29, "every2?: unequal-length lists" };
pxll_string constructed_1068 = {STRING_HEADER(31), 31, "for-each2: unequal-length lists" };
pxll_int constructed_1069[] = {TC_NIL};
pxll_int constructed_1070[] = {TC_NIL};
pxll_string constructed_1071 = {STRING_HEADER(26), 26, "map2: unequal-length lists" };
pxll_int constructed_1072[] = {TC_NIL};
pxll_int constructed_1073[] = {TC_NIL};
pxll_int constructed_1074[] = {TC_NIL};
pxll_string constructed_1075 = {STRING_HEADER(23), 23, "list index out of range" };
pxll_string constructed_1076 = {STRING_HEADER(23), 23, "list index out of range" };
pxll_int constructed_1077[] = {TC_NIL};
pxll_string constructed_1078 = {STRING_HEADER(4), 4, "last" };
pxll_string constructed_1079 = {STRING_HEADER(3), 3, "cdr" };
pxll_string constructed_1080 = {STRING_HEADER(3), 3, "car" };
pxll_string constructed_1081 = {STRING_HEADER(1), 1, " " };
pxll_string constructed_1082 = {STRING_HEADER(17), 17, "lookup-label-code" };
pxll_string constructed_1083 = {STRING_HEADER(25), 25, "vars-get-var: no such var" };
pxll_int pxll_internal_symbols[] = {(96<<8)|TC_VECTOR, UPTR(41,1), UPTR(44,1), UPTR(46,1), UPTR(49,1), UPTR(51,1), UPTR(56,1), UPTR(58,1), UPTR(61,1), UPTR(66,1), UPTR(86,1), UPTR(97,1), UPTR(210,1), UPTR(214,1), UPTR(226,1), UPTR(230,1), UPTR(239,1), UPTR(243,1), UPTR(250,1), UPTR(255,1), UPTR(266,1), UPTR(277,1), UPTR(286,1), UPTR(295,1), UPTR(308,1), UPTR(314,1), UPTR(328,1), UPTR(472,1), UPTR(474,1), UPTR(481,1), UPTR(483,1), UPTR(485,1), UPTR(488,1), UPTR(491,1), UPTR(496,1), UPTR(498,1), UPTR(501,1), UPTR(503,1), UPTR(509,1), UPTR(516,1), UPTR(538,1), UPTR(541,1), UPTR(543,1), UPTR(545,1), UPTR(553,1), UPTR(557,1), UPTR(560,1), UPTR(585,1), UPTR(595,1), UPTR(597,1), UPTR(599,1), UPTR(601,1), UPTR(603,1), UPTR(605,1), UPTR(607,1), UPTR(609,1), UPTR(611,1), UPTR(613,1), UPTR(615,1), UPTR(622,1), UPTR(624,1), UPTR(635,1), UPTR(639,1), UPTR(652,1), UPTR(654,1), UPTR(660,1), UPTR(663,1), UPTR(675,1), UPTR(681,1), UPTR(685,1), UPTR(708,1), UPTR(710,1), UPTR(712,1), UPTR(714,1), UPTR(716,1), UPTR(718,1), UPTR(720,1), UPTR(722,1), UPTR(732,1), UPTR(734,1), UPTR(736,1), UPTR(760,1), UPTR(763,1), UPTR(772,1), UPTR(817,1), UPTR(819,1), UPTR(864,1), UPTR(866,1), UPTR(868,1), UPTR(870,1), UPTR(944,1), UPTR(958,1), UPTR(964,1), UPTR(966,1), UPTR(968,1), UPTR(1027,1), UPTR(1044,1)};

pxll_int
vm (int argc, char * argv[])
{
  register object * lenv = PXLL_NIL;
  register object * k = PXLL_NIL;
// REGISTER_DECLARATIONS //
  register object * r0;
  register object * r1;
  register object * r2;
  register object * r3;
  register object * r4;
  register object * r5;
  register object * r6;
  register object * r7;
  void gc_regs_in (int n) {
    switch (n) {
    case 8: heap1[10] = r7;
    case 7: heap1[9] = r6;
    case 6: heap1[8] = r5;
    case 5: heap1[7] = r4;
    case 4: heap1[6] = r3;
    case 3: heap1[5] = r2;
    case 2: heap1[4] = r1;
    case 1: heap1[3] = r0;
  }}
  void gc_regs_out (int n) {
    switch (n) {
    case 8: r7 = heap0[10];
    case 7: r6 = heap0[9];
    case 6: r5 = heap0[8];
    case 5: r4 = heap0[7];
    case 4: r3 = heap0[6];
    case 3: r2 = heap0[5];
    case 2: r1 = heap0[4];
    case 1: r0 = heap0[3];
  }}
  object * top = PXLL_NIL; // top-level (i.e. 'global') environment
  object * t = 0; // temp - for swaps & building tuples
  object * result;
  object * limit = heap0 + (heap_size - head_room);
  object * freep = heap0;
  int i; // loop counter
  
#define PXLL_RETURN(d)	result = r##d; goto *k[3]

#include "gc.c"

  // check heap is called at the top of each allocating function.
  //  [by locating the check at the top, we avoid considering any
  //   registers as roots of the gc...]
  void check_heap (int nfree) {
    if (freep >= limit) {
      uint64_t t0, t1;
      t0 = rdtsc();
      gc_flip (nfree);
      t1 = rdtsc();
      gc_ticks += t1 - t0;
    }
  }

  void ensure_heap (int nfree, pxll_int size) {
    if (freep + size + 1 >= limit) {
      uint64_t t0, t1;
      t0 = rdtsc();
      gc_flip (nfree);
      t1 = rdtsc();
      gc_ticks += t1 - t0;
    }
  }

  object * allocate (pxll_int tc, pxll_int size) {
    object * save = freep;
    *freep = (object*) (size<<8 | (tc & 0xff));
#if 1
    // at least on the g5, this technique is considerably faster than using memset
    //   in gc_flip() to 'pre-clear' the heap... probably a cache effect...
    while (size--) {
      // this keeps gc from being confused by partially-filled objects.
      *(++freep) = PXLL_NIL;
    }
    ++freep;
#else
    // if you use this version, be sure to set <clear_tospace>!
    freep += size + 1;
#endif
    return save;  
  }

  // this is emitted by the backend for %make-tuple
  object * alloc_no_clear (pxll_int tc, pxll_int size) {
    object * save = freep;
    *freep = (object*) (size<<8 | (tc & 0xff));
    freep += size + 1;
    return save;  
  }

  // gcc inlines/unrolls these nicely, they allow more compact code
  // Note: I'm not using these any longer, because there's a huge
  //   impact on compile times - 11m vs 50m when I tested it.
  inline object * varref (pxll_int depth, pxll_int index)
  {
    object * walk = lenv;
    while (depth--) {
      walk = walk[1];
    }
    return walk[index+2];
  }

  inline void varset (pxll_int depth, pxll_int index, object * val)
  {
    object * walk = lenv;
    while (depth--) {
      walk = walk[1];
    }
    walk[index+2] = val;
  }

  
  // these could probably be written in irken...
  pxll_int dump_image (char * filename, object * closure) {
    FILE * dump_file = fopen (filename, "wb");
    pxll_int offset;
    pxll_int size;
    object * start;
    // do a gc for a compact dump
    closure = gc_dump (closure);
    // for now, start at the front of the heap
    start = heap0;
    size = freep - start;
    offset = (pxll_int) heap0;
    // XXX add endian indicator...
    fprintf (dump_file, "(pxll image %" PRIuPTR " %p)\n", sizeof (pxll_int), start);
    fwrite (&offset, sizeof(pxll_int), 1, dump_file);
    fwrite (&size, sizeof(pxll_int), 1, dump_file);
    fwrite (start, sizeof(pxll_int), size, dump_file);
    fclose (dump_file);
    return size;
  }

  object * load_image (char * filename) {
    FILE * load_file = fopen (filename, "rb");
    if (!load_file) {
      abort();
    } else {
      object * start, * thunk;
      pxll_int size;
      read_header (load_file);	// XXX verify header...
      fread (&start, sizeof(pxll_int), 1, load_file);
      fread (&size, sizeof(pxll_int), 1, load_file);
      fread (heap1, sizeof(pxll_int), size, load_file);
      fprintf (stderr, "size=%d\n", (int) size);
      // relocate heap0
      gc_relocate (4, heap1, heap1 + size, start - heap1);
      // replace roots
      lenv  = (object *) heap1[0];
      k     = (object *) heap1[1];
      top   = (object *) heap1[2];
      thunk = (object *) heap1[3];
      freep = heap1 + size;
      // swap heaps
      { object * temp = heap0; heap0 = heap1; heap1 = temp; }
      return thunk;
    }
  }

  k = allocate (TC_SAVE, 3);
  k[1] = (object *) PXLL_NIL; // top of stack
  k[2] = (object *) PXLL_NIL; // null environment
  k[3] = &&Lreturn; // continuation that will return from this function.
  // --- BEGIN USER PROGRAM ---
  r0 = allocate (TC_TUPLE, 209);
  top = r0;
  r0[1] = lenv; lenv = r0;
  // def list:cons_3
  goto L0;
  FUN_list_3acons_3:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L0:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_list_3acons_3; r1[2] = lenv;
  r0[2] = r1;
  // def sexp:symbol_15
  goto L1;
  FUN_sexp_3asymbol_15:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L1:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sexp_3asymbol_15; r1[2] = lenv;
  r0[3] = r1;
  // def sexp:int_19
  goto L2;
  FUN_sexp_3aint_19:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(5),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L2:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sexp_3aint_19; r1[2] = lenv;
  r0[4] = r1;
  // def vars-get-var_246
  goto L3;
  FUN_vars_get_var_246:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[16];
    r1[2] = r2;
    r2 = top[82];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L4; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L4:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_1083;
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L3:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_vars_get_var_246; r1[2] = lenv;
  r0[5] = r1;
  // def vars-set-flag!_248
  goto L5;
  FUN_vars_set_flag_bang_248:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[3] = r2;
    r2 = top[5];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L6; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
    L6:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,72)];
    r0[2] = r1;
    r0 = ((object***) lenv) [1][1][4];
    r1 = ((object*) lenv) [2];
    r0 = box((pxll_int)unbox(r1)|(1<<unbox(r0)));
    lenv = lenv[1];
    r1 = ((object*) lenv) [2];
    ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,72)] = r0;
    PXLL_RETURN(0);
  L5:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_vars_set_flag_bang_248; r1[2] = lenv;
  r0[6] = r1;
  // def add-var_250
  goto L7;
  FUN_add_var_250:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][3];
    r2 = ((pxll_vector*)r2)->val[16];
    r1[2] = r2;
    r2 = top[82];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L8; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L8:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = allocate (TC_TUPLE, 5);
        r1 = ((object**) lenv) [1][3];
        r1 = ((pxll_vector*)r1)->val[16];
        r0[2] = r1;
        r1 = top[82];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[4] = r1;
        r1 = allocate (UOTAG(25), 5);
        r2 = (object *) 1;
        r1[1] = r2;
        r2 = (object *) 1;
        r1[2] = r2;
        r2 = (object *) 1;
        r1[3] = r2;
        r2 = (object *) 1;
        r1[4] = r2;
        r2 = (object *) 1;
        r1[5] = r2;
        r0[5] = r1;
        r1 = top[72];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L9; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
        L9:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        r1 = ((object**) lenv) [1][3];
        ((pxll_vector*)r1)->val[16] = r0;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L7:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_add_var_250; r1[2] = lenv;
  r0[7] = r1;
  // def add-vars_251
  goto L10;
  FUN_add_vars_251:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    // def add_665
    goto L11;
    FUN_add_665:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][3];
      r0[3] = r1;
      r1 = top[7];
      r0[1] = r1[2]; lenv = r0; goto FUN_add_var_250;
    L11:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_add_665; r1[2] = lenv;
    r0[2] = r1;
    // def search_666
    goto L12;
    FUN_search_666:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(10)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object****) lenv) [1][1][1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L13; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L13:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
        } break;
        case (UOTAG(9)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object****) lenv) [1][1][1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L14; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L14:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
        } break;
        case (UOTAG(7)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object****) lenv) [1][1][1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L15; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L15:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          r0[3] = r1;
          r1 = top[7];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L16; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_add_var_250;
          L16:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_419;
        }
      }
      goto fatbar_419_over;
      fatbar_419:
      r0 = (object *) 14;
      fatbar_419_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object**) lenv) [1][3];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L12:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_search_666; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[1] = r1[2]; lenv = r0; goto FUN_search_666;
  L10:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_add_vars_251; r1[2] = lenv;
  r0[8] = r1;
  // def lookup-label-code_253
  goto L17;
  FUN_lookup_label_code_253:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_678
    goto L18;
    FUN_loop_678:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object*) &constructed_1082;
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][2];
          r0[3] = r1;
          r1 = top[15];
          r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case_noint (r0)) {
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*****) lenv) [1][1][1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = ((object*) lenv) [3];
                PXLL_RETURN(0);
              } else {
                r0 = ((object**) lenv) [1][3];
                lenv = ((object **)lenv)[1][1];
                lenv[2] = r0;
                goto FUN_loop_678;
              }
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L18:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_678; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][3];
    r1 = ((pxll_vector*)r1)->val[5];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_678;
  L17:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_lookup_label_code_253; r1[2] = lenv;
  r0[9] = r1;
  // def terpri_260
  goto L19;
  FUN_terpri_260:
    r0 = (object *) 2562;
    r0 = box((pxll_int)fputc (GET_CHAR(r0), stdout));
    PXLL_RETURN(0);
  L19:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_terpri_260; r1[2] = lenv;
  r0[10] = r1;
  // def =_261
  goto L20;
  FUN__3d_261:
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    PXLL_RETURN(0);
  L20:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN__3d_261; r1[2] = lenv;
  r0[11] = r1;
  // def <_263
  goto L21;
  FUN__3c_263:
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    PXLL_RETURN(0);
  L21:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN__3c_263; r1[2] = lenv;
  r0[12] = r1;
  // def +_269
  goto L22;
  FUN__plus_269:
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = box((pxll_int)unbox(r1)+unbox(r0));
    PXLL_RETURN(0);
  L22:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN__plus_269; r1[2] = lenv;
  r0[13] = r1;
  // def eq?_285
  goto L23;
  FUN_eq_question_285:
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(r1==r0);
    PXLL_RETURN(0);
  L23:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_eq_question_285; r1[2] = lenv;
  r0[14] = r1;
  // def error1_293
  goto L24;
  FUN_error1_293:
    check_heap (0);
    r0 = top[158];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L25; ; k = t;
    lenv = r0[2]; goto *r0[1];
    L25:
    ; lenv = k[2]; k = k[1];
    r0 = ((object*) lenv) [2];
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = ((object*) lenv) [2];
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_1081;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_1081;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = ((object*) lenv) [3];
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = ((object*) lenv) [3];
    PXLL_UNDEFINED; result=r0; goto Lreturn;
    PXLL_RETURN(0);
  L24:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_error1_293; r1[2] = lenv;
  r0[15] = r1;
  // def id_296
  goto L26;
  FUN_id_296:
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L26:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_id_296; r1[2] = lenv;
  r0[16] = r1;
  // def pair->first_300
  goto L27;
  FUN_pair__3efirst_300:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case_noint (r0)) {
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L27:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pair__3efirst_300; r1[2] = lenv;
  r0[17] = r1;
  // def pair->second_301
  goto L28;
  FUN_pair__3esecond_301:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case_noint (r0)) {
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L28:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pair__3esecond_301; r1[2] = lenv;
  r0[18] = r1;
  // def car_309
  goto L29;
  FUN_car_309:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_1080;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_1080;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L29:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_car_309; r1[2] = lenv;
  r0[19] = r1;
  // def cdr_310
  goto L30;
  FUN_cdr_310:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_1079;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_1079;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L30:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_cdr_310; r1[2] = lenv;
  r0[20] = r1;
  // def reverse-onto_311
  goto L31;
  FUN_reverse_onto_311:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[3] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
      } break;
    }
    PXLL_RETURN(0);
  L31:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_reverse_onto_311; r1[2] = lenv;
  r0[21] = r1;
  // def length_314
  goto L32;
  FUN_length_314:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def fun_791
    goto L33;
    FUN_fun_791:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object*) lenv) [3];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object**) lenv) [1][3];
          r1 = (object *) 3;
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_fun_791;
        } break;
      }
      PXLL_RETURN(0);
    L33:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_fun_791; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_fun_791;
  L32:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_length_314; r1[2] = lenv;
  r0[22] = r1;
  // def last_317
  goto L34;
  FUN_last_317:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_1078;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_1078;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_418;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_418_over;
        fatbar_418:
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = top[23];
        r0[1] = r1[2]; lenv = r0; goto FUN_last_317;
        fatbar_418_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L34:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_last_317; r1[2] = lenv;
  r0[23] = r1;
  // def member?_318
  goto L35;
  FUN_member_question_318:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) 6;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][4];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L36; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L36:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = (object *) 262;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = ((object**) lenv) [1][4];
          r0[4] = r1;
          r1 = top[24];
          r0[1] = r1[2]; lenv = r0; goto FUN_member_question_318;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L35:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_member_question_318; r1[2] = lenv;
  r0[24] = r1;
  // def member-eq?_319
  goto L37;
  FUN_member_eq_question_319:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) 6;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) 262;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[25];
          r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L37:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_member_eq_question_319; r1[2] = lenv;
  r0[25] = r1;
  // def remove-eq_320
  goto L38;
  FUN_remove_eq_320:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_1077[0];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][2];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object***) lenv) [1][1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][3];
          r1[3] = r2;
          r2 = top[26];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L39; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_remove_eq_320;
          L39:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L38:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_remove_eq_320; r1[2] = lenv;
  r0[26] = r1;
  // def nth_321
  goto L40;
  FUN_nth_321:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_1076;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_1076;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) 1;
        r1 = ((object**) lenv) [1][3];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } else {
          goto fatbar_417;
        }
        PXLL_RETURN(0);
        goto fatbar_417_over;
        fatbar_417:
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = (object *) 3;
        r2 = ((object**) lenv) [1][3];
        r1 = box((pxll_int)unbox(r2)-unbox(r1));
        r0[3] = r1;
        r1 = top[27];
        r0[1] = r1[2]; lenv = r0; goto FUN_nth_321;
        fatbar_417_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L40:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_nth_321; r1[2] = lenv;
  r0[27] = r1;
  // def index-eq_322
  goto L41;
  FUN_index_eq_322:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_820
    goto L42;
    FUN_loop_820:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = (object*) &constructed_68;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_1075;
          dump_object (r0, 0); fprintf (stdout, "\n");
          r0 = (object*) &constructed_1075;
          PXLL_UNDEFINED; result=r0; goto Lreturn;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object****) lenv) [1][1][1][2];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = ((object**) lenv) [1][2];
            PXLL_RETURN(0);
          } else {
            r0 = (object *) 3;
            r1 = ((object**) lenv) [1][2];
            r0 = box((pxll_int)unbox(r1)+unbox(r0));
            r1 = ((object*) lenv) [3];
            lenv = ((object *)lenv)[1];
            lenv[2] = r0;
            lenv[3] = r1;
            goto FUN_loop_820;
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L42:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_820; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) 1;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_820;
  L41:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_index_eq_322; r1[2] = lenv;
  r0[28] = r1;
  // def slice_323
  goto L43;
  FUN_slice_323:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][3];
    r2 = ((object**) lenv) [1][4];
    r1 = box((pxll_int)unbox(r2)-unbox(r1));
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = (object *) constructed_1073[0];
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_828
      goto L44;
      FUN_loop_828:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][3];
        r1 = ((object*) lenv) [3];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<unbox(r0))) {
          r0 = (object *) 3;
          r1 = ((object*) lenv) [3];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[20];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L45; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
          L45:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r2 = ((object*) lenv) [4];
          lenv[2] = r1;
          lenv[3] = r0;
          lenv[4] = r2;
          goto FUN_loop_828;
        } else {
          r0 = ((object***) lenv) [1][1][4];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<unbox(r0))) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = top[19];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L46; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
            L46:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object**) lenv) [1][4];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = (object *) 3;
            r2 = ((object*) lenv) [3];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = top[20];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L47; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_cdr_310;
            L47:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv[2] = r2;
            lenv[3] = r1;
            lenv[4] = r0;
            goto FUN_loop_828;
          } else {
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [4];
            r0[2] = r1;
            r1 = (object *) constructed_36[0];
            r0[3] = r1;
            r1 = top[21];
            r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L44:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_828; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 4);
      r1 = ((object**) lenv) [1][2];
      r0[2] = r1;
      r1 = (object *) 1;
      r0[3] = r1;
      r1 = (object *) constructed_1074[0];
      r0[4] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_828;
    }
    PXLL_RETURN(0);
  L43:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_slice_323; r1[2] = lenv;
  r0[29] = r1;
  // def range_324
  goto L48;
  FUN_range_324:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_833
    goto L49;
    FUN_loop_833:
      check_heap (0);
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<unbox(r0))) {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } else {
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        r1 = (object *) 3;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)-unbox(r1));
        lenv[2] = r1;
        lenv[3] = r0;
        goto FUN_loop_833;
      }
      PXLL_RETURN(0);
    L49:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_833; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) 3;
    r2 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)unbox(r2)-unbox(r1));
    r0[2] = r1;
    r1 = (object*)TC_NIL;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_833;
  L48:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_range_324; r1[2] = lenv;
  r0[30] = r1;
  // def n-of_325
  goto L50;
  FUN_n_of_325:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_838
    goto L51;
    FUN_loop_838:
      check_heap (0);
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<=unbox(r0))) {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } else {
        r0 = ((object*) lenv) [3];
        r1 = ((object***) lenv) [1][1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        r1 = (object *) 3;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)-unbox(r1));
        lenv[2] = r1;
        lenv[3] = r0;
        goto FUN_loop_838;
      }
      PXLL_RETURN(0);
    L51:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_838; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = (object*)TC_NIL;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_838;
  L50:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_n_of_325; r1[2] = lenv;
  r0[31] = r1;
  // def map_326
  goto L52;
  FUN_map_326:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_1072[0];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L53; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto *r2[1];
        L53:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = ((object*) lenv) [3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L54; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L54:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r0 = ((object*) lenv) [5];
        r1 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L52:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_map_326; r1[2] = lenv;
  r0[32] = r1;
  // def map2_327
  goto L55;
  FUN_map2_327:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = (object *) constructed_1070[0];
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_416;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][4];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][4];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L56; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L56:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 4);
            r2 = ((object***) lenv) [1][1][2];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][3];
            r1[3] = r2;
            r2 = ((object*) lenv) [3];
            r1[4] = r2;
            r2 = top[33];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L57; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map2_327;
            L57:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r0 = ((object*) lenv) [5];
            r1 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_416;
          }
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_416_over;
    fatbar_416:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_1071;
    r0[2] = r1;
    r1 = allocate (UOTAG(1), 2);
    r2 = ((object*) lenv) [3];
    r1[1] = r2;
    r2 = ((object*) lenv) [4];
    r1[2] = r2;
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_416_over: ;
    PXLL_RETURN(0);
  L55:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_map2_327; r1[2] = lenv;
  r0[33] = r1;
  // def filter_328
  goto L58;
  FUN_filter_328:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_1069[0];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L59; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L59:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object***) lenv) [1][1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][3];
          r1[3] = r2;
          r2 = top[34];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L60; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_filter_328;
          L60:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[34];
          r0[1] = r1[2]; lenv = r0; goto FUN_filter_328;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L58:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_filter_328; r1[2] = lenv;
  r0[34] = r1;
  // def for-each_329
  goto L61;
  FUN_for_each_329:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L62; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L62:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object**) lenv) [1][2];
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = top[35];
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      } break;
    }
    PXLL_RETURN(0);
  L61:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_for_each_329; r1[2] = lenv;
  r0[35] = r1;
  // def for-each2_330
  goto L63;
  FUN_for_each2_330:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_415;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][4];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][4];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L64; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L64:
            ; lenv = k[2]; k = k[1];
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object***) lenv) [1][1][2];
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r0[3] = r1;
            r1 = ((object*) lenv) [3];
            r0[4] = r1;
            r1 = top[36];
            r0[1] = r1[2]; lenv = r0; goto FUN_for_each2_330;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_415;
          }
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_415_over;
    fatbar_415:
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_1068;
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = (object*) &constructed_1068;
    PXLL_UNDEFINED; result=r0; goto Lreturn;
    PXLL_RETURN(0);
    fatbar_415_over: ;
    PXLL_RETURN(0);
  L63:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_for_each2_330; r1[2] = lenv;
  r0[36] = r1;
  // def fold_331
  goto L65;
  FUN_fold_331:
    check_heap (0);
    r0 = ((object*) lenv) [4];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L66; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L66:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        r1 = ((object*) lenv) [3];
        r2 = ((object**) lenv) [1][2];
        lenv[2] = r2;
        lenv[3] = r0;
        lenv[4] = r1;
        goto FUN_fold_331;
      } break;
    }
    PXLL_RETURN(0);
  L65:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_fold_331; r1[2] = lenv;
  r0[37] = r1;
  // def foldr_332
  goto L67;
  FUN_foldr_332:
    check_heap (0);
    r0 = ((object*) lenv) [4];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r1[4] = r2;
        r2 = top[38];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L68; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_foldr_332;
        L68:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
      } break;
    }
    PXLL_RETURN(0);
  L67:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_foldr_332; r1[2] = lenv;
  r0[38] = r1;
  // def some?_333
  goto L69;
  FUN_some_question_333:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) 6;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L70; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L70:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = (object *) 262;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[39];
          r0[1] = r1[2]; lenv = r0; goto FUN_some_question_333;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L69:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_some_question_333; r1[2] = lenv;
  r0[39] = r1;
  // def every?_334
  goto L71;
  FUN_every_question_334:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) 262;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L72; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L72:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[40];
          r0[1] = r1[2]; lenv = r0; goto FUN_every_question_334;
        } else {
          r0 = (object *) 6;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L71:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_every_question_334; r1[2] = lenv;
  r0[40] = r1;
  // def every2?_335
  goto L73;
  FUN_every2_question_335:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = (object *) 262;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_414;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][4];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][4];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L74; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L74:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 4);
              r1 = ((object***) lenv) [1][1][2];
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r0[3] = r1;
              r1 = ((object*) lenv) [3];
              r0[4] = r1;
              r1 = top[41];
              r0[1] = r1[2]; lenv = r0; goto FUN_every2_question_335;
            } else {
              r0 = (object *) 6;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_414;
          }
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_414_over;
    fatbar_414:
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_1067;
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = (object*) &constructed_1067;
    PXLL_UNDEFINED; result=r0; goto Lreturn;
    PXLL_RETURN(0);
    fatbar_414_over: ;
    PXLL_RETURN(0);
  L73:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_every2_question_335; r1[2] = lenv;
  r0[41] = r1;
  // def pack_337
  goto L75;
  FUN_pack_337:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def pack2_899
    goto L76;
    FUN_pack2_899:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L77; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L77:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = (object*)TC_NIL;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][4];
          r0[3] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L78; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L78:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          if PXLL_IS_TRUE(r0) {
            r0 = ((object**) lenv) [1][3];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            r1 = ((object**) lenv) [1][4];
            r2 = ((object*) lenv) [3];
            lenv = ((object *)lenv)[1];
            lenv[2] = r2;
            lenv[3] = r0;
            lenv[4] = r1;
            goto FUN_pack2_899;
          } else {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][3];
            r1[2] = r2;
            r2 = (object *) constructed_36[0];
            r1[3] = r2;
            r2 = top[21];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L79; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
            L79:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 4);
            r2 = ((object**) lenv) [1][3];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = (object*)TC_NIL;
            r2[2] = r3;
            r2 = ((object*) lenv) [2];
            r3 = ((object***) lenv) [1][1][2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            lenv = lenv[1];
            r1[3] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[4] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L80; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_pack2_899;
            L80:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L76:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_pack2_899; r1[2] = lenv;
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_1066[0];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = (object*)TC_NIL;
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        r2 = ((object**) lenv) [1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r0[4] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN_pack2_899;
      } break;
    }
    PXLL_RETURN(0);
  L75:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pack_337; r1[2] = lenv;
  r0[42] = r1;
  // def list->vector_339
  goto L81;
  FUN_list__3evector_339:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def recur_912
    goto L82;
    FUN_recur_912:
      check_heap (0);
      r0 = ((object*) lenv) [4];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [2];
          r2 = ((object**) lenv) [1][2];
          range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r0));
          ((pxll_vector*)r2)->val[unbox(r0)] = r1;
          r0 = (object *) 3;
          r1 = ((object**) lenv) [1][3];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          r1 = ((object*) lenv) [3];
          r2 = ((object**) lenv) [1][2];
          lenv = ((object *)lenv)[1];
          lenv[2] = r2;
          lenv[3] = r0;
          lenv[4] = r1;
          goto FUN_recur_912;
        } break;
      }
      PXLL_RETURN(0);
    L82:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_recur_912; r1[2] = lenv;
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_1065[0];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L83; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L83:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = ((object*) lenv) [3];
        ensure_heap (1, unbox(r1));
        r1 = ((object*) lenv) [2];
        r2 = ((object*) lenv) [3];
        if (unbox(r2) == 0) { r1 = (object *) TC_EMPTY_VECTOR; } else {
          t = alloc_no_clear (TC_VECTOR, unbox(r2));
          for (i=0; i<unbox(r2); i++) { t[i+1] = r1; }
          r1 = t;
        }
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object*) lenv) [4];
        r0[2] = r1;
        r1 = (object *) 1;
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r0[4] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN_recur_912;
      } break;
    }
    PXLL_RETURN(0);
  L81:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_list__3evector_339; r1[2] = lenv;
  r0[43] = r1;
  // def sort_340
  goto L84;
  FUN_sort_340:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    // def halve_924
    goto L85;
    FUN_halve_924:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (UOTAG(1), 2);
          r1 = ((object*) lenv) [2];
          r0[1] = r1;
          r1 = (object *) constructed_1063[0];
          r0[2] = r1;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = allocate (UOTAG(1), 2);
              r1 = ((object**) lenv) [1][2];
              r0[1] = r1;
              r1 = (object *) constructed_1064[0];
              r0[2] = r1;
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_413;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_413_over;
          fatbar_413:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L86; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_halve_924;
          L86:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case_noint (r0)) {
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = allocate (UOTAG(1), 2);
              r1 = ((object*) lenv) [3];
              r2 = ((object***) lenv) [1][1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[1] = r1;
              r1 = ((object*) lenv) [2];
              r0[2] = r1;
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
          fatbar_413_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L85:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_halve_924; r1[2] = lenv;
    r0[2] = r1;
    // def merge-sort_925
    goto L87;
    FUN_merge_sort_925:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_412;
            }
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
      goto fatbar_412_over;
      fatbar_412:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = ((object***) lenv) [1][1][2];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L88; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_halve_924;
      L88:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case_noint (r0)) {
        default: {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L89; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_merge_sort_925;
          L89:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L90; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_merge_sort_925;
          L90:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def loop_928_i84
          goto L91;
          FUN_loop_928_i84:
            check_heap (0);
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (TC_NIL): {
                r0 = ((object*) lenv) [3];
                PXLL_RETURN(0);
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object**) lenv) [1][3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = ((object**) lenv) [1][2];
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object***) lenv) [1][1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object***) lenv) [1][1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = allocate (TC_TUPLE, 3);
                    r1 = ((object**) lenv) [1][2];
                    r0[2] = r1;
                    r1 = ((object*) lenv) [2];
                    r0[3] = r1;
                    r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L92; ; k = t;
                    r0[1] = r1[2]; lenv = r0; goto *r1[1];
                    L92:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    if PXLL_IS_TRUE(r0) {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object***) lenv) [1][1][3];
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r3 = ((object**) lenv) [1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[3] = r2;
                      r2 = ((object*****) lenv) [1][1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L93; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_loop_928_i84;
                      L93:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      r1 = ((object***) lenv) [1][1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r1;
                      t[2] = r0;
                      r0 = t;
                      PXLL_RETURN(0);
                    } else {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object***) lenv) [1][1][3];
                      r3 = ((object***) lenv) [1][1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r1[3] = r2;
                      r2 = ((object*****) lenv) [1][1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L94; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_loop_928_i84;
                      L94:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      r1 = ((object**) lenv) [1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r1;
                      t[2] = r0;
                      r0 = t;
                      PXLL_RETURN(0);
                    }
                    PXLL_RETURN(0);
                  } break;
                }
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
          L91:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_loop_928_i84; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][4];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][5];
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          r0[1] = r1[2]; lenv = r0; goto FUN_loop_928_i84;
        } break;
      }
      PXLL_RETURN(0);
      fatbar_412_over: ;
      PXLL_RETURN(0);
    L87:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_merge_sort_925; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][3];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[1] = r1[2]; lenv = r0; goto FUN_merge_sort_925;
  L84:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sort_340; r1[2] = lenv;
  r0[44] = r1;
  // def substring_345
  goto L95;
  FUN_substring_345:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][3];
    r2 = ((object**) lenv) [1][4];
    r1 = box((pxll_int)unbox(r2)-unbox(r1));
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = box((pxll_int)string_tuple_length (unbox(r1)));
    ensure_heap (1, unbox(r1));
    r1 = ((object*) lenv) [2];
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
    r0[3] = r1;
    r0 = ((object**) lenv) [1][3];
    r1 = (object *) 1;
    r2 = ((object**) lenv) [1][2];
    r3 = ((object*) lenv) [2];
    r4 = ((object*) lenv) [3];
    (memcpy (((pxll_string*)(r4))->data+unbox(r1), ((pxll_string*)(r2))->data+unbox(r0), unbox(r3)), PXLL_UNDEFINED);
    r0 = ((object*) lenv) [3];
    PXLL_RETURN(0);
  L95:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_substring_345; r1[2] = lenv;
  r0[45] = r1;
  // def char->string_348
  goto L96;
  FUN_char__3estring_348:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) 3;
    r1 = box((pxll_int)string_tuple_length (unbox(r1)));
    ensure_heap (1, unbox(r1));
    r1 = (object *) 3;
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    range_check (((pxll_string *)(((pxll_string*)(r1))))->len, unbox(r0));
    r0 = (object *) 1;
    r1 = ((object**) lenv) [1][2];
    r2 = ((object*) lenv) [2];
    (((pxll_string*)(r2))->data[unbox(r0)] = GET_CHAR (r1), PXLL_UNDEFINED);
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L96:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_char__3estring_348; r1[2] = lenv;
  r0[46] = r1;
  // def string-concat_352
  goto L97;
  FUN_string_concat_352:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    // def loop_973
    goto L98;
    FUN_loop_973:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object*) lenv) [3];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object***) lenv) [1][1][3];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv = lenv[1];
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_loop_973;
        } break;
      }
      PXLL_RETURN(0);
    L98:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_loop_973; r2[2] = lenv;
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object***) lenv) [1][1][2];
    r1[2] = r2;
    r2 = (object *) 1;
    r1[3] = r2;
    r2 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L99; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_loop_973;
    L99:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = box((pxll_int)string_tuple_length (unbox(r1)));
    ensure_heap (1, unbox(r1));
    r1 = ((object*) lenv) [2];
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_980
    goto L100;
    FUN_loop_980:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object***) lenv) [1][1][3];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r0[2] = r1;
          r0 = (object *) 1;
          r1 = ((object***) lenv) [1][1][3];
          r2 = ((object**) lenv) [1][2];
          r3 = ((object*) lenv) [2];
          r4 = ((object*****) lenv) [1][1][1][1][3];
          r0 = (memcpy (((pxll_string*)(r4))->data+unbox(r1), ((pxll_string*)(r2))->data+unbox(r0), unbox(r3)), PXLL_UNDEFINED);
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object***) lenv) [1][1][3];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv = lenv[1];
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_loop_980;
        } break;
      }
      PXLL_RETURN(0);
    L100:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_980; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_980;
  L97:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_concat_352; r1[2] = lenv;
  r0[47] = r1;
  // def string-join_353
  goto L101;
  FUN_string_join_353:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def sj_987
    goto L102;
    FUN_sj_987:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L103; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L103:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object***) lenv) [1][1][3];
              r3 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[2] = r2;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*) lenv) [2];
              r1[2] = r2;
              r2 = (object *) constructed_36[0];
              r1[3] = r2;
              r2 = top[21];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L104; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
              L104:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              lenv = lenv[1];
              r0[2] = r1;
              r1 = top[47];
              r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
            } break;
            default: {
              goto fatbar_411;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_411_over;
          fatbar_411:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object***) lenv) [1][1][3];
          r2 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_sj_987;
          fatbar_411_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L102:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_sj_987; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][3];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object**) lenv) [1][2];
      r0[2] = r1;
      r1 = top[47];
      r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
    } else {
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object**) lenv) [1][2];
      r0[2] = r1;
      r1 = (object *) constructed_1062[0];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_sj_987;
    }
    PXLL_RETURN(0);
  L101:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_join_353; r1[2] = lenv;
  r0[48] = r1;
  // def string-split_354
  goto L105;
  FUN_string_split_354:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_994
    goto L106;
    FUN_loop_994:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object****) lenv) [1][1][1][2];
      r1 = box((pxll_int)((pxll_string*)(r1))->len);
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object****) lenv) [1][1][1][2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[4] = r2;
        r2 = top[45];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L107; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_substring_345;
        L107:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object**) lenv) [1][4];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = (object *) constructed_36[0];
        r0[3] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r2 = ((object****) lenv) [1][1][1][2];
        range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
        r1 = ((object**) lenv) [1][2];
        r2 = ((object****) lenv) [1][1][1][2];
        r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
        r0[2] = r1;
        r0 = ((object****) lenv) [1][1][1][3];
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 4);
          r2 = ((object****) lenv) [1][1][1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][3];
          r1[3] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[4] = r2;
          r2 = top[45];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L108; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_substring_345;
          L108:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object**) lenv) [1][4];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = (object *) 3;
          r2 = ((object*) lenv) [2];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r2 = (object *) 3;
          r3 = ((object*) lenv) [2];
          r2 = box((pxll_int)unbox(r3)+unbox(r2));
          lenv[2] = r2;
          lenv[3] = r1;
          lenv[4] = r0;
          goto FUN_loop_994;
        } else {
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          r1 = ((object*) lenv) [3];
          r2 = ((object*) lenv) [4];
          lenv[2] = r0;
          lenv[3] = r1;
          lenv[4] = r2;
          goto FUN_loop_994;
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L106:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_994; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r1 = (object *) 1;
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = (object *) constructed_1061[0];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_994;
  L105:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_split_354; r1[2] = lenv;
  r0[49] = r1;
  // def string-compare_355
  goto L109;
  FUN_string_compare_355:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r2 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r1))) {
      r1 = ((object*) lenv) [2];
    } else {
      r1 = ((object*) lenv) [3];
    }
    r2 = ((object**) lenv) [1][3];
    r3 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)memcmp (((pxll_string*)(r3))->data, ((pxll_string*)(r2))->data, unbox(r1)));
    r0[4] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [4];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
      r0 = ((object*) lenv) [3];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = (object *) 1;
        PXLL_RETURN(0);
      } else {
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<unbox(r0))) {
          r0 = (object *) -1;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) 3;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    } else {
      r0 = ((object*) lenv) [4];
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L109:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_compare_355; r1[2] = lenv;
  r0[50] = r1;
  // def string-find_356
  goto L110;
  FUN_string_find_356:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[3] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object*) lenv) [3];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<unbox(r0))) {
      r0 = (object *) -1;
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1007
      goto L111;
      FUN_loop_1007:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][2];
        r1 = ((object*) lenv) [3];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
          r0 = ((object*) lenv) [3];
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)-unbox(r0));
          PXLL_RETURN(0);
        } else {
          r0 = ((object***) lenv) [1][1][3];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
            r0 = (object *) -1;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r2 = ((object*****) lenv) [1][1][1][1][2];
            range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
            r1 = ((object**) lenv) [1][3];
            r2 = ((object*****) lenv) [1][1][1][1][2];
            r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r2 = ((object*****) lenv) [1][1][1][1][3];
            range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
            r1 = ((object**) lenv) [1][2];
            r2 = ((object*****) lenv) [1][1][1][1][3];
            r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 3;
              r1 = ((object*) lenv) [3];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              r1 = (object *) 3;
              r2 = ((object*) lenv) [2];
              r1 = box((pxll_int)unbox(r2)+unbox(r1));
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN_loop_1007;
            } else {
              r0 = (object *) 3;
              r1 = ((object*) lenv) [2];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              r1 = (object *) 1;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1007;
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L111:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1007; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = (object *) 1;
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1007;
    }
    PXLL_RETURN(0);
  L110:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_find_356; r1[2] = lenv;
  r0[51] = r1;
  // def string<?_358
  goto L112;
  FUN_string_3c_question_358:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[3] = r2;
    r2 = top[50];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L113; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
    L113:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    PXLL_RETURN(0);
  L112:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string_3c_question_358; r1[2] = lenv;
  r0[52] = r1;
  // def zero-terminate_360
  goto L114;
  FUN_zero_terminate_360:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r1 = (object *) 3;
    r2 = ((object*) lenv) [2];
    r1 = box((pxll_int)unbox(r2)-unbox(r1));
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r2 = ((object**) lenv) [1][2];
    range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
    r1 = ((object*) lenv) [3];
    r2 = ((object**) lenv) [1][2];
    r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
    r0[4] = r1;
    r0 = (object *) 2;
    r1 = ((object*) lenv) [4];
    r0 = PXLL_TEST(r1==r0);
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = box((pxll_int)((pxll_string*)(r1))->len);
      r0[2] = r1;
      r1 = (object *) 3;
      r2 = ((object*) lenv) [2];
      r1 = box((pxll_int)unbox(r2)+unbox(r1));
      r0[3] = r1;
      r1 = ((object*) lenv) [3];
      r1 = box((pxll_int)string_tuple_length (unbox(r1)));
      ensure_heap (1, unbox(r1));
      r1 = ((object*) lenv) [3];
      r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
      r0[4] = r1;
      r0 = (object *) 1;
      r1 = (object *) 1;
      r2 = ((object**) lenv) [1][2];
      r3 = ((object*) lenv) [2];
      r4 = ((object*) lenv) [4];
      (memcpy (((pxll_string*)(r4))->data+unbox(r1), ((pxll_string*)(r2))->data+unbox(r0), unbox(r3)), PXLL_UNDEFINED);
      r0 = ((object*) lenv) [2];
      r1 = ((object*) lenv) [4];
      range_check (((pxll_string *)(((pxll_string*)(r1))))->len, unbox(r0));
      r0 = ((object*) lenv) [2];
      r1 = (object *) 2;
      r2 = ((object*) lenv) [4];
      (((pxll_string*)(r2))->data[unbox(r0)] = GET_CHAR (r1), PXLL_UNDEFINED);
      r0 = ((object*) lenv) [4];
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L114:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_zero_terminate_360; r1[2] = lenv;
  r0[53] = r1;
  // def list->string_361
  goto L115;
  FUN_list__3estring_361:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L116; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L116:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = box((pxll_int)string_tuple_length (unbox(r1)));
    ensure_heap (1, unbox(r1));
    r1 = ((object*) lenv) [2];
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_1021
    goto L117;
    FUN_loop_1021:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object***) lenv) [1][1][3];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object**) lenv) [1][3];
          r1 = ((object****) lenv) [1][1][1][3];
          range_check (((pxll_string *)(((pxll_string*)(r1))))->len, unbox(r0));
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [2];
          r2 = ((object****) lenv) [1][1][1][3];
          (((pxll_string*)(r2))->data[unbox(r0)] = GET_CHAR (r1), PXLL_UNDEFINED);
          r0 = (object *) 3;
          r1 = ((object**) lenv) [1][3];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_loop_1021;
        } break;
      }
      PXLL_RETURN(0);
    L117:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1021; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_1021;
  L115:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_list__3estring_361; r1[2] = lenv;
  r0[54] = r1;
  // def string->list_362
  goto L118;
  FUN_string__3elist_362:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_1027
    goto L119;
    FUN_loop_1027:
      check_heap (0);
      r0 = (object *) 1;
      r1 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = (object *) 3;
        r2 = ((object**) lenv) [1][3];
        r1 = box((pxll_int)unbox(r2)-unbox(r1));
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r2 = ((object****) lenv) [1][1][1][2];
        range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
        r1 = ((object*) lenv) [2];
        r2 = ((object****) lenv) [1][1][1][2];
        r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][2];
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = (object *) 3;
        r2 = ((object*) lenv) [3];
        r1 = box((pxll_int)unbox(r2)-unbox(r1));
        lenv[2] = r0;
        lenv[3] = r1;
        goto FUN_loop_1027;
      }
      PXLL_RETURN(0);
    L119:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1027; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*)TC_NIL;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_1027;
  L118:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string__3elist_362; r1[2] = lenv;
  r0[55] = r1;
  // def int->string_364
  goto L120;
  FUN_int__3estring_364:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r1 = (object *) 1;
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) 12290;
      r0[2] = r1;
      r1 = top[46];
      r0[1] = r1[2]; lenv = r0; goto FUN_char__3estring_348;
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1036
      goto L121;
      FUN_loop_1036:
        check_heap (0);
        r0 = ((object*) lenv) [2];
        r1 = (object *) 1;
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object *) 1;
          r2 = ((object***) lenv) [1][1][2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r1))) {
            r1 = ((object*) lenv) [3];
            r2 = (object *) 11522;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
          } else {
            r1 = ((object*) lenv) [3];
          }
          r0[2] = r1;
          r1 = top[54];
          r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_361;
        } else {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = (object *) 21;
          r2 = ((object**) lenv) [1][2];
          r1 = box((pxll_int)unbox(r2) % unbox(r1));
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r2 = (object *) 97;
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = ((object*) lenv) [3];
          r1 = TO_CHAR(unbox(r1));
          r0[4] = r1;
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [4];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = (object *) 21;
          r2 = ((object*) lenv) [2];
          r1 = box((pxll_int)unbox(r2)/unbox(r1));
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_loop_1036;
        }
        PXLL_RETURN(0);
      L121:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1036; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r2 = ((object**) lenv) [1][2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r1))) {
        r1 = ((object**) lenv) [1][2];
        r2 = (object *) 1;
        r1 = box((pxll_int)unbox(r2)-unbox(r1));
      } else {
        r1 = ((object**) lenv) [1][2];
      }
      r0[2] = r1;
      r1 = (object *) constructed_1060[0];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1036;
    }
    PXLL_RETURN(0);
  L120:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_int__3estring_364; r1[2] = lenv;
  r0[56] = r1;
  // def int->hex-string_365
  goto L122;
  FUN_int__3ehex_string_365:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_1040
    goto L123;
    FUN_loop_1040:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r1 = (object *) 1;
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = allocate (TC_TUPLE, 2);
        r1 = (object *) 1;
        r2 = ((object***) lenv) [1][1][2];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r1))) {
          r1 = ((object*) lenv) [3];
          r2 = (object *) 11522;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
        } else {
          r1 = ((object*) lenv) [3];
        }
        r0[2] = r1;
        r1 = top[54];
        r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_361;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = (object *) 31;
        r2 = ((object**) lenv) [1][2];
        r1 = box((pxll_int)unbox(r2)&unbox(r1));
        r2 = top[159];
        range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
        r1 = ((pxll_vector*)r2)->val[unbox(r1)];
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = (object *) 9;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)>>unbox(r1));
        lenv[2] = r1;
        lenv[3] = r0;
        goto FUN_loop_1040;
      }
      PXLL_RETURN(0);
    L123:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1040; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) 1;
    r2 = ((object**) lenv) [1][2];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r1))) {
      r1 = ((object**) lenv) [1][2];
      r2 = (object *) 1;
      r1 = box((pxll_int)unbox(r2)-unbox(r1));
    } else {
      r1 = ((object**) lenv) [1][2];
    }
    r0[2] = r1;
    r1 = (object *) constructed_1059[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_1040;
  L122:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_int__3ehex_string_365; r1[2] = lenv;
  r0[57] = r1;
  // def pad_367
  goto L124;
  FUN_pad_367:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][3];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
      r0 = ((object**) lenv) [1][3];
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r2 = ((object***) lenv) [1][1][2];
      r1 = box((pxll_int)unbox(r2)-unbox(r1));
      r0[2] = r1;
      r0 = ((object***) lenv) [1][1][4];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 4);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = allocate (TC_TUPLE, 3);
        r4 = ((object**) lenv) [1][2];
        r3[2] = r4;
        r4 = (object *) 8194;
        r3[3] = r4;
        r4 = top[31];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L125; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_n_of_325;
        L125:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = top[54];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L126; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_list__3estring_361;
        L126:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = ((object****) lenv) [1][1][1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r1 = ((object*) lenv) [4];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 4);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = allocate (TC_TUPLE, 3);
        r4 = ((object**) lenv) [1][2];
        r3[2] = r4;
        r4 = (object *) 8194;
        r3[3] = r4;
        r4 = top[31];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L127; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_n_of_325;
        L127:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = top[54];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L128; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_list__3estring_361;
        L128:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r1 = ((object*) lenv) [4];
        r2 = ((object****) lenv) [1][1][1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      }
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L124:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pad_367; r1[2] = lenv;
  r0[58] = r1;
  // def copy-cstring_372
  goto L129;
  FUN_copy_cstring_372:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)strlen((char*)r1));
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = box((pxll_int)string_tuple_length (unbox(r1)));
    ensure_heap (1, unbox(r1));
    r1 = ((object*) lenv) [2];
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
    r0[3] = r1;
    r0 = ((object**) lenv) [1][2];
    r1 = ((object*) lenv) [2];
    r2 = ((object*) lenv) [3];
    (memcpy (((pxll_string*)(r2))->data, (char*)r0, unbox(r1)), PXLL_UNDEFINED);
    r0 = ((object*) lenv) [3];
    PXLL_RETURN(0);
  L129:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_copy_cstring_372; r1[2] = lenv;
  r0[59] = r1;
  // def open_373
  goto L130;
  FUN_open_373:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[53];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L131; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_zero_terminate_360;
    L131:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r2 = ((object**) lenv) [1][3];
    r3 = ((object**) lenv) [1][4];
    r1 = box((pxll_int)open (((pxll_string*)(r1))->data, unbox(r2), unbox(r3)));
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>=unbox(r0))) {
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_1058;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[53];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L132; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_zero_terminate_360;
      L132:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    }
    PXLL_RETURN(0);
  L130:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_open_373; r1[2] = lenv;
  r0[60] = r1;
  // def file/open-read_380
  goto L133;
  FUN_file_2fopen_read_380:
    check_heap (0);
    r0 = allocate (UOTAG(24), 4);
    r1 = (object *) 32769;
    r1 = box((pxll_int)string_tuple_length (unbox(r1)));
    ensure_heap (1, unbox(r1));
    r1 = (object *) 32769;
    r1 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r1))), ((pxll_string*)(t))->len = unbox(r1), t);
    r0[1] = r1;
    r1 = (object *) 1;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = top[160];
    r1[3] = r2;
    r2 = (object *) 1;
    r1[4] = r2;
    r2 = top[60];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L134; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_open_373;
    L134:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = (object *) 1;
    r0[4] = r1;
    PXLL_RETURN(0);
  L133:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_file_2fopen_read_380; r1[2] = lenv;
  r0[61] = r1;
  // def file/fill-buffer_385
  goto L135;
  FUN_file_2ffill_buffer_385:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 6);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[4] = r1;
    r1 = ((object*) lenv) [3];
    r2 = ((object*) lenv) [4];
    r3 = ((object*) lenv) [2];
    r1 = box((pxll_int)read (unbox(r3), ((pxll_string*)(r1))->data, unbox(r2)));
    r0[5] = r1;
    r1 = ((object*) lenv) [5];
    r0[6] = r1;
    r0 = ((object*) lenv) [6];
    r1 = ((object**) lenv) [1][2];
    ((pxll_vector*)r1)->val[1] = r0;
    r0 = (object *) 1;
    r1 = ((object**) lenv) [1][2];
    ((pxll_vector*)r1)->val[3] = r0;
    r0 = ((object*) lenv) [6];
    PXLL_RETURN(0);
  L135:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_file_2ffill_buffer_385; r1[2] = lenv;
  r0[62] = r1;
  // def file/read-buffer_386
  goto L136;
  FUN_file_2fread_buffer_386:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[3];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[1];
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[2] = r1;
      r0 = ((object**) lenv) [1][2];
      r0 = ((pxll_vector*)r0)->val[1];
      r1 = ((object**) lenv) [1][2];
      ((pxll_vector*)r1)->val[3] = r0;
      r0 = allocate (TC_TUPLE, 4);
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[3] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r0[4] = r1;
      r1 = top[45];
      r0[1] = r1[2]; lenv = r0; goto FUN_substring_345;
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[62];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L137; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_file_2ffill_buffer_385;
      L137:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object*) &constructed_1057;
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object**) lenv) [1][2];
        r2 = ((pxll_vector*)r2)->val[0];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[3] = r2;
        r2 = ((object**) lenv) [1][2];
        r2 = ((pxll_vector*)r2)->val[1];
        r1[4] = r2;
        r2 = top[45];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L138; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_substring_345;
        L138:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = (object *) 1;
        r1 = ((object**) lenv) [1][2];
        ((pxll_vector*)r1)->val[1] = r0;
        r0 = (object *) 1;
        r1 = ((object**) lenv) [1][2];
        ((pxll_vector*)r1)->val[3] = r0;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L136:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_file_2fread_buffer_386; r1[2] = lenv;
  r0[63] = r1;
  // def file/read-char_387
  goto L139;
  FUN_file_2fread_char_387:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[3];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[1];
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[2] = r1;
      r0 = (object *) 3;
      r1 = ((object*) lenv) [2];
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      lenv = lenv[1];
      r1 = ((object*) lenv) [2];
      ((pxll_vector*)r1)->val[3] = r0;
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = (object *) 3;
      r2 = ((object*) lenv) [3];
      r1 = box((pxll_int)unbox(r2)-unbox(r1));
      r0[4] = r1;
      r0 = ((object*) lenv) [4];
      r1 = ((object*) lenv) [2];
      range_check (((pxll_string *)(((pxll_string*)(r1))))->len, unbox(r0));
      r0 = ((object*) lenv) [4];
      r1 = ((object*) lenv) [2];
      r0 = TO_CHAR(((unsigned char *)((pxll_string*)(r1))->data)[unbox(r0)]);
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[62];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L140; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_file_2ffill_buffer_385;
      L140:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 65794;
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[64];
        r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_387;
      }
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L139:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_file_2fread_char_387; r1[2] = lenv;
  r0[64] = r1;
  // def alist/lookup_394
  goto L141;
  FUN_alist_2flookup_394:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(0)): {
        r0 = (object*)UITAG(1);
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = ((object**) lenv) [1][3];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [4];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r0[3] = r1;
          r1 = top[65];
          r0[1] = r1[2]; lenv = r0; goto FUN_alist_2flookup_394;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L141:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_alist_2flookup_394; r1[2] = lenv;
  r0[65] = r1;
  // def alist/get_396
  goto L142;
  FUN_alist_2fget_396:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[3] = r2;
    r2 = top[65];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L143; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
    L143:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object**) lenv) [1][4];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
    }
    PXLL_RETURN(0);
  L142:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_alist_2fget_396; r1[2] = lenv;
  r0[66] = r1;
  // def alist/iterate_397
  goto L144;
  FUN_alist_2fiterate_397:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (UITAG(0)): {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L145; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L145:
        ; lenv = k[2]; k = k[1];
        r0 = ((object*) lenv) [4];
        r1 = ((object**) lenv) [1][2];
        lenv[2] = r1;
        lenv[3] = r0;
        goto FUN_alist_2fiterate_397;
      } break;
    }
    PXLL_RETURN(0);
  L144:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_alist_2fiterate_397; r1[2] = lenv;
  r0[67] = r1;
  // def alist->keys_399
  goto L146;
  FUN_alist__3ekeys_399:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(0)): {
        r0 = (object*)TC_NIL;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = top[68];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L147; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist__3ekeys_399;
        L147:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L146:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_alist__3ekeys_399; r1[2] = lenv;
  r0[68] = r1;
  // def alist/length_401
  goto L148;
  FUN_alist_2flength_401:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(0)): {
        r0 = (object *) 1;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[69];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L149; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flength_401;
        L149:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = (object *) 3;
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L148:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_alist_2flength_401; r1[2] = lenv;
  r0[69] = r1;
  // def tree/skew_405
  goto L150;
  FUN_tree_2fskew_405:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)>unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
    } else {
      r0 = (object *) 6;
    }
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 11);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[3] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[4] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r1 = ((pxll_vector*)r1)->val[1];
      r0[5] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[6] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[7] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[8] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[9] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r2 = ((pxll_vector*)r2)->val[3];
      r1[2] = r2;
      r2 = top[70];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L151; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fskew_405;
      L151:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[10] = r1;
      r1 = allocate (UOTAG(17), 5);
      r2 = ((object*) lenv) [7];
      r1[1] = r2;
      r2 = ((object*) lenv) [9];
      r1[2] = r2;
      r2 = ((object*) lenv) [6];
      r1[3] = r2;
      r2 = ((object*) lenv) [10];
      r1[4] = r2;
      r2 = ((object*) lenv) [8];
      r1[5] = r2;
      r0[11] = r1;
      r0 = allocate (UOTAG(17), 5);
      r1 = ((object*) lenv) [3];
      r0[1] = r1;
      r1 = ((object*) lenv) [5];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[3] = r1;
      r1 = ((object*) lenv) [11];
      r0[4] = r1;
      r1 = ((object*) lenv) [4];
      r0[5] = r1;
      PXLL_RETURN(0);
    } else {
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L150:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2fskew_405; r1[2] = lenv;
  r0[70] = r1;
  // def tree/split_406
  goto L152;
  FUN_tree_2fsplit_406:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[3];
    r1 = ((pxll_vector*)r1)->val[3];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r1 = (object *) 1;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(unbox(r2)==unbox(r1));
      r0[3] = r1;
      r0 = (object *) 6;
      r1 = ((object*) lenv) [3];
      r0 = PXLL_TEST(r1==r0);
      lenv = lenv[1];
    } else {
      r0 = (object *) 6;
    }
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 12);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r2 = (object *) 3;
      r1 = box((pxll_int)unbox(r2)+unbox(r1));
      r0[3] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[4] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[5] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[6] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[7] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[8] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[1];
      r0[9] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r1 = ((pxll_vector*)r1)->val[1];
      r0[10] = r1;
      r1 = allocate (UOTAG(17), 5);
      r2 = ((object*) lenv) [7];
      r1[1] = r2;
      r2 = ((object*) lenv) [9];
      r1[2] = r2;
      r2 = ((object*) lenv) [6];
      r1[3] = r2;
      r2 = ((object*) lenv) [10];
      r1[4] = r2;
      r2 = ((object*) lenv) [8];
      r1[5] = r2;
      r0[11] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r2 = ((pxll_vector*)r2)->val[3];
      r2 = ((pxll_vector*)r2)->val[3];
      r1[2] = r2;
      r2 = top[71];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L153; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fsplit_406;
      L153:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[12] = r1;
      r0 = allocate (UOTAG(17), 5);
      r1 = ((object*) lenv) [4];
      r0[1] = r1;
      r1 = ((object*) lenv) [11];
      r0[2] = r1;
      r1 = ((object*) lenv) [3];
      r0[3] = r1;
      r1 = ((object*) lenv) [12];
      r0[4] = r1;
      r1 = ((object*) lenv) [5];
      r0[5] = r1;
      PXLL_RETURN(0);
    } else {
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L152:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2fsplit_406; r1[2] = lenv;
  r0[71] = r1;
  // def tree/insert_407
  goto L154;
  FUN_tree_2finsert_407:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_1195
    goto L155;
    FUN_loop_1195:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (UOTAG(17), 5);
        r1 = ((object***) lenv) [1][1][4];
        r0[1] = r1;
        r1 = top[165];
        r0[2] = r1;
        r1 = (object *) 3;
        r0[3] = r1;
        r1 = top[165];
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[5] = r1;
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object***) lenv) [1][1][4];
        r2[2] = r3;
        r3 = ((object*) lenv) [2];
        r3 = ((pxll_vector*)r3)->val[0];
        r2[3] = r3;
        r3 = ((object***) lenv) [1][1][3];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L156; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto *r3[1];
        L156:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        if PXLL_IS_TRUE(r2) {
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[2];
          r2[2] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[0];
          r2[3] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[4];
          r2[4] = r3;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][2];
          r4 = ((pxll_vector*)r4)->val[1];
          r3[2] = r4;
          r4 = ((object***) lenv) [1][1][2];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L157; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_loop_1195;
          L157:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[5] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[3];
          r2[6] = r3;
          r2 = allocate (UOTAG(17), 5);
          r3 = ((object*) lenv) [3];
          r2[1] = r3;
          r3 = ((object*) lenv) [5];
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r2[3] = r3;
          r3 = ((object*) lenv) [6];
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r2[5] = r3;
          lenv = lenv[1];
        } else {
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[2];
          r2[2] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[0];
          r2[3] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[4];
          r2[4] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = ((pxll_vector*)r3)->val[1];
          r2[5] = r3;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][2];
          r4 = ((pxll_vector*)r4)->val[3];
          r3[2] = r4;
          r4 = ((object***) lenv) [1][1][2];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L158; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_loop_1195;
          L158:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[6] = r3;
          r2 = allocate (UOTAG(17), 5);
          r3 = ((object*) lenv) [3];
          r2[1] = r3;
          r3 = ((object*) lenv) [5];
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r2[3] = r3;
          r3 = ((object*) lenv) [6];
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r2[5] = r3;
          lenv = lenv[1];
        }
        r1[2] = r2;
        r2 = top[70];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L159; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fskew_405;
        L159:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = top[71];
        r0[1] = r1[2]; lenv = r0; goto FUN_tree_2fsplit_406;
      }
      PXLL_RETURN(0);
    L155:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1195; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_1195;
  L154:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2finsert_407; r1[2] = lenv;
  r0[72] = r1;
  // def tree/member_409
  goto L160;
  FUN_tree_2fmember_409:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def member0_1209
    goto L161;
    FUN_member0_1209:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object*)UITAG(1);
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object***) lenv) [1][1][4];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][3];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L162; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L162:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = ((object*) lenv) [2];
          r0 = ((pxll_vector*)r0)->val[1];
          lenv[2] = r0;
          goto FUN_member0_1209;
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[0];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][4];
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L163; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L163:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          if PXLL_IS_TRUE(r0) {
            r0 = ((object*) lenv) [2];
            r0 = ((pxll_vector*)r0)->val[3];
            lenv[2] = r0;
            goto FUN_member0_1209;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L161:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_member0_1209; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_member0_1209;
  L160:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2fmember_409; r1[2] = lenv;
  r0[73] = r1;
  // def tree/inorder_410
  goto L164;
  FUN_tree_2finorder_410:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def recur_1213
    goto L165;
    FUN_recur_1213:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[1];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L166; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_recur_1213;
        L166:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[4];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L167; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L167:
        ; lenv = k[2]; k = k[1];
        r0 = ((object*) lenv) [2];
        r0 = ((pxll_vector*)r0)->val[3];
        lenv[2] = r0;
        goto FUN_recur_1213;
      }
      PXLL_RETURN(0);
    L165:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_recur_1213; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][3];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_recur_1213;
  L164:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2finorder_410; r1[2] = lenv;
  r0[74] = r1;
  // def tree/reverse_411
  goto L168;
  FUN_tree_2freverse_411:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def recur_1217
    goto L169;
    FUN_recur_1217:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L170; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_recur_1217;
        L170:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[4];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L171; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L171:
        ; lenv = k[2]; k = k[1];
        r0 = ((object*) lenv) [2];
        r0 = ((pxll_vector*)r0)->val[1];
        lenv[2] = r0;
        goto FUN_recur_1217;
      }
      PXLL_RETURN(0);
    L169:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_recur_1217; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][3];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_recur_1217;
  L168:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2freverse_411; r1[2] = lenv;
  r0[75] = r1;
  // def tree/keys_412
  goto L172;
  FUN_tree_2fkeys_412:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_1056[0];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_4448
    goto L173;
    FUN_lambda_4448:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      lenv = lenv[1];
      ((object**) lenv) [1][2] = r0;
      PXLL_RETURN(0);
    L173:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_4448; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = top[75];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L174; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2freverse_411;
    L174:
    ; lenv = k[2]; k = k[1];
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L172:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2fkeys_412; r1[2] = lenv;
  r0[76] = r1;
  // def tree/values_413
  goto L175;
  FUN_tree_2fvalues_413:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_1055[0];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_4465
    goto L176;
    FUN_lambda_4465:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      lenv = lenv[1];
      ((object**) lenv) [1][2] = r0;
      PXLL_RETURN(0);
    L176:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_4465; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = top[75];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L177; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2freverse_411;
    L177:
    ; lenv = k[2]; k = k[1];
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L175:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_tree_2fvalues_413; r1[2] = lenv;
  r0[77] = r1;
  // def symbol->string_416
  goto L178;
  FUN_symbol__3estring_416:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
    PXLL_RETURN(0);
  L178:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_symbol__3estring_416; r1[2] = lenv;
  r0[78] = r1;
  // def intern-symbol_418
  goto L179;
  FUN_intern_symbol_418:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 5);
    r1 = top[167];
    r0[2] = r1;
    r1 = top[52];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][2];
    r2 = UOBJ_GET(r2,0);
    r1[2] = r2;
    r1 = ((object*) lenv) [2];
    lenv = lenv[1];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[5] = r1;
    r1 = top[72];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L180; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
    L180:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    top[167] = r0;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = top[166];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = (object *) 3;
    r0 = box((pxll_int)unbox(r1)+unbox(r0));
    lenv = lenv[1];
    top[166] = r0;
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L179:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_intern_symbol_418; r1[2] = lenv;
  r0[79] = r1;
  // def string->symbol_419
  goto L181;
  FUN_string__3esymbol_419:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = top[167];
    r1[2] = r2;
    r2 = top[52];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L182; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L182:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = top[166];
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        r2 = ((object***) lenv) [1][1][2];
        t = alloc_no_clear (TC_SYMBOL,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[79];
        r0[1] = r1[2]; lenv = r0; goto FUN_intern_symbol_418;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L181:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_string__3esymbol_419; r1[2] = lenv;
  r0[80] = r1;
  // def symbol<?_420
  goto L183;
  FUN_symbol_3c_question_420:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object***) lenv) [1][1][2];
    r2 = UOBJ_GET(r2,0);
    r1[2] = r2;
    r1 = ((object*) lenv) [2];
    lenv = lenv[1];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object***) lenv) [1][1][3];
    r2 = UOBJ_GET(r2,0);
    r1[2] = r2;
    r1 = ((object*) lenv) [2];
    lenv = lenv[1];
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r1[3] = r2;
    r2 = top[50];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L184; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
    L184:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [4];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    PXLL_RETURN(0);
  L183:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_symbol_3c_question_420; r1[2] = lenv;
  r0[81] = r1;
  // def symbol-index<?_421
  goto L185;
  FUN_symbol_index_3c_question_421:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object***) lenv) [1][1][2];
    r2 = UOBJ_GET(r2,1);
    r1[2] = r2;
    r1 = ((object*) lenv) [2];
    lenv = lenv[1];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object***) lenv) [1][1][3];
    r2 = UOBJ_GET(r2,1);
    r1[2] = r2;
    r1 = ((object*) lenv) [2];
    lenv = lenv[1];
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)<unbox(r0));
    PXLL_RETURN(0);
  L185:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_symbol_index_3c_question_421; r1[2] = lenv;
  r0[82] = r1;
  // def char-class_423
  goto L186;
  FUN_char_class_423:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) 513;
    ensure_heap (1, unbox(r1));
    r1 = (object *) 6;
    r2 = (object *) 513;
    if (unbox(r2) == 0) { r1 = (object *) TC_EMPTY_VECTOR; } else {
      t = alloc_no_clear (TC_VECTOR, unbox(r2));
      for (i=0; i<unbox(r2); i++) { t[i+1] = r1; }
      r1 = t;
    }
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    // def in-class?_1255
    goto L187;
    FUN_in_class_question_1255:
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)GET_CHAR(r0));
      r1 = ((object***) lenv) [1][1][2];
      range_check (GET_TUPLE_LENGTH(*(object*)r1), unbox(r0));
      r0 = ((pxll_vector*)r1)->val[unbox(r0)];
      PXLL_RETURN(0);
    L187:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_in_class_question_1255; r1[2] = lenv;
    r0[2] = r1;
    // def loop_1257
    goto L188;
    FUN_loop_1257:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object**) lenv) [1][2];
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)GET_CHAR(r0));
          r1 = (object *) 262;
          r2 = ((object****) lenv) [1][1][1][2];
          range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r0));
          ((pxll_vector*)r2)->val[unbox(r0)] = r1;
          r0 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          goto FUN_loop_1257;
        } break;
      }
      PXLL_RETURN(0);
    L188:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1257; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_1257;
  L186:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_char_class_423; r1[2] = lenv;
  r0[83] = r1;
  // def reader_424
  goto L189;
  FUN_reader_424:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) 65794;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 10);
    r0[1] = lenv; lenv = r0;
    // def skip-whitespace_1267
    goto L190;
    FUN_skip_whitespace_1267:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1288
      goto L191;
      FUN_loop_1288:
        check_heap (0);
        r0 = (object *) 65794;
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) 15106;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def loop_1286_i256
            goto L192;
            FUN_loop_1286_i256:
              check_heap (0);
              r0 = (object *) 2562;
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                goto fatbar_409;
              }
              PXLL_RETURN(0);
              goto fatbar_409_over;
              fatbar_409:
              r0 = (object *) 3330;
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                goto fatbar_410;
              }
              PXLL_RETURN(0);
              goto fatbar_410_over;
              fatbar_410:
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L193; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L193:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
              r0 = ((object*) lenv) [2];
              lenv = lenv[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L194; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L194:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              lenv[2] = r0;
              goto FUN_loop_1286_i256;
              fatbar_410_over: ;
              PXLL_RETURN(0);
              fatbar_409_over: ;
              PXLL_RETURN(0);
            L192:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_loop_1286_i256; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object******) lenv) [1][1][1][1][1][2];
            r1[2] = r2;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L195; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L195:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            ((object******) lenv) [1][1][1][1][1][2] = r1;
            r1 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L196; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_loop_1286_i256;
            L196:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r0[2] = r1;
            r0 = (object *) 65794;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object*****) lenv) [1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L197; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L197:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object****) lenv) [1][1][1][2] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = ((object****) lenv) [1][1][1][2];
            lenv[2] = r0;
            goto FUN_loop_1288;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = top[176];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L198; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L198:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*****) lenv) [1][1][1][1][2];
              r0[2] = r1;
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L199; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L199:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*****) lenv) [1][1][1][1][2] = r0;
              r0 = ((object*) lenv) [2];
              lenv = lenv[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*****) lenv) [1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*****) lenv) [1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L200; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L200:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object****) lenv) [1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object****) lenv) [1][1][1][2];
              lenv[2] = r0;
              goto FUN_loop_1288;
            } else {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L191:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1288; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L201; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L201:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object***) lenv) [1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1288;
    L190:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_skip_whitespace_1267; r1[2] = lenv;
    r0[2] = r1;
    // def read-atom_1268
    goto L202;
    FUN_read_atom_1268:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = (object *) 1;
      r0[3] = r1;
      r1 = (object *) 12290;
      r0[4] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1293
      goto L203;
      FUN_loop_1293:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 65794;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L204; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L204:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object*****) lenv) [1][1][1][1][2] = r0;
        } else {
          r0 = (object *) 14;
        }
        r0 = ((object*****) lenv) [1][1][1][1][2];
        ((object***) lenv) [1][1][4] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][4];
        r0[2] = r1;
        r0 = (object *) 11778;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
        } else {
          r0 = (object *) 14;
        }
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 7;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 11;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r1 = top[177];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L205; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L205:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 11;
            } else {
              r0 = (object *) 7;
            }
          }
        } else {
          goto fatbar_405;
        }
        goto fatbar_405_over;
        fatbar_405:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 5;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 13;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r1 = top[177];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L206; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L206:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 13;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object***) lenv) [1][1][4];
              r0[2] = r1;
              r1 = top[178];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L207; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L207:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 5;
              } else {
                r0 = (object *) 7;
              }
            }
          }
        } else {
          goto fatbar_406;
        }
        goto fatbar_406_over;
        fatbar_406:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 11;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r1 = top[177];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L208; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L208:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 11;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object***) lenv) [1][1][4];
              r0[2] = r1;
              r1 = top[178];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L209; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L209:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 5;
              } else {
                r0 = (object *) 7;
              }
            }
          }
        } else {
          goto fatbar_407;
        }
        goto fatbar_407_over;
        fatbar_407:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 1;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 9;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][4];
            r0[2] = r1;
            r0 = (object *) 11522;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 3;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object***) lenv) [1][1][4];
              r0[2] = r1;
              r1 = top[178];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L210; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L210:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 5;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object***) lenv) [1][1][4];
                r0[2] = r1;
                r1 = top[177];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L211; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L211:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                if PXLL_IS_TRUE(r0) {
                  r0 = (object *) 15;
                } else {
                  r0 = (object *) 7;
                }
              }
            }
          }
        } else {
          goto fatbar_408;
        }
        goto fatbar_408_over;
        fatbar_408:
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_67;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_67;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        fatbar_408_over: ;
        fatbar_407_over: ;
        fatbar_406_over: ;
        fatbar_405_over: ;
        ((object***) lenv) [1][1][2] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 9;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(unbox(r1)<unbox(r0));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          r0[2] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L212; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L212:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object******) lenv) [1][1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[4] = r1;
          r0 = ((object*) lenv) [4];
          r1 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          lenv[2] = r0;
          goto FUN_loop_1293;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[2] = r1;
          r0 = (object *) 9;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)==unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object*) &constructed_68;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
            r0 = (object*) &constructed_1053;
            dump_object (r0, 0); fprintf (stdout, "\n");
            r0 = (object*) &constructed_1053;
            PXLL_UNDEFINED; result=r0; goto Lreturn;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][2];
            r0[2] = r1;
            r0 = (object *) 15;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(unbox(r1)==unbox(r0));
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = ((object******) lenv) [1][1][1][1][1][2];
              r0[2] = r1;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L213; t[4] = r0; k = t;
              lenv = r1[2]; goto *r1[1];
              L213:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              ((object******) lenv) [1][1][1][1][1][2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[4] = r1;
              r0 = ((object*) lenv) [4];
              r1 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              ((object*) lenv) [2] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = allocate (UOTAG(0), 4);
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L214; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L214:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[54];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L215; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3estring_361;
            L215:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[1] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object****) lenv) [1][1][1][2];
            r1[2] = r2;
            r1 = (object *) 13;
            r2 = ((object*) lenv) [2];
            r1 = PXLL_TEST(unbox(r2)==unbox(r1));
            lenv = lenv[1];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L216; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L216:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][3];
            r0[4] = r1;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L203:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1293; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) constructed_1052[0];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1293;
    L202:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_atom_1268; r1[2] = lenv;
    r0[3] = r1;
    // def read_1272
    goto L217;
    FUN_read_1272:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L218; t[4] = r0; k = t;
      lenv = r1[2]; goto FUN_skip_whitespace_1267;
      L218:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object*****) lenv) [1][1][1][1][2];
      r2[2] = r3;
      r2 = (object *) 65794;
      r3 = ((object*) lenv) [2];
      r2 = PXLL_TEST(r3==r2);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r2) {
        r2 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L219; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L219:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        ((object****) lenv) [1][1][1][2] = r2;
      } else {
        r2 = (object *) 14;
      }
      r2 = ((object****) lenv) [1][1][1][2];
      r1[2] = r2;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 10498;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = (object*) &constructed_68;
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r2 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
        r1 = (object*) &constructed_1032;
        dump_object (r1, 0); fprintf (stdout, "\n");
        r1 = (object*) &constructed_1032;
        PXLL_UNDEFINED; result=r1; goto Lreturn;
      } else {
        goto fatbar_365;
      }
      goto fatbar_365_over;
      fatbar_365:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 8962;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L220; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L220:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = (object *) 65794;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = ((object*****) lenv) [1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L221; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L221:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object****) lenv) [1][1][1][2] = r1;
        } else {
          r1 = (object *) 14;
        }
        r1 = ((object****) lenv) [1][1][1][2];
        ((object*) lenv) [2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 10242;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][7];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L222; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_read_list_1277;
          L222:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(7),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_367;
        }
        goto fatbar_367_over;
        fatbar_367:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 29954;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L223; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L223:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object*)UITAG(6);
        } else {
          goto fatbar_368;
        }
        goto fatbar_368_over;
        fatbar_368:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 21762;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L224; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L224:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object*)UITAG(6);
        } else {
          goto fatbar_369;
        }
        goto fatbar_369_over;
        fatbar_369:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 26114;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L225; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L225:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 6;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_370;
        }
        goto fatbar_370_over;
        fatbar_370:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 17922;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L226; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L226:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 6;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_371;
        }
        goto fatbar_371_over;
        fatbar_371:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 29698;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L227; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L227:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 262;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_372;
        }
        goto fatbar_372_over;
        fatbar_372:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 21506;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L228; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L228:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 262;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_373;
        }
        goto fatbar_373_over;
        fatbar_373:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 28418;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L229; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L229:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][9];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L230; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_read_oct_int_1282;
          L230:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_374;
        }
        goto fatbar_374_over;
        fatbar_374:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 20226;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L231; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L231:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][9];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L232; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_read_oct_int_1282;
          L232:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_375;
        }
        goto fatbar_375_over;
        fatbar_375:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 30722;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L233; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L233:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][8];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L234; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_read_hex_int_1281;
          L234:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_376;
        }
        goto fatbar_376_over;
        fatbar_376:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 22530;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L235; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L235:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][8];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L236; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_read_hex_int_1281;
          L236:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_377;
        }
        goto fatbar_377_over;
        fatbar_377:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 23554;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L237; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L237:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L238; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto FUN_read_atom_1268;
          L238:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          switch (get_case_noint (r1)) {
            default: {
              r1 = allocate (TC_TUPLE, 3);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][2];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r2 = UOBJ_GET(r2,2);
              r1[3] = r2;
              r1 = (object *) 3;
              r2 = ((object*) lenv) [3];
              if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = (object *) 1;
                r3 = ((object**) lenv) [1][2];
                range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
                r2 = (object *) 1;
                r3 = ((object**) lenv) [1][2];
                r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
              } else {
                goto fatbar_380;
              }
              goto fatbar_380_over;
              fatbar_380:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1033;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L239; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L239:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 2;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_381;
              }
              goto fatbar_381_over;
              fatbar_381:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1034;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L240; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L240:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 65794;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_382;
              }
              goto fatbar_382_over;
              fatbar_382:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1035;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L241; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L241:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 2306;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_383;
              }
              goto fatbar_383_over;
              fatbar_383:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1036;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L242; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L242:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 3330;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_384;
              }
              goto fatbar_384_over;
              fatbar_384:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1037;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L243; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L243:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 8194;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_385;
              }
              goto fatbar_385_over;
              fatbar_385:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1038;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L244; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L244:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 2562;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                lenv = ((object *)lenv)[1];
                goto fatbar_379;
              }
              fatbar_385_over: ;
              fatbar_384_over: ;
              fatbar_383_over: ;
              fatbar_382_over: ;
              fatbar_381_over: ;
              fatbar_380_over: ;
              lenv = lenv[1];
            } break;
          }
          goto fatbar_379_over;
          fatbar_379:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_1039;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L245; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L245:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_379_over: ;
          lenv = lenv[1];
        } else {
          goto fatbar_378;
        }
        goto fatbar_378_over;
        fatbar_378:
        r1 = allocate (TC_TUPLE, 3);
        r2 = (object*) &constructed_1040;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[15];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L246; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
        L246:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        fatbar_378_over: ;
        fatbar_377_over: ;
        fatbar_376_over: ;
        fatbar_375_over: ;
        fatbar_374_over: ;
        fatbar_373_over: ;
        fatbar_372_over: ;
        fatbar_371_over: ;
        fatbar_370_over: ;
        fatbar_369_over: ;
        fatbar_368_over: ;
        fatbar_367_over: ;
      } else {
        goto fatbar_366;
      }
      goto fatbar_366_over;
      fatbar_366:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 14850;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L247; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L247:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 3);
        r1[1] = lenv; lenv = r1;
        r2 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L248; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto FUN_read_atom_1268;
        L248:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        switch (get_case_noint (r2)) {
          default: {
            r2 = allocate (TC_TUPLE, 4);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][2];
            r3 = UOBJ_GET(r3,0);
            r2[2] = r3;
            r3 = ((object**) lenv) [1][2];
            r3 = UOBJ_GET(r3,1);
            r2[3] = r3;
            r3 = ((object**) lenv) [1][2];
            r3 = UOBJ_GET(r3,3);
            r2[4] = r3;
            r2 = ((object*) lenv) [3];
            switch (get_case (r2)) {
              case ((pxll_int)PXLL_TRUE): {
                r2 = allocate (TC_TUPLE, 3);
                r3 = (object*) &constructed_1041;
                r2[2] = r3;
                r3 = ((object*) lenv) [2];
                r2[3] = r3;
                r3 = top[15];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L249; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_error1_293;
                L249:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
              } break;
              default: {
                r2 = (object *) 1;
                r3 = ((object*) lenv) [4];
                if PXLL_IS_TRUE(PXLL_TEST(r3==r2)) {
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object*) lenv) [2];
                  r2[2] = r3;
                  r3 = top[80];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L250; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_string__3esymbol_419;
                  L250:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                } else {
                  goto fatbar_387;
                }
                goto fatbar_387_over;
                fatbar_387:
                r2 = allocate (TC_TUPLE, 3);
                r3 = (object*) &constructed_1042;
                r2[2] = r3;
                r3 = ((object*) lenv) [2];
                r2[3] = r3;
                r3 = top[15];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L251; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_error1_293;
                L251:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                fatbar_387_over: ;
              } break;
            }
            lenv = lenv[1];
          } break;
        }
        r1[3] = r2;
        r1 = ((object*) lenv) [3];
        r2 = (object *) constructed_49[0];
        t = alloc_no_clear (UOTAG(9),2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_386;
      }
      goto fatbar_386_over;
      fatbar_386:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 11266;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L252; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L252:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = (object *) constructed_1044[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L253; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto FUN_read_1272;
        L253:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_388;
      }
      goto fatbar_388_over;
      fatbar_388:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 9986;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L254; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L254:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = (object *) constructed_870[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L255; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto FUN_read_1272;
        L255:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_389;
      }
      goto fatbar_389_over;
      fatbar_389:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 8706;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L256; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L256:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def loop_1328_i365_i439
        goto L257;
        FUN_loop_1328_i365_i439:
          check_heap (0);
          r0 = (object *) 23554;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L258; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L258:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L259; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L259:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            ((object*) lenv) [2] = r0;
            r0 = (object *) 23554;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L260; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L260:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 23554;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_392;
            }
            PXLL_RETURN(0);
            goto fatbar_392_over;
            fatbar_392:
            r0 = (object *) 8706;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L261; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L261:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 8706;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_393;
            }
            PXLL_RETURN(0);
            goto fatbar_393_over;
            fatbar_393:
            r0 = (object *) 29698;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L262; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L262:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 2306;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_394;
            }
            PXLL_RETURN(0);
            goto fatbar_394_over;
            fatbar_394:
            r0 = (object *) 28162;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L263; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L263:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 2562;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_395;
            }
            PXLL_RETURN(0);
            goto fatbar_395_over;
            fatbar_395:
            r0 = (object *) 29186;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L264; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L264:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 3330;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_396;
            }
            PXLL_RETURN(0);
            goto fatbar_396_over;
            fatbar_396:
            r0 = (object *) 22530;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L265; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L265:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object******) lenv) [1][1][1][1][1][6];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L266; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto FUN_read_hex_code_1275;
              L266:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = ((object**) lenv) [1][3];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_397;
            }
            PXLL_RETURN(0);
            goto fatbar_397_over;
            fatbar_397:
            r0 = (object *) 30722;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L267; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L267:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][2] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][2];
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object******) lenv) [1][1][1][1][1][6];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L268; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto FUN_read_hex_code_1275;
              L268:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = ((object**) lenv) [1][3];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439;
            } else {
              goto fatbar_398;
            }
            PXLL_RETURN(0);
            goto fatbar_398_over;
            fatbar_398:
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_1046;
            r0[2] = r1;
            r1 = ((object*) lenv) [3];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            fatbar_398_over: ;
            PXLL_RETURN(0);
            fatbar_397_over: ;
            PXLL_RETURN(0);
            fatbar_396_over: ;
            PXLL_RETURN(0);
            fatbar_395_over: ;
            PXLL_RETURN(0);
            fatbar_394_over: ;
            PXLL_RETURN(0);
            fatbar_393_over: ;
            PXLL_RETURN(0);
            fatbar_392_over: ;
            PXLL_RETURN(0);
          } else {
            goto fatbar_391;
          }
          PXLL_RETURN(0);
          goto fatbar_391_over;
          fatbar_391:
          r0 = (object *) 8706;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L269; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L269:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L270; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L270:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = top[54];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L271; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3estring_361;
            L271:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(2),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            goto fatbar_399;
          }
          PXLL_RETURN(0);
          goto fatbar_399_over;
          fatbar_399:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r0[2] = r1;
          r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L272; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L272:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
          r0 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L273; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L273:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object******) lenv) [1][1][1][1][1][2] = r0;
          } else {
            r0 = (object *) 14;
          }
          r0 = ((object******) lenv) [1][1][1][1][1][2];
          r1 = ((object*) lenv) [3];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv[2] = r0;
          lenv[3] = r1;
          goto FUN_loop_1328_i365_i439;
          fatbar_399_over: ;
          PXLL_RETURN(0);
          fatbar_391_over: ;
          PXLL_RETURN(0);
        L257:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_loop_1328_i365_i439; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 3);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object******) lenv) [1][1][1][1][1][2];
        r2[2] = r3;
        r2 = (object *) 65794;
        r3 = ((object*) lenv) [2];
        r2 = PXLL_TEST(r3==r2);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r2) {
          r2 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L274; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L274:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          ((object*****) lenv) [1][1][1][1][2] = r2;
        } else {
          r2 = (object *) 14;
        }
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r2 = (object *) constructed_1045[0];
        r1[3] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L275; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_loop_1328_i365_i439;
        L275:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
      } else {
        goto fatbar_390;
      }
      goto fatbar_390_over;
      fatbar_390:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 31490;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L276; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L276:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def loop_1335_i399_i439
        goto L277;
        FUN_loop_1335_i399_i439:
          check_heap (0);
          r0 = ((object*****) lenv) [1][1][1][1][2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L278; ; k = t;
          lenv = r0[2]; goto FUN_skip_whitespace_1267;
          L278:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
          r1[2] = r2;
          r1 = (object *) 65794;
          r2 = ((object*) lenv) [2];
          r1 = PXLL_TEST(r2==r1);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r1) {
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L279; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L279:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            ((object*******) lenv) [1][1][1][1][1][1][2] = r1;
          } else {
            r1 = (object *) 14;
          }
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r0[2] = r1;
          r0 = (object *) 32002;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L280; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L280:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object********) lenv) [1][1][1][1][1][1][1][2] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][2];
            r1[2] = r2;
            r2 = (object *) constructed_36[0];
            r1[3] = r2;
            r2 = top[21];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L281; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
            L281:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(8),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            // def loop_1340_i406_i439
            goto L282;
            FUN_loop_1340_i406_i439:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object*) lenv) [3];
              r0[2] = r1;
              r1 = top[179];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L283; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L283:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 262;
              } else {
                r0 = (object *) 11778;
                r1 = ((object*) lenv) [3];
                r0 = PXLL_TEST(r1==r0);
              }
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                r0[2] = r1;
                r0 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L284; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L284:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2] = r0;
                r0 = ((object*) lenv) [2];
                lenv = lenv[1];
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                r0[2] = r1;
                r0 = (object *) 65794;
                r1 = ((object*) lenv) [2];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L285; ; k = t;
                  lenv = r0[2]; goto *r0[1];
                  L285:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2] = r0;
                } else {
                  r0 = (object *) 14;
                }
                r0 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                r1 = ((object*) lenv) [2];
                r2 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r2 = (object *) 6;
                lenv[2] = r1;
                lenv[3] = r0;
                lenv[4] = r2;
                goto FUN_loop_1340_i406_i439;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object*) lenv) [2];
                r2[2] = r3;
                r3 = (object *) constructed_36[0];
                r2[3] = r3;
                r3 = top[21];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L286; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                L286:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = top[54];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L287; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_list__3estring_361;
                L287:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = top[80];
                r0[1] = r1[2]; lenv = r0; goto FUN_string__3esymbol_419;
              }
              PXLL_RETURN(0);
            L282:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_loop_1340_i406_i439; r2[2] = lenv;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 4);
            r2 = (object *) constructed_1049[0];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
            r2[2] = r3;
            r2 = (object *) 65794;
            r3 = ((object*) lenv) [2];
            r2 = PXLL_TEST(r3==r2);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r2) {
              r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L288; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto *r2[1];
              L288:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              ((object*********) lenv) [1][1][1][1][1][1][1][1][2] = r2;
            } else {
              r2 = (object *) 14;
            }
            r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
            r1[3] = r2;
            r2 = (object *) 6;
            r1[4] = r2;
            r2 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L289; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_loop_1340_i406_i439;
            L289:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r0[2] = r1;
            r0 = (object *) constructed_66[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = (object *) 6;
              t = alloc_no_clear (UOTAG(4),1);
              t[1] = r1;
              r1 = t;
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(0),2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[3] = r1;
              r0 = ((object****) lenv) [1][1][1][2];
              r1 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              lenv = ((object **)lenv)[1][1];
              lenv[2] = r0;
              goto FUN_loop_1335_i399_i439;
            } else {
              r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L290; ; k = t;
              lenv = r0[2]; goto FUN_skip_whitespace_1267;
              L290:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
              r1[2] = r2;
              r1 = (object *) 65794;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(r2==r1);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L291; t[4] = r0; k = t;
                lenv = r1[2]; goto *r1[1];
                L291:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                ((object*********) lenv) [1][1][1][1][1][1][1][1][2] = r1;
              } else {
                r1 = (object *) 14;
              }
              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
              r0[2] = r1;
              r1 = (object *) 15618;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(r2==r1);
              r0[3] = r1;
              r0 = (object *) 6;
              r1 = ((object*) lenv) [3];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object*) &constructed_1048;
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][2];
                r0[3] = r1;
                r1 = top[15];
                r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                r0[2] = r1;
                r0 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L292; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L292:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object*********) lenv) [1][1][1][1][1][1][1][1][2] = r0;
                r0 = ((object*) lenv) [2];
                lenv = lenv[1];
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L293; t[4] = r0; k = t;
                lenv = r1[2]; goto FUN_read_1272;
                L293:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r2 = ((object***) lenv) [1][1][2];
                t = alloc_no_clear (UOTAG(0),2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[2] = r1;
                r0 = ((object*****) lenv) [1][1][1][1][2];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                lenv = ((object ***)lenv)[1][1][1];
                lenv[2] = r0;
                goto FUN_loop_1335_i399_i439;
              }
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        L277:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_loop_1335_i399_i439; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 2);
        r2 = (object *) constructed_1047[0];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L294; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_loop_1335_i399_i439;
        L294:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
      } else {
        goto fatbar_400;
      }
      goto fatbar_400_over;
      fatbar_400:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 10242;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object****) lenv) [1][1][1][7];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L295; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto FUN_read_list_1277;
        L295:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_401;
      }
      goto fatbar_401_over;
      fatbar_401:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = (object*) &constructed_68;
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r2 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
        r1 = (object*) &constructed_1050;
        dump_object (r1, 0); fprintf (stdout, "\n");
        r1 = (object*) &constructed_1050;
        PXLL_UNDEFINED; result=r1; goto Lreturn;
      } else {
        goto fatbar_402;
      }
      goto fatbar_402_over;
      fatbar_402:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][3];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L296; t[4] = r0; t[5] = r1; k = t;
      lenv = r2[2]; goto FUN_read_atom_1268;
      L296:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r1 = ((object*) lenv) [2];
      switch (get_case_noint (r1)) {
        default: {
          r1 = allocate (TC_TUPLE, 5);
          r1[1] = lenv; lenv = r1;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,1);
          r1[3] = r2;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,2);
          r1[4] = r2;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,3);
          r1[5] = r2;
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case ((pxll_int)PXLL_TRUE): {
              r1 = allocate (TC_TUPLE, 5);
              r1[1] = lenv; lenv = r1;
              r2 = (object *) 1;
              r3 = ((object**) lenv) [1][2];
              range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
              r2 = (object *) 1;
              r3 = ((object**) lenv) [1][2];
              r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
              r1[2] = r2;
              r2 = (object *) 11522;
              r3 = ((object*) lenv) [2];
              r2 = PXLL_TEST(r3==r2);
              r1[3] = r2;
              r2 = ((object*) lenv) [3];
              if PXLL_IS_TRUE(r2) {
                r2 = (object *) 3;
              } else {
                r2 = (object *) 1;
              }
              r1[4] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r2[1] = lenv; lenv = r2;
              // def loop_1348_i427_i439
              goto L297;
              FUN_loop_1348_i427_i439:
                check_heap (0);
                r0 = ((object****) lenv) [1][1][1][4];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                  r0 = ((object***) lenv) [1][1][3];
                  if PXLL_IS_TRUE(r0) {
                    r0 = ((object*) lenv) [3];
                    r1 = (object *) 1;
                    r0 = box((pxll_int)unbox(r1)-unbox(r0));
                    PXLL_RETURN(0);
                  } else {
                    r0 = ((object*) lenv) [3];
                    PXLL_RETURN(0);
                  }
                  PXLL_RETURN(0);
                } else {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[169];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r3 = ((object*****) lenv) [1][1][1][1][2];
                  range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
                  r2 = ((object**) lenv) [1][2];
                  r3 = ((object*****) lenv) [1][1][1][1][2];
                  r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
                  r1[3] = r2;
                  r2 = top[65];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L298; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
                  L298:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UITAG(1)): {
                      r0 = allocate (TC_TUPLE, 3);
                      r1 = (object*) &constructed_1051;
                      r0[2] = r1;
                      r1 = ((object*****) lenv) [1][1][1][1][2];
                      r0[3] = r1;
                      r1 = top[15];
                      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                    } break;
                    default: {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = (object *) 21;
                      r2 = ((object****) lenv) [1][1][1][3];
                      r1 = box((pxll_int)unbox(r2)*unbox(r1));
                      r0[2] = r1;
                      r0 = ((object**) lenv) [1][2];
                      r1 = ((object*) lenv) [2];
                      r0 = box((pxll_int)unbox(r1)+unbox(r0));
                      lenv = lenv[1];
                      r1 = (object *) 3;
                      r2 = ((object***) lenv) [1][1][2];
                      r1 = box((pxll_int)unbox(r2)+unbox(r1));
                      lenv = ((object **)lenv)[1][1];
                      lenv[2] = r1;
                      lenv[3] = r0;
                      goto FUN_loop_1348_i427_i439;
                    } break;
                  }
                  PXLL_RETURN(0);
                }
                PXLL_RETURN(0);
              L297:
              r3 = allocate (TC_CLOSURE, 2);
              r3[1] = &&FUN_loop_1348_i427_i439; r3[2] = lenv;
              r2[2] = r3;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][4];
              r2[2] = r3;
              r3 = (object *) 1;
              r2[3] = r3;
              r3 = ((object*) lenv) [2];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L299; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_loop_1348_i427_i439;
              L299:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              lenv = lenv[1];
              r1[5] = r2;
              r1 = ((object*) lenv) [5];
              t = alloc_no_clear (UOTAG(5),1);
              t[1] = r1;
              r1 = t;
              lenv = lenv[1];
            } break;
            default: {
              r1 = (object *) 1;
              r2 = ((object*) lenv) [5];
              if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object**) lenv) [1][2];
                r2[2] = r3;
                r3 = top[80];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L300; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string__3esymbol_419;
                L300:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
              } else {
                goto fatbar_403;
              }
              goto fatbar_403_over;
              fatbar_403:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][2];
              r2 = box((pxll_int)((pxll_string*)(r2))->len);
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][5];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object**) lenv) [1][2];
                r2[2] = r3;
                r3 = top[80];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L301; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string__3esymbol_419;
                L301:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
              } else {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                // def loop_1297_i438_i439
                goto L302;
                FUN_loop_1297_i438_i439:
                  check_heap (0);
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_NIL): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = ((object**) lenv) [1][2];
                          r1[2] = r2;
                          r2 = top[80];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L303; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
                          L303:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r0 = ((object*) lenv) [2];
                          t = alloc_no_clear (UOTAG(1),1);
                          t[1] = r0;
                          r0 = t;
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          goto fatbar_404;
                        }
                      }
                      PXLL_RETURN(0);
                      goto fatbar_404_over;
                      fatbar_404:
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][3];
                      r1[2] = r2;
                      r2 = ((object****) lenv) [1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L304; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_loop_1297_i438_i439;
                      L304:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][2];
                      r1[2] = r2;
                      r2 = top[80];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L305; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
                      L305:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      r1 = ((object*) lenv) [2];
                      t = alloc_no_clear (UOTAG(10),2);
                      t[1] = r1;
                      t[2] = r0;
                      r0 = t;
                      PXLL_RETURN(0);
                      fatbar_404_over: ;
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      r0 = (object*) &constructed_68;
                      r0 = box((pxll_int)((pxll_string*)(r0))->len);
                      r1 = (object*) &constructed_68;
                      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                      r0 = (object*) &constructed_67;
                      dump_object (r0, 0); fprintf (stdout, "\n");
                      r0 = (object*) &constructed_67;
                      PXLL_UNDEFINED; result=r0; goto Lreturn;
                      PXLL_RETURN(0);
                    } break;
                  }
                  PXLL_RETURN(0);
                L302:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_loop_1297_i438_i439; r2[2] = lenv;
                r1[2] = r2;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = allocate (TC_TUPLE, 3);
                r4 = ((object***) lenv) [1][1][2];
                r3[2] = r4;
                r4 = (object *) 11778;
                r3[3] = r4;
                r4 = top[49];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L306; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_string_split_354;
                L306:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[2] = r3;
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object*) lenv) [2];
                r2[2] = r3;
                r3 = (object *) constructed_36[0];
                r2[3] = r3;
                r3 = top[21];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L307; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                L307:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L308; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_loop_1297_i438_i439;
                L308:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                lenv = lenv[1];
              }
              fatbar_403_over: ;
            } break;
          }
          lenv = lenv[1];
        } break;
      }
      lenv = lenv[1];
      fatbar_402_over: ;
      fatbar_401_over: ;
      fatbar_400_over: ;
      fatbar_390_over: ;
      fatbar_389_over: ;
      fatbar_388_over: ;
      fatbar_386_over: ;
      fatbar_366_over: ;
      fatbar_365_over: ;
      lenv = lenv[1];
      r0[2] = r1;
      r0 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L309; ; k = t;
      lenv = r0[2]; goto FUN_skip_whitespace_1267;
      L309:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L310; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L310:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][2];
      r0[2] = r1;
      r0 = (object *) 14850;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r0[2] = r1;
        r0 = ((object******) lenv) [1][1][1][1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L311; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L311:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object*****) lenv) [1][1][1][1][2] = r0;
        r0 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L312; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_read_1272;
        L312:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object***) lenv) [1][1][2];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (UOTAG(9),2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_362;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_362;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_362_over;
        fatbar_362:
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(9)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = (object *) constructed_49[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 7);
              r0[1] = lenv; lenv = r0;
              r1 = (object *) constructed_654[0];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r1;
              r1 = t;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r1;
              r1 = t;
              r0[3] = r1;
              r1 = (object*)TC_NIL;
              r0[4] = r1;
              r1 = ((object*) lenv) [4];
              r2 = ((object*****) lenv) [1][1][1][1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[5] = r1;
              r1 = ((object*) lenv) [5];
              r2 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[6] = r1;
              r1 = ((object*) lenv) [6];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[7] = r1;
              r0 = ((object*) lenv) [7];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r0;
              r0 = t;
              PXLL_RETURN(0);
            } else {
              lenv = ((object *)lenv)[1];
              goto fatbar_363;
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_363;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_363_over;
        fatbar_363:
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_1030;
        r0[2] = r1;
        r1 = allocate (UOTAG(1), 2);
        r2 = ((object***) lenv) [1][1][2];
        r1[1] = r2;
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        fatbar_363_over: ;
        PXLL_RETURN(0);
        fatbar_362_over: ;
        PXLL_RETURN(0);
      } else {
        goto fatbar_361;
      }
      PXLL_RETURN(0);
      goto fatbar_361_over;
      fatbar_361:
      r0 = (object *) 23298;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 9);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object******) lenv) [1][1][1][1][1][2];
        r1[2] = r2;
        r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L313; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L313:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object******) lenv) [1][1][1][1][1][2] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L314; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto FUN_read_1272;
        L314:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L315; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_skip_whitespace_1267;
        L315:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object********) lenv) [1][1][1][1][1][1][1][2];
        r2[2] = r3;
        r2 = (object *) 65794;
        r3 = ((object*) lenv) [2];
        r2 = PXLL_TEST(r3==r2);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r2) {
          r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L316; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L316:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          ((object*******) lenv) [1][1][1][1][1][1][2] = r2;
        } else {
          r2 = (object *) 14;
        }
        r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
        r1[2] = r2;
        r1 = (object *) 23810;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r1[2] = r2;
          r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L317; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L317:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*******) lenv) [1][1][1][1][1][1][2] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = ((object*) lenv) [2];
        } else {
          r1 = (object*) &constructed_68;
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r2 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
          r1 = (object*) &constructed_1031;
          dump_object (r1, 0); fprintf (stdout, "\n");
          r1 = (object*) &constructed_1031;
          PXLL_UNDEFINED; result=r1; goto Lreturn;
        }
        lenv = lenv[1];
        r0[2] = r1;
        r1 = (object *) constructed_295[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[3] = r1;
        r1 = (object *) 6;
        t = alloc_no_clear (UOTAG(4),1);
        t[1] = r1;
        r1 = t;
        r0[4] = r1;
        r1 = (object*)TC_NIL;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[6] = r1;
        r1 = ((object*) lenv) [6];
        r2 = ((object***) lenv) [1][1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[7] = r1;
        r1 = ((object*) lenv) [7];
        r2 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[8] = r1;
        r1 = ((object*) lenv) [8];
        r2 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[9] = r1;
        r0 = ((object*) lenv) [9];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } else {
        goto fatbar_364;
      }
      PXLL_RETURN(0);
      goto fatbar_364_over;
      fatbar_364:
      r0 = ((object**) lenv) [1][2];
      PXLL_RETURN(0);
      fatbar_364_over: ;
      PXLL_RETURN(0);
      fatbar_361_over: ;
      PXLL_RETURN(0);
    L217:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_1272; r1[2] = lenv;
    r0[4] = r1;
    // def read-hex-digit_1274
    goto L318;
    FUN_read_hex_digit_1274:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = top[168];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = top[65];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L319; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
      L319:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(1)): {
          r0 = (object*) &constructed_68;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_1029;
          dump_object (r0, 0); fprintf (stdout, "\n");
          r0 = (object*) &constructed_1029;
          PXLL_UNDEFINED; result=r0; goto Lreturn;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L318:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_digit_1274; r1[2] = lenv;
    r0[5] = r1;
    // def read-hex-code_1275
    goto L320;
    FUN_read_hex_code_1275:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 5);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object****) lenv) [1][1][1][2];
      r2[2] = r3;
      r2 = ((object*****) lenv) [1][1][1][1][2];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L321; t[4] = r0; t[5] = r1; k = t;
      lenv = r2[2]; goto *r2[1];
      L321:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      ((object****) lenv) [1][1][1][2] = r2;
      r2 = ((object*) lenv) [2];
      lenv = lenv[1];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L322; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_read_hex_digit_1274;
      L322:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object****) lenv) [1][1][1][2];
      r2[2] = r3;
      r2 = ((object*****) lenv) [1][1][1][1][2];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L323; t[4] = r0; t[5] = r1; k = t;
      lenv = r2[2]; goto *r2[1];
      L323:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      ((object****) lenv) [1][1][1][2] = r2;
      r2 = ((object*) lenv) [2];
      lenv = lenv[1];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L324; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_read_hex_digit_1274;
      L324:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = (object *) 17;
      r2 = ((object*) lenv) [2];
      r1 = box((pxll_int)unbox(r2)<<unbox(r1));
      r0[4] = r1;
      r1 = ((object*) lenv) [3];
      r2 = ((object*) lenv) [4];
      r1 = box((pxll_int)unbox(r2)+unbox(r1));
      r0[5] = r1;
      r0 = ((object*) lenv) [5];
      r0 = TO_CHAR(unbox(r0));
      PXLL_RETURN(0);
    L320:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_code_1275; r1[2] = lenv;
    r0[6] = r1;
    // def read-list_1277
    goto L325;
    FUN_read_list_1277:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r0 = ((object****) lenv) [1][1][1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L326; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L326:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      ((object***) lenv) [1][1][2] = r0;
      r0 = ((object*) lenv) [2];
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1331
      goto L327;
      FUN_loop_1331:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L328; ; k = t;
        lenv = r0[2]; goto FUN_skip_whitespace_1267;
        L328:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object******) lenv) [1][1][1][1][1][2];
        r1[2] = r2;
        r1 = (object *) 65794;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L329; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L329:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][2] = r1;
        } else {
          r1 = (object *) 14;
        }
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 10498;
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          r0[2] = r1;
          r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L330; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L330:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object******) lenv) [1][1][1][1][1][2] = r0;
          r0 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*****) lenv) [1][1][1][1][4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L331; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_read_1272;
          L331:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object***) lenv) [1][1][2];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = ((object **)lenv)[1][1];
          lenv[2] = r0;
          goto FUN_loop_1331;
        }
        PXLL_RETURN(0);
      L327:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1331; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) constructed_1028[0];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1331;
    L325:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_list_1277; r1[2] = lenv;
    r0[7] = r1;
    // def read-hex-int_1281
    goto L332;
    FUN_read_hex_int_1281:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L333; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L333:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object***) lenv) [1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r1 = (object *) 11522;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L334; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L334:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][2] = r0;
        r0 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0 = (object *) 14;
      } else {
        r0 = (object *) 14;
      }
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1354
      goto L335;
      FUN_loop_1354:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[168];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[65];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L336; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
        L336:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L337; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L337:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object********) lenv) [1][1][1][1][1][1][1][2] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = (object *) 65794;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L338; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L338:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) 33;
            r3 = ((object****) lenv) [1][1][1][2];
            r2 = box((pxll_int)unbox(r3)*unbox(r2));
            r1[2] = r2;
            r1 = ((object**) lenv) [1][2];
            r2 = ((object*) lenv) [2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv = lenv[1];
            lenv = ((object **)lenv)[1][1];
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_1354;
          } break;
          default: {
            r0 = ((object****) lenv) [1][1][1][3];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object**) lenv) [1][2];
              r1 = (object *) 1;
              r0 = box((pxll_int)unbox(r1)-unbox(r0));
              PXLL_RETURN(0);
            } else {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      L335:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1354; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L339; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L339:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][2];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1354;
    L332:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_int_1281; r1[2] = lenv;
    r0[8] = r1;
    // def read-oct-int_1282
    goto L340;
    FUN_read_oct_int_1282:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L341; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L341:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object***) lenv) [1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r1 = (object *) 11522;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L342; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L342:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][2] = r0;
        r0 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0 = (object *) 14;
      } else {
        r0 = (object *) 14;
      }
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1360
      goto L343;
      FUN_loop_1360:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[170];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[65];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L344; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
        L344:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L345; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L345:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object********) lenv) [1][1][1][1][1][1][1][2] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = (object *) 65794;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L346; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L346:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) 17;
            r3 = ((object****) lenv) [1][1][1][2];
            r2 = box((pxll_int)unbox(r3)*unbox(r2));
            r1[2] = r2;
            r1 = ((object**) lenv) [1][2];
            r2 = ((object*) lenv) [2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv = lenv[1];
            lenv = ((object **)lenv)[1][1];
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_1360;
          } break;
          default: {
            r0 = ((object****) lenv) [1][1][1][3];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object**) lenv) [1][2];
              r1 = (object *) 1;
              r0 = box((pxll_int)unbox(r1)-unbox(r0));
              PXLL_RETURN(0);
            } else {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      L343:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1360; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L347; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L347:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][2];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1360;
    L340:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_oct_int_1282; r1[2] = lenv;
    r0[9] = r1;
    // def loop_1367_i575
    goto L348;
    FUN_loop_1367_i575:
      check_heap (0);
      r0 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L349; ; k = t;
      lenv = r0[2]; goto FUN_skip_whitespace_1267;
      L349:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L350; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L350:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][2] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][2];
      r0[2] = r1;
      r0 = (object *) 65794;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(r1==r0);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = (object *) constructed_36[0];
        r0[3] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L351; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_read_1272;
        L351:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(1)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = (object *) constructed_1027[0];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                      r0 = ((object**) lenv) [1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (UOTAG(2)): {
                              r0 = allocate (TC_TUPLE, 2);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r0 = ((object**) lenv) [1][3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r2 = top[85];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L352; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_read_file_425;
                                  L352:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object*) lenv) [2];
                                  r1[2] = r2;
                                  r2 = (object *) constructed_36[0];
                                  r1[3] = r2;
                                  r2 = top[21];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L353; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                  L353:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object*) lenv) [3];
                                  r1[2] = r2;
                                  r2 = (object *) constructed_36[0];
                                  r1[3] = r2;
                                  r2 = top[21];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L354; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                  L354:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                  r0[3] = r1;
                                  r1 = top[21];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L355; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                                  L355:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = result;
                                  lenv = lenv[1];
                                  lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                  lenv[2] = r0;
                                  goto FUN_loop_1367_i575;
                                } break;
                                default: {
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  goto fatbar_360;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_360;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_360;
                        }
                      }
                      PXLL_RETURN(0);
                    } else {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_360;
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_360;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_360;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_360;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_360_over;
        fatbar_360:
        r0 = ((object**) lenv) [1][2];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = ((object *)lenv)[1];
        lenv[2] = r0;
        goto FUN_loop_1367_i575;
        fatbar_360_over: ;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L348:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1367_i575; r1[2] = lenv;
    r0[10] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object *) constructed_1025[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [10];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_1367_i575;
  L189:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_reader_424; r1[2] = lenv;
  r0[84] = r1;
  // def read-file_425
  goto L356;
  FUN_read_file_425:
    check_heap (0);
    r0 = (object*) &constructed_1054;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_1054;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = ((object*) lenv) [2];
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[61];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L357; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_file_2fopen_read_380;
    L357:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    // def lambda_6368
    goto L358;
    FUN_lambda_6368:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = top[64];
      r0[1] = r1[2]; lenv = r0; goto FUN_file_2fread_char_387;
    L358:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_6368; r1[2] = lenv;
    r0[3] = r1;
    r1 = (object *) 65794;
    r0[4] = r1;
    r0 = allocate (TC_TUPLE, 10);
    r0[1] = lenv; lenv = r0;
    // def skip-whitespace_1267_i0
    goto L359;
    FUN_skip_whitespace_1267_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1288_i0
      goto L360;
      FUN_loop_1288_i0:
        check_heap (0);
        r0 = (object *) 65794;
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) 15106;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def loop_1286_i256_i0
            goto L361;
            FUN_loop_1286_i256_i0:
              check_heap (0);
              r0 = (object *) 2562;
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                goto fatbar_358;
              }
              PXLL_RETURN(0);
              goto fatbar_358_over;
              fatbar_358:
              r0 = (object *) 3330;
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                goto fatbar_359;
              }
              PXLL_RETURN(0);
              goto fatbar_359_over;
              fatbar_359:
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L362; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L362:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
              r0 = ((object*) lenv) [2];
              lenv = lenv[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L363; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L363:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              lenv[2] = r0;
              goto FUN_loop_1286_i256_i0;
              fatbar_359_over: ;
              PXLL_RETURN(0);
              fatbar_358_over: ;
              PXLL_RETURN(0);
            L361:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_loop_1286_i256_i0; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object******) lenv) [1][1][1][1][1][4];
            r1[2] = r2;
            r1 = ((object******) lenv) [1][1][1][1][1][3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L364; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L364:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            ((object******) lenv) [1][1][1][1][1][4] = r1;
            r1 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L365; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L365:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*****) lenv) [1][1][1][1][4];
            r0[2] = r1;
            r0 = (object *) 65794;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object****) lenv) [1][1][1][3];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L366; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L366:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object****) lenv) [1][1][1][4] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = ((object****) lenv) [1][1][1][4];
            lenv[2] = r0;
            goto FUN_loop_1288_i0;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = top[176];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L367; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L367:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*****) lenv) [1][1][1][1][4];
              r0[2] = r1;
              r0 = ((object*****) lenv) [1][1][1][1][3];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L368; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L368:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*****) lenv) [1][1][1][1][4] = r0;
              r0 = ((object*) lenv) [2];
              lenv = lenv[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*****) lenv) [1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object****) lenv) [1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L369; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L369:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object****) lenv) [1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object****) lenv) [1][1][1][4];
              lenv[2] = r0;
              goto FUN_loop_1288_i0;
            } else {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L360:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1288_i0; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object***) lenv) [1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L370; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L370:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object***) lenv) [1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object***) lenv) [1][1][4];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L359:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_skip_whitespace_1267_i0; r1[2] = lenv;
    r0[2] = r1;
    // def read-atom_1268_i0
    goto L371;
    FUN_read_atom_1268_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = (object *) 1;
      r0[3] = r1;
      r1 = (object *) 12290;
      r0[4] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1293_i0
      goto L372;
      FUN_loop_1293_i0:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object******) lenv) [1][1][1][1][1][4];
        r0[2] = r1;
        r0 = (object *) 65794;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L373; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L373:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object*****) lenv) [1][1][1][1][4] = r0;
        } else {
          r0 = (object *) 14;
        }
        r0 = ((object*****) lenv) [1][1][1][1][4];
        ((object***) lenv) [1][1][4] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][4];
        r0[2] = r1;
        r0 = (object *) 11778;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
        } else {
          r0 = (object *) 14;
        }
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 7;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 11;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r1 = top[177];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L374; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L374:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 11;
            } else {
              r0 = (object *) 7;
            }
          }
        } else {
          goto fatbar_354;
        }
        goto fatbar_354_over;
        fatbar_354:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 5;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 13;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r1 = top[177];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L375; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L375:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 13;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object***) lenv) [1][1][4];
              r0[2] = r1;
              r1 = top[178];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L376; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L376:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 5;
              } else {
                r0 = (object *) 7;
              }
            }
          }
        } else {
          goto fatbar_355;
        }
        goto fatbar_355_over;
        fatbar_355:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 11;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r1 = top[177];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L377; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L377:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 11;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object***) lenv) [1][1][4];
              r0[2] = r1;
              r1 = top[178];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L378; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L378:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 5;
              } else {
                r0 = (object *) 7;
              }
            }
          }
        } else {
          goto fatbar_356;
        }
        goto fatbar_356_over;
        fatbar_356:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 1;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 9;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][4];
            r0[2] = r1;
            r0 = (object *) 11522;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = (object *) 3;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object***) lenv) [1][1][4];
              r0[2] = r1;
              r1 = top[178];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L379; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L379:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 5;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object***) lenv) [1][1][4];
                r0[2] = r1;
                r1 = top[177];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L380; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L380:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                if PXLL_IS_TRUE(r0) {
                  r0 = (object *) 15;
                } else {
                  r0 = (object *) 7;
                }
              }
            }
          }
        } else {
          goto fatbar_357;
        }
        goto fatbar_357_over;
        fatbar_357:
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_67;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_67;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        fatbar_357_over: ;
        fatbar_356_over: ;
        fatbar_355_over: ;
        fatbar_354_over: ;
        ((object***) lenv) [1][1][2] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[2] = r1;
        r0 = (object *) 9;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(unbox(r1)<unbox(r0));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][4];
          r0[2] = r1;
          r1 = ((object******) lenv) [1][1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L381; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L381:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object******) lenv) [1][1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[4] = r1;
          r0 = ((object*) lenv) [4];
          r1 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          lenv[2] = r0;
          goto FUN_loop_1293_i0;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[2] = r1;
          r0 = (object *) 9;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)==unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object*) &constructed_68;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
            r0 = (object*) &constructed_1053;
            dump_object (r0, 0); fprintf (stdout, "\n");
            r0 = (object*) &constructed_1053;
            PXLL_UNDEFINED; result=r0; goto Lreturn;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][2];
            r0[2] = r1;
            r0 = (object *) 15;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(unbox(r1)==unbox(r0));
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = ((object******) lenv) [1][1][1][1][1][4];
              r0[2] = r1;
              r1 = ((object******) lenv) [1][1][1][1][1][3];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L382; t[4] = r0; k = t;
              lenv = r1[2]; goto *r1[1];
              L382:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              ((object******) lenv) [1][1][1][1][1][4] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[4] = r1;
              r0 = ((object*) lenv) [4];
              r1 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              ((object*) lenv) [2] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = allocate (UOTAG(0), 4);
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L383; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L383:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[54];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L384; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3estring_361;
            L384:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[1] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object****) lenv) [1][1][1][2];
            r1[2] = r2;
            r1 = (object *) 13;
            r2 = ((object*) lenv) [2];
            r1 = PXLL_TEST(unbox(r2)==unbox(r1));
            lenv = lenv[1];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L385; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L385:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][3];
            r0[4] = r1;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L372:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1293_i0; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) constructed_1052[0];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L371:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_atom_1268_i0; r1[2] = lenv;
    r0[3] = r1;
    // def read_1272_i0
    goto L386;
    FUN_read_1272_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L387; t[4] = r0; k = t;
      lenv = r1[2]; goto *r1[1];
      L387:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object*****) lenv) [1][1][1][1][4];
      r2[2] = r3;
      r2 = (object *) 65794;
      r3 = ((object*) lenv) [2];
      r2 = PXLL_TEST(r3==r2);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r2) {
        r2 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L388; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L388:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        ((object****) lenv) [1][1][1][4] = r2;
      } else {
        r2 = (object *) 14;
      }
      r2 = ((object****) lenv) [1][1][1][4];
      r1[2] = r2;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 10498;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = (object*) &constructed_68;
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r2 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
        r1 = (object*) &constructed_1032;
        dump_object (r1, 0); fprintf (stdout, "\n");
        r1 = (object*) &constructed_1032;
        PXLL_UNDEFINED; result=r1; goto Lreturn;
      } else {
        goto fatbar_314;
      }
      goto fatbar_314_over;
      fatbar_314:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 8962;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L389; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L389:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = (object *) 65794;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L390; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L390:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object****) lenv) [1][1][1][4] = r1;
        } else {
          r1 = (object *) 14;
        }
        r1 = ((object****) lenv) [1][1][1][4];
        ((object*) lenv) [2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 10242;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][7];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L391; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L391:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(7),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_316;
        }
        goto fatbar_316_over;
        fatbar_316:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 29954;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L392; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L392:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object*)UITAG(6);
        } else {
          goto fatbar_317;
        }
        goto fatbar_317_over;
        fatbar_317:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 21762;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L393; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L393:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object*)UITAG(6);
        } else {
          goto fatbar_318;
        }
        goto fatbar_318_over;
        fatbar_318:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 26114;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L394; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L394:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 6;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_319;
        }
        goto fatbar_319_over;
        fatbar_319:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 17922;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L395; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L395:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 6;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_320;
        }
        goto fatbar_320_over;
        fatbar_320:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 29698;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L396; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L396:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 262;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_321;
        }
        goto fatbar_321_over;
        fatbar_321:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 21506;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L397; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L397:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = (object *) 262;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
        } else {
          goto fatbar_322;
        }
        goto fatbar_322_over;
        fatbar_322:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 28418;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L398; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L398:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][9];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L399; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L399:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_323;
        }
        goto fatbar_323_over;
        fatbar_323:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 20226;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L400; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L400:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][9];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L401; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L401:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_324;
        }
        goto fatbar_324_over;
        fatbar_324:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 30722;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L402; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L402:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][8];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L403; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L403:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_325;
        }
        goto fatbar_325_over;
        fatbar_325:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 22530;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L404; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L404:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][8];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L405; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L405:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(5),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
        } else {
          goto fatbar_326;
        }
        goto fatbar_326_over;
        fatbar_326:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r1 = (object *) 23554;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*****) lenv) [1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L406; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L406:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L407; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L407:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          switch (get_case_noint (r1)) {
            default: {
              r1 = allocate (TC_TUPLE, 3);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][2];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r2 = UOBJ_GET(r2,2);
              r1[3] = r2;
              r1 = (object *) 3;
              r2 = ((object*) lenv) [3];
              if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = (object *) 1;
                r3 = ((object**) lenv) [1][2];
                range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
                r2 = (object *) 1;
                r3 = ((object**) lenv) [1][2];
                r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
              } else {
                goto fatbar_329;
              }
              goto fatbar_329_over;
              fatbar_329:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1033;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L408; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L408:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 2;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_330;
              }
              goto fatbar_330_over;
              fatbar_330:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1034;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L409; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L409:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 65794;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_331;
              }
              goto fatbar_331_over;
              fatbar_331:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1035;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L410; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L410:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 2306;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_332;
              }
              goto fatbar_332_over;
              fatbar_332:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1036;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L411; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L411:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 3330;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_333;
              }
              goto fatbar_333_over;
              fatbar_333:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1037;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L412; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L412:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 8194;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                goto fatbar_334;
              }
              goto fatbar_334_over;
              fatbar_334:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object*) &constructed_1038;
              r2[3] = r3;
              r3 = top[50];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L413; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
              L413:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) 2562;
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r1;
                r1 = t;
              } else {
                lenv = ((object *)lenv)[1];
                goto fatbar_328;
              }
              fatbar_334_over: ;
              fatbar_333_over: ;
              fatbar_332_over: ;
              fatbar_331_over: ;
              fatbar_330_over: ;
              fatbar_329_over: ;
              lenv = lenv[1];
            } break;
          }
          goto fatbar_328_over;
          fatbar_328:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_1039;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L414; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L414:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_328_over: ;
          lenv = lenv[1];
        } else {
          goto fatbar_327;
        }
        goto fatbar_327_over;
        fatbar_327:
        r1 = allocate (TC_TUPLE, 3);
        r2 = (object*) &constructed_1040;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[15];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L415; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
        L415:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        fatbar_327_over: ;
        fatbar_326_over: ;
        fatbar_325_over: ;
        fatbar_324_over: ;
        fatbar_323_over: ;
        fatbar_322_over: ;
        fatbar_321_over: ;
        fatbar_320_over: ;
        fatbar_319_over: ;
        fatbar_318_over: ;
        fatbar_317_over: ;
        fatbar_316_over: ;
      } else {
        goto fatbar_315;
      }
      goto fatbar_315_over;
      fatbar_315:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 14850;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L416; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L416:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 3);
        r1[1] = lenv; lenv = r1;
        r2 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L417; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L417:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        switch (get_case_noint (r2)) {
          default: {
            r2 = allocate (TC_TUPLE, 4);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][2];
            r3 = UOBJ_GET(r3,0);
            r2[2] = r3;
            r3 = ((object**) lenv) [1][2];
            r3 = UOBJ_GET(r3,1);
            r2[3] = r3;
            r3 = ((object**) lenv) [1][2];
            r3 = UOBJ_GET(r3,3);
            r2[4] = r3;
            r2 = ((object*) lenv) [3];
            switch (get_case (r2)) {
              case ((pxll_int)PXLL_TRUE): {
                r2 = allocate (TC_TUPLE, 3);
                r3 = (object*) &constructed_1041;
                r2[2] = r3;
                r3 = ((object*) lenv) [2];
                r2[3] = r3;
                r3 = top[15];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L418; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_error1_293;
                L418:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
              } break;
              default: {
                r2 = (object *) 1;
                r3 = ((object*) lenv) [4];
                if PXLL_IS_TRUE(PXLL_TEST(r3==r2)) {
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object*) lenv) [2];
                  r2[2] = r3;
                  r3 = top[80];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L419; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_string__3esymbol_419;
                  L419:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                } else {
                  goto fatbar_336;
                }
                goto fatbar_336_over;
                fatbar_336:
                r2 = allocate (TC_TUPLE, 3);
                r3 = (object*) &constructed_1042;
                r2[2] = r3;
                r3 = ((object*) lenv) [2];
                r2[3] = r3;
                r3 = top[15];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L420; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_error1_293;
                L420:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                fatbar_336_over: ;
              } break;
            }
            lenv = lenv[1];
          } break;
        }
        r1[3] = r2;
        r1 = ((object*) lenv) [3];
        r2 = (object *) constructed_49[0];
        t = alloc_no_clear (UOTAG(9),2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_335;
      }
      goto fatbar_335_over;
      fatbar_335:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 11266;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L421; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L421:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = (object *) constructed_1044[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L422; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L422:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_337;
      }
      goto fatbar_337_over;
      fatbar_337:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 9986;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L423; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L423:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = (object *) constructed_870[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L424; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L424:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_338;
      }
      goto fatbar_338_over;
      fatbar_338:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 8706;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L425; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L425:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def loop_1328_i365_i439_i0
        goto L426;
        FUN_loop_1328_i365_i439_i0:
          check_heap (0);
          r0 = (object *) 23554;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L427; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L427:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L428; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L428:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            ((object*) lenv) [2] = r0;
            r0 = (object *) 23554;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L429; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L429:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 23554;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_341;
            }
            PXLL_RETURN(0);
            goto fatbar_341_over;
            fatbar_341:
            r0 = (object *) 8706;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L430; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L430:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 8706;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_342;
            }
            PXLL_RETURN(0);
            goto fatbar_342_over;
            fatbar_342:
            r0 = (object *) 29698;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L431; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L431:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 2306;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_343;
            }
            PXLL_RETURN(0);
            goto fatbar_343_over;
            fatbar_343:
            r0 = (object *) 28162;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L432; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L432:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 2562;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_344;
            }
            PXLL_RETURN(0);
            goto fatbar_344_over;
            fatbar_344:
            r0 = (object *) 29186;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L433; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L433:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = ((object*) lenv) [3];
              r2 = (object *) 3330;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_345;
            }
            PXLL_RETURN(0);
            goto fatbar_345_over;
            fatbar_345:
            r0 = (object *) 22530;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L434; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L434:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object******) lenv) [1][1][1][1][1][6];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L435; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto *r2[1];
              L435:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = ((object**) lenv) [1][3];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_346;
            }
            PXLL_RETURN(0);
            goto fatbar_346_over;
            fatbar_346:
            r0 = (object *) 30722;
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
              r0[2] = r1;
              r0 = (object *) 65794;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = ((object******) lenv) [1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L436; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L436:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object******) lenv) [1][1][1][1][1][4] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = ((object******) lenv) [1][1][1][1][1][4];
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object******) lenv) [1][1][1][1][1][6];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L437; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto *r2[1];
              L437:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = ((object**) lenv) [1][3];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1328_i365_i439_i0;
            } else {
              goto fatbar_347;
            }
            PXLL_RETURN(0);
            goto fatbar_347_over;
            fatbar_347:
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_1046;
            r0[2] = r1;
            r1 = ((object*) lenv) [3];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            fatbar_347_over: ;
            PXLL_RETURN(0);
            fatbar_346_over: ;
            PXLL_RETURN(0);
            fatbar_345_over: ;
            PXLL_RETURN(0);
            fatbar_344_over: ;
            PXLL_RETURN(0);
            fatbar_343_over: ;
            PXLL_RETURN(0);
            fatbar_342_over: ;
            PXLL_RETURN(0);
            fatbar_341_over: ;
            PXLL_RETURN(0);
          } else {
            goto fatbar_340;
          }
          PXLL_RETURN(0);
          goto fatbar_340_over;
          fatbar_340:
          r0 = (object *) 8706;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L438; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L438:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L439; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L439:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = top[54];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L440; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3estring_361;
            L440:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(2),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            goto fatbar_348;
          }
          PXLL_RETURN(0);
          goto fatbar_348_over;
          fatbar_348:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
          r0[2] = r1;
          r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L441; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L441:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
          r0 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 65794;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = ((object******) lenv) [1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L442; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L442:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object******) lenv) [1][1][1][1][1][4] = r0;
          } else {
            r0 = (object *) 14;
          }
          r0 = ((object******) lenv) [1][1][1][1][1][4];
          r1 = ((object*) lenv) [3];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv[2] = r0;
          lenv[3] = r1;
          goto FUN_loop_1328_i365_i439_i0;
          fatbar_348_over: ;
          PXLL_RETURN(0);
          fatbar_340_over: ;
          PXLL_RETURN(0);
        L426:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_loop_1328_i365_i439_i0; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 3);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object******) lenv) [1][1][1][1][1][4];
        r2[2] = r3;
        r2 = (object *) 65794;
        r3 = ((object*) lenv) [2];
        r2 = PXLL_TEST(r3==r2);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r2) {
          r2 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L443; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L443:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          ((object*****) lenv) [1][1][1][1][4] = r2;
        } else {
          r2 = (object *) 14;
        }
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r2 = (object *) constructed_1045[0];
        r1[3] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L444; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto *r2[1];
        L444:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
      } else {
        goto fatbar_339;
      }
      goto fatbar_339_over;
      fatbar_339:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 31490;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L445; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L445:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object*****) lenv) [1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def loop_1335_i399_i439_i0
        goto L446;
        FUN_loop_1335_i399_i439_i0:
          check_heap (0);
          r0 = ((object*****) lenv) [1][1][1][1][2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L447; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L447:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object********) lenv) [1][1][1][1][1][1][1][4];
          r1[2] = r2;
          r1 = (object *) 65794;
          r2 = ((object*) lenv) [2];
          r1 = PXLL_TEST(r2==r1);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r1) {
            r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L448; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L448:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            ((object*******) lenv) [1][1][1][1][1][1][4] = r1;
          } else {
            r1 = (object *) 14;
          }
          r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
          r0[2] = r1;
          r0 = (object *) 32002;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object********) lenv) [1][1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L449; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L449:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object********) lenv) [1][1][1][1][1][1][1][4] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][2];
            r1[2] = r2;
            r2 = (object *) constructed_36[0];
            r1[3] = r2;
            r2 = top[21];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L450; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
            L450:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(8),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            // def loop_1340_i406_i439_i0
            goto L451;
            FUN_loop_1340_i406_i439_i0:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object*) lenv) [3];
              r0[2] = r1;
              r1 = top[179];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L452; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L452:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = (object *) 262;
              } else {
                r0 = (object *) 11778;
                r1 = ((object*) lenv) [3];
                r0 = PXLL_TEST(r1==r0);
              }
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][4];
                r0[2] = r1;
                r0 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L453; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L453:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][4] = r0;
                r0 = ((object*) lenv) [2];
                lenv = lenv[1];
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][4];
                r0[2] = r1;
                r0 = (object *) 65794;
                r1 = ((object*) lenv) [2];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L454; ; k = t;
                  lenv = r0[2]; goto *r0[1];
                  L454:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4] = r0;
                } else {
                  r0 = (object *) 14;
                }
                r0 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
                r1 = ((object*) lenv) [2];
                r2 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r2 = (object *) 6;
                lenv[2] = r1;
                lenv[3] = r0;
                lenv[4] = r2;
                goto FUN_loop_1340_i406_i439_i0;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object*) lenv) [2];
                r2[2] = r3;
                r3 = (object *) constructed_36[0];
                r2[3] = r3;
                r3 = top[21];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L455; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                L455:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = top[54];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L456; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_list__3estring_361;
                L456:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = top[80];
                r0[1] = r1[2]; lenv = r0; goto FUN_string__3esymbol_419;
              }
              PXLL_RETURN(0);
            L451:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_loop_1340_i406_i439_i0; r2[2] = lenv;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 4);
            r2 = (object *) constructed_1049[0];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
            r2[2] = r3;
            r2 = (object *) 65794;
            r3 = ((object*) lenv) [2];
            r2 = PXLL_TEST(r3==r2);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r2) {
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L457; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto *r2[1];
              L457:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              ((object*********) lenv) [1][1][1][1][1][1][1][1][4] = r2;
            } else {
              r2 = (object *) 14;
            }
            r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
            r1[3] = r2;
            r2 = (object *) 6;
            r1[4] = r2;
            r2 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L458; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L458:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r0[2] = r1;
            r0 = (object *) constructed_66[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = (object *) 6;
              t = alloc_no_clear (UOTAG(4),1);
              t[1] = r1;
              r1 = t;
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(0),2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[3] = r1;
              r0 = ((object****) lenv) [1][1][1][2];
              r1 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              lenv = ((object **)lenv)[1][1];
              lenv[2] = r0;
              goto FUN_loop_1335_i399_i439_i0;
            } else {
              r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L459; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L459:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
              r1[2] = r2;
              r1 = (object *) 65794;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(r2==r1);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L460; t[4] = r0; k = t;
                lenv = r1[2]; goto *r1[1];
                L460:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                ((object*********) lenv) [1][1][1][1][1][1][1][1][4] = r1;
              } else {
                r1 = (object *) 14;
              }
              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
              r0[2] = r1;
              r1 = (object *) 15618;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(r2==r1);
              r0[3] = r1;
              r0 = (object *) 6;
              r1 = ((object*) lenv) [3];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object*) &constructed_1048;
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][2];
                r0[3] = r1;
                r1 = top[15];
                r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                r0[2] = r1;
                r0 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L461; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L461:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                ((object*********) lenv) [1][1][1][1][1][1][1][1][4] = r0;
                r0 = ((object*) lenv) [2];
                lenv = lenv[1];
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L462; t[4] = r0; k = t;
                lenv = r1[2]; goto *r1[1];
                L462:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r2 = ((object***) lenv) [1][1][2];
                t = alloc_no_clear (UOTAG(0),2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[2] = r1;
                r0 = ((object*****) lenv) [1][1][1][1][2];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                lenv = ((object ***)lenv)[1][1][1];
                lenv[2] = r0;
                goto FUN_loop_1335_i399_i439_i0;
              }
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        L446:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_loop_1335_i399_i439_i0; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 2);
        r2 = (object *) constructed_1047[0];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L463; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto *r2[1];
        L463:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
      } else {
        goto fatbar_349;
      }
      goto fatbar_349_over;
      fatbar_349:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 10242;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object****) lenv) [1][1][1][7];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L464; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L464:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        lenv = lenv[1];
      } else {
        goto fatbar_350;
      }
      goto fatbar_350_over;
      fatbar_350:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = (object*) &constructed_68;
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r2 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
        r1 = (object*) &constructed_1050;
        dump_object (r1, 0); fprintf (stdout, "\n");
        r1 = (object*) &constructed_1050;
        PXLL_UNDEFINED; result=r1; goto Lreturn;
      } else {
        goto fatbar_351;
      }
      goto fatbar_351_over;
      fatbar_351:
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][3];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L465; t[4] = r0; t[5] = r1; k = t;
      lenv = r2[2]; goto *r2[1];
      L465:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r1 = ((object*) lenv) [2];
      switch (get_case_noint (r1)) {
        default: {
          r1 = allocate (TC_TUPLE, 5);
          r1[1] = lenv; lenv = r1;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,1);
          r1[3] = r2;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,2);
          r1[4] = r2;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,3);
          r1[5] = r2;
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case ((pxll_int)PXLL_TRUE): {
              r1 = allocate (TC_TUPLE, 5);
              r1[1] = lenv; lenv = r1;
              r2 = (object *) 1;
              r3 = ((object**) lenv) [1][2];
              range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
              r2 = (object *) 1;
              r3 = ((object**) lenv) [1][2];
              r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
              r1[2] = r2;
              r2 = (object *) 11522;
              r3 = ((object*) lenv) [2];
              r2 = PXLL_TEST(r3==r2);
              r1[3] = r2;
              r2 = ((object*) lenv) [3];
              if PXLL_IS_TRUE(r2) {
                r2 = (object *) 3;
              } else {
                r2 = (object *) 1;
              }
              r1[4] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r2[1] = lenv; lenv = r2;
              // def loop_1348_i427_i439_i0
              goto L466;
              FUN_loop_1348_i427_i439_i0:
                check_heap (0);
                r0 = ((object****) lenv) [1][1][1][4];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                  r0 = ((object***) lenv) [1][1][3];
                  if PXLL_IS_TRUE(r0) {
                    r0 = ((object*) lenv) [3];
                    r1 = (object *) 1;
                    r0 = box((pxll_int)unbox(r1)-unbox(r0));
                    PXLL_RETURN(0);
                  } else {
                    r0 = ((object*) lenv) [3];
                    PXLL_RETURN(0);
                  }
                  PXLL_RETURN(0);
                } else {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[169];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r3 = ((object*****) lenv) [1][1][1][1][2];
                  range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
                  r2 = ((object**) lenv) [1][2];
                  r3 = ((object*****) lenv) [1][1][1][1][2];
                  r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
                  r1[3] = r2;
                  r2 = top[65];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L467; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
                  L467:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UITAG(1)): {
                      r0 = allocate (TC_TUPLE, 3);
                      r1 = (object*) &constructed_1051;
                      r0[2] = r1;
                      r1 = ((object*****) lenv) [1][1][1][1][2];
                      r0[3] = r1;
                      r1 = top[15];
                      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                    } break;
                    default: {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = (object *) 21;
                      r2 = ((object****) lenv) [1][1][1][3];
                      r1 = box((pxll_int)unbox(r2)*unbox(r1));
                      r0[2] = r1;
                      r0 = ((object**) lenv) [1][2];
                      r1 = ((object*) lenv) [2];
                      r0 = box((pxll_int)unbox(r1)+unbox(r0));
                      lenv = lenv[1];
                      r1 = (object *) 3;
                      r2 = ((object***) lenv) [1][1][2];
                      r1 = box((pxll_int)unbox(r2)+unbox(r1));
                      lenv = ((object **)lenv)[1][1];
                      lenv[2] = r1;
                      lenv[3] = r0;
                      goto FUN_loop_1348_i427_i439_i0;
                    } break;
                  }
                  PXLL_RETURN(0);
                }
                PXLL_RETURN(0);
              L466:
              r3 = allocate (TC_CLOSURE, 2);
              r3[1] = &&FUN_loop_1348_i427_i439_i0; r3[2] = lenv;
              r2[2] = r3;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][4];
              r2[2] = r3;
              r3 = (object *) 1;
              r2[3] = r3;
              r3 = ((object*) lenv) [2];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L468; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto *r3[1];
              L468:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              lenv = lenv[1];
              r1[5] = r2;
              r1 = ((object*) lenv) [5];
              t = alloc_no_clear (UOTAG(5),1);
              t[1] = r1;
              r1 = t;
              lenv = lenv[1];
            } break;
            default: {
              r1 = (object *) 1;
              r2 = ((object*) lenv) [5];
              if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object**) lenv) [1][2];
                r2[2] = r3;
                r3 = top[80];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L469; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string__3esymbol_419;
                L469:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
              } else {
                goto fatbar_352;
              }
              goto fatbar_352_over;
              fatbar_352:
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][2];
              r2 = box((pxll_int)((pxll_string*)(r2))->len);
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][5];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object**) lenv) [1][2];
                r2[2] = r3;
                r3 = top[80];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L470; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string__3esymbol_419;
                L470:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
              } else {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                // def loop_1297_i438_i439_i0
                goto L471;
                FUN_loop_1297_i438_i439_i0:
                  check_heap (0);
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_NIL): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = ((object**) lenv) [1][2];
                          r1[2] = r2;
                          r2 = top[80];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L472; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
                          L472:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r0 = ((object*) lenv) [2];
                          t = alloc_no_clear (UOTAG(1),1);
                          t[1] = r0;
                          r0 = t;
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          goto fatbar_353;
                        }
                      }
                      PXLL_RETURN(0);
                      goto fatbar_353_over;
                      fatbar_353:
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][3];
                      r1[2] = r2;
                      r2 = ((object****) lenv) [1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L473; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L473:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][2];
                      r1[2] = r2;
                      r2 = top[80];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L474; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
                      L474:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      r1 = ((object*) lenv) [2];
                      t = alloc_no_clear (UOTAG(10),2);
                      t[1] = r1;
                      t[2] = r0;
                      r0 = t;
                      PXLL_RETURN(0);
                      fatbar_353_over: ;
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      r0 = (object*) &constructed_68;
                      r0 = box((pxll_int)((pxll_string*)(r0))->len);
                      r1 = (object*) &constructed_68;
                      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                      r0 = (object*) &constructed_67;
                      dump_object (r0, 0); fprintf (stdout, "\n");
                      r0 = (object*) &constructed_67;
                      PXLL_UNDEFINED; result=r0; goto Lreturn;
                      PXLL_RETURN(0);
                    } break;
                  }
                  PXLL_RETURN(0);
                L471:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_loop_1297_i438_i439_i0; r2[2] = lenv;
                r1[2] = r2;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = allocate (TC_TUPLE, 3);
                r4 = ((object***) lenv) [1][1][2];
                r3[2] = r4;
                r4 = (object *) 11778;
                r3[3] = r4;
                r4 = top[49];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L475; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_string_split_354;
                L475:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[2] = r3;
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object*) lenv) [2];
                r2[2] = r3;
                r3 = (object *) constructed_36[0];
                r2[3] = r3;
                r3 = top[21];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L476; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                L476:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L477; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                L477:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                lenv = lenv[1];
              }
              fatbar_352_over: ;
            } break;
          }
          lenv = lenv[1];
        } break;
      }
      lenv = lenv[1];
      fatbar_351_over: ;
      fatbar_350_over: ;
      fatbar_349_over: ;
      fatbar_339_over: ;
      fatbar_338_over: ;
      fatbar_337_over: ;
      fatbar_335_over: ;
      fatbar_315_over: ;
      fatbar_314_over: ;
      lenv = lenv[1];
      r0[2] = r1;
      r0 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L478; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L478:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L479; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L479:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][4];
      r0[2] = r1;
      r0 = (object *) 14850;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object*****) lenv) [1][1][1][1][4];
        r0[2] = r1;
        r0 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L480; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L480:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object*****) lenv) [1][1][1][1][4] = r0;
        r0 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L481; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L481:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object***) lenv) [1][1][2];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (UOTAG(9),2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_311;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_311;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_311_over;
        fatbar_311:
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(9)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = (object *) constructed_49[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 7);
              r0[1] = lenv; lenv = r0;
              r1 = (object *) constructed_654[0];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r1;
              r1 = t;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r1;
              r1 = t;
              r0[3] = r1;
              r1 = (object*)TC_NIL;
              r0[4] = r1;
              r1 = ((object*) lenv) [4];
              r2 = ((object*****) lenv) [1][1][1][1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[5] = r1;
              r1 = ((object*) lenv) [5];
              r2 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[6] = r1;
              r1 = ((object*) lenv) [6];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[7] = r1;
              r0 = ((object*) lenv) [7];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r0;
              r0 = t;
              PXLL_RETURN(0);
            } else {
              lenv = ((object *)lenv)[1];
              goto fatbar_312;
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_312;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_312_over;
        fatbar_312:
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_1030;
        r0[2] = r1;
        r1 = allocate (UOTAG(1), 2);
        r2 = ((object***) lenv) [1][1][2];
        r1[1] = r2;
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        fatbar_312_over: ;
        PXLL_RETURN(0);
        fatbar_311_over: ;
        PXLL_RETURN(0);
      } else {
        goto fatbar_310;
      }
      PXLL_RETURN(0);
      goto fatbar_310_over;
      fatbar_310:
      r0 = (object *) 23298;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 9);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object******) lenv) [1][1][1][1][1][4];
        r1[2] = r2;
        r1 = ((object******) lenv) [1][1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L482; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L482:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object******) lenv) [1][1][1][1][1][4] = r1;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L483; t[4] = r0; t[5] = r1; k = t;
        lenv = r2[2]; goto *r2[1];
        L483:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L484; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L484:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object********) lenv) [1][1][1][1][1][1][1][4];
        r2[2] = r3;
        r2 = (object *) 65794;
        r3 = ((object*) lenv) [2];
        r2 = PXLL_TEST(r3==r2);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r2) {
          r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L485; t[4] = r0; t[5] = r1; k = t;
          lenv = r2[2]; goto *r2[1];
          L485:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          ((object*******) lenv) [1][1][1][1][1][1][4] = r2;
        } else {
          r2 = (object *) 14;
        }
        r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
        r1[2] = r2;
        r1 = (object *) 23810;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
          r1[2] = r2;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L486; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L486:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*******) lenv) [1][1][1][1][1][1][4] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1 = ((object*) lenv) [2];
        } else {
          r1 = (object*) &constructed_68;
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r2 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
          r1 = (object*) &constructed_1031;
          dump_object (r1, 0); fprintf (stdout, "\n");
          r1 = (object*) &constructed_1031;
          PXLL_UNDEFINED; result=r1; goto Lreturn;
        }
        lenv = lenv[1];
        r0[2] = r1;
        r1 = (object *) constructed_295[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[3] = r1;
        r1 = (object *) 6;
        t = alloc_no_clear (UOTAG(4),1);
        t[1] = r1;
        r1 = t;
        r0[4] = r1;
        r1 = (object*)TC_NIL;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[6] = r1;
        r1 = ((object*) lenv) [6];
        r2 = ((object***) lenv) [1][1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[7] = r1;
        r1 = ((object*) lenv) [7];
        r2 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[8] = r1;
        r1 = ((object*) lenv) [8];
        r2 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[9] = r1;
        r0 = ((object*) lenv) [9];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } else {
        goto fatbar_313;
      }
      PXLL_RETURN(0);
      goto fatbar_313_over;
      fatbar_313:
      r0 = ((object**) lenv) [1][2];
      PXLL_RETURN(0);
      fatbar_313_over: ;
      PXLL_RETURN(0);
      fatbar_310_over: ;
      PXLL_RETURN(0);
    L386:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_1272_i0; r1[2] = lenv;
    r0[4] = r1;
    // def read-hex-digit_1274_i0
    goto L487;
    FUN_read_hex_digit_1274_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = top[168];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = top[65];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L488; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
      L488:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(1)): {
          r0 = (object*) &constructed_68;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_1029;
          dump_object (r0, 0); fprintf (stdout, "\n");
          r0 = (object*) &constructed_1029;
          PXLL_UNDEFINED; result=r0; goto Lreturn;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L487:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_digit_1274_i0; r1[2] = lenv;
    r0[5] = r1;
    // def read-hex-code_1275_i0
    goto L489;
    FUN_read_hex_code_1275_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 5);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object****) lenv) [1][1][1][4];
      r2[2] = r3;
      r2 = ((object****) lenv) [1][1][1][3];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L490; t[4] = r0; t[5] = r1; k = t;
      lenv = r2[2]; goto *r2[1];
      L490:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      ((object****) lenv) [1][1][1][4] = r2;
      r2 = ((object*) lenv) [2];
      lenv = lenv[1];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L491; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto *r2[1];
      L491:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object****) lenv) [1][1][1][4];
      r2[2] = r3;
      r2 = ((object****) lenv) [1][1][1][3];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L492; t[4] = r0; t[5] = r1; k = t;
      lenv = r2[2]; goto *r2[1];
      L492:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      ((object****) lenv) [1][1][1][4] = r2;
      r2 = ((object*) lenv) [2];
      lenv = lenv[1];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L493; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto *r2[1];
      L493:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = (object *) 17;
      r2 = ((object*) lenv) [2];
      r1 = box((pxll_int)unbox(r2)<<unbox(r1));
      r0[4] = r1;
      r1 = ((object*) lenv) [3];
      r2 = ((object*) lenv) [4];
      r1 = box((pxll_int)unbox(r2)+unbox(r1));
      r0[5] = r1;
      r0 = ((object*) lenv) [5];
      r0 = TO_CHAR(unbox(r0));
      PXLL_RETURN(0);
    L489:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_code_1275_i0; r1[2] = lenv;
    r0[6] = r1;
    // def read-list_1277_i0
    goto L494;
    FUN_read_list_1277_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][4];
      r0[2] = r1;
      r0 = ((object***) lenv) [1][1][3];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L495; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L495:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      ((object***) lenv) [1][1][4] = r0;
      r0 = ((object*) lenv) [2];
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1331_i0
      goto L496;
      FUN_loop_1331_i0:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L497; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L497:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object******) lenv) [1][1][1][1][1][4];
        r1[2] = r2;
        r1 = (object *) 65794;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L498; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L498:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          ((object*****) lenv) [1][1][1][1][4] = r1;
        } else {
          r1 = (object *) 14;
        }
        r1 = ((object*****) lenv) [1][1][1][1][4];
        r0[2] = r1;
        r0 = (object *) 10498;
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][4];
          r0[2] = r1;
          r0 = ((object******) lenv) [1][1][1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L499; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L499:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object******) lenv) [1][1][1][1][1][4] = r0;
          r0 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object**) lenv) [1][2];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*****) lenv) [1][1][1][1][4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L500; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L500:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object***) lenv) [1][1][2];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = ((object **)lenv)[1][1];
          lenv[2] = r0;
          goto FUN_loop_1331_i0;
        }
        PXLL_RETURN(0);
      L496:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1331_i0; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) constructed_1028[0];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L494:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_list_1277_i0; r1[2] = lenv;
    r0[7] = r1;
    // def read-hex-int_1281_i0
    goto L501;
    FUN_read_hex_int_1281_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object***) lenv) [1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L502; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L502:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object***) lenv) [1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object***) lenv) [1][1][4];
      r0[2] = r1;
      r1 = (object *) 11522;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][4];
        r0[2] = r1;
        r0 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L503; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L503:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][4] = r0;
        r0 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0 = (object *) 14;
      } else {
        r0 = (object *) 14;
      }
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1354_i0
      goto L504;
      FUN_loop_1354_i0:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[168];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[65];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L505; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
        L505:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object********) lenv) [1][1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L506; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L506:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object********) lenv) [1][1][1][1][1][1][1][4] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = (object *) 65794;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L507; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L507:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) 33;
            r3 = ((object****) lenv) [1][1][1][2];
            r2 = box((pxll_int)unbox(r3)*unbox(r2));
            r1[2] = r2;
            r1 = ((object**) lenv) [1][2];
            r2 = ((object*) lenv) [2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv = lenv[1];
            lenv = ((object **)lenv)[1][1];
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_1354_i0;
          } break;
          default: {
            r0 = ((object****) lenv) [1][1][1][3];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object**) lenv) [1][2];
              r1 = (object *) 1;
              r0 = box((pxll_int)unbox(r1)-unbox(r0));
              PXLL_RETURN(0);
            } else {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      L504:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1354_i0; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L508; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L508:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][4];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L501:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_hex_int_1281_i0; r1[2] = lenv;
    r0[8] = r1;
    // def read-oct-int_1282_i0
    goto L509;
    FUN_read_oct_int_1282_i0:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object****) lenv) [1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object***) lenv) [1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L510; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L510:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object***) lenv) [1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object***) lenv) [1][1][4];
      r0[2] = r1;
      r1 = (object *) 11522;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][4];
        r0[2] = r1;
        r0 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L511; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L511:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][4] = r0;
        r0 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0 = (object *) 14;
      } else {
        r0 = (object *) 14;
      }
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1360_i0
      goto L512;
      FUN_loop_1360_i0:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[170];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[65];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L513; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
        L513:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object********) lenv) [1][1][1][1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L514; ; k = t;
            lenv = r0[2]; goto *r0[1];
            L514:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            ((object********) lenv) [1][1][1][1][1][1][1][4] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
            r0[2] = r1;
            r0 = (object *) 65794;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L515; ; k = t;
              lenv = r0[2]; goto *r0[1];
              L515:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              ((object*******) lenv) [1][1][1][1][1][1][4] = r0;
            } else {
              r0 = (object *) 14;
            }
            r0 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) 17;
            r3 = ((object****) lenv) [1][1][1][2];
            r2 = box((pxll_int)unbox(r3)*unbox(r2));
            r1[2] = r2;
            r1 = ((object**) lenv) [1][2];
            r2 = ((object*) lenv) [2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv = lenv[1];
            lenv = ((object **)lenv)[1][1];
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_1360_i0;
          } break;
          default: {
            r0 = ((object****) lenv) [1][1][1][3];
            if PXLL_IS_TRUE(r0) {
              r0 = ((object**) lenv) [1][2];
              r1 = (object *) 1;
              r0 = box((pxll_int)unbox(r1)-unbox(r0));
              PXLL_RETURN(0);
            } else {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      L512:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1360_i0; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L516; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L516:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][4];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L509:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_read_oct_int_1282_i0; r1[2] = lenv;
    r0[9] = r1;
    // def loop_1367_i575_i0
    goto L517;
    FUN_loop_1367_i575_i0:
      check_heap (0);
      r0 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L518; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L518:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object*****) lenv) [1][1][1][1][4];
      r1[2] = r2;
      r1 = (object *) 65794;
      r2 = ((object*) lenv) [2];
      r1 = PXLL_TEST(r2==r1);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object****) lenv) [1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L519; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L519:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        ((object****) lenv) [1][1][1][4] = r1;
      } else {
        r1 = (object *) 14;
      }
      r1 = ((object****) lenv) [1][1][1][4];
      r0[2] = r1;
      r0 = (object *) 65794;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(r1==r0);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = (object *) constructed_36[0];
        r0[3] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L520; t[4] = r0; k = t;
        lenv = r1[2]; goto *r1[1];
        L520:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(1)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = (object *) constructed_1027[0];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                      r0 = ((object**) lenv) [1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (UOTAG(2)): {
                              r0 = allocate (TC_TUPLE, 2);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r0 = ((object**) lenv) [1][3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r2 = top[85];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L521; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_read_file_425;
                                  L521:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object*) lenv) [2];
                                  r1[2] = r2;
                                  r2 = (object *) constructed_36[0];
                                  r1[3] = r2;
                                  r2 = top[21];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L522; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                  L522:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object*) lenv) [3];
                                  r1[2] = r2;
                                  r2 = (object *) constructed_36[0];
                                  r1[3] = r2;
                                  r2 = top[21];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L523; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                  L523:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                  r0[3] = r1;
                                  r1 = top[21];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L524; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                                  L524:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = result;
                                  lenv = lenv[1];
                                  lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                  lenv[2] = r0;
                                  goto FUN_loop_1367_i575_i0;
                                } break;
                                default: {
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  goto fatbar_309;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_309;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_309;
                        }
                      }
                      PXLL_RETURN(0);
                    } else {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_309;
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_309;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_309;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_309;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_309_over;
        fatbar_309:
        r0 = ((object**) lenv) [1][2];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = ((object *)lenv)[1];
        lenv[2] = r0;
        goto FUN_loop_1367_i575_i0;
        fatbar_309_over: ;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L517:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_1367_i575_i0; r1[2] = lenv;
    r0[10] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object *) constructed_1025[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [10];
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
  L356:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_read_file_425; r1[2] = lenv;
  r0[85] = r1;
  // def sexp->symbol_427
  goto L525;
  FUN_sexp__3esymbol_427:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_308;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_308_over;
    fatbar_308:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_1024;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_308_over: ;
    PXLL_RETURN(0);
  L525:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sexp__3esymbol_427; r1[2] = lenv;
  r0[86] = r1;
  // def sexp->int_428
  goto L526;
  FUN_sexp__3eint_428:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_307;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_307_over;
    fatbar_307:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_1023;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_307_over: ;
    PXLL_RETURN(0);
  L526:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sexp__3eint_428; r1[2] = lenv;
  r0[87] = r1;
  // def field=?_429
  goto L527;
  FUN_field_3d_question_429:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,1);
    r0[3] = r1;
    r0 = ((object**) lenv) [1][3];
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][3];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r1 = ((object***) lenv) [1][1][3];
    r1 = UOBJ_GET(r1,1);
    r0[3] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object**) lenv) [1][2];
    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object**) lenv) [1][3];
      r0[2] = r1;
      r1 = ((object*) lenv) [3];
      r0[3] = r1;
      r1 = top[89];
      r0[1] = r1[2]; lenv = r0; goto FUN_sexp_3d_question_430;
    } else {
      r0 = (object *) 6;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
    PXLL_RETURN(0);
    PXLL_RETURN(0);
  L527:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_field_3d_question_429; r1[2] = lenv;
  r0[88] = r1;
  // def sexp=?_430
  goto L528;
  FUN_sexp_3d_question_430:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(6)): {
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (UITAG(6)): {
            r0 = (object *) 262;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            r0 = PXLL_TEST(r1==r0);
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(4)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            r0 = PXLL_TEST(r1==r0);
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(5)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            r0 = PXLL_TEST(unbox(r1)==unbox(r0));
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(2)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = top[50];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L529; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
            L529:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = (object *) 1;
            r1 = ((object*) lenv) [3];
            r0 = PXLL_TEST(unbox(r1)==unbox(r0));
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(3)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            r0 = PXLL_TEST(r1==r0);
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 4);
            r1 = top[89];
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = ((object*) lenv) [2];
            r0[4] = r1;
            r1 = top[41];
            r0[1] = r1[2]; lenv = r0; goto FUN_every2_question_335;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(7)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 4);
            r1 = top[89];
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = ((object*) lenv) [2];
            r0[4] = r1;
            r1 = top[41];
            r0[1] = r1[2]; lenv = r0; goto FUN_every2_question_335;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(8)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(8)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 4);
            r1 = top[88];
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = ((object*) lenv) [2];
            r0[4] = r1;
            r1 = top[41];
            r0[1] = r1[2]; lenv = r0; goto FUN_every2_question_335;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(9)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = ((object*) lenv) [3];
              r1 = ((object**) lenv) [1][3];
              r0 = PXLL_TEST(r1==r0);
              PXLL_RETURN(0);
            } else {
              r0 = (object *) 6;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(10)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = top[89];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L530; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_sexp_3d_question_430;
            L530:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = ((object*) lenv) [3];
              r1 = ((object**) lenv) [1][3];
              r0 = PXLL_TEST(r1==r0);
              PXLL_RETURN(0);
            } else {
              r0 = (object *) 6;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_306;
          }
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_306_over;
    fatbar_306:
    r0 = (object *) 6;
    PXLL_RETURN(0);
    fatbar_306_over: ;
    PXLL_RETURN(0);
  L528:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sexp_3d_question_430; r1[2] = lenv;
  r0[89] = r1;
  // def repr-field_432
  goto L531;
  FUN_repr_field_432:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,1);
    r0[3] = r1;
    r0 = (object *) constructed_66[0];
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
      r0 = (object*) &constructed_1021;
      PXLL_RETURN(0);
    } else {
      goto fatbar_305;
    }
    PXLL_RETURN(0);
    goto fatbar_305_over;
    fatbar_305:
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 6);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][2];
    r2 = allocate (TC_TUPLE, 2);
    r2[1] = lenv; lenv = r2;
    r3 = ((object***) lenv) [1][1][2];
    r3 = UOBJ_GET(r3,0);
    r2[2] = r3;
    r2 = ((object*) lenv) [2];
    lenv = lenv[1];
    r1[2] = r2;
    r2 = allocate (TC_TUPLE, 2);
    r3 = ((object**) lenv) [1][3];
    r2[2] = r3;
    r3 = top[91];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L532; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_repr_433;
    L532:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[3] = r2;
    r2 = (object*)TC_NIL;
    r1[4] = r2;
    r2 = ((object*) lenv) [4];
    r3 = ((object*) lenv) [3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    r1[5] = r2;
    r2 = ((object*) lenv) [5];
    r3 = (object*) &constructed_1022;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    r1[6] = r2;
    r1 = ((object*) lenv) [6];
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    lenv = lenv[1];
    r0[2] = r1;
    r1 = top[47];
    r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
    fatbar_305_over: ;
    PXLL_RETURN(0);
    PXLL_RETURN(0);
  L531:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_repr_field_432; r1[2] = lenv;
  r0[90] = r1;
  // def repr_433
  goto L533;
  FUN_repr_433:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = top[91];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][2];
        r3[3] = r4;
        r4 = top[32];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L534; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
        L534:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*) &constructed_1005;
        r2[3] = r3;
        r3 = top[48];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L535; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
        L535:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_1004;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_1003;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
        PXLL_RETURN(0);
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 4);
        r1[1] = lenv; lenv = r1;
        r2 = (object*)TC_NIL;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r3 = (object*) &constructed_1007;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = ((object**) lenv) [1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r1 = ((object*) lenv) [4];
        r2 = (object*) &constructed_1006;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 4);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[46];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L536; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_char__3estring_348;
        L536:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r1 = ((object*) lenv) [4];
        r2 = (object*) &constructed_1008;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case ((pxll_int)PXLL_TRUE): {
            r0 = (object*) &constructed_1009;
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = (object*) &constructed_1010;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[56];
        r0[1] = r1[2]; lenv = r0; goto FUN_int__3estring_364;
      } break;
      case (UITAG(6)): {
        r0 = (object*) &constructed_1011;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = top[91];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][2];
        r3[3] = r4;
        r4 = top[32];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L537; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
        L537:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*) &constructed_1014;
        r2[3] = r3;
        r3 = top[48];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L538; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
        L538:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_1013;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_1012;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UOTAG(8)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = top[90];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][2];
        r3[3] = r4;
        r4 = top[32];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L539; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
        L539:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*) &constructed_1017;
        r2[3] = r3;
        r3 = top[48];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L540; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
        L540:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_1016;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_1015;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = (object *) constructed_49[0];
        r3 = ((object**) lenv) [1][2];
        if PXLL_IS_TRUE(PXLL_TEST(r3==r2)) {
          r2 = (object*) &constructed_1019;
        } else {
          r2 = ((object**) lenv) [1][2];
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = ((object***) lenv) [1][1][2];
          r3 = UOBJ_GET(r3,0);
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          lenv = lenv[1];
        }
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object***) lenv) [1][1][3];
        r3 = UOBJ_GET(r3,0);
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = (object*) &constructed_1018;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[91];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L541; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_repr_433;
        L541:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object***) lenv) [1][1][3];
        r3 = UOBJ_GET(r3,0);
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = (object*) &constructed_1020;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
    }
    PXLL_RETURN(0);
  L533:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_repr_433; r1[2] = lenv;
  r0[91] = r1;
  // def indent_434
  goto L542;
  FUN_indent_434:
    check_heap (0);
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
      r0 = (object *) 262;
      PXLL_RETURN(0);
    } else {
      goto fatbar_304;
    }
    PXLL_RETURN(0);
    goto fatbar_304_over;
    fatbar_304:
    r0 = (object*) &constructed_1002;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_1002;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object *) 3;
    r2 = ((object*) lenv) [2];
    r1 = box((pxll_int)unbox(r2)-unbox(r1));
    r0[2] = r1;
    r1 = top[92];
    r0[1] = r1[2]; lenv = r0; goto FUN_indent_434;
    fatbar_304_over: ;
    PXLL_RETURN(0);
  L542:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_indent_434; r1[2] = lenv;
  r0[92] = r1;
  // def pp-size-field_435
  goto L543;
  FUN_pp_size_field_435:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r0 = allocate (TC_TUPLE, 7);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,1);
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][2];
    r2 = UOBJ_GET(r2,0);
    r1[2] = r2;
    r1 = ((object*) lenv) [2];
    lenv = lenv[1];
    r0[4] = r1;
    r1 = ((object*) lenv) [4];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[5] = r1;
    r1 = (object *) 3;
    r2 = ((object*) lenv) [5];
    r1 = box((pxll_int)unbox(r2)+unbox(r1));
    r0[6] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [3];
    r1[2] = r2;
    r2 = top[94];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L544; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_pp_size_436;
    L544:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[7] = r1;
    r0 = ((object*) lenv) [7];
    r1 = ((object*) lenv) [6];
    r0 = box((pxll_int)unbox(r1)+unbox(r0));
    PXLL_RETURN(0);
    PXLL_RETURN(0);
  L543:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pp_size_field_435; r1[2] = lenv;
  r0[93] = r1;
  // def pp-size_436
  goto L545;
  FUN_pp_size_436:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = top[13];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[22];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L546; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
        L546:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        r2 = (object *) 3;
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        lenv = lenv[1];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[94];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L547; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L547:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = top[38];
        r0[1] = r1[2]; lenv = r0; goto FUN_foldr_332;
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r2 = UOBJ_GET(r2,0);
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        PXLL_RETURN(0);
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = (object *) 5;
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        PXLL_RETURN(0);
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(3),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = top[91];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L548; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_repr_433;
        L548:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        PXLL_RETURN(0);
      } break;
      case (UOTAG(4)): {
        r0 = (object *) 5;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[56];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L549; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_int__3estring_364;
        L549:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        PXLL_RETURN(0);
      } break;
      case (UITAG(6)): {
        r0 = (object *) 5;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = top[13];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[22];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L550; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
        L550:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        r2 = (object *) 5;
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        lenv = lenv[1];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[94];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L551; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L551:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = top[38];
        r0[1] = r1[2]; lenv = r0; goto FUN_foldr_332;
      } break;
      case (UOTAG(8)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = top[13];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[22];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L552; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
        L552:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = (object *) 3;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        lenv = lenv[1];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[93];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L553; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L553:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = top[38];
        r0[1] = r1[2]; lenv = r0; goto FUN_foldr_332;
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 8);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r2 = UOBJ_GET(r2,0);
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0[4] = r1;
        r1 = ((object*) lenv) [4];
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r0[5] = r1;
        r1 = ((object*) lenv) [3];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][3];
        r2 = UOBJ_GET(r2,0);
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0[6] = r1;
        r1 = ((object*) lenv) [6];
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r0[7] = r1;
        r1 = ((object*) lenv) [7];
        r2 = ((object*) lenv) [5];
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        r0[8] = r1;
        r0 = ((object*) lenv) [8];
        r1 = (object *) 3;
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 7);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[94];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L554; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_pp_size_436;
        L554:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = ((object*) lenv) [3];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][3];
        r2 = UOBJ_GET(r2,0);
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        r1 = box((pxll_int)((pxll_string*)(r1))->len);
        r0[6] = r1;
        r1 = ((object*) lenv) [6];
        r2 = ((object*) lenv) [4];
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        r0[7] = r1;
        r0 = ((object*) lenv) [7];
        r1 = (object *) 3;
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L545:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pp_size_436; r1[2] = lenv;
  r0[94] = r1;
  // def pp_437
  goto L555;
  FUN_pp_437:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][3];
    r1[2] = r2;
    r2 = top[94];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L556; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_pp_size_436;
    L556:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = (object *) 161;
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)<unbox(r0))) {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object***) lenv) [1][1][3];
      r1[2] = r2;
      r2 = top[91];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L557; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_repr_433;
      L557:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      PXLL_RETURN(0);
    } else {
      r0 = ((object**) lenv) [1][3];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object***) lenv) [1][1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = (object*) &constructed_999;
              r0 = box((pxll_int)((pxll_string*)(r0))->len);
              r1 = (object*) &constructed_999;
              r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = (object*) &constructed_1001;
              r0 = box((pxll_int)((pxll_string*)(r0))->len);
              r1 = (object*) &constructed_1001;
              (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object****) lenv) [1][1][1][2];
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = top[95];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L558; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_pp_437;
              L558:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_7143
              goto L559;
              FUN_lambda_7143:
                check_heap (0);
                r0 = top[158];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L560; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L560:
                ; lenv = k[2]; k = k[1];
                r0 = allocate (TC_TUPLE, 2);
                r1 = (object *) 3;
                r2 = ((object*****) lenv) [1][1][1][1][2];
                r1 = box((pxll_int)unbox(r2)+unbox(r1));
                r0[2] = r1;
                r1 = top[92];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L561; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_indent_434;
                L561:
                ; lenv = k[2]; k = k[1];
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object *) 3;
                r2 = ((object*****) lenv) [1][1][1][1][2];
                r1 = box((pxll_int)unbox(r2)+unbox(r1));
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r0[3] = r1;
                r1 = top[95];
                r0[1] = r1[2]; lenv = r0; goto FUN_pp_437;
              L559:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_7143; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object*) lenv) [3];
              r0[3] = r1;
              r1 = top[35];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L562; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
              L562:
              ; lenv = k[2]; k = k[1];
              r0 = (object*) &constructed_1000;
              r0 = box((pxll_int)((pxll_string*)(r0))->len);
              r1 = (object*) &constructed_1000;
              r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_303;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_303_over;
      fatbar_303:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object***) lenv) [1][1][3];
      r1[2] = r2;
      r2 = top[91];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L563; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_repr_433;
      L563:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      PXLL_RETURN(0);
      fatbar_303_over: ;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L555:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pp_437; r1[2] = lenv;
  r0[95] = r1;
  // def matches-list?_439
  goto L564;
  FUN_matches_list_question_439:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = (object *) 262;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_301;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = (object *) constructed_66[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_NIL): {
                      r0 = allocate (TC_TUPLE, 3);
                      // def lambda_7233
                      goto L565;
                      FUN_lambda_7233:
                        check_heap (0);
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = ((object****) lenv) [1][1][1][2];
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        r0[3] = r1;
                        r1 = top[97];
                        r0[1] = r1[2]; lenv = r0; goto FUN_matches_pattern_question_441;
                      L565:
                      r1 = allocate (TC_CLOSURE, 2);
                      r1[1] = &&FUN_lambda_7233; r1[2] = lenv;
                      r0[2] = r1;
                      r1 = ((object****) lenv) [1][1][1][3];
                      r0[3] = r1;
                      r1 = top[40];
                      r0[1] = r1[2]; lenv = r0; goto FUN_every_question_334;
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_302;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  lenv = ((object **)lenv)[1][1];
                  goto fatbar_302;
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_302;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_302;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_302_over;
        fatbar_302:
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = top[97];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L566; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_matches_pattern_question_441;
            L566:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = ((object*) lenv) [3];
              r1 = ((object**) lenv) [1][3];
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN_matches_list_question_439;
            } else {
              r0 = (object *) 6;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_301;
          }
        }
        PXLL_RETURN(0);
        fatbar_302_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_301_over;
    fatbar_301:
    r0 = (object *) 6;
    PXLL_RETURN(0);
    fatbar_301_over: ;
    PXLL_RETURN(0);
  L564:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_matches_list_question_439; r1[2] = lenv;
  r0[96] = r1;
  // def matches-pattern?_441
  goto L567;
  FUN_matches_pattern_question_441:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object***) lenv) [1][1][2];
        r2 = UOBJ_GET(r2,0);
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        lenv = lenv[1];
        r0[2] = r1;
        r1 = (object *) 1;
        r2 = ((object*) lenv) [2];
        range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
        r1 = (object *) 1;
        r2 = ((object*) lenv) [2];
        r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
        r0[3] = r1;
        r0 = (object *) 15362;
        r1 = ((object*) lenv) [3];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = ((object**) lenv) [1][3];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = ((object***) lenv) [1][1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object***) lenv) [1][1][2];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              lenv = lenv[1];
              r0[3] = r1;
              r1 = ((object*) lenv) [3];
              r1 = box((pxll_int)((pxll_string*)(r1))->len);
              r0[4] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = (object *) 3;
              r2 = ((object**) lenv) [1][4];
              r1 = box((pxll_int)unbox(r2)-unbox(r1));
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][3];
              range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][3];
              r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
              r0[3] = r1;
              r0 = (object *) 15874;
              r1 = ((object*) lenv) [3];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 4);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 4);
                r2 = ((object**) lenv) [1][3];
                r1[2] = r2;
                r2 = (object *) 3;
                r1[3] = r2;
                r2 = (object *) 3;
                r3 = ((object**) lenv) [1][4];
                r2 = box((pxll_int)unbox(r3)-unbox(r2));
                r1[4] = r2;
                r2 = top[45];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L568; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_substring_345;
                L568:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object***) lenv) [1][1][2];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                lenv = lenv[1];
                r0[3] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*) lenv) [2];
                r1[2] = r2;
                r2 = ((object*) lenv) [3];
                r1[3] = r2;
                r2 = top[50];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L569; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
                L569:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[4] = r1;
                r0 = (object *) 1;
                r1 = ((object*) lenv) [4];
                r0 = PXLL_TEST(unbox(r1)==unbox(r0));
                PXLL_RETURN(0);
              } else {
                r0 = (object *) 262;
                PXLL_RETURN(0);
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_300;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_300_over;
          fatbar_300:
          r0 = (object *) 6;
          PXLL_RETURN(0);
          fatbar_300_over: ;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) 262;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = top[96];
            r0[1] = r1[2]; lenv = r0; goto FUN_matches_list_question_439;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_299;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_299;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_299_over;
    fatbar_299:
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(r1==r0);
    PXLL_RETURN(0);
    fatbar_299_over: ;
    PXLL_RETURN(0);
  L567:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_matches_pattern_question_441; r1[2] = lenv;
  r0[97] = r1;
  // def get-ellipsis-nestings_442
  goto L570;
  FUN_get_ellipsis_nestings_442:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def dolist_1480
    goto L571;
    FUN_dolist_1480:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = (object *) constructed_66[0];
                  r1 = ((object*) lenv) [2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = ((object**) lenv) [1][3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 4);
                        r0[1] = lenv; lenv = r0;
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = ((object****) lenv) [1][1][1][2];
                        r1[2] = r2;
                        r2 = top[98];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L572; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_get_ellipsis_nestings_442;
                        L572:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        t = alloc_no_clear (UOTAG(0),1);
                        t[1] = r1;
                        r1 = t;
                        r0[3] = r1;
                        r1 = (object*)TC_NIL;
                        r0[4] = r1;
                        r0 = ((object*) lenv) [4];
                        r1 = ((object*) lenv) [3];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object **)lenv)[1][1];
                        goto fatbar_298;
                      }
                    }
                    PXLL_RETURN(0);
                  } else {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_298;
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_298;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_298;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_298_over;
          fatbar_298:
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = top[98];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L573; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_get_ellipsis_nestings_442;
          L573:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L574; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_dolist_1480;
          L574:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L575; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L575:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          fatbar_298_over: ;
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_297;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_297_over;
      fatbar_297:
      r0 = (object *) constructed_998[0];
      PXLL_RETURN(0);
      fatbar_297_over: ;
      PXLL_RETURN(0);
    L571:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_dolist_1480; r1[2] = lenv;
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN_dolist_1480;
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = (object*)TC_NIL;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object***) lenv) [1][1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_296;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_296_over;
    fatbar_296:
    r0 = (object *) constructed_997[0];
    PXLL_RETURN(0);
    fatbar_296_over: ;
    PXLL_RETURN(0);
  L570:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_get_ellipsis_nestings_442; r1[2] = lenv;
  r0[98] = r1;
  // def intersect?_443
  goto L576;
  FUN_intersect_question_443:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            r0 = PXLL_TEST(r1==r0);
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_295;
          }
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 3);
            // def lambda_7497
            goto L577;
            FUN_lambda_7497:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_7493
              goto L578;
              FUN_lambda_7493:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object**) lenv) [1][2];
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r0[3] = r1;
                r1 = top[99];
                r0[1] = r1[2]; lenv = r0; goto FUN_intersect_question_443;
              L578:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_7493; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[3] = r1;
              r1 = top[39];
              r0[1] = r1[2]; lenv = r0; goto FUN_some_question_333;
            L577:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_lambda_7497; r1[2] = lenv;
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = top[39];
            r0[1] = r1[2]; lenv = r0; goto FUN_some_question_333;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_295;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_295;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_295_over;
    fatbar_295:
    r0 = (object *) 6;
    PXLL_RETURN(0);
    fatbar_295_over: ;
    PXLL_RETURN(0);
  L576:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_intersect_question_443; r1[2] = lenv;
  r0[99] = r1;
  // def ellipsis-sub-envs_444
  goto L579;
  FUN_ellipsis_sub_envs_444:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_995[0];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = ((object*****) lenv) [1][1][1][1][2];
                        t = alloc_no_clear (UOTAG(0),1);
                        t[1] = r1;
                        r1 = t;
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][2];
                        r0[3] = r1;
                        r1 = top[99];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L580; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_intersect_question_443;
                        L580:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                        if PXLL_IS_TRUE(r0) {
                          r0 = ((object*) lenv) [2];
                          PXLL_RETURN(0);
                        } else {
                          r0 = ((object****) lenv) [1][1][1][3];
                          r1 = ((object*****) lenv) [1][1][1][1][2];
                          lenv[2] = r1;
                          lenv[3] = r0;
                          goto FUN_ellipsis_sub_envs_444;
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object ****)lenv)[1][1][1][1];
                        goto fatbar_294;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object ***)lenv)[1][1][1];
                    goto fatbar_294;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_294;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_294;
          }
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_294_over;
    fatbar_294:
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_996;
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = (object*) &constructed_996;
    PXLL_UNDEFINED; result=r0; goto Lreturn;
    PXLL_RETURN(0);
    fatbar_294_over: ;
    PXLL_RETURN(0);
  L579:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_ellipsis_sub_envs_444; r1[2] = lenv;
  r0[100] = r1;
  // def get-bindings_445
  goto L581;
  FUN_get_bindings_445:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def dolist_1508
    goto L582;
    FUN_dolist_1508:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = (object *) constructed_66[0];
                  r1 = ((object*) lenv) [2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = ((object**) lenv) [1][3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 10);
                        r0[1] = lenv; lenv = r0;
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = ((object****) lenv) [1][1][1][2];
                        r1[2] = r2;
                        r2 = top[98];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L583; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_get_ellipsis_nestings_442;
                        L583:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        t = alloc_no_clear (UOTAG(0),1);
                        t[1] = r1;
                        r1 = t;
                        r0[3] = r1;
                        r1 = allocate (TC_TUPLE, 3);
                        // def lambda_7613
                        goto L584;
                        FUN_lambda_7613:
                          check_heap (0);
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object******) lenv) [1][1][1][1][1][2];
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][2];
                          r1[3] = r2;
                          r2 = top[101];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L585; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_get_bindings_445;
                          L585:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r0 = ((object*) lenv) [2];
                          t = alloc_no_clear (UOTAG(0),1);
                          t[1] = r0;
                          r0 = t;
                          PXLL_RETURN(0);
                        L584:
                        r2 = allocate (TC_CLOSURE, 2);
                        r2[1] = &&FUN_lambda_7613; r2[2] = lenv;
                        r1[2] = r2;
                        r2 = ((object*****) lenv) [1][1][1][1][3];
                        r1[3] = r2;
                        r2 = top[32];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L586; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                        L586:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[4] = r1;
                        r1 = ((object*) lenv) [4];
                        t = alloc_no_clear (UOTAG(0),1);
                        t[1] = r1;
                        r1 = t;
                        r0[5] = r1;
                        r1 = (object*)TC_NIL;
                        r0[6] = r1;
                        r1 = ((object*) lenv) [6];
                        r2 = ((object*) lenv) [5];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r2;
                        t[2] = r1;
                        r1 = t;
                        r0[7] = r1;
                        r1 = ((object*) lenv) [7];
                        r2 = ((object*) lenv) [3];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r2;
                        t[2] = r1;
                        r1 = t;
                        r0[8] = r1;
                        r1 = ((object*) lenv) [8];
                        t = alloc_no_clear (UOTAG(0),1);
                        t[1] = r1;
                        r1 = t;
                        r0[9] = r1;
                        r1 = (object*)TC_NIL;
                        r0[10] = r1;
                        r0 = ((object*) lenv) [10];
                        r1 = ((object*) lenv) [9];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object **)lenv)[1][1];
                        goto fatbar_293;
                      }
                    }
                    PXLL_RETURN(0);
                  } else {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_293;
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_293;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_293;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_293_over;
          fatbar_293:
          r0 = ((object**) lenv) [1][3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 5);
              r0[1] = lenv; lenv = r0;
              r1 = ((object***) lenv) [1][1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object***) lenv) [1][1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object**) lenv) [1][2];
              r1[2] = r2;
              r2 = ((object*) lenv) [2];
              r1[3] = r2;
              r2 = top[101];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L587; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_get_bindings_445;
              L587:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[4] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object**) lenv) [1][3];
              r1[2] = r2;
              r2 = ((object*) lenv) [3];
              r1[3] = r2;
              r2 = ((object****) lenv) [1][1][1][2];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L588; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_dolist_1508;
              L588:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[5] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*) lenv) [4];
              r1[2] = r2;
              r2 = (object *) constructed_36[0];
              r1[3] = r2;
              r2 = top[21];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L589; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
              L589:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object*) lenv) [5];
              r0[3] = r1;
              r1 = top[21];
              r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_292;
            }
          }
          PXLL_RETURN(0);
          fatbar_293_over: ;
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_292;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_292_over;
      fatbar_292:
      r0 = (object *) constructed_994[0];
      PXLL_RETURN(0);
      fatbar_292_over: ;
      PXLL_RETURN(0);
    L582:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_dolist_1508; r1[2] = lenv;
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 7);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = (object*)TC_NIL;
        r0[3] = r1;
        r1 = ((object*) lenv) [3];
        r2 = ((object***) lenv) [1][1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[4] = r1;
        r1 = ((object*) lenv) [4];
        r2 = ((object***) lenv) [1][1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        r0[6] = r1;
        r1 = (object*)TC_NIL;
        r0[7] = r1;
        r0 = ((object*) lenv) [7];
        r1 = ((object*) lenv) [6];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object***) lenv) [1][1][3];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][2];
            r0[1] = r1[2]; lenv = r0; goto FUN_dolist_1508;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_291;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_291;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_291_over;
    fatbar_291:
    r0 = (object *) constructed_993[0];
    PXLL_RETURN(0);
    fatbar_291_over: ;
    PXLL_RETURN(0);
  L581:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_get_bindings_445; r1[2] = lenv;
  r0[101] = r1;
  // def expand-pattern_447
  goto L590;
  FUN_expand_pattern_447:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[103];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L591; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_expand_list_448;
        L591:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def loop_1524_i709
        goto L592;
        FUN_loop_1524_i709:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = (object*)UITAG(1);
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(0)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(1)): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = ((object**) lenv) [1][2];
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                    r0 = ((object*) lenv) [2];
                                    t = alloc_no_clear (UOTAG(0),1);
                                    t[1] = r0;
                                    r0 = t;
                                    PXLL_RETURN(0);
                                  } else {
                                    r0 = ((object*****) lenv) [1][1][1][1][3];
                                    lenv = ((object *****)lenv)[1][1][1][1][1];
                                    lenv[2] = r0;
                                    goto FUN_loop_1524_i709;
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object ****)lenv)[1][1][1][1];
                                  goto fatbar_290;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_290;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_290;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_290;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_290;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_290_over;
              fatbar_290:
              r0 = ((object*) lenv) [3];
              lenv = ((object *)lenv)[1];
              lenv[2] = r0;
              goto FUN_loop_1524_i709;
              fatbar_290_over: ;
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        L592:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_loop_1524_i709; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][3];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L593; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_loop_1524_i709;
        L593:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = ((object**) lenv) [1][2];
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_289;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_289_over;
    fatbar_289:
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
    fatbar_289_over: ;
    PXLL_RETURN(0);
  L590:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_expand_pattern_447; r1[2] = lenv;
  r0[102] = r1;
  // def expand-list_448
  goto L594;
  FUN_expand_list_448:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def sexp/append_1542
    goto L595;
    FUN_sexp_2fappend_1542:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L596; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L596:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } break;
        default: {
          goto fatbar_288;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_288_over;
      fatbar_288:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_992;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_992;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_288_over: ;
      PXLL_RETURN(0);
    L595:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_sexp_2fappend_1542; r1[2] = lenv;
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = (object *) constructed_66[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 5);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object****) lenv) [1][1][1][2];
                  r1[2] = r2;
                  r2 = top[98];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L597; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_get_ellipsis_nestings_442;
                  L597:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object*) lenv) [2];
                  r1[2] = r2;
                  r2 = ((object******) lenv) [1][1][1][1][1][3];
                  r1[3] = r2;
                  r2 = top[100];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L598; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_ellipsis_sub_envs_444;
                  L598:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (UOTAG(0)): {
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r1 = allocate (TC_TUPLE, 3);
                      // def lambda_7887
                      goto L599;
                      FUN_lambda_7887:
                        check_heap (0);
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = ((object******) lenv) [1][1][1][1][1][2];
                        r0[2] = r1;
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = ((object*) lenv) [2];
                        r1[2] = r2;
                        r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                        r1[3] = r2;
                        r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L600; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_sexp_2fappend_1542;
                        L600:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[3] = r1;
                        r1 = top[102];
                        r0[1] = r1[2]; lenv = r0; goto FUN_expand_pattern_447;
                      L599:
                      r2 = allocate (TC_CLOSURE, 2);
                      r2[1] = &&FUN_lambda_7887; r2[2] = lenv;
                      r1[2] = r2;
                      r2 = ((object*) lenv) [2];
                      r1[3] = r2;
                      r2 = top[32];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L601; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                      L601:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      lenv = lenv[1];
                    } break;
                    default: {
                      goto fatbar_287;
                    }
                  }
                  goto fatbar_287_over;
                  fatbar_287:
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = (object*) &constructed_991;
                  r1[2] = r2;
                  r2 = ((object*) lenv) [3];
                  r1[3] = r2;
                  r2 = top[15];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L602; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                  L602:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  fatbar_287_over: ;
                  r0[4] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object***) lenv) [1][1][3];
                  r1[2] = r2;
                  r2 = ((object******) lenv) [1][1][1][1][1][3];
                  r1[3] = r2;
                  r2 = top[103];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L603; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_expand_list_448;
                  L603:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[5] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object*) lenv) [4];
                  r1[2] = r2;
                  r2 = (object *) constructed_36[0];
                  r1[3] = r2;
                  r2 = top[21];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L604; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                  L604:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object*) lenv) [5];
                  r0[3] = r1;
                  r1 = top[21];
                  r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                } else {
                  lenv = ((object **)lenv)[1][1];
                  goto fatbar_286;
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_286;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_286;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_286_over;
        fatbar_286:
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][3];
        r1[3] = r2;
        r2 = top[102];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L605; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_expand_pattern_447;
        L605:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][3];
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][3];
        r1[3] = r2;
        r2 = top[103];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L606; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_expand_list_448;
        L606:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
        fatbar_286_over: ;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_285;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_285_over;
    fatbar_285:
    r0 = ((object**) lenv) [1][2];
    PXLL_RETURN(0);
    fatbar_285_over: ;
    PXLL_RETURN(0);
  L594:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_expand_list_448; r1[2] = lenv;
  r0[103] = r1;
  // def pred_451
  goto L607;
  FUN_pred_451:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (UOTAG(23), 3);
    r2 = (object*)UITAG(1);
    r1[1] = r2;
    r2 = (object*)UITAG(1);
    r1[2] = r2;
    r2 = (object *) 6;
    r1[3] = r2;
    r0[2] = r1;
    r0 = ((object**) lenv) [1][3];
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][2];
    t = alloc_no_clear (UOTAG(1),3);
    t[1] = r2;
    t[2] = r0;
    t[3] = r1;
    r0 = t;
    PXLL_RETURN(0);
  L607:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pred_451; r1[2] = lenv;
  r0[104] = r1;
  // def new-tvar_453
  goto L608;
  FUN_new_tvar_453:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = top[180];
    r1 = ((pxll_vector*)r1)->val[1];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L609; t[4] = r0; k = t;
    lenv = r1[2]; goto *r1[1];
    L609:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (UOTAG(23), 3);
    r2 = (object*)UITAG(1);
    r1[1] = r2;
    r2 = (object*)UITAG(1);
    r1[2] = r2;
    r2 = (object *) 6;
    r1[3] = r2;
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(0),2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L608:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_new_tvar_453; r1[2] = lenv;
  r0[105] = r1;
  // def rlabel_458
  goto L610;
  FUN_rlabel_458:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) constructed_819[0];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 4);
    r1[1] = lenv; lenv = r1;
    r2 = (object*)TC_NIL;
    r1[2] = r2;
    r2 = ((object*) lenv) [2];
    r3 = ((object**) lenv) [1][4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    r1[3] = r2;
    r2 = ((object*) lenv) [3];
    r3 = ((object**) lenv) [1][3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    r1[4] = r2;
    r1 = ((object*) lenv) [4];
    r2 = ((object**) lenv) [1][2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    lenv = lenv[1];
    r0[3] = r1;
    r1 = top[104];
    r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
  L610:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_rlabel_458; r1[2] = lenv;
  r0[106] = r1;
  // def type-repr_462
  goto L611;
  FUN_type_repr_462:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 4);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[56];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L612; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
        L612:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r1 = ((object*) lenv) [4];
        r2 = (object*) &constructed_983;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) constructed_97[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [3];
                  switch (get_case (r0)) {
                    case (TC_NIL): {
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = allocate (TC_TUPLE, 6);
                      r1[1] = lenv; lenv = r1;
                      r2 = allocate (TC_TUPLE, 2);
                      r3 = ((object**) lenv) [1][2];
                      r2[2] = r3;
                      r3 = top[107];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L613; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_type_repr_462;
                      L613:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[2] = r2;
                      r2 = allocate (TC_TUPLE, 2);
                      r3 = ((object***) lenv) [1][1][2];
                      r2[2] = r3;
                      r3 = top[107];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L614; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_type_repr_462;
                      L614:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[3] = r2;
                      r2 = (object*)TC_NIL;
                      r1[4] = r2;
                      r2 = ((object*) lenv) [4];
                      r3 = ((object*) lenv) [3];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[5] = r2;
                      r2 = ((object*) lenv) [5];
                      r3 = (object*) &constructed_984;
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[6] = r2;
                      r1 = ((object*) lenv) [6];
                      r2 = ((object*) lenv) [2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r2;
                      t[2] = r1;
                      r1 = t;
                      lenv = lenv[1];
                      r0[2] = r1;
                      r1 = top[47];
                      r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_283;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_283;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_283_over;
              fatbar_283:
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 7);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = allocate (TC_TUPLE, 3);
              r4 = top[107];
              r3[2] = r4;
              r4 = ((object**) lenv) [1][3];
              r3[3] = r4;
              r4 = top[32];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L615; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
              L615:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = (object*) &constructed_987;
              r2[3] = r3;
              r3 = top[48];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L616; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
              L616:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = top[107];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L617; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_type_repr_462;
              L617:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r2 = (object*)TC_NIL;
              r1[4] = r2;
              r2 = ((object*) lenv) [4];
              r3 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[5] = r2;
              r2 = ((object*) lenv) [5];
              r3 = (object*) &constructed_986;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[6] = r2;
              r2 = ((object*) lenv) [6];
              r3 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[7] = r2;
              r1 = ((object*) lenv) [7];
              r2 = (object*) &constructed_985;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              r0[2] = r1;
              r1 = top[47];
              r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
              fatbar_283_over: ;
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_282;
            }
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_282;
        }
        PXLL_RETURN(0);
        goto fatbar_282_over;
        fatbar_282:
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = ((object*) lenv) [2];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_284;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_284_over;
        fatbar_284:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 7);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object***) lenv) [1][1][2];
        r3 = UOBJ_GET(r3,0);
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = top[107];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][3];
        r3[3] = r4;
        r4 = top[32];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L618; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
        L618:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*) &constructed_990;
        r2[3] = r3;
        r3 = top[48];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L619; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
        L619:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = (object*) &constructed_989;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r2 = ((object*) lenv) [6];
        r3 = (object*) &constructed_988;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[7] = r2;
        r1 = ((object*) lenv) [7];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        fatbar_284_over: ;
        PXLL_RETURN(0);
        fatbar_282_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L611:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_type_repr_462; r1[2] = lenv;
  r0[107] = r1;
  // def type->trec_464
  goto L620;
  FUN_type__3etrec_464:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L620:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_type__3etrec_464; r1[2] = lenv;
  r0[108] = r1;
  // def type-find_465
  goto L621;
  FUN_type_find_465:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[108];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L622; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_type__3etrec_464;
    L622:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[1];
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = ((object**) lenv) [1][2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[109];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L623; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_type_find_465;
        L623:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        r1 = ((object**) lenv) [1][2];
        ((pxll_vector*)r1)->val[1] = r0;
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L621:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_type_find_465; r1[2] = lenv;
  r0[109] = r1;
  // def type-union_466
  goto L624;
  FUN_type_union_466:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[109];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L625; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_type_find_465;
    L625:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][3];
    r1[2] = r2;
    r2 = top[109];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L626; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_type_find_465;
    L626:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = ((object**) lenv) [1][3];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            r1 = ((object*) lenv) [2];
            ((pxll_vector*)r1)->val[1] = r0;
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,1);
            r0[2] = r1;
            r0 = ((object***) lenv) [1][1][2];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            r1 = ((object*) lenv) [2];
            ((pxll_vector*)r1)->val[1] = r0;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,1);
            r0[2] = r1;
            r0 = ((object***) lenv) [1][1][2];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            r1 = ((object*) lenv) [2];
            ((pxll_vector*)r1)->val[1] = r0;
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][3];
            r1 = UOBJ_GET(r1,2);
            r0[2] = r1;
            r0 = ((object***) lenv) [1][1][2];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            r1 = ((object*) lenv) [2];
            ((pxll_vector*)r1)->val[1] = r0;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L624:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_type_union_466; r1[2] = lenv;
  r0[110] = r1;
  // def parse-type*_467
  goto L627;
  FUN_parse_type_splat_467:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 5);
    r0[1] = lenv; lenv = r0;
    // def arrow-type?_1631
    goto L628;
    FUN_arrow_type_question_1631:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = (object *) 6;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = (object *) constructed_660[0];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = (object *) 262;
                PXLL_RETURN(0);
              } else {
                lenv = ((object *)lenv)[1];
                goto fatbar_281;
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_281;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_281_over;
          fatbar_281:
          r0 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          goto FUN_arrow_type_question_1631;
          fatbar_281_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L628:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_arrow_type_question_1631; r1[2] = lenv;
    r0[2] = r1;
    // def parse-list_1634
    goto L629;
    FUN_parse_list_1634:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = (object *) constructed_870[0];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = ((object**) lenv) [1][3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object***) lenv) [1][1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object***) lenv) [1][1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [2];
                    switch (get_case (r0)) {
                      case (UOTAG(1)): {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r0 = ((object**) lenv) [1][3];
                        switch (get_case (r0)) {
                          case (TC_NIL): {
                            r0 = allocate (TC_TUPLE, 2);
                            r0[1] = lenv; lenv = r0;
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                            r1[2] = r2;
                            r2 = ((object**) lenv) [1][2];
                            r1[3] = r2;
                            r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                            r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
                            r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,11)];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L630; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto *r2[1];
                            L630:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r0 = ((object*) lenv) [2];
                            switch (get_case (r0)) {
                              case (UOTAG(0)): {
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object**) lenv) [1][2];
                                r1 = UOBJ_GET(r1,0);
                                r0[2] = r1;
                                r0 = ((object*) lenv) [2];
                                PXLL_RETURN(0);
                              } break;
                              default: {
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = top[105];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L631; t[4] = r0; k = t;
                                lenv = r1[2]; goto FUN_new_tvar_453;
                                L631:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                r0[2] = r1;
                                r0 = allocate (TC_TUPLE, 4);
                                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                                r0[2] = r1;
                                r1 = ((object***) lenv) [1][1][2];
                                r0[3] = r1;
                                r1 = ((object*) lenv) [2];
                                r0[4] = r1;
                                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                                r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
                                r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,10)];
                                t = allocate (TC_SAVE, 3);
                                t[1] = k; t[2] = lenv; t[3] = &&L632; ; k = t;
                                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                L632:
                                ; lenv = k[2]; k = k[1];
                                r0 = ((object*) lenv) [2];
                                PXLL_RETURN(0);
                              } break;
                            }
                            PXLL_RETURN(0);
                          } break;
                          default: {
                            lenv = ((object **)lenv)[1][1];
                            goto fatbar_278;
                          }
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object *)lenv)[1];
                        goto fatbar_278;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    goto fatbar_278;
                  }
                }
                PXLL_RETURN(0);
                goto fatbar_278_over;
                fatbar_278:
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object*) &constructed_979;
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r0[3] = r1;
                r1 = top[15];
                r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                fatbar_278_over: ;
                PXLL_RETURN(0);
              } else {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_277;
              }
              PXLL_RETURN(0);
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_277;
            }
          }
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_277;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_277_over;
      fatbar_277:
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L633; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_arrow_type_question_1631;
      L633:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        // def loop_1646_i772
        goto L634;
        FUN_loop_1646_i772:
          check_heap (0);
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = (object *) constructed_660[0];
                  r1 = ((object*) lenv) [2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = ((object**) lenv) [1][3];
                    switch (get_case (r0)) {
                      case (TC_PAIR): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object***) lenv) [1][1][3];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object***) lenv) [1][1][3];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = ((object*) lenv) [3];
                        switch (get_case (r0)) {
                          case (TC_NIL): {
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = allocate (TC_TUPLE, 2);
                            r2 = ((object**) lenv) [1][2];
                            r1[2] = r2;
                            r2 = ((object********) lenv) [1][1][1][1][1][1][1][5];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L635; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_parse_1636;
                            L635:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = ((object*****) lenv) [1][1][1][1][2];
                            r1[2] = r2;
                            r2 = (object *) constructed_36[0];
                            r1[3] = r2;
                            r2 = top[21];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L636; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                            L636:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[3] = r1;
                            r0 = allocate (TC_TUPLE, 3);
                            r1 = (object *) constructed_97[0];
                            r0[2] = r1;
                            r1 = ((object*) lenv) [3];
                            r2 = ((object*) lenv) [2];
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r2;
                            t[2] = r1;
                            r1 = t;
                            r0[3] = r1;
                            r1 = top[104];
                            r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
                          } break;
                          default: {
                            lenv = ((object **)lenv)[1][1];
                            goto fatbar_279;
                          }
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object *)lenv)[1];
                        goto fatbar_279;
                      }
                    }
                    PXLL_RETURN(0);
                  } else {
                    lenv = ((object *)lenv)[1];
                    goto fatbar_279;
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_279;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_279_over;
              fatbar_279:
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object**) lenv) [1][2];
              r1[2] = r2;
              r2 = ((object******) lenv) [1][1][1][1][1][5];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L637; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_parse_1636;
              L637:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              r1 = ((object*) lenv) [3];
              lenv = ((object *)lenv)[1];
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_1646_i772;
              fatbar_279_over: ;
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r1 = (object*) &constructed_981;
              r0[2] = r1;
              r1 = ((object***) lenv) [1][1][2];
              r0[3] = r1;
              r1 = top[15];
              r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            } break;
          }
          PXLL_RETURN(0);
        L634:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_loop_1646_i772; r1[2] = lenv;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object *) constructed_980[0];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r0[1] = r1[2]; lenv = r0; goto FUN_loop_1646_i772;
      } else {
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object*) lenv) [2];
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object****) lenv) [1][1][1][5];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][3];
                r1[3] = r2;
                r2 = top[32];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L638; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                L638:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = top[104];
                r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_280;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_280;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_280_over;
        fatbar_280:
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_982;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        fatbar_280_over: ;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
      fatbar_277_over: ;
      PXLL_RETURN(0);
    L629:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_parse_list_1634; r1[2] = lenv;
    r0[3] = r1;
    // def parse-field-list_1635
    goto L639;
    FUN_parse_field_list_1635:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = (object *) constructed_66[0];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = ((object**) lenv) [1][3];
            switch (get_case (r0)) {
              case (TC_NIL): {
                r0 = top[105];
                lenv = r0[2]; goto FUN_new_tvar_453;
              } break;
              default: {
                goto fatbar_276;
              }
            }
            PXLL_RETURN(0);
          } else {
            goto fatbar_276;
          }
          PXLL_RETURN(0);
          goto fatbar_276_over;
          fatbar_276:
          r0 = allocate (TC_TUPLE, 4);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) constructed_546[0];
          r1[3] = r2;
          r2 = top[104];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L640; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
          L640:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object**) lenv) [1][3];
          r2[2] = r3;
          r3 = ((object*****) lenv) [1][1][1][1][5];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L641; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_parse_1636;
          L641:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object *) constructed_543[0];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[3] = r2;
          r2 = top[104];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L642; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
          L642:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L643; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_parse_field_list_1635;
          L643:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = top[106];
          r0[1] = r1[2]; lenv = r0; goto FUN_rlabel_458;
          fatbar_276_over: ;
          PXLL_RETURN(0);
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object *) constructed_560[0];
          r1[2] = r2;
          r2 = (object *) constructed_558[0];
          r1[3] = r2;
          r2 = top[104];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L644; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
          L644:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object *) constructed_557[0];
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          r2 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[3] = r1;
          r1 = top[104];
          r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
        } break;
      }
      PXLL_RETURN(0);
    L639:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_parse_field_list_1635; r1[2] = lenv;
    r0[4] = r1;
    // def parse_1636
    goto L645;
    FUN_parse_1636:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = top[187];
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[65];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L646; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
          L646:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object*) lenv) [2];
              r0[2] = r1;
              r1 = (object *) constructed_977[0];
              r0[3] = r1;
              r1 = top[104];
              r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
            } break;
          }
          PXLL_RETURN(0);
        } break;
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[1] = r1[2]; lenv = r0; goto FUN_parse_list_1634;
        } break;
        case (UOTAG(8)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = ((object***) lenv) [1][1][4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L647; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_parse_field_list_1635;
          L647:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object *) constructed_553[0];
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          r2 = ((object**) lenv) [1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[3] = r1;
          r1 = top[104];
          r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
        } break;
        default: {
          goto fatbar_275;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_275_over;
      fatbar_275:
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_978;
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][2];
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      fatbar_275_over: ;
      PXLL_RETURN(0);
    L645:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_parse_1636; r1[2] = lenv;
    r0[5] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [5];
    r0[1] = r1[2]; lenv = r0; goto FUN_parse_1636;
  L627:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_parse_type_splat_467; r1[2] = lenv;
  r0[111] = r1;
  // def get-record-sig_469
  goto L648;
  FUN_get_record_sig_469:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) constructed_553[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  // def loop_1682
                  goto L649;
                  FUN_loop_1682:
                    check_heap (0);
                    r0 = ((object*) lenv) [2];
                    switch (get_case (r0)) {
                      case (UOTAG(1)): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = (object *) constructed_557[0];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                          r0 = ((object**) lenv) [1][3];
                          PXLL_RETURN(0);
                        } else {
                          goto fatbar_273;
                        }
                        PXLL_RETURN(0);
                        goto fatbar_273_over;
                        fatbar_273:
                        r0 = (object *) constructed_817[0];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_PAIR): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,1);
                                  r0[3] = r1;
                                  r0 = ((object*) lenv) [3];
                                  switch (get_case (r0)) {
                                    case (TC_NIL): {
                                      r0 = ((object****) lenv) [1][1][1][3];
                                      r1 = ((object*) lenv) [2];
                                      lenv = ((object ***)lenv)[1][1][1];
                                      lenv[2] = r1;
                                      lenv[3] = r0;
                                      goto FUN_loop_1682;
                                    } break;
                                    default: {
                                      lenv = ((object **)lenv)[1][1];
                                      goto fatbar_274;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object *)lenv)[1];
                                  goto fatbar_274;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              goto fatbar_274;
                            }
                          }
                          PXLL_RETURN(0);
                        } else {
                          goto fatbar_274;
                        }
                        PXLL_RETURN(0);
                        goto fatbar_274_over;
                        fatbar_274:
                        r0 = (object *) constructed_819[0];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [2];
                              switch (get_case (r0)) {
                                case (UOTAG(1)): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][2];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r0 = ((object**) lenv) [1][3];
                                  switch (get_case (r0)) {
                                    case (TC_PAIR): {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object***) lenv) [1][1][3];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = ((object***) lenv) [1][1][3];
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r0 = ((object*) lenv) [2];
                                      switch (get_case (r0)) {
                                        case (UOTAG(1)): {
                                          r0 = allocate (TC_TUPLE, 2);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object**) lenv) [1][2];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r0 = (object *) constructed_543[0];
                                          r1 = ((object*) lenv) [2];
                                          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                            r0 = ((object**) lenv) [1][3];
                                            switch (get_case (r0)) {
                                              case (TC_PAIR): {
                                                r0 = allocate (TC_TUPLE, 3);
                                                r0[1] = lenv; lenv = r0;
                                                r1 = ((object***) lenv) [1][1][3];
                                                r1 = UOBJ_GET(r1,0);
                                                r0[2] = r1;
                                                r1 = ((object***) lenv) [1][1][3];
                                                r1 = UOBJ_GET(r1,1);
                                                r0[3] = r1;
                                                r0 = ((object*) lenv) [3];
                                                switch (get_case (r0)) {
                                                  case (TC_NIL): {
                                                    r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                                    r1 = ((object****) lenv) [1][1][1][2];
                                                    t = alloc_no_clear (TC_PAIR,2);
                                                    t[1] = r1;
                                                    t[2] = r0;
                                                    r0 = t;
                                                    r1 = ((object*) lenv) [2];
                                                    lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                                    lenv[2] = r1;
                                                    lenv[3] = r0;
                                                    goto FUN_loop_1682;
                                                  } break;
                                                  default: {
                                                    lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                                    goto fatbar_272;
                                                  }
                                                }
                                                PXLL_RETURN(0);
                                              } break;
                                              default: {
                                                lenv = ((object *****)lenv)[1][1][1][1][1];
                                                goto fatbar_272;
                                              }
                                            }
                                            PXLL_RETURN(0);
                                          } else {
                                            lenv = ((object *****)lenv)[1][1][1][1][1];
                                            goto fatbar_272;
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object ****)lenv)[1][1][1][1];
                                          goto fatbar_272;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      lenv = ((object ***)lenv)[1][1][1];
                                      goto fatbar_272;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object **)lenv)[1][1];
                                  goto fatbar_272;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object *)lenv)[1];
                              goto fatbar_272;
                            }
                          }
                          PXLL_RETURN(0);
                        } else {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_272;
                        }
                        PXLL_RETURN(0);
                        fatbar_274_over: ;
                        PXLL_RETURN(0);
                        fatbar_273_over: ;
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        r0 = ((object*) lenv) [3];
                        r1 = (object *) constructed_66[0];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        PXLL_RETURN(0);
                      } break;
                    }
                    PXLL_RETURN(0);
                    goto fatbar_272_over;
                    fatbar_272:
                    r0 = (object *) constructed_973[0];
                    PXLL_RETURN(0);
                    fatbar_272_over: ;
                    PXLL_RETURN(0);
                  L649:
                  r2 = allocate (TC_CLOSURE, 2);
                  r2[1] = &&FUN_loop_1682; r2[2] = lenv;
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object***) lenv) [1][1][2];
                  r1[2] = r2;
                  r2 = (object *) constructed_972[0];
                  r1[3] = r2;
                  r2 = ((object*) lenv) [2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L650; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_loop_1682;
                  L650:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  lenv = lenv[1];
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = top[81];
                  r0[2] = r1;
                  r1 = ((object*) lenv) [2];
                  r0[3] = r1;
                  r1 = top[44];
                  r0[1] = r1[2]; lenv = r0; goto FUN_sort_340;
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_271;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_271;
            }
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_271;
        }
        PXLL_RETURN(0);
        goto fatbar_271_over;
        fatbar_271:
        r0 = (object *) constructed_817[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [3];
                  switch (get_case (r0)) {
                    case (TC_NIL): {
                      r0 = ((object*) lenv) [2];
                      lenv[2] = r0;
                      goto FUN_get_record_sig_469;
                    } break;
                    default: {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_270;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object **)lenv)[1][1];
                  goto fatbar_270;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_270;
            }
          }
          PXLL_RETURN(0);
        } else {
          lenv = ((object *)lenv)[1];
          goto fatbar_270;
        }
        PXLL_RETURN(0);
        fatbar_271_over: ;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_270;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_270_over;
    fatbar_270:
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 5);
    r2[1] = lenv; lenv = r2;
    r3 = allocate (TC_TUPLE, 2);
    r4 = ((object***) lenv) [1][1][2];
    r3[2] = r4;
    r4 = top[107];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L651; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_type_repr_462;
    L651:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    r3 = (object*)TC_NIL;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r4 = (object*) &constructed_976;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[4] = r3;
    r3 = ((object*) lenv) [4];
    r4 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[5] = r3;
    r2 = ((object*) lenv) [5];
    r3 = (object*) &constructed_975;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L652; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L652:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = ((object*) lenv) [2];
    r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_974;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_270_over: ;
    PXLL_RETURN(0);
  L648:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_get_record_sig_469; r1[2] = lenv;
  r0[112] = r1;
  // def make-counter_472
  goto L653;
  FUN_make_counter_472:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    // def inc_1712
    goto L654;
    FUN_inc_1712:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object****) lenv) [1][1][1][2];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = (object *) 3;
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      lenv = lenv[1];
      ((object***) lenv) [1][1][2] = r0;
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
    L654:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_inc_1712; r1[2] = lenv;
    r0[2] = r1;
    // def get_1713
    goto L655;
    FUN_get_1713:
      r0 = ((object**) lenv) [1][2];
      PXLL_RETURN(0);
    L655:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_get_1713; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (UOTAG(22), 2);
    r1 = ((object*) lenv) [3];
    r0[1] = r1;
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    PXLL_RETURN(0);
  L653:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_make_counter_472; r1[2] = lenv;
  r0[113] = r1;
  // def new-match-var_476
  goto L656;
  FUN_new_match_var_476:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 4);
    r2[1] = lenv; lenv = r2;
    r3 = allocate (TC_TUPLE, 2);
    r4 = top[190];
    r4 = ((pxll_vector*)r4)->val[1];
    t = allocate (TC_SAVE, 7);
    t[1] = k; t[2] = lenv; t[3] = &&L657; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
    lenv = r4[2]; goto *r4[1];
    L657:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
    r4 = result;
    r3[2] = r4;
    r4 = top[56];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L658; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
    L658:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    r3 = (object*)TC_NIL;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r4 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[4] = r3;
    r2 = ((object*) lenv) [4];
    r3 = (object*) &constructed_971;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L659; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L659:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = top[80];
    r0[1] = r1[2]; lenv = r0; goto FUN_string__3esymbol_419;
  L656:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_new_match_var_476; r1[2] = lenv;
  r0[114] = r1;
  // def notdotdotdot_477
  goto L660;
  FUN_notdotdotdot_477:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r0 = (object *) constructed_66[0];
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
      r0 = (object *) 6;
      PXLL_RETURN(0);
    } else {
      lenv = ((object *)lenv)[1];
      goto fatbar_269;
    }
    PXLL_RETURN(0);
    PXLL_RETURN(0);
    goto fatbar_269_over;
    fatbar_269:
    r0 = (object *) 262;
    PXLL_RETURN(0);
    fatbar_269_over: ;
    PXLL_RETURN(0);
  L660:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_notdotdotdot_477; r1[2] = lenv;
  r0[115] = r1;
  // def compile-pattern_478
  goto L661;
  FUN_compile_pattern_478:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 20);
    r0[1] = lenv; lenv = r0;
    // def remove-first-pat_1740
    goto L662;
    FUN_remove_first_pat_1740:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,1);
      r0[3] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [3];
          t = alloc_no_clear (UOTAG(0),2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } break;
        default: {
          lenv = ((object *)lenv)[1];
          goto fatbar_268;
        }
      }
      PXLL_RETURN(0);
      PXLL_RETURN(0);
      goto fatbar_268_over;
      fatbar_268:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_970;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_970;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_268_over: ;
      PXLL_RETURN(0);
    L662:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_remove_first_pat_1740; r1[2] = lenv;
    r0[2] = r1;
    // def first-pattern-kind_1741
    goto L663;
    FUN_first_pattern_kind_1741:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = (object *) constructed_868[0];
              PXLL_RETURN(0);
            } break;
            case (UOTAG(1)): {
              r0 = (object *) constructed_964[0];
              PXLL_RETURN(0);
            } break;
            case (UOTAG(2)): {
              r0 = (object *) constructed_474[0];
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = (object *) constructed_958[0];
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
        default: {
          lenv = ((object *)lenv)[1];
          goto fatbar_267;
        }
      }
      PXLL_RETURN(0);
      PXLL_RETURN(0);
      goto fatbar_267_over;
      fatbar_267:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_969;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_969;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_267_over: ;
      PXLL_RETURN(0);
    L663:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_first_pattern_kind_1741; r1[2] = lenv;
    r0[3] = r1;
    // def compare-first-patterns_1742
    goto L664;
    FUN_compare_first_patterns_1742:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = ((object***) lenv) [1][1][3];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L665; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_first_pattern_kind_1741;
      L665:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][3];
      r1[2] = r2;
      r2 = ((object***) lenv) [1][1][3];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L666; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_first_pattern_kind_1741;
      L666:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(r1==r0);
      PXLL_RETURN(0);
    L664:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_compare_first_patterns_1742; r1[2] = lenv;
    r0[4] = r1;
    // def compile-match_1743
    goto L667;
    FUN_compile_match_1743:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = ((object*) lenv) [4];
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              PXLL_RETURN(0);
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_266;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_266_over;
      fatbar_266:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][3];
      r1[2] = r2;
      r2 = ((object***) lenv) [1][1][4];
      r1[3] = r2;
      r2 = top[42];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L668; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_pack_337;
      L668:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[22];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L669; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
      L669:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 3;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object**) lenv) [1][2];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L670; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L670:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = ((object**) lenv) [1][4];
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][6];
        r0[1] = r1[2]; lenv = r0; goto FUN_compile_group_1744;
      } else {
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_9927
        goto L671;
        FUN_lambda_9927:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object***) lenv) [1][1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][4];
          r0[4] = r1;
          r1 = ((object****) lenv) [1][1][1][6];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L672; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_compile_group_1744;
          L672:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          ((object***) lenv) [1][1][4] = r0;
          PXLL_RETURN(0);
        L671:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_9927; r1[2] = lenv;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = (object *) constructed_36[0];
        r1[3] = r2;
        r2 = top[21];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L673; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
        L673:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = top[35];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L674; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        L674:
        ; lenv = k[2]; k = k[1];
        r0 = ((object**) lenv) [1][4];
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
      fatbar_266_over: ;
      PXLL_RETURN(0);
    L667:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_compile_match_1743; r1[2] = lenv;
    r0[5] = r1;
    // def compile-group_1744
    goto L675;
    FUN_compile_group_1744:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object**) lenv) [1][3];
      r2[2] = r3;
      r3 = top[19];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L676; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
      L676:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r2 = ((object***) lenv) [1][1][3];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L677; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_first_pattern_kind_1741;
      L677:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) constructed_958[0];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object****) lenv) [1][1][1][3];
        r3[2] = r4;
        r4 = top[19];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L678; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_car_309;
        L678:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object**) lenv) [1][2];
        r3 = UOBJ_GET(r3,0);
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        lenv = lenv[1];
        lenv = lenv[1];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L679; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L679:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object****) lenv) [1][1][1][9];
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object*) lenv) [2];
        r2[2] = r3;
        r3 = ((object****) lenv) [1][1][1][11];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L680; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_pattern__3efieldpairs_1750;
        L680:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L681; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L681:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_10201_i834
        goto L682;
        FUN_lambda_10201_i834:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = ((object***) lenv) [1][1][2];
          r3 = UOBJ_GET(r3,0);
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[19];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L683; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
          L683:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object******) lenv) [1][1][1][1][1][9];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object*) lenv) [2];
          r2[2] = r3;
          r3 = ((object******) lenv) [1][1][1][1][1][11];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L684; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pattern__3efieldpairs_1750;
          L684:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = top[32];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L685; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
          L685:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 4);
          r2 = top[14];
          r1[2] = r2;
          r2 = ((object***) lenv) [1][1][3];
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r1[4] = r2;
          r2 = top[41];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L686; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_every2_question_335;
          L686:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [4];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_956;
            r0[2] = r1;
            r1 = ((object****) lenv) [1][1][1][3];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
          } else {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        L682:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_10201_i834; r1[2] = lenv;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][3];
        r1[2] = r2;
        r2 = top[20];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L687; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
        L687:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = top[35];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L688; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        L688:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 13);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object****) lenv) [1][1][1][2];
        r1[2] = r2;
        r2 = (object *) 1;
        r1[3] = r2;
        r2 = top[27];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L689; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
        L689:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_10231_i834
        goto L690;
        FUN_lambda_10231_i834:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = ((object***) lenv) [1][1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object****) lenv) [1][1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[3] = r3;
          r3 = (object*)TC_NIL;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = (object*) &constructed_955;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = ((object*) lenv) [6];
          r3 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L691; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L691:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = top[80];
          r0[1] = r1[2]; lenv = r0; goto FUN_string__3esymbol_419;
        L690:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_10231_i834; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L692; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L692:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_10260_i834
        goto L693;
        FUN_lambda_10260_i834:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 6);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object***) lenv) [1][1][2];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*******) lenv) [1][1][1][1][1][1][10];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 2);
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object*) lenv) [2];
          r3[2] = r4;
          r4 = top[19];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L694; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_car_309;
          L694:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = ((object*******) lenv) [1][1][1][1][1][1][11];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L695; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pattern__3efieldpairs_1750;
          L695:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = top[32];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L696; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
          L696:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[20];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L697; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
          L697:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object*) lenv) [3];
          r2[2] = r3;
          r3 = (object *) constructed_36[0];
          r2[3] = r3;
          r3 = top[21];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L698; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
          L698:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r2 = ((object*) lenv) [4];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L699; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L699:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object***) lenv) [1][1][2];
          r2 = UOBJ_GET(r2,1);
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0[6] = r1;
          r0 = ((object*) lenv) [6];
          r1 = ((object*) lenv) [5];
          t = alloc_no_clear (UOTAG(0),2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        L693:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_10260_i834; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L700; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L700:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L701; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L701:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def $loop_1871_i834
        goto L702;
        FUN__24loop_1871_i834:
          check_heap (0);
          r0 = ((object***) lenv) [1][1][5];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [3];
            r0[2] = r1;
            r1 = (object *) constructed_36[0];
            r0[3] = r1;
            r1 = top[21];
            r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          } else {
            r0 = allocate (TC_TUPLE, 10);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][3];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L703; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L703:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[3] = r1;
            r1 = ((object****) lenv) [1][1][1][2];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*****) lenv) [1][1][1][1][3];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L704; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L704:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r1 = ((object*) lenv) [5];
            r2 = ((object*) lenv) [4];
            t = alloc_no_clear (UOTAG(10),2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[6] = r1;
            r1 = (object*)TC_NIL;
            r0[7] = r1;
            r1 = ((object*) lenv) [7];
            r2 = ((object*) lenv) [6];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[8] = r1;
            r1 = ((object*) lenv) [8];
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[9] = r1;
            r1 = ((object*) lenv) [9];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            r0[10] = r1;
            r0 = ((object**) lenv) [1][3];
            r1 = ((object*) lenv) [10];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = (object *) 3;
            r2 = ((object*) lenv) [2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN__24loop_1871_i834;
          }
          PXLL_RETURN(0);
        L702:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN__24loop_1871_i834; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 3);
        r2 = (object *) 1;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L705; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN__24loop_1871_i834;
        L705:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
        r0[6] = r1;
        r1 = (object *) constructed_622[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[7] = r1;
        r1 = ((object*) lenv) [6];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        r0[8] = r1;
        r1 = allocate (TC_TUPLE, 4);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object*****) lenv) [1][1][1][1][2];
        r3[2] = r4;
        r4 = top[20];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L706; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_cdr_310;
        L706:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = ((object**) lenv) [1][3];
        r3[2] = r4;
        r4 = (object *) constructed_36[0];
        r3[3] = r4;
        r4 = top[21];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L707; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
        L707:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = ((object*) lenv) [2];
        r2[3] = r3;
        r3 = top[21];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L708; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
        L708:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        lenv = lenv[1];
        r1[2] = r2;
        r2 = ((object*) lenv) [4];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        r1[4] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][5];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L709; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_compile_match_1743;
        L709:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[9] = r1;
        r1 = (object*)TC_NIL;
        r0[10] = r1;
        r1 = ((object*) lenv) [10];
        r2 = ((object*) lenv) [9];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[11] = r1;
        r1 = ((object*) lenv) [11];
        r2 = ((object*) lenv) [8];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[12] = r1;
        r1 = ((object*) lenv) [12];
        r2 = ((object*) lenv) [7];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[13] = r1;
        r0 = ((object*) lenv) [13];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } else {
        goto fatbar_260;
      }
      PXLL_RETURN(0);
      goto fatbar_260_over;
      fatbar_260:
      r0 = (object *) constructed_474[0];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 9);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object***) lenv) [1][1][3];
        r2[2] = r3;
        r3 = top[19];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L710; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
        L710:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][18];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L711; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_rule__3econstructor_dt_1761;
        L711:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object****) lenv) [1][1][1][18];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][3];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][14];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L712; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_collect_1757;
        L712:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = top[68];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L713; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist__3ekeys_399;
        L713:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][4];
        r2[2] = r3;
        r3 = top[22];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L714; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
        L714:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object *) 3;
        r3 = ((object*) lenv) [2];
        r2 = PXLL_TEST(unbox(r3)==unbox(r2));
        r1[3] = r2;
        r1 = (object *) 6;
        r2 = ((object*) lenv) [3];
        r1 = PXLL_TEST(r2==r1);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_962;
          r1[2] = r2;
          r2 = ((object*) lenv) [4];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L715; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L715:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
        } else {
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object****) lenv) [1][1][1][19];
          r1[2] = r2;
          r2 = ((object***) lenv) [1][1][3];
          r1[3] = r2;
          r2 = ((object****) lenv) [1][1][1][14];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L716; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_collect_1757;
          L716:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
        }
        r0[5] = r1;
        r1 = (object *) 1;
        r0[6] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[2];
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 2);
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object***) lenv) [1][1][3];
        r3[2] = r4;
        r4 = top[19];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L717; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_car_309;
        L717:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = ((object****) lenv) [1][1][1][18];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L718; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_rule__3econstructor_dt_1761;
        L718:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = top[65];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L719; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
        L719:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[7] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object***) lenv) [1][1][4];
        r1[2] = r2;
        r2 = top[188];
        r1[3] = r2;
        r2 = top[89];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L720; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_sexp_3d_question_430;
        L720:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        if PXLL_IS_TRUE(r1) {
          r1 = ((object***) lenv) [1][1][4];
        } else {
          r1 = top[189];
        }
        r0[8] = r1;
        r1 = (object *) constructed_961[0];
        r0[9] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_10922_i872
        goto L721;
        FUN_lambda_10922_i872:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = ((object***) lenv) [1][1][7];
          switch (get_case (r1)) {
            case (UITAG(1)): {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 2);
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = ((object***) lenv) [1][1][3];
              r5 = ((pxll_vector*)r5)->val[4];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L722; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              lenv = r5[2]; goto *r5[1];
              L722:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r4 = ((object*) lenv) [2];
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = ((object**) lenv) [1][2];
              r5 = UOBJ_GET(r5,0);
              r4[2] = r5;
              r4 = ((object*) lenv) [2];
              lenv = lenv[1];
              lenv = lenv[1];
              r3[2] = r4;
              r4 = top[19];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L723; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_car_309;
              L723:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = ((object******) lenv) [1][1][1][1][1][17];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L724; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_pattern__3esubs_1760;
              L724:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = top[22];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L725; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
              L725:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } break;
            default: {
              r1 = allocate (TC_TUPLE, 3);
              r1[1] = lenv; lenv = r1;
              r2 = ((object****) lenv) [1][1][1][7];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r3 = ((object***) lenv) [1][1][2];
              r2[2] = r3;
              r3 = ((object*) lenv) [2];
              r3 = ((pxll_vector*)r3)->val[1];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L726; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto *r3[1];
              L726:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r1 = ((object*) lenv) [3];
              r1 = ((pxll_vector*)r1)->val[0];
              lenv = lenv[1];
            } break;
          }
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[114];
          r1[3] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][20];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L727; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nthunk_1766;
          L727:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          ensure_heap (1, unbox(r1));
          r1 = (object *) 262;
          r2 = ((object*) lenv) [2];
          if (unbox(r2) == 0) { r1 = (object *) TC_EMPTY_VECTOR; } else {
            t = alloc_no_clear (TC_VECTOR, unbox(r2));
            for (i=0; i<unbox(r2); i++) { t[i+1] = r1; }
            r1 = t;
          }
          r0[4] = r1;
          r1 = (object *) constructed_960[0];
          r0[5] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def frob-rule_1932_i872
          goto L728;
          FUN_frob_rule_1932_i872:
            check_heap (0);
            r0 = ((object*) lenv) [2];
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = top[19];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L729; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
            L729:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][17];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L730; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_pattern__3esubs_1760;
            L730:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][4];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L731; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L731:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r2 = ((object*) lenv) [2];
            r1 = PXLL_TEST(unbox(r2)==unbox(r1));
            r0[3] = r1;
            r0 = (object *) 6;
            r1 = ((object*) lenv) [3];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 3);
              r1 = (object*) &constructed_959;
              r0[2] = r1;
              r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
              r0[3] = r1;
              r1 = top[15];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L732; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
              L732:
              ; lenv = k[2]; k = k[1];
              r0 = result;
            } else {
              r0 = (object *) 14;
            }
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = top[19];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L733; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
            L733:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][17];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L734; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_pattern__3esubs_1760;
            L734:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = top[20];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L735; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
            L735:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L736; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L736:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*) lenv) [3];
            r1[3] = r2;
            r2 = top[21];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L737; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
            L737:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = ((object**) lenv) [1][3];
            r2 = ((object*) lenv) [4];
            t = alloc_no_clear (UOTAG(0),2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[5] = r1;
            r0 = ((object*****) lenv) [1][1][1][1][5];
            r1 = ((object*) lenv) [5];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object****) lenv) [1][1][1][5] = r0;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def $loop_1938_i872
            goto L738;
            FUN__24loop_1938_i872:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*****) lenv) [1][1][1][1][4];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L739; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L739:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(1)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = (object *) constructed_639[0];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                      r0 = (object *) 14;
                    } else {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_262;
                    }
                    lenv = lenv[1];
                  } break;
                  default: {
                    goto fatbar_262;
                  }
                }
                goto fatbar_262_over;
                fatbar_262:
                r0 = ((object**) lenv) [1][2];
                r1 = (object *) 6;
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r0));
                ((pxll_vector*)r2)->val[unbox(r0)] = r1;
                fatbar_262_over: ;
                lenv = lenv[1];
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_1938_i872;
              }
              PXLL_RETURN(0);
            L738:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN__24loop_1938_i872; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = (object *) 1;
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[1] = r1[2]; lenv = r0; goto FUN__24loop_1938_i872;
            PXLL_RETURN(0);
          L728:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_frob_rule_1932_i872; r1[2] = lenv;
          r0[2] = r1;
          r0 = (object *) 3;
          r1 = ((object****) lenv) [1][1][1][6];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          ((object****) lenv) [1][1][1][6] = r0;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r1 = ((pxll_vector*)r1)->val[0];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L740; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L740:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L741; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L741:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object***) lenv) [1][1][2];
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          // def $loop_1943_i872
          goto L742;
          FUN__24loop_1943_i872:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][2];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object*) lenv) [3];
              r0[2] = r1;
              r1 = (object *) constructed_36[0];
              r0[3] = r1;
              r1 = top[21];
              r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r2 = ((object******) lenv) [1][1][1][1][1][4];
              range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
              r1 = ((pxll_vector*)r2)->val[unbox(r1)];
              if PXLL_IS_TRUE(r1) {
                r1 = (object *) constructed_639[0];
              } else {
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object******) lenv) [1][1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L743; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L743:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              }
              r0[2] = r1;
              r0 = ((object**) lenv) [1][3];
              r1 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              r1 = (object *) 3;
              r2 = ((object*) lenv) [2];
              r1 = box((pxll_int)unbox(r2)+unbox(r1));
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN__24loop_1943_i872;
            }
            PXLL_RETURN(0);
          L742:
          r2 = allocate (TC_CLOSURE, 2);
          r2[1] = &&FUN__24loop_1943_i872; r2[2] = lenv;
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object *) 1;
          r1[2] = r2;
          r2 = (object*)TC_NIL;
          r1[3] = r2;
          r2 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L744; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN__24loop_1943_i872;
          L744:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 10);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*****) lenv) [1][1][1][1][2];
          r2 = (object *) constructed_49[0];
          t = alloc_no_clear (UOTAG(9),2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = top[3];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][3];
          r1[3] = r2;
          r2 = top[32];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L745; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
          L745:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = ((object*) lenv) [3];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[4] = r1;
          r1 = ((object*) lenv) [4];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          r0[5] = r1;
          r1 = allocate (TC_TUPLE, 4);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
          r3[2] = r4;
          r4 = top[20];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L746; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_cdr_310;
          L746:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = allocate (TC_TUPLE, 3);
          r4 = ((object*****) lenv) [1][1][1][1][3];
          r3[2] = r4;
          r4 = (object *) constructed_36[0];
          r3[3] = r4;
          r4 = top[21];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L747; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
          L747:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r2[3] = r3;
          r3 = top[21];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L748; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
          L748:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][5];
          r1[3] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][8];
          r1[4] = r2;
          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L749; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_compile_match_1743;
          L749:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[6] = r1;
          r1 = (object*)TC_NIL;
          r0[7] = r1;
          r1 = ((object*) lenv) [7];
          r2 = ((object*) lenv) [6];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[8] = r1;
          r1 = ((object*) lenv) [8];
          r2 = ((object*) lenv) [5];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[9] = r1;
          r1 = ((object*) lenv) [9];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          r0[10] = r1;
          r0 = ((object******) lenv) [1][1][1][1][1][9];
          r1 = ((object*) lenv) [10];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object*****) lenv) [1][1][1][1][9] = r0;
          PXLL_RETURN(0);
        L721:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_10922_i872; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [5];
        r0[3] = r1;
        r1 = top[67];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L750; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
        L750:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][7];
        switch (get_case (r1)) {
          case (UOTAG(0)): {
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][7];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r2 = ((pxll_vector*)r2)->val[3];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L751; t[4] = r0; t[5] = r1; k = t;
            lenv = r2[2]; goto *r2[1];
            L751:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = ((object*) lenv) [2];
            r2 = ((object****) lenv) [1][1][1][6];
            r1 = PXLL_TEST(unbox(r2)<unbox(r1));
            lenv = lenv[1];
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 6);
              r1[1] = lenv; lenv = r1;
              r2 = (object *) constructed_635[0];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r2;
              r2 = t;
              r1[2] = r2;
              r2 = (object*)TC_NIL;
              r1[3] = r2;
              r2 = ((object*) lenv) [3];
              r3 = ((object****) lenv) [1][1][1][8];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[4] = r2;
              r2 = ((object*) lenv) [4];
              r3 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[5] = r2;
              r2 = ((object*) lenv) [5];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r2;
              r2 = t;
              r1[6] = r2;
              r1 = ((object****) lenv) [1][1][1][9];
              r2 = ((object*) lenv) [6];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              ((object***) lenv) [1][1][9] = r1;
            } else {
              r1 = (object *) 14;
            }
            r1 = allocate (TC_TUPLE, 12);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) constructed_603[0];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r2;
            r2 = t;
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r2 = ((pxll_vector*)r2)->val[7];
            r1[3] = r2;
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r2;
            r2 = t;
            r1[4] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object******) lenv) [1][1][1][1][1][2];
            r2[2] = r3;
            r3 = top[19];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L752; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
            L752:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[5] = r2;
            r2 = ((object*) lenv) [5];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r2;
            r2 = t;
            r1[6] = r2;
            r2 = (object*)TC_NIL;
            r1[7] = r2;
            r2 = ((object*) lenv) [7];
            r3 = ((object*) lenv) [6];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[8] = r2;
            r2 = ((object*) lenv) [8];
            r3 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[9] = r2;
            r2 = ((object*) lenv) [9];
            r3 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[10] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object****) lenv) [1][1][1][9];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L753; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L753:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[11] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = allocate (TC_TUPLE, 3);
            r4 = ((object*) lenv) [10];
            r3[2] = r4;
            r4 = (object *) constructed_36[0];
            r3[3] = r4;
            r4 = top[21];
            t = allocate (TC_SAVE, 6);
            t[1] = k; t[2] = lenv; t[3] = &&L754; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
            r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
            L754:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
            r3 = result;
            r2[2] = r3;
            r3 = ((object*) lenv) [11];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L755; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L755:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[12] = r2;
            r1 = ((object*) lenv) [12];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            lenv = lenv[1];
            lenv = lenv[1];
          } break;
          default: {
            r1 = allocate (TC_TUPLE, 9);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) constructed_603[0];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r2;
            r2 = t;
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*****) lenv) [1][1][1][1][2];
            r2[2] = r3;
            r3 = top[19];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L756; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
            L756:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r2;
            r2 = t;
            r1[4] = r2;
            r2 = (object*)TC_NIL;
            r1[5] = r2;
            r2 = ((object*) lenv) [5];
            r3 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[6] = r2;
            r2 = ((object*) lenv) [6];
            r3 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[7] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object***) lenv) [1][1][9];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L757; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L757:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[8] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = allocate (TC_TUPLE, 3);
            r4 = ((object*) lenv) [7];
            r3[2] = r4;
            r4 = (object *) constructed_36[0];
            r3[3] = r4;
            r4 = top[21];
            t = allocate (TC_SAVE, 6);
            t[1] = k; t[2] = lenv; t[3] = &&L758; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
            r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
            L758:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
            r3 = result;
            r2[2] = r3;
            r3 = ((object*) lenv) [8];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L759; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L759:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[9] = r2;
            r1 = ((object*) lenv) [9];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            lenv = lenv[1];
          } break;
        }
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = top[188];
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1 = PXLL_TEST(r2==r1);
        r0[2] = r1;
        r0 = (object *) 6;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[3] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][7];
          r0[1] = r1[2]; lenv = r0; goto FUN_fatbar_1745;
        } else {
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      } else {
        goto fatbar_261;
      }
      PXLL_RETURN(0);
      goto fatbar_261_over;
      fatbar_261:
      r0 = (object *) constructed_964[0];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L760; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L760:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_10106_i929
        goto L761;
        FUN_lambda_10106_i929:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[20];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L762; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
          L762:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 4);
          r2 = ((object***) lenv) [1][1][2];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object*) lenv) [2];
          r2[2] = r3;
          r3 = top[19];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L763; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
          L763:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r1[4] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][8];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L764; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_subst_1746;
          L764:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r0 = ((object*) lenv) [5];
          r1 = ((object*) lenv) [4];
          t = alloc_no_clear (UOTAG(0),2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
          PXLL_RETURN(0);
        L761:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_10106_i929; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L765; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L765:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        r1[2] = r2;
        r2 = top[20];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L766; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
        L766:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[4] = r1;
        r1 = ((object****) lenv) [1][1][1][5];
        r0[1] = r1[2]; lenv = r0; goto FUN_compile_match_1743;
      } else {
        goto fatbar_263;
      }
      PXLL_RETURN(0);
      goto fatbar_263_over;
      fatbar_263:
      r0 = (object *) constructed_868[0];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        // def loop_1884_i932
        goto L767;
        FUN_loop_1884_i932:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = ((object*) lenv) [3];
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 5);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 2);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = top[19];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L768; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_car_309;
              L768:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r3 = ((object*) lenv) [2];
              r3 = allocate (TC_TUPLE, 2);
              r3[1] = lenv; lenv = r3;
              r4 = ((object**) lenv) [1][2];
              r4 = UOBJ_GET(r4,0);
              r3[2] = r4;
              r3 = ((object*) lenv) [2];
              lenv = lenv[1];
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[19];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L769; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
              L769:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object******) lenv) [1][1][1][1][1][12];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L770; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_pattern__3eliteral_1754;
              L770:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[4] = r1;
              r1 = ((object*) lenv) [4];
              switch (get_case (r1)) {
                case (UOTAG(2)): {
                  r1 = (object *) constructed_966[0];
                  t = alloc_no_clear (UOTAG(1),1);
                  t[1] = r1;
                  r1 = t;
                } break;
                default: {
                  goto fatbar_265;
                }
              }
              goto fatbar_265_over;
              fatbar_265:
              r1 = (object *) constructed_968[0];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r1;
              r1 = t;
              fatbar_265_over: ;
              r0[5] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r1 = allocate (TC_TUPLE, 20);
              r1[1] = lenv; lenv = r1;
              r2 = (object *) constructed_595[0];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r2;
              r2 = t;
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r3 = ((object******) lenv) [1][1][1][1][1][2];
              r2[2] = r3;
              r3 = top[19];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L771; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
              L771:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r2 = ((object*) lenv) [3];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r2;
              r2 = t;
              r1[4] = r2;
              r2 = (object*)TC_NIL;
              r1[5] = r2;
              r2 = ((object*) lenv) [5];
              r3 = ((object**) lenv) [1][4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[6] = r2;
              r2 = (object *) constructed_870[0];
              t = alloc_no_clear (UOTAG(1),1);
              t[1] = r2;
              r2 = t;
              r1[7] = r2;
              r2 = ((object*) lenv) [6];
              r3 = ((object*) lenv) [7];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[8] = r2;
              r2 = ((object*) lenv) [8];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r2;
              r2 = t;
              r1[9] = r2;
              r2 = (object*)TC_NIL;
              r1[10] = r2;
              r2 = ((object*) lenv) [10];
              r3 = ((object*) lenv) [9];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[11] = r2;
              r2 = ((object*) lenv) [11];
              r3 = ((object*) lenv) [4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[12] = r2;
              r2 = ((object*) lenv) [12];
              r3 = ((object**) lenv) [1][5];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[13] = r2;
              r2 = ((object*) lenv) [13];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r2;
              r2 = t;
              r1[14] = r2;
              r2 = allocate (TC_TUPLE, 4);
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object******) lenv) [1][1][1][1][1][2];
              r3[2] = r4;
              r4 = top[20];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L772; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_cdr_310;
              L772:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 3);
              r4 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r3[2] = r4;
              r4 = ((object**) lenv) [1][2];
              r3[3] = r4;
              r4 = top[32];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L773; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
              L773:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = top[189];
              r2[4] = r3;
              r3 = ((object*******) lenv) [1][1][1][1][1][1][5];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L774; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_compile_match_1743;
              L774:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[15] = r2;
              r2 = (object*)TC_NIL;
              r1[16] = r2;
              r2 = ((object*) lenv) [16];
              r3 = top[189];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[17] = r2;
              r2 = ((object*) lenv) [17];
              r3 = ((object*) lenv) [15];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[18] = r2;
              r2 = ((object*) lenv) [18];
              r3 = ((object*) lenv) [14];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[19] = r2;
              r2 = ((object*) lenv) [19];
              r3 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[20] = r2;
              r1 = ((object*) lenv) [20];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r1;
              r1 = t;
              lenv = lenv[1];
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r0[3] = r1;
              r1 = ((object******) lenv) [1][1][1][1][1][7];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L775; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_fatbar_1745;
              L775:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              r1 = ((object*) lenv) [3];
              lenv = ((object *)lenv)[1];
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN_loop_1884_i932;
            } break;
          }
          PXLL_RETURN(0);
        L767:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_loop_1884_i932; r1[2] = lenv;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object***) lenv) [1][1][3];
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][13];
        r1[3] = r2;
        r2 = top[42];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L776; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_pack_337;
        L776:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r0[1] = r1[2]; lenv = r0; goto FUN_loop_1884_i932;
      } else {
        goto fatbar_264;
      }
      PXLL_RETURN(0);
      goto fatbar_264_over;
      fatbar_264:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_67;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_67;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_264_over: ;
      PXLL_RETURN(0);
      fatbar_263_over: ;
      PXLL_RETURN(0);
      fatbar_261_over: ;
      PXLL_RETURN(0);
      fatbar_260_over: ;
      PXLL_RETURN(0);
    L675:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_compile_group_1744; r1[2] = lenv;
    r0[6] = r1;
    // def fatbar_1745
    goto L777;
    FUN_fatbar_1745:
      check_heap (0);
      r0 = top[189];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } else {
        r0 = top[189];
        r1 = ((object*) lenv) [3];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 8);
          r0[1] = lenv; lenv = r0;
          r1 = (object *) 6;
          t = alloc_no_clear (UOTAG(4),1);
          t[1] = r1;
          r1 = t;
          r0[2] = r1;
          r1 = (object*)TC_NIL;
          r0[3] = r1;
          r1 = ((object*) lenv) [3];
          r2 = ((object**) lenv) [1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[4] = r1;
          r1 = ((object*) lenv) [4];
          r2 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[5] = r1;
          r1 = ((object*) lenv) [5];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[6] = r1;
          r1 = (object *) constructed_498[0];
          t = alloc_no_clear (UOTAG(1),1);
          t[1] = r1;
          r1 = t;
          r0[7] = r1;
          r1 = ((object*) lenv) [6];
          r2 = ((object*) lenv) [7];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[8] = r1;
          r0 = ((object*) lenv) [8];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L777:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_fatbar_1745; r1[2] = lenv;
    r0[7] = r1;
    // def subst_1746
    goto L778;
    FUN_subst_1746:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = (object *) constructed_639[0];
          r2 = ((object**) lenv) [1][2];
          r1 = PXLL_TEST(r2==r1);
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 12);
            r0[1] = lenv; lenv = r0;
            r1 = (object *) constructed_864[0];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][2];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[4] = r1;
            r1 = (object*)TC_NIL;
            r0[5] = r1;
            r1 = ((object*) lenv) [5];
            r2 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[6] = r1;
            r1 = ((object*) lenv) [6];
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[7] = r1;
            r1 = ((object*) lenv) [7];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            r0[8] = r1;
            r1 = (object*)TC_NIL;
            r0[9] = r1;
            r1 = ((object*) lenv) [9];
            r2 = ((object***) lenv) [1][1][4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[10] = r1;
            r1 = ((object*) lenv) [10];
            r2 = ((object*) lenv) [8];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[11] = r1;
            r1 = ((object*) lenv) [11];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[12] = r1;
            r0 = ((object*) lenv) [12];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            r0 = ((object**) lenv) [1][4];
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_259;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_259_over;
      fatbar_259:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_67;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_67;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_259_over: ;
      PXLL_RETURN(0);
    L778:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_subst_1746; r1[2] = lenv;
    r0[8] = r1;
    // def fieldpair->label_1748
    goto L779;
    FUN_fieldpair__3elabel_1748:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
      PXLL_RETURN(0);
    L779:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_fieldpair__3elabel_1748; r1[2] = lenv;
    r0[9] = r1;
    // def fieldpair->pattern_1749
    goto L780;
    FUN_fieldpair__3epattern_1749:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,1);
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      PXLL_RETURN(0);
      PXLL_RETURN(0);
    L780:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_fieldpair__3epattern_1749; r1[2] = lenv;
    r0[10] = r1;
    // def pattern->fieldpairs_1750
    goto L781;
    FUN_pattern__3efieldpairs_1750:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(3)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_258;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_258_over;
      fatbar_258:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_954;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_954;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_258_over: ;
      PXLL_RETURN(0);
    L781:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_pattern__3efieldpairs_1750; r1[2] = lenv;
    r0[11] = r1;
    // def pattern->literal_1754
    goto L782;
    FUN_pattern__3eliteral_1754:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_257;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_257_over;
      fatbar_257:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_953;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_953;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_257_over: ;
      PXLL_RETURN(0);
    L782:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_pattern__3eliteral_1754; r1[2] = lenv;
    r0[12] = r1;
    // def first-literal=?_1755
    goto L783;
    FUN_first_literal_3d_question_1755:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r0 = ((object**) lenv) [1][3];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][3];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object**) lenv) [1][2];
      r2[2] = r3;
      r3 = top[19];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L784; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
      L784:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r2 = ((object****) lenv) [1][1][1][12];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L785; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_pattern__3eliteral_1754;
      L785:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object*) lenv) [2];
      r2[2] = r3;
      r3 = top[19];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L786; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
      L786:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r2 = ((object****) lenv) [1][1][1][12];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L787; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_pattern__3eliteral_1754;
      L787:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = top[89];
      r0[1] = r1[2]; lenv = r0; goto FUN_sexp_3d_question_430;
      PXLL_RETURN(0);
      PXLL_RETURN(0);
    L783:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_first_literal_3d_question_1755; r1[2] = lenv;
    r0[13] = r1;
    // def collect_1757
    goto L788;
    FUN_collect_1757:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop_1893
      goto L789;
      FUN_loop_1893:
        check_heap (0);
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L790; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L790:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = ((object*) lenv) [4];
            r1[3] = r2;
            r2 = top[65];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L791; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
            L791:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r0 = ((object*) lenv) [5];
            switch (get_case (r0)) {
              case (UITAG(1)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = (object *) constructed_952[0];
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 6);
                r1[1] = lenv; lenv = r1;
                // def push_1716_i1034
                goto L792;
                FUN_push_1716_i1034:
                  check_heap (0);
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object****) lenv) [1][1][1][2];
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = ((object**) lenv) [1][2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  ((object***) lenv) [1][1][2] = r0;
                  PXLL_RETURN(0);
                L792:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_push_1716_i1034; r2[2] = lenv;
                r1[2] = r2;
                // def pop_1717_i1034
                goto L793;
                FUN_pop_1717_i1034:
                  check_heap (0);
                  r0 = ((object**) lenv) [1][2];
                  switch (get_case (r0)) {
                    case (TC_NIL): {
                      r0 = (object*) &constructed_68;
                      r0 = box((pxll_int)((pxll_string*)(r0))->len);
                      r1 = (object*) &constructed_68;
                      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                      r0 = (object*) &constructed_951;
                      dump_object (r0, 0); fprintf (stdout, "\n");
                      r0 = (object*) &constructed_951;
                      PXLL_UNDEFINED; result=r0; goto Lreturn;
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      r0 = allocate (TC_TUPLE, 4);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][2];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r1 = ((object*) lenv) [2];
                      r0[4] = r1;
                      r0 = ((object*) lenv) [3];
                      ((object***) lenv) [1][1][2] = r0;
                      r0 = ((object*) lenv) [4];
                      PXLL_RETURN(0);
                    } break;
                  }
                  PXLL_RETURN(0);
                L793:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_pop_1717_i1034; r2[2] = lenv;
                r1[3] = r2;
                // def get_1718_i1034
                goto L794;
                FUN_get_1718_i1034:
                  r0 = ((object**) lenv) [1][2];
                  PXLL_RETURN(0);
                L794:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_get_1718_i1034; r2[2] = lenv;
                r1[4] = r2;
                // def top_1719_i1034
                goto L795;
                FUN_top_1719_i1034:
                  check_heap (0);
                  r0 = ((object**) lenv) [1][2];
                  switch (get_case (r0)) {
                    case (TC_NIL): {
                      r0 = (object*) &constructed_68;
                      r0 = box((pxll_int)((pxll_string*)(r0))->len);
                      r1 = (object*) &constructed_68;
                      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                      r0 = (object*) &constructed_950;
                      dump_object (r0, 0); fprintf (stdout, "\n");
                      r0 = (object*) &constructed_950;
                      PXLL_UNDEFINED; result=r0; goto Lreturn;
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      PXLL_RETURN(0);
                    } break;
                  }
                  PXLL_RETURN(0);
                L795:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_top_1719_i1034; r2[2] = lenv;
                r1[5] = r2;
                // def stack-length_1720_i1034
                goto L796;
                FUN_stack_length_1720_i1034:
                  check_heap (0);
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = ((object**) lenv) [1][2];
                  r0[2] = r1;
                  r1 = top[22];
                  r0[1] = r1[2]; lenv = r0; goto FUN_length_314;
                L796:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_stack_length_1720_i1034; r2[2] = lenv;
                r1[6] = r2;
                r1 = allocate (UOTAG(21), 5);
                r2 = ((object*) lenv) [4];
                r1[1] = r2;
                r2 = ((object*) lenv) [6];
                r1[2] = r2;
                r2 = ((object*) lenv) [3];
                r1[3] = r2;
                r2 = ((object*) lenv) [2];
                r1[4] = r2;
                r2 = ((object*) lenv) [5];
                r1[5] = r2;
                lenv = lenv[1];
                r0[3] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object**) lenv) [1][2];
                r0[2] = r1;
                r1 = ((object*) lenv) [3];
                r1 = ((pxll_vector*)r1)->val[3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L797; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L797:
                ; lenv = k[2]; k = k[1];
                r0 = ((object*) lenv) [3];
                r1 = ((object***) lenv) [1][1][2];
                r2 = ((object**) lenv) [1][4];
                t = alloc_no_clear (UOTAG(1),3);
                t[1] = r2;
                t[2] = r0;
                t[3] = r1;
                r0 = t;
                r1 = ((object**) lenv) [1][3];
                lenv = ((object **)lenv)[1][1];
                lenv[2] = r0;
                lenv[3] = r1;
                goto FUN_loop_1893;
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][5];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object**) lenv) [1][2];
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r1 = ((pxll_vector*)r1)->val[3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L798; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L798:
                ; lenv = k[2]; k = k[1];
                r0 = ((object**) lenv) [1][3];
                r1 = ((object***) lenv) [1][1][2];
                lenv = ((object **)lenv)[1][1];
                lenv[2] = r1;
                lenv[3] = r0;
                goto FUN_loop_1893;
              } break;
            }
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      L789:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop_1893; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*)UITAG(0);
      r0[2] = r1;
      r1 = ((object**) lenv) [1][3];
      r0[3] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop_1893;
    L788:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_collect_1757; r1[2] = lenv;
    r0[14] = r1;
    // def pattern->dt_1758
    goto L799;
    FUN_pattern__3edt_1758:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(2)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_256;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_256_over;
      fatbar_256:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_949;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_949;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_256_over: ;
      PXLL_RETURN(0);
    L799:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_pattern__3edt_1758; r1[2] = lenv;
    r0[15] = r1;
    // def pattern->alt_1759
    goto L800;
    FUN_pattern__3ealt_1759:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(2)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_255;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_255_over;
      fatbar_255:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_948;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_948;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_255_over: ;
      PXLL_RETURN(0);
    L800:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_pattern__3ealt_1759; r1[2] = lenv;
    r0[16] = r1;
    // def pattern->subs_1760
    goto L801;
    FUN_pattern__3esubs_1760:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(2)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,2);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_254;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_254_over;
      fatbar_254:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_947;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_947;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_254_over: ;
      PXLL_RETURN(0);
    L801:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_pattern__3esubs_1760; r1[2] = lenv;
    r0[17] = r1;
    // def rule->constructor-dt_1761
    goto L802;
    FUN_rule__3econstructor_dt_1761:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = top[19];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L803; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
      L803:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][15];
      r0[1] = r1[2]; lenv = r0; goto FUN_pattern__3edt_1758;
      PXLL_RETURN(0);
    L802:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_rule__3econstructor_dt_1761; r1[2] = lenv;
    r0[18] = r1;
    // def rule->constructor-alt_1762
    goto L804;
    FUN_rule__3econstructor_alt_1762:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = top[19];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L805; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
      L805:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][16];
      r0[1] = r1[2]; lenv = r0; goto FUN_pattern__3ealt_1759;
      PXLL_RETURN(0);
    L804:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_rule__3econstructor_alt_1762; r1[2] = lenv;
    r0[19] = r1;
    // def nthunk_1766
    goto L806;
    FUN_nthunk_1766:
      check_heap (0);
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = (object *) constructed_946[0];
        PXLL_RETURN(0);
      } else {
        goto fatbar_253;
      }
      PXLL_RETURN(0);
      goto fatbar_253_over;
      fatbar_253:
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][3];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L807; t[4] = r0; k = t;
      lenv = r1[2]; goto *r1[1];
      L807:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 3);
      r2 = (object *) 3;
      r3 = ((object**) lenv) [1][2];
      r2 = box((pxll_int)unbox(r3)-unbox(r2));
      r1[2] = r2;
      r2 = ((object**) lenv) [1][3];
      r1[3] = r2;
      r2 = ((object***) lenv) [1][1][20];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L808; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_nthunk_1766;
      L808:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      r1 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      PXLL_RETURN(0);
      fatbar_253_over: ;
      PXLL_RETURN(0);
    L806:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_nthunk_1766; r1[2] = lenv;
    r0[20] = r1;
    r0 = allocate (TC_TUPLE, 5);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    // def loop_1790_i1098
    goto L809;
    FUN_loop_1790_i1098:
      check_heap (0);
      r0 = ((object*) lenv) [4];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [3];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = (object *) constructed_660[0];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = ((object**) lenv) [1][3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object***) lenv) [1][1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object***) lenv) [1][1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = allocate (TC_TUPLE, 4);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 3);
                    r2 = ((object*****) lenv) [1][1][1][1][2];
                    r1[2] = r2;
                    r2 = (object *) constructed_36[0];
                    r1[3] = r2;
                    r2 = top[21];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L810; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                    L810:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L811; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto *r2[1];
                    L811:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r1 = ((object*) lenv) [3];
                    r2 = ((object*) lenv) [2];
                    t = alloc_no_clear (UOTAG(0),2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[4] = r1;
                    r0 = ((object*****) lenv) [1][1][1][1][3];
                    r1 = ((object*) lenv) [4];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r1;
                    t[2] = r0;
                    r0 = t;
                    lenv = lenv[1];
                    r1 = ((object*) lenv) [3];
                    r2 = (object *) constructed_940[0];
                    lenv = ((object ***)lenv)[1][1][1];
                    lenv[2] = r2;
                    lenv[3] = r0;
                    lenv[4] = r1;
                    goto FUN_loop_1790_i1098;
                  } break;
                  default: {
                    lenv = ((object *)lenv)[1];
                    goto fatbar_248;
                  }
                }
                PXLL_RETURN(0);
              } else {
                lenv = ((object *)lenv)[1];
                goto fatbar_248;
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_248;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_248_over;
          fatbar_248:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r1[1] = lenv; lenv = r1;
          // def parse-fieldpair_1768_i824_i1098
          goto L812;
          FUN_parse_fieldpair_1768_i824_i1098:
            check_heap (0);
            r0 = ((object*) lenv) [2];
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L813; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_kind_1769_i824_i1098;
            L813:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r0 = ((object*) lenv) [4];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(0),2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            PXLL_RETURN(0);
            PXLL_RETURN(0);
          L812:
          r2 = allocate (TC_CLOSURE, 2);
          r2[1] = &&FUN_parse_fieldpair_1768_i824_i1098; r2[2] = lenv;
          r1[2] = r2;
          // def kind_1769_i824_i1098
          goto L814;
          FUN_kind_1769_i824_i1098:
            check_heap (0);
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              case (UOTAG(8)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object***) lenv) [1][1][2];
                r1[2] = r2;
                r2 = allocate (TC_TUPLE, 3);
                r3 = top[115];
                r2[2] = r3;
                r3 = ((object*) lenv) [2];
                r2[3] = r3;
                r3 = top[34];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L815; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_filter_328;
                L815:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[3] = r2;
                r2 = top[32];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L816; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                L816:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                t = alloc_no_clear (UOTAG(3),1);
                t[1] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              case (UOTAG(4)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(r1) {
                  r1 = (object *) constructed_44[0];
                } else {
                  r1 = (object *) constructed_41[0];
                }
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                r1 = (object *) constructed_941[0];
                r2 = (object *) constructed_46[0];
                t = alloc_no_clear (UOTAG(2),3);
                t[1] = r2;
                t[2] = r0;
                t[3] = r1;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              case (UOTAG(0)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = (object *) constructed_49[0];
                    r1 = (object *) constructed_942[0];
                    r2 = (object *) constructed_51[0];
                    t = alloc_no_clear (UOTAG(2),3);
                    t[1] = r2;
                    t[2] = r0;
                    t[3] = r1;
                    r0 = t;
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [2];
                    switch (get_case (r0)) {
                      case (UOTAG(1)): {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r0 = (object *) constructed_944[0];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r1 = ((object*) lenv) [2];
                                  r0[2] = r1;
                                  r1 = ((object******) lenv) [1][1][1][1][1][3];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_kind_1769_i824_i1098;
                                } break;
                                default: {
                                  lenv = ((object *)lenv)[1];
                                  goto fatbar_252;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              goto fatbar_252;
                            }
                          }
                          PXLL_RETURN(0);
                        } else {
                          goto fatbar_252;
                        }
                        PXLL_RETURN(0);
                        goto fatbar_252_over;
                        fatbar_252:
                        r0 = (object *) constructed_870[0];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [2];
                              switch (get_case (r0)) {
                                case (UOTAG(1)): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][2];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r0 = ((object**) lenv) [1][3];
                                  switch (get_case (r0)) {
                                    case (TC_NIL): {
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][2];
                                      t = alloc_no_clear (UOTAG(1),1);
                                      t[1] = r1;
                                      r1 = t;
                                      r0[2] = r1;
                                      r0 = ((object*) lenv) [2];
                                      t = alloc_no_clear (UOTAG(0),1);
                                      t[1] = r0;
                                      r0 = t;
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      lenv = ((object ***)lenv)[1][1][1];
                                      goto fatbar_251;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object **)lenv)[1][1];
                                  goto fatbar_251;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object *)lenv)[1];
                              goto fatbar_251;
                            }
                          }
                          PXLL_RETURN(0);
                        } else {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_251;
                        }
                        PXLL_RETURN(0);
                        fatbar_252_over: ;
                        PXLL_RETURN(0);
                      } break;
                      case (UOTAG(9)): {
                        r0 = allocate (TC_TUPLE, 4);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = ((object*****) lenv) [1][1][1][1][3];
                        r1[2] = r2;
                        r2 = ((object**) lenv) [1][3];
                        r1[3] = r2;
                        r2 = top[32];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L817; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                        L817:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[4] = r1;
                        r0 = ((object*) lenv) [3];
                        r1 = ((object*) lenv) [4];
                        r2 = ((object*) lenv) [2];
                        t = alloc_no_clear (UOTAG(2),3);
                        t[1] = r2;
                        t[2] = r0;
                        t[3] = r1;
                        r0 = t;
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        goto fatbar_251;
                      }
                    }
                    PXLL_RETURN(0);
                    goto fatbar_251_over;
                    fatbar_251:
                    r0 = allocate (TC_TUPLE, 6);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object*****) lenv) [1][1][1][1][3];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L818; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_kind_1769_i824_i1098;
                    L818:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][3];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r2;
                    r2 = t;
                    r1[2] = r2;
                    r2 = ((object*****) lenv) [1][1][1][1][3];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L819; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_kind_1769_i824_i1098;
                    L819:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r1 = (object*)TC_NIL;
                    r0[4] = r1;
                    r1 = ((object*) lenv) [4];
                    r2 = ((object*) lenv) [3];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[5] = r1;
                    r1 = ((object*) lenv) [5];
                    r2 = ((object*) lenv) [2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[6] = r1;
                    r0 = (object *) constructed_61[0];
                    r1 = ((object*) lenv) [6];
                    r2 = (object *) constructed_51[0];
                    t = alloc_no_clear (UOTAG(2),3);
                    t[1] = r2;
                    t[2] = r0;
                    t[3] = r1;
                    r0 = t;
                    PXLL_RETURN(0);
                    fatbar_251_over: ;
                    PXLL_RETURN(0);
                  } break;
                }
                PXLL_RETURN(0);
                goto fatbar_250_over;
                fatbar_250:
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object*) &constructed_945;
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r0[3] = r1;
                r1 = top[15];
                r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                fatbar_250_over: ;
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_249;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_249_over;
            fatbar_249:
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
            fatbar_249_over: ;
            PXLL_RETURN(0);
          L814:
          r2 = allocate (TC_CLOSURE, 2);
          r2[1] = &&FUN_kind_1769_i824_i1098; r2[2] = lenv;
          r1[3] = r2;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object***) lenv) [1][1][2];
          r1[2] = r2;
          r2 = ((object*) lenv) [3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L820; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_kind_1769_i824_i1098;
          L820:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
          r0[2] = r1;
          r0 = ((object***) lenv) [1][1][2];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = ((object**) lenv) [1][3];
          r2 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          lenv[3] = r1;
          lenv[4] = r2;
          goto FUN_loop_1790_i1098;
          fatbar_248_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L809:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_loop_1790_i1098; r2[2] = lenv;
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 4);
    r2 = (object *) constructed_939[0];
    r1[2] = r2;
    r2 = (object *) constructed_938[0];
    r1[3] = r2;
    r2 = ((object****) lenv) [1][1][1][5];
    r1[4] = r2;
    r2 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L821; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_loop_1790_i1098;
    L821:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 2);
    r2[1] = lenv; lenv = r2;
    r3 = allocate (TC_TUPLE, 2);
    r4 = ((object**) lenv) [1][2];
    r3[2] = r4;
    r4 = top[19];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L822; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_car_309;
    L822:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    r2 = ((object*) lenv) [2];
    r2 = allocate (TC_TUPLE, 2);
    r2[1] = lenv; lenv = r2;
    r3 = ((object**) lenv) [1][2];
    r3 = UOBJ_GET(r3,0);
    r2[2] = r3;
    r2 = ((object*) lenv) [2];
    lenv = lenv[1];
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L823; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L823:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = ((object***) lenv) [1][1][4];
    switch (get_case (r1)) {
      case (TC_NIL): {
        r1 = (object *) 262;
      } break;
      default: {
        goto fatbar_247;
      }
    }
    goto fatbar_247_over;
    fatbar_247:
    r1 = (object *) 6;
    fatbar_247_over: ;
    if PXLL_IS_TRUE(r1) {
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object*) lenv) [3];
      r1[2] = r2;
      r2 = top[114];
      r1[3] = r2;
      r2 = ((object**) lenv) [1][20];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L824; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_nthunk_1766;
      L824:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
    } else {
      r1 = ((object***) lenv) [1][1][4];
    }
    r0[4] = r1;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object*) lenv) [4];
    r1[2] = r2;
    r2 = ((object*) lenv) [2];
    r1[3] = r2;
    r2 = top[188];
    r1[4] = r2;
    r2 = ((object**) lenv) [1][5];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L825; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_compile_match_1743;
    L825:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[5] = r1;
    r0 = allocate (UOTAG(1), 2);
    r1 = ((object*) lenv) [4];
    r0[1] = r1;
    r1 = ((object*) lenv) [5];
    r0[2] = r1;
    PXLL_RETURN(0);
  L661:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_compile_pattern_478; r1[2] = lenv;
  r0[116] = r1;
  // def print-datatype_481
  goto L826;
  FUN_print_datatype_481:
    check_heap (0);
    r0 = (object*) &constructed_937;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_937;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[7];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r0 = dump_object (r0, 0); fprintf (stdout, "\n");
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    // def lambda_13646
    goto L827;
    FUN_lambda_13646:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 8);
      r2[1] = lenv; lenv = r2;
      r3 = ((object***) lenv) [1][1][2];
      r3 = allocate (TC_TUPLE, 2);
      r3[1] = lenv; lenv = r3;
      r4 = ((object****) lenv) [1][1][1][2];
      r4 = UOBJ_GET(r4,0);
      r3[2] = r4;
      r3 = ((object*) lenv) [2];
      lenv = lenv[1];
      r2[2] = r3;
      r3 = allocate (TC_TUPLE, 3);
      r4 = allocate (TC_TUPLE, 3);
      r5 = top[107];
      r4[2] = r5;
      r5 = ((object***) lenv) [1][1][3];
      r5 = ((pxll_vector*)r5)->val[3];
      r4[3] = r5;
      r5 = top[32];
      t = allocate (TC_SAVE, 7);
      t[1] = k; t[2] = lenv; t[3] = &&L828; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
      r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
      L828:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
      r4 = result;
      r3[2] = r4;
      r4 = (object*) &constructed_936;
      r3[3] = r4;
      r4 = top[48];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L829; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
      L829:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[3] = r3;
      r3 = (object*)TC_NIL;
      r2[4] = r3;
      r3 = ((object*) lenv) [4];
      r4 = (object*) &constructed_935;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[5] = r3;
      r3 = ((object*) lenv) [5];
      r4 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[6] = r3;
      r3 = ((object*) lenv) [6];
      r4 = (object*) &constructed_934;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[7] = r3;
      r3 = ((object*) lenv) [7];
      r4 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[8] = r3;
      r2 = ((object*) lenv) [8];
      r3 = (object*) &constructed_933;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L830; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L830:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      PXLL_RETURN(0);
    L827:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_13646; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[6];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L831; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L831:
    ; lenv = k[2]; k = k[1];
    r0 = (object*) &constructed_932;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_932;
    r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    PXLL_RETURN(0);
  L826:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_print_datatype_481; r1[2] = lenv;
  r0[117] = r1;
  // def make-node_486
  goto L832;
  FUN_make_node_486:
    check_heap (0);
    r0 = allocate (UOTAG(20), 6);
    r1 = (object *) 1;
    r0[1] = r1;
    r1 = top[191];
    r1 = ((pxll_vector*)r1)->val[1];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L833; t[4] = r0; k = t;
    lenv = r1[2]; goto *r1[1];
    L833:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 4);
    // def lambda_13664_i1465
    goto L834;
    FUN_lambda_13664_i1465:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[2];
      r0[2] = r1;
      r0 = ((object**) lenv) [1][3];
      r1 = ((object*) lenv) [2];
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      PXLL_RETURN(0);
    L834:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_lambda_13664_i1465; r2[2] = lenv;
    r1[2] = r2;
    r2 = (object *) 3;
    r1[3] = r2;
    r2 = ((object*) lenv) [3];
    r1[4] = r2;
    r2 = top[37];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L835; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_fold_331;
    L835:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[5] = r1;
    r1 = top[192];
    r0[6] = r1;
    PXLL_RETURN(0);
  L832:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_make_node_486; r1[2] = lenv;
  r0[118] = r1;
  // def varref->name_489
  goto L836;
  FUN_varref__3ename_489:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_246;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_246_over;
    fatbar_246:
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_931;
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = (object*) &constructed_931;
    PXLL_UNDEFINED; result=r0; goto Lreturn;
    PXLL_RETURN(0);
    fatbar_246_over: ;
    PXLL_RETURN(0);
  L836:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_varref__3ename_489; r1[2] = lenv;
  r0[119] = r1;
  // def node/if_493
  goto L837;
  FUN_node_2fif_493:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 5);
    r0[1] = lenv; lenv = r0;
    r1 = (object*)TC_NIL;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r2 = ((object**) lenv) [1][3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[4] = r1;
    r1 = ((object*) lenv) [4];
    r2 = ((object**) lenv) [1][2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[5] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*)UITAG(6);
    r0[2] = r1;
    r1 = ((object*) lenv) [5];
    r0[3] = r1;
    r1 = top[118];
    r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
  L837:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_node_2fif_493; r1[2] = lenv;
  r0[120] = r1;
  // def node/fix_498
  goto L838;
  FUN_node_2ffix_498:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = (object*)TC_NIL;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object**) lenv) [1][3];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L839; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L839:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L840; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L840:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    t = alloc_no_clear (UOTAG(10),1);
    t[1] = r1;
    r1 = t;
    r0[2] = r1;
    r1 = ((object*) lenv) [4];
    r0[3] = r1;
    r1 = top[118];
    r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
  L838:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_node_2ffix_498; r1[2] = lenv;
  r0[121] = r1;
  // def node/let_499
  goto L841;
  FUN_node_2flet_499:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = (object*)TC_NIL;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object**) lenv) [1][3];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L842; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L842:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L843; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L843:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    t = alloc_no_clear (UOTAG(9),1);
    t[1] = r1;
    r1 = t;
    r0[2] = r1;
    r1 = ((object*) lenv) [4];
    r0[3] = r1;
    r1 = top[118];
    r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
  L841:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_node_2flet_499; r1[2] = lenv;
  r0[122] = r1;
  // def node/nvcase_500
  goto L844;
  FUN_node_2fnvcase_500:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][6];
    r2 = ((object**) lenv) [1][7];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][5];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][3];
    r2 = ((object**) lenv) [1][4];
    r3 = ((object**) lenv) [1][2];
    t = alloc_no_clear (UOTAG(4),3);
    t[1] = r3;
    t[2] = r1;
    t[3] = r2;
    r1 = t;
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = top[118];
    r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
  L844:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_node_2fnvcase_500; r1[2] = lenv;
  r0[123] = r1;
  // def unpack-fix_504
  goto L845;
  FUN_unpack_fix_504:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = (object *) constructed_36[0];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L846; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L846:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (UOTAG(4), 2);
        r1 = ((object*) lenv) [2];
        r0[1] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = (object *) constructed_36[0];
        r1[3] = r2;
        r2 = top[21];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L847; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
        L847:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_245;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_245_over;
    fatbar_245:
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_930;
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = (object*) &constructed_930;
    PXLL_UNDEFINED; result=r0; goto Lreturn;
    PXLL_RETURN(0);
    fatbar_245_over: ;
    PXLL_RETURN(0);
  L845:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_unpack_fix_504; r1[2] = lenv;
  r0[124] = r1;
  // def literal->string_505
  goto L848;
  FUN_literal__3estring_505:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 6);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = (object *) 8706;
        r2[2] = r3;
        r3 = top[46];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L849; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_char__3estring_348;
        L849:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 2);
        r3 = (object *) 8706;
        r2[2] = r3;
        r3 = top[46];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L850; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_char__3estring_348;
        L850:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object**) lenv) [1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r1 = ((object*) lenv) [6];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[56];
        r0[1] = r1[2]; lenv = r0; goto FUN_int__3estring_364;
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 7);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = (object *) 8962;
        r2[2] = r3;
        r3 = top[46];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L851; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_char__3estring_348;
        L851:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 2);
        r3 = (object *) 23554;
        r2[2] = r3;
        r3 = top[46];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L852; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_char__3estring_348;
        L852:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[46];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L853; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_char__3estring_348;
        L853:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[4] = r2;
        r2 = (object*)TC_NIL;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r2 = ((object*) lenv) [6];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[7] = r2;
        r1 = ((object*) lenv) [7];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
      case (UITAG(3)): {
        r0 = (object*) &constructed_918;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 8);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object***) lenv) [1][1][2];
            r3 = UOBJ_GET(r3,0);
            r2[2] = r3;
            r2 = ((object*) lenv) [2];
            lenv = lenv[1];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][3];
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object***) lenv) [1][1][3];
            r3 = UOBJ_GET(r3,0);
            r2[2] = r3;
            r2 = ((object*) lenv) [2];
            lenv = lenv[1];
            r1[3] = r2;
            r2 = (object*)TC_NIL;
            r1[4] = r2;
            r2 = ((object*) lenv) [4];
            r3 = (object*) &constructed_921;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[5] = r2;
            r2 = ((object*) lenv) [5];
            r3 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[6] = r2;
            r2 = ((object*) lenv) [6];
            r3 = (object*) &constructed_920;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[7] = r2;
            r2 = ((object*) lenv) [7];
            r3 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[8] = r2;
            r1 = ((object*) lenv) [8];
            r2 = (object*) &constructed_919;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            lenv = lenv[1];
            r0[2] = r1;
            r1 = top[47];
            r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
          } break;
          default: {
            goto fatbar_244;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_244_over;
        fatbar_244:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 11);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][2];
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object***) lenv) [1][1][2];
        r3 = UOBJ_GET(r3,0);
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object***) lenv) [1][1][3];
        r3 = UOBJ_GET(r3,0);
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        lenv = lenv[1];
        r1[3] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = top[125];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][4];
        r3[3] = r4;
        r4 = top[32];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L854; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
        L854:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*) &constructed_926;
        r2[3] = r3;
        r3 = top[48];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L855; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
        L855:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[4] = r2;
        r2 = (object*)TC_NIL;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = (object*) &constructed_925;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r2 = ((object*) lenv) [6];
        r3 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[7] = r2;
        r2 = ((object*) lenv) [7];
        r3 = (object*) &constructed_924;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[8] = r2;
        r2 = ((object*) lenv) [8];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[9] = r2;
        r2 = ((object*) lenv) [9];
        r3 = (object*) &constructed_923;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[10] = r2;
        r2 = ((object*) lenv) [10];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[11] = r2;
        r1 = ((object*) lenv) [11];
        r2 = (object*) &constructed_922;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        fatbar_244_over: ;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = top[125];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][2];
        r3[3] = r4;
        r4 = top[32];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L856; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
        L856:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*) &constructed_929;
        r2[3] = r3;
        r3 = top[48];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L857; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_join_353;
        L857:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_928;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_927;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      } break;
    }
    PXLL_RETURN(0);
  L848:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_literal__3estring_505; r1[2] = lenv;
  r0[125] = r1;
  // def flags-repr_506
  goto L858;
  FUN_flags_repr_506:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_2359
    goto L859;
    FUN_loop_2359:
      check_heap (0);
      r0 = (object *) 1;
      r1 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[54];
        r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_361;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = (object *) 3;
        r2 = ((object**) lenv) [1][3];
        r1 = box((pxll_int)unbox(r2)&unbox(r1));
        r0[2] = r1;
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(unbox(r1)==unbox(r0));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = (object *) 3;
          r1 = ((object*) lenv) [3];
          r0 = box((pxll_int)unbox(r1)>>unbox(r0));
          r1 = ((object*) lenv) [2];
          r2 = (object *) 12546;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_loop_2359;
        } else {
          r0 = (object *) 3;
          r1 = ((object*) lenv) [3];
          r0 = box((pxll_int)unbox(r1)>>unbox(r0));
          r1 = ((object*) lenv) [2];
          r2 = (object *) 12290;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_loop_2359;
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L859:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_2359; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) constructed_917[0];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_2359;
  L858:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_flags_repr_506; r1[2] = lenv;
  r0[126] = r1;
  // def pp-node_508
  goto L860;
  FUN_pp_node_508:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def PP_2365
    goto L861;
    FUN__50_50_2365:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object**) lenv) [1][2];
      r3 = ((pxll_vector*)r3)->val[5];
      r2[2] = r3;
      r3 = top[140];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L862; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_apply_subst_531;
      L862:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r2 = top[107];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L863; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_type_repr_462;
      L863:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 12);
      r2[1] = lenv; lenv = r2;
      r3 = allocate (TC_TUPLE, 2);
      r4 = ((object***) lenv) [1][1][2];
      r4 = ((pxll_vector*)r4)->val[1];
      r3[2] = r4;
      r4 = top[56];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L864; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
      L864:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = allocate (TC_TUPLE, 4);
      r4 = (object *) 13;
      r3[2] = r4;
      r4 = ((object*) lenv) [2];
      r3[3] = r4;
      r4 = (object *) 262;
      r3[4] = r4;
      r4 = top[58];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L865; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_pad_367;
      L865:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[3] = r3;
      r3 = allocate (TC_TUPLE, 2);
      r4 = ((object***) lenv) [1][1][2];
      r4 = ((pxll_vector*)r4)->val[2];
      r3[2] = r4;
      r4 = top[56];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L866; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
      L866:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[4] = r3;
      r3 = allocate (TC_TUPLE, 4);
      r4 = (object *) 11;
      r3[2] = r4;
      r4 = ((object*) lenv) [4];
      r3[3] = r4;
      r4 = (object *) 262;
      r3[4] = r4;
      r4 = top[58];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L867; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_pad_367;
      L867:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[5] = r3;
      r3 = top[195];
      r4 = (object *) 5;
      r3 = box((pxll_int)unbox(r4)+unbox(r3));
      r2[6] = r3;
      r3 = ((object***) lenv) [1][1][2];
      r3 = ((pxll_vector*)r3)->val[0];
      r2[7] = r3;
      r3 = allocate (TC_TUPLE, 2);
      r3[1] = lenv; lenv = r3;
      // def loop_2359_i1
      goto L868;
      FUN_loop_2359_i1:
        check_heap (0);
        r0 = (object *) 1;
        r1 = ((object*) lenv) [3];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = top[54];
          r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_361;
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = (object *) 3;
          r2 = ((object**) lenv) [1][3];
          r1 = box((pxll_int)unbox(r2)&unbox(r1));
          r0[2] = r1;
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)==unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = (object *) 3;
            r1 = ((object*) lenv) [3];
            r0 = box((pxll_int)unbox(r1)>>unbox(r0));
            r1 = ((object*) lenv) [2];
            r2 = (object *) 12546;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_2359_i1;
          } else {
            r0 = (object *) 3;
            r1 = ((object*) lenv) [3];
            r0 = box((pxll_int)unbox(r1)>>unbox(r0));
            r1 = ((object*) lenv) [2];
            r2 = (object *) 12290;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_2359_i1;
          }
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L868:
      r4 = allocate (TC_CLOSURE, 2);
      r4[1] = &&FUN_loop_2359_i1; r4[2] = lenv;
      r3[2] = r4;
      r3 = allocate (TC_TUPLE, 3);
      r4 = (object *) constructed_917[0];
      r3[2] = r4;
      r4 = ((object**) lenv) [1][7];
      r3[3] = r4;
      r4 = ((object*) lenv) [2];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L869; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto *r4[1];
      L869:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      lenv = lenv[1];
      r2[8] = r3;
      r3 = allocate (TC_TUPLE, 4);
      r4 = ((object*) lenv) [6];
      r3[2] = r4;
      r4 = ((object*) lenv) [8];
      r3[3] = r4;
      r4 = (object *) 262;
      r3[4] = r4;
      r4 = top[58];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L870; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_pad_367;
      L870:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[9] = r3;
      r3 = (object*)TC_NIL;
      r2[10] = r3;
      r3 = ((object*) lenv) [10];
      r4 = ((object*) lenv) [9];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[11] = r3;
      r3 = ((object*) lenv) [11];
      r4 = ((object*) lenv) [5];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[12] = r3;
      r2 = ((object*) lenv) [12];
      r3 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L871; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L871:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r0 = top[158];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L872; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L872:
      ; lenv = k[2]; k = k[1];
      r0 = ((object*) lenv) [3];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [3];
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object *) 1;
      r1 = ((object**) lenv) [1][3];
      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
        r0 = (object *) 262;
      } else {
        goto fatbar_243;
      }
      goto fatbar_243_over;
      fatbar_243:
      r0 = (object*) &constructed_916;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_916;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) 3;
      r2 = ((object**) lenv) [1][3];
      r1 = box((pxll_int)unbox(r2)-unbox(r1));
      r0[2] = r1;
      r1 = top[92];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L873; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_indent_434;
      L873:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      fatbar_243_over: ;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_884;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = ((object*) lenv) [6];
          r3 = (object*) &constructed_883;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L874; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L874:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [3];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_886;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = ((object*) lenv) [6];
          r3 = (object*) &constructed_885;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L875; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L875:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [3];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(2)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][2];
          r3[2] = r4;
          r4 = top[125];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L876; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_literal__3estring_505;
          L876:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_888;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = ((object*) lenv) [6];
          r3 = (object*) &constructed_887;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L877; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L877:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [3];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(3)): {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 8);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][3];
          r3[2] = r4;
          r4 = top[107];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L878; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_type_repr_462;
          L878:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_891;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object**) lenv) [1][4];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = (object*) &constructed_890;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r3 = ((object*) lenv) [7];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[8] = r3;
          r2 = ((object*) lenv) [8];
          r3 = (object*) &constructed_889;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L879; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L879:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r0 = ((object*) lenv) [5];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [5];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UITAG(5)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = ((object*) lenv) [3];
          r3 = (object*) &constructed_892;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L880; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L880:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UITAG(6)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = ((object*) lenv) [3];
          r3 = (object*) &constructed_893;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L881; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L881:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UITAG(8)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = ((object*) lenv) [3];
          r3 = (object*) &constructed_894;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L882; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L882:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(7)): {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 9);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = allocate (TC_TUPLE, 3);
          r4 = allocate (TC_TUPLE, 3);
          r5 = top[78];
          r4[2] = r5;
          r5 = ((object**) lenv) [1][3];
          r4[3] = r5;
          r5 = top[32];
          t = allocate (TC_SAVE, 7);
          t[1] = k; t[2] = lenv; t[3] = &&L883; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
          r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
          L883:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
          r4 = result;
          r3[2] = r4;
          r4 = (object*) &constructed_898;
          r3[3] = r4;
          r4 = top[48];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L884; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
          L884:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = (object*)TC_NIL;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = (object*) &constructed_897;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r3 = ((object*) lenv) [7];
          r4 = (object*) &constructed_896;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[8] = r3;
          r3 = ((object*) lenv) [8];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[9] = r3;
          r2 = ((object*) lenv) [9];
          r3 = (object*) &constructed_895;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L885; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L885:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r0 = ((object*) lenv) [4];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [4];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(10)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = allocate (TC_TUPLE, 3);
          r5 = top[78];
          r4[2] = r5;
          r5 = ((object**) lenv) [1][2];
          r4[3] = r5;
          r5 = top[32];
          t = allocate (TC_SAVE, 7);
          t[1] = k; t[2] = lenv; t[3] = &&L886; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
          r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
          L886:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
          r4 = result;
          r3[2] = r4;
          r4 = (object*) &constructed_901;
          r3[3] = r4;
          r4 = top[48];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L887; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
          L887:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_900;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = ((object*) lenv) [6];
          r3 = (object*) &constructed_899;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L888; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L888:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [3];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(4)): {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 12);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = allocate (TC_TUPLE, 3);
          r4 = allocate (TC_TUPLE, 3);
          r5 = top[78];
          r4[2] = r5;
          r5 = ((object**) lenv) [1][3];
          r4[3] = r5;
          r5 = top[32];
          t = allocate (TC_SAVE, 7);
          t[1] = k; t[2] = lenv; t[3] = &&L889; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
          r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
          L889:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
          r4 = result;
          r3[2] = r4;
          r4 = (object*) &constructed_907;
          r3[3] = r4;
          r4 = top[48];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L890; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
          L890:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = allocate (TC_TUPLE, 3);
          r4 = allocate (TC_TUPLE, 3);
          r5 = top[56];
          r4[2] = r5;
          r5 = ((object**) lenv) [1][4];
          r4[3] = r5;
          r5 = top[32];
          t = allocate (TC_SAVE, 7);
          t[1] = k; t[2] = lenv; t[3] = &&L891; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
          r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
          L891:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
          r4 = result;
          r3[2] = r4;
          r4 = (object*) &constructed_906;
          r3[3] = r4;
          r4 = top[48];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L892; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
          L892:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[4] = r3;
          r3 = (object*)TC_NIL;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = (object*) &constructed_905;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r3 = ((object*) lenv) [7];
          r4 = ((object*) lenv) [4];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[8] = r3;
          r3 = ((object*) lenv) [8];
          r4 = (object*) &constructed_904;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[9] = r3;
          r3 = ((object*) lenv) [9];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[10] = r3;
          r3 = ((object*) lenv) [10];
          r4 = (object*) &constructed_903;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[11] = r3;
          r3 = ((object*) lenv) [11];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[12] = r3;
          r2 = ((object*) lenv) [12];
          r3 = (object*) &constructed_902;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L893; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L893:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r0 = ((object*) lenv) [5];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [5];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(11)): {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 7);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = ((object**) lenv) [1][3];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][3];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[3] = r3;
          r3 = (object*)TC_NIL;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = (object*) &constructed_909;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r2 = ((object*) lenv) [7];
          r3 = (object*) &constructed_908;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L894; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L894:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r0 = ((object*) lenv) [4];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [4];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        case (UOTAG(12)): {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 9);
          r2[1] = lenv; lenv = r2;
          r3 = ((object**) lenv) [1][2];
          r3 = allocate (TC_TUPLE, 2);
          r3[1] = lenv; lenv = r3;
          r4 = ((object***) lenv) [1][1][2];
          r4 = UOBJ_GET(r4,0);
          r3[2] = r4;
          r3 = ((object*) lenv) [2];
          lenv = lenv[1];
          r2[2] = r3;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][3];
          r3[2] = r4;
          r4 = top[91];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L895; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_repr_433;
          L895:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = (object*)TC_NIL;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = (object*) &constructed_912;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r3 = ((object*) lenv) [7];
          r4 = (object*) &constructed_911;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[8] = r3;
          r3 = ((object*) lenv) [8];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[9] = r3;
          r2 = ((object*) lenv) [9];
          r3 = (object*) &constructed_910;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L896; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L896:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r0 = ((object*) lenv) [4];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [4];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = allocate (TC_TUPLE, 3);
          r5 = top[78];
          r4[2] = r5;
          r5 = ((object**) lenv) [1][2];
          r4[3] = r5;
          r5 = top[32];
          t = allocate (TC_SAVE, 7);
          t[1] = k; t[2] = lenv; t[3] = &&L897; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
          r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
          L897:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
          r4 = result;
          r3[2] = r4;
          r4 = (object*) &constructed_915;
          r3[3] = r4;
          r4 = top[48];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L898; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
          L898:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_914;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = ((object*) lenv) [6];
          r3 = (object*) &constructed_913;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L899; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L899:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [3];
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
        } break;
      }
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][3];
      r2 = (object *) 3;
      r1 = box((pxll_int)unbox(r2)+unbox(r1));
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_14246_i1642
      goto L900;
      FUN_lambda_14246_i1642:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN__50_50_2365;
      L900:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_14246_i1642; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L861:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN__50_50_2365; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L901; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN__50_50_2365;
    L901:
    ; lenv = k[2]; k = k[1];
    r0 = top[158];
    lenv = r0[2]; goto *r0[1];
  L860:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_pp_node_508; r1[2] = lenv;
  r0[127] = r1;
  // def get-formals_509
  goto L902;
  FUN_get_formals_509:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def p_2393
    goto L903;
    FUN_p_2393:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_242;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_242_over;
      fatbar_242:
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_882;
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][2];
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      fatbar_242_over: ;
      PXLL_RETURN(0);
    L903:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_p_2393; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = (object *) constructed_881[0];
    r1[3] = r2;
    r2 = ((object***) lenv) [1][1][2];
    r1[4] = r2;
    r2 = top[37];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L904; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_fold_331;
    L904:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = (object *) constructed_36[0];
    r0[3] = r1;
    r1 = top[21];
    r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
  L902:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_get_formals_509; r1[2] = lenv;
  r0[128] = r1;
  // def unpack-bindings_510
  goto L905;
  FUN_unpack_bindings_510:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_2398
    goto L906;
    FUN_loop_2398:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (UOTAG(1), 2);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L907; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L907:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[1] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [4];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L908; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L908:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UOTAG(1)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = ((object**) lenv) [1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_NIL): {
                              r0 = ((object******) lenv) [1][1][1][1][1][3];
                              r1 = ((object**) lenv) [1][2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r1;
                              t[2] = r0;
                              r0 = t;
                              r1 = ((object******) lenv) [1][1][1][1][1][4];
                              r2 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r2;
                              t[2] = r1;
                              r1 = t;
                              r2 = ((object*****) lenv) [1][1][1][1][3];
                              lenv = ((object *****)lenv)[1][1][1][1][1];
                              lenv[2] = r2;
                              lenv[3] = r0;
                              lenv[4] = r1;
                              goto FUN_loop_2398;
                            } break;
                            default: {
                              lenv = ((object *****)lenv)[1][1][1][1][1];
                              goto fatbar_241;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ****)lenv)[1][1][1][1];
                          goto fatbar_241;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_241;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object **)lenv)[1][1];
                  goto fatbar_241;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_241;
            }
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
      goto fatbar_241_over;
      fatbar_241:
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_880;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      fatbar_241_over: ;
      PXLL_RETURN(0);
    L906:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_2398; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = (object *) constructed_879[0];
    r0[3] = r1;
    r1 = (object *) constructed_878[0];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_2398;
  L905:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_unpack_bindings_510; r1[2] = lenv;
  r0[129] = r1;
  // def parse-cexp-sig_511
  goto L909;
  FUN_parse_cexp_sig_511:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = top[197];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L910; t[4] = r0; k = t;
    lenv = r1[2]; goto *r1[1];
    L910:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object*) lenv) [2];
    r1[3] = r2;
    r2 = top[111];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L911; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_parse_type_splat_467;
    L911:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r0 = allocate (UOTAG(2), 2);
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = ((object*) lenv) [2];
    r2 = ((pxll_vector*)r2)->val[1];
    r2 = ((pxll_vector*)r2)->val[7];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L912; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L912:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[1] = r1;
    r1 = ((object*) lenv) [3];
    r0[2] = r1;
    PXLL_RETURN(0);
  L909:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_parse_cexp_sig_511; r1[2] = lenv;
  r0[130] = r1;
  // def sort-fix-inits_512
  goto L913;
  FUN_sort_fix_inits_512:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 6);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_877[0];
    r0[2] = r1;
    r1 = (object *) constructed_876[0];
    r0[3] = r1;
    r1 = (object *) constructed_875[0];
    r0[4] = r1;
    r1 = (object *) constructed_874[0];
    r0[5] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L914; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L914:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[6] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][6];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def $loop_2421
    goto L915;
    FUN__24loop_2421:
      check_heap (0);
      r0 = ((object***) lenv) [1][1][2];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      } else {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object******) lenv) [1][1][1][1][1][2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[3] = r2;
        r2 = top[27];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L916; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
        L916:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object******) lenv) [1][1][1][1][1][3];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[3] = r2;
        r2 = top[27];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L917; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
        L917:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = ((object*) lenv) [3];
        r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (UOTAG(7)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object******) lenv) [1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object*****) lenv) [1][1][1][1][2] = r0;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object******) lenv) [1][1][1][1][1][4];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object*****) lenv) [1][1][1][1][4] = r0;
          } break;
          default: {
            goto fatbar_240;
          }
        }
        goto fatbar_240_over;
        fatbar_240:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object******) lenv) [1][1][1][1][1][3];
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object*****) lenv) [1][1][1][1][3] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object******) lenv) [1][1][1][1][1][5];
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object*****) lenv) [1][1][1][1][5] = r0;
        fatbar_240_over: ;
        lenv = lenv[1];
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        lenv[2] = r0;
        goto FUN__24loop_2421;
      }
      PXLL_RETURN(0);
    L915:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN__24loop_2421; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object *) 1;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L918; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN__24loop_2421;
    L918:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    lenv = lenv[1];
    lenv = lenv[1];
    r0 = allocate (UOTAG(3), 2);
    r1 = allocate (TC_TUPLE, 5);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [2];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L919; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L919:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[4] = r2;
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [4];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L920; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L920:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[5] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [3];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L921; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L921:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = ((object*) lenv) [5];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L922; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L922:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[1] = r1;
    r1 = allocate (TC_TUPLE, 5);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][4];
    r1[2] = r2;
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [2];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L923; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L923:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[3] = r2;
    r2 = ((object**) lenv) [1][5];
    r1[4] = r2;
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [4];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L924; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L924:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[5] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [3];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L925; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L925:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = ((object*) lenv) [5];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L926; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L926:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[2] = r1;
    PXLL_RETURN(0);
  L913:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_sort_fix_inits_512; r1[2] = lenv;
  r0[131] = r1;
  // def walk_513
  goto L927;
  FUN_walk_513:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_792[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_809[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_809[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_809[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(r1) {
          r1 = (object *) constructed_44[0];
        } else {
          r1 = (object *) constructed_41[0];
        }
        r0[3] = r1;
        r1 = ((object*) lenv) [3];
        r2 = (object *) constructed_860[0];
        r3 = (object *) constructed_46[0];
        t = alloc_no_clear (UOTAG(5),3);
        t[1] = r3;
        t[2] = r1;
        t[3] = r2;
        r1 = t;
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [4];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_809[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UITAG(6)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = (object*)UITAG(3);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_809[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(8)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 4);
        // def lambda_15011
        goto L928;
        FUN_lambda_15011:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          r0 = allocate (TC_TUPLE, 8);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(1),1);
          t[1] = r1;
          r1 = t;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = top[132];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L929; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
          L929:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = (object*)TC_NIL;
          r0[6] = r1;
          r1 = ((object*) lenv) [6];
          r2 = ((object*) lenv) [5];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[7] = r1;
          r1 = ((object*) lenv) [7];
          r2 = ((object**) lenv) [1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[8] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [4];
          r2 = (object *) constructed_491[0];
          t = alloc_no_clear (UOTAG(12),2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[2] = r1;
          r1 = ((object*) lenv) [8];
          r0[3] = r1;
          r1 = top[118];
          r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
          PXLL_RETURN(0);
        L928:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_15011; r1[2] = lenv;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = (object *) 6;
        t = alloc_no_clear (UOTAG(4),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r3 = (object *) constructed_488[0];
        t = alloc_no_clear (UOTAG(12),2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = (object *) constructed_861[0];
        r1[3] = r2;
        r2 = top[118];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L930; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
        L930:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r0[4] = r1;
        r1 = top[38];
        r0[1] = r1[2]; lenv = r0; goto FUN_foldr_332;
      } break;
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 7);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[132];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L931; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
        L931:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = (object*)TC_NIL;
        r0[6] = r1;
        r1 = ((object*) lenv) [6];
        r2 = ((object*) lenv) [5];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[7] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [4];
        r2 = (object *) constructed_485[0];
        t = alloc_no_clear (UOTAG(12),2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = ((object*) lenv) [7];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(7),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = top[134];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L932; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_build_literal_515;
        L932:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_809[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r3 = allocate (TC_TUPLE, 6);
        r3[1] = lenv; lenv = r3;
        r4 = ((object**) lenv) [1][2];
        r4 = allocate (TC_TUPLE, 2);
        r4[1] = lenv; lenv = r4;
        r5 = ((object***) lenv) [1][1][2];
        r5 = UOBJ_GET(r5,0);
        r4[2] = r5;
        r4 = ((object*) lenv) [2];
        lenv = lenv[1];
        r3[2] = r4;
        r4 = ((object**) lenv) [1][3];
        r4 = allocate (TC_TUPLE, 2);
        r4[1] = lenv; lenv = r4;
        r5 = ((object***) lenv) [1][1][3];
        r5 = UOBJ_GET(r5,0);
        r4[2] = r5;
        r4 = ((object*) lenv) [2];
        lenv = lenv[1];
        r3[3] = r4;
        r4 = (object*)TC_NIL;
        r3[4] = r4;
        r4 = ((object*) lenv) [4];
        r5 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r5;
        t[2] = r4;
        r4 = t;
        r3[5] = r4;
        r4 = ((object*) lenv) [5];
        r5 = (object*) &constructed_862;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r5;
        t[2] = r4;
        r4 = t;
        r3[6] = r4;
        r3 = ((object*) lenv) [6];
        r4 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        lenv = lenv[1];
        r2[2] = r3;
        r3 = top[47];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L933; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
        L933:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = top[80];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L934; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
        L934:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [4];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = (object *) constructed_792[0];
        r0[3] = r1;
        r1 = top[118];
        r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = (object *) constructed_864[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(0)): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [2];
                              switch (get_case (r0)) {
                                case (UOTAG(1)): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][2];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r0 = ((object**) lenv) [1][3];
                                  switch (get_case (r0)) {
                                    case (TC_PAIR): {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object***) lenv) [1][1][3];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = ((object***) lenv) [1][1][3];
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r0 = ((object*) lenv) [2];
                                      switch (get_case (r0)) {
                                        case (UOTAG(1)): {
                                          r0 = allocate (TC_TUPLE, 2);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object**) lenv) [1][2];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r0 = ((object**) lenv) [1][3];
                                          switch (get_case (r0)) {
                                            case (TC_NIL): {
                                              r0 = ((object******) lenv) [1][1][1][1][1][3];
                                              switch (get_case (r0)) {
                                                case (TC_PAIR): {
                                                  r0 = allocate (TC_TUPLE, 3);
                                                  r0[1] = lenv; lenv = r0;
                                                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                                  r1 = UOBJ_GET(r1,0);
                                                  r0[2] = r1;
                                                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                                  r1 = UOBJ_GET(r1,1);
                                                  r0[3] = r1;
                                                  r0 = ((object*) lenv) [3];
                                                  switch (get_case (r0)) {
                                                    case (TC_NIL): {
                                                      r0 = allocate (TC_TUPLE, 2);
                                                      r0[1] = lenv; lenv = r0;
                                                      r1 = allocate (TC_TUPLE, 2);
                                                      r2 = ((object**) lenv) [1][2];
                                                      r1[2] = r2;
                                                      r2 = top[132];
                                                      t = allocate (TC_SAVE, 4);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L935; t[4] = r0; k = t;
                                                      r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                                      L935:
                                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                                      r1 = result;
                                                      r0[2] = r1;
                                                      r0 = allocate (TC_TUPLE, 3);
                                                      r1 = ((object***) lenv) [1][1][2];
                                                      r2 = ((object*****) lenv) [1][1][1][1][2];
                                                      t = alloc_no_clear (UOTAG(11),2);
                                                      t[1] = r2;
                                                      t[2] = r1;
                                                      r1 = t;
                                                      r0[2] = r1;
                                                      r1 = allocate (TC_TUPLE, 2);
                                                      r1[1] = lenv; lenv = r1;
                                                      r2 = (object*)TC_NIL;
                                                      r1[2] = r2;
                                                      r1 = ((object*) lenv) [2];
                                                      r2 = ((object**) lenv) [1][2];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r2;
                                                      t[2] = r1;
                                                      r1 = t;
                                                      lenv = lenv[1];
                                                      r0[3] = r1;
                                                      r1 = top[118];
                                                      r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                                                    } break;
                                                    default: {
                                                      lenv = ((object *******)lenv)[1][1][1][1][1][1][1];
                                                      goto fatbar_227;
                                                    }
                                                  }
                                                  PXLL_RETURN(0);
                                                } break;
                                                default: {
                                                  lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                                  goto fatbar_227;
                                                }
                                              }
                                              PXLL_RETURN(0);
                                            } break;
                                            default: {
                                              lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                              goto fatbar_227;
                                            }
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object *****)lenv)[1][1][1][1][1];
                                          goto fatbar_227;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      lenv = ((object ****)lenv)[1][1][1][1];
                                      goto fatbar_227;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_227;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_227;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_227;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_227;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_227;
                }
                PXLL_RETURN(0);
                goto fatbar_227_over;
                fatbar_227:
                r0 = (object *) constructed_866[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(0)): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = ((object*) lenv) [2];
                          r1[2] = r2;
                          r2 = top[129];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L936; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_unpack_bindings_510;
                          L936:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case_noint (r0)) {
                            default: {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = allocate (TC_TUPLE, 4);
                              r1 = ((object*) lenv) [2];
                              r0[2] = r1;
                              r1 = allocate (TC_TUPLE, 3);
                              r2 = top[132];
                              r1[2] = r2;
                              r2 = ((object*) lenv) [3];
                              r1[3] = r2;
                              r2 = top[32];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L937; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                              L937:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[3] = r1;
                              r1 = allocate (TC_TUPLE, 2);
                              r1[1] = lenv; lenv = r1;
                              r2 = allocate (TC_TUPLE, 3);
                              r3 = top[132];
                              r2[2] = r3;
                              r3 = ((object****) lenv) [1][1][1][3];
                              r2[3] = r3;
                              r3 = top[32];
                              t = allocate (TC_SAVE, 5);
                              t[1] = k; t[2] = lenv; t[3] = &&L938; t[4] = r0; t[5] = r1; k = t;
                              r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
                              L938:
                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                              r2 = result;
                              r1[2] = r2;
                              r1 = allocate (TC_TUPLE, 3);
                              r2 = (object*)UITAG(5);
                              r1[2] = r2;
                              r2 = ((object*) lenv) [2];
                              r1[3] = r2;
                              r2 = top[118];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L939; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
                              L939:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              lenv = lenv[1];
                              r0[4] = r1;
                              r1 = top[121];
                              r0[1] = r1[2]; lenv = r0; goto FUN_node_2ffix_498;
                            } break;
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_228;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_228;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_228;
                }
                PXLL_RETURN(0);
                goto fatbar_228_over;
                fatbar_228:
                r0 = (object *) constructed_605[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(0)): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = ((object*) lenv) [2];
                          r1[2] = r2;
                          r2 = top[129];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L940; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_unpack_bindings_510;
                          L940:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case_noint (r0)) {
                            default: {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = allocate (TC_TUPLE, 4);
                              r1 = ((object*) lenv) [2];
                              r0[2] = r1;
                              r1 = allocate (TC_TUPLE, 3);
                              r2 = top[132];
                              r1[2] = r2;
                              r2 = ((object*) lenv) [3];
                              r1[3] = r2;
                              r2 = top[32];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L941; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                              L941:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[3] = r1;
                              r1 = allocate (TC_TUPLE, 2);
                              r1[1] = lenv; lenv = r1;
                              r2 = allocate (TC_TUPLE, 3);
                              r3 = top[132];
                              r2[2] = r3;
                              r3 = ((object****) lenv) [1][1][1][3];
                              r2[3] = r3;
                              r3 = top[32];
                              t = allocate (TC_SAVE, 5);
                              t[1] = k; t[2] = lenv; t[3] = &&L942; t[4] = r0; t[5] = r1; k = t;
                              r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
                              L942:
                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                              r2 = result;
                              r1[2] = r2;
                              r1 = allocate (TC_TUPLE, 3);
                              r2 = (object*)UITAG(5);
                              r1[2] = r2;
                              r2 = ((object*) lenv) [2];
                              r1[3] = r2;
                              r2 = top[118];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L943; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
                              L943:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              lenv = lenv[1];
                              r0[4] = r1;
                              r1 = top[122];
                              r0[1] = r1[2]; lenv = r0; goto FUN_node_2flet_499;
                            } break;
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_229;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_229;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_229;
                }
                PXLL_RETURN(0);
                goto fatbar_229_over;
                fatbar_229:
                r0 = (object *) constructed_685[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(0)): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [2];
                              switch (get_case (r0)) {
                                case (UOTAG(0)): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][2];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = allocate (TC_TUPLE, 2);
                                  r3 = ((object***) lenv) [1][1][2];
                                  r2[2] = r3;
                                  r3 = top[128];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L944; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_get_formals_509;
                                  L944:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[2] = r2;
                                  r2 = allocate (TC_TUPLE, 3);
                                  r3 = top[132];
                                  r2[2] = r3;
                                  r3 = ((object*) lenv) [2];
                                  r2[3] = r3;
                                  r3 = top[32];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L945; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
                                  L945:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[3] = r2;
                                  r2 = top[131];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L946; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_sort_fix_inits_512;
                                  L946:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r0 = ((object*) lenv) [3];
                                  switch (get_case_noint (r0)) {
                                    default: {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][3];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = ((object**) lenv) [1][3];
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r0 = allocate (TC_TUPLE, 4);
                                      r1 = ((object*) lenv) [2];
                                      r0[2] = r1;
                                      r1 = ((object*) lenv) [3];
                                      r0[3] = r1;
                                      r1 = allocate (TC_TUPLE, 2);
                                      r1[1] = lenv; lenv = r1;
                                      r2 = allocate (TC_TUPLE, 3);
                                      r3 = top[132];
                                      r2[2] = r3;
                                      r3 = ((object****) lenv) [1][1][1][3];
                                      r2[3] = r3;
                                      r3 = top[32];
                                      t = allocate (TC_SAVE, 5);
                                      t[1] = k; t[2] = lenv; t[3] = &&L947; t[4] = r0; t[5] = r1; k = t;
                                      r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
                                      L947:
                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                      r2 = result;
                                      r1[2] = r2;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = (object*)UITAG(5);
                                      r1[2] = r2;
                                      r2 = ((object*) lenv) [2];
                                      r1[3] = r2;
                                      r2 = top[118];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L948; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
                                      L948:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      lenv = lenv[1];
                                      r0[4] = r1;
                                      r1 = top[121];
                                      r0[1] = r1[2]; lenv = r0; goto FUN_node_2ffix_498;
                                    } break;
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_230;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_230;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_230;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_230;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_230;
                }
                PXLL_RETURN(0);
                goto fatbar_230_over;
                fatbar_230:
                r0 = (object *) constructed_601[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(1)): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [2];
                              switch (get_case (r0)) {
                                case (UOTAG(0)): {
                                  r0 = allocate (TC_TUPLE, 5);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][2];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object*) lenv) [2];
                                  r1[2] = r2;
                                  r2 = top[128];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L949; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_get_formals_509;
                                  L949:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = top[132];
                                  r1[2] = r2;
                                  r2 = ((object**) lenv) [1][3];
                                  r1[3] = r2;
                                  r2 = top[32];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L950; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                                  L950:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[4] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = (object*)UITAG(5);
                                  r1[2] = r2;
                                  r2 = ((object*) lenv) [4];
                                  r1[3] = r2;
                                  r2 = top[118];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L951; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
                                  L951:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[5] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object*) lenv) [3];
                                  r2 = ((object***) lenv) [1][1][2];
                                  t = alloc_no_clear (UOTAG(7),2);
                                  t[1] = r2;
                                  t[2] = r1;
                                  r1 = t;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = (object*)TC_NIL;
                                  r1[2] = r2;
                                  r1 = ((object*) lenv) [2];
                                  r2 = ((object**) lenv) [1][5];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r2;
                                  t[2] = r1;
                                  r1 = t;
                                  lenv = lenv[1];
                                  r0[3] = r1;
                                  r1 = top[118];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_231;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_231;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_231;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_231;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_231;
                }
                PXLL_RETURN(0);
                goto fatbar_231_over;
                fatbar_231:
                r0 = (object *) constructed_613[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(1)): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_PAIR): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,1);
                                  r0[3] = r1;
                                  r0 = ((object*) lenv) [2];
                                  switch (get_case (r0)) {
                                    case (UOTAG(0)): {
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][2];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r0 = ((object**) lenv) [1][3];
                                      switch (get_case (r0)) {
                                        case (TC_PAIR): {
                                          r0 = allocate (TC_TUPLE, 3);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object***) lenv) [1][1][3];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r1 = ((object***) lenv) [1][1][3];
                                          r1 = UOBJ_GET(r1,1);
                                          r0[3] = r1;
                                          r0 = ((object*) lenv) [2];
                                          switch (get_case (r0)) {
                                            case (UOTAG(0)): {
                                              r0 = allocate (TC_TUPLE, 2);
                                              r0[1] = lenv; lenv = r0;
                                              r1 = ((object**) lenv) [1][2];
                                              r1 = UOBJ_GET(r1,0);
                                              r0[2] = r1;
                                              r0 = ((object**) lenv) [1][3];
                                              switch (get_case (r0)) {
                                                case (TC_PAIR): {
                                                  r0 = allocate (TC_TUPLE, 3);
                                                  r0[1] = lenv; lenv = r0;
                                                  r1 = ((object***) lenv) [1][1][3];
                                                  r1 = UOBJ_GET(r1,0);
                                                  r0[2] = r1;
                                                  r1 = ((object***) lenv) [1][1][3];
                                                  r1 = UOBJ_GET(r1,1);
                                                  r0[3] = r1;
                                                  r0 = ((object*) lenv) [2];
                                                  switch (get_case (r0)) {
                                                    case (UOTAG(0)): {
                                                      r0 = allocate (TC_TUPLE, 2);
                                                      r0[1] = lenv; lenv = r0;
                                                      r1 = ((object**) lenv) [1][2];
                                                      r1 = UOBJ_GET(r1,0);
                                                      r0[2] = r1;
                                                      r0 = ((object**) lenv) [1][3];
                                                      switch (get_case (r0)) {
                                                        case (TC_PAIR): {
                                                          r0 = allocate (TC_TUPLE, 3);
                                                          r0[1] = lenv; lenv = r0;
                                                          r1 = ((object***) lenv) [1][1][3];
                                                          r1 = UOBJ_GET(r1,0);
                                                          r0[2] = r1;
                                                          r1 = ((object***) lenv) [1][1][3];
                                                          r1 = UOBJ_GET(r1,1);
                                                          r0[3] = r1;
                                                          r0 = ((object*) lenv) [3];
                                                          switch (get_case (r0)) {
                                                            case (TC_NIL): {
                                                              r0 = allocate (TC_TUPLE, 7);
                                                              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                                                              r0[2] = r1;
                                                              r1 = allocate (TC_TUPLE, 3);
                                                              r2 = top[86];
                                                              r1[2] = r2;
                                                              r2 = ((object******) lenv) [1][1][1][1][1][2];
                                                              r1[3] = r2;
                                                              r2 = top[32];
                                                              t = allocate (TC_SAVE, 4);
                                                              t[1] = k; t[2] = lenv; t[3] = &&L952; t[4] = r0; k = t;
                                                              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                                                              L952:
                                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                                              r1 = result;
                                                              r0[3] = r1;
                                                              r1 = allocate (TC_TUPLE, 3);
                                                              r2 = top[87];
                                                              r1[2] = r2;
                                                              r2 = ((object****) lenv) [1][1][1][2];
                                                              r1[3] = r2;
                                                              r2 = top[32];
                                                              t = allocate (TC_SAVE, 4);
                                                              t[1] = k; t[2] = lenv; t[3] = &&L953; t[4] = r0; k = t;
                                                              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                                                              L953:
                                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                                              r1 = result;
                                                              r0[4] = r1;
                                                              r1 = allocate (TC_TUPLE, 2);
                                                              r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                                              r1[2] = r2;
                                                              r2 = top[132];
                                                              t = allocate (TC_SAVE, 4);
                                                              t[1] = k; t[2] = lenv; t[3] = &&L954; t[4] = r0; k = t;
                                                              r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                                              L954:
                                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                                              r1 = result;
                                                              r0[5] = r1;
                                                              r1 = allocate (TC_TUPLE, 3);
                                                              r2 = top[132];
                                                              r1[2] = r2;
                                                              r2 = ((object**) lenv) [1][2];
                                                              r1[3] = r2;
                                                              r2 = top[32];
                                                              t = allocate (TC_SAVE, 4);
                                                              t[1] = k; t[2] = lenv; t[3] = &&L955; t[4] = r0; k = t;
                                                              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                                                              L955:
                                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                                              r1 = result;
                                                              r0[6] = r1;
                                                              r1 = allocate (TC_TUPLE, 2);
                                                              r2 = ((object*) lenv) [2];
                                                              r1[2] = r2;
                                                              r2 = top[132];
                                                              t = allocate (TC_SAVE, 4);
                                                              t[1] = k; t[2] = lenv; t[3] = &&L956; t[4] = r0; k = t;
                                                              r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                                              L956:
                                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                                              r1 = result;
                                                              r0[7] = r1;
                                                              r1 = top[123];
                                                              r0[1] = r1[2]; lenv = r0; goto FUN_node_2fnvcase_500;
                                                            } break;
                                                            default: {
                                                              lenv = ((object **********)lenv)[1][1][1][1][1][1][1][1][1][1];
                                                              goto fatbar_232;
                                                            }
                                                          }
                                                          PXLL_RETURN(0);
                                                        } break;
                                                        default: {
                                                          lenv = ((object *********)lenv)[1][1][1][1][1][1][1][1][1];
                                                          goto fatbar_232;
                                                        }
                                                      }
                                                      PXLL_RETURN(0);
                                                    } break;
                                                    default: {
                                                      lenv = ((object ********)lenv)[1][1][1][1][1][1][1][1];
                                                      goto fatbar_232;
                                                    }
                                                  }
                                                  PXLL_RETURN(0);
                                                } break;
                                                default: {
                                                  lenv = ((object *******)lenv)[1][1][1][1][1][1][1];
                                                  goto fatbar_232;
                                                }
                                              }
                                              PXLL_RETURN(0);
                                            } break;
                                            default: {
                                              lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                              goto fatbar_232;
                                            }
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object *****)lenv)[1][1][1][1][1];
                                          goto fatbar_232;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      lenv = ((object ****)lenv)[1][1][1][1];
                                      goto fatbar_232;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_232;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_232;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_232;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_232;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_232;
                }
                PXLL_RETURN(0);
                goto fatbar_232_over;
                fatbar_232:
                r0 = (object *) constructed_615[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (UOTAG(2)): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = ((object***) lenv) [1][1][2];
                              r1[2] = r2;
                              r2 = top[130];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L957; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_parse_cexp_sig_511;
                              L957:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case_noint (r0)) {
                                default: {
                                  r0 = allocate (TC_TUPLE, 4);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,1);
                                  r0[3] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = top[132];
                                  r1[2] = r2;
                                  r2 = ((object***) lenv) [1][1][3];
                                  r1[3] = r2;
                                  r2 = top[32];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L958; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                                  L958:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[4] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object*) lenv) [3];
                                  r2 = ((object**) lenv) [1][2];
                                  r3 = ((object*) lenv) [2];
                                  t = alloc_no_clear (UOTAG(3),3);
                                  t[1] = r3;
                                  t[2] = r1;
                                  t[3] = r2;
                                  r1 = t;
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [4];
                                  r0[3] = r1;
                                  r1 = top[118];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                                } break;
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_233;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_233;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_233;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_233;
                }
                PXLL_RETURN(0);
                goto fatbar_233_over;
                fatbar_233:
                r0 = (object *) constructed_595[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 4);
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object***) lenv) [1][1][2];
                                  r1[2] = r2;
                                  r2 = top[132];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L959; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                  L959:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r2 = top[132];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L960; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                  L960:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object*) lenv) [2];
                                  r1[2] = r2;
                                  r2 = top[132];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L961; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                  L961:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[4] = r1;
                                  r1 = top[120];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_node_2fif_493;
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_234;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_234;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_234;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_234;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_234;
                }
                PXLL_RETURN(0);
                goto fatbar_234_over;
                fatbar_234:
                r0 = (object *) constructed_868[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_NIL): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = ((object**) lenv) [1][2];
                          r1[2] = r2;
                          r2 = top[134];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L962; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_build_literal_515;
                          L962:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r0 = allocate (TC_TUPLE, 3);
                          r1 = ((object*) lenv) [2];
                          t = alloc_no_clear (UOTAG(2),1);
                          t[1] = r1;
                          r1 = t;
                          r0[2] = r1;
                          r1 = (object *) constructed_809[0];
                          r0[3] = r1;
                          r1 = top[118];
                          r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_235;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_235;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_235;
                }
                PXLL_RETURN(0);
                goto fatbar_235_over;
                fatbar_235:
                r0 = (object *) constructed_870[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_NIL): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = ((object**) lenv) [1][2];
                          r1[2] = r2;
                          r2 = top[134];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L963; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_build_literal_515;
                          L963:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r0 = allocate (TC_TUPLE, 3);
                          r1 = ((object*) lenv) [2];
                          t = alloc_no_clear (UOTAG(2),1);
                          t[1] = r1;
                          r1 = t;
                          r0[2] = r1;
                          r1 = (object *) constructed_809[0];
                          r0[3] = r1;
                          r1 = top[118];
                          r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_236;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_236;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_236;
                }
                PXLL_RETURN(0);
                goto fatbar_236_over;
                fatbar_236:
                r0 = (object *) constructed_597[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (UOTAG(1)): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r0 = ((object**) lenv) [1][3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object***) lenv) [1][1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r2 = top[132];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L964; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                                  L964:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object***) lenv) [1][1][2];
                                  t = alloc_no_clear (UOTAG(1),1);
                                  t[1] = r1;
                                  r1 = t;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = (object*)TC_NIL;
                                  r1[2] = r2;
                                  r1 = ((object*) lenv) [2];
                                  r2 = ((object**) lenv) [1][2];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r2;
                                  t[2] = r1;
                                  r1 = t;
                                  lenv = lenv[1];
                                  r0[3] = r1;
                                  r1 = top[118];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_237;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_237;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object *)lenv)[1];
                          goto fatbar_237;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      goto fatbar_237;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_237;
                }
                PXLL_RETURN(0);
                goto fatbar_237_over;
                fatbar_237:
                r0 = (object *) constructed_599[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[132];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][3];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L965; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L965:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = (object*)UITAG(5);
                  r0[2] = r1;
                  r1 = ((object*) lenv) [2];
                  r0[3] = r1;
                  r1 = top[118];
                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_226;
                }
                PXLL_RETURN(0);
                fatbar_237_over: ;
                PXLL_RETURN(0);
                fatbar_236_over: ;
                PXLL_RETURN(0);
                fatbar_235_over: ;
                PXLL_RETURN(0);
                fatbar_234_over: ;
                PXLL_RETURN(0);
                fatbar_233_over: ;
                PXLL_RETURN(0);
                fatbar_232_over: ;
                PXLL_RETURN(0);
                fatbar_231_over: ;
                PXLL_RETURN(0);
                fatbar_230_over: ;
                PXLL_RETURN(0);
                fatbar_229_over: ;
                PXLL_RETURN(0);
                fatbar_228_over: ;
                PXLL_RETURN(0);
                fatbar_227_over: ;
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_226;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_226_over;
            fatbar_226:
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object***) lenv) [1][1][2];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                lenv = lenv[1];
                r0[2] = r1;
                r1 = (object *) 1;
                r2 = ((object*) lenv) [2];
                range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
                r1 = (object *) 1;
                r2 = ((object*) lenv) [2];
                r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
                r0[3] = r1;
                r0 = (object *) 9474;
                r1 = ((object*) lenv) [3];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = ((object**) lenv) [1][3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 4);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object***) lenv) [1][1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = top[132];
                      r1[2] = r2;
                      r2 = ((object*) lenv) [3];
                      r1[3] = r2;
                      r2 = top[32];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L966; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                      L966:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[4] = r1;
                      r0 = allocate (TC_TUPLE, 3);
                      r1 = ((object*) lenv) [2];
                      r2 = ((object**) lenv) [1][2];
                      t = alloc_no_clear (UOTAG(12),2);
                      t[1] = r2;
                      t[2] = r1;
                      r1 = t;
                      r0[2] = r1;
                      r1 = ((object*) lenv) [4];
                      r0[3] = r1;
                      r1 = top[118];
                      r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                    } break;
                    default: {
                      goto fatbar_239;
                    }
                  }
                  PXLL_RETURN(0);
                  goto fatbar_239_over;
                  fatbar_239:
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = (object*) &constructed_871;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][2];
                  r0[3] = r1;
                  r1 = top[15];
                  r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                  fatbar_239_over: ;
                  PXLL_RETURN(0);
                } else {
                  r0 = allocate (TC_TUPLE, 4);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object***) lenv) [1][1][2];
                  r1[2] = r2;
                  r2 = top[132];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L967; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                  L967:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[132];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][3];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L968; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L968:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = ((object*) lenv) [3];
                  r2 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  r0[4] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = (object*)UITAG(8);
                  r0[2] = r1;
                  r1 = ((object*) lenv) [4];
                  r0[3] = r1;
                  r1 = top[118];
                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                }
                PXLL_RETURN(0);
              } break;
              case (UOTAG(9)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = (object *) constructed_49[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 9);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][3];
                  t = alloc_no_clear (UOTAG(1),1);
                  t[1] = r1;
                  r1 = t;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object***) lenv) [1][1][3];
                  r1[2] = r2;
                  r2 = top[22];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L969; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                  L969:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = ((object*) lenv) [3];
                  t = alloc_no_clear (UOTAG(5),1);
                  t[1] = r1;
                  r1 = t;
                  r0[4] = r1;
                  r1 = (object*)TC_NIL;
                  r0[5] = r1;
                  r1 = ((object*) lenv) [5];
                  r2 = ((object*) lenv) [4];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  r0[6] = r1;
                  r1 = ((object*) lenv) [6];
                  r2 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  r0[7] = r1;
                  r1 = ((object*) lenv) [7];
                  t = alloc_no_clear (UOTAG(0),1);
                  t[1] = r1;
                  r1 = t;
                  r0[8] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[132];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][3];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L970; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L970:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[9] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = ((object*) lenv) [8];
                  r2 = (object *) constructed_496[0];
                  t = alloc_no_clear (UOTAG(12),2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  r0[2] = r1;
                  r1 = ((object*) lenv) [9];
                  r0[3] = r1;
                  r1 = top[118];
                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                } else {
                  r0 = allocate (TC_TUPLE, 4);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object***) lenv) [1][1][2];
                  r1[2] = r2;
                  r2 = top[132];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L971; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
                  L971:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[132];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][3];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L972; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L972:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = ((object*) lenv) [3];
                  r2 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  r0[4] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = (object*)UITAG(8);
                  r0[2] = r1;
                  r1 = ((object*) lenv) [4];
                  r0[3] = r1;
                  r1 = top[118];
                  r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
                }
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_238;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_238_over;
            fatbar_238:
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = top[132];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L973; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
            L973:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = top[132];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][3];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L974; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L974:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = ((object*) lenv) [3];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[4] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*)UITAG(8);
            r0[2] = r1;
            r1 = ((object*) lenv) [4];
            r0[3] = r1;
            r1 = top[118];
            r0[1] = r1[2]; lenv = r0; goto FUN_make_node_486;
            fatbar_238_over: ;
            PXLL_RETURN(0);
            fatbar_226_over: ;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_225;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_225_over;
        fatbar_225:
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_872;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        fatbar_225_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_224_over;
    fatbar_224:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_873;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_224_over: ;
    PXLL_RETURN(0);
  L927:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_walk_513; r1[2] = lenv;
  r0[132] = r1;
  // def build-list-literal_514
  goto L975;
  FUN_build_list_literal_514:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(9)): {
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = top[134];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][3];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L976; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L976:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object*) lenv) [4];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(5),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_223;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_223_over;
        fatbar_223:
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = top[134];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L977; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_build_literal_515;
        L977:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object**) lenv) [1][3];
        r1[2] = r2;
        r2 = top[133];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L978; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_build_list_literal_514;
        L978:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = (object*)TC_NIL;
        r0[4] = r1;
        r1 = ((object*) lenv) [4];
        r2 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[6] = r1;
        r0 = (object *) constructed_61[0];
        r1 = ((object*) lenv) [6];
        r2 = (object *) constructed_51[0];
        t = alloc_no_clear (UOTAG(5),3);
        t[1] = r2;
        t[2] = r0;
        t[3] = r1;
        r0 = t;
        PXLL_RETURN(0);
        fatbar_223_over: ;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object *) constructed_49[0];
        r1 = (object *) constructed_859[0];
        r2 = (object *) constructed_51[0];
        t = alloc_no_clear (UOTAG(5),3);
        t[1] = r2;
        t[2] = r0;
        t[3] = r1;
        r0 = t;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L975:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_build_list_literal_514; r1[2] = lenv;
  r0[133] = r1;
  // def build-literal_515
  goto L979;
  FUN_build_literal_515:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(2),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UITAG(6)): {
        r0 = (object*)UITAG(3);
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(4),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[133];
        r0[1] = r1[2]; lenv = r0; goto FUN_build_list_literal_514;
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = top[134];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L980; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L980:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(6),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_222;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_222_over;
    fatbar_222:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_858;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_222_over: ;
    PXLL_RETURN(0);
  L979:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_build_literal_515; r1[2] = lenv;
  r0[134] = r1;
  // def make-vardef_517
  goto L981;
  FUN_make_vardef_517:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 2);
    r3 = allocate (TC_TUPLE, 6);
    r3[1] = lenv; lenv = r3;
    r4 = ((object***) lenv) [1][1][2];
    r4 = allocate (TC_TUPLE, 2);
    r4[1] = lenv; lenv = r4;
    r5 = ((object****) lenv) [1][1][1][2];
    r5 = UOBJ_GET(r5,0);
    r4[2] = r5;
    r4 = ((object*) lenv) [2];
    lenv = lenv[1];
    r3[2] = r4;
    r4 = allocate (TC_TUPLE, 2);
    r5 = ((object***) lenv) [1][1][3];
    r4[2] = r5;
    r5 = top[56];
    t = allocate (TC_SAVE, 7);
    t[1] = k; t[2] = lenv; t[3] = &&L982; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
    r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
    L982:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
    r4 = result;
    r3[3] = r4;
    r4 = (object*)TC_NIL;
    r3[4] = r4;
    r4 = ((object*) lenv) [4];
    r5 = ((object*) lenv) [3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r5;
    t[2] = r4;
    r4 = t;
    r3[5] = r4;
    r4 = ((object*) lenv) [5];
    r5 = (object*) &constructed_857;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r5;
    t[2] = r4;
    r4 = t;
    r3[6] = r4;
    r3 = ((object*) lenv) [6];
    r4 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    lenv = lenv[1];
    r2[2] = r3;
    r3 = top[47];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L983; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
    L983:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = top[80];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L984; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
    L984:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (UOTAG(19), 5);
    r1 = (object *) constructed_856[0];
    r0[1] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = (object *) constructed_855[0];
    r0[4] = r1;
    r1 = ((object**) lenv) [1][3];
    r0[5] = r1;
    PXLL_RETURN(0);
  L981:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_make_vardef_517; r1[2] = lenv;
  r0[135] = r1;
  // def k/free_522
  goto L985;
  FUN_k_2ffree_522:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_854;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_854;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L985:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_k_2ffree_522; r1[2] = lenv;
  r0[136] = r1;
  // def k/target_523
  goto L986;
  FUN_k_2ftarget_523:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_853;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_853;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L986:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_k_2ftarget_523; r1[2] = lenv;
  r0[137] = r1;
  // def print-insn_526
  goto L987;
  FUN_print_insn_526:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    // def mprint-insn_3093
    goto L988;
    FUN_mprint_insn_3093:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r0[3] = r1;
          r1 = top[138];
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
        } break;
        default: {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L988:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_mprint_insn_3093; r1[2] = lenv;
    r0[2] = r1;
    // def print-line_3094
    goto L989;
    FUN_print_line_3094:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r0 = top[158];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L990; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L990:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r1 = top[92];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L991; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_indent_434;
          L991:
          ; lenv = k[2]; k = k[1];
          r0 = (object *) -1;
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
            r0 = (object*) &constructed_851;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_851;
            r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          } else {
            r0 = ((object*) lenv) [2];
            r0 = dump_object (r0, 0);
          }
          r0 = (object*) &constructed_850;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_850;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object**) lenv) [1][2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L992; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L992:
          ; lenv = k[2]; k = k[1];
          r0 = ((object****) lenv) [1][1][1][3];
          r1 = ((object*) lenv) [4];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_print_insn_526;
        } break;
        default: {
          r0 = top[158];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L993; ; k = t;
          lenv = r0[2]; goto *r0[1];
          L993:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r1 = ((object***) lenv) [1][1][3];
          r0[2] = r1;
          r1 = top[92];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L994; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_indent_434;
          L994:
          ; lenv = k[2]; k = k[1];
          r0 = (object*) &constructed_852;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_852;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          lenv = r0[2]; goto *r0[1];
        } break;
      }
      PXLL_RETURN(0);
    L989:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_print_line_3094; r1[2] = lenv;
    r0[3] = r1;
    r0 = ((object**) lenv) [1][2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = top[158];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L995; ; k = t;
        lenv = r0[2]; goto *r0[1];
        L995:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object***) lenv) [1][1][3];
        r0[2] = r1;
        r1 = top[92];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L996; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_indent_434;
        L996:
        ; lenv = k[2]; k = k[1];
        r0 = (object*) &constructed_825;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_825;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_824;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_824;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = ((object*) lenv) [2];
        r0 = dump_object (r0, 0);
        PXLL_RETURN(0);
      } break;
      case (UOTAG(14)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19499
        goto L997;
        FUN_lambda_19499:
          r0 = (object*) &constructed_827;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_827;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L997:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19499; r1[2] = lenv;
        r0[2] = r1;
        r1 = (object*)UITAG(1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(15)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19529
        goto L998;
        FUN_lambda_19529:
          r0 = (object*) &constructed_828;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_828;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L998:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19529; r1[2] = lenv;
        r0[2] = r1;
        r1 = (object*)UITAG(1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19553
        goto L999;
        FUN_lambda_19553:
          check_heap (0);
          r0 = (object*) &constructed_829;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_829;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = top[125];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1000; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_literal__3estring_505;
          L1000:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L999:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19553; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19578
        goto L1001;
        FUN_lambda_19578:
          r0 = (object*) &constructed_830;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_830;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1001:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19578; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,4);
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19618
        goto L1002;
        FUN_lambda_19618:
          check_heap (0);
          r0 = (object*) &constructed_831;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_831;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = top[107];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1003; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_repr_462;
          L1003:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = top[107];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1004; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_repr_462;
          L1004:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [5];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1002:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19618; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [6];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19659
        goto L1005;
        FUN_lambda_19659:
          check_heap (0);
          r0 = (object*) &constructed_832;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_832;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [3];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1006; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
          L1006:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [4];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
        L1005:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19659; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [5];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(6)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19681
        goto L1007;
        FUN_lambda_19681:
          r0 = (object*) &constructed_833;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_833;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          r0 = dump_object (r0, 0);
          PXLL_RETURN(0);
        L1007:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19681; r1[2] = lenv;
        r0[2] = r1;
        r1 = (object*)UITAG(1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19711
        goto L1008;
        FUN_lambda_19711:
          check_heap (0);
          r0 = (object*) &constructed_834;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_834;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [3];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
        L1008:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19711; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(8)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19737
        goto L1009;
        FUN_lambda_19737:
          r0 = (object*) &constructed_835;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_835;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1009:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19737; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19768
        goto L1010;
        FUN_lambda_19768:
          r0 = (object*) &constructed_836;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_836;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1010:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19768; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [5];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(12)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,4);
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19805
        goto L1011;
        FUN_lambda_19805:
          r0 = (object*) &constructed_837;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_837;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [5];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1011:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19805; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [6];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(13)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19838
        goto L1012;
        FUN_lambda_19838:
          r0 = (object*) &constructed_838;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_838;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1012:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19838; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [5];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19860
        goto L1013;
        FUN_lambda_19860:
          r0 = (object*) &constructed_839;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_839;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1013:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19860; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(11)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19885
        goto L1014;
        FUN_lambda_19885:
          r0 = (object*) &constructed_840;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_840;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1014:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19885; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(16)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19906
        goto L1015;
        FUN_lambda_19906:
          r0 = (object*) &constructed_841;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_841;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1015:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19906; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(17)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19926
        goto L1016;
        FUN_lambda_19926:
          r0 = (object*) &constructed_842;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_842;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1016:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19926; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(18)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,4);
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19965
        goto L1017;
        FUN_lambda_19965:
          check_heap (0);
          r0 = (object*) &constructed_843;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_843;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = top[91];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1018; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_repr_433;
          L1018:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][4];
          r1[2] = r2;
          r2 = top[107];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1019; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_repr_462;
          L1019:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
          r0 = ((object*) lenv) [5];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1017:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19965; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [6];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(19)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_19993
        goto L1020;
        FUN_lambda_19993:
          r0 = (object*) &constructed_844;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_844;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1020:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_19993; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(20)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_20032
        goto L1021;
        FUN_lambda_20032:
          check_heap (0);
          r0 = (object*) &constructed_845;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_845;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [3];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1022; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
          L1022:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [4];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
        L1021:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_20032; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [5];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(21)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_20057
        goto L1023;
        FUN_lambda_20057:
          r0 = (object*) &constructed_846;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_846;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          PXLL_RETURN(0);
        L1023:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_20057; r1[2] = lenv;
        r0[2] = r1;
        r1 = (object*)UITAG(1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 7);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,4);
        r0[6] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,5);
        r0[7] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_20108
        goto L1024;
        FUN_lambda_20108:
          check_heap (0);
          r0 = (object*) &constructed_847;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_847;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = top[107];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1025; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_repr_462;
          L1025:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = ((object*) lenv) [2];
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          lenv = lenv[1];
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [5];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1026; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
          L1026:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [6];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = top[138];
          r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
        L1024:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_20108; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [7];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      case (UOTAG(22)): {
        r0 = allocate (TC_TUPLE, 7);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,4);
        r0[6] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,5);
        r0[7] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_20165
        goto L1027;
        FUN_lambda_20165:
          check_heap (0);
          r0 = (object*) &constructed_848;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_848;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_20153
          goto L1028;
          FUN_lambda_20153:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = (object *) 3;
            r2 = ((object****) lenv) [1][1][1][3];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            r0[3] = r1;
            r1 = top[138];
            r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
          L1028:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_20153; r1[2] = lenv;
          r0[2] = r1;
          r1 = ((object*) lenv) [5];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1029; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L1029:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [6];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_mprint_insn_3093;
        L1027:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_20165; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [7];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 7);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,3);
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,4);
        r0[6] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = UOBJ_GET(r1,5);
        r0[7] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_20222
        goto L1030;
        FUN_lambda_20222:
          check_heap (0);
          r0 = (object*) &constructed_849;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_849;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_824;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_824;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [2];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [3];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = ((object*) lenv) [4];
          dump_object (r0, 0);
          r0 = (object*) &constructed_826;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_826;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_20210
          goto L1031;
          FUN_lambda_20210:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = (object *) 3;
            r2 = ((object****) lenv) [1][1][1][3];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            r0[3] = r1;
            r1 = top[138];
            r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
          L1031:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_20210; r1[2] = lenv;
          r0[2] = r1;
          r1 = ((object*) lenv) [5];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1032; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L1032:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [6];
          r0[2] = r1;
          r1 = (object *) 3;
          r2 = ((object***) lenv) [1][1][3];
          r1 = box((pxll_int)unbox(r2)+unbox(r1));
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_mprint_insn_3093;
        L1030:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_20222; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [7];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_print_line_3094;
      } break;
    }
    PXLL_RETURN(0);
  L987:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_print_insn_526; r1[2] = lenv;
  r0[138] = r1;
  // def unify_530
  goto L1033;
  FUN_unify_530:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    // def type-error_3254
    goto L1034;
    FUN_type_error_3254:
      check_heap (0);
      r0 = top[158];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1035; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L1035:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object***) lenv) [1][1][2];
      r0[2] = r1;
      r1 = top[127];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1036; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_pp_node_508;
      L1036:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[140];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1037; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_apply_subst_531;
      L1037:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][3];
      r1[2] = r2;
      r2 = top[140];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1038; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_apply_subst_531;
      L1038:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 8);
      r2[1] = lenv; lenv = r2;
      r3 = allocate (TC_TUPLE, 2);
      r4 = ((object***) lenv) [1][1][2];
      r3[2] = r4;
      r4 = top[107];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L1039; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_type_repr_462;
      L1039:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = allocate (TC_TUPLE, 2);
      r4 = ((object***) lenv) [1][1][3];
      r3[2] = r4;
      r4 = top[107];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L1040; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_type_repr_462;
      L1040:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[3] = r3;
      r3 = (object*)TC_NIL;
      r2[4] = r3;
      r3 = ((object*) lenv) [4];
      r4 = (object*) &constructed_823;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[5] = r3;
      r3 = ((object*) lenv) [5];
      r4 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[6] = r3;
      r3 = ((object*) lenv) [6];
      r4 = (object*) &constructed_822;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[7] = r3;
      r3 = ((object*) lenv) [7];
      r4 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[8] = r3;
      r2 = ((object*) lenv) [8];
      r3 = (object*) &constructed_821;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1041; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L1041:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      lenv = lenv[1];
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_820;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_820;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
    L1034:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_type_error_3254; r1[2] = lenv;
    r0[2] = r1;
    // def U_3255
    goto L1042;
    FUN__55_3255:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 5);
      r0[1] = lenv; lenv = r0;
      r1 = k; // %getcc
      r0[2] = r1;
      // def lambda_20685
      goto L1043;
      FUN_lambda_20685:
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        k = r1; // %putcc
        PXLL_RETURN(0);
      L1043:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_20685; r1[2] = lenv;
      r0[3] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[109];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1044; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_type_find_465;
      L1044:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[4] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][3];
      r1[2] = r2;
      r2 = top[109];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1045; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_type_find_465;
      L1045:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[5] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][5];
      r2 = ((object**) lenv) [1][4];
      r1 = PXLL_TEST(r2==r1);
      r0[2] = r1;
      r0 = (object *) 6;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(r1==r0);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = (object *) 14;
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object**) lenv) [1][5];
            switch (get_case (r0)) {
              case (UOTAG(0)): {
                r0 = (object *) 14;
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][5];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][5];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = (object *) constructed_817[0];
                r1 = ((object**) lenv) [1][2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = (object *) constructed_817[0];
                  r1 = ((object*) lenv) [2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = (object *) 14;
                  } else {
                    goto fatbar_216;
                  }
                  goto fatbar_216_over;
                  fatbar_216:
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object**) lenv) [1][3];
                  r2[2] = r3;
                  r3 = top[19];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L1046; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                  L1046:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][5];
                  r1[3] = r2;
                  r2 = ((object*****) lenv) [1][1][1][1][3];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1047; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__55_3255;
                  L1047:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1048; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1048:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  fatbar_216_over: ;
                } else {
                  goto fatbar_215;
                }
                goto fatbar_215_over;
                fatbar_215:
                r0 = (object *) constructed_817[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object*) lenv) [3];
                  r2[2] = r3;
                  r3 = top[19];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L1049; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                  L1049:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][4];
                  r1[3] = r2;
                  r2 = ((object*****) lenv) [1][1][1][1][3];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1050; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__55_3255;
                  L1050:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1051; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1051:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } else {
                  goto fatbar_217;
                }
                goto fatbar_217_over;
                fatbar_217:
                r0 = (object *) constructed_819[0];
                r1 = ((object**) lenv) [1][2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object***) lenv) [1][1][4];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][5];
                  r1[3] = r2;
                  r2 = ((object*****) lenv) [1][1][1][1][4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1052; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__55_row_3257;
                  L1052:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1053; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1053:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } else {
                  goto fatbar_218;
                }
                goto fatbar_218_over;
                fatbar_218:
                r0 = (object *) constructed_819[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object***) lenv) [1][1][5];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][4];
                  r1[3] = r2;
                  r2 = ((object*****) lenv) [1][1][1][1][4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1054; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__55_row_3257;
                  L1054:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1055; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1055:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } else {
                  goto fatbar_219;
                }
                goto fatbar_219_over;
                fatbar_219:
                r0 = (object *) constructed_557[0];
                r1 = ((object**) lenv) [1][2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object***) lenv) [1][1][4];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][5];
                  r1[3] = r2;
                  r2 = ((object*****) lenv) [1][1][1][1][4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1056; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__55_row_3257;
                  L1056:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1057; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1057:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } else {
                  goto fatbar_220;
                }
                goto fatbar_220_over;
                fatbar_220:
                r0 = (object *) constructed_557[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object***) lenv) [1][1][5];
                  r1[2] = r2;
                  r2 = ((object***) lenv) [1][1][4];
                  r1[3] = r2;
                  r2 = ((object*****) lenv) [1][1][1][1][4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1058; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__55_row_3257;
                  L1058:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object***) lenv) [1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1059; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1059:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } else {
                  goto fatbar_221;
                }
                goto fatbar_221_over;
                fatbar_221:
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r2 = ((object***) lenv) [1][1][2];
                r1 = PXLL_TEST(r2==r1);
                r0[2] = r1;
                r0 = (object *) 6;
                r1 = ((object*) lenv) [2];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = (object *) 262;
                } else {
                  r0 = allocate (TC_TUPLE, 4);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object***) lenv) [1][1][3];
                  r1[2] = r2;
                  r2 = top[22];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1060; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                  L1060:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object**) lenv) [1][3];
                  r1[2] = r2;
                  r2 = top[22];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1061; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                  L1061:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = ((object*) lenv) [3];
                  r2 = ((object*) lenv) [2];
                  r1 = PXLL_TEST(unbox(r2)==unbox(r1));
                  r0[4] = r1;
                  r0 = (object *) 6;
                  r1 = ((object*) lenv) [4];
                  r0 = PXLL_TEST(r1==r0);
                  lenv = lenv[1];
                }
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = ((object****) lenv) [1][1][1][2];
                  r0[2] = r1;
                  r1 = ((object****) lenv) [1][1][1][3];
                  r0[3] = r1;
                  r1 = ((object*****) lenv) [1][1][1][1][2];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1062; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_type_error_3254;
                  L1062:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } else {
                  r0 = (object *) 14;
                }
                fatbar_221_over: ;
                fatbar_220_over: ;
                fatbar_219_over: ;
                fatbar_218_over: ;
                fatbar_217_over: ;
                fatbar_215_over: ;
                lenv = lenv[1];
              } break;
            }
            lenv = lenv[1];
          } break;
        }
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [4];
        r0[2] = r1;
        r1 = ((object*) lenv) [5];
        r0[3] = r1;
        r1 = top[110];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1063; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_type_union_466;
        L1063:
        ; lenv = k[2]; k = k[1];
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,1);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][5];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][5];
                r1 = UOBJ_GET(r1,1);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 4);
                r1 = ((object*****) lenv) [1][1][1][1][3];
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r0[3] = r1;
                r1 = ((object*) lenv) [2];
                r0[4] = r1;
                r1 = top[36];
                r0[1] = r1[2]; lenv = r0; goto FUN_for_each2_330;
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_214;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_214;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_214_over;
        fatbar_214:
        r0 = (object *) 14;
        PXLL_RETURN(0);
        fatbar_214_over: ;
        PXLL_RETURN(0);
      } else {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L1042:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN__55_3255; r1[2] = lenv;
    r0[3] = r1;
    // def U-row_3257
    goto L1064;
    FUN__55_row_3257:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = (object *) constructed_557[0];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = ((object***) lenv) [1][1][3];
                    switch (get_case (r0)) {
                      case (UOTAG(1)): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object****) lenv) [1][1][1][3];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object****) lenv) [1][1][1][3];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = (object *) constructed_557[0];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object***) lenv) [1][1][2];
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[3] = r1;
                                  r1 = ((object******) lenv) [1][1][1][1][1][3];
                                  r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                } break;
                                default: {
                                  lenv = ((object *)lenv)[1];
                                  goto fatbar_211;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              goto fatbar_211;
                            }
                          }
                          PXLL_RETURN(0);
                        } else {
                          goto fatbar_211;
                        }
                        PXLL_RETURN(0);
                        goto fatbar_211_over;
                        fatbar_211:
                        r0 = allocate (TC_TUPLE, 4);
                        r0[1] = lenv; lenv = r0;
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = ((object**) lenv) [1][3];
                        r1[2] = r2;
                        r2 = top[22];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L1065; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                        L1065:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        r0[3] = r1;
                        r1 = allocate (TC_TUPLE, 2);
                        r1[1] = lenv; lenv = r1;
                        // def $loop_3291
                        goto L1066;
                        FUN__24loop_3291:
                          check_heap (0);
                          r0 = ((object***) lenv) [1][1][3];
                          r1 = ((object*) lenv) [2];
                          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                            r0 = allocate (TC_TUPLE, 3);
                            r1 = ((object*) lenv) [3];
                            r0[2] = r1;
                            r1 = (object *) constructed_36[0];
                            r0[3] = r1;
                            r1 = top[21];
                            r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                          } else {
                            r0 = allocate (TC_TUPLE, 2);
                            r0[1] = lenv; lenv = r0;
                            r1 = top[105];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L1067; t[4] = r0; k = t;
                            lenv = r1[2]; goto FUN_new_tvar_453;
                            L1067:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r0 = ((object**) lenv) [1][3];
                            r1 = ((object*) lenv) [2];
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r1;
                            t[2] = r0;
                            r0 = t;
                            lenv = lenv[1];
                            r1 = (object *) 3;
                            r2 = ((object*) lenv) [2];
                            r1 = box((pxll_int)unbox(r2)+unbox(r1));
                            lenv[2] = r1;
                            lenv[3] = r0;
                            goto FUN__24loop_3291;
                          }
                          PXLL_RETURN(0);
                        L1066:
                        r2 = allocate (TC_CLOSURE, 2);
                        r2[1] = &&FUN__24loop_3291; r2[2] = lenv;
                        r1[2] = r2;
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = (object *) 1;
                        r1[2] = r2;
                        r2 = (object*)TC_NIL;
                        r1[3] = r2;
                        r2 = ((object*) lenv) [2];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L1068; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN__24loop_3291;
                        L1068:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        lenv = lenv[1];
                        r0[4] = r1;
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = ((object***) lenv) [1][1][2];
                        r0[2] = r1;
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = ((object**) lenv) [1][2];
                        r1[2] = r2;
                        r2 = ((object*) lenv) [4];
                        r1[3] = r2;
                        r2 = top[104];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L1069; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                        L1069:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[3] = r1;
                        r1 = ((object******) lenv) [1][1][1][1][1][3];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1070; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                        L1070:
                        ; lenv = k[2]; k = k[1];
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r0[2] = r1;
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        // def $loop_3296
                        goto L1071;
                        FUN__24loop_3296:
                          check_heap (0);
                          r0 = ((object***) lenv) [1][1][2];
                          r1 = ((object*) lenv) [2];
                          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                            r0 = (object *) 14;
                            PXLL_RETURN(0);
                          } else {
                            r0 = allocate (TC_TUPLE, 3);
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = ((object*****) lenv) [1][1][1][1][3];
                            r1[2] = r2;
                            r2 = ((object*) lenv) [2];
                            r1[3] = r2;
                            r2 = top[27];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L1072; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                            L1072:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r1 = allocate (TC_TUPLE, 2);
                            r1[1] = lenv; lenv = r1;
                            r2 = allocate (TC_TUPLE, 3);
                            r3 = ((object*****) lenv) [1][1][1][1][4];
                            r2[2] = r3;
                            r3 = ((object**) lenv) [1][2];
                            r2[3] = r3;
                            r3 = top[27];
                            t = allocate (TC_SAVE, 5);
                            t[1] = k; t[2] = lenv; t[3] = &&L1073; t[4] = r0; t[5] = r1; k = t;
                            r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                            L1073:
                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                            r2 = result;
                            r1[2] = r2;
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = (object *) constructed_557[0];
                            r1[2] = r2;
                            r2 = allocate (TC_TUPLE, 2);
                            r2[1] = lenv; lenv = r2;
                            r3 = (object*)TC_NIL;
                            r2[2] = r3;
                            r2 = ((object*) lenv) [2];
                            r3 = ((object**) lenv) [1][2];
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r3;
                            t[2] = r2;
                            r2 = t;
                            lenv = lenv[1];
                            r1[3] = r2;
                            r2 = top[104];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L1074; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                            L1074:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            lenv = lenv[1];
                            r0[3] = r1;
                            r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                            t = allocate (TC_SAVE, 3);
                            t[1] = k; t[2] = lenv; t[3] = &&L1075; ; k = t;
                            r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                            L1075:
                            ; lenv = k[2]; k = k[1];
                            r0 = (object *) 3;
                            r1 = ((object*) lenv) [2];
                            r0 = box((pxll_int)unbox(r1)+unbox(r0));
                            lenv[2] = r0;
                            goto FUN__24loop_3296;
                          }
                          PXLL_RETURN(0);
                        L1071:
                        r1 = allocate (TC_CLOSURE, 2);
                        r1[1] = &&FUN__24loop_3296; r1[2] = lenv;
                        r0[2] = r1;
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = (object *) 1;
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3296;
                        fatbar_211_over: ;
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object *)lenv)[1];
                        goto fatbar_210;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object *)lenv)[1];
                    goto fatbar_210;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_210;
              }
            }
            PXLL_RETURN(0);
          } else {
            goto fatbar_210;
          }
          PXLL_RETURN(0);
          goto fatbar_210_over;
          fatbar_210:
          r0 = (object *) constructed_819[0];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    switch (get_case (r0)) {
                      case (TC_PAIR): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][3];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][3];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = ((object*) lenv) [3];
                        switch (get_case (r0)) {
                          case (TC_NIL): {
                            r0 = ((object*****) lenv) [1][1][1][1][3];
                            switch (get_case (r0)) {
                              case (UOTAG(1)): {
                                r0 = allocate (TC_TUPLE, 3);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object******) lenv) [1][1][1][1][1][3];
                                r1 = UOBJ_GET(r1,0);
                                r0[2] = r1;
                                r1 = ((object******) lenv) [1][1][1][1][1][3];
                                r1 = UOBJ_GET(r1,1);
                                r0[3] = r1;
                                r0 = (object *) constructed_819[0];
                                r1 = ((object*) lenv) [2];
                                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                  r0 = ((object*) lenv) [3];
                                  switch (get_case (r0)) {
                                    case (TC_PAIR): {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][3];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = ((object**) lenv) [1][3];
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r0 = ((object*) lenv) [3];
                                      switch (get_case (r0)) {
                                        case (TC_PAIR): {
                                          r0 = allocate (TC_TUPLE, 3);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object**) lenv) [1][3];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r1 = ((object**) lenv) [1][3];
                                          r1 = UOBJ_GET(r1,1);
                                          r0[3] = r1;
                                          r0 = ((object*) lenv) [3];
                                          switch (get_case (r0)) {
                                            case (TC_PAIR): {
                                              r0 = allocate (TC_TUPLE, 3);
                                              r0[1] = lenv; lenv = r0;
                                              r1 = ((object**) lenv) [1][3];
                                              r1 = UOBJ_GET(r1,0);
                                              r0[2] = r1;
                                              r1 = ((object**) lenv) [1][3];
                                              r1 = UOBJ_GET(r1,1);
                                              r0[3] = r1;
                                              r0 = ((object*) lenv) [3];
                                              switch (get_case (r0)) {
                                                case (TC_NIL): {
                                                  r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                                                  switch (get_case (r0)) {
                                                    case (UOTAG(1)): {
                                                      r0 = allocate (TC_TUPLE, 2);
                                                      r0[1] = lenv; lenv = r0;
                                                      r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                                      r1 = UOBJ_GET(r1,0);
                                                      r0[2] = r1;
                                                      r0 = ((object****) lenv) [1][1][1][2];
                                                      switch (get_case (r0)) {
                                                        case (UOTAG(1)): {
                                                          r0 = allocate (TC_TUPLE, 2);
                                                          r0[1] = lenv; lenv = r0;
                                                          r1 = ((object*****) lenv) [1][1][1][1][2];
                                                          r1 = UOBJ_GET(r1,0);
                                                          r0[2] = r1;
                                                          r0 = ((object*) lenv) [2];
                                                          r1 = ((object**) lenv) [1][2];
                                                          r0 = PXLL_TEST(r1==r0);
                                                          lenv = lenv[1];
                                                        } break;
                                                        default: {
                                                          lenv = ((object *)lenv)[1];
                                                          goto fatbar_213;
                                                        }
                                                      }
                                                      lenv = lenv[1];
                                                    } break;
                                                    default: {
                                                      goto fatbar_213;
                                                    }
                                                  }
                                                  goto fatbar_213_over;
                                                  fatbar_213:
                                                  r0 = (object *) 6;
                                                  fatbar_213_over: ;
                                                  if PXLL_IS_TRUE(r0) {
                                                    r0 = allocate (TC_TUPLE, 3);
                                                    r1 = ((object******) lenv) [1][1][1][1][1][2];
                                                    r0[2] = r1;
                                                    r1 = ((object**) lenv) [1][2];
                                                    r0[3] = r1;
                                                    r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                                                    t = allocate (TC_SAVE, 3);
                                                    t[1] = k; t[2] = lenv; t[3] = &&L1076; ; k = t;
                                                    r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                                    L1076:
                                                    ; lenv = k[2]; k = k[1];
                                                    r0 = allocate (TC_TUPLE, 3);
                                                    r1 = ((object*****) lenv) [1][1][1][1][2];
                                                    r0[2] = r1;
                                                    r1 = ((object*) lenv) [2];
                                                    r0[3] = r1;
                                                    r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                                                    r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                                  } else {
                                                    r0 = allocate (TC_TUPLE, 2);
                                                    r0[1] = lenv; lenv = r0;
                                                    r1 = top[105];
                                                    t = allocate (TC_SAVE, 4);
                                                    t[1] = k; t[2] = lenv; t[3] = &&L1077; t[4] = r0; k = t;
                                                    lenv = r1[2]; goto FUN_new_tvar_453;
                                                    L1077:
                                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                                    r1 = result;
                                                    r0[2] = r1;
                                                    r0 = allocate (TC_TUPLE, 3);
                                                    r1 = ((object******) lenv) [1][1][1][1][1][2];
                                                    r0[2] = r1;
                                                    r1 = allocate (TC_TUPLE, 4);
                                                    r2 = ((object****) lenv) [1][1][1][2];
                                                    r1[2] = r2;
                                                    r2 = ((object***) lenv) [1][1][2];
                                                    r1[3] = r2;
                                                    r2 = ((object*) lenv) [2];
                                                    r1[4] = r2;
                                                    r2 = top[106];
                                                    t = allocate (TC_SAVE, 4);
                                                    t[1] = k; t[2] = lenv; t[3] = &&L1078; t[4] = r0; k = t;
                                                    r1[1] = r2[2]; lenv = r1; goto FUN_rlabel_458;
                                                    L1078:
                                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                                    r1 = result;
                                                    r0[3] = r1;
                                                    r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                                                    t = allocate (TC_SAVE, 3);
                                                    t[1] = k; t[2] = lenv; t[3] = &&L1079; ; k = t;
                                                    r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                                    L1079:
                                                    ; lenv = k[2]; k = k[1];
                                                    r0 = allocate (TC_TUPLE, 3);
                                                    r1 = ((object**) lenv) [1][2];
                                                    r0[2] = r1;
                                                    r1 = allocate (TC_TUPLE, 4);
                                                    r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                                    r1[2] = r2;
                                                    r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
                                                    r1[3] = r2;
                                                    r2 = ((object*) lenv) [2];
                                                    r1[4] = r2;
                                                    r2 = top[106];
                                                    t = allocate (TC_SAVE, 4);
                                                    t[1] = k; t[2] = lenv; t[3] = &&L1080; t[4] = r0; k = t;
                                                    r1[1] = r2[2]; lenv = r1; goto FUN_rlabel_458;
                                                    L1080:
                                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                                    r1 = result;
                                                    r0[3] = r1;
                                                    r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                                                    r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                                  }
                                                  PXLL_RETURN(0);
                                                } break;
                                                default: {
                                                  lenv = ((object ***)lenv)[1][1][1];
                                                  goto fatbar_212;
                                                }
                                              }
                                              PXLL_RETURN(0);
                                            } break;
                                            default: {
                                              lenv = ((object **)lenv)[1][1];
                                              goto fatbar_212;
                                            }
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object *)lenv)[1];
                                          goto fatbar_212;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      goto fatbar_212;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } else {
                                  goto fatbar_212;
                                }
                                PXLL_RETURN(0);
                                goto fatbar_212_over;
                                fatbar_212:
                                r0 = (object *) constructed_557[0];
                                r1 = ((object*) lenv) [2];
                                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object*) lenv) [3];
                                  r1[2] = r2;
                                  r2 = top[19];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1081; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
                                  L1081:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = ((object***) lenv) [1][1][2];
                                  r0[3] = r1;
                                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1082; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                  L1082:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object******) lenv) [1][1][1][1][1][3];
                                  r0[2] = r1;
                                  r1 = ((object**) lenv) [1][2];
                                  r0[3] = r1;
                                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                  r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                } else {
                                  r0 = allocate (TC_TUPLE, 6);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object**) lenv) [1][3];
                                  r1[2] = r2;
                                  r2 = top[22];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1083; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                                  L1083:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[3] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  // def $loop_3315
                                  goto L1084;
                                  FUN__24loop_3315:
                                    check_heap (0);
                                    r0 = ((object***) lenv) [1][1][3];
                                    r1 = ((object*) lenv) [2];
                                    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r1 = ((object*) lenv) [3];
                                      r0[2] = r1;
                                      r1 = (object *) constructed_36[0];
                                      r0[3] = r1;
                                      r1 = top[21];
                                      r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                                    } else {
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = top[105];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1085; t[4] = r0; k = t;
                                      lenv = r1[2]; goto FUN_new_tvar_453;
                                      L1085:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[2] = r1;
                                      r0 = ((object**) lenv) [1][3];
                                      r1 = ((object*) lenv) [2];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r1;
                                      t[2] = r0;
                                      r0 = t;
                                      lenv = lenv[1];
                                      r1 = (object *) 3;
                                      r2 = ((object*) lenv) [2];
                                      r1 = box((pxll_int)unbox(r2)+unbox(r1));
                                      lenv[2] = r1;
                                      lenv[3] = r0;
                                      goto FUN__24loop_3315;
                                    }
                                    PXLL_RETURN(0);
                                  L1084:
                                  r2 = allocate (TC_CLOSURE, 2);
                                  r2[1] = &&FUN__24loop_3315; r2[2] = lenv;
                                  r1[2] = r2;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = (object *) 1;
                                  r1[2] = r2;
                                  r2 = (object*)TC_NIL;
                                  r1[3] = r2;
                                  r2 = ((object*) lenv) [2];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1086; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN__24loop_3315;
                                  L1086:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  lenv = lenv[1];
                                  r0[4] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[5] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  // def $loop_3320
                                  goto L1087;
                                  FUN__24loop_3320:
                                    check_heap (0);
                                    r0 = ((object***) lenv) [1][1][5];
                                    r1 = ((object*) lenv) [2];
                                    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r1 = ((object*) lenv) [3];
                                      r0[2] = r1;
                                      r1 = (object *) constructed_36[0];
                                      r0[3] = r1;
                                      r1 = top[21];
                                      r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                                    } else {
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = top[105];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1088; t[4] = r0; k = t;
                                      lenv = r1[2]; goto FUN_new_tvar_453;
                                      L1088:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[2] = r1;
                                      r0 = ((object**) lenv) [1][3];
                                      r1 = ((object*) lenv) [2];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r1;
                                      t[2] = r0;
                                      r0 = t;
                                      lenv = lenv[1];
                                      r1 = (object *) 3;
                                      r2 = ((object*) lenv) [2];
                                      r1 = box((pxll_int)unbox(r2)+unbox(r1));
                                      lenv[2] = r1;
                                      lenv[3] = r0;
                                      goto FUN__24loop_3320;
                                    }
                                    PXLL_RETURN(0);
                                  L1087:
                                  r2 = allocate (TC_CLOSURE, 2);
                                  r2[1] = &&FUN__24loop_3320; r2[2] = lenv;
                                  r1[2] = r2;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = (object *) 1;
                                  r1[2] = r2;
                                  r2 = (object*)TC_NIL;
                                  r1[3] = r2;
                                  r2 = ((object*) lenv) [2];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1089; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN__24loop_3320;
                                  L1089:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  lenv = lenv[1];
                                  r0[6] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r2 = ((object*) lenv) [4];
                                  r1[3] = r2;
                                  r2 = top[104];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1090; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                                  L1090:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = ((object****) lenv) [1][1][1][2];
                                  r0[3] = r1;
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1091; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                  L1091:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r2 = ((object*) lenv) [6];
                                  r1[3] = r2;
                                  r2 = top[104];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1092; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                                  L1092:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = ((object***) lenv) [1][1][2];
                                  r0[3] = r1;
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1093; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                  L1093:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][2];
                                  r0[2] = r1;
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  // def $loop_3325
                                  goto L1094;
                                  FUN__24loop_3325:
                                    check_heap (0);
                                    r0 = ((object***) lenv) [1][1][2];
                                    r1 = ((object*) lenv) [2];
                                    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                                      r0 = (object *) 14;
                                      PXLL_RETURN(0);
                                    } else {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = ((object*****) lenv) [1][1][1][1][3];
                                      r1[2] = r2;
                                      r2 = ((object*) lenv) [2];
                                      r1[3] = r2;
                                      r2 = top[27];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1095; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                                      L1095:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[2] = r1;
                                      r1 = allocate (TC_TUPLE, 4);
                                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                      r1[2] = r2;
                                      r2 = allocate (TC_TUPLE, 3);
                                      r3 = ((object****) lenv) [1][1][1][4];
                                      r2[2] = r3;
                                      r3 = ((object*) lenv) [2];
                                      r2[3] = r3;
                                      r3 = top[27];
                                      t = allocate (TC_SAVE, 5);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1096; t[4] = r0; t[5] = r1; k = t;
                                      r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                                      L1096:
                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                      r2 = result;
                                      r1[3] = r2;
                                      r2 = allocate (TC_TUPLE, 3);
                                      r3 = ((object****) lenv) [1][1][1][6];
                                      r2[2] = r3;
                                      r3 = ((object*) lenv) [2];
                                      r2[3] = r3;
                                      r3 = top[27];
                                      t = allocate (TC_SAVE, 5);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1097; t[4] = r0; t[5] = r1; k = t;
                                      r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                                      L1097:
                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                      r2 = result;
                                      r1[4] = r2;
                                      r2 = top[106];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1098; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_rlabel_458;
                                      L1098:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[3] = r1;
                                      r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                                      t = allocate (TC_SAVE, 3);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1099; ; k = t;
                                      r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
                                      L1099:
                                      ; lenv = k[2]; k = k[1];
                                      r0 = (object *) 3;
                                      r1 = ((object*) lenv) [2];
                                      r0 = box((pxll_int)unbox(r1)+unbox(r0));
                                      lenv[2] = r0;
                                      goto FUN__24loop_3325;
                                    }
                                    PXLL_RETURN(0);
                                  L1094:
                                  r1 = allocate (TC_CLOSURE, 2);
                                  r1[1] = &&FUN__24loop_3325; r1[2] = lenv;
                                  r0[2] = r1;
                                  r0 = allocate (TC_TUPLE, 2);
                                  r1 = (object *) 1;
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3325;
                                }
                                PXLL_RETURN(0);
                                fatbar_212_over: ;
                                PXLL_RETURN(0);
                              } break;
                              default: {
                                lenv = ((object ****)lenv)[1][1][1][1];
                                goto fatbar_209;
                              }
                            }
                            PXLL_RETURN(0);
                          } break;
                          default: {
                            lenv = ((object ****)lenv)[1][1][1][1];
                            goto fatbar_209;
                          }
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object ***)lenv)[1][1][1];
                        goto fatbar_209;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_209;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_209;
              }
            }
            PXLL_RETURN(0);
          } else {
            lenv = ((object *)lenv)[1];
            goto fatbar_209;
          }
          PXLL_RETURN(0);
          fatbar_210_over: ;
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_209;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_209_over;
      fatbar_209:
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object*) lenv) [3];
      r0[3] = r1;
      r1 = ((object**) lenv) [1][2];
      r0[1] = r1[2]; lenv = r0; goto FUN_type_error_3254;
      fatbar_209_over: ;
      PXLL_RETURN(0);
    L1064:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN__55_row_3257; r1[2] = lenv;
    r0[4] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][3];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][4];
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r0[1] = r1[2]; lenv = r0; goto FUN__55_3255;
  L1033:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_unify_530; r1[2] = lenv;
  r0[139] = r1;
  // def apply-subst_531
  goto L1100;
  FUN_apply_subst_531:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def p_3328
    goto L1101;
    FUN_p_3328:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = top[109];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1102; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_type_find_465;
      L1102:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = top[108];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1103; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_type__3etrec_464;
      L1103:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      r0 = ((pxll_vector*)r0)->val[2];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = top[105];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1104; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_new_tvar_453;
        L1104:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        r1 = ((object**) lenv) [1][3];
        ((pxll_vector*)r1)->val[0] = r0;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } else {
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = (object *) constructed_817[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            } else {
              goto fatbar_208;
            }
            PXLL_RETURN(0);
            goto fatbar_208_over;
            fatbar_208:
            r0 = (object *) 262;
            r1 = ((object**) lenv) [1][3];
            ((pxll_vector*)r1)->val[2] = r0;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*****) lenv) [1][1][1][1][2];
            r2[2] = r3;
            r3 = ((object**) lenv) [1][3];
            r2[3] = r3;
            r3 = top[32];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1105; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
            L1105:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = top[104];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1106; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
            L1106:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = (object *) 6;
            r1 = ((object***) lenv) [1][1][3];
            ((pxll_vector*)r1)->val[2] = r0;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][3];
            r1 = ((pxll_vector*)r1)->val[0];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(0)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object *) constructed_817[0];
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r1[1] = lenv; lenv = r1;
                r2 = (object*)TC_NIL;
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r3 = ((object****) lenv) [1][1][1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                r1[3] = r2;
                r1 = ((object*) lenv) [3];
                r2 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                lenv = lenv[1];
                r0[3] = r1;
                r1 = top[104];
                r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
              } break;
              default: {
                r0 = ((object**) lenv) [1][2];
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
            fatbar_208_over: ;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L1101:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_p_3328; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_p_3328;
  L1100:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_apply_subst_531; r1[2] = lenv;
  r0[140] = r1;
  // def symbol-set-maker_538
  goto L1107;
  FUN_symbol_set_maker_538:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = top[199];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1108; t[4] = r0; k = t;
    lenv = r1[2]; goto *r1[1];
    L1108:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_24561
    goto L1109;
    FUN_lambda_24561:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object**) lenv) [1][2];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[3] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1109:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_24561; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = top[35];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1110; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1110:
    ; lenv = k[2]; k = k[1];
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L1107:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_symbol_set_maker_538; r1[2] = lenv;
  r0[141] = r1;
  // def build-dependency-graph_539
  goto L1111;
  FUN_build_dependency_graph_539:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = top[82];
    r1[2] = r2;
    r2 = top[198];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1112; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1112:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def search_3797
    goto L1113;
    FUN_search_3797:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object***) lenv) [1][1][3];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,10)];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } break;
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object***) lenv) [1][1][3];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,10)];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1114; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L1114:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object***) lenv) [1][1][2];
          r2 = ((pxll_vector*)r2)->val[3];
          r1[2] = r2;
          r2 = top[19];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1115; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
          L1115:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[3] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_search_3797;
        } break;
        case (UOTAG(10)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1116; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L1116:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def $loop_3805
          goto L1117;
          FUN__24loop_3805:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][2];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            } else {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*****) lenv) [1][1][1][1][2];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1118; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1118:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r2 = ((pxll_vector*)r2)->val[3];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1119; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1119:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = (object *) constructed_815[0];
              r1[2] = r2;
              r2 = top[141];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1120; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_symbol_set_maker_538;
              L1120:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[4] = r1;
              r0 = allocate (TC_TUPLE, 4);
              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = ((object*) lenv) [4];
              r0[4] = r1;
              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
              r1 = ((pxll_vector*)r1)->val[1];
              r1 = ((pxll_vector*)r1)->val[0];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L1121; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L1121:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object*) lenv) [3];
              r0[2] = r1;
              r1 = ((object*) lenv) [4];
              r0[3] = r1;
              r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L1122; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_search_3797;
              L1122:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              lenv = lenv[1];
              r0 = (object *) 3;
              r1 = ((object*) lenv) [2];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv[2] = r0;
              goto FUN__24loop_3805;
            }
            PXLL_RETURN(0);
          L1117:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN__24loop_3805; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object *) 1;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1123; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3805;
          L1123:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object***) lenv) [1][1][2];
          r2 = ((pxll_vector*)r2)->val[3];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object*) lenv) [2];
          r2[2] = r3;
          r3 = top[22];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1124; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
          L1124:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1125; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L1125:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[3] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_search_3797;
        } break;
        default: {
          goto fatbar_207;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_207_over;
      fatbar_207:
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_24681
      goto L1126;
      FUN_lambda_24681:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[3] = r1;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN_search_3797;
      L1126:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_24681; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      fatbar_207_over: ;
      PXLL_RETURN(0);
    L1113:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_search_3797; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object *) constructed_814[0];
    r1[2] = r2;
    r2 = top[141];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1127; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_symbol_set_maker_538;
    L1127:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    r1 = (object *) constructed_509[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[4] = r1;
    r1 = ((object***) lenv) [1][1][2];
    r1 = ((pxll_vector*)r1)->val[1];
    r1 = ((pxll_vector*)r1)->val[0];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1128; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1128:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object****) lenv) [1][1][1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = ((object**) lenv) [1][2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1129; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_search_3797;
    L1129:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    lenv = lenv[1];
    r0 = ((object**) lenv) [1][2];
    r1 = ((object***) lenv) [1][1][3];
    ((pxll_vector*)r1)->val[3] = r0;
    PXLL_RETURN(0);
  L1111:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_build_dependency_graph_539; r1[2] = lenv;
  r0[142] = r1;
  // def symbol-add-suffix_548
  goto L1130;
  FUN_symbol_add_suffix_548:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 4);
    r2[1] = lenv; lenv = r2;
    r3 = ((object**) lenv) [1][2];
    r3 = allocate (TC_TUPLE, 2);
    r3[1] = lenv; lenv = r3;
    r4 = ((object***) lenv) [1][1][2];
    r4 = UOBJ_GET(r4,0);
    r3[2] = r4;
    r3 = ((object*) lenv) [2];
    lenv = lenv[1];
    r2[2] = r3;
    r3 = (object*)TC_NIL;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r4 = ((object**) lenv) [1][3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[4] = r3;
    r2 = ((object*) lenv) [4];
    r3 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1131; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L1131:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = top[80];
    r0[1] = r1[2]; lenv = r0; goto FUN_string__3esymbol_419;
  L1130:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_symbol_add_suffix_548; r1[2] = lenv;
  r0[143] = r1;
  // def do-simple-optimizations_551
  goto L1132;
  FUN_do_simple_optimizations_551:
    check_heap (0);
    r0 = (object *) 1;
    top[201] = r0;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_4074
    goto L1133;
    FUN_loop_4074:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 5);
      r2[1] = lenv; lenv = r2;
      r3 = allocate (TC_TUPLE, 2);
      r4 = top[201];
      r3[2] = r4;
      r4 = top[56];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L1134; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
      L1134:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = (object*)TC_NIL;
      r2[3] = r3;
      r3 = ((object*) lenv) [3];
      r4 = (object*) &constructed_813;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[4] = r3;
      r3 = ((object*) lenv) [4];
      r4 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[5] = r3;
      r2 = ((object*) lenv) [5];
      r3 = (object*) &constructed_812;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1135; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L1135:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = top[201];
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)>unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 1;
        top[201] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[145];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1136; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_simpleopt_552;
        L1136:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        lenv[2] = r0;
        goto FUN_loop_4074;
      } else {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L1133:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_4074; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[145];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1137; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
    L1137:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_4074;
  L1132:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_do_simple_optimizations_551; r1[2] = lenv;
  r0[144] = r1;
  // def simpleopt_552
  goto L1138;
  FUN_simpleopt_552:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = k; // %getcc
    r0[2] = r1;
    // def lambda_26817
    goto L1139;
    FUN_lambda_26817:
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][2];
      k = r1; // %putcc
      PXLL_RETURN(0);
    L1139:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_26817; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = top[201];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = (object *) 3;
    r0 = box((pxll_int)unbox(r1)+unbox(r0));
    lenv = lenv[1];
    top[201] = r0;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][2];
    r1 = ((pxll_vector*)r1)->val[4];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object*****) lenv) [1][1][1][1][2];
            r3 = ((pxll_vector*)r3)->val[3];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = top[19];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1140; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
            L1140:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[145];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1141; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
            L1141:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1142; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L1142:
            ; lenv = k[2]; k = k[1];
            r0 = result;
          } break;
          default: {
            goto fatbar_200;
          }
        }
        goto fatbar_200_over;
        fatbar_200:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[124];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1143; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_unpack_fix_504;
        L1143:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = ((object*) lenv) [2];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[4] = r1;
            r1 = ((object*) lenv) [2];
            r1 = ((pxll_vector*)r1)->val[3];
            r0[5] = r1;
            r0 = ((object*) lenv) [4];
            switch (get_case (r0)) {
              case (UOTAG(10)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object**) lenv) [1][5];
                r1[2] = r2;
                r2 = top[124];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1144; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_unpack_fix_504;
                L1144:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case_noint (r0)) {
                  default: {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = allocate (TC_TUPLE, 4);
                    r3 = allocate (TC_TUPLE, 3);
                    r4 = allocate (TC_TUPLE, 3);
                    r5 = ((object*****) lenv) [1][1][1][1][2];
                    r4[2] = r5;
                    r5 = (object *) constructed_36[0];
                    r4[3] = r5;
                    r5 = top[21];
                    t = allocate (TC_SAVE, 7);
                    t[1] = k; t[2] = lenv; t[3] = &&L1145; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                    r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                    L1145:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                    r4 = result;
                    r3[2] = r4;
                    r4 = ((object**) lenv) [1][2];
                    r3[3] = r4;
                    r4 = top[21];
                    t = allocate (TC_SAVE, 6);
                    t[1] = k; t[2] = lenv; t[3] = &&L1146; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                    r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                    L1146:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                    r3 = result;
                    r2[2] = r3;
                    r3 = allocate (TC_TUPLE, 3);
                    r4 = allocate (TC_TUPLE, 3);
                    r5 = ((object***) lenv) [1][1][3];
                    r4[2] = r5;
                    r5 = (object *) constructed_36[0];
                    r4[3] = r5;
                    r5 = top[21];
                    t = allocate (TC_SAVE, 7);
                    t[1] = k; t[2] = lenv; t[3] = &&L1147; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                    r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                    L1147:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                    r4 = result;
                    r3[2] = r4;
                    r4 = ((object*) lenv) [3];
                    r3[3] = r4;
                    r4 = top[21];
                    t = allocate (TC_SAVE, 6);
                    t[1] = k; t[2] = lenv; t[3] = &&L1148; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                    r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                    L1148:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                    r3 = result;
                    r2[3] = r3;
                    r3 = ((object*) lenv) [2];
                    r2[4] = r3;
                    r3 = top[121];
                    t = allocate (TC_SAVE, 5);
                    t[1] = k; t[2] = lenv; t[3] = &&L1149; t[4] = r0; t[5] = r1; k = t;
                    r2[1] = r3[2]; lenv = r2; goto FUN_node_2ffix_498;
                    L1149:
                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                    r2 = result;
                    r1[2] = r2;
                    r2 = top[145];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1150; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
                    L1150:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L1151; ; k = t;
                    r0[1] = r1[2]; lenv = r0; goto *r1[1];
                    L1151:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    lenv = lenv[1];
                  } break;
                }
                lenv = lenv[1];
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_201;
              }
            }
            lenv = lenv[1];
          } break;
        }
        goto fatbar_201_over;
        fatbar_201:
        r0 = (object *) 14;
        fatbar_201_over: ;
        lenv = lenv[1];
        fatbar_200_over: ;
        lenv = lenv[1];
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object*****) lenv) [1][1][1][1][2];
            r3 = ((pxll_vector*)r3)->val[3];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = top[19];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1152; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
            L1152:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[145];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1153; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
            L1153:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1154; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L1154:
            ; lenv = k[2]; k = k[1];
            r0 = result;
          } break;
          default: {
            goto fatbar_202;
          }
        }
        goto fatbar_202_over;
        fatbar_202:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*****) lenv) [1][1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[124];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1155; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_unpack_fix_504;
        L1155:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = ((object*) lenv) [2];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[4] = r1;
            r1 = ((object*) lenv) [2];
            r1 = ((pxll_vector*)r1)->val[3];
            r0[5] = r1;
            r0 = ((object*) lenv) [4];
            switch (get_case (r0)) {
              case (UOTAG(9)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object**) lenv) [1][5];
                r1[2] = r2;
                r2 = top[124];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1156; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_unpack_fix_504;
                L1156:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case_noint (r0)) {
                  default: {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = allocate (TC_TUPLE, 4);
                    r3 = allocate (TC_TUPLE, 3);
                    r4 = allocate (TC_TUPLE, 3);
                    r5 = ((object*****) lenv) [1][1][1][1][2];
                    r4[2] = r5;
                    r5 = (object *) constructed_36[0];
                    r4[3] = r5;
                    r5 = top[21];
                    t = allocate (TC_SAVE, 7);
                    t[1] = k; t[2] = lenv; t[3] = &&L1157; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                    r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                    L1157:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                    r4 = result;
                    r3[2] = r4;
                    r4 = ((object**) lenv) [1][2];
                    r3[3] = r4;
                    r4 = top[21];
                    t = allocate (TC_SAVE, 6);
                    t[1] = k; t[2] = lenv; t[3] = &&L1158; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                    r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                    L1158:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                    r3 = result;
                    r2[2] = r3;
                    r3 = allocate (TC_TUPLE, 3);
                    r4 = allocate (TC_TUPLE, 3);
                    r5 = ((object***) lenv) [1][1][3];
                    r4[2] = r5;
                    r5 = (object *) constructed_36[0];
                    r4[3] = r5;
                    r5 = top[21];
                    t = allocate (TC_SAVE, 7);
                    t[1] = k; t[2] = lenv; t[3] = &&L1159; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                    r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                    L1159:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                    r4 = result;
                    r3[2] = r4;
                    r4 = ((object*) lenv) [3];
                    r3[3] = r4;
                    r4 = top[21];
                    t = allocate (TC_SAVE, 6);
                    t[1] = k; t[2] = lenv; t[3] = &&L1160; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                    r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                    L1160:
                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                    r3 = result;
                    r2[3] = r3;
                    r3 = ((object*) lenv) [2];
                    r2[4] = r3;
                    r3 = top[122];
                    t = allocate (TC_SAVE, 5);
                    t[1] = k; t[2] = lenv; t[3] = &&L1161; t[4] = r0; t[5] = r1; k = t;
                    r2[1] = r3[2]; lenv = r2; goto FUN_node_2flet_499;
                    L1161:
                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                    r2 = result;
                    r1[2] = r2;
                    r2 = top[145];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1162; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
                    L1162:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L1163; ; k = t;
                    r0[1] = r1[2]; lenv = r0; goto *r1[1];
                    L1163:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                    lenv = lenv[1];
                  } break;
                }
                lenv = lenv[1];
              } break;
              default: {
                goto fatbar_203;
              }
            }
            goto fatbar_203_over;
            fatbar_203:
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][3];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1164; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L1164:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def $loop_4103
            goto L1165;
            FUN__24loop_4103:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 4);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object******) lenv) [1][1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1166; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1166:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r1 = ((pxll_vector*)r1)->val[4];
                r0[3] = r1;
                r1 = ((object*) lenv) [2];
                r1 = ((pxll_vector*)r1)->val[3];
                r0[4] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (UOTAG(9)): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][4];
                    r1[2] = r2;
                    r2 = top[124];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1167; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_unpack_fix_504;
                    L1167:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    switch (get_case_noint (r0)) {
                      default: {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][3];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][3];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = allocate (TC_TUPLE, 4);
                        r3 = allocate (TC_TUPLE, 4);
                        r3[1] = lenv; lenv = r3;
                        r4 = allocate (TC_TUPLE, 4);
                        r5 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                        r4[2] = r5;
                        r5 = (object *) 1;
                        r4[3] = r5;
                        r5 = ((object*****) lenv) [1][1][1][1][2];
                        r4[4] = r5;
                        r5 = top[29];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1168; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_slice_323;
                        L1168:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = allocate (TC_TUPLE, 3);
                        r5 = allocate (TC_TUPLE, 3);
                        r6 = ((object*) lenv) [2];
                        r5[2] = r6;
                        r6 = (object *) constructed_36[0];
                        r5[3] = r6;
                        r6 = top[21];
                        t = allocate (TC_SAVE, 8);
                        t[1] = k; t[2] = lenv; t[3] = &&L1169; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                        r5[1] = r6[2]; lenv = r5; goto FUN_reverse_onto_311;
                        L1169:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                        r5 = result;
                        r4[2] = r5;
                        r5 = ((object***) lenv) [1][1][2];
                        r4[3] = r5;
                        r5 = top[21];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1170; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                        L1170:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[3] = r4;
                        r4 = allocate (TC_TUPLE, 4);
                        r5 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                        r4[2] = r5;
                        r5 = ((object*****) lenv) [1][1][1][1][2];
                        r4[3] = r5;
                        r5 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                        r4[4] = r5;
                        r5 = top[29];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1171; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_slice_323;
                        L1171:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[4] = r4;
                        r3 = allocate (TC_TUPLE, 3);
                        r4 = allocate (TC_TUPLE, 3);
                        r5 = ((object*) lenv) [3];
                        r4[2] = r5;
                        r5 = (object *) constructed_36[0];
                        r4[3] = r5;
                        r5 = top[21];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1172; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                        L1172:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = ((object*) lenv) [4];
                        r3[3] = r4;
                        r4 = top[21];
                        t = allocate (TC_SAVE, 6);
                        t[1] = k; t[2] = lenv; t[3] = &&L1173; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                        r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                        L1173:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                        r3 = result;
                        lenv = lenv[1];
                        r2[2] = r3;
                        r3 = allocate (TC_TUPLE, 5);
                        r3[1] = lenv; lenv = r3;
                        r4 = allocate (TC_TUPLE, 4);
                        r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                        r4[2] = r5;
                        r5 = (object *) 1;
                        r4[3] = r5;
                        r5 = ((object*****) lenv) [1][1][1][1][2];
                        r4[4] = r5;
                        r5 = top[29];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1174; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_slice_323;
                        L1174:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = allocate (TC_TUPLE, 3);
                        r5 = allocate (TC_TUPLE, 3);
                        r6 = ((object*) lenv) [2];
                        r5[2] = r6;
                        r6 = (object *) constructed_36[0];
                        r5[3] = r6;
                        r6 = top[21];
                        t = allocate (TC_SAVE, 8);
                        t[1] = k; t[2] = lenv; t[3] = &&L1175; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                        r5[1] = r6[2]; lenv = r5; goto FUN_reverse_onto_311;
                        L1175:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                        r5 = result;
                        r4[2] = r5;
                        r5 = ((object**) lenv) [1][3];
                        r4[3] = r5;
                        r5 = top[21];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1176; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                        L1176:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[3] = r4;
                        r4 = allocate (TC_TUPLE, 4);
                        r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                        r4[2] = r5;
                        r5 = (object *) 3;
                        r6 = ((object*****) lenv) [1][1][1][1][2];
                        r5 = box((pxll_int)unbox(r6)+unbox(r5));
                        r4[3] = r5;
                        r5 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                        r4[4] = r5;
                        r5 = top[29];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1177; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_slice_323;
                        L1177:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[4] = r4;
                        r4 = ((object*) lenv) [4];
                        r5 = ((object**) lenv) [1][2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[5] = r4;
                        r3 = allocate (TC_TUPLE, 3);
                        r4 = allocate (TC_TUPLE, 3);
                        r5 = ((object*) lenv) [3];
                        r4[2] = r5;
                        r5 = (object *) constructed_36[0];
                        r4[3] = r5;
                        r5 = top[21];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L1178; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                        L1178:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = ((object*) lenv) [5];
                        r3[3] = r4;
                        r4 = top[21];
                        t = allocate (TC_SAVE, 6);
                        t[1] = k; t[2] = lenv; t[3] = &&L1179; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                        r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                        L1179:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                        r3 = result;
                        lenv = lenv[1];
                        r2[3] = r3;
                        r3 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                        r2[4] = r3;
                        r3 = top[122];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L1180; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_node_2flet_499;
                        L1180:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                        r1[2] = r2;
                        r2 = top[145];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L1181; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
                        L1181:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1182; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto *r1[1];
                        L1182:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                        lenv = lenv[1];
                      } break;
                    }
                    lenv = lenv[1];
                  } break;
                  default: {
                    goto fatbar_204;
                  }
                }
                goto fatbar_204_over;
                fatbar_204:
                r0 = (object *) 14;
                fatbar_204_over: ;
                lenv = lenv[1];
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_4103;
              }
              PXLL_RETURN(0);
            L1165:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN__24loop_4103; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = (object *) 1;
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1183; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN__24loop_4103;
            L1183:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r0 = (object *) 14;
            lenv = lenv[1];
            fatbar_203_over: ;
            lenv = lenv[1];
          } break;
        }
        lenv = lenv[1];
        fatbar_202_over: ;
        lenv = lenv[1];
      } break;
      case (UITAG(6)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = ((object*) lenv) [2];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[4] = r1;
            r0 = ((object*) lenv) [4];
            switch (get_case (r0)) {
              case (UOTAG(2)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(5)): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = (object *) constructed_46[0];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                      r0 = ((object***) lenv) [1][1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object****) lenv) [1][1][1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object****) lenv) [1][1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = (object *) constructed_44[0];
                                  r1 = ((object***) lenv) [1][1][3];
                                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                    r0 = allocate (TC_TUPLE, 2);
                                    r1 = allocate (TC_TUPLE, 2);
                                    r2 = ((object**) lenv) [1][2];
                                    r1[2] = r2;
                                    r2 = top[145];
                                    t = allocate (TC_SAVE, 4);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1184; t[4] = r0; k = t;
                                    r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
                                    L1184:
                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                    r1 = result;
                                    r0[2] = r1;
                                    r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                    t = allocate (TC_SAVE, 3);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1185; ; k = t;
                                    r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                    L1185:
                                    ; lenv = k[2]; k = k[1];
                                    r0 = result;
                                  } else {
                                    r0 = allocate (TC_TUPLE, 2);
                                    r1 = allocate (TC_TUPLE, 2);
                                    r2 = ((object*) lenv) [2];
                                    r1[2] = r2;
                                    r2 = top[145];
                                    t = allocate (TC_SAVE, 4);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1186; t[4] = r0; k = t;
                                    r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
                                    L1186:
                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                    r1 = result;
                                    r0[2] = r1;
                                    r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                    t = allocate (TC_SAVE, 3);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1187; ; k = t;
                                    r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                    L1187:
                                    ; lenv = k[2]; k = k[1];
                                    r0 = result;
                                  }
                                } break;
                                default: {
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  goto fatbar_205;
                                }
                              }
                              lenv = lenv[1];
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_205;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_205;
                        }
                      }
                    } else {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_205;
                    }
                    lenv = lenv[1];
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_205;
                  }
                }
                lenv = lenv[1];
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_205;
              }
            }
            lenv = lenv[1];
          } break;
          default: {
            goto fatbar_205;
          }
        }
        goto fatbar_205_over;
        fatbar_205:
        r0 = (object *) 14;
        fatbar_205_over: ;
        lenv = lenv[1];
      } break;
      case (UITAG(5)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object****) lenv) [1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1188; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L1188:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = (object *) 3;
        r0 = PXLL_TEST(unbox(r1)==unbox(r0));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object***) lenv) [1][1][2];
          r3 = ((pxll_vector*)r3)->val[3];
          r2[2] = r3;
          r3 = (object *) 1;
          r2[3] = r3;
          r3 = top[27];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1189; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
          L1189:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r2 = top[145];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1190; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
          L1190:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1191; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L1191:
          ; lenv = k[2]; k = k[1];
          r0 = result;
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][2];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1192; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L1192:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def $loop_4126
          goto L1193;
          FUN__24loop_4126:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][2];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            } else {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*****) lenv) [1][1][1][1][2];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1194; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1194:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r1 = ((pxll_vector*)r1)->val[4];
              r0[3] = r1;
              r1 = ((object*) lenv) [2];
              r1 = ((pxll_vector*)r1)->val[3];
              r0[4] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (UITAG(5)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = allocate (TC_TUPLE, 5);
                  r2[1] = lenv; lenv = r2;
                  r3 = allocate (TC_TUPLE, 4);
                  r4 = ((object******) lenv) [1][1][1][1][1][2];
                  r3[2] = r4;
                  r4 = (object *) 1;
                  r3[3] = r4;
                  r4 = ((object***) lenv) [1][1][2];
                  r3[4] = r4;
                  r4 = top[29];
                  t = allocate (TC_SAVE, 6);
                  t[1] = k; t[2] = lenv; t[3] = &&L1195; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                  r3[1] = r4[2]; lenv = r3; goto FUN_slice_323;
                  L1195:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                  r3 = result;
                  r2[2] = r3;
                  r3 = allocate (TC_TUPLE, 3);
                  r4 = allocate (TC_TUPLE, 3);
                  r5 = ((object*) lenv) [2];
                  r4[2] = r5;
                  r5 = (object *) constructed_36[0];
                  r4[3] = r5;
                  r5 = top[21];
                  t = allocate (TC_SAVE, 7);
                  t[1] = k; t[2] = lenv; t[3] = &&L1196; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                  r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                  L1196:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                  r4 = result;
                  r3[2] = r4;
                  r4 = ((object**) lenv) [1][4];
                  r3[3] = r4;
                  r4 = top[21];
                  t = allocate (TC_SAVE, 6);
                  t[1] = k; t[2] = lenv; t[3] = &&L1197; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                  r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                  L1197:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                  r3 = result;
                  r2[3] = r3;
                  r3 = allocate (TC_TUPLE, 4);
                  r4 = ((object******) lenv) [1][1][1][1][1][2];
                  r3[2] = r4;
                  r4 = ((object***) lenv) [1][1][2];
                  r5 = (object *) 3;
                  r4 = box((pxll_int)unbox(r5)+unbox(r4));
                  r3[3] = r4;
                  r4 = ((object******) lenv) [1][1][1][1][1][3];
                  r3[4] = r4;
                  r4 = top[29];
                  t = allocate (TC_SAVE, 6);
                  t[1] = k; t[2] = lenv; t[3] = &&L1198; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                  r3[1] = r4[2]; lenv = r3; goto FUN_slice_323;
                  L1198:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                  r3 = result;
                  r2[4] = r3;
                  r3 = allocate (TC_TUPLE, 3);
                  r4 = allocate (TC_TUPLE, 3);
                  r5 = ((object*) lenv) [3];
                  r4[2] = r5;
                  r5 = (object *) constructed_36[0];
                  r4[3] = r5;
                  r5 = top[21];
                  t = allocate (TC_SAVE, 7);
                  t[1] = k; t[2] = lenv; t[3] = &&L1199; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                  r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
                  L1199:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                  r4 = result;
                  r3[2] = r4;
                  r4 = ((object*) lenv) [4];
                  r3[3] = r4;
                  r4 = top[21];
                  t = allocate (TC_SAVE, 6);
                  t[1] = k; t[2] = lenv; t[3] = &&L1200; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                  r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                  L1200:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                  r3 = result;
                  r2[5] = r3;
                  r2 = allocate (TC_TUPLE, 3);
                  r3 = (object*)UITAG(5);
                  r2[2] = r3;
                  r3 = ((object*) lenv) [5];
                  r2[3] = r3;
                  r3 = top[118];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L1201; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_make_node_486;
                  L1201:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  lenv = lenv[1];
                  r1[2] = r2;
                  r2 = top[145];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1202; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_simpleopt_552;
                  L1202:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1203; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1203:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                } break;
                default: {
                  goto fatbar_206;
                }
              }
              goto fatbar_206_over;
              fatbar_206:
              r0 = (object *) 14;
              fatbar_206_over: ;
              lenv = lenv[1];
              r0 = (object *) 3;
              r1 = ((object*) lenv) [2];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv[2] = r0;
              goto FUN__24loop_4126;
            }
            PXLL_RETURN(0);
          L1193:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN__24loop_4126; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object *) 1;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1204; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN__24loop_4126;
          L1204:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          lenv = lenv[1];
          r0 = (object *) 14;
          lenv = lenv[1];
        }
      } break;
      default: {
        goto fatbar_199;
      }
    }
    goto fatbar_199_over;
    fatbar_199:
    r0 = (object *) 14;
    fatbar_199_over: ;
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = top[201];
    r0[2] = r1;
    r0 = (object *) 3;
    r1 = ((object*) lenv) [2];
    r0 = box((pxll_int)unbox(r1)-unbox(r0));
    lenv = lenv[1];
    top[201] = r0;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = top[145];
    r1[2] = r2;
    r2 = ((object***) lenv) [1][1][2];
    r2 = ((pxll_vector*)r2)->val[3];
    r1[3] = r2;
    r2 = top[32];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1205; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
    L1205:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object***) lenv) [1][1][2];
    ((pxll_vector*)r1)->val[3] = r0;
    r0 = ((object***) lenv) [1][1][2];
    PXLL_RETURN(0);
  L1138:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_simpleopt_552; r1[2] = lenv;
  r0[145] = r1;
  // def do-one-round_555
  goto L1206;
  FUN_do_one_round_555:
    check_heap (0);
    r0 = top[165];
    r1 = ((object*) lenv) [3];
    ((pxll_vector*)r1)->val[16] = r0;
    r0 = top[165];
    r1 = ((object*) lenv) [3];
    ((pxll_vector*)r1)->val[4] = r0;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = top[8];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1207; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_add_vars_251;
    L1207:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) constructed_509[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = top[7];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1208; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_add_var_250;
    L1208:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def walk_3875_i3067_i3070
    goto L1209;
    FUN_walk_3875_i3067_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(8)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object***) lenv) [1][1][2];
          r2 = ((pxll_vector*)r2)->val[3];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1210; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L1210:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[4];
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object******) lenv) [1][1][1][1][1][3];
              r1[2] = r2;
              r2 = ((object*) lenv) [2];
              r1[3] = r2;
              r2 = top[5];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1211; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
              L1211:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object*) lenv) [2];
              r0[2] = r1;
              r1 = ((object****) lenv) [1][1][1][3];
              r0[3] = r1;
              r1 = top[25];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L1212; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
              L1212:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = ((pxll_vector*)r1)->val[1];
                r0[2] = r1;
                r0 = top[153];
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)|(1<<unbox(r0)));
                lenv = lenv[1];
                r1 = ((object*) lenv) [3];
                ((pxll_vector*)r1)->val[1] = r0;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*****) lenv) [1][1][1][1][2];
                r1 = ((pxll_vector*)r1)->val[0];
                r0[2] = r1;
                r0 = top[193];
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)|(1<<unbox(r0)));
                lenv = lenv[1];
                r1 = ((object****) lenv) [1][1][1][2];
                ((pxll_vector*)r1)->val[0] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = ((pxll_vector*)r1)->val[0];
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r1 = (object *) 3;
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv = lenv[1];
              r1 = ((object*) lenv) [3];
              ((pxll_vector*)r1)->val[0] = r0;
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_198;
            }
          }
          goto fatbar_198_over;
          fatbar_198:
          r0 = (object *) 14;
          fatbar_198_over: ;
          lenv = lenv[1];
        } break;
        case (UOTAG(7)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
          r0 = allocate (TC_TUPLE, 5);
          r1 = ((object*****) lenv) [1][1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[4];
          r0[2] = r1;
          r1 = top[82];
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          r0[4] = r1;
          r1 = ((object***) lenv) [1][1][2];
          r0[5] = r1;
          r1 = top[72];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1213; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
          L1213:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          ((pxll_vector*)r1)->val[4] = r0;
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object*****) lenv) [1][1][1][1][3];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[155];
          r0[4] = r1;
          r1 = top[6];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1214; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
          L1214:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
        } break;
        case (UOTAG(12)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = (object *) constructed_214[0];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object *) 262;
          } else {
            r0 = (object *) constructed_210[0];
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
          }
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object*****) lenv) [1][1][1][1][3];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object***) lenv) [1][1][3];
            r1[2] = r2;
            r2 = top[19];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1215; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
            L1215:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = top[157];
            r0[4] = r1;
            r1 = top[6];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1216; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
            L1216:
            ; lenv = k[2]; k = k[1];
            r0 = result;
          } else {
            r0 = (object *) 14;
          }
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_197;
        }
      }
      goto fatbar_197_over;
      fatbar_197:
      r0 = (object *) 14;
      fatbar_197_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_25382_i3067_i3070
      goto L1217;
      FUN_lambda_25382_i3067_i3070:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN_walk_3875_i3067_i3070;
      L1217:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_25382_i3067_i3070; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1209:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_walk_3875_i3067_i3070; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = (object *) constructed_811[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1218; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_walk_3875_i3067_i3070;
    L1218:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    // def add-ref_3893_i3068_i3070
    goto L1219;
    FUN_add_ref_3893_i3068_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object****) lenv) [1][1][1][3];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = top[5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1220; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
      L1220:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,70)];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = (object *) 3;
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      lenv = lenv[1];
      r1 = ((object*) lenv) [2];
      ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,70)] = r0;
      PXLL_RETURN(0);
    L1219:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_add_ref_3893_i3068_i3070; r1[2] = lenv;
    r0[2] = r1;
    // def add-set_3894_i3068_i3070
    goto L1221;
    FUN_add_set_3894_i3068_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object****) lenv) [1][1][1][3];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = top[5];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1222; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
      L1222:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = (object *) 3;
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      lenv = lenv[1];
      r1 = ((object*) lenv) [2];
      ((pxll_vector*)r1)->val[4] = r0;
      PXLL_RETURN(0);
    L1221:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_add_set_3894_i3068_i3070; r1[2] = lenv;
    r0[3] = r1;
    // def walk_3895_i3068_i3070
    goto L1223;
    FUN_walk_3895_i3068_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*****) lenv) [1][1][1][1][3];
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[5];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1224; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
          L1224:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,70)];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = (object *) 3;
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv = lenv[1];
          r1 = ((object*) lenv) [3];
          ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,70)] = r0;
          lenv = lenv[1];
        } break;
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*****) lenv) [1][1][1][1][3];
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[5];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1225; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
          L1225:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = ((pxll_vector*)r1)->val[4];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = (object *) 3;
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv = lenv[1];
          r1 = ((object*) lenv) [3];
          ((pxll_vector*)r1)->val[4] = r0;
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_196;
        }
      }
      goto fatbar_196_over;
      fatbar_196:
      r0 = (object *) 14;
      fatbar_196_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object**) lenv) [1][4];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1223:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_walk_3895_i3068_i3070; r1[2] = lenv;
    r0[4] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [4];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1226; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_walk_3895_i3068_i3070;
    L1226:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_810[0];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    // def find-escaping-functions_4046_i3069_i3070
    goto L1227;
    FUN_find_escaping_functions_4046_i3069_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(7)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r1 = ((pxll_vector*)r1)->val[4];
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (UOTAG(10)): {
              r0 = (object *) 14;
            } break;
            default: {
              goto fatbar_194;
            }
          }
          goto fatbar_194_over;
          fatbar_194:
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object******) lenv) [1][1][1][1][1][3];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[154];
          r0[4] = r1;
          r1 = top[6];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1228; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
          L1228:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object*****) lenv) [1][1][1][1][2] = r0;
          fatbar_194_over: ;
          lenv = lenv[1];
        } break;
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = top[5];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1229; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
          L1229:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,72)];
          r0[3] = r1;
          r0 = top[155];
          r1 = ((object*) lenv) [3];
          r0 = PXLL_TEST((unbox(r1)&(1<<unbox(r0)))>0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][3];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UITAG(8)): {
                r0 = allocate (TC_TUPLE, 4);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*****) lenv) [1][1][1][1][3];
                r1 = ((pxll_vector*)r1)->val[3];
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object*) lenv) [2];
                r1[2] = r2;
                r2 = top[19];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1230; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
                L1230:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = ((object*****) lenv) [1][1][1][1][2];
                r2 = ((object*) lenv) [3];
                r1 = PXLL_TEST(r2==r1);
                r0[4] = r1;
                r0 = (object *) 6;
                r1 = ((object*) lenv) [4];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 4);
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r0[3] = r1;
                  r1 = top[154];
                  r0[4] = r1;
                  r1 = top[6];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1231; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
                  L1231:
                  ; lenv = k[2]; k = k[1];
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = ((object***) lenv) [1][1][2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  ((object******) lenv) [1][1][1][1][1][2] = r0;
                } else {
                  r0 = (object *) 14;
                }
              } break;
              default: {
                goto fatbar_195;
              }
            }
            goto fatbar_195_over;
            fatbar_195:
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = top[154];
            r0[4] = r1;
            r1 = top[6];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1232; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
            L1232:
            ; lenv = k[2]; k = k[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object***) lenv) [1][1][2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object******) lenv) [1][1][1][1][1][2] = r0;
            fatbar_195_over: ;
            lenv = lenv[1];
          } else {
            r0 = (object *) 14;
          }
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_193;
        }
      }
      goto fatbar_193_over;
      fatbar_193:
      r0 = (object *) 14;
      fatbar_193_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_26626_i3069_i3070
      goto L1233;
      FUN_lambda_26626_i3069_i3070:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r0[1] = r1[2]; lenv = r0; goto FUN_find_escaping_functions_4046_i3069_i3070;
      L1233:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_26626_i3069_i3070; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1227:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_find_escaping_functions_4046_i3069_i3070; r1[2] = lenv;
    r0[2] = r1;
    // def maybe-var-escapes_4047_i3069_i3070
    goto L1234;
    FUN_maybe_var_escapes_4047_i3069_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][3];
      r1[3] = r2;
      r2 = top[25];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1235; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_member_eq_question_319;
      L1235:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 6;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(r1==r0);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object****) lenv) [1][1][1][3];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[154];
        r0[4] = r1;
        r1 = top[6];
        r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
      } else {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L1234:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_maybe_var_escapes_4047_i3069_i3070; r1[2] = lenv;
    r0[3] = r1;
    // def find-escaping-variables_4048_i3069_i3070
    goto L1236;
    FUN_find_escaping_variables_4048_i3069_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(7)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1237; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1237:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[21];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1238; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          L1238:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
          lenv = lenv[1];
        } break;
        case (UOTAG(10)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1239; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1239:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[21];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1240; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          L1240:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
          lenv = lenv[1];
        } break;
        case (UOTAG(9)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1241; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1241:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[21];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1242; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          L1242:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
          lenv = lenv[1];
        } break;
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[3] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1243; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_maybe_var_escapes_4047_i3069_i3070;
          L1243:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
        } break;
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[3] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1244; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_maybe_var_escapes_4047_i3069_i3070;
          L1244:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_192;
        }
      }
      goto fatbar_192_over;
      fatbar_192:
      r0 = (object *) 14;
      fatbar_192_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_26711_i3069_i3070
      goto L1245;
      FUN_lambda_26711_i3069_i3070:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[1] = r1[2]; lenv = r0; goto FUN_find_escaping_variables_4048_i3069_i3070;
      L1245:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_26711_i3069_i3070; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[35];
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1236:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_find_escaping_variables_4048_i3069_i3070; r1[2] = lenv;
    r0[4] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = (object *) 1;
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r2;
    r2 = t;
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(2),1);
    t[1] = r2;
    r2 = t;
    r1[2] = r2;
    r2 = (object *) constructed_809[0];
    r1[3] = r2;
    r2 = top[118];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1246; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
    L1246:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1247; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_find_escaping_functions_4046_i3069_i3070;
    L1247:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_26755_i3069_i3070
    goto L1248;
    FUN_lambda_26755_i3069_i3070:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 4);
      r2 = ((object*****) lenv) [1][1][1][1][3];
      r2 = ((pxll_vector*)r2)->val[4];
      r1[2] = r2;
      r2 = top[82];
      r1[3] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[4] = r2;
      r2 = top[73];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1249; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
      L1249:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      switch (get_case (r1)) {
        case (UOTAG(0)): {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
        } break;
        default: {
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_808;
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1250; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1250:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
        } break;
      }
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object*) lenv) [3];
      r0[2] = r1;
      r1 = (object *) constructed_807[0];
      r0[3] = r1;
      r1 = ((object***) lenv) [1][1][4];
      r0[1] = r1[2]; lenv = r0; goto FUN_find_escaping_variables_4048_i3069_i3070;
    L1248:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_26755_i3069_i3070; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = top[35];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1251; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1251:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    lenv = lenv[1];
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = top[142];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1252; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_build_dependency_graph_539;
    L1252:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 5);
    r1[1] = lenv; lenv = r1;
    r2 = allocate (TC_TUPLE, 2);
    r3 = allocate (TC_TUPLE, 3);
    r3[1] = lenv; lenv = r3;
    r4 = ((object***) lenv) [1][1][3];
    r4 = ((pxll_vector*)r4)->val[3];
    r3[2] = r4;
    r4 = allocate (TC_TUPLE, 2);
    r5 = (object *) constructed_806[0];
    r4[2] = r5;
    r5 = top[141];
    t = allocate (TC_SAVE, 7);
    t[1] = k; t[2] = lenv; t[3] = &&L1253; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
    r4[1] = r5[2]; lenv = r4; goto FUN_symbol_set_maker_538;
    L1253:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
    r4 = result;
    r3[3] = r4;
    r3 = allocate (TC_TUPLE, 3);
    r3[1] = lenv; lenv = r3;
    // def walk_4136_i3072
    goto L1254;
    FUN_walk_4136_i3072:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object****) lenv) [1][1][1][2];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = ((object****) lenv) [1][1][1][2];
      r2 = ((pxll_vector*)r2)->val[1];
      r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,11)];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1255; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto *r2[1];
      L1255:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(1)): {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*****) lenv) [1][1][1][1][3];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][2];
          r0[3] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
          r1 = ((pxll_vector*)r1)->val[0];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1256; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L1256:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_27337_i3072
          goto L1257;
          FUN_lambda_27337_i3072:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
            r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
            r2 = ((pxll_vector*)r2)->val[2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1258; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L1258:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = (object *) 6;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r1 = ((object*) lenv) [2];
              r0[2] = r1;
              r1 = ((object*****) lenv) [1][1][1][1][2];
              r0[1] = r1[2]; lenv = r0; goto FUN_walk_4136_i3072;
            } else {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          L1257:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_27337_i3072; r1[2] = lenv;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
          r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,11)];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1259; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto *r2[1];
          L1259:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = top[35];
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        } break;
      }
      PXLL_RETURN(0);
    L1254:
    r4 = allocate (TC_CLOSURE, 2);
    r4[1] = &&FUN_walk_4136_i3072; r4[2] = lenv;
    r3[2] = r4;
    // def trim_4137_i3072
    goto L1260;
    FUN_trim_4137_i3072:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      switch (get_case (r1)) {
        case (UOTAG(10)): {
          r1 = allocate (TC_TUPLE, 5);
          r1[1] = lenv; lenv = r1;
          r2 = ((object**) lenv) [1][2];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object*) lenv) [2];
          r2[2] = r3;
          r3 = top[22];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1261; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
          L1261:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = ((object***) lenv) [1][1][2];
          r2 = ((pxll_vector*)r2)->val[3];
          r1[4] = r2;
          r2 = (object *) constructed_805[0];
          r1[5] = r2;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          // def $loop_4149_i3072
          goto L1262;
          FUN__24loop_4149_i3072:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][2];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object*****) lenv) [1][1][1][1][2];
              r2[2] = r3;
              r3 = ((object**) lenv) [1][2];
              r2[3] = r3;
              r3 = top[27];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1263; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
              L1263:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
              r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
              r2 = ((pxll_vector*)r2)->val[2];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1264; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L1264:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = (object *) 6;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*****) lenv) [1][1][1][1][5];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object****) lenv) [1][1][1][5] = r0;
              } else {
                r0 = (object *) 14;
              }
              r0 = (object *) 3;
              r1 = ((object*) lenv) [2];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv[2] = r0;
              goto FUN__24loop_4149_i3072;
            }
            PXLL_RETURN(0);
          L1262:
          r2 = allocate (TC_CLOSURE, 2);
          r2[1] = &&FUN__24loop_4149_i3072; r2[2] = lenv;
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 2);
          r2 = (object *) 1;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1265; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4149_i3072;
          L1265:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
          lenv = lenv[1];
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object**) lenv) [1][5];
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          switch (get_case (r1)) {
            case (TC_NIL): {
              r1 = (object *) 262;
            } break;
            default: {
              goto fatbar_191;
            }
          }
          goto fatbar_191_over;
          fatbar_191:
          r1 = (object *) 6;
          fatbar_191_over: ;
          lenv = lenv[1];
          if PXLL_IS_TRUE(r1) {
            r1 = ((object***) lenv) [1][1][2];
          } else {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = (object *) constructed_804[0];
            r1[2] = r2;
            r2 = (object *) constructed_803[0];
            r1[3] = r2;
            r2 = allocate (TC_TUPLE, 3);
            // def lambda_27418_i3072
            goto L1266;
            FUN_lambda_27418_i3072:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object***) lenv) [1][1][2];
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = top[27];
              r0[1] = r1[2]; lenv = r0; goto FUN_nth_321;
            L1266:
            r3 = allocate (TC_CLOSURE, 2);
            r3[1] = &&FUN_lambda_27418_i3072; r3[2] = lenv;
            r2[2] = r3;
            r3 = ((object**) lenv) [1][5];
            r2[3] = r3;
            r3 = top[32];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1267; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
            L1267:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][3];
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            // def $loop_4156_i3072
            goto L1268;
            FUN__24loop_4156_i3072:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object**) lenv) [1][2];
                r1[2] = r2;
                r2 = ((object******) lenv) [1][1][1][1][1][5];
                r1[3] = r2;
                r2 = top[25];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1269; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_member_eq_question_319;
                L1269:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = (object *) 6;
                r1 = ((object*) lenv) [2];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object******) lenv) [1][1][1][1][1][2];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r1[3] = r2;
                  r2 = top[27];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1270; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                  L1270:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object*****) lenv) [1][1][1][1][2];
                  r0[3] = r1;
                  r0 = ((object*) lenv) [3];
                  r1 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  ((object****) lenv) [1][1][1][2] = r0;
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object******) lenv) [1][1][1][1][1][4];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r1[3] = r2;
                  r2 = top[27];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1271; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                  L1271:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object*****) lenv) [1][1][1][1][3];
                  r0[3] = r1;
                  r0 = ((object*) lenv) [3];
                  r1 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  ((object****) lenv) [1][1][1][3] = r0;
                } else {
                  r0 = (object *) 14;
                }
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_4156_i3072;
              }
              PXLL_RETURN(0);
            L1268:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN__24loop_4156_i3072; r2[2] = lenv;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object *) 1;
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1272; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4156_i3072;
            L1272:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r1 = allocate (TC_TUPLE, 4);
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1273; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L1273:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1274; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L1274:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[3] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[23];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1275; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_last_317;
            L1275:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[4] = r2;
            r2 = top[121];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1276; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_node_2ffix_498;
            L1276:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
          }
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_190;
        }
      }
      goto fatbar_190_over;
      fatbar_190:
      r1 = ((object**) lenv) [1][2];
      fatbar_190_over: ;
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object***) lenv) [1][1][3];
      r0[2] = r1;
      r1 = ((object*) lenv) [3];
      r1 = ((pxll_vector*)r1)->val[3];
      r0[3] = r1;
      r1 = top[32];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1277; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_map_326;
      L1277:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      r1 = ((object*) lenv) [3];
      ((pxll_vector*)r1)->val[3] = r0;
      r0 = ((object*) lenv) [3];
      PXLL_RETURN(0);
    L1260:
    r4 = allocate (TC_CLOSURE, 2);
    r4[1] = &&FUN_trim_4137_i3072; r4[2] = lenv;
    r3[3] = r4;
    r3 = allocate (TC_TUPLE, 2);
    r4 = (object *) constructed_509[0];
    r3[2] = r4;
    r4 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L1278; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_walk_4136_i3072;
    L1278:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = allocate (TC_TUPLE, 2);
    r4 = ((object****) lenv) [1][1][1][2];
    r3[2] = r4;
    r4 = ((object*) lenv) [3];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L1279; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_trim_4137_i3072;
    L1279:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    lenv = lenv[1];
    lenv = lenv[1];
    r2[2] = r3;
    r3 = top[144];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1280; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_do_simple_optimizations_551;
    L1280:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = allocate (TC_TUPLE, 2);
    r3 = (object *) 1;
    r2[2] = r3;
    r3 = top[113];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1281; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_make_counter_472;
    L1281:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[3] = r2;
    r2 = allocate (TC_TUPLE, 2);
    r3 = (object *) 1;
    r2[2] = r3;
    r3 = top[113];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1282; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_make_counter_472;
    L1282:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[4] = r2;
    r2 = (object*)UITAG(0);
    r1[5] = r2;
    r1 = allocate (TC_TUPLE, 5);
    r1[1] = lenv; lenv = r1;
    // def follow-aliases_3912_i3071
    goto L1283;
    FUN_follow_aliases_3912_i3071:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][3];
      r1[3] = r2;
      r2 = top[65];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1284; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
      L1284:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(1)): {
          r0 = (object*)UITAG(1);
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r1 = ((object****) lenv) [1][1][1][2];
              lenv = ((object ***)lenv)[1][1][1];
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN_follow_aliases_3912_i3071;
            } break;
            default: {
              goto fatbar_189;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_189_over;
          fatbar_189:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (UOTAG(1), 2);
          r2 = ((object****) lenv) [1][1][1][3];
          r1[1] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
          fatbar_189_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1283:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_follow_aliases_3912_i3071; r2[2] = lenv;
    r1[2] = r2;
    // def inline_3914_i3071
    goto L1285;
    FUN_inline_3914_i3071:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = k; // %getcc
      r0[2] = r1;
      // def lambda_25644_i3071
      goto L1286;
      FUN_lambda_25644_i3071:
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        k = r1; // %putcc
        PXLL_RETURN(0);
      L1286:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_25644_i3071; r1[2] = lenv;
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(10)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1287; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L1287:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def $loop_3944_i3071
          goto L1288;
          FUN__24loop_3944_i3071:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][3];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            } else {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object****) lenv) [1][1][1][2];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1289; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1289:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r2 = ((pxll_vector*)r2)->val[3];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1290; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1290:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
              r0[4] = r1;
              r0 = ((object*) lenv) [3];
              r1 = ((object*) lenv) [4];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (UOTAG(1),3);
              t[1] = r2;
              t[2] = r0;
              t[3] = r1;
              r0 = t;
              lenv = lenv[1];
              ((object******) lenv) [1][1][1][1][1][3] = r0;
              r0 = (object *) 3;
              r1 = ((object*) lenv) [2];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv[2] = r0;
              goto FUN__24loop_3944_i3071;
            }
            PXLL_RETURN(0);
          L1288:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN__24loop_3944_i3071; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object *) 1;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1291; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3944_i3071;
          L1291:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          lenv = lenv[1];
        } break;
        case (UITAG(8)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][2];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = (object*) &constructed_68;
              r0 = box((pxll_int)((pxll_string*)(r0))->len);
              r1 = (object*) &constructed_68;
              (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
              r0 = (object*) &constructed_67;
              dump_object (r0, 0); fprintf (stdout, "\n");
              r0 = (object*) &constructed_67;
              PXLL_UNDEFINED; result=r0; goto Lreturn;
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r1 = ((object*) lenv) [2];
              r1 = ((pxll_vector*)r1)->val[4];
              r0[4] = r1;
              r0 = ((object*) lenv) [4];
              switch (get_case (r0)) {
                case (UOTAG(0)): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][4];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object******) lenv) [1][1][1][1][1][3];
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  r1[3] = r2;
                  r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1292; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_follow_aliases_3912_i3071;
                  L1292:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r0 = ((object*) lenv) [3];
                  switch (get_case (r0)) {
                    case (UITAG(1)): {
                      r0 = (object *) 14;
                    } break;
                    default: {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case_noint (r0)) {
                        default: {
                          r0 = allocate (TC_TUPLE, 13);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                          r1[2] = r2;
                          r2 = ((object*) lenv) [2];
                          r1[3] = r2;
                          r2 = top[5];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L1293; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
                          L1293:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[4] = r1;
                          r1 = ((object*) lenv) [4];
                          r1 = ((pxll_vector*)r1)->val[1];
                          r0[5] = r1;
                          r1 = top[154];
                          r2 = ((object*) lenv) [5];
                          r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
                          r0[6] = r1;
                          r1 = ((object*) lenv) [4];
                          r1 = ((pxll_vector*)r1)->val[1];
                          r0[7] = r1;
                          r1 = top[153];
                          r2 = ((object*) lenv) [7];
                          r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
                          r0[8] = r1;
                          r1 = ((object*) lenv) [4];
                          r1 = ((pxll_vector*)r1)->val[1];
                          r0[9] = r1;
                          r1 = top[157];
                          r2 = ((object*) lenv) [9];
                          r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
                          r0[10] = r1;
                          r1 = ((object*) lenv) [4];
                          r1 = ((pxll_vector*)r1)->val[0];
                          r0[11] = r1;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][5];
                          r1[2] = r2;
                          r2 = ((object*) lenv) [2];
                          r1[3] = r2;
                          r2 = top[65];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L1294; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
                          L1294:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[12] = r1;
                          r1 = ((object*) lenv) [12];
                          switch (get_case (r1)) {
                            case (UOTAG(0)): {
                              r1 = allocate (TC_TUPLE, 2);
                              r1[1] = lenv; lenv = r1;
                              r2 = ((object**) lenv) [1][12];
                              r2 = UOBJ_GET(r2,0);
                              r1[2] = r2;
                              r1 = ((object**) lenv) [1][11];
                              r2 = ((object*) lenv) [2];
                              r1 = box((pxll_int)unbox(r2)*unbox(r1));
                              lenv = lenv[1];
                            } break;
                            default: {
                              r1 = ((object*) lenv) [11];
                            } break;
                          }
                          r0[13] = r1;
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][3];
                          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
                          r0[2] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (UOTAG(7)): {
                              r0 = (object *) 262;
                            } break;
                            default: {
                              goto fatbar_188;
                            }
                          }
                          goto fatbar_188_over;
                          fatbar_188:
                          r0 = (object *) 6;
                          fatbar_188_over: ;
                          lenv = lenv[1];
                          if PXLL_IS_TRUE(r0) {
                            r0 = allocate (TC_TUPLE, 4);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object**) lenv) [1][2];
                            r1 = allocate (TC_TUPLE, 2);
                            r1[1] = lenv; lenv = r1;
                            r2 = ((object***) lenv) [1][1][2];
                            r2 = UOBJ_GET(r2,0);
                            r1[2] = r2;
                            r1 = ((object*) lenv) [2];
                            lenv = lenv[1];
                            r0[2] = r1;
                            r1 = (object *) 1;
                            r2 = ((object*) lenv) [2];
                            range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
                            r1 = (object *) 1;
                            r2 = ((object*) lenv) [2];
                            r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
                            r0[3] = r1;
                            r1 = (object *) 24066;
                            r2 = ((object*) lenv) [3];
                            r1 = PXLL_TEST(r2==r1);
                            r0[4] = r1;
                            r0 = (object *) 6;
                            r1 = ((object*) lenv) [4];
                            r0 = PXLL_TEST(r1==r0);
                            lenv = lenv[1];
                            if PXLL_IS_TRUE(r0) {
                              r0 = (object *) 6;
                              r1 = ((object*) lenv) [10];
                              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                r0 = (object *) 1;
                                r1 = ((object*) lenv) [13];
                                if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,74)];
                                  r0[2] = r1;
                                  r0 = top[200];
                                  r1 = ((object*) lenv) [2];
                                  r0 = PXLL_TEST(unbox(r1)<=unbox(r0));
                                  lenv = lenv[1];
                                  if PXLL_IS_TRUE(r0) {
                                    r0 = (object *) 262;
                                  } else {
                                    r0 = (object *) 3;
                                    r1 = ((object*) lenv) [13];
                                    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                                      r0 = (object *) 6;
                                      r1 = ((object*) lenv) [6];
                                      r0 = PXLL_TEST(r1==r0);
                                    } else {
                                      r0 = (object *) 6;
                                    }
                                  }
                                  if PXLL_IS_TRUE(r0) {
                                    r0 = (object *) 6;
                                    r1 = ((object*) lenv) [8];
                                    r0 = PXLL_TEST(r1==r0);
                                  } else {
                                    r0 = (object *) 6;
                                  }
                                } else {
                                  r0 = (object *) 6;
                                }
                              } else {
                                r0 = (object *) 6;
                              }
                            } else {
                              r0 = (object *) 6;
                            }
                          } else {
                            r0 = (object *) 6;
                          }
                          if PXLL_IS_TRUE(r0) {
                            r0 = (object *) 3;
                            r1 = ((object*) lenv) [13];
                            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
                              r1 = ((pxll_vector*)r1)->val[3];
                              r0[2] = r1;
                              r1 = allocate (TC_TUPLE, 3);
                              r2 = ((object*) lenv) [2];
                              r1[2] = r2;
                              r2 = ((object**) lenv) [1][2];
                              r1[3] = r2;
                              r2 = ((object*) lenv) [2];
                              r2 = ((pxll_vector*)r2)->val[1];
                              r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,11)];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L1295; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto *r2[1];
                              L1295:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (UOTAG(0)): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object*) lenv) [2];
                                  r0[2] = r1;
                                  // def lambda_25564_i2951_i3071
                                  goto L1296;
                                  FUN_lambda_25564_i2951_i3071:
                                    check_heap (0);
                                    r0 = allocate (TC_TUPLE, 2);
                                    r0[1] = lenv; lenv = r0;
                                    r1 = allocate (TC_TUPLE, 3);
                                    r2 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][5];
                                    r1[2] = r2;
                                    r2 = ((object**) lenv) [1][2];
                                    r1[3] = r2;
                                    r2 = top[65];
                                    t = allocate (TC_SAVE, 4);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1297; t[4] = r0; k = t;
                                    r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
                                    L1297:
                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                    r1 = result;
                                    r0[2] = r1;
                                    r0 = ((object*) lenv) [2];
                                    switch (get_case (r0)) {
                                      case (UITAG(1)): {
                                        r0 = allocate (TC_TUPLE, 2);
                                        r0[1] = lenv; lenv = r0;
                                        r1 = ((object***************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][5];
                                        r0[2] = r1;
                                        r0 = ((object******) lenv) [1][1][1][1][1][13];
                                        r1 = ((object*) lenv) [2];
                                        r2 = ((object***) lenv) [1][1][2];
                                        t = alloc_no_clear (UOTAG(1),3);
                                        t[1] = r2;
                                        t[2] = r0;
                                        t[3] = r1;
                                        r0 = t;
                                        lenv = lenv[1];
                                        ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][5] = r0;
                                        PXLL_RETURN(0);
                                      } break;
                                      default: {
                                        r0 = (object *) 14;
                                        PXLL_RETURN(0);
                                      } break;
                                    }
                                    PXLL_RETURN(0);
                                  L1296:
                                  r1 = allocate (TC_CLOSURE, 2);
                                  r1[1] = &&FUN_lambda_25564_i2951_i3071; r1[2] = lenv;
                                  r0[3] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
                                  r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,25)];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1298; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                  L1298:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = result;
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  r0 = (object *) 14;
                                } break;
                              }
                              lenv = lenv[1];
                            } else {
                              r0 = (object *) 14;
                            }
                            r0 = allocate (TC_TUPLE, 2);
                            r0[1] = lenv; lenv = r0;
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = ((object**) lenv) [1][3];
                            r1[2] = r2;
                            r2 = ((object*****) lenv) [1][1][1][1][3];
                            r1[3] = r2;
                            r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L1299; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_inline_application_3917_i3071;
                            L1299:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r0 = allocate (TC_TUPLE, 3);
                            r1 = ((object*) lenv) [2];
                            r0[2] = r1;
                            r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
                            r0[3] = r1;
                            r1 = top[8];
                            t = allocate (TC_SAVE, 3);
                            t[1] = k; t[2] = lenv; t[3] = &&L1300; ; k = t;
                            r0[1] = r1[2]; lenv = r0; goto FUN_add_vars_251;
                            L1300:
                            ; lenv = k[2]; k = k[1];
                            r0 = allocate (TC_TUPLE, 2);
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = ((object*) lenv) [2];
                            r1[2] = r2;
                            r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                            r1[3] = r2;
                            r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L1301; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_inline_3914_i3071;
                            L1301:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                            t = allocate (TC_SAVE, 3);
                            t[1] = k; t[2] = lenv; t[3] = &&L1302; ; k = t;
                            r0[1] = r1[2]; lenv = r0; goto *r1[1];
                            L1302:
                            ; lenv = k[2]; k = k[1];
                            r0 = result;
                            lenv = lenv[1];
                          } else {
                            r0 = (object *) 14;
                          }
                          lenv = lenv[1];
                        } break;
                      }
                      lenv = lenv[1];
                    } break;
                  }
                  lenv = lenv[1];
                } break;
                case (UOTAG(7)): {
                  r0 = allocate (TC_TUPLE, 4);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][4];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][4];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object******) lenv) [1][1][1][1][1][2];
                  r3 = ((pxll_vector*)r3)->val[lookup_field((GET_TYPECODE(*r3)-TC_USEROBJ)>>2,75)];
                  r2[2] = r3;
                  r3 = top[19];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L1303; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                  L1303:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][3];
                  r1[3] = r2;
                  r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1304; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_inline_application_3917_i3071;
                  L1304:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[4] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = ((object*) lenv) [4];
                  r0[2] = r1;
                  r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                  r0[3] = r1;
                  r1 = top[8];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1305; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_add_vars_251;
                  L1305:
                  ; lenv = k[2]; k = k[1];
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object*) lenv) [4];
                  r1[2] = r2;
                  r2 = ((object******) lenv) [1][1][1][1][1][3];
                  r1[3] = r2;
                  r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1306; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_inline_3914_i3071;
                  L1306:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = ((object*****) lenv) [1][1][1][1][3];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1307; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                  L1307:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  lenv = lenv[1];
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_187;
                }
              }
              lenv = lenv[1];
            } break;
          }
          goto fatbar_187_over;
          fatbar_187:
          r0 = (object *) 14;
          fatbar_187_over: ;
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_186;
        }
      }
      goto fatbar_186_over;
      fatbar_186:
      r0 = (object *) 14;
      fatbar_186_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_25894_i3071
      goto L1308;
      FUN_lambda_25894_i3071:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[3] = r1;
        r1 = ((object****) lenv) [1][1][1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_inline_3914_i3071;
      L1308:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_25894_i3071; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,75)];
      r0[3] = r1;
      r1 = top[32];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1309; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_map_326;
      L1309:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      r1 = ((object**) lenv) [1][2];
      ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,75)] = r0;
      r0 = ((object**) lenv) [1][2];
      PXLL_RETURN(0);
    L1285:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_inline_3914_i3071; r2[2] = lenv;
    r1[3] = r2;
    // def inline-application_3917_i3071
    goto L1310;
    FUN_inline_application_3917_i3071:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 5);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) constructed_802[0];
      r0[2] = r1;
      r1 = (object *) constructed_801[0];
      r0[3] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object**) lenv) [1][3];
      r1[2] = r2;
      r2 = top[22];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1311; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
      L1311:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[4] = r1;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[5] = r1;
      r0 = ((object*) lenv) [5];
      switch (get_case (r0)) {
        case (UOTAG(7)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][5];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][5];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1312; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L1312:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r2 = ((object***) lenv) [1][1][4];
          r1 = PXLL_TEST(unbox(r2)==unbox(r1));
          r0[3] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [3];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_791;
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][2];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object***) lenv) [1][1][4];
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def $loop_4005_i3071
            goto L1313;
            FUN__24loop_4005_i3071:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 4);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*****) lenv) [1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1314; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1314:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r1[3] = r2;
                r2 = top[5];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1315; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
                L1315:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1316; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1316:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[4] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = ((pxll_vector*)r1)->val[4];
                r0[2] = r1;
                r0 = (object *) 1;
                r1 = ((object*) lenv) [2];
                r0 = PXLL_TEST(unbox(r1)>unbox(r0));
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = ((object***) lenv) [1][1][2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  ((object******) lenv) [1][1][1][1][1][3] = r0;
                } else {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][4];
                  r1 = ((pxll_vector*)r1)->val[4];
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UOTAG(2)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      r1 = ((object****) lenv) [1][1][1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r1;
                      t[2] = r0;
                      r0 = t;
                      lenv = lenv[1];
                      ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
                    } break;
                    case (UOTAG(0)): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
                      r1[2] = r2;
                      r2 = ((object*) lenv) [2];
                      r1[3] = r2;
                      r2 = top[5];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1317; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
                      L1317:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][3];
                      r1 = ((pxll_vector*)r1)->val[4];
                      r0[2] = r1;
                      r0 = (object *) 1;
                      r1 = ((object*) lenv) [2];
                      r0 = PXLL_TEST(unbox(r1)>unbox(r0));
                      lenv = lenv[1];
                      if PXLL_IS_TRUE(r0) {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                        r0[2] = r1;
                        r0 = ((object*) lenv) [2];
                        r1 = ((object*****) lenv) [1][1][1][1][2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        lenv = lenv[1];
                        ((object********) lenv) [1][1][1][1][1][1][1][3] = r0;
                      } else {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                        r0[2] = r1;
                        r0 = ((object*) lenv) [2];
                        r1 = ((object*****) lenv) [1][1][1][1][2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        lenv = lenv[1];
                        ((object********) lenv) [1][1][1][1][1][1][1][2] = r0;
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      goto fatbar_184;
                    }
                  }
                  goto fatbar_184_over;
                  fatbar_184:
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object***) lenv) [1][1][3];
                  r1 = ((pxll_vector*)r1)->val[3];
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = (object *) 3;
                  r0 = PXLL_TEST(unbox(r1)==unbox(r0));
                  lenv = lenv[1];
                  if PXLL_IS_TRUE(r0) {
                    r0 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                    switch (get_case (r0)) {
                      case (TC_PAIR): {
                        r0 = allocate (TC_TUPLE, 4);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        r1 = ((pxll_vector*)r1)->val[4];
                        r0[3] = r1;
                        r1 = ((object*) lenv) [2];
                        r1 = ((pxll_vector*)r1)->val[3];
                        r0[4] = r1;
                        r0 = ((object*) lenv) [3];
                        switch (get_case (r0)) {
                          case (UOTAG(12)): {
                            r0 = allocate (TC_TUPLE, 2);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object**) lenv) [1][3];
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r0 = (object *) constructed_314[0];
                            r1 = ((object*) lenv) [2];
                            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                              r0 = ((object**) lenv) [1][4];
                              switch (get_case (r0)) {
                                case (TC_PAIR): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object***) lenv) [1][1][4];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r1 = ((pxll_vector*)r1)->val[4];
                                  r0[3] = r1;
                                  r0 = ((object*) lenv) [3];
                                  switch (get_case (r0)) {
                                    case (UOTAG(0)): {
                                      r0 = allocate (TC_TUPLE, 4);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][3];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = ((object***************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][3];
                                      r1[2] = r2;
                                      r2 = ((object*) lenv) [2];
                                      r1[3] = r2;
                                      r2 = top[5];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1318; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
                                      L1318:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[3] = r1;
                                      r1 = ((object*) lenv) [3];
                                      r1 = ((pxll_vector*)r1)->val[4];
                                      r0[4] = r1;
                                      r0 = ((object*) lenv) [4];
                                      r1 = (object *) 1;
                                      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
                                      lenv = lenv[1];
                                    } break;
                                    default: {
                                      lenv = ((object ***)lenv)[1][1][1];
                                      goto fatbar_185;
                                    }
                                  }
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object **)lenv)[1][1];
                                  goto fatbar_185;
                                }
                              }
                            } else {
                              lenv = ((object **)lenv)[1][1];
                              goto fatbar_185;
                            }
                            lenv = lenv[1];
                          } break;
                          default: {
                            lenv = ((object *)lenv)[1];
                            goto fatbar_185;
                          }
                        }
                        lenv = lenv[1];
                      } break;
                      default: {
                        goto fatbar_185;
                      }
                    }
                    goto fatbar_185_over;
                    fatbar_185:
                    r0 = (object *) 6;
                    fatbar_185_over: ;
                  } else {
                    r0 = (object *) 6;
                  }
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                    r0[2] = r1;
                    r0 = ((object*) lenv) [2];
                    r1 = ((object****) lenv) [1][1][1][2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r1;
                    t[2] = r0;
                    r0 = t;
                    lenv = lenv[1];
                    ((object*******) lenv) [1][1][1][1][1][1][2] = r0;
                  } else {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                    r0[2] = r1;
                    r0 = ((object*) lenv) [2];
                    r1 = ((object****) lenv) [1][1][1][2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r1;
                    t[2] = r0;
                    r0 = t;
                    lenv = lenv[1];
                    ((object*******) lenv) [1][1][1][1][1][1][3] = r0;
                  }
                  fatbar_184_over: ;
                  lenv = lenv[1];
                }
                lenv = lenv[1];
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_4005_i3071;
              }
              PXLL_RETURN(0);
            L1313:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN__24loop_4005_i3071; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = (object *) 1;
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1319; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN__24loop_4005_i3071;
            L1319:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 6);
            r0[1] = lenv; lenv = r0;
            r1 = (object *) constructed_799[0];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 4);
            r2[1] = lenv; lenv = r2;
            r3 = allocate (TC_TUPLE, 2);
            r4 = ((object*******) lenv) [1][1][1][1][1][1][3];
            r4 = ((pxll_vector*)r4)->val[1];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L1320; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            lenv = r4[2]; goto *r4[1];
            L1320:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = top[56];
            t = allocate (TC_SAVE, 6);
            t[1] = k; t[2] = lenv; t[3] = &&L1321; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
            r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
            L1321:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
            r3 = result;
            r2[2] = r3;
            r3 = (object*)TC_NIL;
            r2[3] = r3;
            r3 = ((object*) lenv) [3];
            r4 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            r2[4] = r3;
            r2 = ((object*) lenv) [4];
            r3 = (object*) &constructed_798;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[47];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1322; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
            L1322:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][2];
            r2 = ((pxll_vector*)r2)->val[3];
            r1[2] = r2;
            r2 = (object *) 1;
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1323; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1323:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            // def append-suffix_3969_i2986_i3071
            goto L1324;
            FUN_append_suffix_3969_i2986_i3071:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 4);
              r2[1] = lenv; lenv = r2;
              r3 = ((object**) lenv) [1][2];
              r3 = allocate (TC_TUPLE, 2);
              r3[1] = lenv; lenv = r3;
              r4 = ((object***) lenv) [1][1][2];
              r4 = UOBJ_GET(r4,0);
              r3[2] = r4;
              r3 = ((object*) lenv) [2];
              lenv = lenv[1];
              r2[2] = r3;
              r3 = (object*)TC_NIL;
              r2[3] = r3;
              r3 = ((object*) lenv) [3];
              r4 = ((object****) lenv) [1][1][1][3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[4] = r3;
              r2 = ((object*) lenv) [4];
              r3 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[2] = r2;
              r2 = top[47];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1325; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
              L1325:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = top[80];
              r0[1] = r1[2]; lenv = r0; goto FUN_string__3esymbol_419;
            L1324:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_append_suffix_3969_i2986_i3071; r2[2] = lenv;
            r1[2] = r2;
            // def rename_3970_i2986_i3071
            goto L1326;
            FUN_rename_3970_i2986_i3071:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              // def get-new-name_3974_i2986_i3071
              goto L1327;
              FUN_get_new_name_3974_i2986_i3071:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object**) lenv) [1][2];
                r1[2] = r2;
                r2 = ((object*****) lenv) [1][1][1][1][2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1328; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_append_suffix_3969_i2986_i3071;
                L1328:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object******) lenv) [1][1][1][1][1][2] = r0;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*****) lenv) [1][1][1][1][3];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object***) lenv) [1][1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object****) lenv) [1][1][1][3] = r0;
                r0 = ((object*) lenv) [2];
                PXLL_RETURN(0);
              L1327:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_get_new_name_3974_i2986_i3071; r1[2] = lenv;
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object***) lenv) [1][1][2];
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r1 = ((pxll_vector*)r1)->val[4];
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (UOTAG(2)): {
                  r0 = ((object*) lenv) [2];
                } break;
                default: {
                  goto fatbar_183;
                }
              }
              goto fatbar_183_over;
              fatbar_183:
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object**) lenv) [1][2];
              r2 = ((pxll_vector*)r2)->val[4];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r2 = ((pxll_vector*)r2)->val[3];
              r1[3] = r2;
              r2 = top[118];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1329; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
              L1329:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = ((object**) lenv) [1][2];
              r0 = ((pxll_vector*)r0)->val[0];
              r1 = ((object*) lenv) [2];
              ((pxll_vector*)r1)->val[0] = r0;
              r0 = ((object**) lenv) [1][2];
              r0 = ((pxll_vector*)r0)->val[5];
              r1 = ((object*) lenv) [2];
              ((pxll_vector*)r1)->val[5] = r0;
              r0 = ((object*) lenv) [2];
              lenv = lenv[1];
              fatbar_183_over: ;
              lenv = lenv[1];
              ((object**) lenv) [1][2] = r0;
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object***) lenv) [1][1][2];
              r1 = ((pxll_vector*)r1)->val[4];
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(9)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object****) lenv) [1][1][1][2];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1330; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L1330:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  t = alloc_no_clear (UOTAG(9),1);
                  t[1] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  r1 = ((object****) lenv) [1][1][1][2];
                  ((pxll_vector*)r1)->val[4] = r0;
                  lenv = lenv[1];
                } break;
                case (UOTAG(10)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object****) lenv) [1][1][1][2];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1331; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L1331:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  t = alloc_no_clear (UOTAG(10),1);
                  t[1] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  r1 = ((object****) lenv) [1][1][1][2];
                  ((pxll_vector*)r1)->val[4] = r0;
                  lenv = lenv[1];
                } break;
                case (UOTAG(7)): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object**) lenv) [1][2];
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1332; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_get_new_name_3974_i2986_i3071;
                  L1332:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object****) lenv) [1][1][1][2];
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][3];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1333; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L1333:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r0 = ((object*) lenv) [3];
                  r1 = ((object*) lenv) [2];
                  t = alloc_no_clear (UOTAG(7),2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  r1 = ((object****) lenv) [1][1][1][2];
                  ((pxll_vector*)r1)->val[4] = r0;
                  lenv = lenv[1];
                } break;
                case (UOTAG(0)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = ((object*) lenv) [2];
                  r0[2] = r1;
                  r1 = ((object****) lenv) [1][1][1][3];
                  r0[3] = r1;
                  r1 = top[25];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1334; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
                  L1334:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][2];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1335; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_append_suffix_3969_i2986_i3071;
                    L1335:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r0 = ((object*) lenv) [2];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    lenv = lenv[1];
                    r1 = ((object****) lenv) [1][1][1][2];
                    ((pxll_vector*)r1)->val[4] = r0;
                  } else {
                    r0 = (object *) 14;
                  }
                  lenv = lenv[1];
                } break;
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = ((object*) lenv) [2];
                  r0[2] = r1;
                  r1 = ((object****) lenv) [1][1][1][3];
                  r0[3] = r1;
                  r1 = top[25];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1336; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
                  L1336:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][2];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1337; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_append_suffix_3969_i2986_i3071;
                    L1337:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r0 = ((object*) lenv) [2];
                    t = alloc_no_clear (UOTAG(1),1);
                    t[1] = r0;
                    r0 = t;
                    lenv = lenv[1];
                    r1 = ((object****) lenv) [1][1][1][2];
                    ((pxll_vector*)r1)->val[4] = r0;
                  } else {
                    r0 = (object *) 14;
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_182;
                }
              }
              goto fatbar_182_over;
              fatbar_182:
              r0 = (object *) 14;
              fatbar_182_over: ;
              lenv = lenv[1];
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_26064_i2986_i3071
              goto L1338;
              FUN_lambda_26064_i2986_i3071:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object*) lenv) [2];
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][3];
                r0[3] = r1;
                r1 = ((object****) lenv) [1][1][1][3];
                r0[1] = r1[2]; lenv = r0; goto FUN_rename_3970_i2986_i3071;
              L1338:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_26064_i2986_i3071; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = ((pxll_vector*)r1)->val[3];
              r0[3] = r1;
              r1 = top[32];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L1339; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_map_326;
              L1339:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              r1 = ((object**) lenv) [1][2];
              ((pxll_vector*)r1)->val[3] = r0;
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            L1326:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_rename_3970_i2986_i3071; r2[2] = lenv;
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][4];
            r1[2] = r2;
            r2 = (object *) constructed_797[0];
            r1[3] = r2;
            r2 = ((object*) lenv) [3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1340; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_rename_3970_i2986_i3071;
            L1340:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r0[5] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object****) lenv) [1][1][1][2];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            switch (get_case (r2)) {
              case (TC_NIL): {
                r2 = (object *) 262;
              } break;
              default: {
                goto fatbar_181;
              }
            }
            goto fatbar_181_over;
            fatbar_181:
            r2 = (object *) 6;
            fatbar_181_over: ;
            r1[3] = r2;
            r1 = (object *) 6;
            r2 = ((object*) lenv) [3];
            r1 = PXLL_TEST(r2==r1);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 3);
              // def lambda_26297_i3071
              goto L1341;
              FUN_lambda_26297_i3071:
                check_heap (0);
                r0 = allocate (UOTAG(1), 2);
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object***) lenv) [1][1][3];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1342; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1342:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[1] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*****) lenv) [1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1343; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1343:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                PXLL_RETURN(0);
              L1341:
              r2 = allocate (TC_CLOSURE, 2);
              r2[1] = &&FUN_lambda_26297_i3071; r2[2] = lenv;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[32];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1344; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
              L1344:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) constructed_796[0];
            }
            r0[6] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object****) lenv) [1][1][1][3];
            r0[2] = r1;
            r1 = (object*)TC_NIL;
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object*) lenv) [5];
              r0[2] = r1;
              r1 = ((object*) lenv) [6];
              r0[3] = r1;
              r1 = ((object*****) lenv) [1][1][1][1][5];
              r0[1] = r1[2]; lenv = r0; goto FUN_substitute_3918_i3071;
            } else {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = (object *) constructed_795[0];
              r0[2] = r1;
              r1 = (object *) constructed_794[0];
              r0[3] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object****) lenv) [1][1][1][3];
              r1[2] = r2;
              r2 = top[22];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1345; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
              L1345:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[4] = r1;
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_26366_i3071
              goto L1346;
              FUN_lambda_26366_i3071:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object*****) lenv) [1][1][1][1][3];
                r2[2] = r3;
                r3 = ((object**) lenv) [1][2];
                r2[3] = r3;
                r3 = top[27];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1347; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                L1347:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 4);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
                r5 = ((pxll_vector*)r5)->val[1];
                t = allocate (TC_SAVE, 8);
                t[1] = k; t[2] = lenv; t[3] = &&L1348; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                lenv = r5[2]; goto *r5[1];
                L1348:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                r5 = result;
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L1349; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L1349:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = (object*)TC_NIL;
                r3[3] = r4;
                r4 = ((object*) lenv) [3];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[4] = r4;
                r3 = ((object*) lenv) [4];
                r4 = (object*) &constructed_793;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1350; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L1350:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[3] = r2;
                r2 = top[143];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1351; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_symbol_add_suffix_548;
                L1351:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object****) lenv) [1][1][1][2];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object***) lenv) [1][1][2] = r0;
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object***) lenv) [1][1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1352; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1352:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = ((object****) lenv) [1][1][1][3];
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object***) lenv) [1][1][3] = r0;
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (UOTAG(1), 2);
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object******) lenv) [1][1][1][1][1][3];
                r2[2] = r3;
                r3 = ((object***) lenv) [1][1][2];
                r2[3] = r3;
                r3 = top[27];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1353; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                L1353:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[1] = r2;
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object**) lenv) [1][2];
                t = alloc_no_clear (UOTAG(0),1);
                t[1] = r3;
                r3 = t;
                r2[2] = r3;
                r3 = (object *) constructed_792[0];
                r2[3] = r3;
                r3 = top[118];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1354; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_make_node_486;
                L1354:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r0[2] = r1;
                r1 = ((object*****) lenv) [1][1][1][1][6];
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object****) lenv) [1][1][1][6] = r0;
                PXLL_RETURN(0);
              L1346:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_26366_i3071; r1[2] = lenv;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object*****) lenv) [1][1][1][1][3];
              r1[2] = r2;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*) lenv) [2];
              r1[2] = r2;
              r2 = (object *) constructed_36[0];
              r1[3] = r2;
              r2 = top[21];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1355; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
              L1355:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              lenv = lenv[1];
              r0[3] = r1;
              r1 = top[35];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L1356; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
              L1356:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object***) lenv) [1][1][5];
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][6];
              r1[3] = r2;
              r2 = ((object*******) lenv) [1][1][1][1][1][1][5];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1357; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_substitute_3918_i3071;
              L1357:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 4);
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object***) lenv) [1][1][2];
              r1[2] = r2;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*) lenv) [2];
              r1[2] = r2;
              r2 = (object *) constructed_36[0];
              r1[3] = r2;
              r2 = top[21];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1358; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
              L1358:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              lenv = lenv[1];
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object***) lenv) [1][1][3];
              r1[2] = r2;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*) lenv) [2];
              r1[2] = r2;
              r2 = (object *) constructed_36[0];
              r1[3] = r2;
              r2 = top[21];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1359; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
              L1359:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              lenv = lenv[1];
              r0[3] = r1;
              r1 = ((object*) lenv) [2];
              r0[4] = r1;
              r1 = top[122];
              r0[1] = r1[2]; lenv = r0; goto FUN_node_2flet_499;
            }
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_180;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_180_over;
      fatbar_180:
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_800;
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      fatbar_180_over: ;
      PXLL_RETURN(0);
    L1310:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_inline_application_3917_i3071; r2[2] = lenv;
    r1[4] = r2;
    // def substitute_3918_i3071
    goto L1360;
    FUN_substitute_3918_i3071:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      // def lookup_4024_i3071
      goto L1361;
      FUN_lookup_4024_i3071:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        // def loop_4027_i3071
        goto L1362;
        FUN_loop_4027_i3071:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = (object*)UITAG(1);
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case_noint (r0)) {
                default: {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = ((object*****) lenv) [1][1][1][1][2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = ((object*) lenv) [3];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } else {
                    r0 = ((object**) lenv) [1][3];
                    lenv = ((object **)lenv)[1][1];
                    lenv[2] = r0;
                    goto FUN_loop_4027_i3071;
                  }
                  PXLL_RETURN(0);
                } break;
              }
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        L1362:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_loop_4027_i3071; r1[2] = lenv;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object****) lenv) [1][1][1][3];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[1] = r1[2]; lenv = r0; goto FUN_loop_4027_i3071;
      L1361:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lookup_4024_i3071; r1[2] = lenv;
      r0[2] = r1;
      // def walk_4025_i3071
      goto L1363;
      FUN_walk_4025_i3071:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = ((pxll_vector*)r1)->val[4];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        switch (get_case (r1)) {
          case (UOTAG(0)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1364; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_lookup_4024_i3071;
            L1364:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r1 = ((object*) lenv) [3];
            switch (get_case (r1)) {
              case (UOTAG(0)): {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object**) lenv) [1][3];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                lenv = lenv[1];
              } break;
              default: {
                r1 = ((object***) lenv) [1][1][2];
              } break;
            }
            lenv = lenv[1];
          } break;
          case (UOTAG(1)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1365; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_lookup_4024_i3071;
            L1365:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r1 = ((object*) lenv) [3];
            switch (get_case (r1)) {
              case (UOTAG(0)): {
                r1 = allocate (TC_TUPLE, 4);
                r1[1] = lenv; lenv = r1;
                r2 = ((object**) lenv) [1][3];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object*) lenv) [2];
                r3 = ((pxll_vector*)r3)->val[lookup_field((GET_TYPECODE(*r3)-TC_USEROBJ)>>2,51)];
                r2[2] = r3;
                r3 = top[119];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1366; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_varref__3ename_489;
                L1366:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[3] = r2;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object****) lenv) [1][1][1][2];
                r3 = ((pxll_vector*)r3)->val[3];
                r2[2] = r3;
                r3 = top[19];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1367; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                L1367:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[4] = r2;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*) lenv) [3];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r2;
                r2 = t;
                r1[2] = r2;
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = (object*)TC_NIL;
                r2[2] = r3;
                r2 = ((object*) lenv) [2];
                r3 = ((object**) lenv) [1][4];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[3] = r2;
                r2 = top[118];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1368; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
                L1368:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                lenv = lenv[1];
              } break;
              default: {
                r1 = ((object***) lenv) [1][1][2];
              } break;
            }
            lenv = lenv[1];
          } break;
          default: {
            goto fatbar_179;
          }
        }
        goto fatbar_179_over;
        fatbar_179:
        r1 = ((object**) lenv) [1][2];
        fatbar_179_over: ;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object***) lenv) [1][1][3];
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[3] = r1;
        r1 = top[32];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1369; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_map_326;
        L1369:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        r1 = ((object*) lenv) [3];
        ((pxll_vector*)r1)->val[3] = r0;
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      L1363:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_walk_4025_i3071; r1[2] = lenv;
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object**) lenv) [1][2];
      r0[2] = r1;
      r1 = ((object*) lenv) [3];
      r0[1] = r1[2]; lenv = r0; goto FUN_walk_4025_i3071;
    L1360:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_substitute_3918_i3071; r2[2] = lenv;
    r1[5] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = (object*)UITAG(0);
    r1[3] = r2;
    r2 = ((object*) lenv) [3];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1370; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_inline_3914_i3071;
    L1370:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    lenv = lenv[1];
    r0[2] = r1;
    r1 = top[144];
    r0[1] = r1[2]; lenv = r0; goto FUN_do_simple_optimizations_551;
  L1206:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_do_one_round_555; r1[2] = lenv;
  r0[146] = r1;
  // def encode-immediate_558
  goto L1371;
  FUN_encode_immediate_558:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = (object *) 3;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)<<unbox(r1));
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = (object *) 3;
        r0 = box((pxll_int)unbox(r1)|unbox(r0));
        PXLL_RETURN(0);
      } break;
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r1 = box((pxll_int)GET_CHAR(r1));
        r0[3] = r1;
        r1 = (object *) 17;
        r2 = ((object*) lenv) [3];
        r1 = box((pxll_int)unbox(r2)<<unbox(r1));
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        r1 = (object *) 5;
        r0 = box((pxll_int)unbox(r1)|unbox(r0));
        PXLL_RETURN(0);
      } break;
      case (UITAG(3)): {
        r0 = (object *) 29;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) constructed_46[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_41[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object *) 13;
            PXLL_RETURN(0);
          } else {
            goto fatbar_178;
          }
          PXLL_RETURN(0);
          goto fatbar_178_over;
          fatbar_178:
          r0 = (object *) constructed_44[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object *) 525;
            PXLL_RETURN(0);
          } else {
            lenv = ((object *)lenv)[1];
            goto fatbar_177;
          }
          PXLL_RETURN(0);
          fatbar_178_over: ;
          PXLL_RETURN(0);
        } else {
          lenv = ((object *)lenv)[1];
          goto fatbar_177;
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_177;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_177_over;
    fatbar_177:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_790;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_177_over: ;
    PXLL_RETURN(0);
  L1371:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_encode_immediate_558; r1[2] = lenv;
  r0[147] = r1;
  // def wrap-in_559
  goto L1372;
  FUN_wrap_in_559:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) constructed_516[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = (object *) constructed_86[0];
                  r1 = ((object*) lenv) [2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = ((object**) lenv) [1][3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = allocate (TC_TUPLE, 4);
                        r1[1] = lenv; lenv = r1;
                        r2 = (object*)TC_NIL;
                        r1[2] = r2;
                        r2 = ((object*) lenv) [2];
                        r3 = (object*) &constructed_777;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r3;
                        t[2] = r2;
                        r2 = t;
                        r1[3] = r2;
                        r2 = ((object*) lenv) [3];
                        r3 = ((object*****) lenv) [1][1][1][1][3];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r3;
                        t[2] = r2;
                        r2 = t;
                        r1[4] = r2;
                        r1 = ((object*) lenv) [4];
                        r2 = (object*) &constructed_776;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r2;
                        t[2] = r1;
                        r1 = t;
                        lenv = lenv[1];
                        r0[2] = r1;
                        r1 = top[47];
                        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                      } break;
                      default: {
                        lenv = ((object **)lenv)[1][1];
                        goto fatbar_166;
                      }
                    }
                    PXLL_RETURN(0);
                  } else {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_166;
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_166;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_166;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_166_over;
          fatbar_166:
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object*) &constructed_778;
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[3] = r1;
          r1 = top[15];
          r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
          fatbar_166_over: ;
          PXLL_RETURN(0);
        } else {
          goto fatbar_165;
        }
        PXLL_RETURN(0);
        goto fatbar_165_over;
        fatbar_165:
        r0 = (object *) constructed_538[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        } else {
          goto fatbar_167;
        }
        PXLL_RETURN(0);
        goto fatbar_167_over;
        fatbar_167:
        r0 = (object *) constructed_734[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        } else {
          goto fatbar_168;
        }
        PXLL_RETURN(0);
        goto fatbar_168_over;
        fatbar_168:
        r0 = (object *) constructed_58[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        } else {
          goto fatbar_169;
        }
        PXLL_RETURN(0);
        goto fatbar_169_over;
        fatbar_169:
        r0 = (object *) constructed_472[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        } else {
          goto fatbar_170;
        }
        PXLL_RETURN(0);
        goto fatbar_170_over;
        fatbar_170:
        r0 = (object *) constructed_97[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        } else {
          goto fatbar_171;
        }
        PXLL_RETURN(0);
        goto fatbar_171_over;
        fatbar_171:
        r0 = (object *) constructed_760[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        } else {
          goto fatbar_172;
        }
        PXLL_RETURN(0);
        goto fatbar_172_over;
        fatbar_172:
        r0 = (object *) constructed_481[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 7);
          r1[1] = lenv; lenv = r1;
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object***) lenv) [1][1][2];
          r2[2] = r3;
          r3 = top[149];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1373; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_irken_type__3ec_type_560;
          L1373:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r2 = (object*)TC_NIL;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = (object*) &constructed_781;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r2 = ((object*) lenv) [4];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[5] = r2;
          r2 = ((object*) lenv) [5];
          r3 = (object*) &constructed_780;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[6] = r2;
          r2 = ((object*) lenv) [6];
          r3 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[7] = r2;
          r1 = ((object*) lenv) [7];
          r2 = (object*) &constructed_779;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_173;
        }
        PXLL_RETURN(0);
        goto fatbar_173_over;
        fatbar_173:
        r0 = (object *) constructed_763[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 3);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r1 = ((object*) lenv) [3];
          r2 = (object*) &constructed_782;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_174;
        }
        PXLL_RETURN(0);
        goto fatbar_174_over;
        fatbar_174:
        r0 = (object *) constructed_86[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = (object*) &constructed_784;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          r2 = (object*) &constructed_783;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_175;
        }
        PXLL_RETURN(0);
        goto fatbar_175_over;
        fatbar_175:
        r0 = (object *) constructed_226[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = (object*) &constructed_786;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          r2 = (object*) &constructed_785;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_176;
        }
        PXLL_RETURN(0);
        goto fatbar_176_over;
        fatbar_176:
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[203];
        r0[3] = r1;
        r1 = top[25];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1374; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
        L1374:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = (object*) &constructed_788;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          r2 = (object*) &constructed_787;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object*) &constructed_789;
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[3] = r1;
          r1 = top[15];
          r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        }
        PXLL_RETURN(0);
        fatbar_176_over: ;
        PXLL_RETURN(0);
        fatbar_175_over: ;
        PXLL_RETURN(0);
        fatbar_174_over: ;
        PXLL_RETURN(0);
        fatbar_173_over: ;
        PXLL_RETURN(0);
        fatbar_172_over: ;
        PXLL_RETURN(0);
        fatbar_171_over: ;
        PXLL_RETURN(0);
        fatbar_170_over: ;
        PXLL_RETURN(0);
        fatbar_169_over: ;
        PXLL_RETURN(0);
        fatbar_168_over: ;
        PXLL_RETURN(0);
        fatbar_167_over: ;
        PXLL_RETURN(0);
        fatbar_165_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1372:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_wrap_in_559; r1[2] = lenv;
  r0[148] = r1;
  // def irken-type->c-type_560
  goto L1375;
  FUN_irken_type__3ec_type_560:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) constructed_772[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 4);
                  r1[1] = lenv; lenv = r1;
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object**) lenv) [1][2];
                  r2[2] = r3;
                  r3 = top[149];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L1376; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_irken_type__3ec_type_560;
                  L1376:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = (object*)TC_NIL;
                  r1[3] = r2;
                  r2 = ((object*) lenv) [3];
                  r3 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r3;
                  t[2] = r2;
                  r2 = t;
                  r1[4] = r2;
                  r1 = ((object*) lenv) [4];
                  r2 = (object*) &constructed_770;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  lenv = lenv[1];
                  r0[2] = r1;
                  r1 = top[47];
                  r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_163;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_163;
            }
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_163;
        }
        PXLL_RETURN(0);
        goto fatbar_163_over;
        fatbar_163:
        r0 = (object *) constructed_481[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = allocate (TC_TUPLE, 5);
                  r1[1] = lenv; lenv = r1;
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object**) lenv) [1][2];
                  r2[2] = r3;
                  r3 = top[149];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L1377; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_irken_type__3ec_type_560;
                  L1377:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = (object*)TC_NIL;
                  r1[3] = r2;
                  r2 = ((object*) lenv) [3];
                  r3 = (object*) &constructed_774;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r3;
                  t[2] = r2;
                  r2 = t;
                  r1[4] = r2;
                  r2 = ((object*) lenv) [4];
                  r3 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r3;
                  t[2] = r2;
                  r2 = t;
                  r1[5] = r2;
                  r1 = ((object*) lenv) [5];
                  r2 = (object*) &constructed_773;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  lenv = lenv[1];
                  r0[2] = r1;
                  r1 = top[47];
                  r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_164;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_164;
            }
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_164;
        }
        PXLL_RETURN(0);
        goto fatbar_164_over;
        fatbar_164:
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = ((object*) lenv) [2];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_162;
          }
        }
        PXLL_RETURN(0);
        fatbar_164_over: ;
        PXLL_RETURN(0);
        fatbar_163_over: ;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_162;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_162_over;
    fatbar_162:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_775;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = top[107];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1378; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_type_repr_462;
    L1378:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_162_over: ;
    PXLL_RETURN(0);
  L1375:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_irken_type__3ec_type_560; r1[2] = lenv;
  r0[149] = r1;
  // def wrap-out_561
  goto L1379;
  FUN_wrap_out_561:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = (object *) constructed_760[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 3);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r1 = ((object*) lenv) [3];
          r2 = (object*) &constructed_758;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_158;
        }
        PXLL_RETURN(0);
        goto fatbar_158_over;
        fatbar_158:
        r0 = (object *) constructed_763[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 3);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r1 = ((object*) lenv) [3];
          r2 = (object*) &constructed_761;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_159;
        }
        PXLL_RETURN(0);
        goto fatbar_159_over;
        fatbar_159:
        r0 = (object *) constructed_46[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = (object*) &constructed_765;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          r2 = (object*) &constructed_764;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_160;
        }
        PXLL_RETURN(0);
        goto fatbar_160_over;
        fatbar_160:
        r0 = (object *) constructed_226[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = (object*) &constructed_767;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          r2 = (object*) &constructed_766;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          goto fatbar_161;
        }
        PXLL_RETURN(0);
        goto fatbar_161_over;
        fatbar_161:
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = top[203];
        r0[3] = r1;
        r1 = top[25];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1380; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
        L1380:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = (object*) &constructed_769;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          r2 = (object*) &constructed_768;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } else {
          r0 = ((object**) lenv) [1][3];
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
        fatbar_161_over: ;
        PXLL_RETURN(0);
        fatbar_160_over: ;
        PXLL_RETURN(0);
        fatbar_159_over: ;
        PXLL_RETURN(0);
        fatbar_158_over: ;
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_157;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_157_over;
    fatbar_157:
    r0 = ((object*) lenv) [3];
    PXLL_RETURN(0);
    fatbar_157_over: ;
    PXLL_RETURN(0);
  L1379:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_wrap_out_561; r1[2] = lenv;
  r0[150] = r1;
  // def cexp-subst_562
  goto L1381;
  FUN_cexp_subst_562:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = (object *) 9474;
    r1[3] = r2;
    r2 = top[49];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1382; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_split_354;
    L1382:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_4213
    goto L1383;
    FUN_loop_4213:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1384; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1384:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = top[47];
          r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = (object*) &constructed_756;
          r1[3] = r2;
          r2 = top[50];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1385; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
          L1385:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 1;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)==unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (TC_NIL): {
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object*) &constructed_754;
                r0[2] = r1;
                r1 = ((object*****) lenv) [1][1][1][1][2];
                r0[3] = r1;
                r1 = top[15];
                r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object****) lenv) [1][1][1][2];
                r2 = (object*) &constructed_755;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                r1 = ((object*) lenv) [3];
                lenv = ((object **)lenv)[1][1];
                lenv[2] = r0;
                lenv[3] = r1;
                goto FUN_loop_4213;
              } break;
            }
            PXLL_RETURN(0);
          } else {
            goto fatbar_156;
          }
          PXLL_RETURN(0);
          goto fatbar_156_over;
          fatbar_156:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = top[169];
          r1[2] = r2;
          r2 = (object *) 1;
          r3 = ((object**) lenv) [1][2];
          range_check (((pxll_string *)(((pxll_string*)(r3))))->len, unbox(r2));
          r2 = (object *) 1;
          r3 = ((object**) lenv) [1][2];
          r2 = TO_CHAR(((unsigned char *)((pxll_string*)(r3))->data)[unbox(r2)]);
          r1[3] = r2;
          r2 = top[65];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1386; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
          L1386:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UITAG(1)): {
              r0 = allocate (TC_TUPLE, 3);
              r1 = (object*) &constructed_757;
              r0[2] = r1;
              r1 = ((object******) lenv) [1][1][1][1][1][2];
              r0[3] = r1;
              r1 = top[15];
              r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 4);
              r2 = ((object****) lenv) [1][1][1][2];
              r1[2] = r2;
              r2 = (object *) 3;
              r1[3] = r2;
              r2 = ((object****) lenv) [1][1][1][2];
              r2 = box((pxll_int)((pxll_string*)(r2))->len);
              r1[4] = r2;
              r2 = top[45];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1387; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_substring_345;
              L1387:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1388; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1388:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = ((object*****) lenv) [1][1][1][1][2];
              r2 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[4] = r1;
              r0 = ((object*) lenv) [4];
              r1 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              r1 = ((object***) lenv) [1][1][3];
              lenv = ((object ***)lenv)[1][1][1];
              lenv[2] = r0;
              lenv[3] = r1;
              goto FUN_loop_4213;
            } break;
          }
          PXLL_RETURN(0);
          fatbar_156_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1383:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_4213; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = allocate (TC_TUPLE, 3);
    r1[1] = lenv; lenv = r1;
    r2 = allocate (TC_TUPLE, 2);
    r3 = ((object***) lenv) [1][1][2];
    r2[2] = r3;
    r3 = top[19];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1389; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
    L1389:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = (object*)TC_NIL;
    r1[3] = r2;
    r1 = ((object*) lenv) [3];
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    lenv = lenv[1];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[20];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1390; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
    L1390:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_4213;
  L1381:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_cexp_subst_562; r1[2] = lenv;
  r0[151] = r1;
  // def getenv-or_572
  goto L1391;
  FUN_getenv_or_572:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[53];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1392; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_zero_terminate_360;
    L1392:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r1 = (object*)getenv (((pxll_string*)(r1))->data);
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST((char*)r1==NULL)) {
      r1 = (object*) &constructed_753;
    } else {
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = top[59];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1393; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_copy_cstring_372;
      L1393:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
    }
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][3];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = (object *) 1;
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = ((object**) lenv) [1][3];
      PXLL_RETURN(0);
    } else {
      r0 = ((object*) lenv) [3];
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1391:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_getenv_or_572; r1[2] = lenv;
  r0[152] = r1;
  r1 = (object *) 1;
  r0[153] = r1;
  r1 = (object *) 3;
  r0[154] = r1;
  r1 = (object *) 5;
  r0[155] = r1;
  r1 = (object *) 7;
  r0[156] = r1;
  r1 = (object *) 11;
  r0[157] = r1;
  r1 = top[10];
  r0[158] = r1;
  r1 = (object *) constructed_752[0];
  r0[159] = r1;
  r1 = box((pxll_int)O_RDONLY);
  r0[160] = r1;
  r1 = box((pxll_int)O_WRONLY);
  r0[161] = r1;
  r1 = box((pxll_int)O_CREAT);
  r0[162] = r1;
  r1 = box((pxll_int)O_TRUNC);
  r0[163] = r1;
  r1 = allocate (TC_TUPLE, 4);
  r1[1] = lenv; lenv = r1;
  r2 = box((pxll_int)argc);
  r1[2] = r2;
  r2 = ((object*) lenv) [2];
  ensure_heap (2, unbox(r2));
  r2 = (object*) &constructed_751;
  r3 = ((object*) lenv) [2];
  if (unbox(r3) == 0) { r2 = (object *) TC_EMPTY_VECTOR; } else {
    t = alloc_no_clear (TC_VECTOR, unbox(r3));
    for (i=0; i<unbox(r3); i++) { t[i+1] = r2; }
    r2 = t;
  }
  r1[3] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r2[1] = lenv; lenv = r2;
  // def loop_4759
  goto L1394;
  FUN_loop_4759:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(unbox(r0)==0)) {
      r0 = ((object***) lenv) [1][1][3];
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) 3;
      r2 = ((object**) lenv) [1][2];
      r1 = box((pxll_int)unbox(r2)-unbox(r1));
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [2];
      r1 = (object*)argv[unbox(r1)];
      r0[2] = r1;
      r1 = top[59];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1395; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_copy_cstring_372;
      L1395:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      r1 = (object *) 3;
      r2 = ((object*) lenv) [2];
      r1 = box((pxll_int)unbox(r2)-unbox(r1));
      r2 = ((object***) lenv) [1][1][3];
      range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
      ((pxll_vector*)r2)->val[unbox(r1)] = r0;
      r0 = (object *) 3;
      r1 = ((object*) lenv) [2];
      r0 = box((pxll_int)unbox(r1)-unbox(r0));
      lenv[2] = r0;
      goto FUN_loop_4759;
    }
    PXLL_RETURN(0);
  L1394:
  r3 = allocate (TC_CLOSURE, 2);
  r3[1] = &&FUN_loop_4759; r3[2] = lenv;
  r2[2] = r3;
  r2 = allocate (TC_TUPLE, 2);
  r3 = ((object**) lenv) [1][2];
  r2[2] = r3;
  r3 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1396; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_loop_4759;
  L1396:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  lenv = lenv[1];
  r1[4] = r2;
  r1 = allocate (UOTAG(18), 2);
  r2 = ((object*) lenv) [2];
  r1[1] = r2;
  r2 = ((object*) lenv) [4];
  r1[2] = r2;
  lenv = lenv[1];
  r0[164] = r1;
  r1 = allocate (UOTAG(17), 5);
  r2 = (object *) 16642;
  r2 = r2;
  r1[1] = r2;
  r2 = top[165];
  r1[2] = r2;
  r2 = (object *) 1;
  r1[3] = r2;
  r2 = top[165];
  r1[4] = r2;
  r2 = (object *) 16642;
  r2 = r2;
  r1[5] = r2;
  r0[165] = r1;
  r1 = (object *) 1;
  r0[166] = r1;
  r1 = top[165];
  r0[167] = r1;
  r1 = (object *) constructed_750[0];
  r0[168] = r1;
  r1 = (object *) constructed_749[0];
  r0[169] = r1;
  r1 = (object *) constructed_748[0];
  r0[170] = r1;
  r1 = (object *) constructed_747[0];
  r0[171] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object*) &constructed_746;
  r1[2] = r2;
  r2 = top[55];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1397; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string__3elist_362;
  L1397:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[172] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object*) &constructed_745;
  r1[2] = r2;
  r2 = top[55];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1398; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string__3elist_362;
  L1398:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[173] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = allocate (TC_TUPLE, 3);
  r3 = top[171];
  r2[2] = r3;
  r3 = (object *) constructed_36[0];
  r2[3] = r3;
  r3 = top[21];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1399; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
  L1399:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  r2 = top[172];
  r1[3] = r2;
  r2 = top[21];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1400; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
  L1400:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[174] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object*) &constructed_744;
  r1[2] = r2;
  r2 = top[55];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1401; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string__3elist_362;
  L1401:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[175] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) constructed_743[0];
  r1[2] = r2;
  r2 = top[83];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1402; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_char_class_423;
  L1402:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[176] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = top[174];
  r1[2] = r2;
  r2 = top[83];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1403; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_char_class_423;
  L1403:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[177] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = top[175];
  r1[2] = r2;
  r2 = top[83];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1404; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_char_class_423;
  L1404:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[178] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = allocate (TC_TUPLE, 2);
  r2[1] = lenv; lenv = r2;
  r3 = allocate (TC_TUPLE, 3);
  r4 = allocate (TC_TUPLE, 3);
  r5 = top[173];
  r4[2] = r5;
  r5 = (object *) constructed_36[0];
  r4[3] = r5;
  r5 = top[21];
  t = allocate (TC_SAVE, 7);
  t[1] = k; t[2] = lenv; t[3] = &&L1405; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
  r4[1] = r5[2]; lenv = r4; goto FUN_reverse_onto_311;
  L1405:
  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
  r4 = result;
  r3[2] = r4;
  r4 = top[175];
  r3[3] = r4;
  r4 = top[21];
  t = allocate (TC_SAVE, 6);
  t[1] = k; t[2] = lenv; t[3] = &&L1406; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
  r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
  L1406:
  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
  r3 = result;
  r2[2] = r3;
  r2 = ((object*) lenv) [2];
  r3 = (object *) 11522;
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r3;
  t[2] = r2;
  r2 = t;
  lenv = lenv[1];
  r1[2] = r2;
  r2 = top[83];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1407; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_char_class_423;
  L1407:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[179] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) 1;
  r1[2] = r2;
  r2 = top[113];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1408; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_make_counter_472;
  L1408:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[180] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_226[0];
  r1[2] = r2;
  r2 = (object *) constructed_742[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1409; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1409:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[181] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_734[0];
  r1[2] = r2;
  r2 = (object *) constructed_741[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1410; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1410:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[182] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_86[0];
  r1[2] = r2;
  r2 = (object *) constructed_740[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1411; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1411:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[183] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_736[0];
  r1[2] = r2;
  r2 = (object *) constructed_739[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1412; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1412:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[184] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_46[0];
  r1[2] = r2;
  r2 = (object *) constructed_738[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1413; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1413:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[185] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_58[0];
  r1[2] = r2;
  r2 = (object *) constructed_737[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1414; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1414:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[186] = r1;
  r1 = allocate (TC_TUPLE, 6);
  r1[1] = lenv; lenv = r1;
  r2 = (object*)UITAG(0);
  r1[2] = r2;
  r2 = top[186];
  r3 = ((object*) lenv) [2];
  r4 = (object *) constructed_58[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r4;
  t[2] = r2;
  t[3] = r3;
  r2 = t;
  r1[3] = r2;
  r2 = top[184];
  r3 = ((object*) lenv) [3];
  r4 = (object *) constructed_736[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r4;
  t[2] = r2;
  t[3] = r3;
  r2 = t;
  r1[4] = r2;
  r2 = top[183];
  r3 = ((object*) lenv) [4];
  r4 = (object *) constructed_86[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r4;
  t[2] = r2;
  t[3] = r3;
  r2 = t;
  r1[5] = r2;
  r2 = top[182];
  r3 = ((object*) lenv) [5];
  r4 = (object *) constructed_734[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r4;
  t[2] = r2;
  t[3] = r3;
  r2 = t;
  r1[6] = r2;
  r1 = top[181];
  r2 = ((object*) lenv) [6];
  r3 = (object *) constructed_226[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r3;
  t[2] = r1;
  t[3] = r2;
  r1 = t;
  lenv = lenv[1];
  r0[187] = r1;
  r1 = allocate (TC_TUPLE, 6);
  r1[1] = lenv; lenv = r1;
  r2 = (object *) constructed_503[0];
  t = alloc_no_clear (UOTAG(1),1);
  t[1] = r2;
  r2 = t;
  r1[2] = r2;
  r2 = (object *) 6;
  t = alloc_no_clear (UOTAG(4),1);
  t[1] = r2;
  r2 = t;
  r1[3] = r2;
  r2 = (object*)TC_NIL;
  r1[4] = r2;
  r2 = ((object*) lenv) [4];
  r3 = ((object*) lenv) [3];
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r3;
  t[2] = r2;
  r2 = t;
  r1[5] = r2;
  r2 = ((object*) lenv) [5];
  r3 = ((object*) lenv) [2];
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r3;
  t[2] = r2;
  r2 = t;
  r1[6] = r2;
  r1 = ((object*) lenv) [6];
  t = alloc_no_clear (UOTAG(0),1);
  t[1] = r1;
  r1 = t;
  lenv = lenv[1];
  r0[188] = r1;
  r1 = allocate (TC_TUPLE, 6);
  r1[1] = lenv; lenv = r1;
  r2 = (object *) constructed_501[0];
  t = alloc_no_clear (UOTAG(1),1);
  t[1] = r2;
  r2 = t;
  r1[2] = r2;
  r2 = (object *) 6;
  t = alloc_no_clear (UOTAG(4),1);
  t[1] = r2;
  r2 = t;
  r1[3] = r2;
  r2 = (object*)TC_NIL;
  r1[4] = r2;
  r2 = ((object*) lenv) [4];
  r3 = ((object*) lenv) [3];
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r3;
  t[2] = r2;
  r2 = t;
  r1[5] = r2;
  r2 = ((object*) lenv) [5];
  r3 = ((object*) lenv) [2];
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r3;
  t[2] = r2;
  r2 = t;
  r1[6] = r2;
  r1 = ((object*) lenv) [6];
  t = alloc_no_clear (UOTAG(0),1);
  t[1] = r1;
  r1 = t;
  lenv = lenv[1];
  r0[189] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) 1;
  r1[2] = r2;
  r2 = top[113];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1415; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_make_counter_472;
  L1415:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[190] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) 1;
  r1[2] = r2;
  r2 = top[113];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1416; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_make_counter_472;
  L1416:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[191] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object *) constructed_732[0];
  r1[2] = r2;
  r2 = (object *) constructed_730[0];
  r1[3] = r2;
  r2 = top[104];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1417; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
  L1417:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[192] = r1;
  r1 = (object *) 1;
  r0[193] = r1;
  r1 = (object *) 5;
  r0[194] = r1;
  r1 = (object *) 7;
  r0[195] = r1;
  r1 = allocate (TC_TUPLE, 5);
  r1[1] = lenv; lenv = r1;
  // def in_3662_i4719
  goto L1418;
  FUN_in_3662_i4719:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3668_i4719
    goto L1419;
    FUN_loop_3668_i4719:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = (object *) 6;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object****) lenv) [1][1][1][3];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object *) 262;
            PXLL_RETURN(0);
          } else {
            r0 = ((object*) lenv) [3];
            lenv = ((object *)lenv)[1];
            lenv[2] = r0;
            goto FUN_loop_3668_i4719;
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1419:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3668_i4719; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3668_i4719;
  L1418:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_in_3662_i4719; r2[2] = lenv;
  r1[2] = r2;
  // def add_3663_i4719
  goto L1420;
  FUN_add_3663_i4719:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[1];
    r1 = ((pxll_vector*)r1)->val[2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1421; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1421:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    if PXLL_IS_TRUE(r0) {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[0];
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      lenv = lenv[1];
      r1 = ((object*) lenv) [2];
      ((pxll_vector*)r1)->val[0] = r0;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1420:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_add_3663_i4719; r2[2] = lenv;
  r1[3] = r2;
  // def get_3664_i4719
  goto L1422;
  FUN_get_3664_i4719:
    r0 = ((object*) lenv) [2];
    r0 = ((pxll_vector*)r0)->val[0];
    PXLL_RETURN(0);
  L1422:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_3664_i4719; r2[2] = lenv;
  r1[4] = r2;
  // def iterate_3665_i4719
  goto L1423;
  FUN_iterate_3665_i4719:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3677_i4719
    goto L1424;
    FUN_loop_3677_i4719:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1425; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L1425:
          ; lenv = k[2]; k = k[1];
          r0 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          goto FUN_loop_3677_i4719;
        } break;
      }
      PXLL_RETURN(0);
    L1424:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3677_i4719; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3677_i4719;
  L1423:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_iterate_3665_i4719; r2[2] = lenv;
  r1[5] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (UOTAG(12), 4);
  r3 = ((object**) lenv) [1][3];
  r2[1] = r3;
  r3 = ((object**) lenv) [1][4];
  r2[2] = r3;
  r3 = ((object**) lenv) [1][2];
  r2[3] = r3;
  r3 = ((object**) lenv) [1][5];
  r2[4] = r3;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def new_3682_i4719
  goto L1426;
  FUN_new_3682_i4719:
    r0 = allocate (UOTAG(16), 2);
    r1 = ((object*) lenv) [2];
    r0[1] = r1;
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    PXLL_RETURN(0);
  L1426:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_new_3682_i4719; r2[2] = lenv;
  r1[2] = r2;
  r1 = ((object*) lenv) [2];
  lenv = lenv[1];
  lenv = lenv[1];
  lenv = lenv[1];
  r0[196] = r1;
  r1 = allocate (TC_TUPLE, 9);
  r1[1] = lenv; lenv = r1;
  // def add_3684_i4720
  goto L1427;
  FUN_add_3684_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[2] = r1;
    r0 = ((object**) lenv) [1][4];
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][3];
    t = alloc_no_clear (UOTAG(1),3);
    t[1] = r2;
    t[2] = r0;
    t[3] = r1;
    r0 = t;
    lenv = lenv[1];
    r1 = ((object*) lenv) [2];
    ((pxll_vector*)r1)->val[0] = r0;
    PXLL_RETURN(0);
  L1427:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_add_3684_i4720; r2[2] = lenv;
  r1[2] = r2;
  // def lookup_3685_i4720
  goto L1428;
  FUN_lookup_3685_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3697_i4720
    goto L1429;
    FUN_loop_3697_i4720:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(0)): {
          r0 = (object*)UITAG(1);
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object****) lenv) [1][1][1][3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = ((object*) lenv) [3];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } else {
            r0 = ((object*) lenv) [4];
            lenv = ((object *)lenv)[1];
            lenv[2] = r0;
            goto FUN_loop_3697_i4720;
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1429:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3697_i4720; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3697_i4720;
  L1428:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lookup_3685_i4720; r2[2] = lenv;
  r1[3] = r2;
  // def lookup*_3686_i4720
  goto L1430;
  FUN_lookup_splat_3686_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[3] = r2;
    r2 = ((object***) lenv) [1][1][3];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1431; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_lookup_3685_i4720;
    L1431:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = ((object**) lenv) [1][4];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1430:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lookup_splat_3686_i4720; r2[2] = lenv;
  r1[4] = r2;
  // def get-error_3687_i4720
  goto L1432;
  FUN_get_error_3687_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[3] = r2;
    r2 = ((object***) lenv) [1][1][3];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1433; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_lookup_3685_i4720;
    L1433:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object**) lenv) [1][4];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1432:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_error_3687_i4720; r2[2] = lenv;
  r1[5] = r2;
  // def iterate_3688_i4720
  goto L1434;
  FUN_iterate_3688_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3714_i4720
    goto L1435;
    FUN_loop_3714_i4720:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(0)): {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1436; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L1436:
          ; lenv = k[2]; k = k[1];
          r0 = ((object*) lenv) [4];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          goto FUN_loop_3714_i4720;
        } break;
      }
      PXLL_RETURN(0);
    L1435:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3714_i4720; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3714_i4720;
  L1434:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_iterate_3688_i4720; r2[2] = lenv;
  r1[6] = r2;
  // def map_3689_i4720
  goto L1437;
  FUN_map_3689_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3721_i4720
    goto L1438;
    FUN_loop_3721_i4720:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (UITAG(0)): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 4);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,2);
          r0[4] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][3];
          r1[3] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1439; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto *r2[1];
          L1439:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object***) lenv) [1][1][2];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = ((object*) lenv) [4];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          lenv[3] = r1;
          goto FUN_loop_3721_i4720;
        } break;
      }
      PXLL_RETURN(0);
    L1438:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3721_i4720; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) constructed_729[0];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3721_i4720;
  L1437:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_map_3689_i4720; r2[2] = lenv;
  r1[7] = r2;
  // def keys_3690_i4720
  goto L1440;
  FUN_keys_3690_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3728_i4720
    goto L1441;
    FUN_loop_3728_i4720:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (UITAG(0)): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,2);
          r0[3] = r1;
          r0 = ((object**) lenv) [1][2];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          lenv[3] = r1;
          goto FUN_loop_3728_i4720;
        } break;
      }
      PXLL_RETURN(0);
    L1441:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3728_i4720; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) constructed_728[0];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3728_i4720;
  L1440:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_keys_3690_i4720; r2[2] = lenv;
  r1[8] = r2;
  // def values_3691_i4720
  goto L1442;
  FUN_values_3691_i4720:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3734_i4720
    goto L1443;
    FUN_loop_3734_i4720:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case (r0)) {
        case (UITAG(0)): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,1);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,2);
          r0[3] = r1;
          r0 = ((object**) lenv) [1][2];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r0;
          lenv[3] = r1;
          goto FUN_loop_3734_i4720;
        } break;
      }
      PXLL_RETURN(0);
    L1443:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3734_i4720; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object *) constructed_727[0];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3734_i4720;
  L1442:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_values_3691_i4720; r2[2] = lenv;
  r1[9] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (UOTAG(14), 8);
  r3 = ((object**) lenv) [1][2];
  r2[1] = r3;
  r3 = ((object**) lenv) [1][3];
  r2[2] = r3;
  r3 = ((object**) lenv) [1][4];
  r2[3] = r3;
  r3 = ((object**) lenv) [1][5];
  r2[4] = r3;
  r3 = ((object**) lenv) [1][6];
  r2[5] = r3;
  r3 = ((object**) lenv) [1][8];
  r2[6] = r3;
  r3 = ((object**) lenv) [1][7];
  r2[7] = r3;
  r3 = ((object**) lenv) [1][9];
  r2[8] = r3;
  r1[2] = r2;
  // def lambda_24304_i4720
  goto L1444;
  FUN_lambda_24304_i4720:
    r0 = allocate (UOTAG(15), 2);
    r1 = (object*)UITAG(0);
    r0[1] = r1;
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    PXLL_RETURN(0);
  L1444:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_lambda_24304_i4720; r1[2] = lenv;
  lenv = lenv[1];
  lenv = lenv[1];
  r0[197] = r1;
  r1 = allocate (TC_TUPLE, 9);
  r1[1] = lenv; lenv = r1;
  // def add_3740_i4721
  goto L1445;
  FUN_add_3740_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[0];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1446; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L1446:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_726;
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 5);
        r1 = ((object**) lenv) [1][2];
        r1 = ((pxll_vector*)r1)->val[2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[4] = r1;
        r1 = ((object**) lenv) [1][4];
        r0[5] = r1;
        r1 = top[72];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1447; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
        L1447:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        r1 = ((object**) lenv) [1][2];
        ((pxll_vector*)r1)->val[2] = r0;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1445:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_add_3740_i4721; r2[2] = lenv;
  r1[2] = r2;
  // def lookup_3741_i4721
  goto L1448;
  FUN_lookup_3741_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r0[4] = r1;
    r1 = top[73];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2fmember_409;
  L1448:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lookup_3741_i4721; r2[2] = lenv;
  r1[3] = r2;
  // def lookup*_3742_i4721
  goto L1449;
  FUN_lookup_splat_3742_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[0];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1450; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L1450:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = ((object**) lenv) [1][4];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1449:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lookup_splat_3742_i4721; r2[2] = lenv;
  r1[4] = r2;
  // def get-error_3743_i4721
  goto L1451;
  FUN_get_error_3743_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[0];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1452; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L1452:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object**) lenv) [1][4];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
    }
    PXLL_RETURN(0);
  L1451:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_error_3743_i4721; r2[2] = lenv;
  r1[5] = r2;
  // def iterate_3744_i4721
  goto L1453;
  FUN_iterate_3744_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [3];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
    r0[3] = r1;
    r1 = top[74];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finorder_410;
  L1453:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_iterate_3744_i4721; r2[2] = lenv;
  r1[6] = r2;
  // def map_3745_i4721
  goto L1454;
  FUN_map_3745_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_725[0];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_24413_i4721
    goto L1455;
    FUN_lambda_24413_i4721:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][3];
      r1[3] = r2;
      r2 = ((object****) lenv) [1][1][1][3];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1456; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto *r2[1];
      L1456:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][2];
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      r1 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      lenv = lenv[1];
      ((object**) lenv) [1][2] = r0;
      PXLL_RETURN(0);
    L1455:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_24413_i4721; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
    r0[3] = r1;
    r1 = top[75];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1457; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2freverse_411;
    L1457:
    ; lenv = k[2]; k = k[1];
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L1454:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_map_3745_i4721; r2[2] = lenv;
  r1[7] = r2;
  // def keys_3746_i4721
  goto L1458;
  FUN_keys_3746_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
    r0[2] = r1;
    r1 = top[76];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2fkeys_412;
  L1458:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_keys_3746_i4721; r2[2] = lenv;
  r1[8] = r2;
  // def values_3747_i4721
  goto L1459;
  FUN_values_3747_i4721:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
    r0[2] = r1;
    r1 = top[77];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2fvalues_413;
  L1459:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_values_3747_i4721; r2[2] = lenv;
  r1[9] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (UOTAG(14), 8);
  r3 = ((object**) lenv) [1][2];
  r2[1] = r3;
  r3 = ((object**) lenv) [1][3];
  r2[2] = r3;
  r3 = ((object**) lenv) [1][4];
  r2[3] = r3;
  r3 = ((object**) lenv) [1][5];
  r2[4] = r3;
  r3 = ((object**) lenv) [1][6];
  r2[5] = r3;
  r3 = ((object**) lenv) [1][8];
  r2[6] = r3;
  r3 = ((object**) lenv) [1][7];
  r2[7] = r3;
  r3 = ((object**) lenv) [1][9];
  r2[8] = r3;
  r1[2] = r2;
  // def lambda_24461_i4721
  goto L1460;
  FUN_lambda_24461_i4721:
    r0 = allocate (UOTAG(13), 3);
    r1 = ((object*) lenv) [2];
    r0[1] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = top[165];
    r0[3] = r1;
    PXLL_RETURN(0);
  L1460:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_lambda_24461_i4721; r1[2] = lenv;
  lenv = lenv[1];
  lenv = lenv[1];
  r0[198] = r1;
  r1 = allocate (TC_TUPLE, 5);
  r1[1] = lenv; lenv = r1;
  // def in_3775_i4722
  goto L1461;
  FUN_in_3775_i4722:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object**) lenv) [1][2];
    r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,51)];
    r1[2] = r2;
    r2 = top[12];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r2 = allocate (TC_TUPLE, 2);
    r2[1] = lenv; lenv = r2;
    r3 = ((object***) lenv) [1][1][3];
    r3 = UOBJ_GET(r3,1);
    r2[2] = r3;
    r2 = ((object*) lenv) [2];
    lenv = lenv[1];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1462; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L1462:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = (object *) 262;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object *) 6;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1461:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_in_3775_i4722; r2[2] = lenv;
  r1[2] = r2;
  // def add_3776_i4722
  goto L1463;
  FUN_add_3776_i4722:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = ((object**) lenv) [1][2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1464; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_in_3775_i4722;
    L1464:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    if PXLL_IS_TRUE(r0) {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 5);
      r1 = ((object*) lenv) [2];
      r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
      r0[2] = r1;
      r1 = top[12];
      r0[3] = r1;
      r1 = ((object*) lenv) [3];
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = ((object**) lenv) [1][3];
      r2 = UOBJ_GET(r2,1);
      r1[2] = r2;
      r1 = ((object*) lenv) [2];
      lenv = lenv[1];
      r0[4] = r1;
      r1 = ((object*) lenv) [3];
      r0[5] = r1;
      r1 = top[72];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1465; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
      L1465:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      r1 = ((object*) lenv) [2];
      ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)] = r0;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1463:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_add_3776_i4722; r2[2] = lenv;
  r1[3] = r2;
  // def get_3777_i4722
  goto L1466;
  FUN_get_3777_i4722:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
    r0[2] = r1;
    r1 = top[77];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2fvalues_413;
  L1466:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_3777_i4722; r2[2] = lenv;
  r1[4] = r2;
  // def iterate_3778_i4722
  goto L1467;
  FUN_iterate_3778_i4722:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_24517_i4722
    goto L1468;
    FUN_lambda_24517_i4722:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [3];
      r0[2] = r1;
      r1 = ((object**) lenv) [1][3];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1468:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_24517_i4722; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,51)];
    r0[3] = r1;
    r1 = top[74];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finorder_410;
  L1467:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_iterate_3778_i4722; r2[2] = lenv;
  r1[5] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (UOTAG(12), 4);
  r3 = ((object**) lenv) [1][3];
  r2[1] = r3;
  r3 = ((object**) lenv) [1][4];
  r2[2] = r3;
  r3 = ((object**) lenv) [1][2];
  r2[3] = r3;
  r3 = ((object**) lenv) [1][5];
  r2[4] = r3;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def new_3790_i4722
  goto L1469;
  FUN_new_3790_i4722:
    r0 = allocate (UOTAG(11), 2);
    r1 = ((object**) lenv) [1][2];
    r0[1] = r1;
    r1 = top[165];
    r0[2] = r1;
    PXLL_RETURN(0);
  L1469:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_new_3790_i4722; r2[2] = lenv;
  r1[2] = r2;
  r1 = ((object*) lenv) [2];
  lenv = lenv[1];
  lenv = lenv[1];
  lenv = lenv[1];
  r0[199] = r1;
  r1 = (object *) 27;
  r0[200] = r1;
  r1 = (object *) 1;
  r0[201] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (TC_TUPLE, 2);
  r3 = (object *) 1;
  r2[2] = r3;
  r3 = top[113];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1470; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_make_counter_472;
  L1470:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  // def lambda_27776
  goto L1471;
  FUN_lambda_27776:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 4);
    r1[1] = lenv; lenv = r1;
    r2 = allocate (TC_TUPLE, 2);
    r3 = ((object**) lenv) [1][2];
    r3 = ((pxll_vector*)r3)->val[1];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L1472; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    lenv = r3[2]; goto *r3[1];
    L1472:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    r3 = top[56];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1473; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
    L1473:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = (object*)TC_NIL;
    r1[3] = r2;
    r2 = ((object*) lenv) [3];
    r3 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    r1[4] = r2;
    r1 = ((object*) lenv) [4];
    r2 = (object*) &constructed_724;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    lenv = lenv[1];
    r0[2] = r1;
    r1 = top[47];
    r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
  L1471:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_lambda_27776; r1[2] = lenv;
  lenv = lenv[1];
  r0[202] = r1;
  r1 = (object *) constructed_723[0];
  r0[203] = r1;
  r1 = (object *) 262;
  r0[204] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = allocate (TC_TUPLE, 2);
  r3 = (object*) &constructed_706;
  r2[2] = r3;
  r3 = top[55];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1474; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_string__3elist_362;
  L1474:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  r2 = top[83];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1475; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_char_class_423;
  L1475:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[205] = r1;
  r1 = (object*) &constructed_705;
  r0[206] = r1;
  r1 = (object*) &constructed_704;
  r0[207] = r1;
  r1 = (object*) &constructed_703;
  r0[208] = r1;
  r1 = (object*) &constructed_702;
  r0[209] = r1;
  r0 = top[165];
  r1 = top[165];
  ((pxll_vector*)r1)->val[1] = r0;
  r0 = top[165];
  r1 = top[165];
  ((pxll_vector*)r1)->val[3] = r0;
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  r1 = (object *) pxll_internal_symbols;
  r0[2] = r1;
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  // def loop_1251_i4726
  goto L1476;
  FUN_loop_1251_i4726:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object****) lenv) [1][1][1][2];
    r1 = box((pxll_int)(r1 == (object*) TC_EMPTY_VECTOR) ? 0 : GET_TUPLE_LENGTH(*r1));
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object**) lenv) [1][2];
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [2];
      r2 = ((object***) lenv) [1][1][2];
      range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
      r1 = ((pxll_vector*)r2)->val[unbox(r1)];
      r0[2] = r1;
      r1 = top[79];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1477; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_intern_symbol_418;
      L1477:
      ; lenv = k[2]; k = k[1];
      r0 = (object *) 3;
      r1 = ((object*) lenv) [2];
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      lenv[2] = r0;
      goto FUN_loop_1251_i4726;
    }
    PXLL_RETURN(0);
  L1476:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_loop_1251_i4726; r1[2] = lenv;
  r0[2] = r1;
  r0 = allocate (TC_TUPLE, 2);
  r1 = (object *) 1;
  r0[2] = r1;
  r1 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L1478; ; k = t;
  r0[1] = r1[2]; lenv = r0; goto FUN_loop_1251_i4726;
  L1478:
  ; lenv = k[2]; k = k[1];
  r0 = result;
  lenv = lenv[1];
  lenv = lenv[1];
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  r1 = top[164];
  r1 = ((pxll_vector*)r1)->val[0];
  r0[2] = r1;
  r0 = (object *) 5;
  r1 = ((object*) lenv) [2];
  r0 = PXLL_TEST(unbox(r1)<unbox(r0));
  lenv = lenv[1];
  if PXLL_IS_TRUE(r0) {
    r0 = (object*) &constructed_68;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_68;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = (object*) &constructed_701;
    dump_object (r0, 0); fprintf (stdout, "\n");
    r0 = (object*) &constructed_701;
    PXLL_UNDEFINED; result=r0; goto Lreturn;
  } else {
    r0 = (object *) 14;
  }
  r0 = allocate (TC_TUPLE, 54);
  r0[1] = lenv; lenv = r0;
  r1 = allocate (UOTAG(8), 17);
  r2 = (object *) constructed_700[0];
  r1[1] = r2;
  r2 = (object *) constructed_699[0];
  r1[2] = r2;
  r2 = (object*)UITAG(0);
  r1[3] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r3 = top[82];
  r2[2] = r3;
  r3 = top[198];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1479; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto *r3[1];
  L1479:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[4] = r2;
  r2 = top[165];
  r1[5] = r2;
  r2 = (object *) constructed_698[0];
  r1[6] = r2;
  r2 = top[165];
  r1[7] = r2;
  r2 = (object *) constructed_697[0];
  r1[8] = r2;
  r2 = (object*)UITAG(0);
  r1[9] = r2;
  r2 = allocate (UOTAG(10), 5);
  r3 = (object*) &constructed_696;
  r2[1] = r3;
  r3 = (object *) 6;
  r2[2] = r3;
  r3 = (object *) 6;
  r2[3] = r3;
  r3 = (object *) 6;
  r2[4] = r3;
  r3 = (object *) 6;
  r2[5] = r3;
  r1[10] = r2;
  r2 = (object *) constructed_695[0];
  r1[11] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r2[1] = lenv; lenv = r2;
  r3 = (object *) -1;
  r2[2] = r3;
  r2 = allocate (TC_TUPLE, 3);
  r2[1] = lenv; lenv = r2;
  // def allocate_2626_i4_i4620_i4727
  goto L1480;
  FUN_allocate_2626_i4_i4620_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_2629_i4_i4620_i4727
    goto L1481;
    FUN_loop_2629_i4_i4620_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 4);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][2];
      r0[3] = r1;
      r1 = top[11];
      r0[4] = r1;
      r1 = top[24];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1482; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_member_question_318;
      L1482:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        lenv[2] = r0;
        goto FUN_loop_2629_i4_i4620_i4727;
      } else {
        r0 = ((object*****) lenv) [1][1][1][1][2];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
          r0 = ((object*) lenv) [2];
        } else {
          r0 = ((object*****) lenv) [1][1][1][1][2];
        }
        ((object*****) lenv) [1][1][1][1][2] = r0;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L1481:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_2629_i4_i4620_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object *) 1;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_2629_i4_i4620_i4727;
  L1480:
  r3 = allocate (TC_CLOSURE, 2);
  r3[1] = &&FUN_allocate_2626_i4_i4620_i4727; r3[2] = lenv;
  r2[2] = r3;
  // def get-max_2627_i4_i4620_i4727
  goto L1483;
  FUN_get_max_2627_i4_i4620_i4727:
    r0 = ((object**) lenv) [1][2];
    PXLL_RETURN(0);
  L1483:
  r3 = allocate (TC_CLOSURE, 2);
  r3[1] = &&FUN_get_max_2627_i4_i4620_i4727; r3[2] = lenv;
  r2[3] = r3;
  r2 = allocate (UOTAG(9), 2);
  r3 = ((object*) lenv) [2];
  r2[1] = r3;
  r3 = ((object*) lenv) [3];
  r2[2] = r3;
  lenv = lenv[1];
  lenv = lenv[1];
  r1[12] = r2;
  r2 = (object *) constructed_694[0];
  r1[13] = r2;
  r2 = (object*) &constructed_693;
  r1[14] = r2;
  r2 = (object*)UITAG(0);
  r1[15] = r2;
  r2 = (object*)UITAG(0);
  r1[16] = r2;
  r2 = top[165];
  r1[17] = r2;
  r0[2] = r1;
  r1 = top[164];
  r1 = ((pxll_vector*)r1)->val[1];
  r0[3] = r1;
  r1 = ((object*) lenv) [2];
  r1 = ((pxll_vector*)r1)->val[9];
  r0[4] = r1;
  r1 = ((object*) lenv) [3];
  r1 = box((pxll_int)(r1 == (object*) TC_EMPTY_VECTOR) ? 0 : GET_TUPLE_LENGTH(*r1));
  r0[5] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def $loop_4706_i4621_i4727
  goto L1484;
  FUN__24loop_4706_i4621_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r0 = ((object****) lenv) [1][1][1][5];
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    } else {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = top[164];
      r1 = ((pxll_vector*)r1)->val[1];
      r2 = ((object**) lenv) [1][2];
      range_check (GET_TUPLE_LENGTH(*(object*)r1), unbox(r2));
      r1 = ((pxll_vector*)r1)->val[unbox(r2)];
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = (object*) &constructed_688;
      r1[3] = r2;
      r2 = top[50];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1485; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
      L1485:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 262;
        r1 = ((object****) lenv) [1][1][1][4];
        ((pxll_vector*)r1)->val[2] = r0;
      } else {
        goto fatbar_151;
      }
      goto fatbar_151_over;
      fatbar_151:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = (object*) &constructed_689;
      r1[3] = r2;
      r2 = top[50];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1486; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
      L1486:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r0[2] = r1;
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = box((pxll_int)unbox(r1)+unbox(r0));
        lenv = lenv[1];
        ((object**) lenv) [1][2] = r0;
        r0 = ((object**) lenv) [1][2];
        r1 = ((object****) lenv) [1][1][1][3];
        range_check (GET_TUPLE_LENGTH(*(object*)r1), unbox(r0));
        r0 = ((pxll_vector*)r1)->val[unbox(r0)];
        r1 = ((object****) lenv) [1][1][1][4];
        ((pxll_vector*)r1)->val[0] = r0;
      } else {
        goto fatbar_152;
      }
      goto fatbar_152_over;
      fatbar_152:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = (object*) &constructed_690;
      r1[3] = r2;
      r2 = top[50];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1487; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
      L1487:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 262;
        r1 = ((object****) lenv) [1][1][1][4];
        ((pxll_vector*)r1)->val[3] = r0;
      } else {
        goto fatbar_153;
      }
      goto fatbar_153_over;
      fatbar_153:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = (object*) &constructed_691;
      r1[3] = r2;
      r2 = top[50];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1488; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
      L1488:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 262;
        r1 = ((object****) lenv) [1][1][1][4];
        ((pxll_vector*)r1)->val[4] = r0;
      } else {
        goto fatbar_154;
      }
      goto fatbar_154_over;
      fatbar_154:
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object**) lenv) [1][2];
      r1[2] = r2;
      r2 = (object*) &constructed_692;
      r1[3] = r2;
      r2 = top[50];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1489; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
      L1489:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(unbox(r1)==unbox(r0));
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = (object *) 262;
        r1 = ((object****) lenv) [1][1][1][4];
        ((pxll_vector*)r1)->val[1] = r0;
      } else {
        goto fatbar_155;
      }
      goto fatbar_155_over;
      fatbar_155:
      r0 = (object *) 14;
      fatbar_155_over: ;
      fatbar_154_over: ;
      fatbar_153_over: ;
      fatbar_152_over: ;
      fatbar_151_over: ;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r0[2] = r1;
      r0 = (object *) 3;
      r1 = ((object*) lenv) [2];
      r0 = box((pxll_int)unbox(r1)+unbox(r0));
      lenv = lenv[1];
      lenv[2] = r0;
      goto FUN__24loop_4706_i4621_i4727;
    }
    PXLL_RETURN(0);
  L1484:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN__24loop_4706_i4621_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) 1;
  r1[2] = r2;
  r2 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1490; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4706_i4621_i4727;
  L1490:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[6] = r1;
  r1 = allocate (TC_TUPLE, 28);
  r1[1] = lenv; lenv = r1;
  // def go_1960_i4622_i4727
  goto L1491;
  FUN_go_1960_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    switch (get_case (r1)) {
      case (UOTAG(0)): {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object***) lenv) [1][1][2];
        r2 = UOBJ_GET(r2,0);
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        switch (get_case (r1)) {
          case (TC_NIL): {
            r1 = (object *) constructed_686[0];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
          } break;
          default: {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = ((object*) lenv) [3];
            switch (get_case (r1)) {
              case (TC_NIL): {
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object*) lenv) [2];
                r1[2] = r2;
                r2 = ((object*****) lenv) [1][1][1][1][8];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1492; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                L1492:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_150;
              }
            }
            lenv = lenv[1];
          } break;
        }
        goto fatbar_150_over;
        fatbar_150:
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = ((object*****) lenv) [1][1][1][1][5];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1493; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_find_declarations_1965_i4622_i4727;
        L1493:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        r1[3] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][6];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1494; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_find_definitions_1966_i4622_i4727;
        L1494:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
        fatbar_150_over: ;
        lenv = lenv[1];
      } break;
      default: {
        goto fatbar_149;
      }
    }
    goto fatbar_149_over;
    fatbar_149:
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object*) &constructed_687;
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = top[15];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1495; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
    L1495:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    fatbar_149_over: ;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object***) lenv) [1][1][20];
    r0[1] = r1[2]; lenv = r0; goto FUN_wrap_with_constructors_1983_i4622_i4727;
  L1491:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_go_1960_i4622_i4727; r2[2] = lenv;
  r1[2] = r2;
  // def wrap-fix_1961_i4622_i4727
  goto L1496;
  FUN_wrap_fix_1961_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1497; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L1497:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)>unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 9);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) constructed_685[0];
      t = alloc_no_clear (UOTAG(1),1);
      t[1] = r1;
      r1 = t;
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r1;
      r1 = t;
      r0[3] = r1;
      r1 = ((object**) lenv) [1][3];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r1;
      r1 = t;
      r0[4] = r1;
      r1 = (object*)TC_NIL;
      r0[5] = r1;
      r1 = ((object*) lenv) [5];
      r2 = ((object**) lenv) [1][4];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[6] = r1;
      r1 = ((object*) lenv) [6];
      r2 = ((object*) lenv) [4];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[7] = r1;
      r1 = ((object*) lenv) [7];
      r2 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[8] = r1;
      r1 = ((object*) lenv) [8];
      r2 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[9] = r1;
      r0 = ((object*) lenv) [9];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r0;
      r0 = t;
      PXLL_RETURN(0);
    } else {
      r0 = ((object*) lenv) [4];
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1496:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_wrap_fix_1961_i4622_i4727; r2[2] = lenv;
  r1[3] = r2;
  // def wrap-definitions_1963_i4622_i4727
  goto L1498;
  FUN_wrap_definitions_1963_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_683[0];
    r0[2] = r1;
    r1 = (object *) constructed_682[0];
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_11238_i4622_i4727
    goto L1499;
    FUN_lambda_11238_i4622_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case_noint (r0)) {
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (UOTAG(1),1);
          t[1] = r1;
          r1 = t;
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object***) lenv) [1][1][2] = r0;
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][3];
          r1[2] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][8];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1500; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
          L1500:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object***) lenv) [1][1][3] = r0;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1499:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_11238_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = top[35];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1501; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L1501:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 4);
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = (object *) constructed_36[0];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1502; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L1502:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    r2 = ((object**) lenv) [1][3];
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = (object *) constructed_36[0];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1503; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L1503:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 3);
    r2[1] = lenv; lenv = r2;
    r3 = (object *) constructed_599[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r3;
    r3 = t;
    r2[2] = r3;
    r3 = ((object***) lenv) [1][1][3];
    r4 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[3] = r3;
    r2 = ((object*) lenv) [3];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = ((object***) lenv) [1][1][8];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1504; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
    L1504:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r1 = ((object***) lenv) [1][1][3];
    r0[1] = r1[2]; lenv = r0; goto FUN_wrap_fix_1961_i4622_i4727;
  L1498:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_wrap_definitions_1963_i4622_i4727; r2[2] = lenv;
  r1[4] = r2;
  // def find-declarations_1965_i4622_i4727
  goto L1505;
  FUN_find_declarations_1965_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def recur_2016_i4622_i4727
    goto L1506;
    FUN_recur_2016_i4622_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [3];
          r0[2] = r1;
          r1 = (object *) constructed_36[0];
          r0[3] = r1;
          r1 = top[21];
          r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UOTAG(1)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = (object *) constructed_675[0];
                      r1 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                        r0 = ((object**) lenv) [1][3];
                        switch (get_case (r0)) {
                          case (TC_PAIR): {
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object***) lenv) [1][1][3];
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = ((object***) lenv) [1][1][3];
                            r1 = UOBJ_GET(r1,1);
                            r0[3] = r1;
                            r0 = ((object*) lenv) [2];
                            switch (get_case (r0)) {
                              case (UOTAG(1)): {
                                r0 = allocate (TC_TUPLE, 4);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object**) lenv) [1][2];
                                r1 = UOBJ_GET(r1,0);
                                r0[2] = r1;
                                r1 = allocate (TC_TUPLE, 2);
                                r1[1] = lenv; lenv = r1;
                                // def loop_2172_i1124_i4622_i4727
                                goto L1507;
                                FUN_loop_2172_i1124_i4622_i4727:
                                  check_heap (0);
                                  r0 = ((object*) lenv) [2];
                                  switch (get_case (r0)) {
                                    case (TC_NIL): {
                                      r0 = (object *) constructed_671[0];
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      r0 = allocate (TC_TUPLE, 3);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][2];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r1 = ((object**) lenv) [1][2];
                                      r1 = UOBJ_GET(r1,1);
                                      r0[3] = r1;
                                      r0 = ((object*) lenv) [3];
                                      switch (get_case (r0)) {
                                        case (TC_PAIR): {
                                          r0 = allocate (TC_TUPLE, 3);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object**) lenv) [1][3];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r1 = ((object**) lenv) [1][3];
                                          r1 = UOBJ_GET(r1,1);
                                          r0[3] = r1;
                                          r0 = ((object*) lenv) [2];
                                          switch (get_case (r0)) {
                                            case (UOTAG(1)): {
                                              r0 = allocate (TC_TUPLE, 2);
                                              r0[1] = lenv; lenv = r0;
                                              r1 = ((object**) lenv) [1][2];
                                              r1 = UOBJ_GET(r1,0);
                                              r0[2] = r1;
                                              r0 = (object *) constructed_660[0];
                                              r1 = ((object*) lenv) [2];
                                              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                                r0 = ((object**) lenv) [1][3];
                                                switch (get_case (r0)) {
                                                  case (TC_PAIR): {
                                                    r0 = allocate (TC_TUPLE, 5);
                                                    r0[1] = lenv; lenv = r0;
                                                    r1 = ((object***) lenv) [1][1][3];
                                                    r1 = UOBJ_GET(r1,0);
                                                    r0[2] = r1;
                                                    r1 = ((object***) lenv) [1][1][3];
                                                    r1 = UOBJ_GET(r1,1);
                                                    r0[3] = r1;
                                                    r1 = allocate (UOTAG(1), 2);
                                                    r2 = ((object****) lenv) [1][1][1][2];
                                                    r1[1] = r2;
                                                    r2 = ((object*) lenv) [2];
                                                    r1[2] = r2;
                                                    r0[4] = r1;
                                                    r1 = allocate (TC_TUPLE, 2);
                                                    r2 = ((object*) lenv) [3];
                                                    r1[2] = r2;
                                                    r2 = ((object******) lenv) [1][1][1][1][1][2];
                                                    t = allocate (TC_SAVE, 4);
                                                    t[1] = k; t[2] = lenv; t[3] = &&L1508; t[4] = r0; k = t;
                                                    r1[1] = r2[2]; lenv = r1; goto FUN_loop_2172_i1124_i4622_i4727;
                                                    L1508:
                                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                                    r1 = result;
                                                    r0[5] = r1;
                                                    r0 = ((object*) lenv) [5];
                                                    r1 = ((object*) lenv) [4];
                                                    t = alloc_no_clear (TC_PAIR,2);
                                                    t[1] = r1;
                                                    t[2] = r0;
                                                    r0 = t;
                                                    PXLL_RETURN(0);
                                                  } break;
                                                  default: {
                                                    lenv = ((object ***)lenv)[1][1][1];
                                                    goto fatbar_146;
                                                  }
                                                }
                                                PXLL_RETURN(0);
                                              } else {
                                                lenv = ((object ***)lenv)[1][1][1];
                                                goto fatbar_146;
                                              }
                                              PXLL_RETURN(0);
                                            } break;
                                            default: {
                                              lenv = ((object **)lenv)[1][1];
                                              goto fatbar_146;
                                            }
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object *)lenv)[1];
                                          goto fatbar_146;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                  }
                                  PXLL_RETURN(0);
                                  goto fatbar_146_over;
                                  fatbar_146:
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = (object*) &constructed_672;
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[3] = r1;
                                  r1 = top[15];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                                  fatbar_146_over: ;
                                  PXLL_RETURN(0);
                                L1507:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_loop_2172_i1124_i4622_i4727; r2[2] = lenv;
                                r1[2] = r2;
                                r1 = allocate (TC_TUPLE, 2);
                                r2 = ((object***) lenv) [1][1][3];
                                r1[2] = r2;
                                r2 = ((object*) lenv) [2];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L1509; t[4] = r0; k = t;
                                r1[1] = r2[2]; lenv = r1; goto FUN_loop_2172_i1124_i4622_i4727;
                                L1509:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                lenv = lenv[1];
                                r0[3] = r1;
                                r1 = allocate (TC_TUPLE, 3);
                                r1[1] = lenv; lenv = r1;
                                // def apply_1558_i1125_i4622_i4727
                                goto L1510;
                                FUN_apply_1558_i1125_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  // def loop_1561_i1125_i4622_i4727
                                  goto L1511;
                                  FUN_loop_1561_i1125_i4622_i4727:
                                    check_heap (0);
                                    r0 = ((object*) lenv) [2];
                                    switch (get_case (r0)) {
                                      case (TC_PAIR): {
                                        r0 = allocate (TC_TUPLE, 3);
                                        r0[1] = lenv; lenv = r0;
                                        r1 = ((object**) lenv) [1][2];
                                        r1 = UOBJ_GET(r1,0);
                                        r0[2] = r1;
                                        r1 = ((object**) lenv) [1][2];
                                        r1 = UOBJ_GET(r1,1);
                                        r0[3] = r1;
                                        r0 = ((object*) lenv) [2];
                                        switch (get_case_noint (r0)) {
                                          default: {
                                            r0 = allocate (TC_TUPLE, 3);
                                            r0[1] = lenv; lenv = r0;
                                            r1 = ((object**) lenv) [1][2];
                                            r1 = UOBJ_GET(r1,0);
                                            r0[2] = r1;
                                            r1 = ((object**) lenv) [1][2];
                                            r1 = UOBJ_GET(r1,1);
                                            r0[3] = r1;
                                            r0 = allocate (TC_TUPLE, 3);
                                            r1 = ((object*) lenv) [2];
                                            r0[2] = r1;
                                            r1 = ((object*****) lenv) [1][1][1][1][2];
                                            r0[3] = r1;
                                            r1 = top[97];
                                            t = allocate (TC_SAVE, 3);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1512; ; k = t;
                                            r0[1] = r1[2]; lenv = r0; goto FUN_matches_pattern_question_441;
                                            L1512:
                                            ; lenv = k[2]; k = k[1];
                                            r0 = result;
                                            if PXLL_IS_TRUE(r0) {
                                              r0 = allocate (TC_TUPLE, 2);
                                              r0[1] = lenv; lenv = r0;
                                              r1 = allocate (TC_TUPLE, 3);
                                              r2 = ((object**) lenv) [1][3];
                                              r1[2] = r2;
                                              r2 = allocate (TC_TUPLE, 3);
                                              r3 = ((object**) lenv) [1][2];
                                              r2[2] = r3;
                                              r3 = ((object******) lenv) [1][1][1][1][1][2];
                                              r2[3] = r3;
                                              r3 = top[101];
                                              t = allocate (TC_SAVE, 5);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1513; t[4] = r0; t[5] = r1; k = t;
                                              r2[1] = r3[2]; lenv = r2; goto FUN_get_bindings_445;
                                              L1513:
                                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                              r2 = result;
                                              r1[3] = r2;
                                              r2 = top[102];
                                              t = allocate (TC_SAVE, 4);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1514; t[4] = r0; k = t;
                                              r1[1] = r2[2]; lenv = r1; goto FUN_expand_pattern_447;
                                              L1514:
                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                              r1 = result;
                                              r0[2] = r1;
                                              r0 = ((object*) lenv) [2];
                                              PXLL_RETURN(0);
                                            } else {
                                              r0 = ((object**) lenv) [1][3];
                                              lenv = ((object **)lenv)[1][1];
                                              lenv[2] = r0;
                                              goto FUN_loop_1561_i1125_i4622_i4727;
                                            }
                                            PXLL_RETURN(0);
                                          } break;
                                        }
                                        PXLL_RETURN(0);
                                      } break;
                                      default: {
                                        r0 = allocate (TC_TUPLE, 3);
                                        r1 = (object*) &constructed_670;
                                        r0[2] = r1;
                                        r1 = allocate (TC_TUPLE, 2);
                                        r2 = ((object***) lenv) [1][1][2];
                                        r1[2] = r2;
                                        r2 = top[91];
                                        t = allocate (TC_SAVE, 4);
                                        t[1] = k; t[2] = lenv; t[3] = &&L1515; t[4] = r0; k = t;
                                        r1[1] = r2[2]; lenv = r1; goto FUN_repr_433;
                                        L1515:
                                        r0 = k[4]; lenv = k[2]; k = k[1];
                                        r1 = result;
                                        r0[3] = r1;
                                        r1 = top[15];
                                        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                                      } break;
                                    }
                                    PXLL_RETURN(0);
                                  L1511:
                                  r1 = allocate (TC_CLOSURE, 2);
                                  r1[1] = &&FUN_loop_1561_i1125_i4622_i4727; r1[2] = lenv;
                                  r0[2] = r1;
                                  r0 = allocate (TC_TUPLE, 2);
                                  r1 = ((object****) lenv) [1][1][1][3];
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_loop_1561_i1125_i4622_i4727;
                                L1510:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_apply_1558_i1125_i4622_i4727; r2[2] = lenv;
                                r1[2] = r2;
                                // def unread-macro_1559_i1125_i4622_i4727
                                goto L1516;
                                FUN_unread_macro_1559_i1125_i4622_i4727:
                                  check_heap (0);
                                  r0 = (object*) &constructed_669;
                                  r0 = box((pxll_int)((pxll_string*)(r0))->len);
                                  r1 = (object*) &constructed_669;
                                  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                                  r0 = ((object**) lenv) [1][2];
                                  dump_object (r0, 0);
                                  r0 = (object*) &constructed_668;
                                  r0 = box((pxll_int)((pxll_string*)(r0))->len);
                                  r1 = (object*) &constructed_668;
                                  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                                  r0 = allocate (TC_TUPLE, 3);
                                  // def lambda_8037_i1125_i4622_i4727
                                  goto L1517;
                                  FUN_lambda_8037_i1125_i4622_i4727:
                                    check_heap (0);
                                    r0 = ((object*) lenv) [2];
                                    switch (get_case_noint (r0)) {
                                      default: {
                                        r0 = allocate (TC_TUPLE, 3);
                                        r0[1] = lenv; lenv = r0;
                                        r1 = ((object**) lenv) [1][2];
                                        r1 = UOBJ_GET(r1,0);
                                        r0[2] = r1;
                                        r1 = ((object**) lenv) [1][2];
                                        r1 = UOBJ_GET(r1,1);
                                        r0[3] = r1;
                                        r0 = allocate (TC_TUPLE, 3);
                                        r1 = (object *) 1;
                                        r0[2] = r1;
                                        r1 = ((object*) lenv) [2];
                                        r0[3] = r1;
                                        r1 = top[95];
                                        t = allocate (TC_SAVE, 3);
                                        t[1] = k; t[2] = lenv; t[3] = &&L1518; ; k = t;
                                        r0[1] = r1[2]; lenv = r0; goto FUN_pp_437;
                                        L1518:
                                        ; lenv = k[2]; k = k[1];
                                        r0 = (object*) &constructed_667;
                                        r0 = box((pxll_int)((pxll_string*)(r0))->len);
                                        r1 = (object*) &constructed_667;
                                        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                                        r0 = allocate (TC_TUPLE, 3);
                                        r1 = (object *) 1;
                                        r0[2] = r1;
                                        r1 = ((object*) lenv) [3];
                                        r0[3] = r1;
                                        r1 = top[95];
                                        t = allocate (TC_SAVE, 3);
                                        t[1] = k; t[2] = lenv; t[3] = &&L1519; ; k = t;
                                        r0[1] = r1[2]; lenv = r0; goto FUN_pp_437;
                                        L1519:
                                        ; lenv = k[2]; k = k[1];
                                        r0 = (object*) &constructed_666;
                                        r0 = box((pxll_int)((pxll_string*)(r0))->len);
                                        r1 = (object*) &constructed_666;
                                        r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                                        PXLL_RETURN(0);
                                      } break;
                                    }
                                    PXLL_RETURN(0);
                                  L1517:
                                  r1 = allocate (TC_CLOSURE, 2);
                                  r1[1] = &&FUN_lambda_8037_i1125_i4622_i4727; r1[2] = lenv;
                                  r0[2] = r1;
                                  r1 = ((object**) lenv) [1][3];
                                  r0[3] = r1;
                                  r1 = top[35];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1520; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
                                  L1520:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = (object*) &constructed_665;
                                  r0 = box((pxll_int)((pxll_string*)(r0))->len);
                                  r1 = (object*) &constructed_665;
                                  r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                                  PXLL_RETURN(0);
                                L1516:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_unread_macro_1559_i1125_i4622_i4727; r2[2] = lenv;
                                r1[3] = r2;
                                r1 = allocate (UOTAG(6), 4);
                                r2 = ((object*) lenv) [2];
                                r1[1] = r2;
                                r2 = ((object**) lenv) [1][2];
                                r1[2] = r2;
                                r2 = ((object**) lenv) [1][3];
                                r1[3] = r2;
                                r2 = ((object*) lenv) [3];
                                r1[4] = r2;
                                lenv = lenv[1];
                                r0[4] = r1;
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                                r1 = ((pxll_vector*)r1)->val[8];
                                r0[2] = r1;
                                r0 = ((object**) lenv) [1][4];
                                r1 = ((object*) lenv) [2];
                                r2 = ((object**) lenv) [1][2];
                                t = alloc_no_clear (UOTAG(1),3);
                                t[1] = r2;
                                t[2] = r0;
                                t[3] = r1;
                                r0 = t;
                                lenv = lenv[1];
                                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                                ((pxll_vector*)r1)->val[8] = r0;
                                lenv = lenv[1];
                              } break;
                              default: {
                                lenv = ((object *)lenv)[1];
                                goto fatbar_145;
                              }
                            }
                            lenv = lenv[1];
                          } break;
                          default: {
                            goto fatbar_145;
                          }
                        }
                        goto fatbar_145_over;
                        fatbar_145:
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = (object*) &constructed_673;
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][3];
                        r0[3] = r1;
                        r1 = top[15];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1521; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                        L1521:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                        fatbar_145_over: ;
                        r0 = ((object*****) lenv) [1][1][1][1][3];
                        r1 = ((object****) lenv) [1][1][1][3];
                        lenv = ((object ****)lenv)[1][1][1][1];
                        lenv[2] = r1;
                        lenv[3] = r0;
                        goto FUN_recur_2016_i4622_i4727;
                      } else {
                        goto fatbar_144;
                      }
                      PXLL_RETURN(0);
                      goto fatbar_144_over;
                      fatbar_144:
                      r0 = (object *) constructed_681[0];
                      r1 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                        r0 = ((object**) lenv) [1][3];
                        switch (get_case (r0)) {
                          case (TC_PAIR): {
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object***) lenv) [1][1][3];
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = ((object***) lenv) [1][1][3];
                            r1 = UOBJ_GET(r1,1);
                            r0[3] = r1;
                            r0 = ((object*) lenv) [2];
                            switch (get_case (r0)) {
                              case (UOTAG(1)): {
                                r0 = allocate (TC_TUPLE, 6);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object**) lenv) [1][2];
                                r1 = UOBJ_GET(r1,0);
                                r0[2] = r1;
                                r1 = top[197];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L1522; t[4] = r0; k = t;
                                lenv = r1[2]; goto *r1[1];
                                L1522:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                r0[3] = r1;
                                r1 = top[197];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L1523; t[4] = r0; k = t;
                                lenv = r1[2]; goto *r1[1];
                                L1523:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                r0[4] = r1;
                                r1 = (object *) 1;
                                r0[5] = r1;
                                r1 = allocate (TC_TUPLE, 8);
                                r1[1] = lenv; lenv = r1;
                                // def get_2186_i1131_i4622_i4727
                                goto L1524;
                                FUN_get_2186_i1131_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 4);
                                  r1 = ((object***) lenv) [1][1][4];
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[3] = r1;
                                  r1 = (object*) &constructed_678;
                                  r0[4] = r1;
                                  r1 = ((object***) lenv) [1][1][4];
                                  r1 = ((pxll_vector*)r1)->val[1];
                                  r1 = ((pxll_vector*)r1)->val[3];
                                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                L1524:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_get_2186_i1131_i4622_i4727; r2[2] = lenv;
                                r1[2] = r2;
                                // def add_2187_i1131_i4622_i4727
                                goto L1525;
                                FUN_add_2187_i1131_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 4);
                                  r1 = ((object***) lenv) [1][1][4];
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r1 = ((pxll_vector*)r1)->val[2];
                                  r0[3] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[4] = r1;
                                  r1 = ((object***) lenv) [1][1][4];
                                  r1 = ((pxll_vector*)r1)->val[1];
                                  r1 = ((pxll_vector*)r1)->val[0];
                                  t = allocate (TC_SAVE, 3);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1526; ; k = t;
                                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                  L1526:
                                  ; lenv = k[2]; k = k[1];
                                  r0 = ((object***) lenv) [1][1][5];
                                  r1 = ((object*) lenv) [2];
                                  ((pxll_vector*)r1)->val[1] = r0;
                                  r0 = ((object***) lenv) [1][1][5];
                                  r1 = (object *) 3;
                                  r0 = box((pxll_int)unbox(r1)+unbox(r0));
                                  ((object***) lenv) [1][1][5] = r0;
                                  PXLL_RETURN(0);
                                L1525:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_add_2187_i1131_i4622_i4727; r2[2] = lenv;
                                r1[3] = r2;
                                // def iterate_2188_i1131_i4622_i4727
                                goto L1527;
                                FUN_iterate_2188_i1131_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object***) lenv) [1][1][4];
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[3] = r1;
                                  r1 = ((object***) lenv) [1][1][4];
                                  r1 = ((pxll_vector*)r1)->val[1];
                                  r1 = ((pxll_vector*)r1)->val[4];
                                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                L1527:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_iterate_2188_i1131_i4622_i4727; r2[2] = lenv;
                                r1[4] = r2;
                                // def get-nalts_2189_i1131_i4622_i4727
                                goto L1528;
                                FUN_get_nalts_2189_i1131_i4622_i4727:
                                  r0 = ((object**) lenv) [1][5];
                                  PXLL_RETURN(0);
                                L1528:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_get_nalts_2189_i1131_i4622_i4727; r2[2] = lenv;
                                r1[5] = r2;
                                // def get-scheme_2190_i1131_i4622_i4727
                                goto L1529;
                                FUN_get_scheme_2190_i1131_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object***) lenv) [1][1][3];
                                  r1[2] = r2;
                                  r2 = ((object***) lenv) [1][1][3];
                                  r2 = ((pxll_vector*)r2)->val[1];
                                  r2 = ((pxll_vector*)r2)->val[7];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1530; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                  L1530:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object*) lenv) [2];
                                  r1[2] = r2;
                                  r2 = (object *) constructed_36[0];
                                  r1[3] = r2;
                                  r2 = top[21];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1531; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                  L1531:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r0 = allocate (UOTAG(2), 2);
                                  r1 = ((object*) lenv) [3];
                                  r0[1] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object***) lenv) [1][1][2];
                                  r1[2] = r2;
                                  r2 = ((object*) lenv) [3];
                                  r1[3] = r2;
                                  r2 = top[104];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1532; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                                  L1532:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  PXLL_RETURN(0);
                                L1529:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_get_scheme_2190_i1131_i4622_i4727; r2[2] = lenv;
                                r1[6] = r2;
                                // def get-tvars_2191_i1131_i4622_i4727
                                goto L1533;
                                FUN_get_tvars_2191_i1131_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object***) lenv) [1][1][3];
                                  r1[2] = r2;
                                  r2 = ((object***) lenv) [1][1][3];
                                  r2 = ((pxll_vector*)r2)->val[1];
                                  r2 = ((pxll_vector*)r2)->val[7];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1534; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                  L1534:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = ((object*) lenv) [2];
                                  r0[2] = r1;
                                  r1 = (object *) constructed_36[0];
                                  r0[3] = r1;
                                  r1 = top[21];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
                                L1533:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_get_tvars_2191_i1131_i4622_i4727; r2[2] = lenv;
                                r1[7] = r2;
                                // def get-alt-scheme_2192_i1131_i4622_i4727
                                goto L1535;
                                FUN_get_alt_scheme_2192_i1131_i4622_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 6);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (TC_TUPLE, 4);
                                  r2 = ((object****) lenv) [1][1][1][4];
                                  r1[2] = r2;
                                  r2 = ((object**) lenv) [1][2];
                                  r1[3] = r2;
                                  r2 = (object*) &constructed_677;
                                  r1[4] = r2;
                                  r2 = ((object****) lenv) [1][1][1][4];
                                  r2 = ((pxll_vector*)r2)->val[1];
                                  r2 = ((pxll_vector*)r2)->val[3];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1536; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                  L1536:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[2] = r1;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r2 = ((object****) lenv) [1][1][1][3];
                                  r1[2] = r2;
                                  r2 = ((object****) lenv) [1][1][1][3];
                                  r2 = ((pxll_vector*)r2)->val[1];
                                  r2 = ((pxll_vector*)r2)->val[7];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1537; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                  L1537:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[3] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object*) lenv) [3];
                                  r1[2] = r2;
                                  r2 = (object *) constructed_36[0];
                                  r1[3] = r2;
                                  r2 = top[21];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1538; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                  L1538:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[4] = r1;
                                  r1 = allocate (TC_TUPLE, 3);
                                  r2 = ((object****) lenv) [1][1][1][2];
                                  r1[2] = r2;
                                  r2 = ((object*) lenv) [4];
                                  r1[3] = r2;
                                  r2 = top[104];
                                  t = allocate (TC_SAVE, 4);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1539; t[4] = r0; k = t;
                                  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                                  L1539:
                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                  r1 = result;
                                  r0[5] = r1;
                                  r1 = allocate (UOTAG(2), 2);
                                  r2 = ((object*) lenv) [4];
                                  r1[1] = r2;
                                  r2 = allocate (TC_TUPLE, 2);
                                  r2[1] = lenv; lenv = r2;
                                  r3 = ((object**) lenv) [1][2];
                                  r3 = ((pxll_vector*)r3)->val[3];
                                  r2[2] = r3;
                                  r2 = allocate (TC_TUPLE, 3);
                                  r3 = (object *) constructed_97[0];
                                  r2[2] = r3;
                                  r3 = ((object*) lenv) [2];
                                  r4 = ((object**) lenv) [1][5];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r4;
                                  t[2] = r3;
                                  r3 = t;
                                  r2[3] = r3;
                                  r3 = top[104];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1540; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                  L1540:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  lenv = lenv[1];
                                  r1[2] = r2;
                                  r0[6] = r1;
                                  r0 = ((object*) lenv) [6];
                                  PXLL_RETURN(0);
                                L1535:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_get_alt_scheme_2192_i1131_i4622_i4727; r2[2] = lenv;
                                r1[8] = r2;
                                r1 = allocate (UOTAG(7), 8);
                                r2 = ((object*) lenv) [3];
                                r1[1] = r2;
                                r2 = ((object*) lenv) [2];
                                r1[2] = r2;
                                r2 = ((object*) lenv) [8];
                                r1[3] = r2;
                                r2 = ((object*) lenv) [5];
                                r1[4] = r2;
                                r2 = ((object*) lenv) [6];
                                r1[5] = r2;
                                r2 = ((object*) lenv) [7];
                                r1[6] = r2;
                                r2 = ((object*) lenv) [4];
                                r1[7] = r2;
                                r2 = ((object**) lenv) [1][2];
                                r1[8] = r2;
                                lenv = lenv[1];
                                r0[6] = r1;
                                r0 = allocate (TC_TUPLE, 3);
                                // def lambda_13070_i1130_i4622_i4727
                                goto L1541;
                                FUN_lambda_13070_i1130_i4622_i4727:
                                  check_heap (0);
                                  r0 = ((object*) lenv) [2];
                                  switch (get_case (r0)) {
                                    case (UOTAG(0)): {
                                      r0 = allocate (TC_TUPLE, 2);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = ((object**) lenv) [1][2];
                                      r1 = UOBJ_GET(r1,0);
                                      r0[2] = r1;
                                      r0 = ((object*) lenv) [2];
                                      switch (get_case (r0)) {
                                        case (TC_PAIR): {
                                          r0 = allocate (TC_TUPLE, 3);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object**) lenv) [1][2];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r1 = ((object**) lenv) [1][2];
                                          r1 = UOBJ_GET(r1,1);
                                          r0[3] = r1;
                                          r0 = ((object*) lenv) [2];
                                          switch (get_case (r0)) {
                                            case (UOTAG(9)): {
                                              r0 = allocate (TC_TUPLE, 3);
                                              r0[1] = lenv; lenv = r0;
                                              r1 = ((object**) lenv) [1][2];
                                              r1 = UOBJ_GET(r1,0);
                                              r0[2] = r1;
                                              r1 = ((object**) lenv) [1][2];
                                              r1 = UOBJ_GET(r1,1);
                                              r0[3] = r1;
                                              r0 = (object *) constructed_49[0];
                                              r1 = ((object*) lenv) [2];
                                              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                                                r0 = allocate (TC_TUPLE, 2);
                                                r1 = allocate (TC_TUPLE, 4);
                                                r2 = ((object*****) lenv) [1][1][1][1][3];
                                                r1[2] = r2;
                                                r2 = ((object*) lenv) [3];
                                                r1[3] = r2;
                                                r2 = ((object**) lenv) [1][3];
                                                r1[4] = r2;
                                                r2 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][22];
                                                t = allocate (TC_SAVE, 4);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1542; t[4] = r0; k = t;
                                                r1[1] = r2[2]; lenv = r1; goto FUN_make_alt_1985_i4622_i4727;
                                                L1542:
                                                r0 = k[4]; lenv = k[2]; k = k[1];
                                                r1 = result;
                                                r0[2] = r1;
                                                r1 = ((object*****) lenv) [1][1][1][1][6];
                                                r1 = ((pxll_vector*)r1)->val[0];
                                                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                                              } else {
                                                lenv = ((object ***)lenv)[1][1][1];
                                                goto fatbar_148;
                                              }
                                              PXLL_RETURN(0);
                                            } break;
                                            default: {
                                              lenv = ((object **)lenv)[1][1];
                                              goto fatbar_148;
                                            }
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object *)lenv)[1];
                                          goto fatbar_148;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      goto fatbar_148;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                  goto fatbar_148_over;
                                  fatbar_148:
                                  r0 = allocate (TC_TUPLE, 3);
                                  r1 = (object*) &constructed_676;
                                  r0[2] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[3] = r1;
                                  r1 = top[15];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                                  fatbar_148_over: ;
                                  PXLL_RETURN(0);
                                L1541:
                                r1 = allocate (TC_CLOSURE, 2);
                                r1[1] = &&FUN_lambda_13070_i1130_i4622_i4727; r1[2] = lenv;
                                r0[2] = r1;
                                r1 = ((object**) lenv) [1][3];
                                r0[3] = r1;
                                r1 = top[35];
                                t = allocate (TC_SAVE, 3);
                                t[1] = k; t[2] = lenv; t[3] = &&L1543; ; k = t;
                                r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
                                L1543:
                                ; lenv = k[2]; k = k[1];
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                                r1 = ((pxll_vector*)r1)->val[2];
                                r0[2] = r1;
                                r0 = ((object**) lenv) [1][6];
                                r1 = ((object*) lenv) [2];
                                r2 = ((object**) lenv) [1][2];
                                t = alloc_no_clear (UOTAG(1),3);
                                t[1] = r2;
                                t[2] = r0;
                                t[3] = r1;
                                r0 = t;
                                lenv = lenv[1];
                                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                                ((pxll_vector*)r1)->val[2] = r0;
                                lenv = lenv[1];
                              } break;
                              default: {
                                lenv = ((object *)lenv)[1];
                                goto fatbar_147;
                              }
                            }
                            lenv = lenv[1];
                          } break;
                          default: {
                            goto fatbar_147;
                          }
                        }
                        goto fatbar_147_over;
                        fatbar_147:
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = (object*) &constructed_679;
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][3];
                        r0[3] = r1;
                        r1 = top[15];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1544; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                        L1544:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                        fatbar_147_over: ;
                        r0 = ((object*****) lenv) [1][1][1][1][3];
                        r1 = ((object****) lenv) [1][1][1][3];
                        lenv = ((object ****)lenv)[1][1][1][1];
                        lenv[2] = r1;
                        lenv[3] = r0;
                        goto FUN_recur_2016_i4622_i4727;
                      } else {
                        lenv = ((object ***)lenv)[1][1][1];
                        goto fatbar_143;
                      }
                      PXLL_RETURN(0);
                      fatbar_144_over: ;
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_143;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_143;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_143;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_143_over;
          fatbar_143:
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          r1 = ((object*) lenv) [3];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_recur_2016_i4622_i4727;
          fatbar_143_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1506:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_recur_2016_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = (object *) constructed_664[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_recur_2016_i4622_i4727;
  L1505:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_find_declarations_1965_i4622_i4727; r2[2] = lenv;
  r1[5] = r2;
  // def find-definitions_1966_i4622_i4727
  goto L1545;
  FUN_find_definitions_1966_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def recur_2027_i4622_i4727
    goto L1546;
    FUN_recur_2027_i4622_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [4];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 3);
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1547; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1547:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1548; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1548:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UOTAG(1)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = (object *) constructed_663[0];
                      r1 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object***) lenv) [1][1][3];
                        switch (get_case (r1)) {
                          case (TC_PAIR): {
                            r1 = allocate (TC_TUPLE, 3);
                            r1[1] = lenv; lenv = r1;
                            r2 = ((object****) lenv) [1][1][1][3];
                            r2 = UOBJ_GET(r2,0);
                            r1[2] = r2;
                            r2 = ((object****) lenv) [1][1][1][3];
                            r2 = UOBJ_GET(r2,1);
                            r1[3] = r2;
                            r1 = ((object*) lenv) [2];
                            switch (get_case (r1)) {
                              case (UOTAG(1)): {
                                r1 = allocate (TC_TUPLE, 2);
                                r1[1] = lenv; lenv = r1;
                                r2 = ((object**) lenv) [1][2];
                                r2 = UOBJ_GET(r2,0);
                                r1[2] = r2;
                                r1 = allocate (TC_TUPLE, 4);
                                r2 = (object *) constructed_660[0];
                                t = alloc_no_clear (UOTAG(1),1);
                                t[1] = r2;
                                r2 = t;
                                r1[2] = r2;
                                r2 = ((object**) lenv) [1][3];
                                r1[3] = r2;
                                r2 = top[89];
                                r1[4] = r2;
                                r2 = top[24];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L1549; t[4] = r0; k = t;
                                r1[1] = r2[2]; lenv = r1; goto FUN_member_question_318;
                                L1549:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                if PXLL_IS_TRUE(r1) {
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = allocate (TC_TUPLE, 5);
                                  r3 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][2];
                                  r2[2] = r3;
                                  r3 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][8];
                                  r2[3] = r3;
                                  r3 = (object *) constructed_657[0];
                                  r2[4] = r3;
                                  r3 = ((object***) lenv) [1][1][3];
                                  r2[5] = r3;
                                  r3 = top[116];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1550; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_compile_pattern_478;
                                  L1550:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[2] = r2;
                                  r1 = ((object*) lenv) [2];
                                  switch (get_case_noint (r1)) {
                                    default: {
                                      r1 = allocate (TC_TUPLE, 3);
                                      r1[1] = lenv; lenv = r1;
                                      r2 = ((object**) lenv) [1][2];
                                      r2 = UOBJ_GET(r2,0);
                                      r1[2] = r2;
                                      r2 = ((object**) lenv) [1][2];
                                      r2 = UOBJ_GET(r2,1);
                                      r1[3] = r2;
                                      r1 = allocate (UOTAG(1), 2);
                                      r2 = ((object***) lenv) [1][1][2];
                                      r1[1] = r2;
                                      r2 = allocate (TC_TUPLE, 11);
                                      r2[1] = lenv; lenv = r2;
                                      r3 = (object *) constructed_601[0];
                                      t = alloc_no_clear (UOTAG(1),1);
                                      t[1] = r3;
                                      r3 = t;
                                      r2[2] = r3;
                                      r3 = ((object****) lenv) [1][1][1][2];
                                      t = alloc_no_clear (UOTAG(1),1);
                                      t[1] = r3;
                                      r3 = t;
                                      r2[3] = r3;
                                      r3 = allocate (TC_TUPLE, 3);
                                      r4 = top[3];
                                      r3[2] = r4;
                                      r4 = ((object**) lenv) [1][2];
                                      r3[3] = r4;
                                      r4 = top[32];
                                      t = allocate (TC_SAVE, 6);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1551; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                      r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
                                      L1551:
                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                      r3 = result;
                                      r2[4] = r3;
                                      r3 = ((object*) lenv) [4];
                                      t = alloc_no_clear (UOTAG(0),1);
                                      t[1] = r3;
                                      r3 = t;
                                      r2[5] = r3;
                                      r3 = allocate (TC_TUPLE, 2);
                                      r4 = ((object**) lenv) [1][3];
                                      r3[2] = r4;
                                      r4 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][8];
                                      t = allocate (TC_SAVE, 6);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1552; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                      r3[1] = r4[2]; lenv = r3; goto FUN_expand_1968_i4622_i4727;
                                      L1552:
                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                      r3 = result;
                                      r2[6] = r3;
                                      r3 = (object*)TC_NIL;
                                      r2[7] = r3;
                                      r3 = ((object*) lenv) [7];
                                      r4 = ((object*) lenv) [6];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r4;
                                      t[2] = r3;
                                      r3 = t;
                                      r2[8] = r3;
                                      r3 = ((object*) lenv) [8];
                                      r4 = ((object*) lenv) [5];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r4;
                                      t[2] = r3;
                                      r3 = t;
                                      r2[9] = r3;
                                      r3 = ((object*) lenv) [9];
                                      r4 = ((object*) lenv) [3];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r4;
                                      t[2] = r3;
                                      r3 = t;
                                      r2[10] = r3;
                                      r3 = ((object*) lenv) [10];
                                      r4 = ((object*) lenv) [2];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r4;
                                      t[2] = r3;
                                      r3 = t;
                                      r2[11] = r3;
                                      r2 = ((object*) lenv) [11];
                                      t = alloc_no_clear (UOTAG(0),1);
                                      t[1] = r2;
                                      r2 = t;
                                      lenv = lenv[1];
                                      r1[2] = r2;
                                      lenv = lenv[1];
                                    } break;
                                  }
                                  lenv = lenv[1];
                                } else {
                                  r1 = allocate (TC_TUPLE, 3);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = allocate (TC_TUPLE, 2);
                                  r3 = ((object***) lenv) [1][1][3];
                                  r2[2] = r3;
                                  r3 = top[22];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1553; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
                                  L1553:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[2] = r2;
                                  r2 = ((object*) lenv) [2];
                                  r3 = (object *) 3;
                                  r2 = PXLL_TEST(unbox(r3)==unbox(r2));
                                  r1[3] = r2;
                                  r1 = (object *) 6;
                                  r2 = ((object*) lenv) [3];
                                  r1 = PXLL_TEST(r2==r1);
                                  lenv = lenv[1];
                                  if PXLL_IS_TRUE(r1) {
                                    r1 = allocate (TC_TUPLE, 3);
                                    r2 = (object*) &constructed_658;
                                    r1[2] = r2;
                                    r2 = ((object*) lenv) [2];
                                    r1[3] = r2;
                                    r2 = top[15];
                                    t = allocate (TC_SAVE, 4);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1554; t[4] = r0; k = t;
                                    r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                                    L1554:
                                    r0 = k[4]; lenv = k[2]; k = k[1];
                                    r1 = result;
                                  } else {
                                    r1 = allocate (UOTAG(1), 2);
                                    r2 = ((object*) lenv) [2];
                                    r1[1] = r2;
                                    r2 = allocate (TC_TUPLE, 2);
                                    r3 = ((object**) lenv) [1][3];
                                    r2[2] = r3;
                                    r3 = top[19];
                                    t = allocate (TC_SAVE, 5);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1555; t[4] = r0; t[5] = r1; k = t;
                                    r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                                    L1555:
                                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                    r2 = result;
                                    r1[2] = r2;
                                  }
                                }
                                lenv = lenv[1];
                              } break;
                              case (UOTAG(0)): {
                                r1 = allocate (TC_TUPLE, 2);
                                r1[1] = lenv; lenv = r1;
                                r2 = ((object**) lenv) [1][2];
                                r2 = UOBJ_GET(r2,0);
                                r1[2] = r2;
                                r1 = ((object*) lenv) [2];
                                switch (get_case (r1)) {
                                  case (TC_PAIR): {
                                    r1 = allocate (TC_TUPLE, 3);
                                    r1[1] = lenv; lenv = r1;
                                    r2 = ((object**) lenv) [1][2];
                                    r2 = UOBJ_GET(r2,0);
                                    r1[2] = r2;
                                    r2 = ((object**) lenv) [1][2];
                                    r2 = UOBJ_GET(r2,1);
                                    r1[3] = r2;
                                    r1 = ((object*) lenv) [2];
                                    switch (get_case (r1)) {
                                      case (UOTAG(1)): {
                                        r1 = allocate (TC_TUPLE, 2);
                                        r1[1] = lenv; lenv = r1;
                                        r2 = ((object**) lenv) [1][2];
                                        r2 = UOBJ_GET(r2,0);
                                        r1[2] = r2;
                                        r1 = allocate (UOTAG(1), 2);
                                        r2 = ((object*) lenv) [2];
                                        r1[1] = r2;
                                        r2 = allocate (TC_TUPLE, 10);
                                        r2[1] = lenv; lenv = r2;
                                        r3 = (object *) constructed_601[0];
                                        t = alloc_no_clear (UOTAG(1),1);
                                        t[1] = r3;
                                        r3 = t;
                                        r2[2] = r3;
                                        r3 = ((object**) lenv) [1][2];
                                        t = alloc_no_clear (UOTAG(1),1);
                                        t[1] = r3;
                                        r3 = t;
                                        r2[3] = r3;
                                        r3 = ((object***) lenv) [1][1][3];
                                        t = alloc_no_clear (UOTAG(0),1);
                                        t[1] = r3;
                                        r3 = t;
                                        r2[4] = r3;
                                        r3 = allocate (TC_TUPLE, 3);
                                        r4 = ((object*****) lenv) [1][1][1][1][3];
                                        r3[2] = r4;
                                        r4 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][4];
                                        r3[3] = r4;
                                        r4 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][6];
                                        t = allocate (TC_SAVE, 6);
                                        t[1] = k; t[2] = lenv; t[3] = &&L1556; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                        r3[1] = r4[2]; lenv = r3; goto FUN_find_definitions_1966_i4622_i4727;
                                        L1556:
                                        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                        r3 = result;
                                        r2[5] = r3;
                                        r3 = (object*)TC_NIL;
                                        r2[6] = r3;
                                        r3 = ((object*) lenv) [6];
                                        r4 = ((object*) lenv) [5];
                                        t = alloc_no_clear (TC_PAIR,2);
                                        t[1] = r4;
                                        t[2] = r3;
                                        r3 = t;
                                        r2[7] = r3;
                                        r3 = ((object*) lenv) [7];
                                        r4 = ((object*) lenv) [4];
                                        t = alloc_no_clear (TC_PAIR,2);
                                        t[1] = r4;
                                        t[2] = r3;
                                        r3 = t;
                                        r2[8] = r3;
                                        r3 = ((object*) lenv) [8];
                                        r4 = ((object*) lenv) [3];
                                        t = alloc_no_clear (TC_PAIR,2);
                                        t[1] = r4;
                                        t[2] = r3;
                                        r3 = t;
                                        r2[9] = r3;
                                        r3 = ((object*) lenv) [9];
                                        r4 = ((object*) lenv) [2];
                                        t = alloc_no_clear (TC_PAIR,2);
                                        t[1] = r4;
                                        t[2] = r3;
                                        r3 = t;
                                        r2[10] = r3;
                                        r2 = ((object*) lenv) [10];
                                        t = alloc_no_clear (UOTAG(0),1);
                                        t[1] = r2;
                                        r2 = t;
                                        lenv = lenv[1];
                                        r1[2] = r2;
                                        lenv = lenv[1];
                                      } break;
                                      default: {
                                        lenv = ((object ***)lenv)[1][1][1];
                                        goto fatbar_142;
                                      }
                                    }
                                    lenv = lenv[1];
                                  } break;
                                  default: {
                                    lenv = ((object **)lenv)[1][1];
                                    goto fatbar_142;
                                  }
                                }
                                lenv = lenv[1];
                              } break;
                              default: {
                                lenv = ((object *)lenv)[1];
                                goto fatbar_142;
                              }
                            }
                            lenv = lenv[1];
                          } break;
                          default: {
                            goto fatbar_142;
                          }
                        }
                        goto fatbar_142_over;
                        fatbar_142:
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = (object*) &constructed_661;
                        r1[2] = r2;
                        r2 = ((object***) lenv) [1][1][3];
                        r1[3] = r2;
                        r2 = top[15];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L1557; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                        L1557:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        fatbar_142_over: ;
                        r0[2] = r1;
                        r0 = ((object******) lenv) [1][1][1][1][1][2];
                        r1 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        lenv = lenv[1];
                        r1 = ((object*****) lenv) [1][1][1][1][3];
                        r2 = ((object****) lenv) [1][1][1][3];
                        lenv = ((object ****)lenv)[1][1][1][1];
                        lenv[2] = r0;
                        lenv[3] = r1;
                        lenv[4] = r2;
                        goto FUN_recur_2027_i4622_i4727;
                      } else {
                        lenv = ((object ***)lenv)[1][1][1];
                        goto fatbar_141;
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_141;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_141;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_141;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_141_over;
          fatbar_141:
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          r1 = ((object*) lenv) [3];
          r2 = ((object**) lenv) [1][2];
          lenv = ((object *)lenv)[1];
          lenv[2] = r2;
          lenv[3] = r0;
          lenv[4] = r1;
          goto FUN_recur_2027_i4622_i4727;
          fatbar_141_over: ;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1546:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_recur_2027_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r1 = (object *) constructed_656[0];
    r0[2] = r1;
    r1 = (object *) constructed_655[0];
    r0[3] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_recur_2027_i4622_i4727;
  L1545:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_find_definitions_1966_i4622_i4727; r2[2] = lenv;
  r1[6] = r2;
  // def expand-field_1967_i4622_i4727
  goto L1558;
  FUN_expand_field_1967_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,0);
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = UOBJ_GET(r1,1);
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [3];
    r1[2] = r2;
    r2 = ((object***) lenv) [1][1][8];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1559; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
    L1559:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r0 = ((object*) lenv) [4];
    r1 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(0),2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    PXLL_RETURN(0);
    PXLL_RETURN(0);
  L1558:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_field_1967_i4622_i4727; r2[2] = lenv;
  r1[7] = r2;
  // def expand_1968_i4622_i4727
  goto L1560;
  FUN_expand_1968_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(2)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(3)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(4)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      case (UITAG(6)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][9];
        r0[1] = r1[2]; lenv = r0; goto FUN_maybe_expand_1969_i4622_i4727;
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object***) lenv) [1][1][8];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1561; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1561:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(7),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(8)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object***) lenv) [1][1][7];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1562; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1562:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        t = alloc_no_clear (UOTAG(8),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(9)): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][8];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1563; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
        L1563:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [4];
        t = alloc_no_clear (UOTAG(10),2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1560:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_1968_i4622_i4727; r2[2] = lenv;
  r1[8] = r2;
  // def maybe-expand_1969_i4622_i4727
  goto L1564;
  FUN_maybe_expand_1969_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) constructed_650[0];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][28];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = top[65];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1565; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
            L1565:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (UOTAG(0)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object***) lenv) [1][1][3];
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object******) lenv) [1][1][1][1][1][2];
                r2 = ((pxll_vector*)r2)->val[8];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[65];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1566; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
                L1566:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(0)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object*****) lenv) [1][1][1][1][2];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r2;
                    r2 = t;
                    r1[2] = r2;
                    r2 = ((object*) lenv) [2];
                    r2 = ((pxll_vector*)r2)->val[0];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1567; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto *r2[1];
                    L1567:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = ((object******) lenv) [1][1][1][1][1][8];
                    r0[1] = r1[2]; lenv = r0; goto FUN_expand_1968_i4622_i4727;
                  } break;
                  default: {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 3);
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    r1[2] = r2;
                    r2 = ((object****) lenv) [1][1][1][3];
                    r1[3] = r2;
                    r2 = top[32];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1568; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                    L1568:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = ((object*) lenv) [2];
                    r2 = ((object****) lenv) [1][1][1][2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } break;
                }
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
          } break;
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(1)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = (object *) constructed_654[0];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                      r0 = ((object**) lenv) [1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (UOTAG(1)): {
                              r0 = allocate (TC_TUPLE, 2);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r0 = ((object**) lenv) [1][3];
                              switch (get_case (r0)) {
                                case (TC_PAIR): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object***) lenv) [1][1][3];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = ((object***) lenv) [1][1][3];
                                  r1 = UOBJ_GET(r1,1);
                                  r0[3] = r1;
                                  r0 = ((object*) lenv) [3];
                                  switch (get_case (r0)) {
                                    case (TC_NIL): {
                                      r0 = allocate (TC_TUPLE, 8);
                                      r0[1] = lenv; lenv = r0;
                                      r1 = (object *) constructed_652[0];
                                      r2 = ((object**) lenv) [1][2];
                                      t = alloc_no_clear (UOTAG(10),2);
                                      t[1] = r2;
                                      t[2] = r1;
                                      r1 = t;
                                      r0[2] = r1;
                                      r1 = ((object***) lenv) [1][1][2];
                                      r2 = ((object*) lenv) [2];
                                      t = alloc_no_clear (UOTAG(10),2);
                                      t[1] = r2;
                                      t[2] = r1;
                                      r1 = t;
                                      r0[3] = r1;
                                      r1 = (object*)TC_NIL;
                                      r0[4] = r1;
                                      r1 = ((object*) lenv) [4];
                                      r2 = ((object**) lenv) [1][2];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r2;
                                      t[2] = r1;
                                      r1 = t;
                                      r0[5] = r1;
                                      r1 = ((object*) lenv) [5];
                                      r2 = ((object*) lenv) [3];
                                      t = alloc_no_clear (TC_PAIR,2);
                                      t[1] = r2;
                                      t[2] = r1;
                                      r1 = t;
                                      r0[6] = r1;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][8];
                                      r1[2] = r2;
                                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                      r1[3] = r2;
                                      r2 = top[32];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1569; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                                      L1569:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[7] = r1;
                                      r1 = allocate (TC_TUPLE, 3);
                                      r2 = allocate (TC_TUPLE, 3);
                                      r3 = ((object*) lenv) [6];
                                      r2[2] = r3;
                                      r3 = (object *) constructed_36[0];
                                      r2[3] = r3;
                                      r3 = top[21];
                                      t = allocate (TC_SAVE, 5);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1570; t[4] = r0; t[5] = r1; k = t;
                                      r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                                      L1570:
                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                      r2 = result;
                                      r1[2] = r2;
                                      r2 = ((object*) lenv) [7];
                                      r1[3] = r2;
                                      r2 = top[21];
                                      t = allocate (TC_SAVE, 4);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1571; t[4] = r0; k = t;
                                      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                                      L1571:
                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                      r1 = result;
                                      r0[8] = r1;
                                      r0 = ((object*) lenv) [8];
                                      t = alloc_no_clear (UOTAG(0),1);
                                      t[1] = r0;
                                      r0 = t;
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                      goto fatbar_140;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  goto fatbar_140;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_140;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_140;
                        }
                      }
                      PXLL_RETURN(0);
                    } else {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_140;
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_140;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_140;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_140;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_140_over;
        fatbar_140:
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object****) lenv) [1][1][1][8];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1572; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1572:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        PXLL_RETURN(0);
        fatbar_140_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1564:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_maybe_expand_1969_i4622_i4727; r2[2] = lenv;
  r1[9] = r2;
  // def expand-if_1970_i4622_i4727
  goto L1573;
  FUN_expand_if_1970_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (TC_NIL): {
                r0 = allocate (TC_TUPLE, 10);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object***) lenv) [1][1][2];
                r1[2] = r2;
                r2 = ((object*****) lenv) [1][1][1][1][8];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1574; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                L1574:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object**) lenv) [1][2];
                r1[2] = r2;
                r2 = ((object*****) lenv) [1][1][1][1][8];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1575; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                L1575:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = (object*)UITAG(6);
                r0[4] = r1;
                r1 = (object*)TC_NIL;
                r0[5] = r1;
                r1 = ((object*) lenv) [5];
                r2 = ((object*) lenv) [4];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[6] = r1;
                r1 = ((object*) lenv) [6];
                r2 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[7] = r1;
                r1 = ((object*) lenv) [7];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[8] = r1;
                r1 = (object *) constructed_595[0];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                r0[9] = r1;
                r1 = ((object*) lenv) [8];
                r2 = ((object*) lenv) [9];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[10] = r1;
                r0 = ((object*) lenv) [10];
                t = alloc_no_clear (UOTAG(0),1);
                t[1] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = allocate (TC_TUPLE, 10);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object****) lenv) [1][1][1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1576; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                    L1576:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object***) lenv) [1][1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1577; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                    L1577:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1578; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                    L1578:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[4] = r1;
                    r1 = (object*)TC_NIL;
                    r0[5] = r1;
                    r1 = ((object*) lenv) [5];
                    r2 = ((object*) lenv) [4];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[6] = r1;
                    r1 = ((object*) lenv) [6];
                    r2 = ((object*) lenv) [3];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[7] = r1;
                    r1 = ((object*) lenv) [7];
                    r2 = ((object*) lenv) [2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[8] = r1;
                    r1 = (object *) constructed_595[0];
                    t = alloc_no_clear (UOTAG(1),1);
                    t[1] = r1;
                    r1 = t;
                    r0[9] = r1;
                    r1 = ((object*) lenv) [8];
                    r2 = ((object*) lenv) [9];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[10] = r1;
                    r0 = ((object*) lenv) [10];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object ***)lenv)[1][1][1];
                    goto fatbar_139;
                  }
                }
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_139;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_139;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_139_over;
    fatbar_139:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_649;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_139_over: ;
    PXLL_RETURN(0);
  L1573:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_if_1970_i4622_i4727; r2[2] = lenv;
  r1[10] = r2;
  // def expand-set!_1971_i4622_i4727
  goto L1579;
  FUN_expand_set_bang_1971_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(10)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object**) lenv) [1][3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = allocate (TC_TUPLE, 10);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object***) lenv) [1][1][3];
                    t = alloc_no_clear (UOTAG(1),1);
                    t[1] = r1;
                    r1 = t;
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object***) lenv) [1][1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1580; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                    L1580:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1581; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                    L1581:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[4] = r1;
                    r1 = (object*)TC_NIL;
                    r0[5] = r1;
                    r1 = ((object*) lenv) [5];
                    r2 = ((object*) lenv) [4];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[6] = r1;
                    r1 = ((object*) lenv) [6];
                    r2 = ((object*) lenv) [3];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[7] = r1;
                    r1 = ((object*) lenv) [7];
                    r2 = ((object*) lenv) [2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[8] = r1;
                    r1 = (object *) constructed_483[0];
                    t = alloc_no_clear (UOTAG(1),1);
                    t[1] = r1;
                    r1 = t;
                    r0[9] = r1;
                    r1 = ((object*) lenv) [8];
                    r2 = ((object*) lenv) [9];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[10] = r1;
                    r0 = ((object*) lenv) [10];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object ***)lenv)[1][1][1];
                    goto fatbar_138;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_138;
              }
            }
            PXLL_RETURN(0);
          } break;
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(1)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = (object *) constructed_295[0];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                      r0 = ((object**) lenv) [1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_PAIR): {
                                  r0 = allocate (TC_TUPLE, 3);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,0);
                                  r0[2] = r1;
                                  r1 = ((object**) lenv) [1][3];
                                  r1 = UOBJ_GET(r1,1);
                                  r0[3] = r1;
                                  r0 = ((object*) lenv) [3];
                                  switch (get_case (r0)) {
                                    case (TC_NIL): {
                                      r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                      switch (get_case (r0)) {
                                        case (TC_PAIR): {
                                          r0 = allocate (TC_TUPLE, 3);
                                          r0[1] = lenv; lenv = r0;
                                          r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                          r1 = UOBJ_GET(r1,0);
                                          r0[2] = r1;
                                          r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                                          r1 = UOBJ_GET(r1,1);
                                          r0[3] = r1;
                                          r0 = ((object*) lenv) [3];
                                          switch (get_case (r0)) {
                                            case (TC_NIL): {
                                              r0 = allocate (TC_TUPLE, 11);
                                              r0[1] = lenv; lenv = r0;
                                              r1 = allocate (TC_TUPLE, 2);
                                              r2 = ((object****) lenv) [1][1][1][2];
                                              r1[2] = r2;
                                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][8];
                                              t = allocate (TC_SAVE, 4);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1582; t[4] = r0; k = t;
                                              r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                                              L1582:
                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                              r1 = result;
                                              r0[2] = r1;
                                              r1 = allocate (TC_TUPLE, 2);
                                              r2 = ((object***) lenv) [1][1][2];
                                              r1[2] = r2;
                                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][8];
                                              t = allocate (TC_SAVE, 4);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1583; t[4] = r0; k = t;
                                              r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                                              L1583:
                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                              r1 = result;
                                              r0[3] = r1;
                                              r1 = allocate (TC_TUPLE, 2);
                                              r2 = ((object**) lenv) [1][2];
                                              r1[2] = r2;
                                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][8];
                                              t = allocate (TC_SAVE, 4);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1584; t[4] = r0; k = t;
                                              r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                                              L1584:
                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                              r1 = result;
                                              r0[4] = r1;
                                              r1 = (object*)TC_NIL;
                                              r0[5] = r1;
                                              r1 = ((object*) lenv) [5];
                                              r2 = ((object*) lenv) [4];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r2;
                                              t[2] = r1;
                                              r1 = t;
                                              r0[6] = r1;
                                              r1 = ((object*) lenv) [6];
                                              r2 = ((object*) lenv) [3];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r2;
                                              t[2] = r1;
                                              r1 = t;
                                              r0[7] = r1;
                                              r1 = ((object*) lenv) [7];
                                              r2 = ((object*) lenv) [2];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r2;
                                              t[2] = r1;
                                              r1 = t;
                                              r0[8] = r1;
                                              r1 = ((object*) lenv) [8];
                                              r2 = ((object*****) lenv) [1][1][1][1][2];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r2;
                                              t[2] = r1;
                                              r1 = t;
                                              r0[9] = r1;
                                              r1 = (object *) constructed_286[0];
                                              t = alloc_no_clear (UOTAG(1),1);
                                              t[1] = r1;
                                              r1 = t;
                                              r0[10] = r1;
                                              r1 = ((object*) lenv) [9];
                                              r2 = ((object*) lenv) [10];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r2;
                                              t[2] = r1;
                                              r1 = t;
                                              r0[11] = r1;
                                              r0 = ((object*) lenv) [11];
                                              t = alloc_no_clear (UOTAG(0),1);
                                              t[1] = r0;
                                              r0 = t;
                                              PXLL_RETURN(0);
                                            } break;
                                            default: {
                                              lenv = ((object ********)lenv)[1][1][1][1][1][1][1][1];
                                              goto fatbar_138;
                                            }
                                          }
                                          PXLL_RETURN(0);
                                        } break;
                                        default: {
                                          lenv = ((object *******)lenv)[1][1][1][1][1][1][1];
                                          goto fatbar_138;
                                        }
                                      }
                                      PXLL_RETURN(0);
                                    } break;
                                    default: {
                                      lenv = ((object *******)lenv)[1][1][1][1][1][1][1];
                                      goto fatbar_138;
                                    }
                                  }
                                  PXLL_RETURN(0);
                                } break;
                                default: {
                                  lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                  goto fatbar_138;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object *****)lenv)[1][1][1][1][1];
                              goto fatbar_138;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ****)lenv)[1][1][1][1];
                          goto fatbar_138;
                        }
                      }
                      PXLL_RETURN(0);
                    } else {
                      lenv = ((object ****)lenv)[1][1][1][1];
                      goto fatbar_138;
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object ***)lenv)[1][1][1];
                    goto fatbar_138;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_138;
              }
            }
            PXLL_RETURN(0);
          } break;
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = allocate (TC_TUPLE, 8);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object***) lenv) [1][1][2];
                    t = alloc_no_clear (UOTAG(1),1);
                    t[1] = r1;
                    r1 = t;
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = ((object******) lenv) [1][1][1][1][1][8];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1585; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                    L1585:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r1 = (object*)TC_NIL;
                    r0[4] = r1;
                    r1 = ((object*) lenv) [4];
                    r2 = ((object*) lenv) [3];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[5] = r1;
                    r1 = ((object*) lenv) [5];
                    r2 = ((object*) lenv) [2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[6] = r1;
                    r1 = (object *) constructed_597[0];
                    t = alloc_no_clear (UOTAG(1),1);
                    t[1] = r1;
                    r1 = t;
                    r0[7] = r1;
                    r1 = ((object*) lenv) [6];
                    r2 = ((object*) lenv) [7];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[8] = r1;
                    r0 = ((object*) lenv) [8];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object ***)lenv)[1][1][1];
                    goto fatbar_138;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_138;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_138;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_138;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_138_over;
    fatbar_138:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_648;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_138_over: ;
    PXLL_RETURN(0);
  L1579:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_set_bang_1971_i4622_i4727; r2[2] = lenv;
  r1[11] = r2;
  // def expand-begin_1972_i4622_i4727
  goto L1586;
  FUN_expand_begin_1972_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_647;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_647;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object***) lenv) [1][1][8];
            r0[1] = r1[2]; lenv = r0; goto FUN_expand_1968_i4622_i4727;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_137;
          }
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
    goto fatbar_137_over;
    fatbar_137:
    r0 = allocate (TC_TUPLE, 4);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object***) lenv) [1][1][8];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = top[32];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1587; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
    L1587:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = (object *) constructed_599[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r2 = ((object*) lenv) [3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[4] = r1;
    r0 = ((object*) lenv) [4];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
    fatbar_137_over: ;
    PXLL_RETURN(0);
  L1586:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_begin_1972_i4622_i4727; r2[2] = lenv;
  r1[12] = r2;
  // def expand-let-splat_1973_i4622_i4727
  goto L1588;
  FUN_expand_let_splat_1973_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 10);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            // def lambda_12028_i4622_i4727
            goto L1589;
            FUN_lambda_12028_i4622_i4727:
              check_heap (0);
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(0)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_NIL): {
                              r0 = allocate (TC_TUPLE, 5);
                              r0[1] = lenv; lenv = r0;
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = ((object**) lenv) [1][2];
                              r1[2] = r2;
                              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][8];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L1590; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                              L1590:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[2] = r1;
                              r1 = (object*)TC_NIL;
                              r0[3] = r1;
                              r1 = ((object*) lenv) [3];
                              r2 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r2;
                              t[2] = r1;
                              r1 = t;
                              r0[4] = r1;
                              r1 = ((object*) lenv) [4];
                              r2 = ((object***) lenv) [1][1][2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r2;
                              t[2] = r1;
                              r1 = t;
                              r0[5] = r1;
                              r0 = ((object*) lenv) [5];
                              t = alloc_no_clear (UOTAG(0),1);
                              t[1] = r0;
                              r0 = t;
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_136;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_136;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_136;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_136;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_136_over;
              fatbar_136:
              r0 = allocate (TC_TUPLE, 3);
              r1 = (object*) &constructed_645;
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = top[15];
              r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
              fatbar_136_over: ;
              PXLL_RETURN(0);
            L1589:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_lambda_12028_i4622_i4727; r2[2] = lenv;
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1591; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L1591:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = ((object*) lenv) [3];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][3];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][4];
            r1[3] = r2;
            r2 = ((object****) lenv) [1][1][1][6];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1592; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_find_definitions_1966_i4622_i4727;
            L1592:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r1 = (object*)TC_NIL;
            r0[6] = r1;
            r1 = ((object*) lenv) [6];
            r2 = ((object*) lenv) [5];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[7] = r1;
            r1 = ((object*) lenv) [7];
            r2 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[8] = r1;
            r1 = (object *) constructed_605[0];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[9] = r1;
            r1 = ((object*) lenv) [8];
            r2 = ((object*) lenv) [9];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[10] = r1;
            r0 = ((object*) lenv) [10];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_135;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_135;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_135_over;
    fatbar_135:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_646;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_135_over: ;
    PXLL_RETURN(0);
  L1588:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_let_splat_1973_i4622_i4727; r2[2] = lenv;
  r1[13] = r2;
  // def expand-lambda_1974_i4622_i4727
  goto L1593;
  FUN_expand_lambda_1974_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = (object *) constructed_585[0];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][4];
        r1[3] = r2;
        r2 = ((object***) lenv) [1][1][6];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1594; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_find_definitions_1966_i4622_i4727;
        L1594:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][16];
        r0[1] = r1[2]; lenv = r0; goto FUN_exp_function_1976_i4622_i4727;
      } break;
      default: {
        goto fatbar_134;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_134_over;
    fatbar_134:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_644;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_134_over: ;
    PXLL_RETURN(0);
  L1593:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_lambda_1974_i4622_i4727; r2[2] = lenv;
  r1[14] = r2;
  // def expand-function_1975_i4622_i4727
  goto L1595;
  FUN_expand_function_1975_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][4];
            r1[3] = r2;
            r2 = ((object****) lenv) [1][1][1][6];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1596; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_find_definitions_1966_i4622_i4727;
            L1596:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = ((object****) lenv) [1][1][1][16];
            r0[1] = r1[2]; lenv = r0; goto FUN_exp_function_1976_i4622_i4727;
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_133;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_133;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_133_over;
    fatbar_133:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_643;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_133_over: ;
    PXLL_RETURN(0);
  L1595:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_function_1975_i4622_i4727; r2[2] = lenv;
  r1[15] = r2;
  // def exp-function_1976_i4622_i4727
  goto L1597;
  FUN_exp_function_1976_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 7);
    r0[1] = lenv; lenv = r0;
    r1 = (object*)TC_NIL;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r2 = ((object**) lenv) [1][4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r1 = ((object*) lenv) [3];
    r2 = ((object**) lenv) [1][3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[4] = r1;
    r1 = ((object*) lenv) [4];
    r2 = ((object**) lenv) [1][2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[5] = r1;
    r1 = (object *) constructed_601[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[6] = r1;
    r1 = ((object*) lenv) [5];
    r2 = ((object*) lenv) [6];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[7] = r1;
    r0 = ((object*) lenv) [7];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L1597:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_exp_function_1976_i4622_i4727; r2[2] = lenv;
  r1[16] = r2;
  // def make-nvget_1978_i4622_i4727
  goto L1598;
  FUN_make_nvget_1978_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 15);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_314[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r2 = ((object**) lenv) [1][2];
    t = alloc_no_clear (UOTAG(9),2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r1 = ((object**) lenv) [1][4];
    t = alloc_no_clear (UOTAG(5),1);
    t[1] = r1;
    r1 = t;
    r0[4] = r1;
    r1 = ((object**) lenv) [1][5];
    t = alloc_no_clear (UOTAG(5),1);
    t[1] = r1;
    r1 = t;
    r0[5] = r1;
    r1 = (object*)TC_NIL;
    r0[6] = r1;
    r1 = ((object*) lenv) [6];
    r2 = ((object*) lenv) [5];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[7] = r1;
    r1 = ((object*) lenv) [7];
    r2 = ((object*) lenv) [4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[8] = r1;
    r1 = ((object*) lenv) [8];
    r2 = ((object*) lenv) [3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[9] = r1;
    r1 = ((object*) lenv) [9];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r1;
    r1 = t;
    r0[10] = r1;
    r1 = ((object**) lenv) [1][6];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[11] = r1;
    r1 = (object*)TC_NIL;
    r0[12] = r1;
    r1 = ((object*) lenv) [12];
    r2 = ((object*) lenv) [11];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[13] = r1;
    r1 = ((object*) lenv) [13];
    r2 = ((object*) lenv) [10];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[14] = r1;
    r1 = ((object*) lenv) [14];
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[15] = r1;
    r0 = ((object*) lenv) [15];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L1598:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_make_nvget_1978_i4622_i4727; r2[2] = lenv;
  r1[17] = r2;
  // def expand-vcase_1979_i4622_i4727
  goto L1599;
  FUN_expand_vcase_1979_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case (r0)) {
                  case (UOTAG(1)): {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 4);
                    r1 = ((object***) lenv) [1][1][2];
                    r0[2] = r1;
                    r1 = ((object*) lenv) [2];
                    r0[3] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r0[4] = r1;
                    r1 = ((object******) lenv) [1][1][1][1][1][19];
                    r0[1] = r1[2]; lenv = r0; goto FUN_expand_vcase_splat_1980_i4622_i4727;
                  } break;
                  default: {
                    lenv = ((object *)lenv)[1];
                    goto fatbar_132;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_132;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_132_over;
            fatbar_132:
            r0 = allocate (TC_TUPLE, 4);
            r1 = (object *) constructed_49[0];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object**) lenv) [1][3];
            r0[4] = r1;
            r1 = ((object****) lenv) [1][1][1][19];
            r0[1] = r1[2]; lenv = r0; goto FUN_expand_vcase_splat_1980_i4622_i4727;
            fatbar_132_over: ;
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_131;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_131;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_131_over;
    fatbar_131:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_642;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_131_over: ;
    PXLL_RETURN(0);
  L1599:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_vcase_1979_i4622_i4727; r2[2] = lenv;
  r1[18] = r2;
  // def expand-vcase*_1980_i4622_i4727
  goto L1600;
  FUN_expand_vcase_splat_1980_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def lambda_12594_i4622_i4727
    goto L1601;
    FUN_lambda_12594_i4622_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) constructed_641[0];
      r0[2] = r1;
      r1 = (object *) constructed_640[0];
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object***) lenv) [1][1][4];
      r1[2] = r2;
      r2 = top[22];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1602; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
      L1602:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def $loop_2156_i4622_i4727
      goto L1603;
      FUN__24loop_2156_i4622_i4727:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][2];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object******) lenv) [1][1][1][1][1][3];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1604; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L1604:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1605; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L1605:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object******) lenv) [1][1][1][1][1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1606; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L1606:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          // def loop_2161_i4622_i4727
          goto L1607;
          FUN_loop_2161_i4622_i4727:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][3];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = allocate (TC_TUPLE, 3);
              r1 = ((object*) lenv) [3];
              r0[2] = r1;
              r1 = (object *) constructed_36[0];
              r0[3] = r1;
              r1 = top[21];
              r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object****) lenv) [1][1][1][2];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[27];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1608; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
              L1608:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = (object *) constructed_639[0];
                  r1 = ((object*) lenv) [2];
                  if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                    r0 = (object *) 3;
                    r1 = ((object***) lenv) [1][1][2];
                    r0 = box((pxll_int)unbox(r1)+unbox(r0));
                    r1 = ((object***) lenv) [1][1][3];
                    lenv = ((object **)lenv)[1][1];
                    lenv[2] = r0;
                    lenv[3] = r1;
                    goto FUN_loop_2161_i4622_i4727;
                  } else {
                    lenv = ((object *)lenv)[1];
                    goto fatbar_130;
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_130;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_130_over;
              fatbar_130:
              r0 = allocate (TC_TUPLE, 6);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 6);
              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
              r1[2] = r2;
              r2 = ((object*****) lenv) [1][1][1][1][4];
              r1[3] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[4] = r2;
              r2 = ((object*****) lenv) [1][1][1][1][3];
              r1[5] = r2;
              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
              r1[6] = r2;
              r2 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][17];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1609; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_make_nvget_1978_i4622_i4727;
              L1609:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = (object*)TC_NIL;
              r0[3] = r1;
              r1 = ((object*) lenv) [3];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[4] = r1;
              r1 = ((object*) lenv) [4];
              r2 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              r0[5] = r1;
              r1 = ((object*) lenv) [5];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r1;
              r1 = t;
              r0[6] = r1;
              r0 = ((object***) lenv) [1][1][3];
              r1 = ((object*) lenv) [6];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              r1 = (object *) 3;
              r2 = ((object**) lenv) [1][2];
              r1 = box((pxll_int)unbox(r2)+unbox(r1));
              lenv = ((object *)lenv)[1];
              lenv[2] = r1;
              lenv[3] = r0;
              goto FUN_loop_2161_i4622_i4727;
              fatbar_130_over: ;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          L1607:
          r2 = allocate (TC_CLOSURE, 2);
          r2[1] = &&FUN_loop_2161_i4622_i4727; r2[2] = lenv;
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object *) 1;
          r1[2] = r2;
          r2 = (object *) constructed_637[0];
          r1[3] = r2;
          r2 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1610; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_loop_2161_i4622_i4727;
          L1610:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
          r0[5] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object*****) lenv) [1][1][1][1][2] = r0;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][5];
          switch (get_case (r1)) {
            case (TC_NIL): {
              r1 = (object *) 262;
            } break;
            default: {
              goto fatbar_129;
            }
          }
          goto fatbar_129_over;
          fatbar_129:
          r1 = (object *) 6;
          fatbar_129_over: ;
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 12);
            r0[1] = lenv; lenv = r0;
            r1 = (object *) constructed_622[0];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[2] = r1;
            r1 = ((object**) lenv) [1][5];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            r0[3] = r1;
            r1 = (object*)TC_NIL;
            r0[4] = r1;
            r1 = ((object*) lenv) [4];
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[5] = r1;
            r1 = ((object*) lenv) [5];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[6] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1611; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1611:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[7] = r1;
            r1 = (object*)TC_NIL;
            r0[8] = r1;
            r1 = ((object*) lenv) [8];
            r2 = ((object*) lenv) [7];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[9] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [6];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1612; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L1612:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*) lenv) [9];
            r1[3] = r2;
            r2 = top[21];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1613; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
            L1613:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[10] = r1;
            r1 = ((object*) lenv) [10];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r1;
            r1 = t;
            r0[11] = r1;
            r1 = ((object******) lenv) [1][1][1][1][1][3];
            r0[12] = r1;
            r0 = ((object*) lenv) [12];
            r1 = ((object*) lenv) [11];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object*****) lenv) [1][1][1][1][3] = r0;
          } else {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1614; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1614:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = ((object******) lenv) [1][1][1][1][1][3];
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object*****) lenv) [1][1][1][1][3] = r0;
          }
          lenv = lenv[1];
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv[2] = r0;
          goto FUN__24loop_2156_i4622_i4727;
        }
        PXLL_RETURN(0);
      L1603:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN__24loop_2156_i4622_i4727; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1615; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN__24loop_2156_i4622_i4727;
      L1615:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 19);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) constructed_613[0];
      t = alloc_no_clear (UOTAG(1),1);
      t[1] = r1;
      r1 = t;
      r0[2] = r1;
      r1 = ((object*****) lenv) [1][1][1][1][2];
      t = alloc_no_clear (UOTAG(1),1);
      t[1] = r1;
      r1 = t;
      r0[3] = r1;
      r1 = ((object*****) lenv) [1][1][1][1][3];
      t = alloc_no_clear (UOTAG(1),1);
      t[1] = r1;
      r1 = t;
      r0[4] = r1;
      r1 = allocate (TC_TUPLE, 3);
      r2 = top[3];
      r1[2] = r2;
      r2 = ((object***) lenv) [1][1][2];
      r1[3] = r2;
      r2 = top[32];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1616; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
      L1616:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[5] = r1;
      r1 = ((object*) lenv) [5];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r1;
      r1 = t;
      r0[6] = r1;
      r1 = allocate (TC_TUPLE, 3);
      r2 = top[4];
      r1[2] = r2;
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object***) lenv) [1][1][2];
      r2[2] = r3;
      r2 = allocate (TC_TUPLE, 3);
      r3 = ((object*) lenv) [2];
      r2[2] = r3;
      r3 = (object *) constructed_36[0];
      r2[3] = r3;
      r3 = top[21];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L1617; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
      L1617:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      lenv = lenv[1];
      r1[3] = r2;
      r2 = top[32];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1618; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
      L1618:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[7] = r1;
      r1 = ((object*) lenv) [7];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r1;
      r1 = t;
      r0[8] = r1;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object******) lenv) [1][1][1][1][1][8];
      r1[2] = r2;
      r2 = allocate (TC_TUPLE, 2);
      r2[1] = lenv; lenv = r2;
      r3 = ((object***) lenv) [1][1][3];
      r2[2] = r3;
      r2 = allocate (TC_TUPLE, 3);
      r3 = ((object*) lenv) [2];
      r2[2] = r3;
      r3 = (object *) constructed_36[0];
      r2[3] = r3;
      r3 = top[21];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L1619; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
      L1619:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      lenv = lenv[1];
      r1[3] = r2;
      r2 = top[32];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1620; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
      L1620:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[9] = r1;
      r1 = ((object*) lenv) [9];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r1;
      r1 = t;
      r0[10] = r1;
      r1 = ((object***) lenv) [1][1][5];
      switch (get_case (r1)) {
        case (UITAG(1)): {
          r1 = top[188];
        } break;
        default: {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object****) lenv) [1][1][1][5];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][8];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1621; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
          L1621:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          lenv = lenv[1];
        } break;
      }
      r0[11] = r1;
      r1 = (object*)TC_NIL;
      r0[12] = r1;
      r1 = ((object*) lenv) [12];
      r2 = ((object*) lenv) [11];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[13] = r1;
      r1 = ((object*) lenv) [13];
      r2 = ((object*) lenv) [10];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[14] = r1;
      r1 = ((object*) lenv) [14];
      r2 = ((object*) lenv) [8];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[15] = r1;
      r1 = ((object*) lenv) [15];
      r2 = ((object*) lenv) [6];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[16] = r1;
      r1 = ((object*) lenv) [16];
      r2 = ((object*) lenv) [4];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[17] = r1;
      r1 = ((object*) lenv) [17];
      r2 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[18] = r1;
      r1 = ((object*) lenv) [18];
      r2 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r2;
      t[2] = r1;
      r1 = t;
      r0[19] = r1;
      r0 = ((object*) lenv) [19];
      t = alloc_no_clear (UOTAG(0),1);
      t[1] = r0;
      r0 = t;
      PXLL_RETURN(0);
    L1601:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_12594_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_2119_i1258_i4622_i4727
    goto L1622;
    FUN_loop_2119_i1258_i4622_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [5];
      switch (get_case (r0)) {
        case (TC_NIL): {
          r0 = allocate (TC_TUPLE, 5);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = ((object*) lenv) [4];
          r0[4] = r1;
          r1 = (object*)UITAG(1);
          r0[5] = r1;
          r1 = ((object***) lenv) [1][1][2];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][5];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][5];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UOTAG(0)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object**) lenv) [1][2];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (UOTAG(9)): {
                              r0 = allocate (TC_TUPLE, 2);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][2];
                              r1 = UOBJ_GET(r1,1);
                              r0[2] = r1;
                              r0 = allocate (TC_TUPLE, 2);
                              r0[1] = lenv; lenv = r0;
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = allocate (TC_TUPLE, 3);
                              r2[1] = lenv; lenv = r2;
                              r3 = (object *) constructed_599[0];
                              t = alloc_no_clear (UOTAG(1),1);
                              t[1] = r3;
                              r3 = t;
                              r2[2] = r3;
                              r3 = ((object******) lenv) [1][1][1][1][1][3];
                              r4 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r4;
                              t[2] = r3;
                              r3 = t;
                              r2[3] = r3;
                              r2 = ((object*) lenv) [3];
                              t = alloc_no_clear (UOTAG(0),1);
                              t[1] = r2;
                              r2 = t;
                              lenv = lenv[1];
                              r1[2] = r2;
                              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][8];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L1623; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_expand_1968_i4622_i4727;
                              L1623:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[2] = r1;
                              r0 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                              r1 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r1;
                              t[2] = r0;
                              r0 = t;
                              lenv = lenv[1];
                              r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                              r2 = ((object**) lenv) [1][3];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r2;
                              t[2] = r1;
                              r1 = t;
                              r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
                              r3 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r3;
                              t[2] = r2;
                              r2 = t;
                              r3 = ((object******) lenv) [1][1][1][1][1][3];
                              lenv = ((object ******)lenv)[1][1][1][1][1][1];
                              lenv[2] = r2;
                              lenv[3] = r1;
                              lenv[4] = r0;
                              lenv[5] = r3;
                              goto FUN_loop_2119_i1258_i4622_i4727;
                            } break;
                            default: {
                              lenv = ((object *****)lenv)[1][1][1][1][1];
                              goto fatbar_128;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ****)lenv)[1][1][1][1];
                          goto fatbar_128;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    case (UOTAG(1)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = (object *) constructed_635[0];
                      r1 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                        r0 = ((object****) lenv) [1][1][1][3];
                        switch (get_case (r0)) {
                          case (TC_NIL): {
                            r0 = allocate (TC_TUPLE, 5);
                            r1 = ((object*****) lenv) [1][1][1][1][2];
                            r0[2] = r1;
                            r1 = ((object*****) lenv) [1][1][1][1][3];
                            r0[3] = r1;
                            r1 = ((object*****) lenv) [1][1][1][1][4];
                            r0[4] = r1;
                            r1 = allocate (TC_TUPLE, 2);
                            r1[1] = lenv; lenv = r1;
                            r2 = allocate (TC_TUPLE, 2);
                            r3 = allocate (TC_TUPLE, 3);
                            r3[1] = lenv; lenv = r3;
                            r4 = (object *) constructed_599[0];
                            t = alloc_no_clear (UOTAG(1),1);
                            t[1] = r4;
                            r4 = t;
                            r3[2] = r4;
                            r4 = ((object****) lenv) [1][1][1][3];
                            r5 = ((object*) lenv) [2];
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r5;
                            t[2] = r4;
                            r4 = t;
                            r3[3] = r4;
                            r3 = ((object*) lenv) [3];
                            t = alloc_no_clear (UOTAG(0),1);
                            t[1] = r3;
                            r3 = t;
                            lenv = lenv[1];
                            r2[2] = r3;
                            r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][8];
                            t = allocate (TC_SAVE, 5);
                            t[1] = k; t[2] = lenv; t[3] = &&L1624; t[4] = r0; t[5] = r1; k = t;
                            r2[1] = r3[2]; lenv = r2; goto FUN_expand_1968_i4622_i4727;
                            L1624:
                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                            r2 = result;
                            r1[2] = r2;
                            r1 = ((object*) lenv) [2];
                            t = alloc_no_clear (UOTAG(0),1);
                            t[1] = r1;
                            r1 = t;
                            lenv = lenv[1];
                            r0[5] = r1;
                            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
                            r0[1] = r1[2]; lenv = r0; goto *r1[1];
                          } break;
                          default: {
                            lenv = ((object ****)lenv)[1][1][1][1];
                            goto fatbar_128;
                          }
                        }
                        PXLL_RETURN(0);
                      } else {
                        lenv = ((object ****)lenv)[1][1][1][1];
                        goto fatbar_128;
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object ***)lenv)[1][1][1];
                      goto fatbar_128;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object **)lenv)[1][1];
                  goto fatbar_128;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_128;
            }
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
      goto fatbar_128_over;
      fatbar_128:
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [5];
      r1[2] = r2;
      r2 = top[19];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1625; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
      L1625:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = top[95];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L1626; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_pp_437;
      L1626:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_636;
      r0[2] = r1;
      r1 = ((object*) lenv) [5];
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      fatbar_128_over: ;
      PXLL_RETURN(0);
    L1622:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_2119_i1258_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 5);
    r1 = (object *) constructed_633[0];
    r0[2] = r1;
    r1 = (object *) constructed_632[0];
    r0[3] = r1;
    r1 = (object *) constructed_631[0];
    r0[4] = r1;
    r1 = ((object***) lenv) [1][1][4];
    r0[5] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_2119_i1258_i4622_i4727;
  L1600:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_vcase_splat_1980_i4622_i4727; r2[2] = lenv;
  r1[19] = r2;
  // def wrap-with-constructors_1983_i4622_i4727
  goto L1627;
  FUN_wrap_with_constructors_1983_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_630[0];
    r0[2] = r1;
    r1 = (object *) constructed_629[0];
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_12859_i4622_i4727
    goto L1628;
    FUN_lambda_12859_i4622_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      // def lambda_12856_i4622_i4727
      goto L1629;
      FUN_lambda_12856_i4622_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r3 = allocate (TC_TUPLE, 7);
        r3[1] = lenv; lenv = r3;
        r4 = ((object****) lenv) [1][1][1][3];
        r4 = ((pxll_vector*)r4)->val[7];
        r3[2] = r4;
        r4 = ((object*) lenv) [2];
        r4 = allocate (TC_TUPLE, 2);
        r4[1] = lenv; lenv = r4;
        r5 = ((object**) lenv) [1][2];
        r5 = UOBJ_GET(r5,0);
        r4[2] = r5;
        r4 = ((object*) lenv) [2];
        lenv = lenv[1];
        r3[3] = r4;
        r4 = ((object***) lenv) [1][1][2];
        r4 = allocate (TC_TUPLE, 2);
        r4[1] = lenv; lenv = r4;
        r5 = ((object****) lenv) [1][1][1][2];
        r5 = UOBJ_GET(r5,0);
        r4[2] = r5;
        r4 = ((object*) lenv) [2];
        lenv = lenv[1];
        r3[4] = r4;
        r4 = (object*)TC_NIL;
        r3[5] = r4;
        r4 = ((object*) lenv) [5];
        r5 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r5;
        t[2] = r4;
        r4 = t;
        r3[6] = r4;
        r4 = ((object*) lenv) [6];
        r5 = (object*) &constructed_628;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r5;
        t[2] = r4;
        r4 = t;
        r3[7] = r4;
        r3 = ((object*) lenv) [7];
        r4 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        lenv = lenv[1];
        r2[2] = r3;
        r3 = top[47];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1630; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
        L1630:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = top[80];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1631; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
        L1631:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object****) lenv) [1][1][1][2];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object***) lenv) [1][1][2] = r0;
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object***) lenv) [1][1][3];
        r2 = ((pxll_vector*)r2)->val[7];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[3] = r2;
        r2 = ((object**) lenv) [1][3];
        r2 = ((pxll_vector*)r2)->val[0];
        r1[4] = r2;
        r2 = ((object******) lenv) [1][1][1][1][1][21];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1632; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_make_constructor_1984_i4622_i4727;
        L1632:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object****) lenv) [1][1][1][3];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object***) lenv) [1][1][3] = r0;
        PXLL_RETURN(0);
      L1629:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_12856_i4622_i4727; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object*) lenv) [3];
      r1 = ((pxll_vector*)r1)->val[6];
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1628:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_12859_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object****) lenv) [1][1][1][2];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[3] = r1;
    r1 = top[67];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1633; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
    L1633:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 4);
    r1 = allocate (TC_TUPLE, 3);
    r2 = top[3];
    r1[2] = r2;
    r2 = ((object*) lenv) [2];
    r1[3] = r2;
    r2 = top[32];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1634; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
    L1634:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = ((object*) lenv) [3];
    r0[3] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[4] = r1;
    r1 = ((object***) lenv) [1][1][3];
    r0[1] = r1[2]; lenv = r0; goto FUN_wrap_fix_1961_i4622_i4727;
  L1627:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_wrap_with_constructors_1983_i4622_i4727; r2[2] = lenv;
  r1[20] = r2;
  // def make-constructor_1984_i4622_i4727
  goto L1635;
  FUN_make_constructor_1984_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 19);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][4];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    // def $loop_2213_i4622_i4727
    goto L1636;
    FUN__24loop_2213_i4622_i4727:
      check_heap (0);
      r0 = ((object***) lenv) [1][1][2];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = (object *) constructed_36[0];
        r0[3] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
      } else {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r3 = allocate (TC_TUPLE, 4);
        r3[1] = lenv; lenv = r3;
        r4 = allocate (TC_TUPLE, 2);
        r5 = ((object***) lenv) [1][1][2];
        r4[2] = r5;
        r5 = top[56];
        t = allocate (TC_SAVE, 7);
        t[1] = k; t[2] = lenv; t[3] = &&L1637; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
        r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
        L1637:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
        r4 = result;
        r3[2] = r4;
        r4 = (object*)TC_NIL;
        r3[3] = r4;
        r4 = ((object*) lenv) [3];
        r5 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r5;
        t[2] = r4;
        r4 = t;
        r3[4] = r4;
        r3 = ((object*) lenv) [4];
        r4 = (object*) &constructed_627;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        lenv = lenv[1];
        r2[2] = r3;
        r3 = top[47];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1638; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
        L1638:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = top[80];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1639; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
        L1639:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = ((object**) lenv) [1][3];
        r1 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = (object *) 3;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        lenv[2] = r1;
        lenv[3] = r0;
        goto FUN__24loop_2213_i4622_i4727;
      }
      PXLL_RETURN(0);
    L1636:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN__24loop_2213_i4622_i4727; r2[2] = lenv;
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object *) 1;
    r1[2] = r2;
    r2 = (object*)TC_NIL;
    r1[3] = r2;
    r2 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1640; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN__24loop_2213_i4622_i4727;
    L1640:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[3] = r1;
    r1 = (object *) constructed_601[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[4] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 2);
    r3 = allocate (TC_TUPLE, 6);
    r3[1] = lenv; lenv = r3;
    r4 = ((object***) lenv) [1][1][2];
    r4 = allocate (TC_TUPLE, 2);
    r4[1] = lenv; lenv = r4;
    r5 = ((object****) lenv) [1][1][1][2];
    r5 = UOBJ_GET(r5,0);
    r4[2] = r5;
    r4 = ((object*) lenv) [2];
    lenv = lenv[1];
    r3[2] = r4;
    r4 = ((object***) lenv) [1][1][3];
    r4 = allocate (TC_TUPLE, 2);
    r4[1] = lenv; lenv = r4;
    r5 = ((object****) lenv) [1][1][1][3];
    r5 = UOBJ_GET(r5,0);
    r4[2] = r5;
    r4 = ((object*) lenv) [2];
    lenv = lenv[1];
    r3[3] = r4;
    r4 = (object*)TC_NIL;
    r3[4] = r4;
    r4 = ((object*) lenv) [4];
    r5 = ((object*) lenv) [3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r5;
    t[2] = r4;
    r4 = t;
    r3[5] = r4;
    r4 = ((object*) lenv) [5];
    r5 = (object*) &constructed_626;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r5;
    t[2] = r4;
    r4 = t;
    r3[6] = r4;
    r3 = ((object*) lenv) [6];
    r4 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    lenv = lenv[1];
    r2[2] = r3;
    r3 = top[47];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1641; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
    L1641:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = top[80];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1642; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
    L1642:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[5] = r1;
    r1 = ((object*) lenv) [5];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[6] = r1;
    r1 = ((object*) lenv) [3];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r1;
    r1 = t;
    r0[7] = r1;
    r1 = (object *) constructed_328[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[8] = r1;
    r1 = ((object**) lenv) [1][3];
    r2 = ((object**) lenv) [1][2];
    t = alloc_no_clear (UOTAG(9),2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[9] = r1;
    r1 = (object*)TC_NIL;
    r0[10] = r1;
    r1 = ((object*) lenv) [10];
    r2 = ((object*) lenv) [9];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[11] = r1;
    r1 = ((object*) lenv) [11];
    r2 = ((object*) lenv) [8];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[12] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [12];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1643; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L1643:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1644; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L1644:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[13] = r1;
    r1 = ((object*) lenv) [13];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r1;
    r1 = t;
    r0[14] = r1;
    r1 = (object*)TC_NIL;
    r0[15] = r1;
    r1 = ((object*) lenv) [15];
    r2 = ((object*) lenv) [14];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[16] = r1;
    r1 = ((object*) lenv) [16];
    r2 = ((object*) lenv) [7];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[17] = r1;
    r1 = ((object*) lenv) [17];
    r2 = ((object*) lenv) [6];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[18] = r1;
    r1 = ((object*) lenv) [18];
    r2 = ((object*) lenv) [4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[19] = r1;
    r0 = ((object*) lenv) [19];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L1635:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_make_constructor_1984_i4622_i4727; r2[2] = lenv;
  r1[21] = r2;
  // def make-alt_1985_i4622_i4727
  goto L1645;
  FUN_make_alt_1985_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    // def lambda_12988_i4622_i4727
    goto L1646;
    FUN_lambda_12988_i4622_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][2];
      r0[3] = r1;
      r1 = top[111];
      r0[1] = r1[2]; lenv = r0; goto FUN_parse_type_splat_467;
    L1646:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_lambda_12988_i4622_i4727; r2[2] = lenv;
    r1[2] = r2;
    r2 = ((object**) lenv) [1][4];
    r1[3] = r2;
    r2 = top[32];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1647; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
    L1647:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1648; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L1648:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r0 = allocate (UOTAG(5), 4);
    r1 = ((object*) lenv) [3];
    r0[1] = r1;
    r1 = (object *) 1;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[4] = r1;
    PXLL_RETURN(0);
  L1645:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_make_alt_1985_i4622_i4727; r2[2] = lenv;
  r1[22] = r2;
  // def expand-match_1991_i4622_i4727
  goto L1649;
  FUN_expand_match_1991_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_2254_i4622_i4727
    goto L1650;
    FUN_loop_2254_i4622_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [4];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][4];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = (object *) constructed_624[0];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 5);
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                r1[2] = r2;
                r2 = ((object*******) lenv) [1][1][1][1][1][1][8];
                r1[3] = r2;
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object****) lenv) [1][1][1][2];
                r2[2] = r3;
                r3 = (object *) constructed_36[0];
                r2[3] = r3;
                r3 = top[21];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1651; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                L1651:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[4] = r2;
                r2 = ((object***) lenv) [1][1][3];
                r1[5] = r2;
                r2 = top[116];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1652; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_compile_pattern_478;
                L1652:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                switch (get_case_noint (r0)) {
                  default: {
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][2];
                    r1 = UOBJ_GET(r1,1);
                    r0[2] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = ((object*****) lenv) [1][1][1][1][3];
                    switch (get_case (r1)) {
                      case (TC_NIL): {
                        r1 = (object *) 262;
                      } break;
                      default: {
                        goto fatbar_127;
                      }
                    }
                    goto fatbar_127_over;
                    fatbar_127:
                    r1 = (object *) 6;
                    fatbar_127_over: ;
                    if PXLL_IS_TRUE(r1) {
                      r1 = ((object*) lenv) [2];
                    } else {
                      r1 = allocate (TC_TUPLE, 7);
                      r1[1] = lenv; lenv = r1;
                      r2 = (object *) constructed_622[0];
                      t = alloc_no_clear (UOTAG(1),1);
                      t[1] = r2;
                      r2 = t;
                      r1[2] = r2;
                      r2 = ((object******) lenv) [1][1][1][1][1][3];
                      t = alloc_no_clear (UOTAG(0),1);
                      t[1] = r2;
                      r2 = t;
                      r1[3] = r2;
                      r2 = (object*)TC_NIL;
                      r1[4] = r2;
                      r2 = ((object*) lenv) [4];
                      r3 = ((object**) lenv) [1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[5] = r2;
                      r2 = ((object*) lenv) [5];
                      r3 = ((object*) lenv) [3];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[6] = r2;
                      r2 = ((object*) lenv) [6];
                      r3 = ((object*) lenv) [2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      r1[7] = r2;
                      r1 = ((object*) lenv) [7];
                      t = alloc_no_clear (UOTAG(0),1);
                      t[1] = r1;
                      r1 = t;
                      lenv = lenv[1];
                    }
                    r0[2] = r1;
                    r1 = ((object********) lenv) [1][1][1][1][1][1][1][8];
                    r0[1] = r1[2]; lenv = r0; goto FUN_expand_1968_i4622_i4727;
                  } break;
                }
                PXLL_RETURN(0);
              } else {
                goto fatbar_126;
              }
              PXLL_RETURN(0);
              goto fatbar_126_over;
              fatbar_126:
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              r1 = ((object***) lenv) [1][1][3];
              r2 = ((object**) lenv) [1][3];
              lenv = ((object **)lenv)[1][1];
              lenv[2] = r0;
              lenv[3] = r1;
              lenv[4] = r2;
              goto FUN_loop_2254_i4622_i4727;
              fatbar_126_over: ;
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_125;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_125_over;
          fatbar_125:
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = top[114];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1653; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_new_match_var_476;
          L1653:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 6);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (UOTAG(1),1);
          t[1] = r1;
          r1 = t;
          r0[2] = r1;
          r1 = (object*)TC_NIL;
          r0[3] = r1;
          r1 = ((object*) lenv) [3];
          r2 = ((object***) lenv) [1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[4] = r1;
          r1 = ((object*) lenv) [4];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[5] = r1;
          r1 = ((object*) lenv) [5];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          r0[6] = r1;
          r0 = ((object****) lenv) [1][1][1][3];
          r1 = ((object*) lenv) [6];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = ((object***) lenv) [1][1][2];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r2 = ((object**) lenv) [1][3];
          lenv = ((object **)lenv)[1][1];
          lenv[2] = r1;
          lenv[3] = r0;
          lenv[4] = r2;
          goto FUN_loop_2254_i4622_i4727;
          fatbar_125_over: ;
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_124;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_124_over;
      fatbar_124:
      r0 = allocate (TC_TUPLE, 3);
      r1 = (object*) &constructed_625;
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][2];
      r0[3] = r1;
      r1 = top[15];
      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      fatbar_124_over: ;
      PXLL_RETURN(0);
    L1650:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_2254_i4622_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r1 = (object *) constructed_620[0];
    r0[2] = r1;
    r1 = (object *) constructed_619[0];
    r0[3] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_2254_i4622_i4727;
  L1649:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_match_1991_i4622_i4727; r2[2] = lenv;
  r1[23] = r2;
  // def expand-cinclude_1992_i4622_i4727
  goto L1654;
  FUN_expand_cinclude_1992_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(2)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][3];
            switch (get_case (r0)) {
              case (TC_NIL): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object******) lenv) [1][1][1][1][1][2];
                r1 = ((pxll_vector*)r1)->val[0];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                r1 = ((object*****) lenv) [1][1][1][1][2];
                ((pxll_vector*)r1)->val[0] = r0;
                r0 = allocate (TC_TUPLE, 4);
                r0[1] = lenv; lenv = r0;
                r1 = (object *) constructed_599[0];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                r0[2] = r1;
                r1 = (object*)TC_NIL;
                r0[3] = r1;
                r1 = ((object*) lenv) [3];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[4] = r1;
                r0 = ((object*) lenv) [4];
                t = alloc_no_clear (UOTAG(0),1);
                t[1] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_123;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_123;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_123;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_123_over;
    fatbar_123:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_618;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_123_over: ;
    PXLL_RETURN(0);
  L1654:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_cinclude_1992_i4622_i4727; r2[2] = lenv;
  r1[24] = r2;
  // def expand-cverbatim_1993_i4622_i4727
  goto L1655;
  FUN_expand_cverbatim_1993_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(2)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object**) lenv) [1][3];
            switch (get_case (r0)) {
              case (TC_NIL): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object******) lenv) [1][1][1][1][1][2];
                r1 = ((pxll_vector*)r1)->val[1];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                r1 = ((object*****) lenv) [1][1][1][1][2];
                ((pxll_vector*)r1)->val[1] = r0;
                r0 = allocate (TC_TUPLE, 4);
                r0[1] = lenv; lenv = r0;
                r1 = (object *) constructed_599[0];
                t = alloc_no_clear (UOTAG(1),1);
                t[1] = r1;
                r1 = t;
                r0[2] = r1;
                r1 = (object*)TC_NIL;
                r0[3] = r1;
                r1 = ((object*) lenv) [3];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[4] = r1;
                r0 = ((object*) lenv) [4];
                t = alloc_no_clear (UOTAG(0),1);
                t[1] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object **)lenv)[1][1];
                goto fatbar_122;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_122;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_122;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_122_over;
    fatbar_122:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_617;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_122_over: ;
    PXLL_RETURN(0);
  L1655:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand_cverbatim_1993_i4622_i4727; r2[2] = lenv;
  r1[25] = r2;
  // def expand-%nvcase_1994_i4622_i4727
  goto L1656;
  FUN_expand__percentnvcase_1994_i4622_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = (object *) constructed_613[0];
    t = alloc_no_clear (UOTAG(1),1);
    t[1] = r1;
    r1 = t;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r2;
    t[2] = r1;
    r1 = t;
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L1656:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand__percentnvcase_1994_i4622_i4727; r2[2] = lenv;
  r1[26] = r2;
  // def expand-%%cexp_1995_i4622_i4727
  goto L1657;
  FUN_expand__percent_percentcexp_1995_i4622_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_PAIR): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 10);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = (object *) constructed_615[0];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r1;
            r1 = t;
            r0[4] = r1;
            r1 = (object*)TC_NIL;
            r0[5] = r1;
            r1 = ((object*) lenv) [5];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[6] = r1;
            r1 = ((object*) lenv) [6];
            r2 = ((object**) lenv) [1][2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[7] = r1;
            r1 = ((object*) lenv) [7];
            r2 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            r0[8] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][8];
            r1[2] = r2;
            r2 = ((object*) lenv) [3];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1658; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L1658:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[9] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [8];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1659; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L1659:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*) lenv) [9];
            r1[3] = r2;
            r2 = top[21];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1660; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
            L1660:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[10] = r1;
            r0 = ((object*) lenv) [10];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          default: {
            lenv = ((object *)lenv)[1];
            goto fatbar_121;
          }
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_121;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_121_over;
    fatbar_121:
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_616;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_121_over: ;
    PXLL_RETURN(0);
  L1657:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_expand__percent_percentcexp_1995_i4622_i4727; r2[2] = lenv;
  r1[27] = r2;
  r2 = allocate (TC_TUPLE, 13);
  r2[1] = lenv; lenv = r2;
  r3 = (object*)UITAG(0);
  r2[2] = r3;
  r3 = ((object**) lenv) [1][27];
  r4 = ((object*) lenv) [2];
  r5 = (object *) constructed_615[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[3] = r3;
  r3 = ((object**) lenv) [1][26];
  r4 = ((object*) lenv) [3];
  r5 = (object *) constructed_613[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[4] = r3;
  r3 = ((object**) lenv) [1][25];
  r4 = ((object*) lenv) [4];
  r5 = (object *) constructed_611[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[5] = r3;
  r3 = ((object**) lenv) [1][24];
  r4 = ((object*) lenv) [5];
  r5 = (object *) constructed_609[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[6] = r3;
  r3 = ((object**) lenv) [1][23];
  r4 = ((object*) lenv) [6];
  r5 = (object *) constructed_607[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[7] = r3;
  r3 = ((object**) lenv) [1][13];
  r4 = ((object*) lenv) [7];
  r5 = (object *) constructed_605[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[8] = r3;
  r3 = ((object**) lenv) [1][18];
  r4 = ((object*) lenv) [8];
  r5 = (object *) constructed_603[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[9] = r3;
  r3 = ((object**) lenv) [1][15];
  r4 = ((object*) lenv) [9];
  r5 = (object *) constructed_601[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[10] = r3;
  r3 = ((object**) lenv) [1][14];
  r4 = ((object*) lenv) [10];
  r5 = (object *) constructed_585[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[11] = r3;
  r3 = ((object**) lenv) [1][12];
  r4 = ((object*) lenv) [11];
  r5 = (object *) constructed_599[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[12] = r3;
  r3 = ((object**) lenv) [1][11];
  r4 = ((object*) lenv) [12];
  r5 = (object *) constructed_597[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r5;
  t[2] = r3;
  t[3] = r4;
  r3 = t;
  r2[13] = r3;
  r2 = ((object**) lenv) [1][10];
  r3 = ((object*) lenv) [13];
  r4 = (object *) constructed_595[0];
  t = alloc_no_clear (UOTAG(1),3);
  t[1] = r4;
  t[2] = r2;
  t[3] = r3;
  r2 = t;
  lenv = lenv[1];
  r1[28] = r2;
  r1 = ((object*) lenv) [2];
  lenv = lenv[1];
  r0[7] = r1;
  r1 = top[164];
  r1 = ((pxll_vector*)r1)->val[1];
  r2 = (object *) 3;
  range_check (GET_TUPLE_LENGTH(*(object*)r1), unbox(r2));
  r1 = ((pxll_vector*)r1)->val[unbox(r2)];
  r0[8] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [8];
  r1[2] = r2;
  r2 = (object *) 11778;
  r1[3] = r2;
  r2 = top[49];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1661; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string_split_354;
  L1661:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[9] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [9];
  r1[2] = r2;
  r2 = (object *) constructed_36[0];
  r1[3] = r2;
  r2 = top[21];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1662; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
  L1662:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[10] = r1;
  r1 = allocate (TC_TUPLE, 4);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (TC_TUPLE, 2);
  r3 = ((object**) lenv) [1][10];
  r2[2] = r3;
  r3 = top[19];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1663; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
  L1663:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  r2 = allocate (TC_TUPLE, 3);
  r3 = ((object*) lenv) [2];
  r2[2] = r3;
  r3 = (object*) &constructed_593;
  r2[3] = r3;
  r3 = top[50];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1664; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_string_compare_355;
  L1664:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[3] = r2;
  r2 = (object *) 1;
  r3 = ((object*) lenv) [3];
  r2 = PXLL_TEST(unbox(r3)==unbox(r2));
  r1[4] = r2;
  r1 = (object *) 6;
  r2 = ((object*) lenv) [4];
  r1 = PXLL_TEST(r2==r1);
  lenv = lenv[1];
  if PXLL_IS_TRUE(r1) {
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object*) &constructed_591;
    r1[2] = r2;
    r2 = ((object*) lenv) [8];
    r1[3] = r2;
    r2 = top[15];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1665; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
    L1665:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
  } else {
    r1 = allocate (TC_TUPLE, 3);
    r2 = allocate (TC_TUPLE, 2);
    r2[1] = lenv; lenv = r2;
    r3 = allocate (TC_TUPLE, 2);
    r4 = ((object**) lenv) [1][10];
    r3[2] = r4;
    r4 = top[20];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L1666; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_cdr_310;
    L1666:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    r2 = allocate (TC_TUPLE, 3);
    r3 = ((object*) lenv) [2];
    r2[2] = r3;
    r3 = (object *) constructed_36[0];
    r2[3] = r3;
    r3 = top[21];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1667; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
    L1667:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = (object*) &constructed_592;
    r1[3] = r2;
    r2 = top[48];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1668; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_join_353;
    L1668:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
  }
  r0[11] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = allocate (TC_TUPLE, 3);
  r2[1] = lenv; lenv = r2;
  r3 = (object*)TC_NIL;
  r2[2] = r3;
  r3 = ((object*) lenv) [2];
  r4 = (object*) &constructed_590;
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r4;
  t[2] = r3;
  r3 = t;
  r2[3] = r3;
  r2 = ((object*) lenv) [3];
  r3 = ((object**) lenv) [1][11];
  t = alloc_no_clear (TC_PAIR,2);
  t[1] = r3;
  t[2] = r2;
  r2 = t;
  lenv = lenv[1];
  r1[2] = r2;
  r2 = top[47];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1669; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
  L1669:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[12] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = ((object*) lenv) [8];
  r1[2] = r2;
  r2 = top[85];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1670; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_read_file_425;
  L1670:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[13] = r1;
  r1 = ((object*) lenv) [13];
  r0[14] = r1;
  r1 = allocate (TC_TUPLE, 4);
  r2 = top[2];
  r1[2] = r2;
  r2 = ((object*) lenv) [14];
  r1[3] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r3 = ((object*) lenv) [2];
  r3 = ((pxll_vector*)r3)->val[13];
  r2[2] = r3;
  r3 = top[85];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1671; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_read_file_425;
  L1671:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[4] = r2;
  r2 = top[38];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1672; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_foldr_332;
  L1672:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[15] = r1;
  r1 = ((object*) lenv) [15];
  r0[16] = r1;
  r1 = ((object*) lenv) [16];
  t = alloc_no_clear (UOTAG(0),1);
  t[1] = r1;
  r1 = t;
  r0[17] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = ((object*) lenv) [17];
  r1[2] = r2;
  r2 = ((object*) lenv) [7];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1673; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto *r2[1];
  L1673:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[18] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = ((object*) lenv) [18];
  r1[2] = r2;
  r2 = top[132];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1674; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_walk_513;
  L1674:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[19] = r1;
  r1 = allocate (TC_TUPLE, 4);
  r1[1] = lenv; lenv = r1;
  // def shadow_2600_i4629_i4727
  goto L1675;
  FUN_shadow_2600_i4629_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*)TC_NIL;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][2];
        r2 = ((pxll_vector*)r2)->val[0];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][2];
        r1[3] = r2;
        r2 = top[25];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1676; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_member_eq_question_319;
        L1676:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = (object *) 6;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object***) lenv) [1][1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][3];
          r1[3] = r2;
          r2 = ((object****) lenv) [1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1677; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_shadow_2600_i4629_i4727;
          L1677:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = ((object*) lenv) [3];
          r1 = ((object**) lenv) [1][2];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_shadow_2600_i4629_i4727;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1675:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_shadow_2600_i4629_i4727; r2[2] = lenv;
  r1[2] = r2;
  // def lookup_2601_i4629_i4727
  goto L1678;
  FUN_lookup_2601_i4629_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object***) lenv) [1][1][2];
        r0 = PXLL_TEST(r1==r0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = ((object*) lenv) [2];
          r0 = ((pxll_vector*)r0)->val[1];
          PXLL_RETURN(0);
        } else {
          r0 = ((object*) lenv) [3];
          r1 = ((object**) lenv) [1][2];
          lenv = ((object *)lenv)[1];
          lenv[2] = r1;
          lenv[3] = r0;
          goto FUN_lookup_2601_i4629_i4727;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L1678:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lookup_2601_i4629_i4727; r2[2] = lenv;
  r1[3] = r2;
  // def walk_2602_i4629_i4727
  goto L1679;
  FUN_walk_2602_i4629_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = k; // %getcc
    r0[2] = r1;
    // def lambda_16505_i4629_i4727
    goto L1680;
    FUN_lambda_16505_i4629_i4727:
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][2];
      k = r1; // %putcc
      PXLL_RETURN(0);
    L1680:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_16505_i4629_i4727; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][2];
    r1 = ((pxll_vector*)r1)->val[4];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object****) lenv) [1][1][1][3];
        r0[3] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1681; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_shadow_2600_i4629_i4727;
        L1681:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][3] = r0;
        lenv = lenv[1];
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object****) lenv) [1][1][1][3];
        r0[3] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1682; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_shadow_2600_i4629_i4727;
        L1682:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][3] = r0;
        lenv = lenv[1];
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object****) lenv) [1][1][1][3];
        r0[3] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1683; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_shadow_2600_i4629_i4727;
        L1683:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][3] = r0;
        lenv = lenv[1];
      } break;
      case (UOTAG(11)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (UOTAG(4), 2);
        r2 = ((object**) lenv) [1][2];
        r1[1] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object**) lenv) [1][3];
        r2[2] = r3;
        r3 = ((object*****) lenv) [1][1][1][1][3];
        r2[3] = r3;
        r3 = ((object******) lenv) [1][1][1][1][1][3];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1684; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_lookup_2601_i4629_i4727;
        L1684:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r0[2] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object****) lenv) [1][1][1][3] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 3);
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object****) lenv) [1][1][1][2];
        r3 = ((pxll_vector*)r3)->val[3];
        r2[2] = r3;
        r3 = top[19];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1685; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
        L1685:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][3];
        r1[3] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1686; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_walk_2602_i4629_i4727;
        L1686:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1687; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L1687:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        lenv = lenv[1];
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][3];
        r1[3] = r2;
        r2 = ((object******) lenv) [1][1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1688; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lookup_2601_i4629_i4727;
        L1688:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = ((object****) lenv) [1][1][1][2];
        ((pxll_vector*)r1)->val[4] = r0;
        lenv = lenv[1];
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][3];
        r1[3] = r2;
        r2 = ((object******) lenv) [1][1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1689; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lookup_2601_i4629_i4727;
        L1689:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(1),1);
        t[1] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = ((object****) lenv) [1][1][1][2];
        ((pxll_vector*)r1)->val[4] = r0;
        lenv = lenv[1];
      } break;
      default: {
        goto fatbar_120;
      }
    }
    goto fatbar_120_over;
    fatbar_120:
    r0 = (object *) 14;
    fatbar_120_over: ;
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_16608_i4629_i4727
    goto L1690;
    FUN_lambda_16608_i4629_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][3];
      r0[3] = r1;
      r1 = ((object****) lenv) [1][1][1][4];
      r0[1] = r1[2]; lenv = r0; goto FUN_walk_2602_i4629_i4727;
    L1690:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_16608_i4629_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[3];
    r0[3] = r1;
    r1 = top[32];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1691; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_map_326;
    L1691:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    r1 = ((object**) lenv) [1][2];
    ((pxll_vector*)r1)->val[3] = r0;
    r0 = ((object**) lenv) [1][2];
    PXLL_RETURN(0);
  L1679:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_walk_2602_i4629_i4727; r2[2] = lenv;
  r1[4] = r2;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object**) lenv) [1][19];
  r1[2] = r2;
  r2 = (object *) constructed_589[0];
  r1[3] = r2;
  r2 = ((object*) lenv) [4];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1692; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_walk_2602_i4629_i4727;
  L1692:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[20] = r1;
  r1 = (object *) 1;
  t = alloc_no_clear (UOTAG(5),1);
  t[1] = r1;
  r1 = t;
  ((object*) lenv) [17] = r1;
  r0[21] = r1;
  r1 = (object *) 1;
  t = alloc_no_clear (UOTAG(5),1);
  t[1] = r1;
  r1 = t;
  ((object*) lenv) [18] = r1;
  r0[22] = r1;
  r1 = (object *) constructed_588[0];
  ((object*) lenv) [13] = r1;
  r0[23] = r1;
  r1 = (object *) constructed_587[0];
  ((object*) lenv) [15] = r1;
  r0[24] = r1;
  r1 = top[165];
  r0[25] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) 1;
  r1[2] = r2;
  r2 = top[113];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1693; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_make_counter_472;
  L1693:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[26] = r1;
  r1 = allocate (TC_TUPLE, 4);
  r1[1] = lenv; lenv = r1;
  // def add_2547_i1750_i4632_i4727
  goto L1694;
  FUN_add_2547_i1750_i4632_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object****) lenv) [1][1][1][26];
    r2 = ((pxll_vector*)r2)->val[1];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L1695; t[4] = r0; t[5] = r1; k = t;
    lenv = r2[2]; goto *r2[1];
    L1695:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[3] = r2;
    r2 = top[135];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1696; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_make_vardef_517;
    L1696:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 5);
    r1 = ((object****) lenv) [1][1][1][25];
    r0[2] = r1;
    r1 = top[82];
    r0[3] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[4] = r1;
    r1 = ((object*) lenv) [2];
    r0[5] = r1;
    r1 = top[72];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1697; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
    L1697:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    ((object****) lenv) [1][1][1][25] = r0;
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L1694:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_add_2547_i1750_i4632_i4727; r2[2] = lenv;
  r1[2] = r2;
  // def lookup_2548_i1750_i4632_i4727
  goto L1698;
  FUN_lookup_2548_i1750_i4632_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object***) lenv) [1][1][25];
    r0[2] = r1;
    r1 = top[82];
    r0[3] = r1;
    r1 = ((object*) lenv) [2];
    r0[4] = r1;
    r1 = top[73];
    r0[1] = r1[2]; lenv = r0; goto FUN_tree_2fmember_409;
  L1698:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lookup_2548_i1750_i4632_i4727; r2[2] = lenv;
  r1[3] = r2;
  // def get_2549_i1750_i4632_i4727
  goto L1699;
  FUN_get_2549_i1750_i4632_i4727:
    r0 = ((object**) lenv) [1][25];
    PXLL_RETURN(0);
  L1699:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_2549_i1750_i4632_i4727; r2[2] = lenv;
  r1[4] = r2;
  r1 = allocate (UOTAG(3), 3);
  r2 = ((object*) lenv) [2];
  r1[1] = r2;
  r2 = ((object*) lenv) [4];
  r1[2] = r2;
  r2 = ((object*) lenv) [3];
  r1[3] = r2;
  lenv = lenv[1];
  r0[27] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def rename_2556_i4632_i4727
  goto L1700;
  FUN_rename_2556_i4632_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def lookup_2562_i4632_i4727
    goto L1701;
    FUN_lookup_2562_i4632_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def loop0_2564_i4632_i4727
      goto L1702;
      FUN_loop0_2564_i4632_i4727:
        check_heap (0);
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = (object*)UITAG(1);
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def loop1_2568_i4632_i4727
            goto L1703;
            FUN_loop1_2568_i4632_i4727:
              check_heap (0);
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = ((object***) lenv) [1][1][3];
                  lenv = ((object ***)lenv)[1][1][1];
                  lenv[2] = r0;
                  goto FUN_loop0_2564_i4632_i4727;
                } break;
                default: {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,16)];
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                  r0 = PXLL_TEST(r1==r0);
                  lenv = lenv[1];
                  if PXLL_IS_TRUE(r0) {
                    r0 = ((object*) lenv) [2];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } else {
                    r0 = ((object*) lenv) [3];
                    lenv = ((object *)lenv)[1];
                    lenv[2] = r0;
                    goto FUN_loop1_2568_i4632_i4727;
                  }
                  PXLL_RETURN(0);
                } break;
              }
              PXLL_RETURN(0);
            L1703:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_loop1_2568_i4632_i4727; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[1] = r1[2]; lenv = r0; goto FUN_loop1_2568_i4632_i4727;
          } break;
        }
        PXLL_RETURN(0);
      L1702:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_loop0_2564_i4632_i4727; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object****) lenv) [1][1][1][3];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r0[1] = r1[2]; lenv = r0; goto FUN_loop0_2564_i4632_i4727;
    L1701:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lookup_2562_i4632_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][2];
    r1 = ((pxll_vector*)r1)->val[4];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object******) lenv) [1][1][1][1][1][27];
        r2 = ((pxll_vector*)r2)->val[0];
        r1[2] = r2;
        r2 = ((object*) lenv) [3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1704; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1704:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1705; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lookup_2562_i4632_i4727;
        L1705:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        switch (get_case (r1)) {
          case (UITAG(1)): {
            r1 = (object *) constructed_585[0];
            r2 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 4);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object*****) lenv) [1][1][1][1][2];
              r5 = ((pxll_vector*)r5)->val[1];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L1706; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L1706:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r3 = ((object*) lenv) [4];
              r4 = (object*) &constructed_583;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1707; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L1707:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = top[80];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1708; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_string__3esymbol_419;
              L1708:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = ((object*) lenv) [2];
            }
          } break;
          default: {
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][5];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r1 = ((object*) lenv) [2];
            r1 = ((pxll_vector*)r1)->val[2];
            lenv = lenv[1];
          } break;
        }
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_16211_i4632_i4727
        goto L1709;
        FUN_lambda_16211_i4632_i4727:
          r0 = ((object*) lenv) [2];
          r0 = ((pxll_vector*)r0)->val[2];
          PXLL_RETURN(0);
        L1709:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_16211_i4632_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object**) lenv) [1][4];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1710; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1710:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][6];
        t = alloc_no_clear (UOTAG(7),2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = ((object****) lenv) [1][1][1][2];
        ((pxll_vector*)r1)->val[4] = r0;
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        r2 = ((object**) lenv) [1][4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_16096_i1759_i4632_i4727
        goto L1711;
        FUN_lambda_16096_i1759_i4632_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r0[3] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_rename_2556_i4632_i4727;
        L1711:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_16096_i1759_i4632_i4727; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[35];
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      } break;
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object******) lenv) [1][1][1][1][1][27];
        r2 = ((pxll_vector*)r2)->val[0];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1712; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1712:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_16244_i4632_i4727
        goto L1713;
        FUN_lambda_16244_i4632_i4727:
          r0 = ((object*) lenv) [2];
          r0 = ((pxll_vector*)r0)->val[2];
          PXLL_RETURN(0);
        L1713:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_16244_i4632_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1714; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1714:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(10),1);
        t[1] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = ((object****) lenv) [1][1][1][2];
        ((pxll_vector*)r1)->val[4] = r0;
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][3];
        r2 = ((object**) lenv) [1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_16096_i1762_i4632_i4727
        goto L1715;
        FUN_lambda_16096_i1762_i4632_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r0[3] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_rename_2556_i4632_i4727;
        L1715:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_16096_i1762_i4632_i4727; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[35];
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object****) lenv) [1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[124];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1716; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_unpack_fix_504;
        L1716:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = (object *) constructed_586[0];
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1717; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L1717:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][5];
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def $loop_2590_i4632_i4727
            goto L1718;
            FUN__24loop_2590_i4632_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 3);
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object****) lenv) [1][1][1][3];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1719; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L1719:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object*****) lenv) [1][1][1][1][4];
                r1[2] = r2;
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                lenv = lenv[1];
                r0[3] = r1;
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L1720; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_rename_2556_i4632_i4727;
                L1720:
                ; lenv = k[2]; k = k[1];
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 3);
                r3 = ((object******) lenv) [1][1][1][1][1][2];
                r2[2] = r3;
                r3 = ((object**) lenv) [1][2];
                r2[3] = r3;
                r3 = top[27];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L1721; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                L1721:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][27];
                r2 = ((pxll_vector*)r2)->val[0];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1722; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                L1722:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = ((object*****) lenv) [1][1][1][1][4];
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object****) lenv) [1][1][1][4] = r0;
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_2590_i4632_i4727;
              }
              PXLL_RETURN(0);
            L1718:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN__24loop_2590_i4632_i4727; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = (object *) 1;
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1723; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN__24loop_2590_i4632_i4727;
            L1723:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            // def lambda_16328_i4632_i4727
            goto L1724;
            FUN_lambda_16328_i4632_i4727:
              r0 = ((object*) lenv) [2];
              r0 = ((pxll_vector*)r0)->val[2];
              PXLL_RETURN(0);
            L1724:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_lambda_16328_i4632_i4727; r2[2] = lenv;
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object***) lenv) [1][1][4];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = (object *) constructed_36[0];
            r2[3] = r3;
            r3 = top[21];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1725; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
            L1725:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1726; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L1726:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(9),1);
            t[1] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = ((object*****) lenv) [1][1][1][1][2];
            ((pxll_vector*)r1)->val[4] = r0;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][4];
            r1[2] = r2;
            r1 = ((object******) lenv) [1][1][1][1][1][3];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            lenv = lenv[1];
            r0[3] = r1;
            r1 = ((object******) lenv) [1][1][1][1][1][2];
            r0[1] = r1[2]; lenv = r0; goto FUN_rename_2556_i4632_i4727;
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1727; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lookup_2562_i4632_i4727;
        L1727:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (UITAG(1)): {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = ((pxll_vector*)r1)->val[2];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(0),1);
            t[1] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = ((object*****) lenv) [1][1][1][1][2];
            ((pxll_vector*)r1)->val[4] = r0;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1728; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lookup_2562_i4632_i4727;
        L1728:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UITAG(1)): {
            r0 = (object *) 14;
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = ((pxll_vector*)r1)->val[2];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(1),1);
            t[1] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = ((object******) lenv) [1][1][1][1][1][2];
            ((pxll_vector*)r1)->val[4] = r0;
            lenv = lenv[1];
          } break;
        }
        lenv = lenv[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_16096_i1773_i4632_i4727
        goto L1729;
        FUN_lambda_16096_i1773_i4632_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object******) lenv) [1][1][1][1][1][3];
          r0[3] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r0[1] = r1[2]; lenv = r0; goto FUN_rename_2556_i4632_i4727;
        L1729:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_16096_i1773_i4632_i4727; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[35];
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      } break;
      default: {
        goto fatbar_119;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_119_over;
    fatbar_119:
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object****) lenv) [1][1][1][2];
    r1 = ((pxll_vector*)r1)->val[3];
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_16096_i1774_i4632_i4727
    goto L1730;
    FUN_lambda_16096_i1774_i4632_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object*****) lenv) [1][1][1][1][3];
      r0[3] = r1;
      r1 = ((object******) lenv) [1][1][1][1][1][2];
      r0[1] = r1[2]; lenv = r0; goto FUN_rename_2556_i4632_i4727;
    L1730:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_16096_i1774_i4632_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = top[35];
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    fatbar_119_over: ;
    PXLL_RETURN(0);
  L1700:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_rename_2556_i4632_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object**) lenv) [1][20];
  r1[2] = r2;
  r2 = (object *) constructed_582[0];
  r1[3] = r2;
  r2 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1731; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_rename_2556_i4632_i4727;
  L1731:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[28] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [20];
  r1[2] = r2;
  r2 = ((object*) lenv) [2];
  r1[3] = r2;
  r2 = top[146];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1732; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_do_one_round_555;
  L1732:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[29] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [29];
  r1[2] = r2;
  r2 = ((object*) lenv) [2];
  r1[3] = r2;
  r2 = top[146];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1733; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_do_one_round_555;
  L1733:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[30] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object*)UITAG(5);
  r1[2] = r2;
  r2 = (object *) constructed_581[0];
  r1[3] = r2;
  r2 = top[118];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1734; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
  L1734:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  ((object*) lenv) [29] = r1;
  r0[31] = r1;
  r1 = ((object*) lenv) [30];
  r0[32] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def search_3887_i4634_i4727
  goto L1735;
  FUN_search_3887_i4634_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[3];
    r0[3] = r1;
    r1 = top[40];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1736; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_every_question_334;
    L1736:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][2];
      r1 = ((pxll_vector*)r1)->val[4];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      switch (get_case (r1)) {
        case (UITAG(8)): {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = ((object***) lenv) [1][1][2];
          r2 = ((pxll_vector*)r2)->val[0];
          r1[2] = r2;
          r1 = top[193];
          r2 = ((object*) lenv) [2];
          r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
          lenv = lenv[1];
        } break;
        default: {
          goto fatbar_118;
        }
      }
      goto fatbar_118_over;
      fatbar_118:
      r1 = (object *) 262;
      fatbar_118_over: ;
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[2] = r1;
        r0 = top[194];
        r1 = ((object*) lenv) [2];
        r0 = box((pxll_int)unbox(r1)|(1<<unbox(r0)));
        lenv = lenv[1];
        r1 = ((object**) lenv) [1][2];
        ((pxll_vector*)r1)->val[0] = r0;
      } else {
        r0 = (object *) 14;
      }
      r0 = ((object*) lenv) [3];
      PXLL_RETURN(0);
    } else {
      r0 = (object *) 6;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1735:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_search_3887_i4634_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r2 = ((object**) lenv) [1][32];
  r1[2] = r2;
  r2 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1737; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_search_3887_i4634_i4727;
  L1737:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[33] = r1;
  r1 = ((object*) lenv) [2];
  r1 = ((pxll_vector*)r1)->val[9];
  r1 = ((pxll_vector*)r1)->val[4];
  if PXLL_IS_TRUE(r1) {
    r1 = (object*) &constructed_580;
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r2 = (object*) &constructed_580;
    (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [30];
    r1[2] = r2;
    r2 = top[127];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1738; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_pp_node_508;
    L1738:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = (object *) 14;
  } else {
    r1 = (object *) 14;
  }
  r0[34] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [30];
  r1[2] = r2;
  r2 = ((object*) lenv) [2];
  r1[3] = r2;
  r2 = top[142];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1739; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_build_dependency_graph_539;
  L1739:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[35] = r1;
  r1 = (object*) &constructed_579;
  r1 = box((pxll_int)((pxll_string*)(r1))->len);
  r2 = (object*) &constructed_579;
  r1 = (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
  r0[36] = r1;
  r1 = ((object*) lenv) [2];
  r1 = ((pxll_vector*)r1)->val[3];
  r0[37] = r1;
  r1 = (object *) constructed_578[0];
  r0[38] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object *) constructed_577[0];
  r1[2] = r2;
  r2 = top[141];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1740; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_symbol_set_maker_538;
  L1740:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[39] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def visit0_3827_i4637_i4727
  goto L1741;
  FUN_visit0_3827_i4637_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object***) lenv) [1][1][39];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = ((object***) lenv) [1][1][39];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
    r1 = ((pxll_vector*)r1)->val[0];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1742; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1742:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object****) lenv) [1][1][1][37];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = ((object****) lenv) [1][1][1][37];
    r2 = ((pxll_vector*)r2)->val[1];
    r2 = ((pxll_vector*)r2)->val[1];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1743; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1743:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = (object *) 14;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        // def lambda_24864_i4637_i4727
        goto L1744;
        FUN_lambda_24864_i4637_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*******) lenv) [1][1][1][1][1][1][39];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][39];
          r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
          r2 = ((pxll_vector*)r2)->val[2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1745; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto *r2[1];
          L1745:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r0[1] = r1[2]; lenv = r0; goto FUN_visit0_3827_i4637_i4727;
          } else {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        L1744:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_24864_i4637_i4727; r1[2] = lenv;
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
        r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,25)];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1746; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L1746:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        lenv = lenv[1];
      } break;
    }
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object****) lenv) [1][1][1][38];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object**) lenv) [1][2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    lenv = lenv[1];
    ((object***) lenv) [1][1][38] = r0;
    PXLL_RETURN(0);
  L1741:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_visit0_3827_i4637_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object**) lenv) [1][37];
  r1[2] = r2;
  // def lambda_24899_i4637_i4727
  goto L1747;
  FUN_lambda_24899_i4637_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object****) lenv) [1][1][1][39];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = ((object****) lenv) [1][1][1][39];
    r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
    r2 = ((pxll_vector*)r2)->val[2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1748; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1748:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = (object *) 6;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(r1==r0);
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object**) lenv) [1][2];
      r0[1] = r1[2]; lenv = r0; goto FUN_visit0_3827_i4637_i4727;
    } else {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1747:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lambda_24899_i4637_i4727; r2[2] = lenv;
  r1[3] = r2;
  r2 = ((object**) lenv) [1][37];
  r2 = ((pxll_vector*)r2)->val[1];
  r2 = ((pxll_vector*)r2)->val[4];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1749; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto *r2[1];
  L1749:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = allocate (TC_TUPLE, 6);
  r1[1] = lenv; lenv = r1;
  r2 = allocate (TC_TUPLE, 2);
  r3 = top[82];
  r2[2] = r3;
  r3 = top[198];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1750; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto *r3[1];
  L1750:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  r2 = allocate (TC_TUPLE, 3);
  r3 = ((object***) lenv) [1][1][37];
  r2[2] = r3;
  // def lambda_24734_i2884_i4637_i4727
  goto L1751;
  FUN_lambda_24734_i2884_i4637_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object *) constructed_576[0];
    r1[2] = r2;
    r2 = top[141];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1752; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_symbol_set_maker_538;
    L1752:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[1];
    r1 = ((pxll_vector*)r1)->val[0];
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
  L1751:
  r3 = allocate (TC_CLOSURE, 2);
  r3[1] = &&FUN_lambda_24734_i2884_i4637_i4727; r3[2] = lenv;
  r2[3] = r3;
  r3 = ((object***) lenv) [1][1][37];
  r3 = ((pxll_vector*)r3)->val[1];
  r3 = ((pxll_vector*)r3)->val[4];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1753; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto *r3[1];
  L1753:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = allocate (TC_TUPLE, 3);
  r3 = ((object***) lenv) [1][1][37];
  r2[2] = r3;
  // def lambda_24785_i2884_i4637_i4727
  goto L1754;
  FUN_lambda_24785_i2884_i4637_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_24777_i2884_i4637_i4727
    goto L1755;
    FUN_lambda_24777_i2884_i4637_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object****) lenv) [1][1][1][2];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = ((object****) lenv) [1][1][1][2];
      r2 = ((pxll_vector*)r2)->val[1];
      r2 = ((pxll_vector*)r2)->val[1];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1756; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto *r2[1];
      L1756:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UITAG(1)): {
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object****) lenv) [1][1][1][2];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object****) lenv) [1][1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[141];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1757; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_symbol_set_maker_538;
          L1757:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r1 = ((pxll_vector*)r1)->val[1];
          r1 = ((pxll_vector*)r1)->val[0];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
          r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,10)];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } break;
      }
      PXLL_RETURN(0);
    L1755:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_24777_i2884_i4637_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [3];
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
    r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,11)];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1758; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1758:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = top[35];
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
  L1754:
  r3 = allocate (TC_CLOSURE, 2);
  r3[1] = &&FUN_lambda_24785_i2884_i4637_i4727; r3[2] = lenv;
  r2[3] = r3;
  r3 = ((object***) lenv) [1][1][37];
  r3 = ((pxll_vector*)r3)->val[1];
  r3 = ((pxll_vector*)r3)->val[4];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1759; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto *r3[1];
  L1759:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = ((object*) lenv) [2];
  r1[3] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r3 = (object *) constructed_575[0];
  r2[2] = r3;
  r3 = top[141];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1760; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_symbol_set_maker_538;
  L1760:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[4] = r2;
  r2 = (object *) constructed_574[0];
  r1[5] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r3 = (object *) constructed_573[0];
  r2[2] = r3;
  r3 = top[141];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L1761; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_symbol_set_maker_538;
  L1761:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[6] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def visit1_3838_i4637_i4727
  goto L1762;
  FUN_visit1_3838_i4637_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object***) lenv) [1][1][4];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = ((object***) lenv) [1][1][4];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
    r1 = ((pxll_vector*)r1)->val[0];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1763; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L1763:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object****) lenv) [1][1][1][3];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = ((object****) lenv) [1][1][1][3];
    r2 = ((pxll_vector*)r2)->val[1];
    r2 = ((pxll_vector*)r2)->val[1];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1764; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1764:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = (object *) 14;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        // def lambda_24945_i4637_i4727
        goto L1765;
        FUN_lambda_24945_i4637_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
          r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
          r2 = ((pxll_vector*)r2)->val[2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1766; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto *r2[1];
          L1766:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r0[1] = r1[2]; lenv = r0; goto FUN_visit1_3838_i4637_i4727;
          } else {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        L1765:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_24945_i4637_i4727; r1[2] = lenv;
        r0[3] = r1;
        r1 = ((object*) lenv) [2];
        r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
        r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,25)];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1767; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
        L1767:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        lenv = lenv[1];
      } break;
    }
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object***) lenv) [1][1][6];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[3] = r1;
    r1 = ((object***) lenv) [1][1][6];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,50)];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,10)];
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
  L1762:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_visit1_3838_i4637_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def $loop_3843_i4637_i4727
  goto L1768;
  FUN__24loop_3843_i4637_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = ((object******) lenv) [1][1][1][1][1][38];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    switch (get_case (r1)) {
      case (TC_NIL): {
        r1 = (object *) 262;
      } break;
      default: {
        goto fatbar_117;
      }
    }
    goto fatbar_117_over;
    fatbar_117:
    r1 = (object *) 6;
    fatbar_117_over: ;
    r0[3] = r1;
    r0 = (object *) 6;
    r1 = ((object*) lenv) [3];
    r0 = PXLL_TEST(r1==r0);
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object******) lenv) [1][1][1][1][1][38];
      switch (get_case (r1)) {
        case (TC_NIL): {
          r1 = (object*) &constructed_68;
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r2 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
          r1 = (object*) &constructed_572;
          dump_object (r1, 0); fprintf (stdout, "\n");
          r1 = (object*) &constructed_572;
          PXLL_UNDEFINED; result=r1; goto Lreturn;
        } break;
        default: {
          r1 = allocate (TC_TUPLE, 3);
          r1[1] = lenv; lenv = r1;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][38];
          r2 = UOBJ_GET(r2,0);
          r1[2] = r2;
          r2 = ((object*******) lenv) [1][1][1][1][1][1][38];
          r2 = UOBJ_GET(r2,1);
          r1[3] = r2;
          r1 = ((object*) lenv) [3];
          ((object*******) lenv) [1][1][1][1][1][1][38] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
        } break;
      }
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object*****) lenv) [1][1][1][1][4];
      r1[2] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[3] = r2;
      r2 = ((object*****) lenv) [1][1][1][1][4];
      r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
      r2 = ((pxll_vector*)r2)->val[2];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L1769; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto *r2[1];
      L1769:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = (object *) 6;
      r1 = ((object*) lenv) [2];
      r0 = PXLL_TEST(r1==r0);
      lenv = lenv[1];
      if PXLL_IS_TRUE(r0) {
        r0 = allocate (TC_TUPLE, 2);
        r1 = (object *) constructed_571[0];
        r0[2] = r1;
        r1 = top[141];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1770; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_symbol_set_maker_538;
        L1770:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        ((object****) lenv) [1][1][1][6] = r0;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1771; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_visit1_3838_i4637_i4727;
        L1771:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*****) lenv) [1][1][1][1][6];
        r1[2] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][6];
        r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,50)];
        r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,11)];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1772; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto *r2[1];
        L1772:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][5];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object****) lenv) [1][1][1][5] = r0;
      } else {
        r0 = (object *) 14;
      }
      lenv = lenv[1];
      goto FUN__24loop_3843_i4637_i4727;
    } else {
      r0 = (object *) 14;
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1768:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN__24loop_3843_i4637_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L1773; t[4] = r0; k = t;
  lenv = r1[2]; goto FUN__24loop_3843_i4637_i4727;
  L1773:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r1 = ((object**) lenv) [1][5];
  lenv = lenv[1];
  lenv = lenv[1];
  lenv = lenv[1];
  r0[40] = r1;
  r1 = ((object*) lenv) [2];
  r1 = ((pxll_vector*)r1)->val[9];
  r1 = ((pxll_vector*)r1)->val[4];
  if PXLL_IS_TRUE(r1) {
    r1 = ((object*) lenv) [40];
    dump_object (r1, 0); fprintf (stdout, "\n");
    r1 = (object *) 14;
  } else {
    r1 = (object *) 14;
  }
  r0[41] = r1;
  r1 = ((object*) lenv) [40];
  r2 = ((object*) lenv) [2];
  ((pxll_vector*)r2)->val[12] = r1;
  r0[42] = r1;
  r1 = (object*) &constructed_570;
  r1 = box((pxll_int)((pxll_string*)(r1))->len);
  r2 = (object*) &constructed_570;
  r1 = (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
  r0[43] = r1;
  r1 = ((object*) lenv) [30];
  r0[44] = r1;
  r1 = allocate (TC_TUPLE, 14);
  r1[1] = lenv; lenv = r1;
  // def instantiate-type-scheme_3343_i4640_i4727
  goto L1774;
  FUN_instantiate_type_scheme_3343_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    // def lambda_21470_i4640_i4727
    goto L1775;
    FUN_lambda_21470_i4640_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = top[105];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1776; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_new_tvar_453;
          L1776:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r1 = ((object**) lenv) [1][3];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(1),3);
          t[1] = r2;
          t[2] = r0;
          t[3] = r1;
          r0 = t;
          PXLL_RETURN(0);
        } break;
        default: {
          goto fatbar_116;
        }
      }
      PXLL_RETURN(0);
      goto fatbar_116_over;
      fatbar_116:
      r0 = (object*) &constructed_68;
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = (object*) &constructed_68;
      (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      r0 = (object*) &constructed_569;
      dump_object (r0, 0); fprintf (stdout, "\n");
      r0 = (object*) &constructed_569;
      PXLL_UNDEFINED; result=r0; goto Lreturn;
      PXLL_RETURN(0);
      fatbar_116_over: ;
      PXLL_RETURN(0);
    L1775:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_lambda_21470_i4640_i4727; r2[2] = lenv;
    r1[2] = r2;
    r2 = (object*)UITAG(0);
    r1[3] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[4] = r2;
    r2 = top[38];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1777; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_foldr_332;
    L1777:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def walk_3380_i4640_i4727
    goto L1778;
    FUN_walk_3380_i4640_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(1)): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object*) lenv) [2];
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object***) lenv) [1][1][2];
          r1[2] = r2;
          r2 = ((object*) lenv) [3];
          r1[3] = r2;
          r2 = top[32];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1779; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
          L1779:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r1 = top[104];
          r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object****) lenv) [1][1][1][2];
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[65];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1780; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
          L1780:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = ((object**) lenv) [1][2];
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1778:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_walk_3380_i4640_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object***) lenv) [1][1][3];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_walk_3380_i4640_i4727;
  L1774:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_instantiate_type_scheme_3343_i4640_i4727; r2[2] = lenv;
  r1[2] = r2;
  // def occurs-in-type_3344_i4640_i4727
  goto L1781;
  FUN_occurs_in_type_3344_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = k; // %getcc
    r0[2] = r1;
    // def lambda_21530_i4640_i4727
    goto L1782;
    FUN_lambda_21530_i4640_i4727:
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][2];
      k = r1; // %putcc
      PXLL_RETURN(0);
    L1782:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_21530_i4640_i4727; r1[2] = lenv;
    r0[3] = r1;
    r0 = ((object**) lenv) [1][3];
    r1 = ((object**) lenv) [1][2];
    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
      r0 = (object *) 262;
      PXLL_RETURN(0);
    } else {
      r0 = ((object**) lenv) [1][3];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = (object *) 6;
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object***) lenv) [1][1][3];
          r1 = UOBJ_GET(r1,1);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_21551_i4640_i4727
          goto L1783;
          FUN_lambda_21551_i4640_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object****) lenv) [1][1][1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1784; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_occurs_in_type_3344_i4640_i4727;
            L1784:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r1 = (object *) 262;
              r0[2] = r1;
              r1 = ((object***) lenv) [1][1][3];
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
            } else {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          L1783:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_21551_i4640_i4727; r1[2] = lenv;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1785; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L1785:
          ; lenv = k[2]; k = k[1];
          r0 = (object *) 6;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    }
    PXLL_RETURN(0);
  L1781:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_occurs_in_type_3344_i4640_i4727; r2[2] = lenv;
  r1[3] = r2;
  // def occurs-free-in-tenv_3345_i4640_i4727
  goto L1786;
  FUN_occurs_free_in_tenv_3345_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = k; // %getcc
    r0[2] = r1;
    // def lambda_21571_i4640_i4727
    goto L1787;
    FUN_lambda_21571_i4640_i4727:
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][2];
      k = r1; // %putcc
      PXLL_RETURN(0);
    L1787:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_21571_i4640_i4727; r1[2] = lenv;
    r0[3] = r1;
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_21602_i4640_i4727
    goto L1788;
    FUN_lambda_21602_i4640_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [3];
      switch (get_case_noint (r0)) {
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][3];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = top[25];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1789; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_member_eq_question_319;
          L1789:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object****) lenv) [1][1][1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [3];
            r0[3] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][3];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1790; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_occurs_in_type_3344_i4640_i4727;
            L1790:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r1 = (object *) 262;
              r0[2] = r1;
              r1 = ((object***) lenv) [1][1][3];
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
            } else {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } else {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L1788:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_21602_i4640_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][3];
    r0[3] = r1;
    r1 = top[67];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1791; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
    L1791:
    ; lenv = k[2]; k = k[1];
    r0 = (object *) 6;
    PXLL_RETURN(0);
  L1786:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_occurs_free_in_tenv_3345_i4640_i4727; r2[2] = lenv;
  r1[4] = r2;
  // def build-type-scheme_3346_i4640_i4727
  goto L1792;
  FUN_build_type_scheme_3346_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object *) constructed_568[0];
    r1[2] = r2;
    r2 = top[196];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1793; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1793:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def find-generic-tvars_3406_i4640_i4727
    goto L1794;
    FUN_find_generic_tvars_3406_i4640_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (UOTAG(0)): {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][3];
          r1[3] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1795; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_occurs_free_in_tenv_3345_i4640_i4727;
          L1795:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object***) lenv) [1][1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object***) lenv) [1][1][2];
            r1 = ((pxll_vector*)r1)->val[1];
            r1 = ((pxll_vector*)r1)->val[0];
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
          } else {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          }
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object***) lenv) [1][1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[35];
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        } break;
      }
      PXLL_RETURN(0);
    L1794:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_find_generic_tvars_3406_i4640_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object****) lenv) [1][1][1][2];
    r1[2] = r2;
    r2 = top[140];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1796; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_apply_subst_531;
    L1796:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L1797; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_find_generic_tvars_3406_i4640_i4727;
    L1797:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (UOTAG(2), 2);
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object***) lenv) [1][1][2];
    r1[2] = r2;
    r2 = ((object***) lenv) [1][1][2];
    r2 = ((pxll_vector*)r2)->val[1];
    r2 = ((pxll_vector*)r2)->val[1];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L1798; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L1798:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[1] = r1;
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    PXLL_RETURN(0);
  L1792:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_build_type_scheme_3346_i4640_i4727; r2[2] = lenv;
  r1[5] = r2;
  // def type-of*_3347_i4640_i4727
  goto L1799;
  FUN_type_of_splat_3347_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[4];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object*) lenv) [2];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[4] = r1;
        r1 = ((object****) lenv) [1][1][1][7];
        r0[1] = r1[2]; lenv = r0; goto FUN_type_of_literal_3349_i4640_i4727;
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object*) lenv) [3];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1800; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_instantiate_type_scheme_3343_i4640_i4727;
        L1800:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = (object *) constructed_97[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = allocate (TC_TUPLE, 5);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object**) lenv) [1][3];
              r1[2] = r2;
              r2 = top[22];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1801; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
              L1801:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = (object *) 3;
              r2 = ((object*) lenv) [2];
              r1 = box((pxll_int)unbox(r2)-unbox(r1));
              r0[3] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object*****) lenv) [1][1][1][1][2];
              r2 = ((pxll_vector*)r2)->val[3];
              r1[2] = r2;
              r2 = top[22];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1802; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
              L1802:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[4] = r1;
              r1 = ((object*) lenv) [4];
              r2 = ((object*) lenv) [3];
              r1 = PXLL_TEST(unbox(r2)==unbox(r1));
              r0[5] = r1;
              r0 = (object *) 6;
              r1 = ((object*) lenv) [5];
              r0 = PXLL_TEST(r1==r0);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 3);
                r1 = (object*) &constructed_519;
                r0[2] = r1;
                r1 = ((object****) lenv) [1][1][1][2];
                r0[3] = r1;
                r1 = top[15];
                r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
              } else {
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = allocate (TC_TUPLE, 3);
                    r1 = (object*) &constructed_520;
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r0[3] = r1;
                    r1 = top[15];
                    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                  } break;
                  default: {
                    r0 = allocate (TC_TUPLE, 4);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r1 = allocate (TC_TUPLE, 3);
                    // def lambda_22061_i2295_i4640_i4727
                    goto L1803;
                    FUN_lambda_22061_i2295_i4640_i4727:
                      check_heap (0);
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object**) lenv) [1][2];
                      r1[2] = r2;
                      r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
                      r1[3] = r2;
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][6];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1804; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
                      L1804:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r0 = ((object*) lenv) [2];
                      r1 = ((object**) lenv) [1][2];
                      ((pxll_vector*)r1)->val[5] = r0;
                      r0 = ((object*) lenv) [2];
                      PXLL_RETURN(0);
                    L1803:
                    r2 = allocate (TC_CLOSURE, 2);
                    r2[1] = &&FUN_lambda_22061_i2295_i4640_i4727; r2[2] = lenv;
                    r1[2] = r2;
                    r2 = ((object*****) lenv) [1][1][1][1][2];
                    r2 = ((pxll_vector*)r2)->val[3];
                    r1[3] = r2;
                    r2 = top[32];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1805; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                    L1805:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[4] = r1;
                    r0 = allocate (TC_TUPLE, 4);
                    // def lambda_22074_i2295_i4640_i4727
                    goto L1806;
                    FUN_lambda_22074_i2295_i4640_i4727:
                      check_heap (0);
                      r0 = allocate (TC_TUPLE, 4);
                      r1 = ((object******) lenv) [1][1][1][1][1][2];
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object*) lenv) [2];
                      r1[2] = r2;
                      r2 = ((object*******) lenv) [1][1][1][1][1][1][8];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1807; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_unraw_3350_i4640_i4727;
                      L1807:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r1 = ((object*) lenv) [3];
                      r0[4] = r1;
                      r1 = top[139];
                      r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                    L1806:
                    r1 = allocate (TC_CLOSURE, 2);
                    r1[1] = &&FUN_lambda_22074_i2295_i4640_i4727; r1[2] = lenv;
                    r0[2] = r1;
                    r1 = ((object*) lenv) [3];
                    r0[3] = r1;
                    r1 = ((object*) lenv) [4];
                    r0[4] = r1;
                    r1 = top[36];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L1808; ; k = t;
                    r0[1] = r1[2]; lenv = r0; goto FUN_for_each2_330;
                    L1808:
                    ; lenv = k[2]; k = k[1];
                    r0 = ((object*) lenv) [2];
                    PXLL_RETURN(0);
                  } break;
                }
                PXLL_RETURN(0);
              }
              PXLL_RETURN(0);
            } else {
              lenv = ((object *)lenv)[1];
              goto fatbar_81;
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_81;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_81_over;
        fatbar_81:
        r0 = ((object*) lenv) [4];
        PXLL_RETURN(0);
        fatbar_81_over: ;
        PXLL_RETURN(0);
      } break;
      case (UITAG(6)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_22108_i2301_i4640_i4727
        goto L1809;
        FUN_lambda_22108_i2301_i4640_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][3];
          r1[3] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][6];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1810; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
          L1810:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          ((pxll_vector*)r1)->val[5] = r0;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        L1809:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_22108_i2301_i4640_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1811; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1811:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 4);
                        r1 = ((object******) lenv) [1][1][1][1][1][2];
                        r0[2] = r1;
                        r1 = ((object***) lenv) [1][1][2];
                        r0[3] = r1;
                        r1 = top[185];
                        r0[4] = r1;
                        r1 = top[139];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1812; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                        L1812:
                        ; lenv = k[2]; k = k[1];
                        r0 = allocate (TC_TUPLE, 4);
                        r1 = ((object******) lenv) [1][1][1][1][1][2];
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][2];
                        r0[3] = r1;
                        r1 = ((object*) lenv) [2];
                        r0[4] = r1;
                        r1 = top[139];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1813; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                        L1813:
                        ; lenv = k[2]; k = k[1];
                        r0 = ((object*) lenv) [2];
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object ***)lenv)[1][1][1];
                        goto fatbar_82;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_82;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_82;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_82;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_82_over;
        fatbar_82:
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_521;
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        fatbar_82_over: ;
        PXLL_RETURN(0);
      } break;
      case (UITAG(5)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        // def loop_3482_i2303_i4640_i4727
        goto L1814;
        FUN_loop_3482_i2303_i4640_i4727:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_NIL): {
              r0 = (object*) &constructed_68;
              r0 = box((pxll_int)((pxll_string*)(r0))->len);
              r1 = (object*) &constructed_68;
              (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
              r0 = (object*) &constructed_522;
              dump_object (r0, 0); fprintf (stdout, "\n");
              r0 = (object*) &constructed_522;
              PXLL_UNDEFINED; result=r0; goto Lreturn;
              PXLL_RETURN(0);
            } break;
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object**) lenv) [1][2];
                  r1[2] = r2;
                  r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  r1[3] = r2;
                  r2 = ((object********) lenv) [1][1][1][1][1][1][1][6];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1815; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
                  L1815:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = ((object*) lenv) [2];
                  r1 = ((object**) lenv) [1][2];
                  ((pxll_vector*)r1)->val[5] = r0;
                  r0 = ((object*) lenv) [2];
                  PXLL_RETURN(0);
                } break;
                default: {
                  goto fatbar_83;
                }
              }
              PXLL_RETURN(0);
              goto fatbar_83_over;
              fatbar_83:
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object**) lenv) [1][2];
              r1[2] = r2;
              r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
              r1[3] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][6];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1816; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
              L1816:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r1 = ((object**) lenv) [1][2];
              ((pxll_vector*)r1)->val[5] = r0;
              r0 = ((object*) lenv) [2];
              lenv = lenv[1];
              r0 = ((object*) lenv) [3];
              lenv = ((object *)lenv)[1];
              lenv[2] = r0;
              goto FUN_loop_3482_i2303_i4640_i4727;
              fatbar_83_over: ;
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        L1814:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_loop_3482_i2303_i4640_i4727; r1[2] = lenv;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object**) lenv) [1][2];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[1] = r1[2]; lenv = r0; goto FUN_loop_3482_i2303_i4640_i4727;
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1817; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L1817:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[4] = r1;
        r1 = (object *) constructed_524[0];
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_22233_i2307_i4640_i4727
        goto L1818;
        FUN_lambda_22233_i2307_i4640_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = top[105];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1819; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_new_tvar_453;
          L1819:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object***) lenv) [1][1][5];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          ((object***) lenv) [1][1][5] = r0;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (UOTAG(2), 2);
          r2 = (object *) constructed_523[0];
          r1[1] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object****) lenv) [1][1][1][4];
          r2 = ((object***) lenv) [1][1][2];
          t = alloc_no_clear (UOTAG(1),3);
          t[1] = r2;
          t[2] = r0;
          t[3] = r1;
          r0 = t;
          lenv = lenv[1];
          ((object***) lenv) [1][1][4] = r0;
          PXLL_RETURN(0);
        L1818:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_22233_i2307_i4640_i4727; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[35];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1820; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        L1820:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][3];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][4];
        r1[3] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][6];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1821; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
        L1821:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r2 = ((object**) lenv) [1][3];
        ((pxll_vector*)r2)->val[5] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][5];
        r1[2] = r2;
        r2 = (object *) constructed_36[0];
        r1[3] = r2;
        r2 = top[21];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1822; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
        L1822:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object *) constructed_97[0];
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r2 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        r0[3] = r1;
        r1 = top[104];
        r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][3];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][9];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1823; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_apply_tenv_3356_i4640_i4727;
        L1823:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1824; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L1824:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][3];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][9];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1825; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_apply_tenv_3356_i4640_i4727;
        L1825:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][3];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][6];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1826; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
        L1826:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        r2 = ((object*) lenv) [3];
        ((pxll_vector*)r2)->val[5] = r1;
        r1 = ((object*) lenv) [5];
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object***) lenv) [1][1][2];
        r0[2] = r1;
        r1 = ((object*) lenv) [4];
        r0[3] = r1;
        r1 = ((object*) lenv) [6];
        r0[4] = r1;
        r1 = top[139];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1827; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
        L1827:
        ; lenv = k[2]; k = k[1];
        r0 = top[184];
        PXLL_RETURN(0);
      } break;
      case (UITAG(8)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_22322_i2317_i4640_i4727
        goto L1828;
        FUN_lambda_22322_i2317_i4640_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][3];
          r1[3] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][6];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1829; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
          L1829:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          ((pxll_vector*)r1)->val[5] = r0;
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        L1828:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_22322_i2317_i4640_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object***) lenv) [1][1][2];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1830; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L1830:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = top[105];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1831; t[4] = r0; k = t;
            lenv = r1[2]; goto FUN_new_tvar_453;
            L1831:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object****) lenv) [1][1][1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = (object *) constructed_97[0];
            r1[2] = r2;
            r2 = ((object*) lenv) [3];
            r3 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[3] = r2;
            r2 = top[104];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1832; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
            L1832:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = top[139];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1833; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
            L1833:
            ; lenv = k[2]; k = k[1];
            r0 = ((object*) lenv) [4];
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = (object*) &constructed_68;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
            r0 = (object*) &constructed_525;
            dump_object (r0, 0); fprintf (stdout, "\n");
            r0 = (object*) &constructed_525;
            PXLL_UNDEFINED; result=r0; goto Lreturn;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 17);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[4] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[12];
        r0[5] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1834; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L1834:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[6] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = top[82];
        r1[2] = r2;
        r2 = top[198];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1835; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto *r2[1];
        L1835:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[7] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [6];
        r1[2] = r2;
        r2 = top[30];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1836; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_range_324;
        L1836:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[8] = r1;
        r1 = (object *) constructed_533[0];
        r0[9] = r1;
        r1 = (object *) constructed_532[0];
        r0[10] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = ((object**) lenv) [1][6];
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def $loop_3855_i2323_i4640_i4727
        goto L1837;
        FUN__24loop_3855_i2323_i4640_i4727:
          check_heap (0);
          r0 = ((object***) lenv) [1][1][2];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object****) lenv) [1][1][1][7];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][2];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1838; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1838:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = allocate (UOTAG(2), 2);
            r2 = (object *) 6;
            r1[1] = r2;
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r0[4] = r1;
            r1 = ((object****) lenv) [1][1][1][7];
            r1 = ((pxll_vector*)r1)->val[1];
            r1 = ((pxll_vector*)r1)->val[0];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1839; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto *r1[1];
            L1839:
            ; lenv = k[2]; k = k[1];
            r0 = (object *) 3;
            r1 = ((object*) lenv) [2];
            r0 = box((pxll_int)unbox(r1)+unbox(r0));
            lenv[2] = r0;
            goto FUN__24loop_3855_i2323_i4640_i4727;
          }
          PXLL_RETURN(0);
        L1837:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN__24loop_3855_i2323_i4640_i4727; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 2);
        r2 = (object *) 1;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1840; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN__24loop_3855_i2323_i4640_i4727;
        L1840:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
        lenv = lenv[1];
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_25150_i2323_i4640_i4727
        goto L1841;
        FUN_lambda_25150_i2323_i4640_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object***) lenv) [1][1][10];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1842; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L1842:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 1;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)>unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = ((object**) lenv) [1][9];
            r1 = ((object**) lenv) [1][10];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            ((object**) lenv) [1][9] = r0;
            r0 = (object *) constructed_531[0];
            ((object**) lenv) [1][10] = r0;
          } else {
            r0 = (object *) 14;
          }
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_25145_i2323_i4640_i4727
          goto L1843;
          FUN_lambda_25145_i2323_i4640_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][7];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = ((object****) lenv) [1][1][1][7];
            r2 = ((pxll_vector*)r2)->val[1];
            r2 = ((pxll_vector*)r2)->val[1];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1844; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L1844:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UITAG(1)): {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = ((pxll_vector*)r1)->val[0];
                r0[2] = r1;
                r0 = (object *) 6;
                r1 = ((object*) lenv) [2];
                r0 = PXLL_TEST(r1==r0);
                lenv = lenv[1];
                if PXLL_IS_TRUE(r0) {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = ((pxll_vector*)r1)->val[1];
                  r0[2] = r1;
                  r0 = ((object******) lenv) [1][1][1][1][1][10];
                  r1 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r1;
                  t[2] = r0;
                  r0 = t;
                  lenv = lenv[1];
                  ((object*****) lenv) [1][1][1][1][10] = r0;
                  r0 = (object *) 262;
                  r1 = ((object*) lenv) [2];
                  ((pxll_vector*)r1)->val[0] = r0;
                  r0 = allocate (TC_TUPLE, 3);
                  r1 = ((object*) lenv) [2];
                  r1 = ((pxll_vector*)r1)->val[1];
                  r0[2] = r1;
                  r1 = ((object*****) lenv) [1][1][1][1][8];
                  r0[3] = r1;
                  r1 = top[26];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L1845; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_remove_eq_320;
                  L1845:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  ((object*****) lenv) [1][1][1][1][8] = r0;
                  PXLL_RETURN(0);
                } else {
                  r0 = (object *) 14;
                  PXLL_RETURN(0);
                }
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
          L1843:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_25145_i2323_i4640_i4727; r1[2] = lenv;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = top[35];
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        L1841:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_25150_i2323_i4640_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object*) lenv) [5];
        r1[3] = r2;
        r2 = top[35];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1846; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_for_each_329;
        L1846:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][10];
        r2[2] = r3;
        r3 = top[22];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1847; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
        L1847:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = (object *) 1;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(unbox(r2)>unbox(r1));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = ((object*) lenv) [9];
          r2 = ((object*) lenv) [10];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          ((object*) lenv) [9] = r1;
        } else {
          r1 = (object *) 14;
        }
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][8];
        r2[2] = r3;
        r3 = top[22];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1848; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
        L1848:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = (object *) 1;
        r2 = ((object*) lenv) [2];
        r1 = PXLL_TEST(unbox(r2)>unbox(r1));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r1) {
          r1 = ((object*) lenv) [9];
          r2 = ((object*) lenv) [8];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          ((object*) lenv) [9] = r1;
        } else {
          r1 = (object *) 14;
        }
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [9];
        r1[2] = r2;
        r2 = (object *) constructed_36[0];
        r1[3] = r2;
        r2 = top[21];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1849; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
        L1849:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[11] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1850; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L1850:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[12] = r1;
        r1 = (object *) constructed_530[0];
        r0[13] = r1;
        r1 = (object *) constructed_529[0];
        r0[14] = r1;
        r1 = (object *) constructed_528[0];
        r0[15] = r1;
        r1 = (object *) 1;
        r0[16] = r1;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_25243_i2322_i4640_i4727
        goto L1851;
        FUN_lambda_25243_i2322_i4640_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = (object *) constructed_527[0];
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_25229_i2322_i4640_i4727
          goto L1852;
          FUN_lambda_25229_i2322_i4640_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*****) lenv) [1][1][1][1][2];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1853; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1853:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*****) lenv) [1][1][1][1][13];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object****) lenv) [1][1][1][13] = r0;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*****) lenv) [1][1][1][1][4];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1854; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1854:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*****) lenv) [1][1][1][1][14];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object****) lenv) [1][1][1][14] = r0;
            r0 = ((object**) lenv) [1][2];
            r1 = ((object****) lenv) [1][1][1][16];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            ((object**) lenv) [1][2] = r0;
            r0 = (object *) 3;
            r1 = ((object****) lenv) [1][1][1][16];
            r0 = box((pxll_int)unbox(r1)+unbox(r0));
            ((object****) lenv) [1][1][1][16] = r0;
            PXLL_RETURN(0);
          L1852:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_25229_i2322_i4640_i4727; r1[2] = lenv;
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L1855; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L1855:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1856; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L1856:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object****) lenv) [1][1][1][15];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          ((object***) lenv) [1][1][15] = r0;
          PXLL_RETURN(0);
        L1851:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_25243_i2322_i4640_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object*) lenv) [11];
        r1[3] = r2;
        r2 = top[35];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1857; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_for_each_329;
        L1857:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = allocate (UOTAG(5), 4);
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object*) lenv) [13];
        r2[2] = r3;
        r3 = (object *) constructed_36[0];
        r2[3] = r3;
        r3 = top[21];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1858; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
        L1858:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[1] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object*) lenv) [14];
        r2[2] = r3;
        r3 = (object *) constructed_36[0];
        r2[3] = r3;
        r3 = top[21];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1859; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
        L1859:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object*) lenv) [4];
        r2[2] = r3;
        r3 = ((object*) lenv) [12];
        r2[3] = r3;
        r3 = top[27];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1860; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
        L1860:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object*) lenv) [15];
        r2[2] = r3;
        r3 = (object *) constructed_36[0];
        r2[3] = r3;
        r3 = top[21];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L1861; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
        L1861:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[4] = r2;
        r0[17] = r1;
        r0 = ((object*) lenv) [17];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 10);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][17];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][17];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = ((object**) lenv) [1][17];
            r1 = UOBJ_GET(r1,2);
            r0[4] = r1;
            r1 = ((object**) lenv) [1][17];
            r1 = UOBJ_GET(r1,3);
            r0[5] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1862; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L1862:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[6] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = top[43];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1863; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3evector_339;
            L1863:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[7] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = top[43];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1864; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3evector_339;
            L1864:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[8] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            r3 = ((object**) lenv) [1][6];
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            // def $loop_3528_i2321_i4640_i4727
            goto L1865;
            FUN__24loop_3528_i2321_i4640_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object*) lenv) [3];
                r0[2] = r1;
                r1 = (object *) constructed_36[0];
                r0[3] = r1;
                r1 = top[21];
                r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = top[105];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1866; t[4] = r0; k = t;
                lenv = r1[2]; goto FUN_new_tvar_453;
                L1866:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object**) lenv) [1][3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                r1 = (object *) 3;
                r2 = ((object*) lenv) [2];
                r1 = box((pxll_int)unbox(r2)+unbox(r1));
                lenv[2] = r1;
                lenv[3] = r0;
                goto FUN__24loop_3528_i2321_i4640_i4727;
              }
              PXLL_RETURN(0);
            L1865:
            r3 = allocate (TC_CLOSURE, 2);
            r3[1] = &&FUN__24loop_3528_i2321_i4640_i4727; r3[2] = lenv;
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = (object *) 1;
            r2[2] = r3;
            r3 = (object*)TC_NIL;
            r2[3] = r3;
            r3 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L1867; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN__24loop_3528_i2321_i4640_i4727;
            L1867:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[43];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1868; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_list__3evector_339;
            L1868:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[9] = r1;
            r1 = ((object*) lenv) [6];
            ensure_heap (1, unbox(r1));
            r1 = top[192];
            r2 = ((object*) lenv) [6];
            if (unbox(r2) == 0) { r1 = (object *) TC_EMPTY_VECTOR; } else {
              t = alloc_no_clear (TC_VECTOR, unbox(r2));
              for (i=0; i<unbox(r2); i++) { t[i+1] = r1; }
              r1 = t;
            }
            r0[10] = r1;
            r0 = allocate (TC_TUPLE, 3);
            // def lambda_22494_i2321_i4640_i4727
            goto L1869;
            FUN_lambda_22494_i2321_i4640_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 4);
              // def lambda_22434_i2321_i4640_i4727
              goto L1870;
              FUN_lambda_22434_i2321_i4640_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r2 = ((object*****) lenv) [1][1][1][1][7];
                range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
                r1 = ((pxll_vector*)r2)->val[unbox(r1)];
                r0[2] = r1;
                r1 = allocate (UOTAG(2), 2);
                r2 = (object *) constructed_526[0];
                r1[1] = r2;
                r2 = ((object**) lenv) [1][2];
                r3 = ((object*****) lenv) [1][1][1][1][9];
                range_check (GET_TUPLE_LENGTH(*(object*)r3), unbox(r2));
                r2 = ((pxll_vector*)r3)->val[unbox(r2)];
                r1[2] = r2;
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                r1 = ((object**) lenv) [1][3];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),3);
                t[1] = r2;
                t[2] = r0;
                t[3] = r1;
                r0 = t;
                PXLL_RETURN(0);
              L1870:
              r2 = allocate (TC_CLOSURE, 2);
              r2[1] = &&FUN_lambda_22434_i2321_i4640_i4727; r2[2] = lenv;
              r1[2] = r2;
              r2 = ((object****) lenv) [1][1][1][3];
              r1[3] = r2;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object**) lenv) [1][2];
              r2[2] = r3;
              r3 = (object *) constructed_36[0];
              r2[3] = r3;
              r3 = top[21];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1871; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
              L1871:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[4] = r2;
              r2 = top[38];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L1872; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_foldr_332;
              L1872:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_22466_i2321_i4640_i4727
              goto L1873;
              FUN_lambda_22466_i2321_i4640_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 6);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r2 = ((object*****) lenv) [1][1][1][1][8];
                range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
                r1 = ((pxll_vector*)r2)->val[unbox(r1)];
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object*) lenv) [2];
                r1[2] = r2;
                r2 = ((object***) lenv) [1][1][2];
                r1[3] = r2;
                r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][6];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1874; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
                L1874:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = ((object*) lenv) [3];
                r2 = ((object*) lenv) [2];
                ((pxll_vector*)r2)->val[5] = r1;
                r1 = ((object*) lenv) [3];
                r0[4] = r1;
                r1 = allocate (TC_TUPLE, 4);
                r2 = ((object**) lenv) [1][2];
                r3 = ((object*****) lenv) [1][1][1][1][8];
                range_check (GET_TUPLE_LENGTH(*(object*)r3), unbox(r2));
                r2 = ((pxll_vector*)r3)->val[unbox(r2)];
                r1[2] = r2;
                r2 = ((object*) lenv) [4];
                r1[3] = r2;
                r2 = ((object**) lenv) [1][2];
                r3 = ((object*****) lenv) [1][1][1][1][9];
                range_check (GET_TUPLE_LENGTH(*(object*)r3), unbox(r2));
                r2 = ((pxll_vector*)r3)->val[unbox(r2)];
                r1[4] = r2;
                r2 = top[139];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1875; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_unify_530;
                L1875:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[5] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object*) lenv) [4];
                r1[2] = r2;
                r2 = top[140];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1876; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_apply_subst_531;
                L1876:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[6] = r1;
                r0 = ((object**) lenv) [1][2];
                r1 = ((object*) lenv) [6];
                r2 = ((object*****) lenv) [1][1][1][1][10];
                range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r0));
                ((pxll_vector*)r2)->val[unbox(r0)] = r1;
                PXLL_RETURN(0);
              L1873:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_22466_i2321_i4640_i4727; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[3] = r1;
              r1 = top[35];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L1877; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
              L1877:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_22487_i2321_i4640_i4727
              goto L1878;
              FUN_lambda_22487_i2321_i4640_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object**) lenv) [1][2];
                r3 = ((object*****) lenv) [1][1][1][1][10];
                range_check (GET_TUPLE_LENGTH(*(object*)r3), unbox(r2));
                r2 = ((pxll_vector*)r3)->val[unbox(r2)];
                r1[2] = r2;
                r2 = ((object******) lenv) [1][1][1][1][1][3];
                r1[3] = r2;
                r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L1879; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_build_type_scheme_3346_i4640_i4727;
                L1879:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object***) lenv) [1][1][2];
                r2 = ((object******) lenv) [1][1][1][1][1][7];
                range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
                r1 = ((pxll_vector*)r2)->val[unbox(r1)];
                r0[2] = r1;
                r0 = ((object**) lenv) [1][2];
                r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(1),3);
                t[1] = r2;
                t[2] = r0;
                t[3] = r1;
                r0 = t;
                lenv = lenv[1];
                ((object******) lenv) [1][1][1][1][1][3] = r0;
                PXLL_RETURN(0);
              L1878:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_22487_i2321_i4640_i4727; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[3] = r1;
              r1 = top[35];
              r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
            L1869:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_lambda_22494_i2321_i4640_i4727; r1[2] = lenv;
            r0[2] = r1;
            r1 = ((object*) lenv) [5];
            r0[3] = r1;
            r1 = top[35];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L1880; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
            L1880:
            ; lenv = k[2]; k = k[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][4];
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][3];
            r1[3] = r2;
            r2 = ((object******) lenv) [1][1][1][1][1][6];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1881; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
            L1881:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][4];
            ((pxll_vector*)r1)->val[5] = r0;
            r0 = ((object*) lenv) [2];
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1882; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L1882:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        // def $loop_3549_i2355_i4640_i4727
        goto L1883;
        FUN__24loop_3549_i2355_i4640_i4727:
          check_heap (0);
          r0 = ((object***) lenv) [1][1][2];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*****) lenv) [1][1][1][1][5];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1884; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1884:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*****) lenv) [1][1][1][1][2];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1885; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L1885:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = ((object*****) lenv) [1][1][1][1][3];
            r1[3] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][6];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1886; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
            L1886:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = ((object*) lenv) [4];
            r2 = ((object*) lenv) [2];
            ((pxll_vector*)r2)->val[5] = r1;
            r1 = ((object*) lenv) [4];
            r0[5] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (UOTAG(2), 2);
            r2 = (object *) constructed_534[0];
            r1[1] = r2;
            r2 = ((object**) lenv) [1][5];
            r1[2] = r2;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object******) lenv) [1][1][1][1][1][3];
            r2 = ((object**) lenv) [1][3];
            t = alloc_no_clear (UOTAG(1),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            lenv = lenv[1];
            ((object*****) lenv) [1][1][1][1][3] = r0;
            lenv = lenv[1];
            r0 = (object *) 3;
            r1 = ((object*) lenv) [2];
            r0 = box((pxll_int)unbox(r1)+unbox(r0));
            lenv[2] = r0;
            goto FUN__24loop_3549_i2355_i4640_i4727;
          }
          PXLL_RETURN(0);
        L1883:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN__24loop_3549_i2355_i4640_i4727; r1[2] = lenv;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = (object *) 1;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L1887; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3549_i2355_i4640_i4727;
        L1887:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        lenv = lenv[1];
        lenv = lenv[1];
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][5];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][4];
        r1[3] = r2;
        r2 = top[27];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1888; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
        L1888:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][3];
        r1[3] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][6];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1889; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
        L1889:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        ((pxll_vector*)r1)->val[5] = r0;
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(12)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[4] = r1;
        r1 = (object *) constructed_210[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object*****) lenv) [1][1][1][1][13];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = ((object*) lenv) [3];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 5);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_538[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1890; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1890:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*****) lenv) [1][1][1][1][13];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1891; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1891:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_85;
        }
        goto fatbar_85_over;
        fatbar_85:
        r1 = (object *) constructed_214[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_538[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1892; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1892:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1893; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1893:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_86;
        }
        goto fatbar_86_over;
        fatbar_86:
        r1 = (object *) constructed_230[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object*****) lenv) [1][1][1][1][13];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = ((object*) lenv) [3];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_481[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1894; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1894:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*****) lenv) [1][1][1][1][13];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [6];
          r4 = top[184];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1895; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1895:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_87;
        }
        goto fatbar_87_over;
        fatbar_87:
        r1 = (object *) constructed_239[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 5);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_481[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1896; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1896:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1897; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1897:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_88;
        }
        goto fatbar_88_over;
        fatbar_88:
        r1 = (object *) constructed_243[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object*****) lenv) [1][1][1][1][13];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = ((object*) lenv) [3];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object*****) lenv) [1][1][1][1][13];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1898; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1898:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_89;
        }
        goto fatbar_89_over;
        fatbar_89:
        r1 = (object *) constructed_250[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object**) lenv) [1][3];
          r2[2] = r3;
          r3 = top[197];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1899; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          lenv = r3[2]; goto *r3[1];
          L1899:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = top[111];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1900; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_parse_type_splat_467;
          L1900:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = allocate (UOTAG(2), 2);
          r2 = (object *) constructed_539[0];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 4);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_481[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object***) lenv) [1][1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1901; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1901:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1902; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1902:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
          lenv = lenv[1];
        } else {
          goto fatbar_90;
        }
        goto fatbar_90_over;
        fatbar_90:
        r1 = (object *) constructed_314[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case (UOTAG(0)): {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object**) lenv) [1][2];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [2];
                  switch (get_case (r1)) {
                    case (UOTAG(9)): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object**) lenv) [1][3];
                      switch (get_case (r1)) {
                        case (TC_PAIR): {
                          r1 = allocate (TC_TUPLE, 3);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object***) lenv) [1][1][3];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r2 = ((object***) lenv) [1][1][3];
                          r2 = UOBJ_GET(r2,1);
                          r1[3] = r2;
                          r1 = ((object*) lenv) [2];
                          switch (get_case (r1)) {
                            case (UOTAG(5)): {
                              r1 = allocate (TC_TUPLE, 2);
                              r1[1] = lenv; lenv = r1;
                              r2 = ((object**) lenv) [1][2];
                              r2 = UOBJ_GET(r2,0);
                              r1[2] = r2;
                              r1 = ((object**) lenv) [1][3];
                              switch (get_case (r1)) {
                                case (TC_PAIR): {
                                  r1 = allocate (TC_TUPLE, 3);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = ((object***) lenv) [1][1][3];
                                  r2 = UOBJ_GET(r2,0);
                                  r1[2] = r2;
                                  r2 = ((object***) lenv) [1][1][3];
                                  r2 = UOBJ_GET(r2,1);
                                  r1[3] = r2;
                                  r1 = ((object*) lenv) [2];
                                  switch (get_case (r1)) {
                                    case (UOTAG(5)): {
                                      r1 = allocate (TC_TUPLE, 2);
                                      r1[1] = lenv; lenv = r1;
                                      r2 = ((object**) lenv) [1][2];
                                      r2 = UOBJ_GET(r2,0);
                                      r1[2] = r2;
                                      r1 = ((object**) lenv) [1][3];
                                      switch (get_case (r1)) {
                                        case (TC_NIL): {
                                          r1 = (object *) constructed_49[0];
                                          r2 = ((object*****) lenv) [1][1][1][1][2];
                                          if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                                            r1 = allocate (TC_TUPLE, 6);
                                            r1[1] = lenv; lenv = r1;
                                            r2 = allocate (TC_TUPLE, 2);
                                            r3 = ((object**) lenv) [1][2];
                                            r2[2] = r3;
                                            r3 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][10];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1903; t[4] = r0; t[5] = r1; k = t;
                                            r2[1] = r3[2]; lenv = r2; goto FUN_n_tvars_3366_i4640_i4727;
                                            L1903:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            r1[2] = r2;
                                            r2 = top[105];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1904; t[4] = r0; t[5] = r1; k = t;
                                            lenv = r2[2]; goto FUN_new_tvar_453;
                                            L1904:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            r1[3] = r2;
                                            r2 = allocate (TC_TUPLE, 3);
                                            r3 = ((object******) lenv) [1][1][1][1][1][3];
                                            r2[2] = r3;
                                            r3 = (object *) constructed_546[0];
                                            r2[3] = r3;
                                            r3 = top[104];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1905; t[4] = r0; t[5] = r1; k = t;
                                            r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                            L1905:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            r1[4] = r2;
                                            r2 = allocate (TC_TUPLE, 4);
                                            r3 = ((object*) lenv) [4];
                                            r2[2] = r3;
                                            r3 = allocate (TC_TUPLE, 2);
                                            r3[1] = lenv; lenv = r3;
                                            r4 = (object *) 3;
                                            r5 = ((object***) lenv) [1][1][2];
                                            if PXLL_IS_TRUE(PXLL_TEST(unbox(r5)>unbox(r4))) {
                                              r4 = allocate (TC_TUPLE, 3);
                                              r5 = (object *) constructed_545[0];
                                              r4[2] = r5;
                                              r5 = ((object**) lenv) [1][2];
                                              r4[3] = r5;
                                              r5 = top[104];
                                              t = allocate (TC_SAVE, 7);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1906; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                              r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
                                              L1906:
                                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                              r4 = result;
                                            } else {
                                              r4 = allocate (TC_TUPLE, 2);
                                              r5 = ((object**) lenv) [1][2];
                                              r4[2] = r5;
                                              r5 = top[19];
                                              t = allocate (TC_SAVE, 7);
                                              t[1] = k; t[2] = lenv; t[3] = &&L1907; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                              r4[1] = r5[2]; lenv = r4; goto FUN_car_309;
                                              L1907:
                                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                              r4 = result;
                                            }
                                            r3[2] = r4;
                                            r3 = allocate (TC_TUPLE, 3);
                                            r4 = (object *) constructed_543[0];
                                            r3[2] = r4;
                                            r4 = allocate (TC_TUPLE, 2);
                                            r4[1] = lenv; lenv = r4;
                                            r5 = (object*)TC_NIL;
                                            r4[2] = r5;
                                            r4 = ((object*) lenv) [2];
                                            r5 = ((object**) lenv) [1][2];
                                            t = alloc_no_clear (TC_PAIR,2);
                                            t[1] = r5;
                                            t[2] = r4;
                                            r4 = t;
                                            lenv = lenv[1];
                                            r3[3] = r4;
                                            r4 = top[104];
                                            t = allocate (TC_SAVE, 6);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1908; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                            r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
                                            L1908:
                                            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                            r3 = result;
                                            lenv = lenv[1];
                                            r2[3] = r3;
                                            r3 = ((object*) lenv) [3];
                                            r2[4] = r3;
                                            r3 = top[106];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1909; t[4] = r0; t[5] = r1; k = t;
                                            r2[1] = r3[2]; lenv = r2; goto FUN_rlabel_458;
                                            L1909:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            r1[5] = r2;
                                            r2 = allocate (TC_TUPLE, 3);
                                            r3 = (object *) constructed_541[0];
                                            r2[2] = r3;
                                            r3 = allocate (TC_TUPLE, 2);
                                            r3[1] = lenv; lenv = r3;
                                            r4 = (object*)TC_NIL;
                                            r3[2] = r4;
                                            r3 = ((object*) lenv) [2];
                                            r4 = ((object**) lenv) [1][5];
                                            t = alloc_no_clear (TC_PAIR,2);
                                            t[1] = r4;
                                            t[2] = r3;
                                            r3 = t;
                                            lenv = lenv[1];
                                            r2[3] = r3;
                                            r3 = top[104];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1910; t[4] = r0; t[5] = r1; k = t;
                                            r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                            L1910:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            r1[6] = r2;
                                            r1 = allocate (UOTAG(2), 2);
                                            r2 = ((object*) lenv) [2];
                                            r3 = ((object*) lenv) [3];
                                            t = alloc_no_clear (TC_PAIR,2);
                                            t[1] = r3;
                                            t[2] = r2;
                                            r2 = t;
                                            r1[1] = r2;
                                            r2 = allocate (TC_TUPLE, 4);
                                            r2[1] = lenv; lenv = r2;
                                            r3 = allocate (TC_TUPLE, 3);
                                            r4 = ((object**) lenv) [1][2];
                                            r3[2] = r4;
                                            r4 = ((object*****) lenv) [1][1][1][1][2];
                                            r3[3] = r4;
                                            r4 = top[27];
                                            t = allocate (TC_SAVE, 6);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1911; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                            r3[1] = r4[2]; lenv = r3; goto FUN_nth_321;
                                            L1911:
                                            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                            r3 = result;
                                            r2[2] = r3;
                                            r3 = (object*)TC_NIL;
                                            r2[3] = r3;
                                            r3 = ((object*) lenv) [3];
                                            r4 = ((object**) lenv) [1][6];
                                            t = alloc_no_clear (TC_PAIR,2);
                                            t[1] = r4;
                                            t[2] = r3;
                                            r3 = t;
                                            r2[4] = r3;
                                            r2 = allocate (TC_TUPLE, 3);
                                            r3 = (object *) constructed_97[0];
                                            r2[2] = r3;
                                            r3 = ((object*) lenv) [4];
                                            r4 = ((object*) lenv) [2];
                                            t = alloc_no_clear (TC_PAIR,2);
                                            t[1] = r4;
                                            t[2] = r3;
                                            r3 = t;
                                            r2[3] = r3;
                                            r3 = top[104];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1912; t[4] = r0; t[5] = r1; k = t;
                                            r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                            L1912:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            lenv = lenv[1];
                                            r1[2] = r2;
                                            lenv = lenv[1];
                                          } else {
                                            r1 = allocate (TC_TUPLE, 2);
                                            r1[1] = lenv; lenv = r1;
                                            r2 = allocate (TC_TUPLE, 3);
                                            r3 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][2];
                                            r3 = ((pxll_vector*)r3)->val[2];
                                            r2[2] = r3;
                                            r3 = ((object******) lenv) [1][1][1][1][1][2];
                                            r2[3] = r3;
                                            r3 = top[65];
                                            t = allocate (TC_SAVE, 5);
                                            t[1] = k; t[2] = lenv; t[3] = &&L1913; t[4] = r0; t[5] = r1; k = t;
                                            r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flookup_394;
                                            L1913:
                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                            r2 = result;
                                            r1[2] = r2;
                                            r1 = ((object*) lenv) [2];
                                            switch (get_case (r1)) {
                                              case (UITAG(1)): {
                                                r1 = allocate (TC_TUPLE, 3);
                                                r2 = (object*) &constructed_547;
                                                r1[2] = r2;
                                                r2 = ((object******) lenv) [1][1][1][1][1][2];
                                                r1[3] = r2;
                                                r2 = top[15];
                                                t = allocate (TC_SAVE, 4);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1914; t[4] = r0; k = t;
                                                r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                                                L1914:
                                                r0 = k[4]; lenv = k[2]; k = k[1];
                                                r1 = result;
                                              } break;
                                              default: {
                                                r1 = allocate (TC_TUPLE, 5);
                                                r1[1] = lenv; lenv = r1;
                                                r2 = ((object**) lenv) [1][2];
                                                r2 = UOBJ_GET(r2,0);
                                                r1[2] = r2;
                                                r2 = allocate (TC_TUPLE, 2);
                                                r3 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                                r2[2] = r3;
                                                r3 = ((object*) lenv) [2];
                                                r3 = ((pxll_vector*)r3)->val[1];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1915; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto *r3[1];
                                                L1915:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                r1[3] = r2;
                                                r2 = ((object*) lenv) [2];
                                                r2 = ((pxll_vector*)r2)->val[5];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1916; t[4] = r0; t[5] = r1; k = t;
                                                lenv = r2[2]; goto *r2[1];
                                                L1916:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                r1[4] = r2;
                                                r2 = allocate (TC_TUPLE, 3);
                                                r3 = ((object*******) lenv) [1][1][1][1][1][1][2];
                                                r2[2] = r3;
                                                r3 = ((object*) lenv) [4];
                                                r2[3] = r3;
                                                r3 = top[104];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1917; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                                L1917:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                r1[5] = r2;
                                                r1 = allocate (UOTAG(2), 2);
                                                r2 = ((object*) lenv) [4];
                                                r1[1] = r2;
                                                r2 = allocate (TC_TUPLE, 4);
                                                r2[1] = lenv; lenv = r2;
                                                r3 = allocate (TC_TUPLE, 3);
                                                r4 = ((object**) lenv) [1][3];
                                                r4 = ((pxll_vector*)r4)->val[3];
                                                r3[2] = r4;
                                                r4 = ((object******) lenv) [1][1][1][1][1][2];
                                                r3[3] = r4;
                                                r4 = top[27];
                                                t = allocate (TC_SAVE, 6);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1918; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                                r3[1] = r4[2]; lenv = r3; goto FUN_nth_321;
                                                L1918:
                                                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                                r3 = result;
                                                r2[2] = r3;
                                                r3 = (object*)TC_NIL;
                                                r2[3] = r3;
                                                r3 = ((object*) lenv) [3];
                                                r4 = ((object**) lenv) [1][5];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r4;
                                                t[2] = r3;
                                                r3 = t;
                                                r2[4] = r3;
                                                r2 = allocate (TC_TUPLE, 3);
                                                r3 = (object *) constructed_97[0];
                                                r2[2] = r3;
                                                r3 = ((object*) lenv) [4];
                                                r4 = ((object*) lenv) [2];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r4;
                                                t[2] = r3;
                                                r3 = t;
                                                r2[3] = r3;
                                                r3 = top[104];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L1919; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                                L1919:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                lenv = lenv[1];
                                                r1[2] = r2;
                                                lenv = lenv[1];
                                              } break;
                                            }
                                            lenv = lenv[1];
                                          }
                                        } break;
                                        default: {
                                          lenv = ((object *******)lenv)[1][1][1][1][1][1][1];
                                          goto fatbar_92;
                                        }
                                      }
                                      lenv = lenv[1];
                                    } break;
                                    default: {
                                      lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                      goto fatbar_92;
                                    }
                                  }
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  goto fatbar_92;
                                }
                              }
                              lenv = lenv[1];
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_92;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_92;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_92;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_92;
                }
              }
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_92;
            }
          }
          goto fatbar_92_over;
          fatbar_92:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_548;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1920; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1920:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_92_over: ;
        } else {
          goto fatbar_91;
        }
        goto fatbar_91_over;
        fatbar_91:
        r1 = (object *) constructed_496[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case (UOTAG(0)): {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object**) lenv) [1][2];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object**) lenv) [1][2];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [2];
                  switch (get_case (r1)) {
                    case (UOTAG(1)): {
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r1 = ((object**) lenv) [1][3];
                      switch (get_case (r1)) {
                        case (TC_PAIR): {
                          r1 = allocate (TC_TUPLE, 3);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object***) lenv) [1][1][3];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r2 = ((object***) lenv) [1][1][3];
                          r2 = UOBJ_GET(r2,1);
                          r1[3] = r2;
                          r1 = ((object*) lenv) [2];
                          switch (get_case (r1)) {
                            case (UOTAG(5)): {
                              r1 = allocate (TC_TUPLE, 2);
                              r1[1] = lenv; lenv = r1;
                              r2 = ((object**) lenv) [1][2];
                              r2 = UOBJ_GET(r2,0);
                              r1[2] = r2;
                              r1 = ((object**) lenv) [1][3];
                              switch (get_case (r1)) {
                                case (TC_NIL): {
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = allocate (TC_TUPLE, 3);
                                  r3 = ((object****) lenv) [1][1][1][2];
                                  r2[2] = r3;
                                  r3 = (object *) constructed_546[0];
                                  r2[3] = r3;
                                  r3 = top[104];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1921; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                  L1921:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[2] = r2;
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = allocate (TC_TUPLE, 3);
                                  r3 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                                  r3 = ((pxll_vector*)r3)->val[15];
                                  r2[2] = r3;
                                  r3 = ((object*****) lenv) [1][1][1][1][2];
                                  r2[3] = r3;
                                  r3 = top[65];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1922; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flookup_394;
                                  L1922:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[2] = r2;
                                  r1 = ((object*) lenv) [2];
                                  switch (get_case (r1)) {
                                    case (UOTAG(0)): {
                                      r1 = (object *) 14;
                                    } break;
                                    default: {
                                      r1 = allocate (TC_TUPLE, 2);
                                      r1[1] = lenv; lenv = r1;
                                      r2 = allocate (TC_TUPLE, 2);
                                      r3 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][2];
                                      r3 = ((pxll_vector*)r3)->val[15];
                                      r2[2] = r3;
                                      r3 = top[69];
                                      t = allocate (TC_SAVE, 5);
                                      t[1] = k; t[2] = lenv; t[3] = &&L1923; t[4] = r0; t[5] = r1; k = t;
                                      r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flength_401;
                                      L1923:
                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                      r2 = result;
                                      r1[2] = r2;
                                      r1 = allocate (TC_TUPLE, 2);
                                      r1[1] = lenv; lenv = r1;
                                      r2 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][2];
                                      r2 = ((pxll_vector*)r2)->val[15];
                                      r1[2] = r2;
                                      r1 = ((object**) lenv) [1][2];
                                      r2 = ((object*) lenv) [2];
                                      r3 = ((object*******) lenv) [1][1][1][1][1][1][2];
                                      t = alloc_no_clear (UOTAG(1),3);
                                      t[1] = r3;
                                      t[2] = r1;
                                      t[3] = r2;
                                      r1 = t;
                                      lenv = lenv[1];
                                      r2 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][2];
                                      ((pxll_vector*)r2)->val[15] = r1;
                                      lenv = lenv[1];
                                    } break;
                                  }
                                  lenv = lenv[1];
                                  r1 = (object *) 3;
                                  r2 = ((object**) lenv) [1][2];
                                  if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                                    r1 = allocate (UOTAG(2), 2);
                                    r2 = allocate (TC_TUPLE, 3);
                                    r2[1] = lenv; lenv = r2;
                                    r3 = (object*)TC_NIL;
                                    r2[2] = r3;
                                    r3 = ((object*) lenv) [2];
                                    r4 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][13];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r4;
                                    t[2] = r3;
                                    r3 = t;
                                    r2[3] = r3;
                                    r2 = ((object*) lenv) [3];
                                    r3 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][12];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r3;
                                    t[2] = r2;
                                    r2 = t;
                                    lenv = lenv[1];
                                    r1[1] = r2;
                                    r2 = allocate (TC_TUPLE, 5);
                                    r2[1] = lenv; lenv = r2;
                                    r3 = allocate (TC_TUPLE, 4);
                                    r4 = ((object**) lenv) [1][2];
                                    r3[2] = r4;
                                    r4 = allocate (TC_TUPLE, 3);
                                    r5 = (object *) constructed_543[0];
                                    r4[2] = r5;
                                    r5 = allocate (TC_TUPLE, 2);
                                    r5[1] = lenv; lenv = r5;
                                    r6 = (object*)TC_NIL;
                                    r5[2] = r6;
                                    r5 = ((object*) lenv) [2];
                                    r6 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][12];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r6;
                                    t[2] = r5;
                                    r5 = t;
                                    lenv = lenv[1];
                                    r4[3] = r5;
                                    r5 = top[104];
                                    t = allocate (TC_SAVE, 7);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1924; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                    r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
                                    L1924:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                    r4 = result;
                                    r3[3] = r4;
                                    r4 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][13];
                                    r3[4] = r4;
                                    r4 = top[106];
                                    t = allocate (TC_SAVE, 6);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1925; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                    r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
                                    L1925:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                    r3 = result;
                                    r2[2] = r3;
                                    r3 = allocate (TC_TUPLE, 3);
                                    r4 = (object *) constructed_541[0];
                                    r3[2] = r4;
                                    r4 = allocate (TC_TUPLE, 2);
                                    r4[1] = lenv; lenv = r4;
                                    r5 = (object*)TC_NIL;
                                    r4[2] = r5;
                                    r4 = ((object*) lenv) [2];
                                    r5 = ((object**) lenv) [1][2];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r5;
                                    t[2] = r4;
                                    r4 = t;
                                    lenv = lenv[1];
                                    r3[3] = r4;
                                    r4 = top[104];
                                    t = allocate (TC_SAVE, 6);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1926; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                    r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
                                    L1926:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                    r3 = result;
                                    r2[3] = r3;
                                    r3 = (object*)TC_NIL;
                                    r2[4] = r3;
                                    r3 = ((object*) lenv) [4];
                                    r4 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][12];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r4;
                                    t[2] = r3;
                                    r3 = t;
                                    r2[5] = r3;
                                    r2 = allocate (TC_TUPLE, 3);
                                    r3 = (object *) constructed_97[0];
                                    r2[2] = r3;
                                    r3 = ((object*) lenv) [5];
                                    r4 = ((object*) lenv) [3];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r4;
                                    t[2] = r3;
                                    r3 = t;
                                    r2[3] = r3;
                                    r3 = top[104];
                                    t = allocate (TC_SAVE, 5);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1927; t[4] = r0; t[5] = r1; k = t;
                                    r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                    L1927:
                                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                    r2 = result;
                                    lenv = lenv[1];
                                    r1[2] = r2;
                                  } else {
                                    goto fatbar_95;
                                  }
                                  goto fatbar_95_over;
                                  fatbar_95:
                                  r1 = (object *) 1;
                                  r2 = ((object**) lenv) [1][2];
                                  if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                                    r1 = allocate (UOTAG(2), 2);
                                    r2 = allocate (TC_TUPLE, 2);
                                    r2[1] = lenv; lenv = r2;
                                    r3 = (object*)TC_NIL;
                                    r2[2] = r3;
                                    r2 = ((object*) lenv) [2];
                                    r3 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][12];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r3;
                                    t[2] = r2;
                                    r2 = t;
                                    lenv = lenv[1];
                                    r1[1] = r2;
                                    r2 = allocate (TC_TUPLE, 3);
                                    r2[1] = lenv; lenv = r2;
                                    r3 = allocate (TC_TUPLE, 4);
                                    r4 = ((object**) lenv) [1][2];
                                    r3[2] = r4;
                                    r4 = allocate (TC_TUPLE, 2);
                                    r4[1] = lenv; lenv = r4;
                                    r5 = allocate (TC_TUPLE, 3);
                                    r6 = (object *) constructed_545[0];
                                    r5[2] = r6;
                                    r6 = (object *) constructed_550[0];
                                    r5[3] = r6;
                                    r6 = top[104];
                                    t = allocate (TC_SAVE, 8);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1928; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                                    r5[1] = r6[2]; lenv = r5; goto FUN_pred_451;
                                    L1928:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                                    r5 = result;
                                    r4[2] = r5;
                                    r4 = allocate (TC_TUPLE, 3);
                                    r5 = (object *) constructed_543[0];
                                    r4[2] = r5;
                                    r5 = allocate (TC_TUPLE, 2);
                                    r5[1] = lenv; lenv = r5;
                                    r6 = (object*)TC_NIL;
                                    r5[2] = r6;
                                    r5 = ((object*) lenv) [2];
                                    r6 = ((object**) lenv) [1][2];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r6;
                                    t[2] = r5;
                                    r5 = t;
                                    lenv = lenv[1];
                                    r4[3] = r5;
                                    r5 = top[104];
                                    t = allocate (TC_SAVE, 7);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1929; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                    r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
                                    L1929:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                    r4 = result;
                                    lenv = lenv[1];
                                    r3[3] = r4;
                                    r4 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][12];
                                    r3[4] = r4;
                                    r4 = top[106];
                                    t = allocate (TC_SAVE, 6);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1930; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                    r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
                                    L1930:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                    r3 = result;
                                    r2[2] = r3;
                                    r3 = allocate (TC_TUPLE, 3);
                                    r4 = (object *) constructed_541[0];
                                    r3[2] = r4;
                                    r4 = allocate (TC_TUPLE, 2);
                                    r4[1] = lenv; lenv = r4;
                                    r5 = (object*)TC_NIL;
                                    r4[2] = r5;
                                    r4 = ((object*) lenv) [2];
                                    r5 = ((object**) lenv) [1][2];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r5;
                                    t[2] = r4;
                                    r4 = t;
                                    lenv = lenv[1];
                                    r3[3] = r4;
                                    r4 = top[104];
                                    t = allocate (TC_SAVE, 6);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1931; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                    r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
                                    L1931:
                                    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                    r3 = result;
                                    r2[3] = r3;
                                    r2 = allocate (TC_TUPLE, 3);
                                    r3 = (object *) constructed_97[0];
                                    r2[2] = r3;
                                    r3 = (object *) constructed_549[0];
                                    r4 = ((object*) lenv) [3];
                                    t = alloc_no_clear (TC_PAIR,2);
                                    t[1] = r4;
                                    t[2] = r3;
                                    r3 = t;
                                    r2[3] = r3;
                                    r3 = top[104];
                                    t = allocate (TC_SAVE, 5);
                                    t[1] = k; t[2] = lenv; t[3] = &&L1932; t[4] = r0; t[5] = r1; k = t;
                                    r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                    L1932:
                                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                    r2 = result;
                                    lenv = lenv[1];
                                    r1[2] = r2;
                                  } else {
                                    goto fatbar_96;
                                  }
                                  goto fatbar_96_over;
                                  fatbar_96:
                                  r1 = allocate (TC_TUPLE, 3);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = top[105];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1933; t[4] = r0; t[5] = r1; k = t;
                                  lenv = r2[2]; goto FUN_new_tvar_453;
                                  L1933:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[2] = r2;
                                  r2 = allocate (TC_TUPLE, 2);
                                  r3 = ((object***) lenv) [1][1][2];
                                  r2[2] = r3;
                                  r3 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][10];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1934; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_n_tvars_3366_i4640_i4727;
                                  L1934:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  r1[3] = r2;
                                  r1 = allocate (UOTAG(2), 2);
                                  r2 = ((object*) lenv) [3];
                                  r3 = ((object*) lenv) [2];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r3;
                                  t[2] = r2;
                                  r2 = t;
                                  r1[1] = r2;
                                  r2 = allocate (TC_TUPLE, 3);
                                  r2[1] = lenv; lenv = r2;
                                  r3 = allocate (TC_TUPLE, 4);
                                  r4 = ((object***) lenv) [1][1][2];
                                  r3[2] = r4;
                                  r4 = allocate (TC_TUPLE, 2);
                                  r4[1] = lenv; lenv = r4;
                                  r5 = allocate (TC_TUPLE, 3);
                                  r6 = (object *) constructed_545[0];
                                  r5[2] = r6;
                                  r6 = ((object***) lenv) [1][1][3];
                                  r5[3] = r6;
                                  r6 = top[104];
                                  t = allocate (TC_SAVE, 8);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1935; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                                  r5[1] = r6[2]; lenv = r5; goto FUN_pred_451;
                                  L1935:
                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                                  r5 = result;
                                  r4[2] = r5;
                                  r4 = allocate (TC_TUPLE, 3);
                                  r5 = (object *) constructed_543[0];
                                  r4[2] = r5;
                                  r5 = allocate (TC_TUPLE, 2);
                                  r5[1] = lenv; lenv = r5;
                                  r6 = (object*)TC_NIL;
                                  r5[2] = r6;
                                  r5 = ((object*) lenv) [2];
                                  r6 = ((object**) lenv) [1][2];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r6;
                                  t[2] = r5;
                                  r5 = t;
                                  lenv = lenv[1];
                                  r4[3] = r5;
                                  r5 = top[104];
                                  t = allocate (TC_SAVE, 7);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1936; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                  r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
                                  L1936:
                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                  r4 = result;
                                  lenv = lenv[1];
                                  r3[3] = r4;
                                  r4 = ((object**) lenv) [1][2];
                                  r3[4] = r4;
                                  r4 = top[106];
                                  t = allocate (TC_SAVE, 6);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1937; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                  r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
                                  L1937:
                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                  r3 = result;
                                  r2[2] = r3;
                                  r3 = allocate (TC_TUPLE, 3);
                                  r4 = (object *) constructed_541[0];
                                  r3[2] = r4;
                                  r4 = allocate (TC_TUPLE, 2);
                                  r4[1] = lenv; lenv = r4;
                                  r5 = (object*)TC_NIL;
                                  r4[2] = r5;
                                  r4 = ((object*) lenv) [2];
                                  r5 = ((object**) lenv) [1][2];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r5;
                                  t[2] = r4;
                                  r4 = t;
                                  lenv = lenv[1];
                                  r3[3] = r4;
                                  r4 = top[104];
                                  t = allocate (TC_SAVE, 6);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1938; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                                  r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
                                  L1938:
                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                                  r3 = result;
                                  r2[3] = r3;
                                  r2 = allocate (TC_TUPLE, 3);
                                  r3 = (object *) constructed_97[0];
                                  r2[2] = r3;
                                  r3 = ((object**) lenv) [1][3];
                                  r4 = ((object*) lenv) [3];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r4;
                                  t[2] = r3;
                                  r3 = t;
                                  r2[3] = r3;
                                  r3 = top[104];
                                  t = allocate (TC_SAVE, 5);
                                  t[1] = k; t[2] = lenv; t[3] = &&L1939; t[4] = r0; t[5] = r1; k = t;
                                  r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                                  L1939:
                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                  r2 = result;
                                  lenv = lenv[1];
                                  r1[2] = r2;
                                  lenv = lenv[1];
                                  fatbar_96_over: ;
                                  fatbar_95_over: ;
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  goto fatbar_94;
                                }
                              }
                              lenv = lenv[1];
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_94;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_94;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_94;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_94;
                }
              }
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_94;
            }
          }
          goto fatbar_94_over;
          fatbar_94:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_548;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1940; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1940:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_94_over: ;
        } else {
          goto fatbar_93;
        }
        goto fatbar_93_over;
        fatbar_93:
        r1 = (object *) constructed_328[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case (UOTAG(9)): {
              r1 = allocate (TC_TUPLE, 4);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,1);
              r1[3] = r2;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object******) lenv) [1][1][1][1][1][2];
              r3 = ((pxll_vector*)r3)->val[2];
              r2[2] = r3;
              r3 = ((object*) lenv) [2];
              r2[3] = r3;
              r3 = top[65];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1941; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flookup_394;
              L1941:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[4] = r2;
              r1 = ((object*) lenv) [4];
              switch (get_case (r1)) {
                case (UITAG(1)): {
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = (object*) &constructed_551;
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  r1[3] = r2;
                  r2 = top[15];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1942; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                  L1942:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                } break;
                default: {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object**) lenv) [1][4];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object**) lenv) [1][3];
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  r2 = ((pxll_vector*)r2)->val[2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1943; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                  L1943:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  lenv = lenv[1];
                } break;
              }
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_98;
            }
          }
          goto fatbar_98_over;
          fatbar_98:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_548;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1944; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1944:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_98_over: ;
        } else {
          goto fatbar_97;
        }
        goto fatbar_97_over;
        fatbar_97:
        r1 = (object *) constructed_483[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case (UOTAG(1)): {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r1 = allocate (UOTAG(2), 2);
              r2 = allocate (TC_TUPLE, 3);
              r2[1] = lenv; lenv = r2;
              r3 = (object*)TC_NIL;
              r2[2] = r3;
              r3 = ((object*) lenv) [2];
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[3] = r3;
              r2 = ((object*) lenv) [3];
              r3 = ((object******) lenv) [1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[1] = r2;
              r2 = allocate (TC_TUPLE, 6);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 4);
              r4 = allocate (TC_TUPLE, 3);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = (object *) constructed_546[0];
              r4[3] = r5;
              r5 = top[104];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L1945; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
              L1945:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 3);
              r5 = (object *) constructed_543[0];
              r4[2] = r5;
              r5 = allocate (TC_TUPLE, 2);
              r5[1] = lenv; lenv = r5;
              r6 = (object*)TC_NIL;
              r5[2] = r6;
              r5 = ((object*) lenv) [2];
              r6 = ((object*******) lenv) [1][1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r6;
              t[2] = r5;
              r5 = t;
              lenv = lenv[1];
              r4[3] = r5;
              r5 = top[104];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L1946; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
              L1946:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              r3[4] = r4;
              r4 = top[106];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1947; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
              L1947:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 3);
              r4 = (object *) constructed_553[0];
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = (object*)TC_NIL;
              r4[2] = r5;
              r4 = ((object*) lenv) [2];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              lenv = lenv[1];
              r3[3] = r4;
              r4 = top[104];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1948; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
              L1948:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = (object*)TC_NIL;
              r2[4] = r3;
              r3 = ((object*) lenv) [4];
              r4 = ((object******) lenv) [1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[5] = r3;
              r3 = ((object*) lenv) [5];
              r4 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[6] = r3;
              r2 = allocate (TC_TUPLE, 3);
              r3 = (object *) constructed_97[0];
              r2[2] = r3;
              r3 = ((object*) lenv) [6];
              r4 = top[184];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[3] = r3;
              r3 = top[104];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1949; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
              L1949:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              lenv = lenv[1];
              r1[2] = r2;
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_100;
            }
          }
          goto fatbar_100_over;
          fatbar_100:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_548;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1950; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1950:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_100_over: ;
        } else {
          goto fatbar_99;
        }
        goto fatbar_99_over;
        fatbar_99:
        r1 = (object *) constructed_485[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case (UOTAG(1)): {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r1 = allocate (UOTAG(2), 2);
              r2 = allocate (TC_TUPLE, 3);
              r2[1] = lenv; lenv = r2;
              r3 = (object*)TC_NIL;
              r2[2] = r3;
              r3 = ((object*) lenv) [2];
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[3] = r3;
              r2 = ((object*) lenv) [3];
              r3 = ((object******) lenv) [1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[1] = r2;
              r2 = allocate (TC_TUPLE, 5);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 4);
              r4 = allocate (TC_TUPLE, 3);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = (object *) constructed_546[0];
              r4[3] = r5;
              r5 = top[104];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L1951; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
              L1951:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 3);
              r5 = (object *) constructed_543[0];
              r4[2] = r5;
              r5 = allocate (TC_TUPLE, 2);
              r5[1] = lenv; lenv = r5;
              r6 = (object*)TC_NIL;
              r5[2] = r6;
              r5 = ((object*) lenv) [2];
              r6 = ((object*******) lenv) [1][1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r6;
              t[2] = r5;
              r5 = t;
              lenv = lenv[1];
              r4[3] = r5;
              r5 = top[104];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L1952; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
              L1952:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              r3[4] = r4;
              r4 = top[106];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1953; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
              L1953:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 3);
              r4 = (object *) constructed_553[0];
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = (object*)TC_NIL;
              r4[2] = r5;
              r4 = ((object*) lenv) [2];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              lenv = lenv[1];
              r3[3] = r4;
              r4 = top[104];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1954; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
              L1954:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = (object*)TC_NIL;
              r2[4] = r3;
              r3 = ((object*) lenv) [4];
              r4 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[5] = r3;
              r2 = allocate (TC_TUPLE, 3);
              r3 = (object *) constructed_97[0];
              r2[2] = r3;
              r3 = ((object*) lenv) [5];
              r4 = ((object******) lenv) [1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[3] = r3;
              r3 = top[104];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1955; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
              L1955:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              lenv = lenv[1];
              r1[2] = r2;
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_102;
            }
          }
          goto fatbar_102_over;
          fatbar_102:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_548;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1956; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1956:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_102_over: ;
        } else {
          goto fatbar_101;
        }
        goto fatbar_101_over;
        fatbar_101:
        r1 = (object *) constructed_491[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = ((object*) lenv) [3];
          switch (get_case (r1)) {
            case (UOTAG(1)): {
              r1 = allocate (TC_TUPLE, 3);
              r1[1] = lenv; lenv = r1;
              r2 = ((object**) lenv) [1][3];
              r2 = UOBJ_GET(r2,0);
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object*) lenv) [2];
              r2[2] = r3;
              r3 = (object *) constructed_546[0];
              r2[3] = r3;
              r3 = top[104];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1957; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
              L1957:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r1 = allocate (UOTAG(2), 2);
              r2 = allocate (TC_TUPLE, 4);
              r2[1] = lenv; lenv = r2;
              r3 = (object*)TC_NIL;
              r2[2] = r3;
              r3 = ((object*) lenv) [2];
              r4 = ((object******) lenv) [1][1][1][1][1][14];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[3] = r3;
              r3 = ((object*) lenv) [3];
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[4] = r3;
              r2 = ((object*) lenv) [4];
              r3 = ((object******) lenv) [1][1][1][1][1][12];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[1] = r2;
              r2 = allocate (TC_TUPLE, 8);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 4);
              r4 = ((object**) lenv) [1][3];
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 3);
              r5 = (object *) constructed_543[0];
              r4[2] = r5;
              r5 = allocate (TC_TUPLE, 2);
              r5[1] = lenv; lenv = r5;
              r6 = (object*)TC_NIL;
              r5[2] = r6;
              r5 = ((object*) lenv) [2];
              r6 = ((object*******) lenv) [1][1][1][1][1][1][14];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r6;
              t[2] = r5;
              r5 = t;
              lenv = lenv[1];
              r4[3] = r5;
              r5 = top[104];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L1958; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_pred_451;
              L1958:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              r3[4] = r4;
              r4 = top[106];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1959; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
              L1959:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 3);
              r4 = (object *) constructed_553[0];
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = (object*)TC_NIL;
              r4[2] = r5;
              r4 = ((object*) lenv) [2];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              lenv = lenv[1];
              r3[3] = r4;
              r4 = top[104];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1960; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
              L1960:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = allocate (TC_TUPLE, 4);
              r4 = ((object**) lenv) [1][3];
              r3[2] = r4;
              r4 = ((object******) lenv) [1][1][1][1][1][12];
              r3[3] = r4;
              r4 = ((object******) lenv) [1][1][1][1][1][13];
              r3[4] = r4;
              r4 = top[106];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1961; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_rlabel_458;
              L1961:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[4] = r3;
              r3 = allocate (TC_TUPLE, 3);
              r4 = (object *) constructed_553[0];
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = (object*)TC_NIL;
              r4[2] = r5;
              r4 = ((object*) lenv) [2];
              r5 = ((object**) lenv) [1][4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              lenv = lenv[1];
              r3[3] = r4;
              r4 = top[104];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L1962; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
              L1962:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[5] = r3;
              r3 = (object*)TC_NIL;
              r2[6] = r3;
              r3 = ((object*) lenv) [6];
              r4 = ((object******) lenv) [1][1][1][1][1][14];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[7] = r3;
              r3 = ((object*) lenv) [7];
              r4 = ((object*) lenv) [5];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[8] = r3;
              r2 = allocate (TC_TUPLE, 3);
              r3 = (object *) constructed_97[0];
              r2[2] = r3;
              r3 = ((object*) lenv) [8];
              r4 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[3] = r3;
              r3 = top[104];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L1963; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
              L1963:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              lenv = lenv[1];
              r1[2] = r2;
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_104;
            }
          }
          goto fatbar_104_over;
          fatbar_104:
          r1 = allocate (TC_TUPLE, 3);
          r2 = (object*) &constructed_548;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = top[15];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L1964; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
          L1964:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          fatbar_104_over: ;
        } else {
          goto fatbar_103;
        }
        goto fatbar_103_over;
        fatbar_103:
        r1 = (object *) constructed_255[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = (object *) constructed_554[0];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [3];
          r4 = top[184];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1965; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1965:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_105;
        }
        goto fatbar_105_over;
        fatbar_105:
        r1 = (object *) constructed_488[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = (object *) constructed_561[0];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 4);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_560[0];
          r3[2] = r4;
          r4 = (object *) constructed_558[0];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1966; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1966:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_557[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object**) lenv) [1][2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1967; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1967:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_553[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object**) lenv) [1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1968; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1968:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[4] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = (object *) constructed_555[0];
          r4 = ((object*) lenv) [4];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1969; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1969:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_106;
        }
        goto fatbar_106_over;
        fatbar_106:
        r1 = (object *) constructed_286[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 6);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_472[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1970; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1970:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [6];
          r4 = top[184];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1971; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1971:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_107;
        }
        goto fatbar_107_over;
        fatbar_107:
        r1 = (object *) constructed_295[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 5);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_472[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1972; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1972:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1973; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1973:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_108;
        }
        goto fatbar_108_over;
        fatbar_108:
        r1 = (object *) constructed_308[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 5);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 3);
          r4 = (object *) constructed_472[0];
          r3[2] = r4;
          r4 = allocate (TC_TUPLE, 2);
          r4[1] = lenv; lenv = r4;
          r5 = (object*)TC_NIL;
          r4[2] = r5;
          r4 = ((object*) lenv) [2];
          r5 = ((object******) lenv) [1][1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r5;
          t[2] = r4;
          r4 = t;
          lenv = lenv[1];
          r3[3] = r4;
          r4 = top[104];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L1974; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_pred_451;
          L1974:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = top[181];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1975; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1975:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_109;
        }
        goto fatbar_109_over;
        fatbar_109:
        r1 = (object *) constructed_503[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = (object *) constructed_562[0];
          r4 = ((object****) lenv) [1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1976; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1976:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
        } else {
          goto fatbar_110;
        }
        goto fatbar_110_over;
        fatbar_110:
        r1 = (object *) constructed_501[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = (object *) constructed_563[0];
          r4 = ((object****) lenv) [1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1977; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1977:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
        } else {
          goto fatbar_111;
        }
        goto fatbar_111_over;
        fatbar_111:
        r1 = (object *) constructed_498[0];
        r2 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
          r1 = allocate (UOTAG(2), 2);
          r2 = allocate (TC_TUPLE, 2);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r2 = ((object*) lenv) [2];
          r3 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[1] = r2;
          r2 = allocate (TC_TUPLE, 4);
          r2[1] = lenv; lenv = r2;
          r3 = (object*)TC_NIL;
          r2[2] = r3;
          r3 = ((object*) lenv) [2];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r2 = allocate (TC_TUPLE, 3);
          r3 = (object *) constructed_97[0];
          r2[2] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*****) lenv) [1][1][1][1][12];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[3] = r3;
          r3 = top[104];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L1978; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
          L1978:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          lenv = lenv[1];
          r1[2] = r2;
        } else {
          goto fatbar_112;
        }
        goto fatbar_112_over;
        fatbar_112:
        r1 = allocate (TC_TUPLE, 3);
        r2 = (object*) &constructed_564;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[15];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L1979; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
        L1979:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        fatbar_112_over: ;
        fatbar_111_over: ;
        fatbar_110_over: ;
        fatbar_109_over: ;
        fatbar_108_over: ;
        fatbar_107_over: ;
        fatbar_106_over: ;
        fatbar_105_over: ;
        fatbar_103_over: ;
        fatbar_101_over: ;
        fatbar_99_over: ;
        fatbar_97_over: ;
        fatbar_93_over: ;
        fatbar_91_over: ;
        fatbar_90_over: ;
        fatbar_89_over: ;
        fatbar_88_over: ;
        fatbar_87_over: ;
        fatbar_86_over: ;
        fatbar_85_over: ;
        r0[5] = r1;
        r0 = ((object*) lenv) [5];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][5];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][5];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = ((object*) lenv) [3];
            r1[3] = r2;
            r2 = ((object*****) lenv) [1][1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L1980; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_instantiate_type_scheme_3343_i4640_i4727;
            L1980:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r0 = ((object*) lenv) [4];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = (object *) constructed_97[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object*) lenv) [3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object**) lenv) [1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = allocate (TC_TUPLE, 4);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][3];
                      r1[2] = r2;
                      r2 = top[22];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1981; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                      L1981:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object*****) lenv) [1][1][1][1][4];
                      r1[2] = r2;
                      r2 = top[22];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1982; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                      L1982:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r1 = ((object*) lenv) [3];
                      r2 = ((object*) lenv) [2];
                      r1 = PXLL_TEST(unbox(r2)==unbox(r1));
                      r0[4] = r1;
                      r0 = (object *) 6;
                      r1 = ((object*) lenv) [4];
                      r0 = PXLL_TEST(r1==r0);
                      lenv = lenv[1];
                      if PXLL_IS_TRUE(r0) {
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = (object*) &constructed_535;
                        r0[2] = r1;
                        r1 = ((object****) lenv) [1][1][1][4];
                        r0[3] = r1;
                        r1 = top[15];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L1983; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                        L1983:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                      } else {
                        r0 = (object *) 14;
                      }
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][3];
                      r1[2] = r2;
                      r2 = top[22];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1984; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                      L1984:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      // def $loop_3655_i2361_i4640_i4727
                      goto L1985;
                      FUN__24loop_3655_i2361_i4640_i4727:
                        check_heap (0);
                        r0 = ((object***) lenv) [1][1][2];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                          r0 = (object *) 14;
                          PXLL_RETURN(0);
                        } else {
                          r0 = allocate (TC_TUPLE, 5);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][2];
                          r1[3] = r2;
                          r2 = top[27];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L1986; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                          L1986:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object*) lenv) [2];
                          r1[2] = r2;
                          r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                          r1[3] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][6];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L1987; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
                          L1987:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[3] = r1;
                          r1 = ((object*) lenv) [3];
                          r2 = ((object*) lenv) [2];
                          ((pxll_vector*)r2)->val[5] = r1;
                          r1 = ((object*) lenv) [3];
                          r0[4] = r1;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object*****) lenv) [1][1][1][1][3];
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][2];
                          r1[3] = r2;
                          r2 = top[27];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L1988; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                          L1988:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[5] = r1;
                          r0 = allocate (TC_TUPLE, 4);
                          r1 = ((object*) lenv) [2];
                          r0[2] = r1;
                          r1 = ((object*) lenv) [4];
                          r0[3] = r1;
                          r1 = ((object*) lenv) [5];
                          r0[4] = r1;
                          r1 = top[139];
                          t = allocate (TC_SAVE, 3);
                          t[1] = k; t[2] = lenv; t[3] = &&L1989; ; k = t;
                          r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                          L1989:
                          ; lenv = k[2]; k = k[1];
                          r0 = result;
                          lenv = lenv[1];
                          r0 = (object *) 3;
                          r1 = ((object*) lenv) [2];
                          r0 = box((pxll_int)unbox(r1)+unbox(r0));
                          lenv[2] = r0;
                          goto FUN__24loop_3655_i2361_i4640_i4727;
                        }
                        PXLL_RETURN(0);
                      L1985:
                      r1 = allocate (TC_CLOSURE, 2);
                      r1[1] = &&FUN__24loop_3655_i2361_i4640_i4727; r1[2] = lenv;
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = (object *) 1;
                      r0[2] = r1;
                      r1 = ((object*) lenv) [2];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L1990; ; k = t;
                      r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3655_i2361_i4640_i4727;
                      L1990:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      lenv = lenv[1];
                      lenv = lenv[1];
                      r0 = ((object*) lenv) [2];
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_84;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_84;
                }
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_84;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_84_over;
            fatbar_84:
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_536;
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            fatbar_84_over: ;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = (object *) constructed_49[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][2];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 7);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r1 = top[105];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1991; t[4] = r0; k = t;
                  lenv = r1[2]; goto FUN_new_tvar_453;
                  L1991:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[4] = r1;
                  r1 = ((object*) lenv) [2];
                  r1 = ((pxll_vector*)r1)->val[4];
                  r0[5] = r1;
                  r1 = ((object*) lenv) [5];
                  switch (get_case (r1)) {
                    case (UOTAG(12)): {
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][5];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r1 = (object *) constructed_503[0];
                      r2 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                        r1 = (object *) 6;
                      } else {
                        lenv = ((object *)lenv)[1];
                        goto fatbar_114;
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      goto fatbar_114;
                    }
                  }
                  goto fatbar_114_over;
                  fatbar_114:
                  r1 = (object *) 262;
                  fatbar_114_over: ;
                  r0[6] = r1;
                  r1 = ((object*) lenv) [6];
                  if PXLL_IS_TRUE(r1) {
                    r1 = allocate (TC_TUPLE, 2);
                    r1[1] = lenv; lenv = r1;
                    r2 = allocate (TC_TUPLE, 3);
                    r3 = (object *) constructed_560[0];
                    r2[2] = r3;
                    r3 = (object *) constructed_558[0];
                    r2[3] = r3;
                    r3 = top[104];
                    t = allocate (TC_SAVE, 5);
                    t[1] = k; t[2] = lenv; t[3] = &&L1992; t[4] = r0; t[5] = r1; k = t;
                    r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                    L1992:
                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                    r2 = result;
                    r1[2] = r2;
                    r1 = allocate (TC_TUPLE, 3);
                    r2 = (object *) constructed_557[0];
                    r1[2] = r2;
                    r2 = allocate (TC_TUPLE, 2);
                    r2[1] = lenv; lenv = r2;
                    r3 = (object*)TC_NIL;
                    r2[2] = r3;
                    r2 = ((object*) lenv) [2];
                    r3 = ((object**) lenv) [1][2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r3;
                    t[2] = r2;
                    r2 = t;
                    lenv = lenv[1];
                    r1[3] = r2;
                    r2 = top[104];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1993; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                    L1993:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    lenv = lenv[1];
                  } else {
                    r1 = top[105];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L1994; t[4] = r0; k = t;
                    lenv = r1[2]; goto FUN_new_tvar_453;
                    L1994:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                  }
                  r0[7] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object*****) lenv) [1][1][1][1][3];
                  r1[2] = r2;
                  r2 = top[22];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L1995; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                  L1995:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  // def $loop_3590_i2551_i4640_i4727
                  goto L1996;
                  FUN__24loop_3590_i2551_i4640_i4727:
                    check_heap (0);
                    r0 = ((object***) lenv) [1][1][2];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                      r0 = (object *) 14;
                      PXLL_RETURN(0);
                    } else {
                      r0 = allocate (TC_TUPLE, 5);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object*****) lenv) [1][1][1][1][3];
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r1[3] = r2;
                      r2 = top[27];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1997; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                      L1997:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r1[3] = r2;
                      r2 = top[27];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1998; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                      L1998:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r1[3] = r2;
                      r2 = top[27];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L1999; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                      L1999:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[4] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object*) lenv) [4];
                      r1[2] = r2;
                      r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][10];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2000; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_n_tvars_3366_i4640_i4727;
                      L2000:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[5] = r1;
                      r0 = allocate (TC_TUPLE, 4);
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object*) lenv) [3];
                      r1[2] = r2;
                      r2 = (object *) constructed_546[0];
                      r1[3] = r2;
                      r2 = top[104];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2001; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                      L2001:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = (object *) 3;
                      r3 = ((object**) lenv) [1][4];
                      if PXLL_IS_TRUE(PXLL_TEST(unbox(r3)==unbox(r2))) {
                        r2 = allocate (TC_TUPLE, 2);
                        r3 = ((object**) lenv) [1][5];
                        r2[2] = r3;
                        r3 = top[19];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2002; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                        L2002:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                      } else {
                        r2 = allocate (TC_TUPLE, 3);
                        r3 = (object *) constructed_545[0];
                        r2[2] = r3;
                        r3 = ((object**) lenv) [1][5];
                        r2[3] = r3;
                        r3 = top[104];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2003; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_pred_451;
                        L2003:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                      }
                      r1[2] = r2;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = (object *) constructed_543[0];
                      r1[2] = r2;
                      r2 = allocate (TC_TUPLE, 2);
                      r2[1] = lenv; lenv = r2;
                      r3 = (object*)TC_NIL;
                      r2[2] = r3;
                      r2 = ((object*) lenv) [2];
                      r3 = ((object**) lenv) [1][2];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      lenv = lenv[1];
                      r1[3] = r2;
                      r2 = top[104];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2004; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                      L2004:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      lenv = lenv[1];
                      r0[3] = r1;
                      r1 = ((object*****) lenv) [1][1][1][1][7];
                      r0[4] = r1;
                      r1 = top[106];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2005; ; k = t;
                      r0[1] = r1[2]; lenv = r0; goto FUN_rlabel_458;
                      L2005:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      ((object*****) lenv) [1][1][1][1][7] = r0;
                      r0 = allocate (TC_TUPLE, 4);
                      r1 = ((object*) lenv) [2];
                      r0[2] = r1;
                      r1 = ((object*****) lenv) [1][1][1][1][4];
                      r0[3] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = allocate (TC_TUPLE, 3);
                      r3 = ((object**) lenv) [1][2];
                      r2[2] = r3;
                      r3 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                      r2[3] = r3;
                      r3 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][6];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2006; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_type_of_splat_3347_i4640_i4727;
                      L2006:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[2] = r2;
                      r1 = ((object*) lenv) [2];
                      r2 = ((object**) lenv) [1][2];
                      ((pxll_vector*)r2)->val[5] = r1;
                      r1 = ((object*) lenv) [2];
                      lenv = lenv[1];
                      r0[4] = r1;
                      r1 = top[139];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2007; ; k = t;
                      r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                      L2007:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      lenv = lenv[1];
                      r0 = (object *) 3;
                      r1 = ((object*) lenv) [2];
                      r0 = box((pxll_int)unbox(r1)+unbox(r0));
                      lenv[2] = r0;
                      goto FUN__24loop_3590_i2551_i4640_i4727;
                    }
                    PXLL_RETURN(0);
                  L1996:
                  r1 = allocate (TC_CLOSURE, 2);
                  r1[1] = &&FUN__24loop_3590_i2551_i4640_i4727; r1[2] = lenv;
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r1 = (object *) 1;
                  r0[2] = r1;
                  r1 = ((object*) lenv) [2];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L2008; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3590_i2551_i4640_i4727;
                  L2008:
                  ; lenv = k[2]; k = k[1];
                  r0 = result;
                  lenv = lenv[1];
                  lenv = lenv[1];
                  r0 = ((object*) lenv) [6];
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 4);
                    r1 = ((object*) lenv) [2];
                    r0[2] = r1;
                    r1 = ((object*) lenv) [4];
                    r0[3] = r1;
                    r1 = allocate (TC_TUPLE, 2);
                    r1[1] = lenv; lenv = r1;
                    r2 = allocate (TC_TUPLE, 3);
                    r3 = ((object**) lenv) [1][2];
                    r2[2] = r3;
                    r3 = ((object*******) lenv) [1][1][1][1][1][1][3];
                    r2[3] = r3;
                    r3 = ((object********) lenv) [1][1][1][1][1][1][1][6];
                    t = allocate (TC_SAVE, 5);
                    t[1] = k; t[2] = lenv; t[3] = &&L2009; t[4] = r0; t[5] = r1; k = t;
                    r2[1] = r3[2]; lenv = r2; goto FUN_type_of_splat_3347_i4640_i4727;
                    L2009:
                    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                    r2 = result;
                    r1[2] = r2;
                    r1 = ((object*) lenv) [2];
                    r2 = ((object**) lenv) [1][2];
                    ((pxll_vector*)r2)->val[5] = r1;
                    r1 = ((object*) lenv) [2];
                    lenv = lenv[1];
                    r0[4] = r1;
                    r1 = top[139];
                    t = allocate (TC_SAVE, 3);
                    t[1] = k; t[2] = lenv; t[3] = &&L2010; ; k = t;
                    r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                    L2010:
                    ; lenv = k[2]; k = k[1];
                    r0 = result;
                  } else {
                    r0 = (object *) 14;
                  }
                  r0 = allocate (TC_TUPLE, 4);
                  r1 = ((object**) lenv) [1][2];
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = (object *) constructed_541[0];
                  r1[2] = r2;
                  r2 = allocate (TC_TUPLE, 2);
                  r2[1] = lenv; lenv = r2;
                  r3 = (object*)TC_NIL;
                  r2[2] = r3;
                  r2 = ((object*) lenv) [2];
                  r3 = ((object**) lenv) [1][7];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r3;
                  t[2] = r2;
                  r2 = t;
                  lenv = lenv[1];
                  r1[3] = r2;
                  r2 = top[104];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2011; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                  L2011:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = allocate (TC_TUPLE, 3);
                  r3 = ((object***) lenv) [1][1][2];
                  r2[2] = r3;
                  r3 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  r2[3] = r3;
                  r3 = ((object********) lenv) [1][1][1][1][1][1][1][6];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2012; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_type_of_splat_3347_i4640_i4727;
                  L2012:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r1 = ((object*) lenv) [2];
                  r2 = ((object***) lenv) [1][1][2];
                  ((pxll_vector*)r2)->val[5] = r1;
                  r1 = ((object*) lenv) [2];
                  lenv = lenv[1];
                  r0[4] = r1;
                  r1 = top[139];
                  t = allocate (TC_SAVE, 3);
                  t[1] = k; t[2] = lenv; t[3] = &&L2013; ; k = t;
                  r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                  L2013:
                  ; lenv = k[2]; k = k[1];
                  r0 = ((object*) lenv) [4];
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_113;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_113;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_113_over;
          fatbar_113:
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object*) &constructed_565;
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r0[3] = r1;
          r1 = top[15];
          r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
          fatbar_113_over: ;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 10);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 4);
          r2 = ((object******) lenv) [1][1][1][1][1][2];
          r2 = ((pxll_vector*)r2)->val[2];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = (object*) &constructed_566;
          r1[4] = r2;
          r2 = top[66];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2014; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_alist_2fget_396;
          L2014:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][2];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2015; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2015:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = (object *) 3;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2016; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2016:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object*) lenv) [3];
          r2[2] = r3;
          r3 = top[20];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2017; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_cdr_310;
          L2017:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r2 = top[20];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2018; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
          L2018:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[6] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [4];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][3];
          r1[3] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][6];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2019; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_of_splat_3347_i4640_i4727;
          L2019:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[7] = r1;
          r1 = ((object*) lenv) [7];
          r2 = ((object*) lenv) [4];
          ((pxll_vector*)r2)->val[5] = r1;
          r1 = ((object*) lenv) [7];
          r0[8] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2020; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L2020:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[9] = r1;
          r1 = top[105];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2021; t[4] = r0; k = t;
          lenv = r1[2]; goto FUN_new_tvar_453;
          L2021:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[10] = r1;
          r0 = ((object*) lenv) [9];
          switch (get_case_noint (r0)) {
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][9];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][9];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = (object *) 262;
                } break;
                default: {
                  goto fatbar_115;
                }
              }
              goto fatbar_115_over;
              fatbar_115:
              r0 = (object *) 6;
              fatbar_115_over: ;
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 4);
                r1 = ((object*****) lenv) [1][1][1][1][2];
                r0[2] = r1;
                r1 = ((object**) lenv) [1][8];
                r0[3] = r1;
                r1 = ((object*) lenv) [3];
                r0[4] = r1;
                r1 = top[139];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2022; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                L2022:
                ; lenv = k[2]; k = k[1];
                r0 = result;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object**) lenv) [1][2];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][3];
                r1[3] = r2;
                r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2023; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_instantiate_type_scheme_3343_i4640_i4727;
                L2023:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 4);
                r1 = ((object******) lenv) [1][1][1][1][1][2];
                r0[2] = r1;
                r1 = ((object***) lenv) [1][1][8];
                r0[3] = r1;
                r1 = ((object*) lenv) [2];
                r0[4] = r1;
                r1 = top[139];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2024; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                L2024:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                lenv = lenv[1];
              }
              lenv = lenv[1];
            } break;
          }
          r0 = allocate (TC_TUPLE, 3);
          // def lambda_22668_i2564_i4640_i4727
          goto L2025;
          FUN_lambda_22668_i2564_i4640_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object**) lenv) [1][10];
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = ((object******) lenv) [1][1][1][1][1][3];
            r2[3] = r3;
            r3 = ((object*******) lenv) [1][1][1][1][1][1][6];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2026; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_type_of_splat_3347_i4640_i4727;
            L2026:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = ((object*) lenv) [2];
            r2 = ((object**) lenv) [1][2];
            ((pxll_vector*)r2)->val[5] = r1;
            r1 = ((object*) lenv) [2];
            lenv = lenv[1];
            r0[4] = r1;
            r1 = top[139];
            r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
          L2025:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_lambda_22668_i2564_i4640_i4727; r1[2] = lenv;
          r0[2] = r1;
          r1 = ((object*) lenv) [6];
          r0[3] = r1;
          r1 = top[35];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2027; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
          L2027:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object*) lenv) [5];
          r0[2] = r1;
          r1 = ((object*) lenv) [10];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r1[1] = lenv; lenv = r1;
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object**) lenv) [1][5];
          r2[2] = r3;
          r3 = ((object*****) lenv) [1][1][1][1][3];
          r2[3] = r3;
          r3 = ((object******) lenv) [1][1][1][1][1][6];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2028; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_type_of_splat_3347_i4640_i4727;
          L2028:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r1 = ((object*) lenv) [2];
          r2 = ((object**) lenv) [1][5];
          ((pxll_vector*)r2)->val[5] = r1;
          r1 = ((object*) lenv) [2];
          lenv = lenv[1];
          r0[4] = r1;
          r1 = top[139];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2029; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
          L2029:
          ; lenv = k[2]; k = k[1];
          r0 = ((object*) lenv) [10];
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_80;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_80_over;
    fatbar_80:
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object**) lenv) [1][2];
    r0[2] = r1;
    r1 = top[127];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2030; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_pp_node_508;
    L2030:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 3);
    r1 = (object*) &constructed_567;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][2];
    r0[3] = r1;
    r1 = top[15];
    r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
    fatbar_80_over: ;
    PXLL_RETURN(0);
  L1799:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_type_of_splat_3347_i4640_i4727; r2[2] = lenv;
  r1[6] = r2;
  // def type-of-literal_3349_i4640_i4727
  goto L2031;
  FUN_type_of_literal_3349_i4640_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = top[183];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = top[181];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(2)): {
        r0 = top[182];
        PXLL_RETURN(0);
      } break;
      case (UITAG(3)): {
        r0 = top[184];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(4)): {
        r0 = top[186];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(5)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object****) lenv) [1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[2];
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = (object*) &constructed_518;
        r1[4] = r2;
        r2 = top[66];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2032; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2fget_396;
        L2032:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = ((object*) lenv) [5];
        r2 = ((pxll_vector*)r2)->val[2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2033; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto *r2[1];
        L2033:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[6] = r1;
        r0 = ((object*) lenv) [6];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][6];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][6];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = ((object*) lenv) [3];
            r1[3] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2034; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_instantiate_type_scheme_3343_i4640_i4727;
            L2034:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r0 = ((object*) lenv) [4];
            switch (get_case (r0)) {
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][4];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = (object *) constructed_97[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object*) lenv) [3];
                  switch (get_case (r0)) {
                    case (TC_PAIR): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object**) lenv) [1][3];
                      r1 = UOBJ_GET(r1,1);
                      r0[3] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][3];
                      r1[2] = r2;
                      r2 = top[22];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2035; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                      L2035:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      // def $loop_3445_i4640_i4727
                      goto L2036;
                      FUN__24loop_3445_i4640_i4727:
                        check_heap (0);
                        r0 = ((object***) lenv) [1][1][2];
                        r1 = ((object*) lenv) [2];
                        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                          r0 = (object *) 14;
                          PXLL_RETURN(0);
                        } else {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 4);
                          r2 = allocate (TC_TUPLE, 3);
                          r3 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                          r2[2] = r3;
                          r3 = ((object**) lenv) [1][2];
                          r2[3] = r3;
                          r3 = top[27];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2037; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                          L2037:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                          r1[3] = r2;
                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                          r1[4] = r2;
                          r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][7];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2038; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_type_of_literal_3349_i4640_i4727;
                          L2038:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r0 = allocate (TC_TUPLE, 4);
                          r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
                          r0[2] = r1;
                          r1 = ((object*) lenv) [2];
                          r0[3] = r1;
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object*****) lenv) [1][1][1][1][3];
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][2];
                          r1[3] = r2;
                          r2 = top[27];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2039; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                          L2039:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[4] = r1;
                          r1 = top[139];
                          t = allocate (TC_SAVE, 3);
                          t[1] = k; t[2] = lenv; t[3] = &&L2040; ; k = t;
                          r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
                          L2040:
                          ; lenv = k[2]; k = k[1];
                          r0 = result;
                          lenv = lenv[1];
                          r0 = (object *) 3;
                          r1 = ((object*) lenv) [2];
                          r0 = box((pxll_int)unbox(r1)+unbox(r0));
                          lenv[2] = r0;
                          goto FUN__24loop_3445_i4640_i4727;
                        }
                        PXLL_RETURN(0);
                      L2036:
                      r1 = allocate (TC_CLOSURE, 2);
                      r1[1] = &&FUN__24loop_3445_i4640_i4727; r1[2] = lenv;
                      r0[2] = r1;
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = (object *) 1;
                      r0[2] = r1;
                      r1 = ((object*) lenv) [2];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2041; ; k = t;
                      r0[1] = r1[2]; lenv = r0; goto FUN__24loop_3445_i4640_i4727;
                      L2041:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      lenv = lenv[1];
                      lenv = lenv[1];
                      r0 = ((object*) lenv) [2];
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_79;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_79;
                }
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_79;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_79_over;
            fatbar_79:
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_517;
            r0[2] = r1;
            r1 = ((object*) lenv) [4];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            fatbar_79_over: ;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = top[105];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2042; t[4] = r0; k = t;
        lenv = r1[2]; goto FUN_new_tvar_453;
        L2042:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_21960_i4640_i4727
        goto L2043;
        FUN_lambda_21960_i4640_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 4);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = ((object****) lenv) [1][1][1][3];
          r1[3] = r2;
          r2 = ((object****) lenv) [1][1][1][4];
          r1[4] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][7];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2044; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_type_of_literal_3349_i4640_i4727;
          L2044:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 4);
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          r0[4] = r1;
          r1 = top[139];
          r0[1] = r1[2]; lenv = r0; goto FUN_unify_530;
        L2043:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_21960_i4640_i4727; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[35];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L2045; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        L2045:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object *) constructed_472[0];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = (object*)TC_NIL;
        r1[2] = r2;
        r1 = ((object*) lenv) [2];
        r2 = ((object**) lenv) [1][3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[3] = r1;
        r1 = top[104];
        r0[1] = r1[2]; lenv = r0; goto FUN_pred_451;
      } break;
    }
    PXLL_RETURN(0);
  L2031:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_type_of_literal_3349_i4640_i4727; r2[2] = lenv;
  r1[7] = r2;
  // def unraw_3350_i4640_i4727
  goto L2046;
  FUN_unraw_3350_i4640_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = (object *) constructed_516[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (TC_PAIR): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case (r0)) {
                case (TC_NIL): {
                  r0 = ((object*) lenv) [2];
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object **)lenv)[1][1];
                  goto fatbar_78;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              lenv = ((object *)lenv)[1];
              goto fatbar_78;
            }
          }
          PXLL_RETURN(0);
        } else {
          lenv = ((object *)lenv)[1];
          goto fatbar_78;
        }
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_78;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_78_over;
    fatbar_78:
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
    fatbar_78_over: ;
    PXLL_RETURN(0);
  L2046:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_unraw_3350_i4640_i4727; r2[2] = lenv;
  r1[8] = r2;
  // def apply-tenv_3356_i4640_i4727
  goto L2047;
  FUN_apply_tenv_3356_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][3];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = top[65];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2048; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
    L2048:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UITAG(1)): {
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_514;
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case_noint (r0)) {
          default: {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object*) lenv) [3];
            r0[3] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r0[1] = r1[2]; lenv = r0; goto FUN_instantiate_type_scheme_3343_i4640_i4727;
          } break;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2047:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_apply_tenv_3356_i4640_i4727; r2[2] = lenv;
  r1[9] = r2;
  // def n-tvars_3366_i4640_i4727
  goto L2049;
  FUN_n_tvars_3366_i4640_i4727:
    check_heap (0);
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
      r0 = (object *) constructed_513[0];
      PXLL_RETURN(0);
    } else {
      goto fatbar_77;
    }
    PXLL_RETURN(0);
    goto fatbar_77_over;
    fatbar_77:
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = top[105];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2050; t[4] = r0; k = t;
    lenv = r1[2]; goto FUN_new_tvar_453;
    L2050:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = (object *) 3;
    r3 = ((object**) lenv) [1][2];
    r2 = box((pxll_int)unbox(r3)-unbox(r2));
    r1[2] = r2;
    r2 = ((object***) lenv) [1][1][10];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2051; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_n_tvars_3366_i4640_i4727;
    L2051:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    PXLL_RETURN(0);
    fatbar_77_over: ;
    PXLL_RETURN(0);
  L2049:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_n_tvars_3366_i4640_i4727; r2[2] = lenv;
  r1[10] = r2;
  // def apply-subst-to-program_3370_i4640_i4727
  goto L2052;
  FUN_apply_subst_to_program_3370_i4640_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[5];
    r0[2] = r1;
    r1 = top[140];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2053; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_apply_subst_531;
    L2053:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    r1 = ((object*) lenv) [2];
    ((pxll_vector*)r1)->val[5] = r0;
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object**) lenv) [1][11];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[3];
    r0[3] = r1;
    r1 = top[35];
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
  L2052:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_apply_subst_to_program_3370_i4640_i4727; r2[2] = lenv;
  r1[11] = r2;
  r2 = top[105];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2054; t[4] = r0; t[5] = r1; k = t;
  lenv = r2[2]; goto FUN_new_tvar_453;
  L2054:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[12] = r2;
  r2 = top[105];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2055; t[4] = r0; t[5] = r1; k = t;
  lenv = r2[2]; goto FUN_new_tvar_453;
  L2055:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[13] = r2;
  r2 = top[105];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2056; t[4] = r0; t[5] = r1; k = t;
  lenv = r2[2]; goto FUN_new_tvar_453;
  L2056:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[14] = r2;
  r1 = allocate (TC_TUPLE, 4);
  r1[1] = lenv; lenv = r1;
  r2 = (object*)UITAG(0);
  r1[2] = r2;
  r2 = allocate (TC_TUPLE, 3);
  r3 = ((object***) lenv) [1][1][44];
  r2[2] = r3;
  r3 = ((object*) lenv) [2];
  r2[3] = r3;
  r3 = ((object**) lenv) [1][6];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2057; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_type_of_splat_3347_i4640_i4727;
  L2057:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[3] = r2;
  r2 = ((object*) lenv) [3];
  r3 = ((object***) lenv) [1][1][44];
  ((pxll_vector*)r3)->val[5] = r2;
  r2 = ((object*) lenv) [3];
  r1[4] = r2;
  r1 = allocate (TC_TUPLE, 2);
  r2 = ((object***) lenv) [1][1][44];
  r2 = ((pxll_vector*)r2)->val[5];
  r1[2] = r2;
  r2 = top[140];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2058; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_apply_subst_531;
  L2058:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r2 = ((object***) lenv) [1][1][44];
  ((pxll_vector*)r2)->val[5] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object**) lenv) [1][11];
  r1[2] = r2;
  r2 = ((object***) lenv) [1][1][44];
  r2 = ((pxll_vector*)r2)->val[3];
  r1[3] = r2;
  r2 = top[35];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2059; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_for_each_329;
  L2059:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = ((object*) lenv) [4];
  lenv = lenv[1];
  lenv = lenv[1];
  r0[45] = r1;
  r1 = ((object*) lenv) [2];
  r1 = ((pxll_vector*)r1)->val[9];
  r1 = ((pxll_vector*)r1)->val[4];
  if PXLL_IS_TRUE(r1) {
    r1 = (object*) &constructed_512;
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r2 = (object*) &constructed_512;
    (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [30];
    r1[2] = r2;
    r2 = top[127];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2060; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_pp_node_508;
    L2060:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = top[158];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2061; t[4] = r0; k = t;
    lenv = r1[2]; goto *r1[1];
    L2061:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = (object *) 14;
  } else {
    r1 = (object *) 14;
  }
  r0[46] = r1;
  r1 = (object*) &constructed_511;
  r1 = box((pxll_int)((pxll_string*)(r1))->len);
  r2 = (object*) &constructed_511;
  r1 = (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
  r0[47] = r1;
  r1 = ((object*) lenv) [30];
  r0[48] = r1;
  r1 = (object *) constructed_510[0];
  r0[49] = r1;
  r1 = allocate (TC_TUPLE, 22);
  r1[1] = lenv; lenv = r1;
  // def cont_2644_i4643_i4727
  goto L2062;
  FUN_cont_2644_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 3);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object****) lenv) [1][1][1][2];
    r2 = ((pxll_vector*)r2)->val[11];
    r2 = ((pxll_vector*)r2)->val[0];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2063; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L2063:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][3];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2064; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto *r2[1];
    L2064:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r0 = ((object**) lenv) [1][2];
    r1 = ((object*) lenv) [3];
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(0),3);
    t[1] = r2;
    t[2] = r0;
    t[3] = r1;
    r0 = t;
    PXLL_RETURN(0);
  L2062:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_cont_2644_i4643_i4727; r2[2] = lenv;
  r1[2] = r2;
  // def compile_2646_i4643_i4727
  goto L2065;
  FUN_compile_2646_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 3);
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [5];
      r1[2] = r2;
      r2 = top[136];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2066; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
      L2066:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object**) lenv) [1][19];
      r0[3] = r1;
      r1 = ((object**) lenv) [1][2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2067; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_cont_2644_i4643_i4727;
      L2067:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      ((object*) lenv) [5] = r0;
    } else {
      r0 = (object *) 14;
    }
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][3];
    r1 = ((pxll_vector*)r1)->val[4];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(2)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[1];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[4] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(4)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = ((object*****) lenv) [1][1][1][1][6];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2068; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_get_symbol_index_2649_i4643_i4727;
            L2068:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = (object *) constructed_58[0];
            r1 = ((object**) lenv) [1][4];
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (UOTAG(2),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = ((object**) lenv) [1][3];
            r1[3] = r2;
            r2 = ((object*****) lenv) [1][1][1][1][5];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2069; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_get_literal_index_2648_i4643_i4727;
            L2069:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = (object *) constructed_86[0];
            r1 = ((object**) lenv) [1][4];
            r2 = ((object*) lenv) [3];
            t = alloc_no_clear (UOTAG(2),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          case (UOTAG(6)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][7];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2070; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_scan_literals_2650_i4643_i4727;
            L2070:
            ; lenv = k[2]; k = k[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][2];
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][3];
            r1[3] = r2;
            r2 = ((object******) lenv) [1][1][1][1][1][5];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2071; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_get_literal_index_2648_i4643_i4727;
            L2071:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = (object *) constructed_472[0];
            r1 = ((object***) lenv) [1][1][4];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(2),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          case (UOTAG(5)): {
            r0 = allocate (TC_TUPLE, 4);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,2);
            r0[4] = r1;
            r0 = (object *) constructed_46[0];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
              r0 = ((object**) lenv) [1][4];
              r1 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(1),2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              PXLL_RETURN(0);
            } else {
              goto fatbar_54;
            }
            PXLL_RETURN(0);
            goto fatbar_54_over;
            fatbar_54:
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [4];
            r0[2] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][7];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2072; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_scan_literals_2650_i4643_i4727;
            L2072:
            ; lenv = k[2]; k = k[1];
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][2];
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][3];
            r1[3] = r2;
            r2 = ((object******) lenv) [1][1][1][1][1][5];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2073; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_get_literal_index_2648_i4643_i4727;
            L2073:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = (object *) constructed_474[0];
            r1 = ((object***) lenv) [1][1][4];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(2),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
            fatbar_54_over: ;
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_53;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_53_over;
        fatbar_53:
        r0 = ((object*) lenv) [4];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(1),2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        PXLL_RETURN(0);
        fatbar_53_over: ;
        PXLL_RETURN(0);
      } break;
      case (UITAG(5)): {
        r0 = allocate (TC_TUPLE, 5);
        r1 = ((object**) lenv) [1][2];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][4];
        r0[4] = r1;
        r1 = ((object**) lenv) [1][5];
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][8];
        r0[1] = r1[2]; lenv = r0; goto FUN_c_sequence_2652_i4643_i4727;
      } break;
      case (UITAG(6)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][5];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[137];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2074; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_k_2ftarget_523;
        L2074:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[136];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2075; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
        L2075:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[5] = r1;
        r0 = ((object*) lenv) [5];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][5];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][5];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (TC_PAIR): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object****) lenv) [1][1][1][2];
                        r1 = ((pxll_vector*)r1)->val[4];
                        r0[2] = r1;
                        r0 = ((object*) lenv) [2];
                        switch (get_case (r0)) {
                          case (UOTAG(3)): {
                            r0 = allocate (TC_TUPLE, 5);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object**) lenv) [1][2];
                            r1 = UOBJ_GET(r1,1);
                            r0[2] = r1;
                            r1 = ((object**) lenv) [1][2];
                            r1 = UOBJ_GET(r1,2);
                            r0[3] = r1;
                            r1 = allocate (TC_TUPLE, 2);
                            r2 = ((object******) lenv) [1][1][1][1][1][2];
                            r1[2] = r2;
                            r2 = top[136];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L2076; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
                            L2076:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[4] = r1;
                            r1 = allocate (TC_TUPLE, 2);
                            r2 = ((object******) lenv) [1][1][1][1][1][2];
                            r1[2] = r2;
                            r2 = top[137];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L2077; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_k_2ftarget_523;
                            L2077:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[5] = r1;
                            r0 = allocate (TC_TUPLE, 2);
                            r0[1] = lenv; lenv = r0;
                            // def finish_2777_i1800_i4643_i4727
                            goto L2078;
                            FUN_finish_2777_i1800_i4643_i4727:
                              check_heap (0);
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = allocate (TC_TUPLE, 5);
                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                              r1[2] = r2;
                              r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
                              r1[3] = r2;
                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][4];
                              r1[4] = r2;
                              r2 = allocate (TC_TUPLE, 3);
                              r3 = ((object****) lenv) [1][1][1][4];
                              r2[2] = r3;
                              // def lambda_17385_i1800_i4643_i4727
                              goto L2079;
                              FUN_lambda_17385_i1800_i4643_i4727:
                                check_heap (0);
                                r0 = ((object*****) lenv) [1][1][1][1][5];
                                r1 = ((object*) lenv) [2];
                                t = alloc_no_clear (UOTAG(6),2);
                                t[1] = r1;
                                t[2] = r0;
                                r0 = t;
                                PXLL_RETURN(0);
                              L2079:
                              r3 = allocate (TC_CLOSURE, 2);
                              r3[1] = &&FUN_lambda_17385_i1800_i4643_i4727; r3[2] = lenv;
                              r2[3] = r3;
                              r3 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                              t = allocate (TC_SAVE, 5);
                              t[1] = k; t[2] = lenv; t[3] = &&L2080; t[4] = r0; t[5] = r1; k = t;
                              r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
                              L2080:
                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                              r2 = result;
                              r1[5] = r2;
                              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2081; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
                              L2081:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[2] = r1;
                              r1 = allocate (TC_TUPLE, 5);
                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                              r1[2] = r2;
                              r2 = ((object******) lenv) [1][1][1][1][1][2];
                              r1[3] = r2;
                              r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][4];
                              r1[4] = r2;
                              r2 = allocate (TC_TUPLE, 3);
                              r3 = ((object****) lenv) [1][1][1][4];
                              r2[2] = r3;
                              // def lambda_17399_i1800_i4643_i4727
                              goto L2082;
                              FUN_lambda_17399_i1800_i4643_i4727:
                                check_heap (0);
                                r0 = ((object*****) lenv) [1][1][1][1][5];
                                r1 = ((object*) lenv) [2];
                                t = alloc_no_clear (UOTAG(6),2);
                                t[1] = r1;
                                t[2] = r0;
                                r0 = t;
                                PXLL_RETURN(0);
                              L2082:
                              r3 = allocate (TC_CLOSURE, 2);
                              r3[1] = &&FUN_lambda_17399_i1800_i4643_i4727; r3[2] = lenv;
                              r2[3] = r3;
                              r3 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                              t = allocate (TC_SAVE, 5);
                              t[1] = k; t[2] = lenv; t[3] = &&L2083; t[4] = r0; t[5] = r1; k = t;
                              r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
                              L2083:
                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                              r2 = result;
                              r1[5] = r2;
                              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][3];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2084; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
                              L2084:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[3] = r1;
                              r0 = ((object*) lenv) [2];
                              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                              r2 = ((object****) lenv) [1][1][1][2];
                              r3 = ((object****) lenv) [1][1][1][3];
                              r4 = ((object*) lenv) [3];
                              r5 = ((object**) lenv) [1][2];
                              t = alloc_no_clear (UOTAG(5),6);
                              t[1] = r5;
                              t[2] = r2;
                              t[3] = r3;
                              t[4] = r0;
                              t[5] = r4;
                              t[6] = r1;
                              r0 = t;
                              PXLL_RETURN(0);
                            L2078:
                            r1 = allocate (TC_CLOSURE, 2);
                            r1[1] = &&FUN_finish_2777_i1800_i4643_i4727; r1[2] = lenv;
                            r0[2] = r1;
                            r0 = allocate (TC_TUPLE, 5);
                            r1 = ((object******) lenv) [1][1][1][1][1][2];
                            r1 = ((pxll_vector*)r1)->val[3];
                            r0[2] = r1;
                            r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                            r0[3] = r1;
                            r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
                            r0[4] = r1;
                            r1 = ((object*) lenv) [2];
                            r0[5] = r1;
                            r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][12];
                            r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_2663_i4643_i4727;
                          } break;
                          default: {
                            goto fatbar_56;
                          }
                        }
                        PXLL_RETURN(0);
                        goto fatbar_56_over;
                        fatbar_56:
                        r0 = allocate (TC_TUPLE, 5);
                        r1 = (object *) 6;
                        r0[2] = r1;
                        r1 = ((object****) lenv) [1][1][1][2];
                        r0[3] = r1;
                        r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
                        r0[4] = r1;
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = ((object*****) lenv) [1][1][1][1][4];
                        r1[2] = r2;
                        // def lambda_17330_i1799_i4643_i4727
                        goto L2085;
                        FUN_lambda_17330_i1799_i4643_i4727:
                          check_heap (0);
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = allocate (TC_TUPLE, 5);
                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                          r1[2] = r2;
                          r2 = ((object*****) lenv) [1][1][1][1][2];
                          r1[3] = r2;
                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                          r1[4] = r2;
                          r2 = allocate (TC_TUPLE, 3);
                          r3 = ((object*******) lenv) [1][1][1][1][1][1][4];
                          r2[2] = r3;
                          // def lambda_17310_i1799_i4643_i4727
                          goto L2086;
                          FUN_lambda_17310_i1799_i4643_i4727:
                            check_heap (0);
                            r0 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                            r1 = ((object*) lenv) [2];
                            t = alloc_no_clear (UOTAG(6),2);
                            t[1] = r1;
                            t[2] = r0;
                            r0 = t;
                            PXLL_RETURN(0);
                          L2086:
                          r3 = allocate (TC_CLOSURE, 2);
                          r3[1] = &&FUN_lambda_17310_i1799_i4643_i4727; r3[2] = lenv;
                          r2[3] = r3;
                          r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2087; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
                          L2087:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[5] = r2;
                          r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2088; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
                          L2088:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[2] = r1;
                          r1 = allocate (TC_TUPLE, 5);
                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                          r1[2] = r2;
                          r2 = ((object****) lenv) [1][1][1][2];
                          r1[3] = r2;
                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                          r1[4] = r2;
                          r2 = allocate (TC_TUPLE, 3);
                          r3 = ((object*******) lenv) [1][1][1][1][1][1][4];
                          r2[2] = r3;
                          // def lambda_17324_i1799_i4643_i4727
                          goto L2089;
                          FUN_lambda_17324_i1799_i4643_i4727:
                            check_heap (0);
                            r0 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                            r1 = ((object*) lenv) [2];
                            t = alloc_no_clear (UOTAG(6),2);
                            t[1] = r1;
                            t[2] = r0;
                            r0 = t;
                            PXLL_RETURN(0);
                          L2089:
                          r3 = allocate (TC_CLOSURE, 2);
                          r3[1] = &&FUN_lambda_17324_i1799_i4643_i4727; r3[2] = lenv;
                          r2[3] = r3;
                          r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2090; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
                          L2090:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[5] = r2;
                          r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2091; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
                          L2091:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          r0[3] = r1;
                          r0 = ((object*******) lenv) [1][1][1][1][1][1][2];
                          r1 = ((object*) lenv) [2];
                          r2 = ((object*) lenv) [3];
                          r3 = ((object**) lenv) [1][2];
                          t = alloc_no_clear (UOTAG(4),4);
                          t[1] = r3;
                          t[2] = r1;
                          t[3] = r2;
                          t[4] = r0;
                          r0 = t;
                          PXLL_RETURN(0);
                        L2085:
                        r2 = allocate (TC_CLOSURE, 2);
                        r2[1] = &&FUN_lambda_17330_i1799_i4643_i4727; r2[2] = lenv;
                        r1[3] = r2;
                        r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2092; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
                        L2092:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[5] = r1;
                        r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                        r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
                        fatbar_56_over: ;
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        lenv = ((object ***)lenv)[1][1][1];
                        goto fatbar_55;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object **)lenv)[1][1];
                    goto fatbar_55;
                  }
                }
                PXLL_RETURN(0);
              } break;
              default: {
                lenv = ((object *)lenv)[1];
                goto fatbar_55;
              }
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_55;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_55_over;
        fatbar_55:
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_475;
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
        fatbar_55_over: ;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(7)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[1];
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][3];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2093; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L2093:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 4);
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*****) lenv) [1][1][1][1][49];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2094; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L2094:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = top[156];
        r0[4] = r1;
        r1 = top[6];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L2095; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
        L2095:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object******) lenv) [1][1][1][1][1][49];
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object*****) lenv) [1][1][1][1][49] = r0;
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 5);
        r2 = (object *) 262;
        r1[2] = r2;
        r2 = ((object**) lenv) [1][5];
        r1[3] = r2;
        r2 = ((object**) lenv) [1][3];
        switch (get_case (r2)) {
          case (TC_NIL): {
            r2 = ((object****) lenv) [1][1][1][4];
          } break;
          default: {
            goto fatbar_57;
          }
        }
        goto fatbar_57_over;
        fatbar_57:
        r2 = ((object****) lenv) [1][1][1][4];
        r3 = ((object**) lenv) [1][3];
        t = alloc_no_clear (UOTAG(1),2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        fatbar_57_over: ;
        r1[4] = r2;
        r2 = allocate (TC_TUPLE, 3);
        r3 = (object *) constructed_477[0];
        r2[2] = r3;
        r3 = ((object*****) lenv) [1][1][1][1][19];
        r2[3] = r3;
        r3 = ((object*****) lenv) [1][1][1][1][2];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2096; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
        L2096:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[5] = r2;
        r2 = ((object*****) lenv) [1][1][1][1][3];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2097; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
        L2097:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r2 = ((object**) lenv) [1][6];
        r3 = ((object**) lenv) [1][2];
        t = alloc_no_clear (UOTAG(7),3);
        t[1] = r3;
        t[2] = r1;
        t[3] = r2;
        r1 = t;
        r0[3] = r1;
        r0 = ((object******) lenv) [1][1][1][1][1][49];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = (object*) &constructed_68;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
            r0 = (object*) &constructed_476;
            dump_object (r0, 0); fprintf (stdout, "\n");
            r0 = (object*) &constructed_476;
            PXLL_UNDEFINED; result=r0; goto Lreturn;
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][49];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][49];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            ((object*******) lenv) [1][1][1][1][1][1][49] = r0;
            r0 = ((object*) lenv) [2];
            lenv = lenv[1];
          } break;
        }
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = (object *) 1;
        r1[3] = r2;
        r2 = ((object***) lenv) [1][1][4];
        r1[4] = r2;
        r2 = ((object****) lenv) [1][1][1][11];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2098; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lexical_address_2658_i4643_i4727;
        L2098:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        switch (get_case_noint (r0)) {
          case (UOTAG(7)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = (object *) -1;
            r1 = ((object**) lenv) [1][3];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(19),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          case (UOTAG(1)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object**) lenv) [1][3];
            r2 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(8),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object**) lenv) [1][3];
            r2 = (object *) -1;
            t = alloc_no_clear (UOTAG(8),3);
            t[1] = r2;
            t[2] = r0;
            t[3] = r1;
            r0 = t;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 6);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][3];
        r2 = ((pxll_vector*)r2)->val[3];
        r1[2] = r2;
        r2 = top[19];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2099; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
        L2099:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = (object *) 1;
        r1[3] = r2;
        r2 = ((object***) lenv) [1][1][4];
        r1[4] = r2;
        r2 = ((object****) lenv) [1][1][1][11];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2100; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_lexical_address_2658_i4643_i4727;
        L2100:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = ((object*) lenv) [5];
        switch (get_case_noint (r1)) {
          case (UOTAG(1)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][5];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object**) lenv) [1][5];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            // def lambda_17669_i1816_i4643_i4727
            goto L2101;
            FUN_lambda_17669_i1816_i4643_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][4];
              r1 = ((object**) lenv) [1][3];
              r2 = ((object*) lenv) [2];
              r3 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(9),4);
              t[1] = r3;
              t[2] = r1;
              t[3] = r2;
              t[4] = r0;
              r0 = t;
              PXLL_RETURN(0);
            L2101:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_lambda_17669_i1816_i4643_i4727; r1[2] = lenv;
            lenv = lenv[1];
          } break;
          case (UOTAG(6)): {
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][5];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            // def lambda_17683_i1816_i4643_i4727
            goto L2102;
            FUN_lambda_17683_i1816_i4643_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][4];
              r1 = ((object**) lenv) [1][2];
              r2 = ((object*) lenv) [2];
              r3 = (object *) -1;
              t = alloc_no_clear (UOTAG(9),4);
              t[1] = r3;
              t[2] = r1;
              t[3] = r2;
              t[4] = r0;
              r0 = t;
              PXLL_RETURN(0);
            L2102:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_lambda_17683_i1816_i4643_i4727; r1[2] = lenv;
            lenv = lenv[1];
          } break;
          default: {
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][5];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            // def lambda_17695_i1816_i4643_i4727
            goto L2103;
            FUN_lambda_17695_i1816_i4643_i4727:
              check_heap (0);
              r0 = ((object**) lenv) [1][2];
              r1 = ((object***) lenv) [1][1][4];
              r2 = ((object*) lenv) [2];
              t = alloc_no_clear (UOTAG(19),3);
              t[1] = r2;
              t[2] = r0;
              t[3] = r1;
              r0 = t;
              PXLL_RETURN(0);
            L2103:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_lambda_17695_i1816_i4643_i4727; r1[2] = lenv;
            lenv = lenv[1];
          } break;
        }
        r0[6] = r1;
        r0 = allocate (TC_TUPLE, 5);
        r1 = (object *) 6;
        r0[2] = r1;
        r1 = ((object*) lenv) [3];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object*) lenv) [4];
        r2[2] = r3;
        r3 = top[136];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2104; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_k_2ffree_522;
        L2104:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = ((object*) lenv) [6];
        r1[3] = r2;
        r2 = ((object****) lenv) [1][1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2105; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
        L2105:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = ((object****) lenv) [1][1][1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[5] = r1;
        r0 = allocate (TC_TUPLE, 5);
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[3] = r1;
        r1 = ((object*) lenv) [5];
        r0[4] = r1;
        // def lambda_17967_i1820_i4643_i4727
        goto L2106;
        FUN_lambda_17967_i1820_i4643_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*****) lenv) [1][1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[2] = r1;
          r0 = ((object**) lenv) [1][2];
          r1 = ((object*) lenv) [2];
          r2 = ((object***) lenv) [1][1][4];
          r3 = ((object***) lenv) [1][1][5];
          r4 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (UOTAG(3),5);
          t[1] = r4;
          t[2] = r1;
          t[3] = r2;
          t[4] = r0;
          t[5] = r3;
          r0 = t;
          PXLL_RETURN(0);
        L2106:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_17967_i1820_i4643_i4727; r1[2] = lenv;
        r0[5] = r1;
        r1 = ((object****) lenv) [1][1][1][12];
        r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_2663_i4643_i4727;
      } break;
      case (UITAG(8)): {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object***) lenv) [1][1][5];
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_PAIR): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r0 = ((object****) lenv) [1][1][1][2];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = ((pxll_vector*)r1)->val[4];
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (UOTAG(0)): {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
                  r1 = ((pxll_vector*)r1)->val[0];
                  r0[2] = r1;
                  r0 = top[193];
                  r1 = ((object*) lenv) [2];
                  r0 = PXLL_TEST((unbox(r1)&(1<<unbox(r0)))>0);
                  lenv = lenv[1];
                  if PXLL_IS_TRUE(r0) {
                    r0 = allocate (TC_TUPLE, 5);
                    r0[1] = lenv; lenv = r0;
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][49];
                    r1[2] = r2;
                    r2 = top[19];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L2107; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
                    L2107:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = allocate (TC_TUPLE, 3);
                    r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                    r1[2] = r2;
                    r2 = ((object*) lenv) [2];
                    r1[3] = r2;
                    r2 = top[5];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L2108; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
                    L2108:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[3] = r1;
                    r1 = ((object*) lenv) [3];
                    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,72)];
                    r0[4] = r1;
                    r1 = top[154];
                    r2 = ((object*) lenv) [4];
                    r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
                    r0[5] = r1;
                    r0 = (object *) 6;
                    r1 = ((object*) lenv) [5];
                    r0 = PXLL_TEST(r1==r0);
                    lenv = lenv[1];
                  } else {
                    r0 = (object *) 6;
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_60;
                }
              }
              goto fatbar_60_over;
              fatbar_60:
              r0 = (object *) 6;
              fatbar_60_over: ;
              lenv = lenv[1];
            } else {
              r0 = (object *) 6;
            }
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object**) lenv) [1][2];
              r2 = ((pxll_vector*)r2)->val[4];
              r1[2] = r2;
              r2 = top[119];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2109; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_varref__3ename_489;
              L2109:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 4);
              r2 = ((object*) lenv) [2];
              r1[2] = r2;
              r2 = (object *) 1;
              r1[3] = r2;
              r2 = ((object*****) lenv) [1][1][1][1][4];
              r1[4] = r2;
              r2 = ((object******) lenv) [1][1][1][1][1][11];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2110; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_lexical_address_2658_i4643_i4727;
              L2110:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              switch (get_case_noint (r0)) {
                case (UOTAG(7)): {
                  r0 = (object*) &constructed_68;
                  r0 = box((pxll_int)((pxll_string*)(r0))->len);
                  r1 = (object*) &constructed_68;
                  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                  r0 = (object*) &constructed_478;
                  dump_object (r0, 0); fprintf (stdout, "\n");
                  r0 = (object*) &constructed_478;
                  PXLL_UNDEFINED; result=r0; goto Lreturn;
                  PXLL_RETURN(0);
                } break;
                case (UOTAG(1)): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = allocate (TC_TUPLE, 5);
                  r1 = ((object***) lenv) [1][1][3];
                  r0[2] = r1;
                  r1 = ((object******) lenv) [1][1][1][1][1][4];
                  r0[3] = r1;
                  r1 = ((object****) lenv) [1][1][1][2];
                  r0[4] = r1;
                  // def lambda_18221_i4643_i4727_i8
                  goto L2111;
                  FUN_lambda_18221_i4643_i4727_i8:
                    check_heap (0);
                    r0 = ((object***) lenv) [1][1][2];
                    r1 = ((object*) lenv) [2];
                    r2 = ((object**) lenv) [1][2];
                    t = alloc_no_clear (UOTAG(15),3);
                    t[1] = r2;
                    t[2] = r0;
                    t[3] = r1;
                    r0 = t;
                    PXLL_RETURN(0);
                  L2111:
                  r1 = allocate (TC_CLOSURE, 2);
                  r1[1] = &&FUN_lambda_18221_i4643_i4727_i8; r1[2] = lenv;
                  r0[5] = r1;
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][12];
                  r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_2663_i4643_i4727;
                } break;
                default: {
                  r0 = allocate (TC_TUPLE, 2);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][3];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r0 = allocate (TC_TUPLE, 5);
                  r1 = ((object***) lenv) [1][1][3];
                  r0[2] = r1;
                  r1 = ((object******) lenv) [1][1][1][1][1][4];
                  r0[3] = r1;
                  r1 = ((object****) lenv) [1][1][1][2];
                  r0[4] = r1;
                  // def lambda_18221_i4643_i4727_i9
                  goto L2112;
                  FUN_lambda_18221_i4643_i4727_i9:
                    check_heap (0);
                    r0 = ((object***) lenv) [1][1][2];
                    r1 = ((object*) lenv) [2];
                    r2 = (object *) -1;
                    t = alloc_no_clear (UOTAG(15),3);
                    t[1] = r2;
                    t[2] = r0;
                    t[3] = r1;
                    r0 = t;
                    PXLL_RETURN(0);
                  L2112:
                  r1 = allocate (TC_CLOSURE, 2);
                  r1[1] = &&FUN_lambda_18221_i4643_i4727_i9; r1[2] = lenv;
                  r0[5] = r1;
                  r1 = ((object*******) lenv) [1][1][1][1][1][1][12];
                  r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_2663_i4643_i4727;
                } break;
              }
              PXLL_RETURN(0);
            } else {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = ((object*****) lenv) [1][1][1][1][2];
              if PXLL_IS_TRUE(r1) {
                r1 = ((object******) lenv) [1][1][1][1][1][21];
              } else {
                r1 = ((object******) lenv) [1][1][1][1][1][20];
              }
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = ((pxll_vector*)r1)->val[4];
              r0[3] = r1;
              r1 = ((object*) lenv) [3];
              switch (get_case (r1)) {
                case (UOTAG(0)): {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object**) lenv) [1][3];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = allocate (TC_TUPLE, 3);
                  r3 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                  r2[2] = r3;
                  r3 = ((object**) lenv) [1][2];
                  r2[3] = r3;
                  r3 = top[5];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2113; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_vars_get_var_246;
                  L2113:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,72)];
                  r1[3] = r2;
                  r1 = top[155];
                  r2 = ((object*) lenv) [3];
                  r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
                  lenv = lenv[1];
                  if PXLL_IS_TRUE(r1) {
                    r1 = ((object*) lenv) [2];
                    t = alloc_no_clear (UOTAG(0),1);
                    t[1] = r1;
                    r1 = t;
                  } else {
                    r1 = (object*)UITAG(1);
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_59;
                }
              }
              goto fatbar_59_over;
              fatbar_59:
              r1 = (object*)UITAG(1);
              fatbar_59_over: ;
              r0[4] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              // def make-call_2922_i1822_i4643_i4727
              goto L2114;
              FUN_make_call_2922_i1822_i4643_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 5);
                r1 = (object *) 6;
                r0[2] = r1;
                r1 = ((object****) lenv) [1][1][1][2];
                r0[3] = r1;
                r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
                r0[4] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = allocate (TC_TUPLE, 2);
                r4 = ((object******) lenv) [1][1][1][1][1][2];
                r3[2] = r4;
                r4 = top[136];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2115; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_k_2ffree_522;
                L2115:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[2] = r3;
                r2 = ((object*) lenv) [2];
                r3 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[2] = r2;
                // def lambda_18334_i1822_i4643_i4727
                goto L2116;
                FUN_lambda_18334_i1822_i4643_i4727:
                  check_heap (0);
                  r0 = allocate (TC_TUPLE, 5);
                  r1 = ((object****) lenv) [1][1][1][4];
                  r0[2] = r1;
                  r1 = ((object*) lenv) [2];
                  r0[3] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r0[4] = r1;
                  r1 = ((object******) lenv) [1][1][1][1][1][2];
                  r0[5] = r1;
                  r1 = ((object****) lenv) [1][1][1][2];
                  r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L2116:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_lambda_18334_i1822_i4643_i4727; r2[2] = lenv;
                r1[3] = r2;
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2117; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
                L2117:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[5] = r1;
                r1 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
              L2114:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_make_call_2922_i1822_i4643_i4727; r1[2] = lenv;
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object****) lenv) [1][1][1][3];
              r1[2] = r2;
              r2 = top[22];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2118; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
              L2118:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = (object *) 1;
              r1 = ((object*) lenv) [2];
              r0 = PXLL_TEST(unbox(r1)>unbox(r0));
              lenv = lenv[1];
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object*****) lenv) [1][1][1][1][2];
                r2[2] = r3;
                r3 = top[136];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2119; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_k_2ffree_522;
                L2119:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2120; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
                L2120:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 4);
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][49];
                r1[2] = r2;
                r2 = top[19];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2121; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
                L2121:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = top[156];
                r0[4] = r1;
                r1 = top[6];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2122; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
                L2122:
                ; lenv = k[2]; k = k[1];
                r0 = ((object****) lenv) [1][1][1][3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = ((object*) lenv) [2];
                    r1 = (object *) 1;
                    t = alloc_no_clear (UOTAG(10),2);
                    t[1] = r1;
                    t[2] = r0;
                    r0 = t;
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    goto fatbar_58;
                  }
                }
                PXLL_RETURN(0);
                goto fatbar_58_over;
                fatbar_58:
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object*****) lenv) [1][1][1][1][3];
                r1[2] = r2;
                r2 = top[22];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2123; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                L2123:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object**) lenv) [1][2];
                r2[2] = r3;
                r3 = top[136];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2124; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_k_2ffree_522;
                L2124:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                // def lambda_18411_i1834_i4643_i4727
                goto L2125;
                FUN_lambda_18411_i1834_i4643_i4727:
                  check_heap (0);
                  r0 = allocate (TC_TUPLE, 8);
                  r1 = (object *) 1;
                  r0[2] = r1;
                  r1 = (object *) 3;
                  r0[3] = r1;
                  r1 = ((object******) lenv) [1][1][1][1][1][3];
                  r0[4] = r1;
                  r1 = ((object*) lenv) [2];
                  r0[5] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object****) lenv) [1][1][1][2];
                  r2[2] = r3;
                  r3 = top[136];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2126; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_k_2ffree_522;
                  L2126:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r1 = ((object*) lenv) [2];
                  r2 = ((object**) lenv) [1][2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r2;
                  t[2] = r1;
                  r1 = t;
                  lenv = lenv[1];
                  r0[6] = r1;
                  r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                  r0[7] = r1;
                  r1 = ((object***) lenv) [1][1][2];
                  r0[8] = r1;
                  r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][16];
                  r0[1] = r1[2]; lenv = r0; goto FUN_compile_store_args_2671_i4643_i4727;
                L2125:
                r2 = allocate (TC_CLOSURE, 2);
                r2[1] = &&FUN_lambda_18411_i1834_i4643_i4727; r2[2] = lenv;
                r1[3] = r2;
                r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2127; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
                L2127:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r0 = ((object*) lenv) [3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(10),2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                PXLL_RETURN(0);
                fatbar_58_over: ;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r1 = (object *) -1;
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r0[1] = r1[2]; lenv = r0; goto FUN_make_call_2922_i1822_i4643_i4727;
              }
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = (object*) &constructed_68;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
            r0 = (object*) &constructed_479;
            dump_object (r0, 0); fprintf (stdout, "\n");
            r0 = (object*) &constructed_479;
            PXLL_UNDEFINED; result=r0; goto Lreturn;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(12)): {
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[5] = r1;
        r0 = (object *) constructed_230[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][5];
          r1[2] = r2;
          r2 = (object *) 5;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2128; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2128:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][5];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2129; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2129:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = ((object*) lenv) [4];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[5] = r1;
          r0 = ((object*) lenv) [5];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][5];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][5];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = (object *) constructed_481[0];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_PAIR): {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = ((object*) lenv) [3];
                    switch (get_case (r0)) {
                      case (TC_NIL): {
                        r0 = allocate (TC_TUPLE, 4);
                        r0[1] = lenv; lenv = r0;
                        r1 = (object*)TC_NIL;
                        r0[2] = r1;
                        r1 = ((object*) lenv) [2];
                        r2 = ((object****) lenv) [1][1][1][3];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r2;
                        t[2] = r1;
                        r1 = t;
                        r0[3] = r1;
                        r1 = allocate (TC_TUPLE, 3);
                        r2 = (object *) constructed_97[0];
                        r1[2] = r2;
                        r2 = ((object*) lenv) [3];
                        r3 = ((object**) lenv) [1][2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r3;
                        t[2] = r2;
                        r2 = t;
                        r1[3] = r2;
                        r2 = top[104];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2130; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_pred_451;
                        L2130:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[4] = r1;
                        r0 = allocate (TC_TUPLE, 7);
                        r1 = ((object*****) lenv) [1][1][1][1][5];
                        r0[2] = r1;
                        r1 = ((object*****) lenv) [1][1][1][1][2];
                        r0[3] = r1;
                        r1 = ((object*****) lenv) [1][1][1][1][3];
                        r0[4] = r1;
                        r1 = ((object*) lenv) [4];
                        r0[5] = r1;
                        r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
                        r0[6] = r1;
                        r1 = ((object*****) lenv) [1][1][1][1][4];
                        r0[7] = r1;
                        r1 = ((object********) lenv) [1][1][1][1][1][1][1][14];
                        r0[1] = r1[2]; lenv = r0; goto FUN_c_primargs_2665_i4643_i4727;
                      } break;
                      default: {
                        lenv = ((object **)lenv)[1][1];
                        goto fatbar_62;
                      }
                    }
                    PXLL_RETURN(0);
                  } break;
                  default: {
                    lenv = ((object *)lenv)[1];
                    goto fatbar_62;
                  }
                }
                PXLL_RETURN(0);
              } else {
                lenv = ((object *)lenv)[1];
                goto fatbar_62;
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_62;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_62_over;
          fatbar_62:
          r0 = (object*) &constructed_68;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_67;
          dump_object (r0, 0); fprintf (stdout, "\n");
          r0 = (object*) &constructed_67;
          PXLL_UNDEFINED; result=r0; goto Lreturn;
          PXLL_RETURN(0);
          fatbar_62_over: ;
          PXLL_RETURN(0);
        } else {
          goto fatbar_61;
        }
        PXLL_RETURN(0);
        goto fatbar_61_over;
        fatbar_61:
        r0 = (object *) constructed_483[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 6);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][5];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2131; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2131:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = top[112];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2132; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_get_record_sig_469;
          L2132:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = top[3];
          r1[2] = r2;
          r2 = ((object*) lenv) [4];
          r1[3] = r2;
          r2 = top[32];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2133; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
          L2133:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = ((object*) lenv) [5];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          r0[6] = r1;
          r0 = allocate (TC_TUPLE, 7);
          r1 = ((object**) lenv) [1][5];
          r0[2] = r1;
          r1 = (object *) constructed_266[0];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = ((object**) lenv) [1][6];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[4] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[5] = r1;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[6] = r1;
          r1 = ((object**) lenv) [1][4];
          r0[7] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][14];
          r0[1] = r1[2]; lenv = r0; goto FUN_c_primargs_2665_i4643_i4727;
        } else {
          goto fatbar_63;
        }
        PXLL_RETURN(0);
        goto fatbar_63_over;
        fatbar_63:
        r0 = (object *) constructed_485[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 6);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][5];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2134; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2134:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = top[112];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2135; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_get_record_sig_469;
          L2135:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = top[3];
          r1[2] = r2;
          r2 = ((object*) lenv) [4];
          r1[3] = r2;
          r2 = top[32];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2136; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
          L2136:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = ((object*) lenv) [5];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          r0[6] = r1;
          r0 = allocate (TC_TUPLE, 7);
          r1 = ((object**) lenv) [1][5];
          r0[2] = r1;
          r1 = (object *) constructed_277[0];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 4);
          r1[1] = lenv; lenv = r1;
          r2 = (object*)TC_NIL;
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r3 = ((object**) lenv) [1][6];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r3 = ((object***) lenv) [1][1][3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          r1[4] = r2;
          r1 = ((object*) lenv) [4];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r1;
          r1 = t;
          lenv = lenv[1];
          r0[4] = r1;
          r1 = ((object****) lenv) [1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[5] = r1;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[6] = r1;
          r1 = ((object**) lenv) [1][4];
          r0[7] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][14];
          r0[1] = r1[2]; lenv = r0; goto FUN_c_primargs_2665_i4643_i4727;
        } else {
          goto fatbar_64;
        }
        PXLL_RETURN(0);
        goto fatbar_64_over;
        fatbar_64:
        r0 = (object *) constructed_491[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def loop_3038_i1860_i4643_i4727
          goto L2137;
          FUN_loop_3038_i1860_i4643_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            switch (get_case (r0)) {
              case (UOTAG(12)): {
                r0 = allocate (TC_TUPLE, 3);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r1 = UOBJ_GET(r1,1);
                r0[3] = r1;
                r0 = (object *) constructed_488[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = allocate (TC_TUPLE, 9);
                  r0[1] = lenv; lenv = r0;
                  r1 = allocate (TC_TUPLE, 3);
                  // def lambda_19071_i1860_i4643_i4727
                  goto L2138;
                  FUN_lambda_19071_i1860_i4643_i4727:
                    check_heap (0);
                    r0 = ((object*) lenv) [2];
                    switch (get_case_noint (r0)) {
                      default: {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r0 = ((object**) lenv) [1][3];
                        switch (get_case_noint (r0)) {
                          default: {
                            r0 = allocate (TC_TUPLE, 5);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object***) lenv) [1][1][3];
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = ((object**) lenv) [1][2];
                            r1 = allocate (TC_TUPLE, 2);
                            r1[1] = lenv; lenv = r1;
                            r2 = ((object***) lenv) [1][1][2];
                            r2 = UOBJ_GET(r2,0);
                            r1[2] = r2;
                            r1 = ((object*) lenv) [2];
                            lenv = lenv[1];
                            r0[3] = r1;
                            r1 = ((object*) lenv) [2];
                            r1 = allocate (TC_TUPLE, 2);
                            r1[1] = lenv; lenv = r1;
                            r2 = ((object**) lenv) [1][2];
                            r2 = UOBJ_GET(r2,0);
                            r1[2] = r2;
                            r1 = ((object*) lenv) [2];
                            lenv = lenv[1];
                            r0[4] = r1;
                            r1 = allocate (TC_TUPLE, 3);
                            r2 = ((object*) lenv) [3];
                            r1[2] = r2;
                            r2 = ((object*) lenv) [4];
                            r1[3] = r2;
                            r2 = top[50];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L2139; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
                            L2139:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[5] = r1;
                            r0 = (object *) 1;
                            r1 = ((object*) lenv) [5];
                            r0 = PXLL_TEST(unbox(r1)<unbox(r0));
                            PXLL_RETURN(0);
                          } break;
                        }
                        PXLL_RETURN(0);
                      } break;
                    }
                    PXLL_RETURN(0);
                  L2138:
                  r2 = allocate (TC_CLOSURE, 2);
                  r2[1] = &&FUN_lambda_19071_i1860_i4643_i4727; r2[2] = lenv;
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r1[3] = r2;
                  r2 = top[44];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2140; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_sort_340;
                  L2140:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[2] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[17];
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2141; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L2141:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[3] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = top[18];
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  r1[3] = r2;
                  r2 = top[32];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2142; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
                  L2142:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[4] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  // def loop_3075_i1863_i4643_i4727
                  goto L2143;
                  FUN_loop_3075_i1863_i4643_i4727:
                    check_heap (0);
                    r0 = ((object*) lenv) [2];
                    switch (get_case (r0)) {
                      case (TC_PAIR): {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,0);
                        r0[2] = r1;
                        r1 = ((object**) lenv) [1][2];
                        r1 = UOBJ_GET(r1,1);
                        r0[3] = r1;
                        r0 = ((object*) lenv) [2];
                        switch (get_case_noint (r0)) {
                          default: {
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = ((object**) lenv) [1][2];
                            r1 = UOBJ_GET(r1,0);
                            r0[2] = r1;
                            r1 = ((object**) lenv) [1][2];
                            r1 = UOBJ_GET(r1,1);
                            r0[3] = r1;
                            r0 = allocate (TC_TUPLE, 3);
                            r0[1] = lenv; lenv = r0;
                            r1 = allocate (TC_TUPLE, 2);
                            r2 = ((object**) lenv) [1][2];
                            r1[2] = r2;
                            r2 = top[22];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L2144; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                            L2144:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r1 = allocate (TC_TUPLE, 2);
                            r2 = ((object******) lenv) [1][1][1][1][1][3];
                            r1[2] = r2;
                            r2 = top[22];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L2145; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                            L2145:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[3] = r1;
                            r0 = ((object*) lenv) [3];
                            r1 = ((object*) lenv) [2];
                            r0 = PXLL_TEST(unbox(r1)==unbox(r0));
                            lenv = lenv[1];
                            if PXLL_IS_TRUE(r0) {
                              r0 = allocate (TC_TUPLE, 4);
                              r1 = top[14];
                              r0[2] = r1;
                              r1 = ((object*) lenv) [2];
                              r0[3] = r1;
                              r1 = ((object*****) lenv) [1][1][1][1][3];
                              r0[4] = r1;
                              r1 = top[41];
                              t = allocate (TC_SAVE, 3);
                              t[1] = k; t[2] = lenv; t[3] = &&L2146; ; k = t;
                              r0[1] = r1[2]; lenv = r0; goto FUN_every2_question_335;
                              L2146:
                              ; lenv = k[2]; k = k[1];
                              r0 = result;
                            } else {
                              r0 = (object *) 6;
                            }
                            if PXLL_IS_TRUE(r0) {
                              r0 = ((object*) lenv) [3];
                              PXLL_RETURN(0);
                            } else {
                              r0 = ((object**) lenv) [1][3];
                              lenv = ((object **)lenv)[1][1];
                              lenv[2] = r0;
                              goto FUN_loop_3075_i1863_i4643_i4727;
                            }
                            PXLL_RETURN(0);
                          } break;
                        }
                        PXLL_RETURN(0);
                      } break;
                      default: {
                        r0 = allocate (TC_TUPLE, 2);
                        r0[1] = lenv; lenv = r0;
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][2];
                        r2 = ((pxll_vector*)r2)->val[10];
                        r1[2] = r2;
                        r2 = top[22];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2147; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                        L2147:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r0 = allocate (TC_TUPLE, 3);
                        r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][18];
                        r0[2] = r1;
                        r1 = ((object****) lenv) [1][1][1][3];
                        r0[3] = r1;
                        r1 = top[35];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L2148; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
                        L2148:
                        ; lenv = k[2]; k = k[1];
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = allocate (UOTAG(1), 2);
                        r2 = ((object*****) lenv) [1][1][1][1][3];
                        r1[1] = r2;
                        r2 = ((object**) lenv) [1][2];
                        r1[2] = r2;
                        r0[2] = r1;
                        r1 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][2];
                        r1 = ((pxll_vector*)r1)->val[10];
                        r0[3] = r1;
                        r0 = ((object*) lenv) [3];
                        r1 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r1;
                        t[2] = r0;
                        r0 = t;
                        lenv = lenv[1];
                        r1 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][2];
                        ((pxll_vector*)r1)->val[10] = r0;
                        r0 = ((object*) lenv) [2];
                        PXLL_RETURN(0);
                      } break;
                    }
                    PXLL_RETURN(0);
                  L2143:
                  r2 = allocate (TC_CLOSURE, 2);
                  r2[1] = &&FUN_loop_3075_i1863_i4643_i4727; r2[2] = lenv;
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                  r2 = ((pxll_vector*)r2)->val[10];
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2149; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_loop_3075_i1863_i4643_i4727;
                  L2149:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  lenv = lenv[1];
                  r0[5] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object******) lenv) [1][1][1][1][1][4];
                  r1[2] = r2;
                  r2 = top[136];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2150; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
                  L2150:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[6] = r1;
                  r1 = ((object*) lenv) [5];
                  t = alloc_no_clear (UOTAG(1),1);
                  t[1] = r1;
                  r1 = t;
                  r0[7] = r1;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = ((object*) lenv) [4];
                  r1[2] = r2;
                  r2 = top[22];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2151; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                  L2151:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[8] = r1;
                  r1 = allocate (TC_TUPLE, 3);
                  r2 = ((object*) lenv) [6];
                  r1[2] = r2;
                  // def lambda_19110_i1860_i4643_i4727
                  goto L2152;
                  FUN_lambda_19110_i1860_i4643_i4727:
                    check_heap (0);
                    r0 = allocate (TC_TUPLE, 8);
                    r1 = (object *) 1;
                    r0[2] = r1;
                    r1 = (object *) 1;
                    r0[3] = r1;
                    r1 = ((object**) lenv) [1][4];
                    r0[4] = r1;
                    r1 = ((object*) lenv) [2];
                    r0[5] = r1;
                    r1 = ((object**) lenv) [1][6];
                    r2 = ((object*) lenv) [2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r2;
                    t[2] = r1;
                    r1 = t;
                    r0[6] = r1;
                    r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                    r0[7] = r1;
                    r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
                    r0[8] = r1;
                    r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][16];
                    r0[1] = r1[2]; lenv = r0; goto FUN_compile_store_args_2671_i4643_i4727;
                  L2152:
                  r2 = allocate (TC_CLOSURE, 2);
                  r2[1] = &&FUN_lambda_19110_i1860_i4643_i4727; r2[2] = lenv;
                  r1[3] = r2;
                  r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2153; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
                  L2153:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  r0[9] = r1;
                  r0 = ((object*) lenv) [8];
                  r1 = ((object*) lenv) [9];
                  r2 = ((object*) lenv) [7];
                  t = alloc_no_clear (UOTAG(11),3);
                  t[1] = r2;
                  t[2] = r0;
                  t[3] = r1;
                  r0 = t;
                  PXLL_RETURN(0);
                } else {
                  goto fatbar_67;
                }
                PXLL_RETURN(0);
                goto fatbar_67_over;
                fatbar_67:
                r0 = (object *) constructed_491[0];
                r1 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                  r0 = ((object*) lenv) [3];
                  switch (get_case (r0)) {
                    case (UOTAG(1)): {
                      r0 = allocate (TC_TUPLE, 3);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][3];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r1 = ((object****) lenv) [1][1][1][2];
                      r1 = ((pxll_vector*)r1)->val[3];
                      r0[3] = r1;
                      r0 = ((object*) lenv) [3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 3);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,0);
                          r0[2] = r1;
                          r1 = ((object**) lenv) [1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[3] = r1;
                          r0 = ((object*) lenv) [3];
                          switch (get_case (r0)) {
                            case (TC_PAIR): {
                              r0 = allocate (TC_TUPLE, 3);
                              r0[1] = lenv; lenv = r0;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,0);
                              r0[2] = r1;
                              r1 = ((object**) lenv) [1][3];
                              r1 = UOBJ_GET(r1,1);
                              r0[3] = r1;
                              r0 = ((object*) lenv) [3];
                              switch (get_case (r0)) {
                                case (TC_NIL): {
                                  r0 = allocate (TC_TUPLE, 2);
                                  r0[1] = lenv; lenv = r0;
                                  r1 = allocate (UOTAG(1), 2);
                                  r2 = ((object****) lenv) [1][1][1][2];
                                  r1[1] = r2;
                                  r2 = ((object**) lenv) [1][2];
                                  r1[2] = r2;
                                  r0[2] = r1;
                                  r0 = ((object*******) lenv) [1][1][1][1][1][1][3];
                                  r1 = ((object*) lenv) [2];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r1;
                                  t[2] = r0;
                                  r0 = t;
                                  lenv = lenv[1];
                                  r1 = ((object**) lenv) [1][2];
                                  lenv = ((object *****)lenv)[1][1][1][1][1];
                                  lenv[2] = r1;
                                  lenv[3] = r0;
                                  goto FUN_loop_3038_i1860_i4643_i4727;
                                } break;
                                default: {
                                  lenv = ((object **)lenv)[1][1];
                                  goto fatbar_68;
                                }
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object *)lenv)[1];
                              goto fatbar_68;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          goto fatbar_68;
                        }
                      }
                      PXLL_RETURN(0);
                      goto fatbar_68_over;
                      fatbar_68:
                      r0 = allocate (TC_TUPLE, 3);
                      r1 = (object*) &constructed_489;
                      r0[2] = r1;
                      r1 = ((object****) lenv) [1][1][1][2];
                      r0[3] = r1;
                      r1 = top[15];
                      r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
                      fatbar_68_over: ;
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_66;
                    }
                  }
                  PXLL_RETURN(0);
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_66;
                }
                PXLL_RETURN(0);
                fatbar_67_over: ;
                PXLL_RETURN(0);
              } break;
              default: {
                goto fatbar_66;
              }
            }
            PXLL_RETURN(0);
            goto fatbar_66_over;
            fatbar_66:
            r0 = (object*) &constructed_68;
            r0 = box((pxll_int)((pxll_string*)(r0))->len);
            r1 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
            r0 = (object*) &constructed_492;
            dump_object (r0, 0); fprintf (stdout, "\n");
            r0 = (object*) &constructed_492;
            PXLL_UNDEFINED; result=r0; goto Lreturn;
            PXLL_RETURN(0);
            fatbar_66_over: ;
            PXLL_RETURN(0);
          L2137:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_loop_3038_i1860_i4643_i4727; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = ((object****) lenv) [1][1][1][3];
          r0[2] = r1;
          r1 = (object *) constructed_486[0];
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          r0[1] = r1[2]; lenv = r0; goto FUN_loop_3038_i1860_i4643_i4727;
        } else {
          goto fatbar_65;
        }
        PXLL_RETURN(0);
        goto fatbar_65_over;
        fatbar_65:
        r0 = (object *) constructed_496[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object*) lenv) [3];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              switch (get_case (r0)) {
                case (TC_PAIR): {
                  r0 = allocate (TC_TUPLE, 3);
                  r0[1] = lenv; lenv = r0;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,0);
                  r0[2] = r1;
                  r1 = ((object**) lenv) [1][2];
                  r1 = UOBJ_GET(r1,1);
                  r0[3] = r1;
                  r0 = ((object*) lenv) [2];
                  switch (get_case (r0)) {
                    case (UOTAG(1)): {
                      r0 = allocate (TC_TUPLE, 2);
                      r0[1] = lenv; lenv = r0;
                      r1 = ((object**) lenv) [1][2];
                      r1 = UOBJ_GET(r1,0);
                      r0[2] = r1;
                      r0 = ((object**) lenv) [1][3];
                      switch (get_case (r0)) {
                        case (TC_PAIR): {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = ((object***) lenv) [1][1][3];
                          r1 = UOBJ_GET(r1,1);
                          r0[2] = r1;
                          r0 = ((object*) lenv) [2];
                          switch (get_case (r0)) {
                            case (TC_NIL): {
                              r0 = allocate (TC_TUPLE, 4);
                              r0[1] = lenv; lenv = r0;
                              r1 = allocate (TC_TUPLE, 4);
                              r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                              r2 = ((pxll_vector*)r2)->val[15];
                              r1[2] = r2;
                              r2 = ((object***) lenv) [1][1][2];
                              r1[3] = r2;
                              r2 = (object*) &constructed_493;
                              r1[4] = r2;
                              r2 = top[66];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2154; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_alist_2fget_396;
                              L2154:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[2] = r1;
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = ((object******) lenv) [1][1][1][1][1][4];
                              r1[2] = r2;
                              r2 = top[136];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2155; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
                              L2155:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[3] = r1;
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = ((object******) lenv) [1][1][1][1][1][5];
                              r1[2] = r2;
                              r2 = top[22];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2156; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                              L2156:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                              r0[4] = r1;
                              r0 = (object *) 1;
                              r1 = ((object*) lenv) [4];
                              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
                                r0 = allocate (TC_TUPLE, 4);
                                r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                                r0[2] = r1;
                                r1 = allocate (TC_TUPLE, 2);
                                r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][49];
                                r1[2] = r2;
                                r2 = top[19];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L2157; t[4] = r0; k = t;
                                r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
                                L2157:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                r0[3] = r1;
                                r1 = top[156];
                                r0[4] = r1;
                                r1 = top[6];
                                t = allocate (TC_SAVE, 3);
                                t[1] = k; t[2] = lenv; t[3] = &&L2158; ; k = t;
                                r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
                                L2158:
                                ; lenv = k[2]; k = k[1];
                                r0 = result;
                              } else {
                                r0 = (object *) 14;
                              }
                              r0 = (object *) 1;
                              r1 = ((object*) lenv) [4];
                              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
                                r0 = allocate (TC_TUPLE, 3);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object**) lenv) [1][2];
                                t = alloc_no_clear (UOTAG(1),1);
                                t[1] = r1;
                                r1 = t;
                                r0[2] = r1;
                                r1 = allocate (TC_TUPLE, 3);
                                r2 = ((object**) lenv) [1][3];
                                r1[2] = r2;
                                // def lambda_18998_i1875_i4643_i4727
                                goto L2159;
                                FUN_lambda_18998_i1875_i4643_i4727:
                                  check_heap (0);
                                  r0 = allocate (TC_TUPLE, 8);
                                  r1 = (object *) 1;
                                  r0[2] = r1;
                                  r1 = (object *) 1;
                                  r0[3] = r1;
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][5];
                                  r0[4] = r1;
                                  r1 = ((object*) lenv) [2];
                                  r0[5] = r1;
                                  r1 = ((object***) lenv) [1][1][3];
                                  r2 = ((object*) lenv) [2];
                                  t = alloc_no_clear (TC_PAIR,2);
                                  t[1] = r2;
                                  t[2] = r1;
                                  r1 = t;
                                  r0[6] = r1;
                                  r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
                                  r0[7] = r1;
                                  r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                                  r0[8] = r1;
                                  r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][16];
                                  r0[1] = r1[2]; lenv = r0; goto FUN_compile_store_args_2671_i4643_i4727;
                                L2159:
                                r2 = allocate (TC_CLOSURE, 2);
                                r2[1] = &&FUN_lambda_18998_i1875_i4643_i4727; r2[2] = lenv;
                                r1[3] = r2;
                                r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                                t = allocate (TC_SAVE, 4);
                                t[1] = k; t[2] = lenv; t[3] = &&L2160; t[4] = r0; k = t;
                                r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
                                L2160:
                                r0 = k[4]; lenv = k[2]; k = k[1];
                                r1 = result;
                                r0[3] = r1;
                                r0 = ((object**) lenv) [1][4];
                                r1 = ((object*) lenv) [3];
                                r2 = ((object*) lenv) [2];
                                t = alloc_no_clear (UOTAG(11),3);
                                t[1] = r2;
                                t[2] = r0;
                                t[3] = r1;
                                r0 = t;
                                PXLL_RETURN(0);
                              } else {
                                r0 = allocate (TC_TUPLE, 2);
                                r0[1] = lenv; lenv = r0;
                                r1 = ((object**) lenv) [1][2];
                                t = alloc_no_clear (UOTAG(1),1);
                                t[1] = r1;
                                r1 = t;
                                r0[2] = r1;
                                r0 = (object *) 1;
                                r1 = ((object*******) lenv) [1][1][1][1][1][1][4];
                                r2 = ((object*) lenv) [2];
                                t = alloc_no_clear (UOTAG(11),3);
                                t[1] = r2;
                                t[2] = r0;
                                t[3] = r1;
                                r0 = t;
                                PXLL_RETURN(0);
                              }
                              PXLL_RETURN(0);
                            } break;
                            default: {
                              lenv = ((object ****)lenv)[1][1][1][1];
                              goto fatbar_70;
                            }
                          }
                          PXLL_RETURN(0);
                        } break;
                        default: {
                          lenv = ((object ***)lenv)[1][1][1];
                          goto fatbar_70;
                        }
                      }
                      PXLL_RETURN(0);
                    } break;
                    default: {
                      lenv = ((object **)lenv)[1][1];
                      goto fatbar_70;
                    }
                  }
                  PXLL_RETURN(0);
                } break;
                default: {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_70;
                }
              }
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_70;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_70_over;
          fatbar_70:
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object*) &constructed_494;
          r0[2] = r1;
          r1 = ((object*) lenv) [3];
          r0[3] = r1;
          r1 = top[15];
          r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
          fatbar_70_over: ;
          PXLL_RETURN(0);
        } else {
          goto fatbar_69;
        }
        PXLL_RETURN(0);
        goto fatbar_69_over;
        fatbar_69:
        r0 = (object *) constructed_328[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][5];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2161; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L2161:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 1;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)>unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 4);
            r1 = ((object*****) lenv) [1][1][1][1][2];
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*****) lenv) [1][1][1][1][49];
            r1[2] = r2;
            r2 = top[19];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2162; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
            L2162:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = top[156];
            r0[4] = r1;
            r1 = top[6];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2163; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
            L2163:
            ; lenv = k[2]; k = k[1];
            r0 = result;
          } else {
            r0 = (object *) 14;
          }
          r0 = allocate (TC_TUPLE, 7);
          r1 = ((object*) lenv) [5];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object*) lenv) [3];
          r0[4] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[5] = r1;
          r1 = ((object***) lenv) [1][1][4];
          r0[6] = r1;
          r1 = ((object*) lenv) [4];
          r0[7] = r1;
          r1 = ((object****) lenv) [1][1][1][14];
          r0[1] = r1[2]; lenv = r0; goto FUN_c_primargs_2665_i4643_i4727;
        } else {
          goto fatbar_71;
        }
        PXLL_RETURN(0);
        goto fatbar_71_over;
        fatbar_71:
        r0 = (object *) constructed_498[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 7);
          r0[1] = lenv; lenv = r0;
          r1 = ((object*****) lenv) [1][1][1][1][22];
          r1 = ((pxll_vector*)r1)->val[1];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2164; t[4] = r0; k = t;
          lenv = r1[2]; goto *r1[1];
          L2164:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][4];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(3),2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][4];
          r1[2] = r2;
          r2 = top[136];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2165; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
          L2165:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][4];
          r1[2] = r2;
          r2 = top[137];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2166; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_k_2ftarget_523;
          L2166:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = allocate (TC_TUPLE, 5);
          r2 = ((object****) lenv) [1][1][1][2];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object**) lenv) [1][5];
          r2[2] = r3;
          r3 = (object *) 1;
          r2[3] = r3;
          r3 = top[27];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2167; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
          L2167:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = ((object*) lenv) [3];
          r1[4] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object*) lenv) [4];
          r2[2] = r3;
          // def lambda_18866_i1888_i4643_i4727
          goto L2168;
          FUN_lambda_18866_i1888_i4643_i4727:
            check_heap (0);
            r0 = ((object**) lenv) [1][5];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(6),2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            PXLL_RETURN(0);
          L2168:
          r3 = allocate (TC_CLOSURE, 2);
          r3[1] = &&FUN_lambda_18866_i1888_i4643_i4727; r3[2] = lenv;
          r2[3] = r3;
          r3 = ((object*****) lenv) [1][1][1][1][2];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2169; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
          L2169:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[5] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2170; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
          L2170:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[6] = r1;
          r1 = allocate (TC_TUPLE, 5);
          r2 = ((object****) lenv) [1][1][1][2];
          r1[2] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object**) lenv) [1][5];
          r2[2] = r3;
          r3 = (object *) 3;
          r2[3] = r3;
          r3 = top[27];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2171; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
          L2171:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[3] = r2;
          r2 = ((object****) lenv) [1][1][1][4];
          r1[4] = r2;
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object*) lenv) [4];
          r2[2] = r3;
          // def lambda_18883_i1888_i4643_i4727
          goto L2172;
          FUN_lambda_18883_i1888_i4643_i4727:
            check_heap (0);
            r0 = ((object**) lenv) [1][5];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (UOTAG(6),2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            PXLL_RETURN(0);
          L2172:
          r3 = allocate (TC_CLOSURE, 2);
          r3[1] = &&FUN_lambda_18883_i1888_i4643_i4727; r3[2] = lenv;
          r2[3] = r3;
          r3 = ((object*****) lenv) [1][1][1][1][2];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2173; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
          L2173:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[5] = r2;
          r2 = ((object*****) lenv) [1][1][1][1][3];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2174; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
          L2174:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[7] = r1;
          r0 = ((object**) lenv) [1][4];
          r1 = ((object*) lenv) [6];
          r2 = ((object*) lenv) [7];
          r3 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(20),4);
          t[1] = r3;
          t[2] = r1;
          t[3] = r2;
          t[4] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          goto fatbar_72;
        }
        PXLL_RETURN(0);
        goto fatbar_72_over;
        fatbar_72:
        r0 = (object *) constructed_501[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def loop_3016_i1894_i4643_i4727
          goto L2175;
          FUN_loop_3016_i1894_i4643_i4727:
            check_heap (0);
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (UITAG(0)): {
                r0 = (object*) &constructed_68;
                r0 = box((pxll_int)((pxll_string*)(r0))->len);
                r1 = (object*) &constructed_68;
                (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                r0 = (object*) &constructed_499;
                dump_object (r0, 0); fprintf (stdout, "\n");
                r0 = (object*) &constructed_499;
                PXLL_UNDEFINED; result=r0; goto Lreturn;
                PXLL_RETURN(0);
              } break;
              case (UOTAG(1)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,1);
                r0[2] = r1;
                r0 = (object *) 3;
                r1 = ((object**) lenv) [1][2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                r1 = ((object*) lenv) [2];
                lenv = ((object *)lenv)[1];
                lenv[2] = r0;
                lenv[3] = r1;
                goto FUN_loop_3016_i1894_i4643_i4727;
              } break;
              case (UOTAG(2)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,2);
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                lenv = ((object *)lenv)[1];
                lenv[2] = r1;
                lenv[3] = r0;
                goto FUN_loop_3016_i1894_i4643_i4727;
              } break;
              default: {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = ((object**) lenv) [1][2];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(21),2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
          L2175:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_loop_3016_i1894_i4643_i4727; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r1 = (object *) 1;
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][4];
          r0[3] = r1;
          r1 = ((object*) lenv) [2];
          r0[1] = r1[2]; lenv = r0; goto FUN_loop_3016_i1894_i4643_i4727;
        } else {
          goto fatbar_73;
        }
        PXLL_RETURN(0);
        goto fatbar_73_over;
        fatbar_73:
        r0 = allocate (TC_TUPLE, 7);
        r1 = ((object*) lenv) [5];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = ((object*) lenv) [3];
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[5];
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[6] = r1;
        r1 = ((object*) lenv) [4];
        r0[7] = r1;
        r1 = ((object****) lenv) [1][1][1][14];
        r0[1] = r1[2]; lenv = r0; goto FUN_c_primargs_2665_i4643_i4727;
        fatbar_73_over: ;
        PXLL_RETURN(0);
        fatbar_72_over: ;
        PXLL_RETURN(0);
        fatbar_71_over: ;
        PXLL_RETURN(0);
        fatbar_69_over: ;
        PXLL_RETURN(0);
        fatbar_65_over: ;
        PXLL_RETURN(0);
        fatbar_64_over: ;
        PXLL_RETURN(0);
        fatbar_63_over: ;
        PXLL_RETURN(0);
        fatbar_61_over: ;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(4)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = (object *) constructed_49[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (TC_TUPLE, 7);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][5];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = top[136];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2176; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
          L2176:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2177; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2177:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [2];
          r1[2] = r2;
          r2 = (object *) 3;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2178; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2178:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[6] = r1;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 2);
          r3 = ((object*) lenv) [2];
          r2[2] = r3;
          r3 = top[20];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2179; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_cdr_310;
          L2179:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r2 = top[20];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2180; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
          L2180:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[7] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def finish_2993_i1899_i4643_i4727
          goto L2181;
          FUN_finish_2993_i1899_i4643_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 6);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][4];
            r1[2] = r2;
            // def lambda_18760_i1899_i4643_i4727
            goto L2182;
            FUN_lambda_18760_i1899_i4643_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object******) lenv) [1][1][1][1][1][3];
              r1[2] = r2;
              r2 = top[137];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2183; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_k_2ftarget_523;
              L2183:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r1 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(6),2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              PXLL_RETURN(0);
            L2182:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_lambda_18760_i1899_i4643_i4727; r2[2] = lenv;
            r1[3] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2184; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
            L2184:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            // def lambda_18770_i1899_i4643_i4727
            goto L2185;
            FUN_lambda_18770_i1899_i4643_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 5);
              r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = ((object********) lenv) [1][1][1][1][1][1][1][4];
              r0[4] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[5] = r1;
              r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
              r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
            L2185:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_lambda_18770_i1899_i4643_i4727; r2[2] = lenv;
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][7];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2186; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L2186:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = ((object****) lenv) [1][1][1][6];
            r1 = ((pxll_vector*)r1)->val[4];
            r0[4] = r1;
            r1 = ((object*) lenv) [4];
            switch (get_case (r1)) {
              case (UOTAG(12)): {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object**) lenv) [1][4];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r1 = (object *) constructed_503[0];
                r2 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                  r1 = (object *) 6;
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_75;
                }
                lenv = lenv[1];
              } break;
              default: {
                goto fatbar_75;
              }
            }
            goto fatbar_75_over;
            fatbar_75:
            r1 = (object *) 262;
            fatbar_75_over: ;
            r0[5] = r1;
            r1 = ((object*) lenv) [5];
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 5);
              r3 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              r2[2] = r3;
              r3 = ((object*****) lenv) [1][1][1][1][6];
              r2[3] = r3;
              r3 = ((object********) lenv) [1][1][1][1][1][1][1][4];
              r2[4] = r3;
              r3 = ((object**) lenv) [1][2];
              r2[5] = r3;
              r3 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2187; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_compile_2646_i4643_i4727;
              L2187:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              t = alloc_no_clear (UOTAG(0),1);
              t[1] = r1;
              r1 = t;
              lenv = lenv[1];
            } else {
              r1 = (object*)UITAG(1);
            }
            r0[6] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object****) lenv) [1][1][1][3];
            r2 = ((object*****) lenv) [1][1][1][1][3];
            r3 = ((object*****) lenv) [1][1][1][1][4];
            r4 = ((object*) lenv) [6];
            r5 = ((object**) lenv) [1][2];
            t = alloc_no_clear (UOTAG(23),6);
            t[1] = r5;
            t[2] = r2;
            t[3] = r3;
            t[4] = r0;
            t[5] = r4;
            t[6] = r1;
            r0 = t;
            PXLL_RETURN(0);
          L2181:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_finish_2993_i1899_i4643_i4727; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 5);
          r1 = (object *) 6;
          r0[2] = r1;
          r1 = ((object**) lenv) [1][5];
          r0[3] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][4];
          r0[4] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][4];
          r1[2] = r2;
          r2 = ((object*) lenv) [2];
          r1[3] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][2];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2188; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
          L2188:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[5] = r1;
          r1 = ((object******) lenv) [1][1][1][1][1][3];
          r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
        } else {
          goto fatbar_74;
        }
        PXLL_RETURN(0);
        goto fatbar_74_over;
        fatbar_74:
        r0 = allocate (TC_TUPLE, 5);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[2] = r1;
        r1 = ((object****) lenv) [1][1][1][5];
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [3];
        r1[2] = r2;
        r2 = top[136];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2189; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
        L2189:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object******) lenv) [1][1][1][1][1][2];
        r2 = ((pxll_vector*)r2)->val[2];
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[3] = r2;
        r2 = top[65];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2190; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
        L2190:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r0 = ((object*) lenv) [5];
        switch (get_case (r0)) {
          case (UITAG(1)): {
            r0 = allocate (TC_TUPLE, 3);
            r1 = (object*) &constructed_504;
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r0[3] = r1;
            r1 = top[15];
            r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
          } break;
          default: {
            r0 = allocate (TC_TUPLE, 5);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][5];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = (object *) 1;
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2191; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L2191:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = (object *) 3;
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2192; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L2192:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = top[20];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2193; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_cdr_310;
            L2193:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = top[20];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2194; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
            L2194:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            // def finish_2974_i1905_i4643_i4727
            goto L2195;
            FUN_finish_2974_i1905_i4643_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object*****) lenv) [1][1][1][1][4];
              r1[2] = r2;
              // def lambda_18643_i1905_i4643_i4727
              goto L2196;
              FUN_lambda_18643_i1905_i4643_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object*******) lenv) [1][1][1][1][1][1][3];
                r1[2] = r2;
                r2 = top[137];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2197; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_k_2ftarget_523;
                L2197:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (UOTAG(6),2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                PXLL_RETURN(0);
              L2196:
              r2 = allocate (TC_CLOSURE, 2);
              r2[1] = &&FUN_lambda_18643_i1905_i4643_i4727; r2[2] = lenv;
              r1[3] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2198; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
              L2198:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 3);
              // def lambda_18653_i1905_i4643_i4727
              goto L2199;
              FUN_lambda_18653_i1905_i4643_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 5);
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                r0[2] = r1;
                r1 = ((object*) lenv) [2];
                r0[3] = r1;
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                r0[4] = r1;
                r1 = ((object**) lenv) [1][2];
                r0[5] = r1;
                r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][3];
                r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
              L2199:
              r2 = allocate (TC_CLOSURE, 2);
              r2[1] = &&FUN_lambda_18653_i1905_i4643_i4727; r2[2] = lenv;
              r1[2] = r2;
              r2 = ((object****) lenv) [1][1][1][5];
              r1[3] = r2;
              r2 = top[32];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2200; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
              L2200:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = allocate (TC_TUPLE, 4);
              r1[1] = lenv; lenv = r1;
              r2 = ((object*****) lenv) [1][1][1][1][2];
              r2 = ((pxll_vector*)r2)->val[3];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2201; t[4] = r0; t[5] = r1; k = t;
              lenv = r2[2]; goto *r2[1];
              L2201:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r3 = ((object**) lenv) [1][3];
              r2[2] = r3;
              r3 = top[22];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2202; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
              L2202:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r2 = ((object*) lenv) [3];
              r3 = ((object*) lenv) [2];
              r2 = PXLL_TEST(unbox(r3)==unbox(r2));
              r1[4] = r2;
              r1 = (object *) 6;
              r2 = ((object*) lenv) [4];
              r1 = PXLL_TEST(r2==r1);
              lenv = lenv[1];
              if PXLL_IS_TRUE(r1) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object*****) lenv) [1][1][1][1][4];
                r2 = ((pxll_vector*)r2)->val[4];
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                switch (get_case (r1)) {
                  case (UOTAG(12)): {
                    r1 = allocate (TC_TUPLE, 2);
                    r1[1] = lenv; lenv = r1;
                    r2 = ((object**) lenv) [1][2];
                    r2 = UOBJ_GET(r2,0);
                    r1[2] = r2;
                    r1 = (object *) constructed_503[0];
                    r2 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = (object*) &constructed_505;
                      r1[2] = r2;
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                      r1[3] = r2;
                      r2 = top[15];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2203; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                      L2203:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                    } else {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_76;
                    }
                    lenv = lenv[1];
                  } break;
                  default: {
                    goto fatbar_76;
                  }
                }
                goto fatbar_76_over;
                fatbar_76:
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 5);
                r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                r2[2] = r3;
                r3 = ((object******) lenv) [1][1][1][1][1][4];
                r2[3] = r3;
                r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][4];
                r2[4] = r3;
                r3 = ((object***) lenv) [1][1][2];
                r2[5] = r3;
                r3 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][3];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2204; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_compile_2646_i4643_i4727;
                L2204:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(0),1);
                t[1] = r1;
                r1 = t;
                lenv = lenv[1];
                fatbar_76_over: ;
                lenv = lenv[1];
              } else {
                r1 = (object*)UITAG(1);
              }
              r0[4] = r1;
              r0 = ((object*) lenv) [3];
              r1 = ((object*****) lenv) [1][1][1][1][3];
              r2 = ((object******) lenv) [1][1][1][1][1][2];
              r3 = ((object******) lenv) [1][1][1][1][1][3];
              r4 = ((object*) lenv) [4];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(22),6);
              t[1] = r5;
              t[2] = r2;
              t[3] = r3;
              t[4] = r0;
              t[5] = r4;
              t[6] = r1;
              r0 = t;
              PXLL_RETURN(0);
            L2195:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_finish_2974_i1905_i4643_i4727; r1[2] = lenv;
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 5);
            r1 = (object *) 6;
            r0[2] = r1;
            r1 = ((object**) lenv) [1][3];
            r0[3] = r1;
            r1 = ((object******) lenv) [1][1][1][1][1][4];
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][4];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2205; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
            L2205:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
            r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
          } break;
        }
        PXLL_RETURN(0);
        fatbar_74_over: ;
        PXLL_RETURN(0);
      } break;
      case (UOTAG(10)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 6);
        r1 = ((object***) lenv) [1][1][2];
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r1 = ((pxll_vector*)r1)->val[3];
        r0[4] = r1;
        r1 = ((object***) lenv) [1][1][4];
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][5];
        r0[6] = r1;
        r1 = ((object****) lenv) [1][1][1][17];
        r0[1] = r1[2]; lenv = r0; goto FUN_c_let_splat_2673_i4643_i4727;
      } break;
      case (UOTAG(9)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object****) lenv) [1][1][1][3];
        r1 = ((pxll_vector*)r1)->val[0];
        r0[2] = r1;
        r0 = top[194];
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST((unbox(r1)&(1<<unbox(r0)))>0);
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2206; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L2206:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = (object *) 11;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(unbox(r1)<unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            // def lambda_18174_i1913_i4643_i4727
            goto L2207;
            FUN_lambda_18174_i1913_i4643_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][2];
              r1[3] = r2;
              r2 = top[5];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2208; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_vars_get_var_246;
              L2208:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,72)];
              r0[3] = r1;
              r0 = top[154];
              r1 = ((object*) lenv) [3];
              r0 = PXLL_TEST((unbox(r1)&(1<<unbox(r0)))>0);
              PXLL_RETURN(0);
            L2207:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_lambda_18174_i1913_i4643_i4727; r2[2] = lenv;
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            r1[3] = r2;
            r2 = top[39];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2209; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_some_question_333;
            L2209:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = (object *) 6;
            r1 = ((object*) lenv) [2];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
          } else {
            r0 = (object *) 6;
          }
        } else {
          r0 = (object *) 6;
        }
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object****) lenv) [1][1][1][3];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[2] = r1;
          r1 = ((object****) lenv) [1][1][1][5];
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def loop_2943_i1918_i4643_i4727
          goto L2210;
          FUN_loop_2943_i1918_i4643_i4727:
            check_heap (0);
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2211; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L2211:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = (object *) 1;
            r0 = PXLL_TEST(unbox(r1)==unbox(r0));
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (TC_TUPLE, 5);
              r1 = ((object******) lenv) [1][1][1][1][1][2];
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object*) lenv) [3];
              r1[2] = r2;
              r2 = top[19];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2212; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
              L2212:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = ((object*) lenv) [4];
              r0[4] = r1;
              r1 = ((object***) lenv) [1][1][3];
              switch (get_case (r1)) {
                case (UOTAG(0)): {
                  r1 = allocate (TC_TUPLE, 5);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,2);
                  r1[4] = r2;
                  r2 = allocate (TC_TUPLE, 3);
                  r3 = allocate (TC_TUPLE, 3);
                  r4 = ((object*) lenv) [3];
                  r3[2] = r4;
                  r4 = (object *) constructed_36[0];
                  r3[3] = r4;
                  r4 = top[21];
                  t = allocate (TC_SAVE, 6);
                  t[1] = k; t[2] = lenv; t[3] = &&L2213; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                  r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
                  L2213:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                  r3 = result;
                  r2[2] = r3;
                  r3 = ((object**) lenv) [1][5];
                  r2[3] = r3;
                  r3 = top[21];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2214; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
                  L2214:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[5] = r2;
                  r1 = ((object*) lenv) [5];
                  r2 = ((object*) lenv) [4];
                  r3 = ((object*) lenv) [2];
                  t = alloc_no_clear (UOTAG(0),3);
                  t[1] = r3;
                  t[2] = r1;
                  t[3] = r2;
                  r1 = t;
                  lenv = lenv[1];
                } break;
                default: {
                  r1 = (object*) &constructed_68;
                  r1 = box((pxll_int)((pxll_string*)(r1))->len);
                  r2 = (object*) &constructed_68;
                  (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
                  r1 = (object*) &constructed_507;
                  dump_object (r1, 0); fprintf (stdout, "\n");
                  r1 = (object*) &constructed_507;
                  PXLL_UNDEFINED; result=r1; goto Lreturn;
                } break;
              }
              r0[5] = r1;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
              r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
            } else {
              r0 = allocate (TC_TUPLE, 5);
              r1 = (object *) 6;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object*) lenv) [3];
              r1[2] = r2;
              r2 = top[19];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2215; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
              L2215:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = ((object*) lenv) [4];
              r0[4] = r1;
              r1 = allocate (TC_TUPLE, 3);
              r2 = allocate (TC_TUPLE, 2);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object****) lenv) [1][1][1][3];
              r3[2] = r4;
              r4 = top[136];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2216; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_k_2ffree_522;
              L2216:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r2 = allocate (TC_TUPLE, 3);
              r3 = allocate (TC_TUPLE, 3);
              r4 = ((object**) lenv) [1][5];
              r3[2] = r4;
              r4 = (object *) constructed_36[0];
              r3[3] = r4;
              r4 = top[21];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2217; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
              L2217:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = ((object*) lenv) [2];
              r2[3] = r3;
              r3 = top[21];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2218; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
              L2218:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              lenv = lenv[1];
              r1[2] = r2;
              // def lambda_18514_i1918_i4643_i4727
              goto L2219;
              FUN_lambda_18514_i1918_i4643_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 5);
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object**) lenv) [1][2];
                r1[2] = r2;
                r2 = top[20];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2220; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
                L2220:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = ((object**) lenv) [1][3];
                r1[2] = r2;
                r2 = top[20];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2221; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
                L2221:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object***) lenv) [1][1][2];
                r2[2] = r3;
                r3 = top[19];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2222; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_car_309;
                L2222:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = ((object**) lenv) [1][2];
                r2 = ((object***) lenv) [1][1][4];
                r3 = ((object*) lenv) [2];
                t = alloc_no_clear (UOTAG(2),3);
                t[1] = r3;
                t[2] = r1;
                t[3] = r2;
                r1 = t;
                lenv = lenv[1];
                r0[4] = r1;
                r1 = ((object**) lenv) [1][5];
                r2 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                r0[5] = r1;
                r1 = ((object***) lenv) [1][1][2];
                r0[1] = r1[2]; lenv = r0; goto FUN_loop_2943_i1918_i4643_i4727;
              L2219:
              r2 = allocate (TC_CLOSURE, 2);
              r2[1] = &&FUN_lambda_18514_i1918_i4643_i4727; r2[2] = lenv;
              r1[3] = r2;
              r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2223; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
              L2223:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[5] = r1;
              r1 = ((object*******) lenv) [1][1][1][1][1][1][3];
              r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
            }
            PXLL_RETURN(0);
          L2210:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN_loop_2943_i1918_i4643_i4727; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 5);
          r1 = ((object***) lenv) [1][1][2];
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r0[3] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][4];
          r0[4] = r1;
          r1 = (object *) constructed_506[0];
          r0[5] = r1;
          r1 = ((object*) lenv) [2];
          r0[1] = r1[2]; lenv = r0; goto FUN_loop_2943_i1918_i4643_i4727;
        } else {
          r0 = allocate (TC_TUPLE, 6);
          r1 = ((object***) lenv) [1][1][2];
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r1 = ((pxll_vector*)r1)->val[3];
          r0[4] = r1;
          r1 = ((object***) lenv) [1][1][4];
          r0[5] = r1;
          r1 = ((object***) lenv) [1][1][5];
          r0[6] = r1;
          r1 = ((object****) lenv) [1][1][1][17];
          r0[1] = r1[2]; lenv = r0; goto FUN_c_let_splat_2673_i4643_i4727;
        }
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_67;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_67;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2065:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_compile_2646_i4643_i4727; r2[2] = lenv;
  r1[3] = r2;
  // def add-literal_2647_i4643_i4727
  goto L2224;
  FUN_add_literal_2647_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object****) lenv) [1][1][1][2];
    r2 = ((pxll_vector*)r2)->val[7];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2225; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L2225:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object*****) lenv) [1][1][1][1][2];
    r1 = ((pxll_vector*)r1)->val[7];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object***) lenv) [1][1][2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    lenv = lenv[1];
    r1 = ((object****) lenv) [1][1][1][2];
    ((pxll_vector*)r1)->val[7] = r0;
    r0 = ((object*) lenv) [2];
    PXLL_RETURN(0);
  L2224:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_add_literal_2647_i4643_i4727; r2[2] = lenv;
  r1[4] = r2;
  // def get-literal-index_2648_i4643_i4727
  goto L2226;
  FUN_get_literal_index_2648_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 4);
    r2 = ((object****) lenv) [1][1][1][2];
    r2 = ((pxll_vector*)r2)->val[6];
    r1[2] = r2;
    r2 = top[12];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    r1[4] = r2;
    r2 = top[73];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2227; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_tree_2fmember_409;
    L2227:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2228; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_add_literal_2647_i4643_i4727;
        L2228:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 5);
        r1 = ((object*****) lenv) [1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[6];
        r0[2] = r1;
        r1 = top[12];
        r0[3] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[4] = r1;
        r1 = ((object*) lenv) [2];
        r0[5] = r1;
        r1 = top[72];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L2229; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_tree_2finsert_407;
        L2229:
        ; lenv = k[2]; k = k[1];
        r0 = result;
        r1 = ((object*****) lenv) [1][1][1][1][2];
        ((pxll_vector*)r1)->val[6] = r0;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2226:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_literal_index_2648_i4643_i4727; r2[2] = lenv;
  r1[5] = r2;
  // def get-symbol-index_2649_i4643_i4727
  goto L2230;
  FUN_get_symbol_index_2649_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object****) lenv) [1][1][1][2];
    r2 = ((pxll_vector*)r2)->val[14];
    r1[2] = r2;
    r2 = ((object**) lenv) [1][2];
    r1[3] = r2;
    r2 = top[65];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2231; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
    L2231:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (UOTAG(0)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = ((object****) lenv) [1][1][1][2];
        r3 = allocate (TC_TUPLE, 2);
        r3[1] = lenv; lenv = r3;
        r4 = ((object*****) lenv) [1][1][1][1][2];
        r4 = UOBJ_GET(r4,0);
        r3[2] = r4;
        r3 = ((object*) lenv) [2];
        lenv = lenv[1];
        r2[2] = r3;
        r2 = ((object*) lenv) [2];
        t = alloc_no_clear (UOTAG(0),1);
        t[1] = r2;
        r2 = t;
        lenv = lenv[1];
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2232; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_add_literal_2647_i4643_i4727;
        L2232:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object***) lenv) [1][1][2];
        t = alloc_no_clear (UOTAG(4),1);
        t[1] = r2;
        r2 = t;
        r1[2] = r2;
        r2 = ((object****) lenv) [1][1][1][4];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2233; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_add_literal_2647_i4643_i4727;
        L2233:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[14];
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        r1 = ((object*) lenv) [2];
        r2 = ((object****) lenv) [1][1][1][2];
        t = alloc_no_clear (UOTAG(1),3);
        t[1] = r2;
        t[2] = r0;
        t[3] = r1;
        r0 = t;
        lenv = lenv[1];
        r1 = ((object*****) lenv) [1][1][1][1][2];
        ((pxll_vector*)r1)->val[14] = r0;
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2230:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_get_symbol_index_2649_i4643_i4727; r2[2] = lenv;
  r1[6] = r2;
  // def scan-literals_2650_i4643_i4727
  goto L2234;
  FUN_scan_literals_2650_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object *) 1;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(4)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object****) lenv) [1][1][1][6];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2235; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_get_symbol_index_2649_i4643_i4727;
            L2235:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
          } break;
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object****) lenv) [1][1][1][4];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2236; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_add_literal_2647_i4643_i4727;
            L2236:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
          } break;
          case (UOTAG(5)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,2);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object****) lenv) [1][1][1][7];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2237; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_scan_literals_2650_i4643_i4727;
            L2237:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
          } break;
          case (UOTAG(6)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = ((object*) lenv) [2];
            r0[2] = r1;
            r1 = ((object****) lenv) [1][1][1][7];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2238; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_scan_literals_2650_i4643_i4727;
            L2238:
            ; lenv = k[2]; k = k[1];
            r0 = result;
            lenv = lenv[1];
          } break;
          default: {
            goto fatbar_52;
          }
        }
        goto fatbar_52_over;
        fatbar_52:
        r0 = (object *) 1;
        fatbar_52_over: ;
        r0 = ((object*) lenv) [3];
        lenv = ((object *)lenv)[1];
        lenv[2] = r0;
        goto FUN_scan_literals_2650_i4643_i4727;
      } break;
    }
    PXLL_RETURN(0);
  L2234:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_scan_literals_2650_i4643_i4727; r2[2] = lenv;
  r1[7] = r2;
  // def c-sequence_2652_i4643_i4727
  goto L2239;
  FUN_c_sequence_2652_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*) &constructed_68;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_68;
        (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
        r0 = (object*) &constructed_470;
        dump_object (r0, 0); fprintf (stdout, "\n");
        r0 = (object*) &constructed_470;
        PXLL_UNDEFINED; result=r0; goto Lreturn;
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][3];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        switch (get_case (r0)) {
          case (TC_NIL): {
            r0 = allocate (TC_TUPLE, 5);
            r1 = ((object**) lenv) [1][2];
            r0[2] = r1;
            r1 = ((object*) lenv) [2];
            r0[3] = r1;
            r1 = ((object**) lenv) [1][4];
            r0[4] = r1;
            r1 = ((object**) lenv) [1][5];
            r0[5] = r1;
            r1 = ((object***) lenv) [1][1][3];
            r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
          } break;
          default: {
            goto fatbar_51;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_51_over;
        fatbar_51:
        r0 = allocate (TC_TUPLE, 5);
        r1 = (object *) 6;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][4];
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object***) lenv) [1][1][5];
        r2[2] = r3;
        r3 = top[136];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2240; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_k_2ffree_522;
        L2240:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 5);
        r3 = ((object***) lenv) [1][1][2];
        r2[2] = r3;
        r3 = ((object**) lenv) [1][3];
        r2[3] = r3;
        r3 = ((object***) lenv) [1][1][4];
        r2[4] = r3;
        r3 = ((object***) lenv) [1][1][5];
        r2[5] = r3;
        r3 = ((object****) lenv) [1][1][1][8];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2241; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_c_sequence_2652_i4643_i4727;
        L2241:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r1 = ((object*) lenv) [2];
        r2 = ((object*) lenv) [3];
        r3 = (object *) -1;
        t = alloc_no_clear (UOTAG(0),3);
        t[1] = r3;
        t[2] = r1;
        t[3] = r2;
        r1 = t;
        lenv = lenv[1];
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
        fatbar_51_over: ;
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2239:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_c_sequence_2652_i4643_i4727; r2[2] = lenv;
  r1[8] = r2;
  // def extend-lenv_2655_i4643_i4727
  goto L2242;
  FUN_extend_lenv_2655_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = ((object*) lenv) [3];
        PXLL_RETURN(0);
      } break;
      default: {
        goto fatbar_50;
      }
    }
    PXLL_RETURN(0);
    goto fatbar_50_over;
    fatbar_50:
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(1),2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    PXLL_RETURN(0);
    fatbar_50_over: ;
    PXLL_RETURN(0);
  L2242:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_extend_lenv_2655_i4643_i4727; r2[2] = lenv;
  r1[9] = r2;
  // def search-rib_2657_i4643_i4727
  goto L2243;
  FUN_search_rib_2657_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [4];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = (object*)UITAG(1);
        PXLL_RETURN(0);
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = ((object**) lenv) [1][3];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) 3;
          r1 = ((object**) lenv) [1][3];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          r1 = ((object*) lenv) [3];
          r2 = ((object**) lenv) [1][2];
          lenv = ((object *)lenv)[1];
          lenv[2] = r2;
          lenv[3] = r0;
          lenv[4] = r1;
          goto FUN_search_rib_2657_i4643_i4727;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2243:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_search_rib_2657_i4643_i4727; r2[2] = lenv;
  r1[10] = r2;
  // def lexical-address_2658_i4643_i4727
  goto L2244;
  FUN_lexical_address_2658_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [4];
    switch (get_case (r0)) {
      case (UITAG(0)): {
        r0 = allocate (TC_TUPLE, 3);
        r1 = (object*) &constructed_469;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = top[15];
        r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
      } break;
      case (UOTAG(1)): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 4);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = (object *) 1;
        r1[3] = r2;
        r2 = ((object*) lenv) [2];
        r1[4] = r2;
        r2 = ((object***) lenv) [1][1][10];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2245; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_search_rib_2657_i4643_i4727;
        L2245:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[4] = r1;
        r0 = ((object*) lenv) [4];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][4];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = (object*)UITAG(0);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            r1 = ((object***) lenv) [1][1][3];
            r0 = PXLL_TEST(r1==r0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r0) {
              r0 = allocate (UOTAG(6), 1);
              r1 = ((object*) lenv) [2];
              r0[1] = r1;
              PXLL_RETURN(0);
            } else {
              r0 = allocate (UOTAG(1), 2);
              r1 = ((object***) lenv) [1][1][3];
              r0[1] = r1;
              r1 = ((object*) lenv) [2];
              r0[2] = r1;
              PXLL_RETURN(0);
            }
            PXLL_RETURN(0);
          } break;
          default: {
            r0 = (object *) 3;
            r1 = ((object**) lenv) [1][3];
            r0 = box((pxll_int)unbox(r1)+unbox(r0));
            r1 = ((object*) lenv) [3];
            r2 = ((object**) lenv) [1][2];
            lenv = ((object *)lenv)[1];
            lenv[2] = r2;
            lenv[3] = r0;
            lenv[4] = r1;
            goto FUN_lexical_address_2658_i4643_i4727;
          } break;
        }
        PXLL_RETURN(0);
      } break;
      case (UOTAG(3)): {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,1);
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        r1 = ((object*) lenv) [2];
        r2 = ((object**) lenv) [1][2];
        lenv = ((object *)lenv)[1];
        lenv[2] = r2;
        lenv[3] = r0;
        lenv[4] = r1;
        goto FUN_lexical_address_2658_i4643_i4727;
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r1 = ((object**) lenv) [1][4];
        r1 = UOBJ_GET(r1,2);
        r0[4] = r1;
        r0 = ((object*) lenv) [2];
        r1 = ((object**) lenv) [1][2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = allocate (UOTAG(7), 1);
          r1 = ((object*) lenv) [3];
          r0[1] = r1;
          PXLL_RETURN(0);
        } else {
          r0 = ((object**) lenv) [1][3];
          r1 = ((object*) lenv) [4];
          r2 = ((object**) lenv) [1][2];
          lenv = ((object *)lenv)[1];
          lenv[2] = r2;
          lenv[3] = r0;
          lenv[4] = r1;
          goto FUN_lexical_address_2658_i4643_i4727;
        }
        PXLL_RETURN(0);
      } break;
    }
    PXLL_RETURN(0);
  L2244:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_lexical_address_2658_i4643_i4727; r2[2] = lenv;
  r1[11] = r2;
  // def collect-primargs_2663_i4643_i4727
  goto L2246;
  FUN_collect_primargs_2663_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 6);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2247; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L2247:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r1[1] = lenv; lenv = r1;
    // def $loop_2860_i4643_i4727
    goto L2248;
    FUN__24loop_2860_i4643_i4727:
      check_heap (0);
      r0 = ((object***) lenv) [1][1][2];
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
        r0 = allocate (TC_TUPLE, 3);
        r1 = ((object*) lenv) [3];
        r0[2] = r1;
        r1 = (object *) constructed_36[0];
        r0[3] = r1;
        r1 = top[21];
        r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
      } else {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (UOTAG(1), 2);
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object*****) lenv) [1][1][1][1][2];
        r2[2] = r3;
        r3 = ((object**) lenv) [1][2];
        r2[3] = r3;
        r3 = top[27];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2249; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
        L2249:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[1] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r0[2] = r1;
        r0 = ((object**) lenv) [1][3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        r1 = (object *) 3;
        r2 = ((object*) lenv) [2];
        r1 = box((pxll_int)unbox(r2)+unbox(r1));
        lenv[2] = r1;
        lenv[3] = r0;
        goto FUN__24loop_2860_i4643_i4727;
      }
      PXLL_RETURN(0);
    L2248:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN__24loop_2860_i4643_i4727; r2[2] = lenv;
    r1[2] = r2;
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object *) 1;
    r1[2] = r2;
    r2 = (object*)TC_NIL;
    r1[3] = r2;
    r2 = ((object*) lenv) [2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2250; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN__24loop_2860_i4643_i4727;
    L2250:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    lenv = lenv[1];
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 3);
    // def lambda_18032_i4643_i4727
    goto L2251;
    FUN_lambda_18032_i4643_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case_noint (r0)) {
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r0 = ((object**) lenv) [1][3];
          switch (get_case_noint (r0)) {
            default: {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = ((object***) lenv) [1][1][3];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = ((pxll_vector*)r1)->val[2];
              r0[3] = r1;
              r1 = ((object*) lenv) [2];
              r1 = ((pxll_vector*)r1)->val[2];
              r0[4] = r1;
              r0 = ((object*) lenv) [4];
              r1 = ((object*) lenv) [3];
              r0 = PXLL_TEST(unbox(r1)>unbox(r0));
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L2251:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_lambda_18032_i4643_i4727; r2[2] = lenv;
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r1[3] = r2;
    r2 = top[44];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2252; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_sort_340;
    L2252:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = top[18];
    r1[2] = r2;
    r2 = ((object*) lenv) [4];
    r1[3] = r2;
    r2 = top[32];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2253; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
    L2253:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[5] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = top[17];
    r1[2] = r2;
    r2 = ((object*) lenv) [4];
    r1[3] = r2;
    r2 = top[32];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2254; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
    L2254:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[6] = r1;
    r0 = allocate (TC_TUPLE, 7);
    r1 = ((object*) lenv) [6];
    r0[2] = r1;
    r1 = (object *) constructed_468[0];
    r0[3] = r1;
    r1 = ((object*) lenv) [5];
    r0[4] = r1;
    r1 = ((object**) lenv) [1][3];
    r0[5] = r1;
    r1 = ((object**) lenv) [1][4];
    r0[6] = r1;
    r1 = ((object**) lenv) [1][5];
    r0[7] = r1;
    r1 = ((object***) lenv) [1][1][13];
    r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_splat_2664_i4643_i4727;
  L2246:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_collect_primargs_2663_i4643_i4727; r2[2] = lenv;
  r1[12] = r2;
  // def collect-primargs*_2664_i4643_i4727
  goto L2255;
  FUN_collect_primargs_splat_2664_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    switch (get_case (r0)) {
      case (TC_NIL): {
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][3];
        r1[2] = r2;
        r2 = (object *) constructed_36[0];
        r1[3] = r2;
        r2 = top[21];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2256; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
        L2256:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object**) lenv) [1][4];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2257; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L2257:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        // def $loop_2879_i4643_i4727
        goto L2258;
        FUN__24loop_2879_i4643_i4727:
          check_heap (0);
          r0 = ((object***) lenv) [1][1][3];
          r1 = ((object*) lenv) [2];
          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
            r0 = allocate (TC_TUPLE, 3);
            r1 = ((object*) lenv) [3];
            r0[2] = r1;
            r1 = (object *) constructed_36[0];
            r0[3] = r1;
            r1 = top[21];
            r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][2];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = ((object*****) lenv) [1][1][1][1][4];
            r2[3] = r3;
            r3 = top[28];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2259; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_index_eq_322;
            L2259:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = top[27];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2260; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
            L2260:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r0 = ((object**) lenv) [1][3];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = (object *) 3;
            r2 = ((object*) lenv) [2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN__24loop_2879_i4643_i4727;
          }
          PXLL_RETURN(0);
        L2258:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN__24loop_2879_i4643_i4727; r2[2] = lenv;
        r1[2] = r2;
        r1 = allocate (TC_TUPLE, 3);
        r2 = (object *) 1;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2261; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN__24loop_2879_i4643_i4727;
        L2261:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        lenv = lenv[1];
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r1 = ((object*) lenv) [4];
        r0[2] = r1;
        r1 = ((object**) lenv) [1][7];
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
      } break;
      default: {
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,0);
        r0[2] = r1;
        r1 = ((object**) lenv) [1][2];
        r1 = UOBJ_GET(r1,1);
        r0[3] = r1;
        r0 = allocate (TC_TUPLE, 5);
        r1 = (object *) 6;
        r0[2] = r1;
        r1 = ((object*) lenv) [2];
        r0[3] = r1;
        r1 = ((object**) lenv) [1][5];
        r0[4] = r1;
        r1 = allocate (TC_TUPLE, 3);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object***) lenv) [1][1][6];
        r3[2] = r4;
        r4 = top[136];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2262; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_k_2ffree_522;
        L2262:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r2 = allocate (TC_TUPLE, 3);
        r3 = allocate (TC_TUPLE, 3);
        r4 = ((object*) lenv) [2];
        r3[2] = r4;
        r4 = (object *) constructed_36[0];
        r3[3] = r4;
        r4 = top[21];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2263; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_reverse_onto_311;
        L2263:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = ((object***) lenv) [1][1][3];
        r2[3] = r3;
        r3 = top[21];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2264; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
        L2264:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        lenv = lenv[1];
        r1[2] = r2;
        // def lambda_18129_i4643_i4727
        goto L2265;
        FUN_lambda_18129_i4643_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 7);
          r1 = ((object**) lenv) [1][3];
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][3];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[3] = r1;
          r1 = ((object***) lenv) [1][1][4];
          r0[4] = r1;
          r1 = ((object***) lenv) [1][1][5];
          r0[5] = r1;
          r1 = ((object***) lenv) [1][1][6];
          r0[6] = r1;
          r1 = ((object***) lenv) [1][1][7];
          r0[7] = r1;
          r1 = ((object****) lenv) [1][1][1][13];
          r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_splat_2664_i4643_i4727;
        L2265:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_18129_i4643_i4727; r2[2] = lenv;
        r1[3] = r2;
        r2 = ((object***) lenv) [1][1][2];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2266; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
        L2266:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[5] = r1;
        r1 = ((object***) lenv) [1][1][3];
        r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
      } break;
    }
    PXLL_RETURN(0);
  L2255:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_collect_primargs_splat_2664_i4643_i4727; r2[2] = lenv;
  r1[13] = r2;
  // def c-primargs_2665_i4643_i4727
  goto L2267;
  FUN_c_primargs_2665_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 5);
    r1 = ((object*) lenv) [2];
    r0[2] = r1;
    r1 = ((object*) lenv) [6];
    r0[3] = r1;
    r1 = ((object*) lenv) [7];
    r0[4] = r1;
    // def lambda_18152_i4643_i4727
    goto L2268;
    FUN_lambda_18152_i4643_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      r1 = ((object**) lenv) [1][4];
      r2 = ((object**) lenv) [1][5];
      r3 = ((object**) lenv) [1][7];
      r4 = ((object**) lenv) [1][3];
      t = alloc_no_clear (UOTAG(18),5);
      t[1] = r4;
      t[2] = r1;
      t[3] = r2;
      t[4] = r0;
      t[5] = r3;
      r0 = t;
      PXLL_RETURN(0);
    L2268:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_18152_i4643_i4727; r1[2] = lenv;
    r0[5] = r1;
    r1 = ((object**) lenv) [1][12];
    r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_2663_i4643_i4727;
  L2267:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_c_primargs_2665_i4643_i4727; r2[2] = lenv;
  r1[14] = r2;
  // def c-trcall_2668_i4643_i4727
  goto L2269;
  FUN_c_trcall_2668_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 5);
    r1 = ((object*) lenv) [4];
    r0[2] = r1;
    r1 = ((object*) lenv) [5];
    r0[3] = r1;
    r1 = ((object*) lenv) [6];
    r0[4] = r1;
    // def lambda_18221_i4643_i4727
    goto L2270;
    FUN_lambda_18221_i4643_i4727:
      check_heap (0);
      r0 = ((object**) lenv) [1][3];
      r1 = ((object*) lenv) [2];
      r2 = ((object**) lenv) [1][2];
      t = alloc_no_clear (UOTAG(15),3);
      t[1] = r2;
      t[2] = r0;
      t[3] = r1;
      r0 = t;
      PXLL_RETURN(0);
    L2270:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_18221_i4643_i4727; r1[2] = lenv;
    r0[5] = r1;
    r1 = ((object**) lenv) [1][12];
    r0[1] = r1[2]; lenv = r0; goto FUN_collect_primargs_2663_i4643_i4727;
  L2269:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_c_trcall_2668_i4643_i4727; r2[2] = lenv;
  r1[15] = r2;
  // def compile-store-args_2671_i4643_i4727
  goto L2271;
  FUN_compile_store_args_2671_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 5);
    r1 = (object *) 6;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [4];
    r1[2] = r2;
    r2 = top[19];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2272; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
    L2272:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = ((object*) lenv) [7];
    r0[4] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object*) lenv) [6];
    r1[2] = r2;
    // def lambda_18459_i4643_i4727
    goto L2273;
    FUN_lambda_18459_i4643_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object****) lenv) [1][1][1][4];
      r2[2] = r3;
      r3 = top[20];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L2274; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_cdr_310;
      L2274:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r1 = ((object*) lenv) [2];
      switch (get_case (r1)) {
        case (TC_NIL): {
          r1 = (object *) 262;
        } break;
        default: {
          goto fatbar_49;
        }
      }
      goto fatbar_49_over;
      fatbar_49:
      r1 = (object *) 6;
      fatbar_49_over: ;
      lenv = lenv[1];
      if PXLL_IS_TRUE(r1) {
        r1 = ((object***) lenv) [1][1][8];
      } else {
        r1 = allocate (TC_TUPLE, 2);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 8);
        r3 = (object *) 3;
        r4 = ((object****) lenv) [1][1][1][2];
        r3 = box((pxll_int)unbox(r4)+unbox(r3));
        r2[2] = r3;
        r3 = ((object****) lenv) [1][1][1][3];
        r2[3] = r3;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object****) lenv) [1][1][1][4];
        r3[2] = r4;
        r4 = top[20];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2275; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_cdr_310;
        L2275:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[4] = r3;
        r3 = ((object****) lenv) [1][1][1][5];
        r2[5] = r3;
        r3 = ((object****) lenv) [1][1][1][6];
        r2[6] = r3;
        r3 = ((object****) lenv) [1][1][1][7];
        r2[7] = r3;
        r3 = ((object****) lenv) [1][1][1][8];
        r2[8] = r3;
        r3 = ((object*****) lenv) [1][1][1][1][16];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2276; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_compile_store_args_2671_i4643_i4727;
        L2276:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r1 = ((object****) lenv) [1][1][1][6];
        r2 = ((object*) lenv) [2];
        r3 = (object *) -1;
        t = alloc_no_clear (UOTAG(0),3);
        t[1] = r3;
        t[2] = r1;
        t[3] = r2;
        r1 = t;
        lenv = lenv[1];
      }
      r0[2] = r1;
      r0 = ((object***) lenv) [1][1][2];
      r1 = ((object**) lenv) [1][2];
      r2 = ((object***) lenv) [1][1][5];
      r3 = ((object*) lenv) [2];
      r4 = ((object***) lenv) [1][1][3];
      t = alloc_no_clear (UOTAG(12),5);
      t[1] = r4;
      t[2] = r1;
      t[3] = r2;
      t[4] = r0;
      t[5] = r3;
      r0 = t;
      PXLL_RETURN(0);
    L2273:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_lambda_18459_i4643_i4727; r2[2] = lenv;
    r1[3] = r2;
    r2 = ((object**) lenv) [1][2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2277; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
    L2277:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[5] = r1;
    r1 = ((object**) lenv) [1][3];
    r0[1] = r1[2]; lenv = r0; goto FUN_compile_2646_i4643_i4727;
  L2271:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_compile_store_args_2671_i4643_i4727; r2[2] = lenv;
  r1[16] = r2;
  // def c-let-splat_2673_i4643_i4727
  goto L2278;
  FUN_c_let_splat_2673_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 9);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][4];
    r1[2] = r2;
    r2 = (object *) constructed_36[0];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2279; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L2279:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = top[19];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2280; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
    L2280:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object*) lenv) [2];
    r1[2] = r2;
    r2 = top[20];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2281; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_cdr_310;
    L2281:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object*) lenv) [4];
    r1[2] = r2;
    r2 = (object *) constructed_36[0];
    r1[3] = r2;
    r2 = top[21];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2282; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
    L2282:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[5] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][3];
    r1[2] = r2;
    r2 = top[22];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2283; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
    L2283:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[6] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object**) lenv) [1][6];
    r1[2] = r2;
    r2 = top[136];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2284; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_k_2ffree_522;
    L2284:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[7] = r1;
    r1 = allocate (TC_TUPLE, 5);
    r2 = ((object**) lenv) [1][2];
    r1[2] = r2;
    r2 = ((object*) lenv) [3];
    r1[3] = r2;
    r2 = ((object**) lenv) [1][3];
    switch (get_case (r2)) {
      case (TC_NIL): {
        r2 = ((object**) lenv) [1][5];
      } break;
      default: {
        goto fatbar_48;
      }
    }
    goto fatbar_48_over;
    fatbar_48:
    r2 = ((object**) lenv) [1][5];
    r3 = ((object**) lenv) [1][3];
    t = alloc_no_clear (UOTAG(1),2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    fatbar_48_over: ;
    r1[4] = r2;
    r2 = allocate (TC_TUPLE, 3);
    r3 = allocate (TC_TUPLE, 2);
    r4 = ((object**) lenv) [1][6];
    r3[2] = r4;
    r4 = top[136];
    t = allocate (TC_SAVE, 6);
    t[1] = k; t[2] = lenv; t[3] = &&L2285; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
    r3[1] = r4[2]; lenv = r3; goto FUN_k_2ffree_522;
    L2285:
    r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
    r3 = result;
    r2[2] = r3;
    // def lambda_18565_i4643_i4727
    goto L2286;
    FUN_lambda_18565_i4643_i4727:
      check_heap (0);
      r0 = ((object***) lenv) [1][1][6];
      r1 = ((object*) lenv) [2];
      t = alloc_no_clear (UOTAG(17),2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      PXLL_RETURN(0);
    L2286:
    r3 = allocate (TC_CLOSURE, 2);
    r3[1] = &&FUN_lambda_18565_i4643_i4727; r3[2] = lenv;
    r2[3] = r3;
    r3 = ((object***) lenv) [1][1][2];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L2287; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
    L2287:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[5] = r2;
    r2 = ((object***) lenv) [1][1][3];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2288; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
    L2288:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[8] = r1;
    r1 = ((object*) lenv) [7];
    r2 = ((object*) lenv) [8];
    r3 = (object *) -1;
    t = alloc_no_clear (UOTAG(0),3);
    t[1] = r3;
    t[2] = r1;
    t[3] = r2;
    r1 = t;
    r0[9] = r1;
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object****) lenv) [1][1][1][2];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object****) lenv) [1][1][1][49];
    r1[2] = r2;
    r2 = top[19];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2289; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
    L2289:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = top[156];
    r0[4] = r1;
    r1 = top[6];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2290; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
    L2290:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 3);
    r2 = ((object**) lenv) [1][7];
    r1[2] = r2;
    // def lambda_18598_i4643_i4727
    goto L2291;
    FUN_lambda_18598_i4643_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 8);
      r2 = (object *) 1;
      r1[2] = r2;
      r2 = (object *) 3;
      r1[3] = r2;
      r2 = ((object****) lenv) [1][1][1][5];
      r1[4] = r2;
      r2 = ((object**) lenv) [1][2];
      r1[5] = r2;
      r2 = ((object****) lenv) [1][1][1][7];
      r3 = ((object**) lenv) [1][2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      r1[6] = r2;
      r2 = ((object*****) lenv) [1][1][1][1][3];
      switch (get_case (r2)) {
        case (TC_NIL): {
          r2 = ((object*****) lenv) [1][1][1][1][5];
        } break;
        default: {
          goto fatbar_47;
        }
      }
      goto fatbar_47_over;
      fatbar_47:
      r2 = ((object*****) lenv) [1][1][1][1][5];
      r3 = ((object*****) lenv) [1][1][1][1][3];
      t = alloc_no_clear (UOTAG(1),2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      fatbar_47_over: ;
      r1[7] = r2;
      r2 = ((object****) lenv) [1][1][1][9];
      r1[8] = r2;
      r2 = ((object******) lenv) [1][1][1][1][1][16];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2292; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_compile_store_args_2671_i4643_i4727;
      L2292:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][7];
      r2 = ((object*) lenv) [2];
      r3 = (object *) -1;
      t = alloc_no_clear (UOTAG(0),3);
      t[1] = r3;
      t[2] = r1;
      t[3] = r2;
      r1 = t;
      r0[3] = r1;
      r0 = ((object*) lenv) [3];
      r1 = ((object**) lenv) [1][2];
      t = alloc_no_clear (UOTAG(16),2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      PXLL_RETURN(0);
    L2291:
    r2 = allocate (TC_CLOSURE, 2);
    r2[1] = &&FUN_lambda_18598_i4643_i4727; r2[2] = lenv;
    r1[3] = r2;
    r2 = ((object****) lenv) [1][1][1][2];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2293; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_cont_2644_i4643_i4727;
    L2293:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r1 = ((object**) lenv) [1][6];
    t = alloc_no_clear (UOTAG(10),2);
    t[1] = r1;
    t[2] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L2278:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_c_let_splat_2673_i4643_i4727; r2[2] = lenv;
  r1[17] = r2;
  // def record-label-tag_2681_i4643_i4727
  goto L2294;
  FUN_record_label_tag_2681_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_3065_i4643_i4727
    goto L2295;
    FUN_loop_3065_i4643_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case (r0)) {
        case (TC_PAIR): {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case_noint (r0)) {
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = ((object*****) lenv) [1][1][1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = ((object**) lenv) [1][3];
                lenv = ((object **)lenv)[1][1];
                lenv[2] = r0;
                goto FUN_loop_3065_i4643_i4727;
              }
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        } break;
        default: {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object******) lenv) [1][1][1][1][1][2];
          r2 = ((pxll_vector*)r2)->val[5];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2296; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L2296:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (UOTAG(1), 2);
          r2 = ((object*****) lenv) [1][1][1][1][2];
          r1[1] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r0[2] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
          r1 = ((pxll_vector*)r1)->val[5];
          r0[3] = r1;
          r0 = ((object*) lenv) [3];
          r1 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r1;
          t[2] = r0;
          r0 = t;
          lenv = lenv[1];
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          ((pxll_vector*)r1)->val[5] = r0;
          PXLL_RETURN(0);
        } break;
      }
      PXLL_RETURN(0);
    L2295:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_3065_i4643_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object****) lenv) [1][1][1][2];
    r1 = ((pxll_vector*)r1)->val[5];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_3065_i4643_i4727;
  L2294:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_record_label_tag_2681_i4643_i4727; r2[2] = lenv;
  r1[18] = r2;
  // def gen-return_2684_i4643_i4727
  goto L2297;
  FUN_gen_return_2684_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(0),1);
    t[1] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L2297:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_gen_return_2684_i4643_i4727; r2[2] = lenv;
  r1[19] = r2;
  // def gen-invoke_2685_i4643_i4727
  goto L2298;
  FUN_gen_invoke_2685_i4643_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 4);
    r1 = ((object***) lenv) [1][1][2];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = ((object***) lenv) [1][1][49];
    r1[2] = r2;
    r2 = top[19];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2299; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_car_309;
    L2299:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = top[156];
    r0[4] = r1;
    r1 = top[6];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2300; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_vars_set_flag_bang_248;
    L2300:
    ; lenv = k[2]; k = k[1];
    r0 = ((object*) lenv) [5];
    r1 = ((object*) lenv) [3];
    r2 = ((object*) lenv) [4];
    r3 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(13),4);
    t[1] = r3;
    t[2] = r1;
    t[3] = r2;
    t[4] = r0;
    r0 = t;
    PXLL_RETURN(0);
  L2298:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_gen_invoke_2685_i4643_i4727; r2[2] = lenv;
  r1[20] = r2;
  // def gen-tail_2686_i4643_i4727
  goto L2301;
  FUN_gen_tail_2686_i4643_i4727:
    check_heap (0);
    r0 = ((object*) lenv) [3];
    r1 = ((object*) lenv) [4];
    r2 = ((object*) lenv) [2];
    t = alloc_no_clear (UOTAG(14),3);
    t[1] = r2;
    t[2] = r0;
    t[3] = r1;
    r0 = t;
    PXLL_RETURN(0);
  L2301:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_gen_tail_2686_i4643_i4727; r2[2] = lenv;
  r1[21] = r2;
  r2 = allocate (TC_TUPLE, 2);
  r3 = (object *) 1;
  r2[2] = r3;
  r3 = top[113];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2302; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_make_counter_472;
  L2302:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[22] = r2;
  r1 = allocate (TC_TUPLE, 5);
  r2 = (object *) 262;
  r1[2] = r2;
  r2 = ((object**) lenv) [1][48];
  r1[3] = r2;
  r2 = (object*)UITAG(0);
  r1[4] = r2;
  r2 = allocate (TC_TUPLE, 3);
  r3 = (object *) constructed_467[0];
  r2[2] = r3;
  r3 = ((object*) lenv) [19];
  r2[3] = r3;
  r3 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2303; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_cont_2644_i4643_i4727;
  L2303:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[5] = r2;
  r2 = ((object*) lenv) [3];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2304; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_compile_2646_i4643_i4727;
  L2304:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[50] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = (object*)UITAG(5);
  r1[2] = r2;
  r2 = (object *) constructed_466[0];
  r1[3] = r2;
  r2 = top[118];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2305; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_make_node_486;
  L2305:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  ((object*) lenv) [30] = r1;
  r0[51] = r1;
  r1 = allocate (UOTAG(1), 3);
  r2 = (object *) 32769;
  r2 = box((pxll_int)string_tuple_length (unbox(r2)));
  ensure_heap (2, unbox(r2));
  r2 = (object *) 32769;
  r2 = (t=alloc_no_clear (TC_STRING, string_tuple_length (unbox(r2))), ((pxll_string*)(t))->len = unbox(r2), t);
  r1[1] = r2;
  r2 = allocate (TC_TUPLE, 4);
  r3 = ((object*) lenv) [12];
  r2[2] = r3;
  r3 = allocate (TC_TUPLE, 2);
  r3[1] = lenv; lenv = r3;
  r4 = top[162];
  r5 = top[161];
  r4 = box((pxll_int)unbox(r5)|unbox(r4));
  r3[2] = r4;
  r3 = ((object*) lenv) [2];
  r4 = top[163];
  r3 = box((pxll_int)unbox(r4)|unbox(r3));
  lenv = lenv[1];
  r2[3] = r3;
  r3 = (object *) 841;
  r2[4] = r3;
  r3 = top[60];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2306; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_open_373;
  L2306:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  r2 = (object *) 1;
  r1[3] = r2;
  r0[52] = r1;
  r1 = (object *) 3;
  r0[53] = r1;
  r1 = allocate (TC_TUPLE, 7);
  r1[1] = lenv; lenv = r1;
  // def write-indent_4165_i4646_i4727
  goto L2307;
  FUN_write_indent_4165_i4646_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    // def loop_4171_i4646_i4727
    goto L2308;
    FUN_loop_4171_i4646_i4727:
      check_heap (0);
      r0 = (object *) 1;
      r1 = ((object*) lenv) [2];
      if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object*****) lenv) [1][1][1][1][52];
        r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,6)];
        r0[2] = r1;
        r0 = (object*) &constructed_465;
        r0 = box((pxll_int)((pxll_string*)(r0))->len);
        r1 = (object*) &constructed_465;
        r2 = ((object*) lenv) [2];
        r0 = box((pxll_int)write (unbox(r2), ((pxll_string*)(r1))->data, unbox(r0)));
        lenv = lenv[1];
        r0 = (object *) 3;
        r1 = ((object*) lenv) [2];
        r0 = box((pxll_int)unbox(r1)-unbox(r0));
        lenv[2] = r0;
        goto FUN_loop_4171_i4646_i4727;
      } else {
        r0 = (object *) 14;
        PXLL_RETURN(0);
      }
      PXLL_RETURN(0);
    L2308:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_loop_4171_i4646_i4727; r1[2] = lenv;
    r0[2] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object***) lenv) [1][1][53];
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r0[1] = r1[2]; lenv = r0; goto FUN_loop_4171_i4646_i4727;
  L2307:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_write_indent_4165_i4646_i4727; r2[2] = lenv;
  r1[2] = r2;
  // def write-string_4166_i4646_i4727
  goto L2309;
  FUN_write_string_4166_i4646_i4727:
    check_heap (0);
    r0 = ((object**) lenv) [1][2];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2310; ; k = t;
    lenv = r0[2]; goto FUN_write_indent_4165_i4646_i4727;
    L2310:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object****) lenv) [1][1][1][52];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,6)];
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = ((object**) lenv) [1][2];
    r2 = ((object*) lenv) [2];
    r0 = box((pxll_int)write (unbox(r2), ((pxll_string*)(r1))->data, unbox(r0)));
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object****) lenv) [1][1][1][52];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,6)];
    r0[2] = r1;
    r0 = (object*) &constructed_464;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_464;
    r2 = ((object*) lenv) [2];
    r0 = box((pxll_int)write (unbox(r2), ((pxll_string*)(r1))->data, unbox(r0)));
    lenv = lenv[1];
    r0 = (object *) 14;
    PXLL_RETURN(0);
  L2309:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_write_string_4166_i4646_i4727; r2[2] = lenv;
  r1[3] = r2;
  // def copy_4167_i4646_i4727
  goto L2311;
  FUN_copy_4167_i4646_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object****) lenv) [1][1][1][52];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,6)];
    r0[2] = r1;
    r0 = ((object**) lenv) [1][2];
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = ((object**) lenv) [1][2];
    r2 = ((object*) lenv) [2];
    r0 = box((pxll_int)write (unbox(r2), ((pxll_string*)(r1))->data, unbox(r0)));
    PXLL_RETURN(0);
  L2311:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_copy_4167_i4646_i4727; r2[2] = lenv;
  r1[4] = r2;
  // def indent_4168_i4646_i4727
  goto L2312;
  FUN_indent_4168_i4646_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][53];
    r0[2] = r1;
    r0 = (object *) 3;
    r1 = ((object*) lenv) [2];
    r0 = box((pxll_int)unbox(r1)+unbox(r0));
    lenv = lenv[1];
    ((object**) lenv) [1][53] = r0;
    PXLL_RETURN(0);
  L2312:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_indent_4168_i4646_i4727; r2[2] = lenv;
  r1[5] = r2;
  // def dedent_4169_i4646_i4727
  goto L2313;
  FUN_dedent_4169_i4646_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][53];
    r0[2] = r1;
    r0 = (object *) 3;
    r1 = ((object*) lenv) [2];
    r0 = box((pxll_int)unbox(r1)-unbox(r0));
    lenv = lenv[1];
    ((object**) lenv) [1][53] = r0;
    PXLL_RETURN(0);
  L2313:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_dedent_4169_i4646_i4727; r2[2] = lenv;
  r1[6] = r2;
  // def close-file_4170_i4646_i4727
  goto L2314;
  FUN_close_file_4170_i4646_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object***) lenv) [1][1][52];
    r1 = ((pxll_vector*)r1)->val[lookup_field((GET_TYPECODE(*r1)-TC_USEROBJ)>>2,6)];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r0 = box((pxll_int)close (unbox(r0)));
    PXLL_RETURN(0);
  L2314:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_close_file_4170_i4646_i4727; r2[2] = lenv;
  r1[7] = r2;
  r1 = allocate (UOTAG(0), 5);
  r2 = ((object*) lenv) [7];
  r1[1] = r2;
  r2 = ((object*) lenv) [4];
  r1[2] = r2;
  r2 = ((object*) lenv) [6];
  r1[3] = r2;
  r2 = ((object*) lenv) [5];
  r1[4] = r2;
  r2 = ((object*) lenv) [3];
  r1[5] = r2;
  lenv = lenv[1];
  r0[54] = r1;
  r0 = ((object*) lenv) [2];
  r0 = ((pxll_vector*)r0)->val[9];
  r0 = ((pxll_vector*)r0)->val[4];
  if PXLL_IS_TRUE(r0) {
    r0 = (object*) &constructed_463;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_463;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 3);
    r1 = ((object*) lenv) [50];
    r0[2] = r1;
    r1 = (object *) 1;
    r0[3] = r1;
    r1 = top[138];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2315; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_print_insn_526;
    L2315:
    ; lenv = k[2]; k = k[1];
    r0 = top[158];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2316; ; k = t;
    lenv = r0[2]; goto *r0[1];
    L2316:
    ; lenv = k[2]; k = k[1];
    r0 = (object*) &constructed_462;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_462;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_34348_i4727
    goto L2317;
    FUN_lambda_34348_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [3];
      r0[2] = r1;
      r1 = top[117];
      r0[1] = r1[2]; lenv = r0; goto FUN_print_datatype_481;
    L2317:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_34348_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[2];
    r0[3] = r1;
    r1 = top[67];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2318; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
    L2318:
    ; lenv = k[2]; k = k[1];
    r0 = (object*) &constructed_461;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_461;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[5];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r0 = dump_object (r0, 0); fprintf (stdout, "\n");
    lenv = lenv[1];
    r0 = (object*) &constructed_460;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_460;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[10];
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r0 = dump_object (r0, 0); fprintf (stdout, "\n");
    lenv = lenv[1];
    r0 = (object*) &constructed_459;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_459;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_34396_i4727
    goto L2319;
    FUN_lambda_34396_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 8);
      r2[1] = lenv; lenv = r2;
      r3 = allocate (TC_TUPLE, 2);
      r4 = ((object***) lenv) [1][1][3];
      r3[2] = r4;
      r4 = top[56];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L2320; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
      L2320:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = ((object***) lenv) [1][1][2];
      r3 = allocate (TC_TUPLE, 2);
      r3[1] = lenv; lenv = r3;
      r4 = ((object****) lenv) [1][1][1][2];
      r4 = UOBJ_GET(r4,0);
      r3[2] = r4;
      r3 = ((object*) lenv) [2];
      lenv = lenv[1];
      r2[3] = r3;
      r3 = (object*)TC_NIL;
      r2[4] = r3;
      r3 = ((object*) lenv) [4];
      r4 = (object*) &constructed_458;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[5] = r3;
      r3 = ((object*) lenv) [5];
      r4 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[6] = r3;
      r3 = ((object*) lenv) [6];
      r4 = (object*) &constructed_457;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[7] = r3;
      r3 = ((object*) lenv) [7];
      r4 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[8] = r3;
      r2 = ((object*) lenv) [8];
      r3 = (object*) &constructed_456;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2321; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L2321:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      PXLL_RETURN(0);
    L2319:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_34396_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[14];
    r0[3] = r1;
    r1 = top[67];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2322; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
    L2322:
    ; lenv = k[2]; k = k[1];
    r0 = (object*) &constructed_455;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_455;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_34430_i4727
    goto L2323;
    FUN_lambda_34430_i4727:
      check_heap (0);
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 8);
      r2[1] = lenv; lenv = r2;
      r3 = allocate (TC_TUPLE, 2);
      r4 = ((object***) lenv) [1][1][3];
      r3[2] = r4;
      r4 = top[56];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L2324; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
      L2324:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = ((object***) lenv) [1][1][2];
      r3 = allocate (TC_TUPLE, 2);
      r3[1] = lenv; lenv = r3;
      r4 = ((object****) lenv) [1][1][1][2];
      r4 = UOBJ_GET(r4,0);
      r3[2] = r4;
      r3 = ((object*) lenv) [2];
      lenv = lenv[1];
      r2[3] = r3;
      r3 = (object*)TC_NIL;
      r2[4] = r3;
      r3 = ((object*) lenv) [4];
      r4 = (object*) &constructed_454;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[5] = r3;
      r3 = ((object*) lenv) [5];
      r4 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[6] = r3;
      r3 = ((object*) lenv) [6];
      r4 = (object*) &constructed_453;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[7] = r3;
      r3 = ((object*) lenv) [7];
      r4 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[8] = r3;
      r2 = ((object*) lenv) [8];
      r3 = (object*) &constructed_452;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2325; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L2325:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r0 = ((object*) lenv) [2];
      r0 = box((pxll_int)((pxll_string*)(r0))->len);
      r1 = ((object*) lenv) [2];
      r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
      PXLL_RETURN(0);
    L2323:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_34430_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[15];
    r0[3] = r1;
    r1 = top[67];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2326; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
    L2326:
    ; lenv = k[2]; k = k[1];
    r0 = (object *) 14;
  } else {
    r0 = (object *) 14;
  }
  r0 = (object*) &constructed_451;
  r0 = box((pxll_int)((pxll_string*)(r0))->len);
  r1 = (object*) &constructed_451;
  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
  r0 = (object*) &constructed_450;
  r0 = box((pxll_int)((pxll_string*)(r0))->len);
  r1 = (object*) &constructed_450;
  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
  r0 = ((object*) lenv) [12];
  r0 = box((pxll_int)((pxll_string*)(r0))->len);
  r1 = ((object*) lenv) [12];
  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
  r0 = top[158];
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L2327; ; k = t;
  lenv = r0[2]; goto *r0[1];
  L2327:
  ; lenv = k[2]; k = k[1];
  r0 = allocate (TC_TUPLE, 3);
  // def lambda_34467_i4727
  goto L2328;
  FUN_lambda_34467_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 4);
    r2[1] = lenv; lenv = r2;
    r3 = (object*)TC_NIL;
    r2[2] = r3;
    r3 = ((object*) lenv) [2];
    r4 = (object*) &constructed_449;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r4 = ((object**) lenv) [1][2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[4] = r3;
    r2 = ((object*) lenv) [4];
    r3 = (object*) &constructed_448;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2329; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L2329:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r1 = ((object**) lenv) [1][54];
    r1 = ((pxll_vector*)r1)->val[4];
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
  L2328:
  r1 = allocate (TC_CLOSURE, 2);
  r1[1] = &&FUN_lambda_34467_i4727; r1[2] = lenv;
  r0[2] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = ((object**) lenv) [1][2];
  r2 = ((pxll_vector*)r2)->val[0];
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [2];
  r1[2] = r2;
  r2 = (object *) constructed_36[0];
  r1[3] = r2;
  r2 = top[21];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2330; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
  L2330:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[3] = r1;
  r1 = top[35];
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L2331; ; k = t;
  r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
  L2331:
  ; lenv = k[2]; k = k[1];
  r0 = allocate (TC_TUPLE, 3);
  r1 = ((object*) lenv) [54];
  r1 = ((pxll_vector*)r1)->val[4];
  r0[2] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  r2 = ((object**) lenv) [1][2];
  r2 = ((pxll_vector*)r2)->val[1];
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 3);
  r2 = ((object*) lenv) [2];
  r1[2] = r2;
  r2 = (object *) constructed_36[0];
  r1[3] = r2;
  r2 = top[21];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2332; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
  L2332:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[3] = r1;
  r1 = top[35];
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L2333; ; k = t;
  r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
  L2333:
  ; lenv = k[2]; k = k[1];
  r0 = allocate (TC_TUPLE, 6);
  r0[1] = lenv; lenv = r0;
  r1 = allocate (TC_TUPLE, 2);
  r2 = (object*) &constructed_447;
  r1[2] = r2;
  r2 = top[61];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2334; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_file_2fopen_read_380;
  L2334:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[2] = r1;
  r1 = allocate (TC_TUPLE, 2);
  r1[1] = lenv; lenv = r1;
  // def loop_4683_i4578_i4680_i4727
  goto L2335;
  FUN_loop_4683_i4578_i4680_i4727:
    check_heap (0);
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = box((pxll_int)((pxll_string*)(r1))->len);
    r0[2] = r1;
    r0 = (object *) 1;
    r1 = ((object*) lenv) [2];
    r0 = PXLL_TEST(unbox(r1)==unbox(r0));
    lenv = lenv[1];
    if PXLL_IS_TRUE(r0) {
      r0 = allocate (TC_TUPLE, 2);
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object*) lenv) [3];
      r1[2] = r2;
      r2 = (object *) constructed_36[0];
      r1[3] = r2;
      r2 = top[21];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2336; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
      L2336:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = top[47];
      r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
    } else {
      r0 = ((object*) lenv) [3];
      r1 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r1;
      t[2] = r0;
      r0 = t;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object***) lenv) [1][1][2];
      r1[2] = r2;
      r2 = top[63];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2337; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_file_2fread_buffer_386;
      L2337:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      lenv[2] = r1;
      lenv[3] = r0;
      goto FUN_loop_4683_i4578_i4680_i4727;
    }
    PXLL_RETURN(0);
  L2335:
  r2 = allocate (TC_CLOSURE, 2);
  r2[1] = &&FUN_loop_4683_i4578_i4680_i4727; r2[2] = lenv;
  r1[2] = r2;
  r1 = allocate (TC_TUPLE, 3);
  r2 = allocate (TC_TUPLE, 2);
  r3 = ((object**) lenv) [1][2];
  r2[2] = r3;
  r3 = top[63];
  t = allocate (TC_SAVE, 5);
  t[1] = k; t[2] = lenv; t[3] = &&L2338; t[4] = r0; t[5] = r1; k = t;
  r2[1] = r3[2]; lenv = r2; goto FUN_file_2fread_buffer_386;
  L2338:
  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
  r2 = result;
  r1[2] = r2;
  r2 = (object *) constructed_446[0];
  r1[3] = r2;
  r2 = ((object*) lenv) [2];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2339; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_loop_4683_i4578_i4680_i4727;
  L2339:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  lenv = lenv[1];
  r0[3] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = top[206];
  r1[2] = r2;
  r2 = ((object*) lenv) [3];
  r1[3] = r2;
  r2 = top[51];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2340; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string_find_356;
  L2340:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[4] = r1;
  r1 = allocate (TC_TUPLE, 3);
  r2 = top[207];
  r1[2] = r2;
  r2 = ((object*) lenv) [3];
  r1[3] = r2;
  r2 = top[51];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2341; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_string_find_356;
  L2341:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[5] = r1;
  r1 = (object *) -1;
  r2 = ((object*) lenv) [4];
  if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
    r1 = (object *) 262;
  } else {
    r1 = (object *) -1;
    r2 = ((object*) lenv) [5];
    r1 = PXLL_TEST(unbox(r2)==unbox(r1));
  }
  if PXLL_IS_TRUE(r1) {
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object*) &constructed_445;
    r1[2] = r2;
    r2 = allocate (UOTAG(1), 2);
    r3 = ((object*) lenv) [4];
    r2[1] = r3;
    r3 = ((object*) lenv) [5];
    r2[2] = r3;
    r1[3] = r2;
    r2 = top[15];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2342; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
    L2342:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
  } else {
    r1 = allocate (TC_TUPLE, 8);
    r1[1] = lenv; lenv = r1;
    r2 = top[206];
    r2 = box((pxll_int)((pxll_string*)(r2))->len);
    r1[2] = r2;
    r2 = ((object*) lenv) [2];
    r3 = ((object**) lenv) [1][4];
    r2 = box((pxll_int)unbox(r3)+unbox(r2));
    r1[3] = r2;
    r2 = top[207];
    r2 = box((pxll_int)((pxll_string*)(r2))->len);
    r1[4] = r2;
    r2 = ((object*) lenv) [4];
    r3 = ((object**) lenv) [1][5];
    r2 = box((pxll_int)unbox(r3)+unbox(r2));
    r1[5] = r2;
    r2 = allocate (TC_TUPLE, 4);
    r3 = ((object**) lenv) [1][3];
    r2[2] = r3;
    r3 = (object *) 1;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r2[4] = r3;
    r3 = top[45];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L2343; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_substring_345;
    L2343:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[6] = r2;
    r2 = allocate (TC_TUPLE, 4);
    r3 = ((object**) lenv) [1][3];
    r2[2] = r3;
    r3 = ((object*) lenv) [3];
    r2[3] = r3;
    r3 = ((object*) lenv) [5];
    r2[4] = r3;
    r3 = top[45];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L2344; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_substring_345;
    L2344:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[7] = r2;
    r2 = allocate (TC_TUPLE, 4);
    r3 = ((object**) lenv) [1][3];
    r2[2] = r3;
    r3 = ((object*) lenv) [5];
    r2[3] = r3;
    r3 = ((object**) lenv) [1][3];
    r3 = box((pxll_int)((pxll_string*)(r3))->len);
    r2[4] = r3;
    r3 = top[45];
    t = allocate (TC_SAVE, 5);
    t[1] = k; t[2] = lenv; t[3] = &&L2345; t[4] = r0; t[5] = r1; k = t;
    r2[1] = r3[2]; lenv = r2; goto FUN_substring_345;
    L2345:
    r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
    r2 = result;
    r1[8] = r2;
    r1 = allocate (UOTAG(8), 3);
    r2 = ((object*) lenv) [6];
    r1[1] = r2;
    r2 = ((object*) lenv) [7];
    r1[2] = r2;
    r2 = ((object*) lenv) [8];
    r1[3] = r2;
    lenv = lenv[1];
  }
  r0[6] = r1;
  r0 = ((object*) lenv) [6];
  switch (get_case_noint (r0)) {
    default: {
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][6];
      r1 = UOBJ_GET(r1,0);
      r0[2] = r1;
      r1 = ((object**) lenv) [1][6];
      r1 = UOBJ_GET(r1,1);
      r0[3] = r1;
      r1 = ((object**) lenv) [1][6];
      r1 = UOBJ_GET(r1,2);
      r0[4] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [2];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][54];
      r1 = ((pxll_vector*)r1)->val[1];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2346; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2346:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 8);
      r0[1] = lenv; lenv = r0;
      r1 = ((object****) lenv) [1][1][1][2];
      r1 = ((pxll_vector*)r1)->val[7];
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = (object *) constructed_36[0];
      r1[3] = r2;
      r2 = top[21];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2347; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
      L2347:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[3] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r2 = ((object*) lenv) [3];
      r1[2] = r2;
      r2 = top[22];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2348; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
      L2348:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[4] = r1;
      r1 = (object*)UITAG(0);
      r0[5] = r1;
      r1 = (object *) constructed_444[0];
      r0[6] = r1;
      r1 = (object *) 1;
      r0[7] = r1;
      r1 = (object *) 1;
      r0[8] = r1;
      r0 = allocate (TC_TUPLE, 4);
      r0[1] = lenv; lenv = r0;
      // def uohead_4620_i4681_i4727
      goto L2349;
      FUN_uohead_4620_i4681_i4727:
        check_heap (0);
        r0 = (object *) constructed_51[0];
        r1 = ((object*) lenv) [3];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_61[0];
          r1 = ((object*) lenv) [4];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_440;
            PXLL_RETURN(0);
          } else {
            goto fatbar_46;
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_46;
        }
        PXLL_RETURN(0);
        goto fatbar_46_over;
        fatbar_46:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 8);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r2[2] = r3;
        r3 = top[56];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2350; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
        L2350:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][5];
        r2[2] = r3;
        r3 = top[56];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2351; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
        L2351:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[3] = r2;
        r2 = (object*)TC_NIL;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = (object*) &constructed_443;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r2 = ((object*) lenv) [5];
        r3 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[6] = r2;
        r2 = ((object*) lenv) [6];
        r3 = (object*) &constructed_442;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[7] = r2;
        r2 = ((object*) lenv) [7];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[8] = r2;
        r1 = ((object*) lenv) [8];
        r2 = (object*) &constructed_441;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        fatbar_46_over: ;
        PXLL_RETURN(0);
      L2349:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_uohead_4620_i4681_i4727; r1[2] = lenv;
      r0[2] = r1;
      // def uitag_4621_i4681_i4727
      goto L2352;
      FUN_uitag_4621_i4681_i4727:
        check_heap (0);
        r0 = (object *) constructed_51[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_49[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_437;
            PXLL_RETURN(0);
          } else {
            goto fatbar_45;
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_45;
        }
        PXLL_RETURN(0);
        goto fatbar_45_over;
        fatbar_45:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][4];
        r2[2] = r3;
        r3 = top[56];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2353; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
        L2353:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_439;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_438;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        fatbar_45_over: ;
        PXLL_RETURN(0);
      L2352:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_uitag_4621_i4681_i4727; r1[2] = lenv;
      r0[3] = r1;
      // def walk_4622_i4681_i4727
      goto L2354;
      FUN_walk_4622_i4681_i4727:
        check_heap (0);
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(5)): {
            r0 = allocate (TC_TUPLE, 7);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,1);
            r0[3] = r1;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,2);
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 4);
            r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r2 = ((pxll_vector*)r2)->val[2];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = (object*) &constructed_425;
            r1[4] = r2;
            r2 = top[66];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2355; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_alist_2fget_396;
            L2355:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = ((object*) lenv) [5];
            r2 = ((pxll_vector*)r2)->val[1];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2356; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2356:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[6] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [4];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2357; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L2357:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[7] = r1;
            r0 = (object *) 1;
            r1 = ((object*) lenv) [7];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>unbox(r0))) {
              r0 = allocate (TC_TUPLE, 4);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 3);
              r2 = ((object****) lenv) [1][1][1][4];
              r1[2] = r2;
              r2 = ((object**) lenv) [1][4];
              r1[3] = r2;
              r2 = top[32];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2358; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
              L2358:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object*****) lenv) [1][1][1][1][6];
              r1[2] = r2;
              r2 = top[22];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2359; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
              L2359:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[3] = r1;
              r1 = ((object*) lenv) [3];
              r2 = (object *) 3;
              r1 = box((pxll_int)unbox(r2)+unbox(r1));
              r0[4] = r1;
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = allocate (TC_TUPLE, 5);
              r2 = ((object***) lenv) [1][1][7];
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = ((object***) lenv) [1][1][3];
              r1[4] = r2;
              r2 = ((object***) lenv) [1][1][6];
              r2 = ((pxll_vector*)r2)->val[1];
              r1[5] = r2;
              r2 = ((object*****) lenv) [1][1][1][1][2];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2360; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_uohead_4620_i4681_i4727;
              L2360:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object******) lenv) [1][1][1][1][1][6];
              r0[3] = r1;
              r0 = ((object*) lenv) [3];
              r1 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              ((object*****) lenv) [1][1][1][1][6] = r0;
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_33094_i4681_i4727
              goto L2361;
              FUN_lambda_33094_i4681_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object*******) lenv) [1][1][1][1][1][1][6];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object******) lenv) [1][1][1][1][1][6] = r0;
                PXLL_RETURN(0);
              L2361:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_33094_i4681_i4727; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r0[3] = r1;
              r1 = top[35];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L2362; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
              L2362:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 8);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 2);
              r3 = ((object******) lenv) [1][1][1][1][1][7];
              r2[2] = r3;
              r3 = top[56];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2363; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
              L2363:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = allocate (TC_TUPLE, 2);
              r3 = ((object**) lenv) [1][4];
              r2[2] = r3;
              r3 = top[56];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2364; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
              L2364:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[3] = r2;
              r2 = (object*)TC_NIL;
              r1[4] = r2;
              r2 = ((object*) lenv) [4];
              r3 = (object*) &constructed_424;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[5] = r2;
              r2 = ((object*) lenv) [5];
              r3 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[6] = r2;
              r2 = ((object*) lenv) [6];
              r3 = (object*) &constructed_423;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[7] = r2;
              r2 = ((object*) lenv) [7];
              r3 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              r1[8] = r2;
              r1 = ((object*) lenv) [8];
              r2 = (object*) &constructed_422;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r2;
              t[2] = r1;
              r1 = t;
              lenv = lenv[1];
              r0[2] = r1;
              r1 = top[47];
              r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
            } else {
              r0 = allocate (TC_TUPLE, 4);
              r1 = ((object*) lenv) [2];
              r0[2] = r1;
              r1 = ((object*) lenv) [3];
              r0[3] = r1;
              r1 = ((object*) lenv) [6];
              r1 = ((pxll_vector*)r1)->val[1];
              r0[4] = r1;
              r1 = ((object***) lenv) [1][1][3];
              r0[1] = r1[2]; lenv = r0; goto FUN_uitag_4621_i4681_i4727;
            }
            PXLL_RETURN(0);
          } break;
          case (UOTAG(6)): {
            r0 = allocate (TC_TUPLE, 6);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object***) lenv) [1][1][4];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = top[32];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2365; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
            L2365:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2366; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L2366:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[4] = r1;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object****) lenv) [1][1][1][6];
            r1[2] = r2;
            r2 = top[22];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2367; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
            L2367:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[5] = r1;
            r1 = ((object*) lenv) [5];
            r2 = (object *) 3;
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            r0[6] = r1;
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 5);
            r2[1] = lenv; lenv = r2;
            r3 = allocate (TC_TUPLE, 2);
            r4 = ((object***) lenv) [1][1][4];
            r3[2] = r4;
            r4 = top[56];
            t = allocate (TC_SAVE, 6);
            t[1] = k; t[2] = lenv; t[3] = &&L2368; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
            r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
            L2368:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
            r3 = result;
            r2[2] = r3;
            r3 = (object*)TC_NIL;
            r2[3] = r3;
            r3 = ((object*) lenv) [3];
            r4 = (object*) &constructed_430;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            r2[4] = r3;
            r3 = ((object*) lenv) [4];
            r4 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            r2[5] = r3;
            r2 = ((object*) lenv) [5];
            r3 = (object*) &constructed_429;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            lenv = lenv[1];
            r1[2] = r2;
            r2 = top[47];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2369; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
            L2369:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[2] = r1;
            r1 = ((object*****) lenv) [1][1][1][1][6];
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            ((object****) lenv) [1][1][1][6] = r0;
            r0 = allocate (TC_TUPLE, 3);
            // def lambda_33178_i4681_i4727
            goto L2370;
            FUN_lambda_33178_i4681_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object******) lenv) [1][1][1][1][1][6];
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r1 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r1;
              t[2] = r0;
              r0 = t;
              lenv = lenv[1];
              ((object*****) lenv) [1][1][1][1][6] = r0;
              PXLL_RETURN(0);
            L2370:
            r1 = allocate (TC_CLOSURE, 2);
            r1[1] = &&FUN_lambda_33178_i4681_i4727; r1[2] = lenv;
            r0[2] = r1;
            r1 = ((object*) lenv) [3];
            r0[3] = r1;
            r1 = top[35];
            t = allocate (TC_SAVE, 3);
            t[1] = k; t[2] = lenv; t[3] = &&L2371; ; k = t;
            r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
            L2371:
            ; lenv = k[2]; k = k[1];
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 8);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*****) lenv) [1][1][1][1][7];
            r2[2] = r3;
            r3 = top[56];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2372; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
            L2372:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][6];
            r2[2] = r3;
            r3 = top[56];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2373; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
            L2373:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = (object*)TC_NIL;
            r1[4] = r2;
            r2 = ((object*) lenv) [4];
            r3 = (object*) &constructed_428;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[5] = r2;
            r2 = ((object*) lenv) [5];
            r3 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[6] = r2;
            r2 = ((object*) lenv) [6];
            r3 = (object*) &constructed_427;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[7] = r2;
            r2 = ((object*) lenv) [7];
            r3 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[8] = r2;
            r1 = ((object*) lenv) [8];
            r2 = (object*) &constructed_426;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            lenv = lenv[1];
            r0[2] = r1;
            r1 = top[47];
            r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
          } break;
          case (UOTAG(4)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 4);
            r2 = ((object*******) lenv) [1][1][1][1][1][1][2];
            r2 = ((pxll_vector*)r2)->val[14];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = (object*) &constructed_433;
            r1[4] = r2;
            r2 = top[66];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2374; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_alist_2fget_396;
            L2374:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r1 = allocate (TC_TUPLE, 5);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r3 = top[56];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2375; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
            L2375:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = (object*)TC_NIL;
            r1[3] = r2;
            r2 = ((object*) lenv) [3];
            r3 = (object*) &constructed_432;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[4] = r2;
            r2 = ((object*) lenv) [4];
            r3 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r3;
            t[2] = r2;
            r2 = t;
            r1[5] = r2;
            r1 = ((object*) lenv) [5];
            r2 = (object*) &constructed_431;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r2;
            t[2] = r1;
            r1 = t;
            lenv = lenv[1];
            r0[2] = r1;
            r1 = top[47];
            r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
          } break;
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 3);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,0);
            r0[2] = r1;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object****) lenv) [1][1][1][5];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = top[65];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2376; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
            L2376:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            r0[3] = r1;
            r0 = ((object*) lenv) [3];
            switch (get_case (r0)) {
              case (UOTAG(0)): {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object**) lenv) [1][3];
                r1 = UOBJ_GET(r1,0);
                r0[2] = r1;
                r0 = allocate (TC_TUPLE, 2);
                r1 = allocate (TC_TUPLE, 5);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object**) lenv) [1][2];
                r2[2] = r3;
                r3 = top[56];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2377; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
                L2377:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = (object*)TC_NIL;
                r1[3] = r2;
                r2 = ((object*) lenv) [3];
                r3 = (object*) &constructed_435;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                r1[4] = r2;
                r2 = ((object*) lenv) [4];
                r3 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                r1[5] = r2;
                r1 = ((object*) lenv) [5];
                r2 = (object*) &constructed_434;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r2;
                t[2] = r1;
                r1 = t;
                lenv = lenv[1];
                r0[2] = r1;
                r1 = top[47];
                r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
              } break;
              default: {
                r0 = (object*) &constructed_68;
                r0 = box((pxll_int)((pxll_string*)(r0))->len);
                r1 = (object*) &constructed_68;
                (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
                r0 = (object*) &constructed_436;
                dump_object (r0, 0); fprintf (stdout, "\n");
                r0 = (object*) &constructed_436;
                PXLL_UNDEFINED; result=r0; goto Lreturn;
                PXLL_RETURN(0);
              } break;
            }
            PXLL_RETURN(0);
          } break;
          default: {
            goto fatbar_44;
          }
        }
        PXLL_RETURN(0);
        goto fatbar_44_over;
        fatbar_44:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object*) lenv) [2];
        r1[2] = r2;
        r2 = top[147];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2378; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_encode_immediate_558;
        L2378:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = top[56];
        r0[1] = r1[2]; lenv = r0; goto FUN_int__3estring_364;
        fatbar_44_over: ;
        PXLL_RETURN(0);
      L2354:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_walk_4622_i4681_i4727; r1[2] = lenv;
      r0[4] = r1;
      r0 = ((object*****) lenv) [1][1][1][1][54];
      r0 = ((pxll_vector*)r0)->val[2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2379; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L2379:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object***) lenv) [1][1][4];
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def $loop_4651_i4681_i4727
      goto L2380;
      FUN__24loop_4651_i4681_i4727:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][2];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } else {
          r0 = (object *) constructed_421[0];
          ((object*****) lenv) [1][1][1][1][6] = r0;
          r0 = ((object*) lenv) [2];
          ((object*****) lenv) [1][1][1][1][7] = r0;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object******) lenv) [1][1][1][1][1][3];
          r1[2] = r2;
          r2 = ((object**) lenv) [1][2];
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2381; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2381:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(0)): {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object*) lenv) [2];
              r1 = box((pxll_int)((pxll_string*)(r1))->len);
              r0[3] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object********) lenv) [1][1][1][1][1][1][1][5];
              r0[2] = r1;
              r0 = ((object****) lenv) [1][1][1][2];
              r1 = ((object*) lenv) [2];
              r2 = ((object**) lenv) [1][2];
              t = alloc_no_clear (UOTAG(1),3);
              t[1] = r2;
              t[2] = r0;
              t[3] = r1;
              r0 = t;
              lenv = lenv[1];
              ((object*******) lenv) [1][1][1][1][1][1][5] = r0;
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 14);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object****) lenv) [1][1][1][2];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2382; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2382:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object**) lenv) [1][3];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2383; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2383:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object**) lenv) [1][3];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2384; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2384:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[4] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r3[1] = lenv; lenv = r3;
              // def loop_4666_i4483_i4681_i4727
              goto L2385;
              FUN_loop_4666_i4483_i4681_i4727:
                check_heap (0);
                r0 = ((object*) lenv) [3];
                switch (get_case (r0)) {
                  case (TC_NIL): {
                    r0 = allocate (TC_TUPLE, 2);
                    r1 = allocate (TC_TUPLE, 3);
                    r2 = ((object*) lenv) [2];
                    r1[2] = r2;
                    r2 = (object *) constructed_36[0];
                    r1[3] = r2;
                    r2 = top[21];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L2386; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
                    L2386:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    r0[2] = r1;
                    r1 = top[47];
                    r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                  } break;
                  default: {
                    r0 = allocate (TC_TUPLE, 3);
                    r0[1] = lenv; lenv = r0;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,0);
                    r0[2] = r1;
                    r1 = ((object**) lenv) [1][3];
                    r1 = UOBJ_GET(r1,1);
                    r0[3] = r1;
                    r0 = allocate (TC_TUPLE, 2);
                    r0[1] = lenv; lenv = r0;
                    r1 = (object *) 8706;
                    r2 = ((object**) lenv) [1][2];
                    if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                      r1 = (object*) &constructed_401;
                    } else {
                      goto fatbar_36;
                    }
                    goto fatbar_36_over;
                    fatbar_36:
                    r1 = (object *) 23554;
                    r2 = ((object**) lenv) [1][2];
                    if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                      r1 = (object*) &constructed_402;
                    } else {
                      goto fatbar_37;
                    }
                    goto fatbar_37_over;
                    fatbar_37:
                    r1 = (object *) 2306;
                    r2 = ((object**) lenv) [1][2];
                    if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                      r1 = (object*) &constructed_403;
                    } else {
                      goto fatbar_38;
                    }
                    goto fatbar_38_over;
                    fatbar_38:
                    r1 = (object *) 2562;
                    r2 = ((object**) lenv) [1][2];
                    if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                      r1 = (object*) &constructed_404;
                    } else {
                      goto fatbar_39;
                    }
                    goto fatbar_39_over;
                    fatbar_39:
                    r1 = (object *) 3330;
                    r2 = ((object**) lenv) [1][2];
                    if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                      r1 = (object*) &constructed_405;
                    } else {
                      goto fatbar_40;
                    }
                    goto fatbar_40_over;
                    fatbar_40:
                    r1 = allocate (TC_TUPLE, 2);
                    r2 = ((object**) lenv) [1][2];
                    r1[2] = r2;
                    r2 = top[205];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L2387; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto *r2[1];
                    L2387:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                    if PXLL_IS_TRUE(r1) {
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object**) lenv) [1][2];
                      r1[2] = r2;
                      r2 = top[46];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2388; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_char__3estring_348;
                      L2388:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                    } else {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object***) lenv) [1][1][2];
                      r2 = box((pxll_int)GET_CHAR(r2));
                      r1[2] = r2;
                      r2 = allocate (TC_TUPLE, 2);
                      r2[1] = lenv; lenv = r2;
                      // def loop_1044_i4492_i4681_i4727
                      goto L2389;
                      FUN_loop_1044_i4492_i4681_i4727:
                        check_heap (0);
                        r0 = ((object*) lenv) [2];
                        r1 = (object *) 1;
                        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                          r0 = allocate (TC_TUPLE, 2);
                          r1 = (object *) 1;
                          r2 = ((object***) lenv) [1][1][2];
                          if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)<unbox(r1))) {
                            r1 = ((object*) lenv) [3];
                            r2 = (object *) 11522;
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r2;
                            t[2] = r1;
                            r1 = t;
                          } else {
                            r1 = ((object*) lenv) [3];
                          }
                          r0[2] = r1;
                          r1 = top[54];
                          r0[1] = r1[2]; lenv = r0; goto FUN_list__3estring_361;
                        } else {
                          r0 = allocate (TC_TUPLE, 2);
                          r0[1] = lenv; lenv = r0;
                          r1 = (object *) 15;
                          r2 = ((object**) lenv) [1][2];
                          r1 = box((pxll_int)unbox(r2)&unbox(r1));
                          r2 = top[159];
                          range_check (GET_TUPLE_LENGTH(*(object*)r2), unbox(r1));
                          r1 = ((pxll_vector*)r2)->val[unbox(r1)];
                          r0[2] = r1;
                          r0 = ((object**) lenv) [1][3];
                          r1 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r1;
                          t[2] = r0;
                          r0 = t;
                          lenv = lenv[1];
                          r1 = (object *) 7;
                          r2 = ((object*) lenv) [2];
                          r1 = box((pxll_int)unbox(r2)>>unbox(r1));
                          lenv[2] = r1;
                          lenv[3] = r0;
                          goto FUN_loop_1044_i4492_i4681_i4727;
                        }
                        PXLL_RETURN(0);
                      L2389:
                      r3 = allocate (TC_CLOSURE, 2);
                      r3[1] = &&FUN_loop_1044_i4492_i4681_i4727; r3[2] = lenv;
                      r2[2] = r3;
                      r2 = allocate (TC_TUPLE, 3);
                      r3 = (object *) 1;
                      r4 = ((object**) lenv) [1][2];
                      if PXLL_IS_TRUE(PXLL_TEST(unbox(r4)<unbox(r3))) {
                        r3 = ((object**) lenv) [1][2];
                        r4 = (object *) 1;
                        r3 = box((pxll_int)unbox(r4)-unbox(r3));
                      } else {
                        r3 = ((object**) lenv) [1][2];
                      }
                      r2[2] = r3;
                      r3 = (object *) constructed_410[0];
                      r2[3] = r3;
                      r3 = ((object*) lenv) [2];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2390; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_loop_1044_i4492_i4681_i4727;
                      L2390:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      lenv = lenv[1];
                      r1[3] = r2;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = allocate (TC_TUPLE, 5);
                      r2[1] = lenv; lenv = r2;
                      r3 = ((object**) lenv) [1][3];
                      r3 = box((pxll_int)((pxll_string*)(r3))->len);
                      r2[2] = r3;
                      r3 = (object *) 5;
                      r4 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r4==r3)) {
                        r3 = (object*) &constructed_406;
                      } else {
                        goto fatbar_41;
                      }
                      goto fatbar_41_over;
                      fatbar_41:
                      r3 = (object *) 3;
                      r4 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r4==r3)) {
                        r3 = (object*) &constructed_407;
                      } else {
                        goto fatbar_42;
                      }
                      goto fatbar_42_over;
                      fatbar_42:
                      r3 = (object *) 1;
                      r4 = ((object*) lenv) [2];
                      if PXLL_IS_TRUE(PXLL_TEST(r4==r3)) {
                        r3 = (object*) &constructed_408;
                      } else {
                        goto fatbar_43;
                      }
                      goto fatbar_43_over;
                      fatbar_43:
                      r3 = allocate (TC_TUPLE, 3);
                      r4 = (object*) &constructed_409;
                      r3[2] = r4;
                      r4 = ((object****) lenv) [1][1][1][2];
                      r3[3] = r4;
                      r4 = top[15];
                      t = allocate (TC_SAVE, 6);
                      t[1] = k; t[2] = lenv; t[3] = &&L2391; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                      r3[1] = r4[2]; lenv = r3; goto FUN_error1_293;
                      L2391:
                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                      r3 = result;
                      fatbar_43_over: ;
                      fatbar_42_over: ;
                      fatbar_41_over: ;
                      r2[3] = r3;
                      r3 = (object*)TC_NIL;
                      r2[4] = r3;
                      r3 = ((object*) lenv) [4];
                      r4 = ((object**) lenv) [1][3];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r4;
                      t[2] = r3;
                      r3 = t;
                      r2[5] = r3;
                      r2 = ((object*) lenv) [5];
                      r3 = ((object*) lenv) [3];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r3;
                      t[2] = r2;
                      r2 = t;
                      lenv = lenv[1];
                      r1[2] = r2;
                      r2 = top[47];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2392; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                      L2392:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      lenv = lenv[1];
                    }
                    fatbar_40_over: ;
                    fatbar_39_over: ;
                    fatbar_38_over: ;
                    fatbar_37_over: ;
                    fatbar_36_over: ;
                    r0[2] = r1;
                    r0 = ((object***) lenv) [1][1][2];
                    r1 = ((object*) lenv) [2];
                    t = alloc_no_clear (TC_PAIR,2);
                    t[1] = r1;
                    t[2] = r0;
                    r0 = t;
                    lenv = lenv[1];
                    r1 = ((object*) lenv) [3];
                    lenv = ((object *)lenv)[1];
                    lenv[2] = r0;
                    lenv[3] = r1;
                    goto FUN_loop_4666_i4483_i4681_i4727;
                  } break;
                }
                PXLL_RETURN(0);
              L2385:
              r4 = allocate (TC_CLOSURE, 2);
              r4[1] = &&FUN_loop_4666_i4483_i4681_i4727; r4[2] = lenv;
              r3[2] = r4;
              r3 = allocate (TC_TUPLE, 3);
              r4 = (object *) constructed_400[0];
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][2];
              r4[2] = r5;
              r5 = top[55];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2393; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string__3elist_362;
              L2393:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = ((object*) lenv) [2];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2394; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_loop_4666_i4483_i4681_i4727;
              L2394:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              lenv = lenv[1];
              r2[5] = r3;
              r3 = (object*)TC_NIL;
              r2[6] = r3;
              r3 = ((object*) lenv) [6];
              r4 = (object*) &constructed_399;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[7] = r3;
              r3 = ((object*) lenv) [7];
              r4 = ((object*) lenv) [5];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[8] = r3;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_398;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[9] = r3;
              r3 = ((object*) lenv) [9];
              r4 = ((object*) lenv) [4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[10] = r3;
              r3 = ((object*) lenv) [10];
              r4 = (object*) &constructed_397;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[11] = r3;
              r3 = ((object*) lenv) [11];
              r4 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[12] = r3;
              r3 = ((object*) lenv) [12];
              r4 = (object*) &constructed_396;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[13] = r3;
              r3 = ((object*) lenv) [13];
              r4 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[14] = r3;
              r2 = ((object*) lenv) [14];
              r3 = (object*) &constructed_395;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[2] = r2;
              r2 = top[47];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2395; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
              L2395:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
              r1 = ((pxll_vector*)r1)->val[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L2396; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L2396:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              lenv = lenv[1];
            } break;
            case (UOTAG(4)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 4);
              r2[1] = lenv; lenv = r2;
              r3 = ((object**) lenv) [1][2];
              r3 = allocate (TC_TUPLE, 2);
              r3[1] = lenv; lenv = r3;
              r4 = ((object***) lenv) [1][1][2];
              r4 = UOBJ_GET(r4,0);
              r3[2] = r4;
              r3 = ((object*) lenv) [2];
              lenv = lenv[1];
              r2[2] = r3;
              r3 = (object*)TC_NIL;
              r2[3] = r3;
              r3 = ((object*) lenv) [3];
              r4 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[4] = r3;
              r2 = ((object*) lenv) [4];
              r3 = (object*) &constructed_416;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[2] = r2;
              r2 = top[47];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2397; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
              L2397:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
              r1 = ((pxll_vector*)r1)->val[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L2398; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L2398:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 14);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object****) lenv) [1][1][1][2];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2399; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2399:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object****) lenv) [1][1][1][2];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2400; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2400:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r4 = allocate (TC_TUPLE, 2);
              r4[1] = lenv; lenv = r4;
              r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][7];
              r4[2] = r5;
              r4 = (object *) 3;
              r5 = ((object*) lenv) [2];
              r4 = box((pxll_int)unbox(r5)-unbox(r4));
              lenv = lenv[1];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2401; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2401:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[4] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object********) lenv) [1][1][1][1][1][1][1][8];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2402; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2402:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[5] = r3;
              r3 = (object*)TC_NIL;
              r2[6] = r3;
              r3 = ((object*) lenv) [6];
              r4 = (object*) &constructed_415;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[7] = r3;
              r3 = ((object*) lenv) [7];
              r4 = ((object*) lenv) [5];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[8] = r3;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_414;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[9] = r3;
              r3 = ((object*) lenv) [9];
              r4 = ((object*) lenv) [4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[10] = r3;
              r3 = ((object*) lenv) [10];
              r4 = (object*) &constructed_413;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[11] = r3;
              r3 = ((object*) lenv) [11];
              r4 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[12] = r3;
              r3 = ((object*) lenv) [12];
              r4 = (object*) &constructed_412;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[13] = r3;
              r3 = ((object*) lenv) [13];
              r4 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[14] = r3;
              r2 = ((object*) lenv) [14];
              r3 = (object*) &constructed_411;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[2] = r2;
              r2 = top[47];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2403; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
              L2403:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
              r1 = ((pxll_vector*)r1)->val[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L2404; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L2404:
              ; lenv = k[2]; k = k[1];
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object********) lenv) [1][1][1][1][1][1][1][8];
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              r1 = (object *) 3;
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv = lenv[1];
              ((object*******) lenv) [1][1][1][1][1][1][8] = r0;
              lenv = lenv[1];
            } break;
            default: {
              goto fatbar_35;
            }
          }
          goto fatbar_35_over;
          fatbar_35:
          r0 = allocate (TC_TUPLE, 5);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 3);
          r3 = ((object*******) lenv) [1][1][1][1][1][1][3];
          r2[2] = r3;
          r3 = ((object***) lenv) [1][1][2];
          r2[3] = r3;
          r3 = top[27];
          t = allocate (TC_SAVE, 5);
          t[1] = k; t[2] = lenv; t[3] = &&L2405; t[4] = r0; t[5] = r1; k = t;
          r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
          L2405:
          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
          r2 = result;
          r1[2] = r2;
          r2 = ((object******) lenv) [1][1][1][1][1][4];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2406; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_walk_4622_i4681_i4727;
          L2406:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][6];
          r0[3] = r1;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object*) lenv) [3];
          r1[2] = r2;
          r2 = (object *) constructed_36[0];
          r1[3] = r2;
          r2 = top[21];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2407; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
          L2407:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[4] = r1;
          r1 = ((object*) lenv) [4];
          r2 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r2;
          t[2] = r1;
          r1 = t;
          r0[5] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 8);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object****) lenv) [1][1][1][2];
          r3[2] = r4;
          r4 = top[56];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L2408; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
          L2408:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = allocate (TC_TUPLE, 3);
          r4 = allocate (TC_TUPLE, 3);
          r5 = top[16];
          r4[2] = r5;
          r5 = ((object**) lenv) [1][5];
          r4[3] = r5;
          r5 = top[32];
          t = allocate (TC_SAVE, 7);
          t[1] = k; t[2] = lenv; t[3] = &&L2409; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
          r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
          L2409:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
          r4 = result;
          r3[2] = r4;
          r4 = (object*) &constructed_420;
          r3[3] = r4;
          r4 = top[48];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L2410; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
          L2410:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = (object*)TC_NIL;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_419;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = (object*) &constructed_418;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r3 = ((object*) lenv) [7];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[8] = r3;
          r2 = ((object*) lenv) [8];
          r3 = (object*) &constructed_417;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2411; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L2411:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
          r1 = ((pxll_vector*)r1)->val[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2412; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L2412:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          fatbar_35_over: ;
          lenv = lenv[1];
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv[2] = r0;
          goto FUN__24loop_4651_i4681_i4727;
        }
        PXLL_RETURN(0);
      L2380:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN__24loop_4651_i4681_i4727; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2413; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN__24loop_4651_i4681_i4727;
      L2413:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = (object *) constructed_394[0];
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_33521_i4681_i4727
      goto L2414;
      FUN_lambda_33521_i4681_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 3);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 5);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object***) lenv) [1][1][3];
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2415; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2415:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = (object*)TC_NIL;
        r2[3] = r3;
        r3 = ((object*) lenv) [3];
        r4 = (object*) &constructed_393;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[4] = r3;
        r3 = ((object*) lenv) [4];
        r4 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[5] = r3;
        r2 = ((object*) lenv) [5];
        r3 = (object*) &constructed_392;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        lenv = lenv[1];
        r1[2] = r2;
        r2 = top[47];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2416; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
        L2416:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object***) lenv) [1][1][2];
        r0[3] = r1;
        r0 = ((object*) lenv) [3];
        r1 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r1;
        t[2] = r0;
        r0 = t;
        lenv = lenv[1];
        ((object**) lenv) [1][2] = r0;
        PXLL_RETURN(0);
      L2414:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_33521_i4681_i4727; r1[2] = lenv;
      r0[2] = r1;
      r1 = ((object******) lenv) [1][1][1][1][1][2];
      r1 = ((pxll_vector*)r1)->val[14];
      r0[3] = r1;
      r1 = top[67];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2417; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_alist_2fiterate_397;
      L2417:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = allocate (TC_TUPLE, 2);
      r2 = allocate (TC_TUPLE, 8);
      r2[1] = lenv; lenv = r2;
      r3 = allocate (TC_TUPLE, 2);
      r4 = allocate (TC_TUPLE, 2);
      r5 = ((object**) lenv) [1][2];
      r4[2] = r5;
      r5 = top[22];
      t = allocate (TC_SAVE, 7);
      t[1] = k; t[2] = lenv; t[3] = &&L2418; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
      r4[1] = r5[2]; lenv = r4; goto FUN_length_314;
      L2418:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
      r4 = result;
      r3[2] = r4;
      r4 = top[56];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L2419; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
      L2419:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = allocate (TC_TUPLE, 3);
      r4 = allocate (TC_TUPLE, 3);
      r5 = top[16];
      r4[2] = r5;
      r5 = ((object**) lenv) [1][2];
      r4[3] = r5;
      r5 = top[32];
      t = allocate (TC_SAVE, 7);
      t[1] = k; t[2] = lenv; t[3] = &&L2420; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
      r4[1] = r5[2]; lenv = r4; goto FUN_map_326;
      L2420:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
      r4 = result;
      r3[2] = r4;
      r4 = (object*) &constructed_391;
      r3[3] = r4;
      r4 = top[48];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L2421; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_string_join_353;
      L2421:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[3] = r3;
      r3 = (object*)TC_NIL;
      r2[4] = r3;
      r3 = ((object*) lenv) [4];
      r4 = (object*) &constructed_390;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[5] = r3;
      r3 = ((object*) lenv) [5];
      r4 = ((object*) lenv) [3];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[6] = r3;
      r3 = ((object*) lenv) [6];
      r4 = (object*) &constructed_389;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[7] = r3;
      r3 = ((object*) lenv) [7];
      r4 = ((object*) lenv) [2];
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r4;
      t[2] = r3;
      r3 = t;
      r2[8] = r3;
      r2 = ((object*) lenv) [8];
      r3 = (object*) &constructed_388;
      t = alloc_no_clear (TC_PAIR,2);
      t[1] = r3;
      t[2] = r2;
      r2 = t;
      lenv = lenv[1];
      r1[2] = r2;
      r2 = top[47];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2422; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
      L2422:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object******) lenv) [1][1][1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2423; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2423:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      r0 = ((object*****) lenv) [1][1][1][1][54];
      r0 = ((pxll_vector*)r0)->val[3];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2424; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L2424:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [3];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][54];
      r1 = ((pxll_vector*)r1)->val[1];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2425; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2425:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 3);
      r0[1] = lenv; lenv = r0;
      r1 = ((object****) lenv) [1][1][1][2];
      r1 = ((pxll_vector*)r1)->val[11];
      r1 = ((pxll_vector*)r1)->val[1];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2426; t[4] = r0; k = t;
      lenv = r1[2]; goto *r1[1];
      L2426:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      r2 = (object *) 3;
      r1 = box((pxll_int)unbox(r2)+unbox(r1));
      r0[3] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      r1 = ((object**) lenv) [1][3];
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r0[1] = lenv; lenv = r0;
      // def $loop_4608_i4682_i4727
      goto L2427;
      FUN__24loop_4608_i4682_i4727:
        check_heap (0);
        r0 = ((object***) lenv) [1][1][2];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        } else {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 5);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][2];
          r3[2] = r4;
          r4 = top[56];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L2428; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
          L2428:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = (object*) &constructed_387;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r2 = ((object*) lenv) [5];
          r3 = (object*) &constructed_386;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2429; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L2429:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*******) lenv) [1][1][1][1][1][1][54];
          r1 = ((pxll_vector*)r1)->val[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2430; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L2430:
          ; lenv = k[2]; k = k[1];
          r0 = (object *) 3;
          r1 = ((object*) lenv) [2];
          r0 = box((pxll_int)unbox(r1)+unbox(r0));
          lenv[2] = r0;
          goto FUN__24loop_4608_i4682_i4727;
        }
        PXLL_RETURN(0);
      L2427:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN__24loop_4608_i4682_i4727; r1[2] = lenv;
      r0[2] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object *) 1;
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2431; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN__24loop_4608_i4682_i4727;
      L2431:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_385;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2432; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2432:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_384;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2433; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2433:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_32883_i4682_i4727
      goto L2434;
      FUN_lambda_32883_i4682_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 11);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r4 = (object *) 3;
        r5 = ((object**) lenv) [1][2];
        r4 = box((pxll_int)unbox(r5)+unbox(r4));
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2435; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2435:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = allocate (TC_TUPLE, 2);
        r4 = (object *) 7;
        r5 = ((object**) lenv) [1][2];
        r4 = box((pxll_int)unbox(r5)+unbox(r4));
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2436; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2436:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[3] = r3;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object**) lenv) [1][2];
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2437; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2437:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[4] = r3;
        r3 = (object*)TC_NIL;
        r2[5] = r3;
        r3 = ((object*) lenv) [5];
        r4 = (object*) &constructed_383;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[6] = r3;
        r3 = ((object*) lenv) [6];
        r4 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[7] = r3;
        r3 = ((object*) lenv) [7];
        r4 = (object*) &constructed_382;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[8] = r3;
        r3 = ((object*) lenv) [8];
        r4 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[9] = r3;
        r3 = ((object*) lenv) [9];
        r4 = (object*) &constructed_381;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[10] = r3;
        r3 = ((object*) lenv) [10];
        r4 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[11] = r3;
        r2 = ((object*) lenv) [11];
        r3 = (object*) &constructed_380;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        lenv = lenv[1];
        r1[2] = r2;
        r2 = top[47];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2438; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
        L2438:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][54];
        r1 = ((pxll_vector*)r1)->val[4];
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2434:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_32883_i4682_i4727; r1[2] = lenv;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object**) lenv) [1][3];
      r2[2] = r3;
      r3 = top[30];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L2439; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_range_324;
      L2439:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = (object *) constructed_36[0];
      r1[3] = r2;
      r2 = top[21];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2440; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
      L2440:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      lenv = lenv[1];
      r0[3] = r1;
      r1 = top[35];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2441; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      L2441:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_379;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2442; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2442:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_378;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2443; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2443:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_377;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2444; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2444:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 3);
      // def lambda_32944_i4682_i4727
      goto L2445;
      FUN_lambda_32944_i4682_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 11);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r4 = (object *) 3;
        r5 = ((object**) lenv) [1][2];
        r4 = box((pxll_int)unbox(r5)+unbox(r4));
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2446; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2446:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[2] = r3;
        r3 = allocate (TC_TUPLE, 2);
        r4 = ((object**) lenv) [1][2];
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2447; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2447:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[3] = r3;
        r3 = allocate (TC_TUPLE, 2);
        r4 = (object *) 7;
        r5 = ((object**) lenv) [1][2];
        r4 = box((pxll_int)unbox(r5)+unbox(r4));
        r3[2] = r4;
        r4 = top[56];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2448; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
        L2448:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        r2[4] = r3;
        r3 = (object*)TC_NIL;
        r2[5] = r3;
        r3 = ((object*) lenv) [5];
        r4 = (object*) &constructed_376;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[6] = r3;
        r3 = ((object*) lenv) [6];
        r4 = ((object*) lenv) [4];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[7] = r3;
        r3 = ((object*) lenv) [7];
        r4 = (object*) &constructed_375;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[8] = r3;
        r3 = ((object*) lenv) [8];
        r4 = ((object*) lenv) [3];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[9] = r3;
        r3 = ((object*) lenv) [9];
        r4 = (object*) &constructed_374;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[10] = r3;
        r3 = ((object*) lenv) [10];
        r4 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r4;
        t[2] = r3;
        r3 = t;
        r2[11] = r3;
        r2 = ((object*) lenv) [11];
        r3 = (object*) &constructed_373;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        lenv = lenv[1];
        r1[2] = r2;
        r2 = top[47];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2449; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
        L2449:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = ((object*****) lenv) [1][1][1][1][54];
        r1 = ((pxll_vector*)r1)->val[4];
        r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2445:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_lambda_32944_i4682_i4727; r1[2] = lenv;
      r0[2] = r1;
      r1 = allocate (TC_TUPLE, 2);
      r1[1] = lenv; lenv = r1;
      r2 = allocate (TC_TUPLE, 2);
      r3 = ((object**) lenv) [1][3];
      r2[2] = r3;
      r3 = top[30];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L2450; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_range_324;
      L2450:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[2] = r2;
      r1 = allocate (TC_TUPLE, 3);
      r2 = ((object*) lenv) [2];
      r1[2] = r2;
      r2 = (object *) constructed_36[0];
      r1[3] = r2;
      r2 = top[21];
      t = allocate (TC_SAVE, 4);
      t[1] = k; t[2] = lenv; t[3] = &&L2451; t[4] = r0; k = t;
      r1[1] = r2[2]; lenv = r1; goto FUN_reverse_onto_311;
      L2451:
      r0 = k[4]; lenv = k[2]; k = k[1];
      r1 = result;
      lenv = lenv[1];
      r0[3] = r1;
      r1 = top[35];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2452; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
      L2452:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_372;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2453; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2453:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object*) lenv) [4];
      r0[2] = r1;
      r1 = ((object***) lenv) [1][1][54];
      r1 = ((pxll_vector*)r1)->val[1];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2454; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2454:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 8);
      r0[1] = lenv; lenv = r0;
      // def emit_4226_i4683_i4727
      goto L2455;
      FUN_emit_4226_i4683_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = ((object**) lenv) [1][2];
        switch (get_case (r1)) {
          case (UOTAG(0)): {
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2456; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2456:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_74;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_73;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2457; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2457:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2458; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2458:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = (object*)UITAG(1);
            lenv = lenv[1];
          } break;
          case (UOTAG(1)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2459; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2459:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = top[147];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2460; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_encode_immediate_558;
            L2460:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = (object *) -1;
            r3 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r3)==unbox(r2))) {
              r2 = (object*) &constructed_77;
            } else {
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 4);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2461; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2461:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r3 = ((object*) lenv) [4];
              r4 = (object*) &constructed_78;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2462; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2462:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
            }
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 6);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][3];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2463; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2463:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_76;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r4 = ((object*) lenv) [5];
            r5 = (object*) &constructed_75;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[6] = r4;
            r3 = ((object*) lenv) [6];
            r4 = ((object**) lenv) [1][4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2464; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2464:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2465; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2465:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = ((object*) lenv) [3];
            lenv = lenv[1];
          } break;
          case (UOTAG(2)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2466; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2466:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = (object *) 1;
            r2 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
              r1 = (object *) constructed_86[0];
              r2 = ((object**) lenv) [1][3];
              if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 8);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2467; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2467:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object***) lenv) [1][1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2468; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2468:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[3] = r4;
                r4 = (object*)TC_NIL;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = (object*) &constructed_81;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r4 = ((object*) lenv) [5];
                r5 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[6] = r4;
                r4 = ((object*) lenv) [6];
                r5 = (object*) &constructed_80;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[7] = r4;
                r4 = ((object*) lenv) [7];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[8] = r4;
                r3 = ((object*) lenv) [8];
                r4 = (object*) &constructed_79;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2469; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2469:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
                r2 = ((pxll_vector*)r2)->val[4];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2470; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                L2470:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              } else {
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 8);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2471; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2471:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object***) lenv) [1][1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2472; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2472:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[3] = r4;
                r4 = (object*)TC_NIL;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = (object*) &constructed_84;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r4 = ((object*) lenv) [5];
                r5 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[6] = r4;
                r4 = ((object*) lenv) [6];
                r5 = (object*) &constructed_83;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[7] = r4;
                r4 = ((object*) lenv) [7];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[8] = r4;
                r3 = ((object*) lenv) [8];
                r4 = (object*) &constructed_82;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2473; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2473:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
                r2 = ((pxll_vector*)r2)->val[4];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2474; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                L2474:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              }
            } else {
              r1 = (object *) 14;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [4];
            lenv = lenv[1];
          } break;
          case (UOTAG(4)): {
            r1 = allocate (TC_TUPLE, 5);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2475; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2475:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_90;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_89;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2476; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2476:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2477; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2477:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r1 = ((pxll_vector*)r1)->val[3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2478; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L2478:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2479; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
            L2479:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r1 = ((pxll_vector*)r1)->val[2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2480; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L2480:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object*) &constructed_88;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2481; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2481:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r1 = ((pxll_vector*)r1)->val[3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2482; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L2482:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [4];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2483; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
            L2483:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r1 = ((pxll_vector*)r1)->val[2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2484; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L2484:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object*) &constructed_87;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2485; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2485:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*) lenv) [5];
            lenv = lenv[1];
          } break;
          case (UOTAG(5)): {
            r1 = allocate (TC_TUPLE, 7);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,4);
            r1[6] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,5);
            r1[7] = r2;
            r1 = ((object*) lenv) [3];
            switch (get_case (r1)) {
              case (UOTAG(1)): {
                r1 = allocate (TC_TUPLE, 3);
                r1[1] = lenv; lenv = r1;
                r2 = ((object**) lenv) [1][3];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r2 = ((object**) lenv) [1][3];
                r2 = UOBJ_GET(r2,1);
                r1[3] = r2;
                r1 = (object *) constructed_97[0];
                r2 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 6);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r2 = allocate (TC_TUPLE, 3);
                      // def lambda_29135_i3217_i4683_i4727
                      goto L2486;
                      FUN_lambda_29135_i3217_i4683_i4727:
                        check_heap (0);
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = allocate (TC_TUPLE, 4);
                        r1[1] = lenv; lenv = r1;
                        r2 = allocate (TC_TUPLE, 2);
                        r3 = ((object**) lenv) [1][2];
                        r2[2] = r3;
                        r3 = top[56];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2487; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
                        L2487:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                        r1[2] = r2;
                        r2 = (object*)TC_NIL;
                        r1[3] = r2;
                        r2 = ((object*) lenv) [3];
                        r3 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r3;
                        t[2] = r2;
                        r2 = t;
                        r1[4] = r2;
                        r1 = ((object*) lenv) [4];
                        r2 = (object*) &constructed_95;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r2;
                        t[2] = r1;
                        r1 = t;
                        lenv = lenv[1];
                        r0[2] = r1;
                        r1 = top[47];
                        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                      L2486:
                      r3 = allocate (TC_CLOSURE, 2);
                      r3[1] = &&FUN_lambda_29135_i3217_i4683_i4727; r3[2] = lenv;
                      r2[2] = r3;
                      r3 = ((object***) lenv) [1][1][2];
                      r2[3] = r3;
                      r3 = top[32];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2488; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_map_326;
                      L2488:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[4] = r2;
                      r2 = allocate (TC_TUPLE, 4);
                      r3 = top[148];
                      r2[2] = r3;
                      r3 = ((object*) lenv) [3];
                      r2[3] = r3;
                      r3 = ((object*) lenv) [4];
                      r2[4] = r3;
                      r3 = top[33];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2489; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_map2_327;
                      L2489:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[5] = r2;
                      r2 = allocate (TC_TUPLE, 3);
                      r3 = ((object*) lenv) [2];
                      r2[2] = r3;
                      r3 = allocate (TC_TUPLE, 3);
                      r4 = ((object***) lenv) [1][1][4];
                      r3[2] = r4;
                      r4 = ((object*) lenv) [5];
                      r3[3] = r4;
                      r4 = top[151];
                      t = allocate (TC_SAVE, 6);
                      t[1] = k; t[2] = lenv; t[3] = &&L2490; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                      r3[1] = r4[2]; lenv = r3; goto FUN_cexp_subst_562;
                      L2490:
                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                      r3 = result;
                      r2[3] = r3;
                      r3 = top[150];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2491; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_wrap_out_561;
                      L2491:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[6] = r2;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = allocate (TC_TUPLE, 2);
                      r3 = allocate (TC_TUPLE, 4);
                      r3[1] = lenv; lenv = r3;
                      r4 = (object*)TC_NIL;
                      r3[2] = r4;
                      r4 = ((object*) lenv) [2];
                      r5 = (object*) &constructed_94;
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r5;
                      t[2] = r4;
                      r4 = t;
                      r3[3] = r4;
                      r4 = ((object*) lenv) [3];
                      r5 = ((object**) lenv) [1][6];
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r5;
                      t[2] = r4;
                      r4 = t;
                      r3[4] = r4;
                      r3 = ((object*) lenv) [4];
                      r4 = (object*) &constructed_93;
                      t = alloc_no_clear (TC_PAIR,2);
                      t[1] = r4;
                      t[2] = r3;
                      r3 = t;
                      lenv = lenv[1];
                      r2[2] = r3;
                      r3 = top[47];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2492; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                      L2492:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[2] = r2;
                      r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r2 = ((pxll_vector*)r2)->val[4];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2493; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L2493:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[3];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2494; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2494:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object***) lenv) [1][1][5];
                      r1[2] = r2;
                      r2 = ((object******) lenv) [1][1][1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2495; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
                      L2495:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2496; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2496:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = (object*) &constructed_92;
                      r1[2] = r2;
                      r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r2 = ((pxll_vector*)r2)->val[4];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2497; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L2497:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[3];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2498; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2498:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object***) lenv) [1][1][6];
                      r1[2] = r2;
                      r2 = ((object******) lenv) [1][1][1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2499; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
                      L2499:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2500; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2500:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = (object*) &constructed_91;
                      r1[2] = r2;
                      r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                      r2 = ((pxll_vector*)r2)->val[4];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2501; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L2501:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_6;
                    }
                  }
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_6;
                }
                lenv = lenv[1];
              } break;
              default: {
                goto fatbar_6;
              }
            }
            goto fatbar_6_over;
            fatbar_6:
            r1 = (object*) &constructed_68;
            r1 = box((pxll_int)((pxll_string*)(r1))->len);
            r2 = (object*) &constructed_68;
            (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
            r1 = (object*) &constructed_67;
            dump_object (r1, 0); fprintf (stdout, "\n");
            r1 = (object*) &constructed_67;
            PXLL_UNDEFINED; result=r1; goto Lreturn;
            fatbar_6_over: ;
            r1 = ((object*) lenv) [7];
            lenv = lenv[1];
          } break;
          case (UOTAG(6)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object*) lenv) [2];
            r1[2] = r2;
            r2 = ((object*) lenv) [3];
            r1[3] = r2;
            r2 = ((object****) lenv) [1][1][1][3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2502; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_move_4227_i4683_i4727;
            L2502:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = (object*)UITAG(1);
            lenv = lenv[1];
          } break;
          case (UOTAG(3)): {
            r1 = allocate (TC_TUPLE, 6);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,4);
            r1[6] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][6];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2503; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2503:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            switch (get_case (r2)) {
              case (UOTAG(1)): {
                r2 = allocate (TC_TUPLE, 3);
                r2[1] = lenv; lenv = r2;
                r3 = ((object***) lenv) [1][1][2];
                r3 = UOBJ_GET(r3,0);
                r2[2] = r3;
                r3 = ((object***) lenv) [1][1][2];
                r3 = UOBJ_GET(r3,1);
                r2[3] = r3;
                r2 = (object *) constructed_97[0];
                r3 = ((object*) lenv) [2];
                if PXLL_IS_TRUE(PXLL_TEST(r3==r2)) {
                  r2 = ((object*) lenv) [3];
                  switch (get_case (r2)) {
                    case (TC_PAIR): {
                      r2 = allocate (TC_TUPLE, 5);
                      r2[1] = lenv; lenv = r2;
                      r3 = ((object**) lenv) [1][3];
                      r3 = UOBJ_GET(r3,0);
                      r2[2] = r3;
                      r3 = ((object**) lenv) [1][3];
                      r3 = UOBJ_GET(r3,1);
                      r2[3] = r3;
                      r3 = allocate (TC_TUPLE, 3);
                      // def lambda_29246_i3229_i4683_i4727
                      goto L2504;
                      FUN_lambda_29246_i3229_i4683_i4727:
                        check_heap (0);
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = allocate (TC_TUPLE, 4);
                        r1[1] = lenv; lenv = r1;
                        r2 = allocate (TC_TUPLE, 2);
                        r3 = ((object**) lenv) [1][2];
                        r2[2] = r3;
                        r3 = top[56];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2505; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
                        L2505:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                        r1[2] = r2;
                        r2 = (object*)TC_NIL;
                        r1[3] = r2;
                        r2 = ((object*) lenv) [3];
                        r3 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r3;
                        t[2] = r2;
                        r2 = t;
                        r1[4] = r2;
                        r1 = ((object*) lenv) [4];
                        r2 = (object*) &constructed_102;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r2;
                        t[2] = r1;
                        r1 = t;
                        lenv = lenv[1];
                        r0[2] = r1;
                        r1 = top[47];
                        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
                      L2504:
                      r4 = allocate (TC_CLOSURE, 2);
                      r4[1] = &&FUN_lambda_29246_i3229_i4683_i4727; r4[2] = lenv;
                      r3[2] = r4;
                      r4 = ((object****) lenv) [1][1][1][5];
                      r3[3] = r4;
                      r4 = top[32];
                      t = allocate (TC_SAVE, 6);
                      t[1] = k; t[2] = lenv; t[3] = &&L2506; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                      r3[1] = r4[2]; lenv = r3; goto FUN_map_326;
                      L2506:
                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                      r3 = result;
                      r2[4] = r3;
                      r3 = allocate (TC_TUPLE, 4);
                      r4 = top[148];
                      r3[2] = r4;
                      r4 = ((object*) lenv) [3];
                      r3[3] = r4;
                      r4 = ((object*) lenv) [4];
                      r3[4] = r4;
                      r4 = top[33];
                      t = allocate (TC_SAVE, 6);
                      t[1] = k; t[2] = lenv; t[3] = &&L2507; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                      r3[1] = r4[2]; lenv = r3; goto FUN_map2_327;
                      L2507:
                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                      r3 = result;
                      r2[5] = r3;
                      r2 = allocate (TC_TUPLE, 3);
                      r3 = ((object****) lenv) [1][1][1][3];
                      r2[2] = r3;
                      r3 = allocate (TC_TUPLE, 3);
                      r4 = ((object****) lenv) [1][1][1][4];
                      r3[2] = r4;
                      r4 = ((object*) lenv) [5];
                      r3[3] = r4;
                      r4 = top[151];
                      t = allocate (TC_SAVE, 6);
                      t[1] = k; t[2] = lenv; t[3] = &&L2508; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                      r3[1] = r4[2]; lenv = r3; goto FUN_cexp_subst_562;
                      L2508:
                      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                      r3 = result;
                      r2[3] = r3;
                      r3 = top[150];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2509; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto FUN_wrap_out_561;
                      L2509:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_7;
                    }
                  }
                } else {
                  lenv = ((object *)lenv)[1];
                  goto fatbar_7;
                }
                lenv = lenv[1];
              } break;
              default: {
                goto fatbar_7;
              }
            }
            goto fatbar_7_over;
            fatbar_7:
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = ((object**) lenv) [1][4];
            r2[3] = r3;
            r3 = top[150];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2510; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_wrap_out_561;
            L2510:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            fatbar_7_over: ;
            r1[3] = r2;
            r1 = (object *) -1;
            r2 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 3);
              r3[1] = lenv; lenv = r3;
              r4 = (object*)TC_NIL;
              r3[2] = r4;
              r4 = ((object*) lenv) [2];
              r5 = (object*) &constructed_98;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[3] = r4;
              r3 = ((object*) lenv) [3];
              r4 = ((object**) lenv) [1][3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2511; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2511:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2512; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2512:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 7);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2513; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2513:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = (object*) &constructed_101;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = ((object**) lenv) [1][3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = (object*) &constructed_100;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r3 = ((object*) lenv) [7];
              r4 = (object*) &constructed_99;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2514; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2514:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2515; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2515:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [6];
            lenv = lenv[1];
          } break;
          case (UOTAG(7)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2516; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2516:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][2];
            r2[2] = r3;
            r3 = ((object*****) lenv) [1][1][1][1][4];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2517; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_gen_function_label_4234_i4683_i4727;
            L2517:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = top[202];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2518; t[4] = r0; t[5] = r1; k = t;
            lenv = r2[2]; goto *r2[1];
            L2518:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 4);
            r3[1] = lenv; lenv = r3;
            r4 = ((object***) lenv) [1][1][2];
            r4 = allocate (TC_TUPLE, 2);
            r4[1] = lenv; lenv = r4;
            r5 = ((object****) lenv) [1][1][1][2];
            r5 = UOBJ_GET(r5,0);
            r4[2] = r5;
            r4 = ((object*) lenv) [2];
            lenv = lenv[1];
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r3 = ((object*) lenv) [4];
            r4 = (object*) &constructed_116;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2519; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2519:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2520; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2520:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 4);
            r3[1] = lenv; lenv = r3;
            r4 = (object*)TC_NIL;
            r3[2] = r4;
            r4 = ((object*) lenv) [2];
            r5 = (object*) &constructed_115;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = ((object**) lenv) [1][4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r3 = ((object*) lenv) [4];
            r4 = (object*) &constructed_114;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2521; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2521:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2522; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2522:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 3);
            r3[1] = lenv; lenv = r3;
            r4 = (object*)TC_NIL;
            r3[2] = r4;
            r4 = ((object*) lenv) [2];
            r5 = (object*) &constructed_113;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[3] = r4;
            r3 = ((object*) lenv) [3];
            r4 = ((object**) lenv) [1][3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2523; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2523:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2524; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2524:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r1 = ((pxll_vector*)r1)->val[3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2525; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L2525:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r1 = ((pxll_vector*)r1)->val[9];
            r1 = ((pxll_vector*)r1)->val[3];
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 4);
              r3[1] = lenv; lenv = r3;
              r4 = (object*)TC_NIL;
              r3[2] = r4;
              r4 = ((object*) lenv) [2];
              r5 = (object*) &constructed_112;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = ((object**) lenv) [1][3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r3 = ((object*) lenv) [4];
              r4 = (object*) &constructed_111;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2526; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2526:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2527; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2527:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 14;
            }
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
            r2[2] = r3;
            r3 = ((object***) lenv) [1][1][2];
            r2[3] = r3;
            r3 = top[5];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2528; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_vars_get_var_246;
            L2528:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r2 = ((pxll_vector*)r2)->val[lookup_field((GET_TYPECODE(*r2)-TC_USEROBJ)>>2,72)];
            r1[3] = r2;
            r1 = top[156];
            r2 = ((object*) lenv) [3];
            r1 = PXLL_TEST((unbox(r2)&(1<<unbox(r1)))>0);
            lenv = lenv[1];
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = (object*) &constructed_110;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2529; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2529:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 14;
            }
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][3];
            r1[2] = r2;
            r2 = ((object*****) lenv) [1][1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2530; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
            L2530:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r1 = ((pxll_vector*)r1)->val[2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2531; t[4] = r0; k = t;
            lenv = r1[2]; goto *r1[1];
            L2531:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 3);
            r3[1] = lenv; lenv = r3;
            r4 = (object*)TC_NIL;
            r3[2] = r4;
            r4 = ((object*) lenv) [2];
            r5 = (object*) &constructed_109;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[3] = r4;
            r3 = ((object*) lenv) [3];
            r4 = ((object**) lenv) [1][4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2532; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2532:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2533; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2533:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2534; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2534:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_108;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_107;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2535; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2535:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2536; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2536:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 10);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2537; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2537:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2538; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2538:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[3] = r4;
            r4 = (object*)TC_NIL;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = (object*) &constructed_106;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r4 = ((object*) lenv) [5];
            r5 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[6] = r4;
            r4 = ((object*) lenv) [6];
            r5 = (object*) &constructed_105;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[7] = r4;
            r4 = ((object*) lenv) [7];
            r5 = ((object**) lenv) [1][3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[8] = r4;
            r4 = ((object*) lenv) [8];
            r5 = (object*) &constructed_104;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[9] = r4;
            r4 = ((object*) lenv) [9];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[10] = r4;
            r3 = ((object*) lenv) [10];
            r4 = (object*) &constructed_103;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2539; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2539:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2540; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2540:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = ((object*) lenv) [4];
            lenv = lenv[1];
          } break;
          case (UOTAG(8)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2541; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2541:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = (object *) 1;
            r2 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = (object *) -1;
              r3 = ((object***) lenv) [1][1][2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r3)==unbox(r2))) {
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 5);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object****) lenv) [1][1][1][3];
                r6 = (object *) 5;
                r5 = box((pxll_int)unbox(r6)+unbox(r5));
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2542; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2542:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = (object*)TC_NIL;
                r3[3] = r4;
                r4 = ((object*) lenv) [3];
                r5 = (object*) &constructed_120;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r3 = ((object*) lenv) [5];
                r4 = (object*) &constructed_119;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2543; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2543:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
              } else {
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 11);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = allocate (TC_TUPLE, 3);
                r6 = ((object****) lenv) [1][1][1][2];
                r5[2] = r6;
                r6 = (object*) &constructed_126;
                r5[3] = r6;
                r6 = top[31];
                t = allocate (TC_SAVE, 8);
                t[1] = k; t[2] = lenv; t[3] = &&L2544; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
                L2544:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                r5 = result;
                r4[2] = r5;
                r5 = top[47];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2545; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
                L2545:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = allocate (TC_TUPLE, 3);
                r6 = ((object****) lenv) [1][1][1][2];
                r5[2] = r6;
                r6 = (object*) &constructed_125;
                r5[3] = r6;
                r6 = top[31];
                t = allocate (TC_SAVE, 8);
                t[1] = k; t[2] = lenv; t[3] = &&L2546; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
                L2546:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                r5 = result;
                r4[2] = r5;
                r5 = top[47];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2547; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
                L2547:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[3] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = (object *) 5;
                r6 = ((object****) lenv) [1][1][1][3];
                r5 = box((pxll_int)unbox(r6)+unbox(r5));
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2548; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2548:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[4] = r4;
                r4 = (object*)TC_NIL;
                r3[5] = r4;
                r4 = ((object*) lenv) [5];
                r5 = (object*) &constructed_124;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[6] = r4;
                r4 = ((object*) lenv) [6];
                r5 = ((object*) lenv) [4];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[7] = r4;
                r4 = ((object*) lenv) [7];
                r5 = (object*) &constructed_123;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[8] = r4;
                r4 = ((object*) lenv) [8];
                r5 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[9] = r4;
                r4 = ((object*) lenv) [9];
                r5 = (object*) &constructed_122;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[10] = r4;
                r4 = ((object*) lenv) [10];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[11] = r4;
                r3 = ((object*) lenv) [11];
                r4 = (object*) &constructed_121;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2549; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2549:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
              }
              r1[2] = r2;
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 6);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2550; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2550:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_118;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r3 = ((object*) lenv) [6];
              r4 = (object*) &constructed_117;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2551; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2551:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2552; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2552:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              lenv = lenv[1];
            } else {
              r1 = (object *) 14;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [4];
            lenv = lenv[1];
          } break;
          case (UOTAG(9)): {
            r1 = allocate (TC_TUPLE, 5);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r1 = (object *) -1;
            r2 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 8);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][3];
              r6 = (object *) 5;
              r5 = box((pxll_int)unbox(r6)+unbox(r5));
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2553; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2553:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][4];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2554; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2554:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = (object*)TC_NIL;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_129;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_128;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_127;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2555; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2555:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2556; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2556:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 14);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object**) lenv) [1][2];
              r5[2] = r6;
              r6 = (object*) &constructed_136;
              r5[3] = r6;
              r6 = top[31];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2557; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
              L2557:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[47];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2558; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
              L2558:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object**) lenv) [1][2];
              r5[2] = r6;
              r6 = (object*) &constructed_135;
              r5[3] = r6;
              r6 = top[31];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2559; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
              L2559:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[47];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2560; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
              L2560:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = (object *) 5;
              r6 = ((object**) lenv) [1][3];
              r5 = box((pxll_int)unbox(r6)+unbox(r5));
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2561; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2561:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[4] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][4];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2562; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2562:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[5] = r4;
              r4 = (object*)TC_NIL;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_134;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [5];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r4 = ((object*) lenv) [8];
              r5 = (object*) &constructed_133;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[9] = r4;
              r4 = ((object*) lenv) [9];
              r5 = ((object*) lenv) [4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[10] = r4;
              r4 = ((object*) lenv) [10];
              r5 = (object*) &constructed_132;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[11] = r4;
              r4 = ((object*) lenv) [11];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[12] = r4;
              r4 = ((object*) lenv) [12];
              r5 = (object*) &constructed_131;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[13] = r4;
              r4 = ((object*) lenv) [13];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[14] = r4;
              r3 = ((object*) lenv) [14];
              r4 = (object*) &constructed_130;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2563; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2563:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2564; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2564:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            }
            r1 = ((object*) lenv) [5];
            lenv = lenv[1];
          } break;
          case (UOTAG(10)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2565; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2565:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 8);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2566; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2566:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = allocate (TC_TUPLE, 2);
            r5 = (object *) 3;
            r6 = ((object***) lenv) [1][1][2];
            r5 = box((pxll_int)unbox(r6)+unbox(r5));
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2567; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2567:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[3] = r4;
            r4 = (object*)TC_NIL;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = (object*) &constructed_141;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r4 = ((object*) lenv) [5];
            r5 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[6] = r4;
            r4 = ((object*) lenv) [6];
            r5 = (object*) &constructed_140;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[7] = r4;
            r4 = ((object*) lenv) [7];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[8] = r4;
            r3 = ((object*) lenv) [8];
            r4 = (object*) &constructed_139;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2568; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2568:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2569; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2569:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = top[204];
            if PXLL_IS_TRUE(r1) {
              r1 = (object *) 6;
              top[204] = r1;
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 5);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2570; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2570:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = (object*) &constructed_138;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r3 = ((object*) lenv) [5];
              r4 = (object*) &constructed_137;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2571; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2571:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2572; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2572:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 14;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [3];
            lenv = lenv[1];
          } break;
          case (UOTAG(11)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2573; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2573:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object**) lenv) [1][2];
            switch (get_case (r2)) {
              case (UOTAG(0)): {
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = ((object***) lenv) [1][1][2];
                r3 = UOBJ_GET(r3,0);
                r2[2] = r3;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object*) lenv) [2];
                r2[2] = r3;
                r3 = top[56];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2574; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
                L2574:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                lenv = lenv[1];
              } break;
              default: {
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = ((object***) lenv) [1][1][2];
                r3 = UOBJ_GET(r3,0);
                r2[2] = r3;
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 6);
                r3[1] = lenv; lenv = r3;
                r4 = (object *) 1;
                r5 = ((object****) lenv) [1][1][1][3];
                if PXLL_IS_TRUE(PXLL_TEST(unbox(r5)==unbox(r4))) {
                  r4 = (object*) &constructed_150;
                } else {
                  r4 = (object*) &constructed_151;
                }
                r3[2] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2575; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2575:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[3] = r4;
                r4 = (object*)TC_NIL;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = (object*) &constructed_149;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r4 = ((object*) lenv) [5];
                r5 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[6] = r4;
                r3 = ((object*) lenv) [6];
                r4 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2576; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2576:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                lenv = lenv[1];
              } break;
            }
            r1[3] = r2;
            r1 = (object *) 1;
            r2 = ((object**) lenv) [1][3];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 7);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2577; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2577:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = (object*) &constructed_144;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = ((object**) lenv) [1][3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = (object*) &constructed_143;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r3 = ((object*) lenv) [7];
              r4 = (object*) &constructed_142;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2578; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2578:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2579; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2579:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 10);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2580; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2580:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][3];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2581; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2581:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = (object*)TC_NIL;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_148;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_147;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object**) lenv) [1][3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r4 = ((object*) lenv) [8];
              r5 = (object*) &constructed_146;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[9] = r4;
              r4 = ((object*) lenv) [9];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[10] = r4;
              r3 = ((object*) lenv) [10];
              r4 = (object*) &constructed_145;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2582; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2582:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2583; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2583:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [4];
            lenv = lenv[1];
          } break;
          case (UOTAG(12)): {
            r1 = allocate (TC_TUPLE, 6);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,4);
            r1[6] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 11);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][4];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2584; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2584:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = allocate (TC_TUPLE, 2);
            r5 = allocate (TC_TUPLE, 2);
            r5[1] = lenv; lenv = r5;
            r6 = ((object***) lenv) [1][1][2];
            r7 = ((object***) lenv) [1][1][5];
            r6 = box((pxll_int)unbox(r7)+unbox(r6));
            r5[2] = r6;
            r5 = ((object*) lenv) [2];
            r6 = (object *) 3;
            r5 = box((pxll_int)unbox(r6)+unbox(r5));
            lenv = lenv[1];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2585; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2585:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[3] = r4;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][3];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2586; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2586:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[4] = r4;
            r4 = (object*)TC_NIL;
            r3[5] = r4;
            r4 = ((object*) lenv) [5];
            r5 = (object*) &constructed_155;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[6] = r4;
            r4 = ((object*) lenv) [6];
            r5 = ((object*) lenv) [4];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[7] = r4;
            r4 = ((object*) lenv) [7];
            r5 = (object*) &constructed_154;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[8] = r4;
            r4 = ((object*) lenv) [8];
            r5 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[9] = r4;
            r4 = ((object*) lenv) [9];
            r5 = (object*) &constructed_153;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[10] = r4;
            r4 = ((object*) lenv) [10];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[11] = r4;
            r3 = ((object*) lenv) [11];
            r4 = (object*) &constructed_152;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2587; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2587:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2588; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2588:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*) lenv) [6];
            lenv = lenv[1];
          } break;
          case (UOTAG(13)): {
            r1 = allocate (TC_TUPLE, 5);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r1 = allocate (TC_TUPLE, 5);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 3);
            r3 = top[12];
            r2[2] = r3;
            r3 = allocate (TC_TUPLE, 2);
            r4 = ((object**) lenv) [1][5];
            r3[2] = r4;
            r4 = top[136];
            t = allocate (TC_SAVE, 6);
            t[1] = k; t[2] = lenv; t[3] = &&L2589; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
            r3[1] = r4[2]; lenv = r3; goto FUN_k_2ffree_522;
            L2589:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
            r3 = result;
            r2[3] = r3;
            r3 = top[44];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2590; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_sort_340;
            L2590:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = top[202];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2591; t[4] = r0; t[5] = r1; k = t;
            lenv = r2[2]; goto *r2[1];
            L2591:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object*) lenv) [2];
            r2[2] = r3;
            r3 = top[22];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2592; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
            L2592:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[4] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][5];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2593; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2593:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[5] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][4];
            r6 = (object *) 7;
            r5 = box((pxll_int)unbox(r6)+unbox(r5));
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2594; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2594:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_181;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_180;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2595; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2595:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2596; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2596:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][4];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            // def $loop_4424_i3362_i4683_i4727
            goto L2597;
            FUN__24loop_4424_i3362_i4683_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object*) lenv) [3];
                r0[2] = r1;
                r1 = (object *) constructed_36[0];
                r0[3] = r1;
                r1 = top[21];
                r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 7);
                r2[1] = lenv; lenv = r2;
                r3 = allocate (TC_TUPLE, 2);
                r4 = (object *) 9;
                r5 = ((object***) lenv) [1][1][2];
                r4 = box((pxll_int)unbox(r5)+unbox(r4));
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2598; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2598:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[2] = r3;
                r3 = allocate (TC_TUPLE, 2);
                r4 = allocate (TC_TUPLE, 3);
                r5 = ((object******) lenv) [1][1][1][1][1][2];
                r4[2] = r5;
                r5 = ((object***) lenv) [1][1][2];
                r4[3] = r5;
                r5 = top[27];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2599; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_nth_321;
                L2599:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2600; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2600:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[3] = r3;
                r3 = (object*)TC_NIL;
                r2[4] = r3;
                r3 = ((object*) lenv) [4];
                r4 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[5] = r3;
                r3 = ((object*) lenv) [5];
                r4 = (object*) &constructed_179;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[6] = r3;
                r3 = ((object*) lenv) [6];
                r4 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[7] = r3;
                r2 = ((object*) lenv) [7];
                r3 = (object*) &constructed_178;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = top[47];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2601; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                L2601:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object**) lenv) [1][3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                r1 = (object *) 3;
                r2 = ((object*) lenv) [2];
                r1 = box((pxll_int)unbox(r2)+unbox(r1));
                lenv[2] = r1;
                lenv[3] = r0;
                goto FUN__24loop_4424_i3362_i4683_i4727;
              }
              PXLL_RETURN(0);
            L2597:
            r3 = allocate (TC_CLOSURE, 2);
            r3[1] = &&FUN__24loop_4424_i3362_i4683_i4727; r3[2] = lenv;
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = (object *) 1;
            r2[2] = r3;
            r3 = (object*)TC_NIL;
            r2[3] = r3;
            r3 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2602; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN__24loop_4424_i3362_i4683_i4727;
            L2602:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 7);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 3);
            r5 = ((object**) lenv) [1][3];
            r4[2] = r5;
            r5 = (object*) &constructed_177;
            r4[3] = r5;
            r5 = top[48];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2603; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_string_join_353;
            L2603:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_176;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r4 = ((object*) lenv) [5];
            r5 = (object*) &constructed_175;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[6] = r4;
            r4 = ((object*) lenv) [6];
            r5 = ((object***) lenv) [1][1][3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[7] = r4;
            r3 = ((object*) lenv) [7];
            r4 = (object*) &constructed_174;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2604; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2604:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2605; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2605:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            switch (get_case (r2)) {
              case (UITAG(1)): {
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 5);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object****) lenv) [1][1][1][3];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2606; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2606:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = (object*)TC_NIL;
                r3[3] = r4;
                r4 = ((object*) lenv) [3];
                r5 = (object*) &constructed_171;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r3 = ((object*) lenv) [5];
                r4 = (object*) &constructed_170;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2607; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2607:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
              } break;
              default: {
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = ((object****) lenv) [1][1][1][2];
                r3 = UOBJ_GET(r3,0);
                r2[2] = r3;
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 5);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2608; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_gen_function_label_4234_i4683_i4727;
                L2608:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = (object*)TC_NIL;
                r3[3] = r4;
                r4 = ((object*) lenv) [3];
                r5 = (object*) &constructed_173;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r3 = ((object*) lenv) [5];
                r4 = (object*) &constructed_172;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2609; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2609:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                lenv = lenv[1];
              } break;
            }
            r1[2] = r2;
            r1 = (object *) 1;
            r2 = ((object***) lenv) [1][1][4];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 12);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object****) lenv) [1][1][1][4];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2610; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2610:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object****) lenv) [1][1][1][3];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2611; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2611:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object****) lenv) [1][1][1][4];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2612; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2612:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[4] = r4;
              r4 = (object*)TC_NIL;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_167;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r4 = ((object*) lenv) [8];
              r5 = (object*) &constructed_166;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[9] = r4;
              r4 = ((object*) lenv) [9];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[10] = r4;
              r4 = ((object*) lenv) [10];
              r5 = (object*) &constructed_165;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[11] = r4;
              r4 = ((object*) lenv) [11];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[12] = r4;
              r3 = ((object*) lenv) [12];
              r4 = (object*) &constructed_164;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2613; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2613:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2614; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2614:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 6);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object****) lenv) [1][1][1][3];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2615; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2615:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_169;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r3 = ((object*) lenv) [6];
              r4 = (object*) &constructed_168;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2616; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2616:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2617; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2617:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            }
            lenv = lenv[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 3);
            r3[1] = lenv; lenv = r3;
            r4 = (object*)TC_NIL;
            r3[2] = r4;
            r4 = ((object*) lenv) [2];
            r5 = (object*) &constructed_163;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[3] = r4;
            r3 = ((object*) lenv) [3];
            r4 = ((object**) lenv) [1][3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2618; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2618:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2619; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2619:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][4];
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r2[1] = lenv; lenv = r2;
            // def $loop_4431_i3362_i4683_i4727
            goto L2620;
            FUN__24loop_4431_i3362_i4683_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object*) lenv) [3];
                r0[2] = r1;
                r1 = (object *) constructed_36[0];
                r0[3] = r1;
                r1 = top[21];
                r0[1] = r1[2]; lenv = r0; goto FUN_reverse_onto_311;
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 8);
                r2[1] = lenv; lenv = r2;
                r3 = allocate (TC_TUPLE, 2);
                r4 = allocate (TC_TUPLE, 3);
                r5 = ((object******) lenv) [1][1][1][1][1][2];
                r4[2] = r5;
                r5 = ((object***) lenv) [1][1][2];
                r4[3] = r5;
                r5 = top[27];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2621; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_nth_321;
                L2621:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2622; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2622:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[2] = r3;
                r3 = allocate (TC_TUPLE, 2);
                r4 = (object *) 9;
                r5 = ((object***) lenv) [1][1][2];
                r4 = box((pxll_int)unbox(r5)+unbox(r4));
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2623; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2623:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[3] = r3;
                r3 = (object*)TC_NIL;
                r2[4] = r3;
                r3 = ((object*) lenv) [4];
                r4 = (object*) &constructed_162;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[5] = r3;
                r3 = ((object*) lenv) [5];
                r4 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[6] = r3;
                r3 = ((object*) lenv) [6];
                r4 = (object*) &constructed_161;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[7] = r3;
                r3 = ((object*) lenv) [7];
                r4 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[8] = r3;
                r2 = ((object*) lenv) [8];
                r3 = (object*) &constructed_160;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = top[47];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2624; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                L2624:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r0 = ((object**) lenv) [1][3];
                r1 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                r1 = (object *) 3;
                r2 = ((object*) lenv) [2];
                r1 = box((pxll_int)unbox(r2)+unbox(r1));
                lenv[2] = r1;
                lenv[3] = r0;
                goto FUN__24loop_4431_i3362_i4683_i4727;
              }
              PXLL_RETURN(0);
            L2620:
            r3 = allocate (TC_CLOSURE, 2);
            r3[1] = &&FUN__24loop_4431_i3362_i4683_i4727; r3[2] = lenv;
            r2[2] = r3;
            r2 = allocate (TC_TUPLE, 3);
            r3 = (object *) 1;
            r2[2] = r3;
            r3 = (object*)TC_NIL;
            r2[3] = r3;
            r3 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2625; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN__24loop_4431_i3362_i4683_i4727;
            L2625:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            lenv = lenv[1];
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 4);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 3);
            r5 = ((object**) lenv) [1][3];
            r4[2] = r5;
            r5 = (object*) &constructed_159;
            r4[3] = r5;
            r5 = top[48];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2626; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_string_join_353;
            L2626:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_158;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r3 = ((object*) lenv) [4];
            r4 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2627; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2627:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2628; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2628:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = (object *) 1;
            r2 = ((object*) lenv) [5];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 5);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object**) lenv) [1][5];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2629; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2629:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = (object*) &constructed_157;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r3 = ((object*) lenv) [5];
              r4 = (object*) &constructed_156;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2630; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2630:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2631; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2631:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 14;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [5];
            lenv = lenv[1];
          } break;
          case (UOTAG(14)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            switch (get_case (r2)) {
              case (UITAG(1)): {
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 5);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object***) lenv) [1][1][3];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2632; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2632:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = (object*)TC_NIL;
                r3[3] = r4;
                r4 = ((object*) lenv) [3];
                r5 = (object*) &constructed_189;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r3 = ((object*) lenv) [5];
                r4 = (object*) &constructed_188;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2633; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2633:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
              } break;
              default: {
                r2 = allocate (TC_TUPLE, 2);
                r2[1] = lenv; lenv = r2;
                r3 = ((object***) lenv) [1][1][2];
                r3 = UOBJ_GET(r3,0);
                r2[2] = r3;
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 5);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = ((object*******) lenv) [1][1][1][1][1][1][4];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2634; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_gen_function_label_4234_i4683_i4727;
                L2634:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = (object*)TC_NIL;
                r3[3] = r4;
                r4 = ((object*) lenv) [3];
                r5 = (object*) &constructed_191;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r3 = ((object*) lenv) [5];
                r4 = (object*) &constructed_190;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2635; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2635:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                lenv = lenv[1];
              } break;
            }
            r1[2] = r2;
            r1 = (object *) 1;
            r2 = ((object**) lenv) [1][4];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 12);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][4];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2636; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2636:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][3];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2637; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2637:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][4];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2638; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2638:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[4] = r4;
              r4 = (object*)TC_NIL;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_185;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [4];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r4 = ((object*) lenv) [8];
              r5 = (object*) &constructed_184;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[9] = r4;
              r4 = ((object*) lenv) [9];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[10] = r4;
              r4 = ((object*) lenv) [10];
              r5 = (object*) &constructed_183;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[11] = r4;
              r4 = ((object*) lenv) [11];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[12] = r4;
              r3 = ((object*) lenv) [12];
              r4 = (object*) &constructed_182;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2639; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2639:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2640; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2640:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 6);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][3];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2641; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2641:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = ((object**) lenv) [1][2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_187;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r3 = ((object*) lenv) [6];
              r4 = (object*) &constructed_186;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2642; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2642:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2643; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2643:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            }
            lenv = lenv[1];
            r1 = (object*)UITAG(1);
            lenv = lenv[1];
          } break;
          case (UOTAG(15)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[22];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2644; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
            L2644:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = (object *) 3;
            r3 = ((object**) lenv) [1][2];
            r2 = box((pxll_int)unbox(r3)-unbox(r2));
            r1[3] = r2;
            r1 = (object *) 1;
            r2 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
              r1 = (object *) 3;
              r2 = ((object*) lenv) [3];
              r1 = box((pxll_int)unbox(r2)+unbox(r1));
              ((object*) lenv) [3] = r1;
            } else {
              r1 = (object *) 14;
            }
            r1 = (object *) 1;
            r2 = ((object*) lenv) [3];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 8);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object**) lenv) [1][3];
              r5[2] = r6;
              r6 = (object*) &constructed_201;
              r5[3] = r6;
              r6 = top[31];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2645; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
              L2645:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[47];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2646; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
              L2646:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object**) lenv) [1][3];
              r5[2] = r6;
              r6 = (object*) &constructed_200;
              r5[3] = r6;
              r6 = top[31];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2647; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
              L2647:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[47];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2648; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
              L2648:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = (object*)TC_NIL;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_199;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_198;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_197;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2649; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2649:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2650; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2650:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 14;
            }
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            // def $loop_4441_i3456_i4683_i4727
            goto L2651;
            FUN__24loop_4441_i3456_i4683_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 2);
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 8);
                r2[1] = lenv; lenv = r2;
                r3 = allocate (TC_TUPLE, 2);
                r4 = ((object**) lenv) [1][2];
                r5 = (object *) 5;
                r4 = box((pxll_int)unbox(r5)+unbox(r4));
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2652; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2652:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[2] = r3;
                r3 = allocate (TC_TUPLE, 2);
                r4 = allocate (TC_TUPLE, 3);
                r5 = ((object******) lenv) [1][1][1][1][1][4];
                r4[2] = r5;
                r5 = ((object**) lenv) [1][2];
                r4[3] = r5;
                r5 = top[27];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2653; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_nth_321;
                L2653:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2654; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2654:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[3] = r3;
                r3 = (object*)TC_NIL;
                r2[4] = r3;
                r3 = ((object*) lenv) [4];
                r4 = (object*) &constructed_196;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[5] = r3;
                r3 = ((object*) lenv) [5];
                r4 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[6] = r3;
                r3 = ((object*) lenv) [6];
                r4 = (object*) &constructed_195;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[7] = r3;
                r3 = ((object*) lenv) [7];
                r4 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[8] = r3;
                r2 = ((object*) lenv) [8];
                r3 = (object*) &constructed_194;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = top[47];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2655; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                L2655:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                r1 = ((pxll_vector*)r1)->val[4];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2656; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L2656:
                ; lenv = k[2]; k = k[1];
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_4441_i3456_i4683_i4727;
              }
              PXLL_RETURN(0);
            L2651:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN__24loop_4441_i3456_i4683_i4727; r2[2] = lenv;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object *) 1;
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2657; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4441_i3456_i4683_i4727;
            L2657:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object***) lenv) [1][1][3];
            r4[2] = r5;
            r5 = ((object******) lenv) [1][1][1][1][1][4];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2658; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_gen_function_label_4234_i4683_i4727;
            L2658:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_193;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_192;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2659; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2659:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2660; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2660:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = (object*)UITAG(1);
            lenv = lenv[1];
          } break;
          case (UOTAG(16)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 8);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2661; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2661:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2662; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2662:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[3] = r4;
            r4 = (object*)TC_NIL;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = (object*) &constructed_204;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r4 = ((object*) lenv) [5];
            r5 = ((object*) lenv) [3];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[6] = r4;
            r4 = ((object*) lenv) [6];
            r5 = (object*) &constructed_203;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[7] = r4;
            r4 = ((object*) lenv) [7];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[8] = r4;
            r3 = ((object*) lenv) [8];
            r4 = (object*) &constructed_202;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2663; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2663:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2664; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2664:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*) lenv) [3];
            lenv = lenv[1];
          } break;
          case (UOTAG(17)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][3];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2665; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2665:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object*) &constructed_205;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2666; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2666:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 3);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            r1[3] = r2;
            r2 = ((object*****) lenv) [1][1][1][1][3];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2667; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_move_4227_i4683_i4727;
            L2667:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = ((object*) lenv) [3];
            lenv = lenv[1];
          } break;
          case (UOTAG(18)): {
            r1 = allocate (TC_TUPLE, 6);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,4);
            r1[6] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            // def primop-error_4470_i3489_i4683_i4727
            goto L2668;
            FUN_primop_error_4470_i3489_i4683_i4727:
              check_heap (0);
              r0 = allocate (TC_TUPLE, 3);
              r1 = (object*) &constructed_208;
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r0[3] = r1;
              r1 = top[15];
              r0[1] = r1[2]; lenv = r0; goto FUN_error1_293;
            L2668:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN_primop_error_4470_i3489_i4683_i4727; r2[2] = lenv;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object***) lenv) [1][1][6];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2669; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2669:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object***) lenv) [1][1][5];
            r2[2] = r3;
            r3 = top[22];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2670; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
            L2670:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r1 = (object *) constructed_210[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [3];
                      switch (get_case (r1)) {
                        case (TC_NIL): {
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 5);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2671; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2671:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = (object*)TC_NIL;
                          r3[3] = r4;
                          r4 = ((object*) lenv) [3];
                          r5 = (object*) &constructed_207;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r3 = ((object*) lenv) [5];
                          r4 = (object*) &constructed_206;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2672; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2672:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2673; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2673:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = allocate (TC_TUPLE, 3);
                          r2 = ((object*) lenv) [2];
                          r1[2] = r2;
                          r2 = ((object***) lenv) [1][1][2];
                          r1[3] = r2;
                          r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2674; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto FUN_move_4227_i4683_i4727;
                          L2674:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_9;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_9;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_9;
                }
              }
              goto fatbar_9_over;
              fatbar_9:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2675; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2675:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_9_over: ;
            } else {
              goto fatbar_8;
            }
            goto fatbar_8_over;
            fatbar_8:
            r1 = (object *) constructed_214[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_NIL): {
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = allocate (TC_TUPLE, 5);
                  r3[1] = lenv; lenv = r3;
                  r4 = allocate (TC_TUPLE, 2);
                  r5 = ((object**) lenv) [1][2];
                  r4[2] = r5;
                  r5 = top[56];
                  t = allocate (TC_SAVE, 7);
                  t[1] = k; t[2] = lenv; t[3] = &&L2676; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                  L2676:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                  r4 = result;
                  r3[2] = r4;
                  r4 = (object*)TC_NIL;
                  r3[3] = r4;
                  r4 = ((object*) lenv) [3];
                  r5 = (object*) &constructed_212;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r5;
                  t[2] = r4;
                  r4 = t;
                  r3[4] = r4;
                  r4 = ((object*) lenv) [4];
                  r5 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r5;
                  t[2] = r4;
                  r4 = t;
                  r3[5] = r4;
                  r3 = ((object*) lenv) [5];
                  r4 = (object*) &constructed_211;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r4;
                  t[2] = r3;
                  r3 = t;
                  lenv = lenv[1];
                  r2[2] = r3;
                  r3 = top[47];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2677; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                  L2677:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
                  r2 = ((pxll_vector*)r2)->val[4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2678; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                  L2678:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                } break;
                default: {
                  goto fatbar_11;
                }
              }
              goto fatbar_11_over;
              fatbar_11:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2679; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2679:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_11_over: ;
            } else {
              goto fatbar_10;
            }
            goto fatbar_10_over;
            fatbar_10:
            r1 = (object *) constructed_230[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [3];
                      switch (get_case (r1)) {
                        case (TC_PAIR): {
                          r1 = allocate (TC_TUPLE, 3);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object**) lenv) [1][3];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][3];
                          r2 = UOBJ_GET(r2,1);
                          r1[3] = r2;
                          r1 = ((object*) lenv) [3];
                          switch (get_case (r1)) {
                            case (TC_NIL): {
                              r1 = ((object******) lenv) [1][1][1][1][1][4];
                              switch (get_case (r1)) {
                                case (UOTAG(1)): {
                                  r1 = allocate (TC_TUPLE, 3);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
                                  r2 = UOBJ_GET(r2,0);
                                  r1[2] = r2;
                                  r2 = ((object*******) lenv) [1][1][1][1][1][1][4];
                                  r2 = UOBJ_GET(r2,1);
                                  r1[3] = r2;
                                  r1 = (object *) constructed_97[0];
                                  r2 = ((object*) lenv) [2];
                                  if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
                                    r1 = ((object*) lenv) [3];
                                    switch (get_case (r1)) {
                                      case (TC_PAIR): {
                                        r1 = allocate (TC_TUPLE, 3);
                                        r1[1] = lenv; lenv = r1;
                                        r2 = ((object**) lenv) [1][3];
                                        r2 = UOBJ_GET(r2,0);
                                        r1[2] = r2;
                                        r2 = ((object**) lenv) [1][3];
                                        r2 = UOBJ_GET(r2,1);
                                        r1[3] = r2;
                                        r1 = ((object*) lenv) [3];
                                        switch (get_case (r1)) {
                                          case (TC_PAIR): {
                                            r1 = allocate (TC_TUPLE, 3);
                                            r1[1] = lenv; lenv = r1;
                                            r2 = ((object**) lenv) [1][3];
                                            r2 = UOBJ_GET(r2,0);
                                            r1[2] = r2;
                                            r2 = ((object**) lenv) [1][3];
                                            r2 = UOBJ_GET(r2,1);
                                            r1[3] = r2;
                                            r1 = ((object*) lenv) [3];
                                            switch (get_case (r1)) {
                                              case (TC_NIL): {
                                                r1 = allocate (TC_TUPLE, 4);
                                                r1[1] = lenv; lenv = r1;
                                                r2 = allocate (TC_TUPLE, 2);
                                                r3 = allocate (TC_TUPLE, 4);
                                                r3[1] = lenv; lenv = r3;
                                                r4 = allocate (TC_TUPLE, 2);
                                                r5 = ((object******) lenv) [1][1][1][1][1][2];
                                                r4[2] = r5;
                                                r5 = top[56];
                                                t = allocate (TC_SAVE, 7);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2680; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                L2680:
                                                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                r4 = result;
                                                r3[2] = r4;
                                                r4 = (object*)TC_NIL;
                                                r3[3] = r4;
                                                r4 = ((object*) lenv) [3];
                                                r5 = ((object*) lenv) [2];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[4] = r4;
                                                r3 = ((object*) lenv) [4];
                                                r4 = (object*) &constructed_228;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r4;
                                                t[2] = r3;
                                                r3 = t;
                                                lenv = lenv[1];
                                                r2[2] = r3;
                                                r3 = top[47];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2681; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                L2681:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                r1[2] = r2;
                                                r2 = ((object***) lenv) [1][1][2];
                                                switch (get_case (r2)) {
                                                  case (UOTAG(1)): {
                                                    r2 = allocate (TC_TUPLE, 2);
                                                    r2[1] = lenv; lenv = r2;
                                                    r3 = ((object****) lenv) [1][1][1][2];
                                                    r3 = UOBJ_GET(r3,0);
                                                    r2[2] = r3;
                                                    r2 = ((object***) lenv) [1][1][2];
                                                    switch (get_case (r2)) {
                                                      case (UOTAG(1)): {
                                                        r2 = allocate (TC_TUPLE, 2);
                                                        r2[1] = lenv; lenv = r2;
                                                        r3 = ((object****) lenv) [1][1][1][2];
                                                        r3 = UOBJ_GET(r3,0);
                                                        r2[2] = r3;
                                                        r2 = (object *) constructed_226[0];
                                                        r3 = ((object*) lenv) [2];
                                                        if PXLL_IS_TRUE(PXLL_TEST(r3==r2)) {
                                                          r2 = allocate (TC_TUPLE, 3);
                                                          r3 = ((object**) lenv) [1][2];
                                                          r2[2] = r3;
                                                          r3 = top[203];
                                                          r2[3] = r3;
                                                          r3 = top[25];
                                                          t = allocate (TC_SAVE, 5);
                                                          t[1] = k; t[2] = lenv; t[3] = &&L2682; t[4] = r0; t[5] = r1; k = t;
                                                          r2[1] = r3[2]; lenv = r2; goto FUN_member_eq_question_319;
                                                          L2682:
                                                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                          r2 = result;
                                                          if PXLL_IS_TRUE(r2) {
                                                            r2 = allocate (TC_TUPLE, 2);
                                                            r3 = allocate (TC_TUPLE, 7);
                                                            r3[1] = lenv; lenv = r3;
                                                            r4 = ((object***) lenv) [1][1][2];
                                                            r4 = allocate (TC_TUPLE, 2);
                                                            r4[1] = lenv; lenv = r4;
                                                            r5 = ((object****) lenv) [1][1][1][2];
                                                            r5 = UOBJ_GET(r5,0);
                                                            r4[2] = r5;
                                                            r4 = ((object*) lenv) [2];
                                                            lenv = lenv[1];
                                                            r3[2] = r4;
                                                            r4 = (object*)TC_NIL;
                                                            r3[3] = r4;
                                                            r4 = ((object*) lenv) [3];
                                                            r5 = (object*) &constructed_223;
                                                            t = alloc_no_clear (TC_PAIR,2);
                                                            t[1] = r5;
                                                            t[2] = r4;
                                                            r4 = t;
                                                            r3[4] = r4;
                                                            r4 = ((object*) lenv) [4];
                                                            r5 = ((object****) lenv) [1][1][1][2];
                                                            t = alloc_no_clear (TC_PAIR,2);
                                                            t[1] = r5;
                                                            t[2] = r4;
                                                            r4 = t;
                                                            r3[5] = r4;
                                                            r4 = ((object*) lenv) [5];
                                                            r5 = (object*) &constructed_222;
                                                            t = alloc_no_clear (TC_PAIR,2);
                                                            t[1] = r5;
                                                            t[2] = r4;
                                                            r4 = t;
                                                            r3[6] = r4;
                                                            r4 = ((object*) lenv) [6];
                                                            r5 = ((object*) lenv) [2];
                                                            t = alloc_no_clear (TC_PAIR,2);
                                                            t[1] = r5;
                                                            t[2] = r4;
                                                            r4 = t;
                                                            r3[7] = r4;
                                                            r3 = ((object*) lenv) [7];
                                                            r4 = (object*) &constructed_221;
                                                            t = alloc_no_clear (TC_PAIR,2);
                                                            t[1] = r4;
                                                            t[2] = r3;
                                                            r3 = t;
                                                            lenv = lenv[1];
                                                            r2[2] = r3;
                                                            r3 = top[47];
                                                            t = allocate (TC_SAVE, 5);
                                                            t[1] = k; t[2] = lenv; t[3] = &&L2683; t[4] = r0; t[5] = r1; k = t;
                                                            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                            L2683:
                                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                            r2 = result;
                                                          } else {
                                                            r2 = allocate (TC_TUPLE, 3);
                                                            r3 = (object*) &constructed_224;
                                                            r2[2] = r3;
                                                            r3 = ((object**) lenv) [1][2];
                                                            r2[3] = r3;
                                                            r3 = top[15];
                                                            t = allocate (TC_SAVE, 5);
                                                            t[1] = k; t[2] = lenv; t[3] = &&L2684; t[4] = r0; t[5] = r1; k = t;
                                                            r2[1] = r3[2]; lenv = r2; goto FUN_error1_293;
                                                            L2684:
                                                            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                            r2 = result;
                                                          }
                                                        } else {
                                                          lenv = ((object **)lenv)[1][1];
                                                          goto fatbar_14;
                                                        }
                                                        lenv = lenv[1];
                                                      } break;
                                                      default: {
                                                        lenv = ((object *)lenv)[1];
                                                        goto fatbar_14;
                                                      }
                                                    }
                                                    lenv = lenv[1];
                                                  } break;
                                                  default: {
                                                    goto fatbar_14;
                                                  }
                                                }
                                                goto fatbar_14_over;
                                                fatbar_14:
                                                r2 = allocate (TC_TUPLE, 3);
                                                r3 = (object*) &constructed_227;
                                                r2[2] = r3;
                                                r3 = allocate (UOTAG(1), 2);
                                                r4 = ((object***) lenv) [1][1][2];
                                                r3[1] = r4;
                                                r4 = ((object**) lenv) [1][2];
                                                r3[2] = r4;
                                                r2[3] = r3;
                                                r3 = top[15];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2685; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto FUN_error1_293;
                                                L2685:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                fatbar_14_over: ;
                                                r1[3] = r2;
                                                r2 = allocate (TC_TUPLE, 2);
                                                r3 = allocate (TC_TUPLE, 11);
                                                r3[1] = lenv; lenv = r3;
                                                r4 = allocate (TC_TUPLE, 2);
                                                r5 = ((object****) lenv) [1][1][1][2];
                                                r4[2] = r5;
                                                r5 = top[107];
                                                t = allocate (TC_SAVE, 7);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2686; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                r4[1] = r5[2]; lenv = r4; goto FUN_type_repr_462;
                                                L2686:
                                                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                r4 = result;
                                                r3[2] = r4;
                                                r4 = allocate (TC_TUPLE, 2);
                                                r5 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                                r4[2] = r5;
                                                r5 = top[56];
                                                t = allocate (TC_SAVE, 7);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2687; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                L2687:
                                                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                r4 = result;
                                                r3[3] = r4;
                                                r4 = allocate (TC_TUPLE, 2);
                                                r5 = ((object*******) lenv) [1][1][1][1][1][1][2];
                                                r4[2] = r5;
                                                r5 = top[56];
                                                t = allocate (TC_SAVE, 7);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2688; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                L2688:
                                                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                r4 = result;
                                                r3[4] = r4;
                                                r4 = (object*)TC_NIL;
                                                r3[5] = r4;
                                                r4 = ((object*) lenv) [5];
                                                r5 = (object*) &constructed_220;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[6] = r4;
                                                r4 = ((object*) lenv) [6];
                                                r5 = ((object*) lenv) [4];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[7] = r4;
                                                r4 = ((object*) lenv) [7];
                                                r5 = (object*) &constructed_219;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[8] = r4;
                                                r4 = ((object*) lenv) [8];
                                                r5 = ((object*) lenv) [3];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[9] = r4;
                                                r4 = ((object*) lenv) [9];
                                                r5 = (object*) &constructed_218;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[10] = r4;
                                                r4 = ((object*) lenv) [10];
                                                r5 = ((object*) lenv) [2];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[11] = r4;
                                                r3 = ((object*) lenv) [11];
                                                r4 = (object*) &constructed_217;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r4;
                                                t[2] = r3;
                                                r3 = t;
                                                lenv = lenv[1];
                                                r2[2] = r3;
                                                r3 = top[47];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2689; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                L2689:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                r1[4] = r2;
                                                r1 = allocate (TC_TUPLE, 2);
                                                r2 = allocate (TC_TUPLE, 2);
                                                r3 = allocate (TC_TUPLE, 5);
                                                r3[1] = lenv; lenv = r3;
                                                r4 = (object*)TC_NIL;
                                                r3[2] = r4;
                                                r4 = ((object*) lenv) [2];
                                                r5 = (object*) &constructed_216;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[3] = r4;
                                                r4 = ((object*) lenv) [3];
                                                r5 = ((object**) lenv) [1][3];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[4] = r4;
                                                r4 = ((object*) lenv) [4];
                                                r5 = (object*) &constructed_215;
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r5;
                                                t[2] = r4;
                                                r4 = t;
                                                r3[5] = r4;
                                                r3 = ((object*) lenv) [5];
                                                r4 = ((object**) lenv) [1][4];
                                                t = alloc_no_clear (TC_PAIR,2);
                                                t[1] = r4;
                                                t[2] = r3;
                                                r3 = t;
                                                lenv = lenv[1];
                                                r2[2] = r3;
                                                r3 = top[47];
                                                t = allocate (TC_SAVE, 5);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2690; t[4] = r0; t[5] = r1; k = t;
                                                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                L2690:
                                                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                r2 = result;
                                                r1[2] = r2;
                                                r2 = ((object****************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                                                r2 = ((pxll_vector*)r2)->val[4];
                                                t = allocate (TC_SAVE, 4);
                                                t[1] = k; t[2] = lenv; t[3] = &&L2691; t[4] = r0; k = t;
                                                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                                L2691:
                                                r0 = k[4]; lenv = k[2]; k = k[1];
                                                r1 = result;
                                                lenv = lenv[1];
                                              } break;
                                              default: {
                                                lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                                goto fatbar_13;
                                              }
                                            }
                                            lenv = lenv[1];
                                          } break;
                                          default: {
                                            lenv = ((object *****)lenv)[1][1][1][1][1];
                                            goto fatbar_13;
                                          }
                                        }
                                        lenv = lenv[1];
                                      } break;
                                      default: {
                                        lenv = ((object ****)lenv)[1][1][1][1];
                                        goto fatbar_13;
                                      }
                                    }
                                  } else {
                                    lenv = ((object ****)lenv)[1][1][1][1];
                                    goto fatbar_13;
                                  }
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object ***)lenv)[1][1][1];
                                  goto fatbar_13;
                                }
                              }
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_13;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_13;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_13;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_13;
                }
              }
              goto fatbar_13_over;
              fatbar_13:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2692; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2692:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_13_over: ;
            } else {
              goto fatbar_12;
            }
            goto fatbar_12_over;
            fatbar_12:
            r1 = (object *) constructed_239[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [3];
                      switch (get_case (r1)) {
                        case (TC_NIL): {
                          r1 = allocate (TC_TUPLE, 2);
                          r1[1] = lenv; lenv = r1;
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 11);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object*******) lenv) [1][1][1][1][1][1][4];
                          r4[2] = r5;
                          r5 = top[107];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2693; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_type_repr_462;
                          L2693:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object****) lenv) [1][1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2694; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2694:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[3] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2695; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2695:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[4] = r4;
                          r4 = (object*)TC_NIL;
                          r3[5] = r4;
                          r4 = ((object*) lenv) [5];
                          r5 = (object*) &constructed_237;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[6] = r4;
                          r4 = ((object*) lenv) [6];
                          r5 = ((object*) lenv) [4];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[7] = r4;
                          r4 = ((object*) lenv) [7];
                          r5 = (object*) &constructed_236;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[8] = r4;
                          r4 = ((object*) lenv) [8];
                          r5 = ((object*) lenv) [3];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[9] = r4;
                          r4 = ((object*) lenv) [9];
                          r5 = (object*) &constructed_235;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[10] = r4;
                          r4 = ((object*) lenv) [10];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[11] = r4;
                          r3 = ((object*) lenv) [11];
                          r4 = (object*) &constructed_234;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2696; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2696:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 8);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object*****) lenv) [1][1][1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2697; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2697:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = allocate (TC_TUPLE, 3);
                          r5 = ((object*******) lenv) [1][1][1][1][1][1][4];
                          r4[2] = r5;
                          r5 = ((object**) lenv) [1][2];
                          r4[3] = r5;
                          r5 = top[150];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2698; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_wrap_out_561;
                          L2698:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[3] = r4;
                          r4 = (object*)TC_NIL;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = (object*) &constructed_233;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r4 = ((object*) lenv) [5];
                          r5 = ((object*) lenv) [3];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[6] = r4;
                          r4 = ((object*) lenv) [6];
                          r5 = (object*) &constructed_232;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[7] = r4;
                          r4 = ((object*) lenv) [7];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[8] = r4;
                          r3 = ((object*) lenv) [8];
                          r4 = (object*) &constructed_231;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2699; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2699:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2700; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2700:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_16;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_16;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_16;
                }
              }
              goto fatbar_16_over;
              fatbar_16:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2701; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2701:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_16_over: ;
            } else {
              goto fatbar_15;
            }
            goto fatbar_15_over;
            fatbar_15:
            r1 = (object *) constructed_243[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 5);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 2);
              r6 = ((object****) lenv) [1][1][1][5];
              r5[2] = r6;
              r6 = top[19];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2702; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_car_309;
              L2702:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2703; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2703:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = (object*)TC_NIL;
              r3[3] = r4;
              r4 = ((object*) lenv) [3];
              r5 = (object*) &constructed_241;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r3 = ((object*) lenv) [5];
              r4 = (object*) &constructed_240;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2704; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2704:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2705; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2705:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              goto fatbar_17;
            }
            goto fatbar_17_over;
            fatbar_17:
            r1 = (object *) constructed_250[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object****) lenv) [1][1][1][3];
              r2[2] = r3;
              r3 = top[197];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2706; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              lenv = r3[2]; goto *r3[1];
              L2706:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = top[111];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2707; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_parse_type_splat_467;
              L2707:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = (object *) 1;
              r2 = ((object**) lenv) [1][2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 11);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object***) lenv) [1][1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2708; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2708:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = top[149];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2709; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_irken_type__3ec_type_560;
                L2709:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[3] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = allocate (TC_TUPLE, 2);
                r6 = ((object*****) lenv) [1][1][1][1][5];
                r5[2] = r6;
                r6 = top[19];
                t = allocate (TC_SAVE, 8);
                t[1] = k; t[2] = lenv; t[3] = &&L2710; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                r5[1] = r6[2]; lenv = r5; goto FUN_car_309;
                L2710:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                r5 = result;
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2711; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2711:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[4] = r4;
                r4 = (object*)TC_NIL;
                r3[5] = r4;
                r4 = ((object*) lenv) [5];
                r5 = (object*) &constructed_247;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[6] = r4;
                r4 = ((object*) lenv) [6];
                r5 = ((object*) lenv) [4];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[7] = r4;
                r4 = ((object*) lenv) [7];
                r5 = (object*) &constructed_246;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[8] = r4;
                r4 = ((object*) lenv) [8];
                r5 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[9] = r4;
                r4 = ((object*) lenv) [9];
                r5 = (object*) &constructed_245;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[10] = r4;
                r4 = ((object*) lenv) [10];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[11] = r4;
                r3 = ((object*) lenv) [11];
                r4 = (object*) &constructed_244;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2712; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2712:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
                r2 = ((pxll_vector*)r2)->val[4];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2713; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                L2713:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              } else {
                r1 = allocate (TC_TUPLE, 3);
                r2 = (object*) &constructed_248;
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r1[3] = r2;
                r2 = top[15];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2714; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                L2714:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              }
              lenv = lenv[1];
            } else {
              goto fatbar_18;
            }
            goto fatbar_18_over;
            fatbar_18:
            r1 = (object *) constructed_255[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 8);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 2);
              r6 = allocate (TC_TUPLE, 2);
              r7 = ((object****) lenv) [1][1][1][6];
              r6[2] = r7;
              r7 = top[136];
              t = allocate (TC_SAVE, 9);
              t[1] = k; t[2] = lenv; t[3] = &&L2715; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; t[9] = r5; k = t;
              r6[1] = r7[2]; lenv = r6; goto FUN_k_2ffree_522;
              L2715:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; r5 = k[9]; lenv = k[2]; k = k[1];
              r6 = result;
              r5[2] = r6;
              r6 = top[22];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2716; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_length_314;
              L2716:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2717; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2717:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 2);
              r6 = ((object****) lenv) [1][1][1][5];
              r5[2] = r6;
              r6 = top[19];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2718; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_car_309;
              L2718:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2719; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2719:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = (object*)TC_NIL;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_253;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_252;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_251;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2720; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2720:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2721; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2721:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              goto fatbar_19;
            }
            goto fatbar_19_over;
            fatbar_19:
            r1 = (object *) constructed_266[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][3];
              switch (get_case (r1)) {
                case (UOTAG(0)): {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r1 = ((object*) lenv) [2];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [2];
                      switch (get_case (r1)) {
                        case (UOTAG(1)): {
                          r1 = allocate (TC_TUPLE, 2);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object**) lenv) [1][2];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r1 = ((object**) lenv) [1][3];
                          switch (get_case (r1)) {
                            case (TC_PAIR): {
                              r1 = allocate (TC_TUPLE, 3);
                              r1[1] = lenv; lenv = r1;
                              r2 = ((object***) lenv) [1][1][3];
                              r2 = UOBJ_GET(r2,0);
                              r1[2] = r2;
                              r2 = ((object***) lenv) [1][1][3];
                              r2 = UOBJ_GET(r2,1);
                              r1[3] = r2;
                              r1 = ((object*) lenv) [2];
                              switch (get_case (r1)) {
                                case (UOTAG(0)): {
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = ((object**) lenv) [1][2];
                                  r2 = UOBJ_GET(r2,0);
                                  r1[2] = r2;
                                  r1 = ((object**) lenv) [1][3];
                                  switch (get_case (r1)) {
                                    case (TC_NIL): {
                                      r1 = ((object********) lenv) [1][1][1][1][1][1][1][5];
                                      switch (get_case (r1)) {
                                        case (TC_PAIR): {
                                          r1 = allocate (TC_TUPLE, 3);
                                          r1[1] = lenv; lenv = r1;
                                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                                          r2 = UOBJ_GET(r2,0);
                                          r1[2] = r2;
                                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                                          r2 = UOBJ_GET(r2,1);
                                          r1[3] = r2;
                                          r1 = ((object*) lenv) [3];
                                          switch (get_case (r1)) {
                                            case (TC_PAIR): {
                                              r1 = allocate (TC_TUPLE, 3);
                                              r1[1] = lenv; lenv = r1;
                                              r2 = ((object**) lenv) [1][3];
                                              r2 = UOBJ_GET(r2,0);
                                              r1[2] = r2;
                                              r2 = ((object**) lenv) [1][3];
                                              r2 = UOBJ_GET(r2,1);
                                              r1[3] = r2;
                                              r1 = ((object*) lenv) [3];
                                              switch (get_case (r1)) {
                                                case (TC_NIL): {
                                                  r1 = allocate (TC_TUPLE, 3);
                                                  r1[1] = lenv; lenv = r1;
                                                  r2 = allocate (TC_TUPLE, 3);
                                                  r3 = ((object******) lenv) [1][1][1][1][1][2];
                                                  r2[2] = r3;
                                                  r3 = ((object*****************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][2];
                                                  r2[3] = r3;
                                                  r3 = top[9];
                                                  t = allocate (TC_SAVE, 5);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2722; t[4] = r0; t[5] = r1; k = t;
                                                  r2[1] = r3[2]; lenv = r2; goto FUN_lookup_label_code_253;
                                                  L2722:
                                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                  r2 = result;
                                                  r1[2] = r2;
                                                  r2 = allocate (TC_TUPLE, 2);
                                                  r3 = ((object****) lenv) [1][1][1][2];
                                                  r2[2] = r3;
                                                  r3 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][5];
                                                  t = allocate (TC_SAVE, 5);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2723; t[4] = r0; t[5] = r1; k = t;
                                                  r2[1] = r3[2]; lenv = r2; goto FUN_guess_record_type_4247_i4683_i4727;
                                                  L2723:
                                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                  r2 = result;
                                                  r1[3] = r2;
                                                  r1 = ((object*) lenv) [3];
                                                  switch (get_case (r1)) {
                                                    case (UOTAG(0)): {
                                                      r1 = allocate (TC_TUPLE, 2);
                                                      r1[1] = lenv; lenv = r1;
                                                      r2 = ((object**) lenv) [1][3];
                                                      r2 = UOBJ_GET(r2,0);
                                                      r1[2] = r2;
                                                      r1 = allocate (TC_TUPLE, 2);
                                                      r2 = allocate (TC_TUPLE, 2);
                                                      r3 = allocate (TC_TUPLE, 11);
                                                      r3[1] = lenv; lenv = r3;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = ((object*****) lenv) [1][1][1][1][2];
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2724; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2724:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[2] = r4;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = allocate (TC_TUPLE, 3);
                                                      r6 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                                      r5[2] = r6;
                                                      r6 = ((object**) lenv) [1][2];
                                                      r5[3] = r6;
                                                      r6 = top[28];
                                                      t = allocate (TC_SAVE, 8);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2725; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                                                      r5[1] = r6[2]; lenv = r5; goto FUN_index_eq_322;
                                                      L2725:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                                                      r5 = result;
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2726; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2726:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[3] = r4;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = ((object****) lenv) [1][1][1][2];
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2727; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2727:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[4] = r4;
                                                      r4 = (object*)TC_NIL;
                                                      r3[5] = r4;
                                                      r4 = ((object*) lenv) [5];
                                                      r5 = (object*) &constructed_259;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[6] = r4;
                                                      r4 = ((object*) lenv) [6];
                                                      r5 = ((object*) lenv) [4];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[7] = r4;
                                                      r4 = ((object*) lenv) [7];
                                                      r5 = (object*) &constructed_258;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[8] = r4;
                                                      r4 = ((object*) lenv) [8];
                                                      r5 = ((object*) lenv) [3];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[9] = r4;
                                                      r4 = ((object*) lenv) [9];
                                                      r5 = (object*) &constructed_257;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[10] = r4;
                                                      r4 = ((object*) lenv) [10];
                                                      r5 = ((object*) lenv) [2];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[11] = r4;
                                                      r3 = ((object*) lenv) [11];
                                                      r4 = (object*) &constructed_256;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r4;
                                                      t[2] = r3;
                                                      r3 = t;
                                                      lenv = lenv[1];
                                                      r2[2] = r3;
                                                      r3 = top[47];
                                                      t = allocate (TC_SAVE, 5);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2728; t[4] = r0; t[5] = r1; k = t;
                                                      r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                      L2728:
                                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                      r2 = result;
                                                      r1[2] = r2;
                                                      r2 = ((object******************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                                                      r2 = ((pxll_vector*)r2)->val[4];
                                                      t = allocate (TC_SAVE, 4);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2729; t[4] = r0; k = t;
                                                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                                      L2729:
                                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                                      r1 = result;
                                                      lenv = lenv[1];
                                                    } break;
                                                    default: {
                                                      r1 = allocate (TC_TUPLE, 2);
                                                      r2 = allocate (TC_TUPLE, 2);
                                                      r3 = allocate (TC_TUPLE, 14);
                                                      r3[1] = lenv; lenv = r3;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = ((object****) lenv) [1][1][1][2];
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2730; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2730:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[2] = r4;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = ((object****) lenv) [1][1][1][2];
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2731; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2731:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[3] = r4;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = ((object**) lenv) [1][2];
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2732; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2732:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[4] = r4;
                                                      r4 = allocate (TC_TUPLE, 2);
                                                      r5 = ((object***) lenv) [1][1][2];
                                                      r4[2] = r5;
                                                      r5 = top[56];
                                                      t = allocate (TC_SAVE, 7);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2733; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                      r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                      L2733:
                                                      r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                      r4 = result;
                                                      r3[5] = r4;
                                                      r4 = (object*)TC_NIL;
                                                      r3[6] = r4;
                                                      r4 = ((object*) lenv) [6];
                                                      r5 = (object*) &constructed_264;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[7] = r4;
                                                      r4 = ((object*) lenv) [7];
                                                      r5 = ((object*) lenv) [5];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[8] = r4;
                                                      r4 = ((object*) lenv) [8];
                                                      r5 = (object*) &constructed_263;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[9] = r4;
                                                      r4 = ((object*) lenv) [9];
                                                      r5 = ((object*) lenv) [4];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[10] = r4;
                                                      r4 = ((object*) lenv) [10];
                                                      r5 = (object*) &constructed_262;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[11] = r4;
                                                      r4 = ((object*) lenv) [11];
                                                      r5 = ((object*) lenv) [3];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[12] = r4;
                                                      r4 = ((object*) lenv) [12];
                                                      r5 = (object*) &constructed_261;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[13] = r4;
                                                      r4 = ((object*) lenv) [13];
                                                      r5 = ((object*) lenv) [2];
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r5;
                                                      t[2] = r4;
                                                      r4 = t;
                                                      r3[14] = r4;
                                                      r3 = ((object*) lenv) [14];
                                                      r4 = (object*) &constructed_260;
                                                      t = alloc_no_clear (TC_PAIR,2);
                                                      t[1] = r4;
                                                      t[2] = r3;
                                                      r3 = t;
                                                      lenv = lenv[1];
                                                      r2[2] = r3;
                                                      r3 = top[47];
                                                      t = allocate (TC_SAVE, 5);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2734; t[4] = r0; t[5] = r1; k = t;
                                                      r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                      L2734:
                                                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                      r2 = result;
                                                      r1[2] = r2;
                                                      r2 = ((object*****************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                                                      r2 = ((pxll_vector*)r2)->val[4];
                                                      t = allocate (TC_SAVE, 4);
                                                      t[1] = k; t[2] = lenv; t[3] = &&L2735; t[4] = r0; k = t;
                                                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                                      L2735:
                                                      r0 = k[4]; lenv = k[2]; k = k[1];
                                                      r1 = result;
                                                    } break;
                                                  }
                                                  lenv = lenv[1];
                                                } break;
                                                default: {
                                                  lenv = ((object *******)lenv)[1][1][1][1][1][1][1];
                                                  goto fatbar_21;
                                                }
                                              }
                                              lenv = lenv[1];
                                            } break;
                                            default: {
                                              lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                              goto fatbar_21;
                                            }
                                          }
                                          lenv = lenv[1];
                                        } break;
                                        default: {
                                          lenv = ((object *****)lenv)[1][1][1][1][1];
                                          goto fatbar_21;
                                        }
                                      }
                                    } break;
                                    default: {
                                      lenv = ((object *****)lenv)[1][1][1][1][1];
                                      goto fatbar_21;
                                    }
                                  }
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object ****)lenv)[1][1][1][1];
                                  goto fatbar_21;
                                }
                              }
                              lenv = lenv[1];
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_21;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_21;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_21;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_21;
                }
              }
              goto fatbar_21_over;
              fatbar_21:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2736; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2736:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_21_over: ;
            } else {
              goto fatbar_20;
            }
            goto fatbar_20_over;
            fatbar_20:
            r1 = (object *) constructed_277[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][3];
              switch (get_case (r1)) {
                case (UOTAG(0)): {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r1 = ((object*) lenv) [2];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [2];
                      switch (get_case (r1)) {
                        case (UOTAG(1)): {
                          r1 = allocate (TC_TUPLE, 2);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object**) lenv) [1][2];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r1 = ((object**) lenv) [1][3];
                          switch (get_case (r1)) {
                            case (TC_PAIR): {
                              r1 = allocate (TC_TUPLE, 3);
                              r1[1] = lenv; lenv = r1;
                              r2 = ((object***) lenv) [1][1][3];
                              r2 = UOBJ_GET(r2,0);
                              r1[2] = r2;
                              r2 = ((object***) lenv) [1][1][3];
                              r2 = UOBJ_GET(r2,1);
                              r1[3] = r2;
                              r1 = ((object*) lenv) [2];
                              switch (get_case (r1)) {
                                case (UOTAG(0)): {
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = ((object**) lenv) [1][2];
                                  r2 = UOBJ_GET(r2,0);
                                  r1[2] = r2;
                                  r1 = ((object**) lenv) [1][3];
                                  switch (get_case (r1)) {
                                    case (TC_NIL): {
                                      r1 = ((object********) lenv) [1][1][1][1][1][1][1][5];
                                      switch (get_case (r1)) {
                                        case (TC_PAIR): {
                                          r1 = allocate (TC_TUPLE, 3);
                                          r1[1] = lenv; lenv = r1;
                                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                                          r2 = UOBJ_GET(r2,0);
                                          r1[2] = r2;
                                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                                          r2 = UOBJ_GET(r2,1);
                                          r1[3] = r2;
                                          r1 = ((object*) lenv) [3];
                                          switch (get_case (r1)) {
                                            case (TC_NIL): {
                                              r1 = allocate (TC_TUPLE, 3);
                                              r1[1] = lenv; lenv = r1;
                                              r2 = allocate (TC_TUPLE, 3);
                                              r3 = ((object*****) lenv) [1][1][1][1][2];
                                              r2[2] = r3;
                                              r3 = ((object****************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][2];
                                              r2[3] = r3;
                                              r3 = top[9];
                                              t = allocate (TC_SAVE, 5);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2737; t[4] = r0; t[5] = r1; k = t;
                                              r2[1] = r3[2]; lenv = r2; goto FUN_lookup_label_code_253;
                                              L2737:
                                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                              r2 = result;
                                              r1[2] = r2;
                                              r2 = allocate (TC_TUPLE, 2);
                                              r3 = ((object***) lenv) [1][1][2];
                                              r2[2] = r3;
                                              r3 = ((object*************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][5];
                                              t = allocate (TC_SAVE, 5);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2738; t[4] = r0; t[5] = r1; k = t;
                                              r2[1] = r3[2]; lenv = r2; goto FUN_guess_record_type_4247_i4683_i4727;
                                              L2738:
                                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                              r2 = result;
                                              r1[3] = r2;
                                              r1 = ((object*) lenv) [3];
                                              switch (get_case (r1)) {
                                                case (UOTAG(0)): {
                                                  r1 = allocate (TC_TUPLE, 2);
                                                  r1[1] = lenv; lenv = r1;
                                                  r2 = ((object**) lenv) [1][3];
                                                  r2 = UOBJ_GET(r2,0);
                                                  r1[2] = r2;
                                                  r1 = allocate (TC_TUPLE, 2);
                                                  r2 = allocate (TC_TUPLE, 2);
                                                  r3 = allocate (TC_TUPLE, 11);
                                                  r3[1] = lenv; lenv = r3;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2739; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2739:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[2] = r4;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = ((object****) lenv) [1][1][1][2];
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2740; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2740:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[3] = r4;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = allocate (TC_TUPLE, 3);
                                                  r6 = ((object*******) lenv) [1][1][1][1][1][1][2];
                                                  r5[2] = r6;
                                                  r6 = ((object**) lenv) [1][2];
                                                  r5[3] = r6;
                                                  r6 = top[28];
                                                  t = allocate (TC_SAVE, 8);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2741; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
                                                  r5[1] = r6[2]; lenv = r5; goto FUN_index_eq_322;
                                                  L2741:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
                                                  r5 = result;
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2742; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2742:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[4] = r4;
                                                  r4 = (object*)TC_NIL;
                                                  r3[5] = r4;
                                                  r4 = ((object*) lenv) [5];
                                                  r5 = (object*) &constructed_270;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[6] = r4;
                                                  r4 = ((object*) lenv) [6];
                                                  r5 = ((object*) lenv) [4];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[7] = r4;
                                                  r4 = ((object*) lenv) [7];
                                                  r5 = (object*) &constructed_269;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[8] = r4;
                                                  r4 = ((object*) lenv) [8];
                                                  r5 = ((object*) lenv) [3];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[9] = r4;
                                                  r4 = ((object*) lenv) [9];
                                                  r5 = (object*) &constructed_268;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[10] = r4;
                                                  r4 = ((object*) lenv) [10];
                                                  r5 = ((object*) lenv) [2];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[11] = r4;
                                                  r3 = ((object*) lenv) [11];
                                                  r4 = (object*) &constructed_267;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r4;
                                                  t[2] = r3;
                                                  r3 = t;
                                                  lenv = lenv[1];
                                                  r2[2] = r3;
                                                  r3 = top[47];
                                                  t = allocate (TC_SAVE, 5);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2743; t[4] = r0; t[5] = r1; k = t;
                                                  r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                  L2743:
                                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                  r2 = result;
                                                  r1[2] = r2;
                                                  r2 = ((object*****************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                                                  r2 = ((pxll_vector*)r2)->val[4];
                                                  t = allocate (TC_SAVE, 4);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2744; t[4] = r0; k = t;
                                                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                                  L2744:
                                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                                  r1 = result;
                                                  lenv = lenv[1];
                                                } break;
                                                default: {
                                                  r1 = allocate (TC_TUPLE, 2);
                                                  r2 = allocate (TC_TUPLE, 2);
                                                  r3 = allocate (TC_TUPLE, 14);
                                                  r3[1] = lenv; lenv = r3;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2745; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2745:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[2] = r4;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = ((object***) lenv) [1][1][2];
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2746; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2746:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[3] = r4;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = ((object***) lenv) [1][1][2];
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2747; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2747:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[4] = r4;
                                                  r4 = allocate (TC_TUPLE, 2);
                                                  r5 = ((object**) lenv) [1][2];
                                                  r4[2] = r5;
                                                  r5 = top[56];
                                                  t = allocate (TC_SAVE, 7);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2748; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                                  L2748:
                                                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                                  r4 = result;
                                                  r3[5] = r4;
                                                  r4 = (object*)TC_NIL;
                                                  r3[6] = r4;
                                                  r4 = ((object*) lenv) [6];
                                                  r5 = (object*) &constructed_275;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[7] = r4;
                                                  r4 = ((object*) lenv) [7];
                                                  r5 = ((object*) lenv) [5];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[8] = r4;
                                                  r4 = ((object*) lenv) [8];
                                                  r5 = (object*) &constructed_274;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[9] = r4;
                                                  r4 = ((object*) lenv) [9];
                                                  r5 = ((object*) lenv) [4];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[10] = r4;
                                                  r4 = ((object*) lenv) [10];
                                                  r5 = (object*) &constructed_273;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[11] = r4;
                                                  r4 = ((object*) lenv) [11];
                                                  r5 = ((object*) lenv) [3];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[12] = r4;
                                                  r4 = ((object*) lenv) [12];
                                                  r5 = (object*) &constructed_272;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[13] = r4;
                                                  r4 = ((object*) lenv) [13];
                                                  r5 = ((object*) lenv) [2];
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r5;
                                                  t[2] = r4;
                                                  r4 = t;
                                                  r3[14] = r4;
                                                  r3 = ((object*) lenv) [14];
                                                  r4 = (object*) &constructed_271;
                                                  t = alloc_no_clear (TC_PAIR,2);
                                                  t[1] = r4;
                                                  t[2] = r3;
                                                  r3 = t;
                                                  lenv = lenv[1];
                                                  r2[2] = r3;
                                                  r3 = top[47];
                                                  t = allocate (TC_SAVE, 5);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2749; t[4] = r0; t[5] = r1; k = t;
                                                  r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                                  L2749:
                                                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                                  r2 = result;
                                                  r1[2] = r2;
                                                  r2 = ((object****************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                                                  r2 = ((pxll_vector*)r2)->val[4];
                                                  t = allocate (TC_SAVE, 4);
                                                  t[1] = k; t[2] = lenv; t[3] = &&L2750; t[4] = r0; k = t;
                                                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                                  L2750:
                                                  r0 = k[4]; lenv = k[2]; k = k[1];
                                                  r1 = result;
                                                } break;
                                              }
                                              lenv = lenv[1];
                                            } break;
                                            default: {
                                              lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                              goto fatbar_23;
                                            }
                                          }
                                          lenv = lenv[1];
                                        } break;
                                        default: {
                                          lenv = ((object *****)lenv)[1][1][1][1][1];
                                          goto fatbar_23;
                                        }
                                      }
                                    } break;
                                    default: {
                                      lenv = ((object *****)lenv)[1][1][1][1][1];
                                      goto fatbar_23;
                                    }
                                  }
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object ****)lenv)[1][1][1][1];
                                  goto fatbar_23;
                                }
                              }
                              lenv = lenv[1];
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_23;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_23;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_23;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_23;
                }
              }
              goto fatbar_23_over;
              fatbar_23:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2751; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2751:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_23_over: ;
            } else {
              goto fatbar_22;
            }
            goto fatbar_22_over;
            fatbar_22:
            r1 = (object *) constructed_286[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [3];
                      switch (get_case (r1)) {
                        case (TC_PAIR): {
                          r1 = allocate (TC_TUPLE, 3);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object**) lenv) [1][3];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][3];
                          r2 = UOBJ_GET(r2,1);
                          r1[3] = r2;
                          r1 = ((object*) lenv) [3];
                          switch (get_case (r1)) {
                            case (TC_NIL): {
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = allocate (TC_TUPLE, 2);
                              r3 = allocate (TC_TUPLE, 8);
                              r3[1] = lenv; lenv = r3;
                              r4 = allocate (TC_TUPLE, 2);
                              r5 = ((object****) lenv) [1][1][1][2];
                              r4[2] = r5;
                              r5 = top[56];
                              t = allocate (TC_SAVE, 7);
                              t[1] = k; t[2] = lenv; t[3] = &&L2752; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                              L2752:
                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                              r4 = result;
                              r3[2] = r4;
                              r4 = allocate (TC_TUPLE, 2);
                              r5 = ((object***) lenv) [1][1][2];
                              r4[2] = r5;
                              r5 = top[56];
                              t = allocate (TC_SAVE, 7);
                              t[1] = k; t[2] = lenv; t[3] = &&L2753; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                              L2753:
                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                              r4 = result;
                              r3[3] = r4;
                              r4 = (object*)TC_NIL;
                              r3[4] = r4;
                              r4 = ((object*) lenv) [4];
                              r5 = (object*) &constructed_284;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[5] = r4;
                              r4 = ((object*) lenv) [5];
                              r5 = ((object*) lenv) [3];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[6] = r4;
                              r4 = ((object*) lenv) [6];
                              r5 = (object*) &constructed_283;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[7] = r4;
                              r4 = ((object*) lenv) [7];
                              r5 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[8] = r4;
                              r3 = ((object*) lenv) [8];
                              r4 = (object*) &constructed_282;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r4;
                              t[2] = r3;
                              r3 = t;
                              lenv = lenv[1];
                              r2[2] = r3;
                              r3 = top[47];
                              t = allocate (TC_SAVE, 5);
                              t[1] = k; t[2] = lenv; t[3] = &&L2754; t[4] = r0; t[5] = r1; k = t;
                              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                              L2754:
                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                              r2 = result;
                              r1[2] = r2;
                              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                              r2 = ((pxll_vector*)r2)->val[4];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2755; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto *r2[1];
                              L2755:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = allocate (TC_TUPLE, 2);
                              r2 = allocate (TC_TUPLE, 2);
                              r3 = allocate (TC_TUPLE, 11);
                              r3[1] = lenv; lenv = r3;
                              r4 = allocate (TC_TUPLE, 2);
                              r5 = ((object****) lenv) [1][1][1][2];
                              r4[2] = r5;
                              r5 = top[56];
                              t = allocate (TC_SAVE, 7);
                              t[1] = k; t[2] = lenv; t[3] = &&L2756; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                              L2756:
                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                              r4 = result;
                              r3[2] = r4;
                              r4 = allocate (TC_TUPLE, 2);
                              r5 = ((object***) lenv) [1][1][2];
                              r4[2] = r5;
                              r5 = top[56];
                              t = allocate (TC_SAVE, 7);
                              t[1] = k; t[2] = lenv; t[3] = &&L2757; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                              L2757:
                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                              r4 = result;
                              r3[3] = r4;
                              r4 = allocate (TC_TUPLE, 2);
                              r5 = ((object**) lenv) [1][2];
                              r4[2] = r5;
                              r5 = top[56];
                              t = allocate (TC_SAVE, 7);
                              t[1] = k; t[2] = lenv; t[3] = &&L2758; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                              L2758:
                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                              r4 = result;
                              r3[4] = r4;
                              r4 = (object*)TC_NIL;
                              r3[5] = r4;
                              r4 = ((object*) lenv) [5];
                              r5 = (object*) &constructed_281;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[6] = r4;
                              r4 = ((object*) lenv) [6];
                              r5 = ((object*) lenv) [4];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[7] = r4;
                              r4 = ((object*) lenv) [7];
                              r5 = (object*) &constructed_280;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[8] = r4;
                              r4 = ((object*) lenv) [8];
                              r5 = ((object*) lenv) [3];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[9] = r4;
                              r4 = ((object*) lenv) [9];
                              r5 = (object*) &constructed_279;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[10] = r4;
                              r4 = ((object*) lenv) [10];
                              r5 = ((object*) lenv) [2];
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r5;
                              t[2] = r4;
                              r4 = t;
                              r3[11] = r4;
                              r3 = ((object*) lenv) [11];
                              r4 = (object*) &constructed_278;
                              t = alloc_no_clear (TC_PAIR,2);
                              t[1] = r4;
                              t[2] = r3;
                              r3 = t;
                              lenv = lenv[1];
                              r2[2] = r3;
                              r3 = top[47];
                              t = allocate (TC_SAVE, 5);
                              t[1] = k; t[2] = lenv; t[3] = &&L2759; t[4] = r0; t[5] = r1; k = t;
                              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                              L2759:
                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                              r2 = result;
                              r1[2] = r2;
                              r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                              r2 = ((pxll_vector*)r2)->val[4];
                              t = allocate (TC_SAVE, 4);
                              t[1] = k; t[2] = lenv; t[3] = &&L2760; t[4] = r0; k = t;
                              r1[1] = r2[2]; lenv = r1; goto *r2[1];
                              L2760:
                              r0 = k[4]; lenv = k[2]; k = k[1];
                              r1 = result;
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_25;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_25;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_25;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_25;
                }
              }
              goto fatbar_25_over;
              fatbar_25:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2761; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2761:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_25_over: ;
            } else {
              goto fatbar_24;
            }
            goto fatbar_24_over;
            fatbar_24:
            r1 = (object *) constructed_295[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [3];
                      switch (get_case (r1)) {
                        case (TC_NIL): {
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 8);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2762; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2762:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object**) lenv) [1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2763; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2763:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[3] = r4;
                          r4 = (object*)TC_NIL;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = (object*) &constructed_293;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r4 = ((object*) lenv) [5];
                          r5 = ((object*) lenv) [3];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[6] = r4;
                          r4 = ((object*) lenv) [6];
                          r5 = (object*) &constructed_292;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[7] = r4;
                          r4 = ((object*) lenv) [7];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[8] = r4;
                          r3 = ((object*) lenv) [8];
                          r4 = (object*) &constructed_291;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2764; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2764:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2765; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2765:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 11);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object****) lenv) [1][1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2766; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2766:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2767; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2767:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[3] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object**) lenv) [1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2768; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2768:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[4] = r4;
                          r4 = (object*)TC_NIL;
                          r3[5] = r4;
                          r4 = ((object*) lenv) [5];
                          r5 = (object*) &constructed_290;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[6] = r4;
                          r4 = ((object*) lenv) [6];
                          r5 = ((object*) lenv) [4];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[7] = r4;
                          r4 = ((object*) lenv) [7];
                          r5 = (object*) &constructed_289;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[8] = r4;
                          r4 = ((object*) lenv) [8];
                          r5 = ((object*) lenv) [3];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[9] = r4;
                          r4 = ((object*) lenv) [9];
                          r5 = (object*) &constructed_288;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[10] = r4;
                          r4 = ((object*) lenv) [10];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[11] = r4;
                          r3 = ((object*) lenv) [11];
                          r4 = (object*) &constructed_287;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2769; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2769:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2770; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2770:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_27;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_27;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_27;
                }
              }
              goto fatbar_27_over;
              fatbar_27:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2771; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2771:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_27_over: ;
            } else {
              goto fatbar_26;
            }
            goto fatbar_26_over;
            fatbar_26:
            r1 = (object *) constructed_308[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][5];
              switch (get_case (r1)) {
                case (TC_PAIR): {
                  r1 = allocate (TC_TUPLE, 3);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][5];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r1 = ((object*) lenv) [3];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][3];
                      r2 = UOBJ_GET(r2,1);
                      r1[3] = r2;
                      r1 = ((object*) lenv) [3];
                      switch (get_case (r1)) {
                        case (TC_NIL): {
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 8);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2772; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2772:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object****) lenv) [1][1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2773; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2773:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[3] = r4;
                          r4 = (object*)TC_NIL;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = (object*) &constructed_306;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r4 = ((object*) lenv) [5];
                          r5 = ((object*) lenv) [3];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[6] = r4;
                          r4 = ((object*) lenv) [6];
                          r5 = (object*) &constructed_305;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[7] = r4;
                          r4 = ((object*) lenv) [7];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[8] = r4;
                          r3 = ((object*) lenv) [8];
                          r4 = (object*) &constructed_304;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2774; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2774:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2775; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2775:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 5);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2776; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2776:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = (object*)TC_NIL;
                          r3[3] = r4;
                          r4 = ((object*) lenv) [3];
                          r5 = (object*) &constructed_303;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r3 = ((object*) lenv) [5];
                          r4 = (object*) &constructed_302;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2777; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2777:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2778; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2778:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 8);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object***) lenv) [1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2779; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2779:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object**) lenv) [1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2780; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2780:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[3] = r4;
                          r4 = (object*)TC_NIL;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = (object*) &constructed_301;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r4 = ((object*) lenv) [5];
                          r5 = ((object*) lenv) [3];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[6] = r4;
                          r4 = ((object*) lenv) [6];
                          r5 = (object*) &constructed_300;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[7] = r4;
                          r4 = ((object*) lenv) [7];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[8] = r4;
                          r3 = ((object*) lenv) [8];
                          r4 = (object*) &constructed_299;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2781; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2781:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2782; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2782:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = allocate (TC_TUPLE, 2);
                          r3 = allocate (TC_TUPLE, 5);
                          r3[1] = lenv; lenv = r3;
                          r4 = allocate (TC_TUPLE, 2);
                          r5 = ((object****) lenv) [1][1][1][2];
                          r4[2] = r5;
                          r5 = top[56];
                          t = allocate (TC_SAVE, 7);
                          t[1] = k; t[2] = lenv; t[3] = &&L2783; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                          r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                          L2783:
                          r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                          r4 = result;
                          r3[2] = r4;
                          r4 = (object*)TC_NIL;
                          r3[3] = r4;
                          r4 = ((object*) lenv) [3];
                          r5 = (object*) &constructed_298;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[4] = r4;
                          r4 = ((object*) lenv) [4];
                          r5 = ((object*) lenv) [2];
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r5;
                          t[2] = r4;
                          r4 = t;
                          r3[5] = r4;
                          r3 = ((object*) lenv) [5];
                          r4 = (object*) &constructed_297;
                          t = alloc_no_clear (TC_PAIR,2);
                          t[1] = r4;
                          t[2] = r3;
                          r3 = t;
                          lenv = lenv[1];
                          r2[2] = r3;
                          r3 = top[47];
                          t = allocate (TC_SAVE, 5);
                          t[1] = k; t[2] = lenv; t[3] = &&L2784; t[4] = r0; t[5] = r1; k = t;
                          r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                          L2784:
                          r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                          r2 = result;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2785; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2785:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = allocate (TC_TUPLE, 2);
                          r2 = (object*) &constructed_296;
                          r1[2] = r2;
                          r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                          r2 = ((pxll_vector*)r2)->val[4];
                          t = allocate (TC_SAVE, 4);
                          t[1] = k; t[2] = lenv; t[3] = &&L2786; t[4] = r0; k = t;
                          r1[1] = r2[2]; lenv = r1; goto *r2[1];
                          L2786:
                          r0 = k[4]; lenv = k[2]; k = k[1];
                          r1 = result;
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_29;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_29;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_29;
                }
              }
              goto fatbar_29_over;
              fatbar_29:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2787; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2787:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_29_over: ;
            } else {
              goto fatbar_28;
            }
            goto fatbar_28_over;
            fatbar_28:
            r1 = (object *) constructed_314[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][3];
              switch (get_case (r1)) {
                case (UOTAG(0)): {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r1 = ((object*) lenv) [2];
                  switch (get_case (r1)) {
                    case (TC_PAIR): {
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][2];
                      r2 = UOBJ_GET(r2,1);
                      r1[2] = r2;
                      r1 = ((object*) lenv) [2];
                      switch (get_case (r1)) {
                        case (TC_PAIR): {
                          r1 = allocate (TC_TUPLE, 3);
                          r1[1] = lenv; lenv = r1;
                          r2 = ((object**) lenv) [1][2];
                          r2 = UOBJ_GET(r2,0);
                          r1[2] = r2;
                          r2 = ((object**) lenv) [1][2];
                          r2 = UOBJ_GET(r2,1);
                          r1[3] = r2;
                          r1 = ((object*) lenv) [2];
                          switch (get_case (r1)) {
                            case (UOTAG(5)): {
                              r1 = allocate (TC_TUPLE, 2);
                              r1[1] = lenv; lenv = r1;
                              r2 = ((object**) lenv) [1][2];
                              r2 = UOBJ_GET(r2,0);
                              r1[2] = r2;
                              r1 = ((object**) lenv) [1][3];
                              switch (get_case (r1)) {
                                case (TC_PAIR): {
                                  r1 = allocate (TC_TUPLE, 2);
                                  r1[1] = lenv; lenv = r1;
                                  r2 = ((object***) lenv) [1][1][3];
                                  r2 = UOBJ_GET(r2,1);
                                  r1[2] = r2;
                                  r1 = ((object*) lenv) [2];
                                  switch (get_case (r1)) {
                                    case (TC_NIL): {
                                      r1 = ((object********) lenv) [1][1][1][1][1][1][1][5];
                                      switch (get_case (r1)) {
                                        case (TC_PAIR): {
                                          r1 = allocate (TC_TUPLE, 3);
                                          r1[1] = lenv; lenv = r1;
                                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                                          r2 = UOBJ_GET(r2,0);
                                          r1[2] = r2;
                                          r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                                          r2 = UOBJ_GET(r2,1);
                                          r1[3] = r2;
                                          r1 = ((object*) lenv) [3];
                                          switch (get_case (r1)) {
                                            case (TC_NIL): {
                                              r1 = allocate (TC_TUPLE, 2);
                                              r2 = allocate (TC_TUPLE, 2);
                                              r3 = allocate (TC_TUPLE, 11);
                                              r3[1] = lenv; lenv = r3;
                                              r4 = allocate (TC_TUPLE, 2);
                                              r5 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                                              r4[2] = r5;
                                              r5 = top[56];
                                              t = allocate (TC_SAVE, 7);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2788; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                              L2788:
                                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                              r4 = result;
                                              r3[2] = r4;
                                              r4 = allocate (TC_TUPLE, 2);
                                              r5 = ((object**) lenv) [1][2];
                                              r4[2] = r5;
                                              r5 = top[56];
                                              t = allocate (TC_SAVE, 7);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2789; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                              L2789:
                                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                              r4 = result;
                                              r3[3] = r4;
                                              r4 = allocate (TC_TUPLE, 2);
                                              r5 = ((object****) lenv) [1][1][1][2];
                                              r4[2] = r5;
                                              r5 = top[56];
                                              t = allocate (TC_SAVE, 7);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2790; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                                              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                                              L2790:
                                              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                                              r4 = result;
                                              r3[4] = r4;
                                              r4 = (object*)TC_NIL;
                                              r3[5] = r4;
                                              r4 = ((object*) lenv) [5];
                                              r5 = (object*) &constructed_312;
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r5;
                                              t[2] = r4;
                                              r4 = t;
                                              r3[6] = r4;
                                              r4 = ((object*) lenv) [6];
                                              r5 = ((object*) lenv) [4];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r5;
                                              t[2] = r4;
                                              r4 = t;
                                              r3[7] = r4;
                                              r4 = ((object*) lenv) [7];
                                              r5 = (object*) &constructed_311;
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r5;
                                              t[2] = r4;
                                              r4 = t;
                                              r3[8] = r4;
                                              r4 = ((object*) lenv) [8];
                                              r5 = ((object*) lenv) [3];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r5;
                                              t[2] = r4;
                                              r4 = t;
                                              r3[9] = r4;
                                              r4 = ((object*) lenv) [9];
                                              r5 = (object*) &constructed_310;
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r5;
                                              t[2] = r4;
                                              r4 = t;
                                              r3[10] = r4;
                                              r4 = ((object*) lenv) [10];
                                              r5 = ((object*) lenv) [2];
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r5;
                                              t[2] = r4;
                                              r4 = t;
                                              r3[11] = r4;
                                              r3 = ((object*) lenv) [11];
                                              r4 = (object*) &constructed_309;
                                              t = alloc_no_clear (TC_PAIR,2);
                                              t[1] = r4;
                                              t[2] = r3;
                                              r3 = t;
                                              lenv = lenv[1];
                                              r2[2] = r3;
                                              r3 = top[47];
                                              t = allocate (TC_SAVE, 5);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2791; t[4] = r0; t[5] = r1; k = t;
                                              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                                              L2791:
                                              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                                              r2 = result;
                                              r1[2] = r2;
                                              r2 = ((object***************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                                              r2 = ((pxll_vector*)r2)->val[4];
                                              t = allocate (TC_SAVE, 4);
                                              t[1] = k; t[2] = lenv; t[3] = &&L2792; t[4] = r0; k = t;
                                              r1[1] = r2[2]; lenv = r1; goto *r2[1];
                                              L2792:
                                              r0 = k[4]; lenv = k[2]; k = k[1];
                                              r1 = result;
                                            } break;
                                            default: {
                                              lenv = ((object ******)lenv)[1][1][1][1][1][1];
                                              goto fatbar_31;
                                            }
                                          }
                                          lenv = lenv[1];
                                        } break;
                                        default: {
                                          lenv = ((object *****)lenv)[1][1][1][1][1];
                                          goto fatbar_31;
                                        }
                                      }
                                    } break;
                                    default: {
                                      lenv = ((object *****)lenv)[1][1][1][1][1];
                                      goto fatbar_31;
                                    }
                                  }
                                  lenv = lenv[1];
                                } break;
                                default: {
                                  lenv = ((object ****)lenv)[1][1][1][1];
                                  goto fatbar_31;
                                }
                              }
                              lenv = lenv[1];
                            } break;
                            default: {
                              lenv = ((object ***)lenv)[1][1][1];
                              goto fatbar_31;
                            }
                          }
                          lenv = lenv[1];
                        } break;
                        default: {
                          lenv = ((object **)lenv)[1][1];
                          goto fatbar_31;
                        }
                      }
                      lenv = lenv[1];
                    } break;
                    default: {
                      lenv = ((object *)lenv)[1];
                      goto fatbar_31;
                    }
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_31;
                }
              }
              goto fatbar_31_over;
              fatbar_31:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2793; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2793:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_31_over: ;
            } else {
              goto fatbar_30;
            }
            goto fatbar_30_over;
            fatbar_30:
            r1 = (object *) constructed_328[0];
            r2 = ((object***) lenv) [1][1][2];
            if PXLL_IS_TRUE(PXLL_TEST(r2==r1)) {
              r1 = ((object***) lenv) [1][1][3];
              switch (get_case (r1)) {
                case (UOTAG(9)): {
                  r1 = allocate (TC_TUPLE, 4);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r2 = ((object****) lenv) [1][1][1][3];
                  r2 = UOBJ_GET(r2,1);
                  r1[3] = r2;
                  r2 = allocate (TC_TUPLE, 3);
                  r3 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][2];
                  r3 = ((pxll_vector*)r3)->val[2];
                  r2[2] = r3;
                  r3 = ((object*) lenv) [2];
                  r2[3] = r3;
                  r3 = top[65];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2794; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flookup_394;
                  L2794:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[4] = r2;
                  r1 = ((object*) lenv) [4];
                  switch (get_case (r1)) {
                    case (UITAG(1)): {
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = (object*) &constructed_315;
                      r1[2] = r2;
                      r2 = ((object*) lenv) [2];
                      r1[3] = r2;
                      r2 = top[15];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2795; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                      L2795:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                    } break;
                    default: {
                      r1 = allocate (TC_TUPLE, 3);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object**) lenv) [1][4];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r2 = allocate (TC_TUPLE, 2);
                      r3 = ((object**) lenv) [1][3];
                      r2[2] = r3;
                      r3 = ((object*) lenv) [2];
                      r3 = ((pxll_vector*)r3)->val[1];
                      t = allocate (TC_SAVE, 5);
                      t[1] = k; t[2] = lenv; t[3] = &&L2796; t[4] = r0; t[5] = r1; k = t;
                      r2[1] = r3[2]; lenv = r2; goto *r3[1];
                      L2796:
                      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                      r2 = result;
                      r1[3] = r2;
                      r1 = (object *) 1;
                      r2 = ((object***) lenv) [1][1][3];
                      if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = allocate (TC_TUPLE, 2);
                        r3 = allocate (TC_TUPLE, 8);
                        r3[1] = lenv; lenv = r3;
                        r4 = allocate (TC_TUPLE, 2);
                        r5 = ((object****) lenv) [1][1][1][2];
                        r4[2] = r5;
                        r5 = top[56];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L2797; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                        L2797:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = allocate (TC_TUPLE, 4);
                        r5 = ((object***) lenv) [1][1][2];
                        r4[2] = r5;
                        r5 = ((object***) lenv) [1][1][3];
                        r4[3] = r5;
                        r5 = ((object**) lenv) [1][3];
                        r5 = ((pxll_vector*)r5)->val[1];
                        r4[4] = r5;
                        r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][7];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L2798; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_get_uitag_4249_i4683_i4727;
                        L2798:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[3] = r4;
                        r4 = (object*)TC_NIL;
                        r3[4] = r4;
                        r4 = ((object*) lenv) [4];
                        r5 = (object*) &constructed_318;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[5] = r4;
                        r4 = ((object*) lenv) [5];
                        r5 = ((object*) lenv) [3];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[6] = r4;
                        r4 = ((object*) lenv) [6];
                        r5 = (object*) &constructed_317;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[7] = r4;
                        r4 = ((object*) lenv) [7];
                        r5 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[8] = r4;
                        r3 = ((object*) lenv) [8];
                        r4 = (object*) &constructed_316;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r4;
                        t[2] = r3;
                        r3 = t;
                        lenv = lenv[1];
                        r2[2] = r3;
                        r3 = top[47];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2799; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                        L2799:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                        r1[2] = r2;
                        r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                        r2 = ((pxll_vector*)r2)->val[4];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2800; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto *r2[1];
                        L2800:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                      } else {
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = allocate (TC_TUPLE, 2);
                        r3 = allocate (TC_TUPLE, 8);
                        r3[1] = lenv; lenv = r3;
                        r4 = allocate (TC_TUPLE, 4);
                        r5 = ((object***) lenv) [1][1][2];
                        r4[2] = r5;
                        r5 = ((object***) lenv) [1][1][3];
                        r4[3] = r5;
                        r5 = ((object**) lenv) [1][3];
                        r5 = ((pxll_vector*)r5)->val[1];
                        r4[4] = r5;
                        r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][6];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L2801; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_get_uotag_4248_i4683_i4727;
                        L2801:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = allocate (TC_TUPLE, 2);
                        r5 = ((object****) lenv) [1][1][1][3];
                        r4[2] = r5;
                        r5 = top[56];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L2802; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                        L2802:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[3] = r4;
                        r4 = (object*)TC_NIL;
                        r3[4] = r4;
                        r4 = ((object*) lenv) [4];
                        r5 = (object*) &constructed_326;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[5] = r4;
                        r4 = ((object*) lenv) [5];
                        r5 = ((object*) lenv) [3];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[6] = r4;
                        r4 = ((object*) lenv) [6];
                        r5 = (object*) &constructed_325;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[7] = r4;
                        r4 = ((object*) lenv) [7];
                        r5 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[8] = r4;
                        r3 = ((object*) lenv) [8];
                        r4 = (object*) &constructed_324;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r4;
                        t[2] = r3;
                        r3 = t;
                        lenv = lenv[1];
                        r2[2] = r3;
                        r3 = top[47];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2803; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                        L2803:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                        r1[2] = r2;
                        r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                        r2 = ((pxll_vector*)r2)->val[4];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2804; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto *r2[1];
                        L2804:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = allocate (TC_TUPLE, 2);
                        r1[1] = lenv; lenv = r1;
                        r2 = ((object****) lenv) [1][1][1][3];
                        r1[2] = r2;
                        r1 = allocate (TC_TUPLE, 2);
                        r1[1] = lenv; lenv = r1;
                        // def $loop_4551_i3489_i4683_i4727
                        goto L2805;
                        FUN__24loop_4551_i3489_i4683_i4727:
                          check_heap (0);
                          r0 = ((object***) lenv) [1][1][2];
                          r1 = ((object*) lenv) [2];
                          if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                            r0 = (object *) 14;
                            PXLL_RETURN(0);
                          } else {
                            r0 = allocate (TC_TUPLE, 2);
                            r1 = allocate (TC_TUPLE, 2);
                            r2 = allocate (TC_TUPLE, 8);
                            r2[1] = lenv; lenv = r2;
                            r3 = allocate (TC_TUPLE, 2);
                            r4 = (object *) 3;
                            r5 = ((object**) lenv) [1][2];
                            r4 = box((pxll_int)unbox(r5)+unbox(r4));
                            r3[2] = r4;
                            r4 = top[56];
                            t = allocate (TC_SAVE, 6);
                            t[1] = k; t[2] = lenv; t[3] = &&L2806; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                            r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                            L2806:
                            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                            r3 = result;
                            r2[2] = r3;
                            r3 = allocate (TC_TUPLE, 2);
                            r4 = allocate (TC_TUPLE, 3);
                            r5 = ((object*********) lenv) [1][1][1][1][1][1][1][1][5];
                            r4[2] = r5;
                            r5 = ((object**) lenv) [1][2];
                            r4[3] = r5;
                            r5 = top[27];
                            t = allocate (TC_SAVE, 7);
                            t[1] = k; t[2] = lenv; t[3] = &&L2807; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                            r4[1] = r5[2]; lenv = r4; goto FUN_nth_321;
                            L2807:
                            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                            r4 = result;
                            r3[2] = r4;
                            r4 = top[56];
                            t = allocate (TC_SAVE, 6);
                            t[1] = k; t[2] = lenv; t[3] = &&L2808; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                            r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                            L2808:
                            r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                            r3 = result;
                            r2[3] = r3;
                            r3 = (object*)TC_NIL;
                            r2[4] = r3;
                            r3 = ((object*) lenv) [4];
                            r4 = (object*) &constructed_323;
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r4;
                            t[2] = r3;
                            r3 = t;
                            r2[5] = r3;
                            r3 = ((object*) lenv) [5];
                            r4 = ((object*) lenv) [3];
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r4;
                            t[2] = r3;
                            r3 = t;
                            r2[6] = r3;
                            r3 = ((object*) lenv) [6];
                            r4 = (object*) &constructed_322;
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r4;
                            t[2] = r3;
                            r3 = t;
                            r2[7] = r3;
                            r3 = ((object*) lenv) [7];
                            r4 = ((object*) lenv) [2];
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r4;
                            t[2] = r3;
                            r3 = t;
                            r2[8] = r3;
                            r2 = ((object*) lenv) [8];
                            r3 = (object*) &constructed_321;
                            t = alloc_no_clear (TC_PAIR,2);
                            t[1] = r3;
                            t[2] = r2;
                            r2 = t;
                            lenv = lenv[1];
                            r1[2] = r2;
                            r2 = top[47];
                            t = allocate (TC_SAVE, 4);
                            t[1] = k; t[2] = lenv; t[3] = &&L2809; t[4] = r0; k = t;
                            r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                            L2809:
                            r0 = k[4]; lenv = k[2]; k = k[1];
                            r1 = result;
                            r0[2] = r1;
                            r1 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                            r1 = ((pxll_vector*)r1)->val[4];
                            t = allocate (TC_SAVE, 3);
                            t[1] = k; t[2] = lenv; t[3] = &&L2810; ; k = t;
                            r0[1] = r1[2]; lenv = r0; goto *r1[1];
                            L2810:
                            ; lenv = k[2]; k = k[1];
                            r0 = (object *) 3;
                            r1 = ((object*) lenv) [2];
                            r0 = box((pxll_int)unbox(r1)+unbox(r0));
                            lenv[2] = r0;
                            goto FUN__24loop_4551_i3489_i4683_i4727;
                          }
                          PXLL_RETURN(0);
                        L2805:
                        r2 = allocate (TC_CLOSURE, 2);
                        r2[1] = &&FUN__24loop_4551_i3489_i4683_i4727; r2[2] = lenv;
                        r1[2] = r2;
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = (object *) 1;
                        r1[2] = r2;
                        r2 = ((object*) lenv) [2];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2811; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4551_i3489_i4683_i4727;
                        L2811:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        lenv = lenv[1];
                        lenv = lenv[1];
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = allocate (TC_TUPLE, 2);
                        r3 = allocate (TC_TUPLE, 5);
                        r3[1] = lenv; lenv = r3;
                        r4 = allocate (TC_TUPLE, 2);
                        r5 = ((object****) lenv) [1][1][1][2];
                        r4[2] = r5;
                        r5 = top[56];
                        t = allocate (TC_SAVE, 7);
                        t[1] = k; t[2] = lenv; t[3] = &&L2812; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                        r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                        L2812:
                        r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                        r4 = result;
                        r3[2] = r4;
                        r4 = (object*)TC_NIL;
                        r3[3] = r4;
                        r4 = ((object*) lenv) [3];
                        r5 = (object*) &constructed_320;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[4] = r4;
                        r4 = ((object*) lenv) [4];
                        r5 = ((object*) lenv) [2];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r5;
                        t[2] = r4;
                        r4 = t;
                        r3[5] = r4;
                        r3 = ((object*) lenv) [5];
                        r4 = (object*) &constructed_319;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r4;
                        t[2] = r3;
                        r3 = t;
                        lenv = lenv[1];
                        r2[2] = r3;
                        r3 = top[47];
                        t = allocate (TC_SAVE, 5);
                        t[1] = k; t[2] = lenv; t[3] = &&L2813; t[4] = r0; t[5] = r1; k = t;
                        r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                        L2813:
                        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                        r2 = result;
                        r1[2] = r2;
                        r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                        r2 = ((pxll_vector*)r2)->val[4];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2814; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto *r2[1];
                        L2814:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                      }
                      lenv = lenv[1];
                    } break;
                  }
                  lenv = lenv[1];
                } break;
                default: {
                  goto fatbar_33;
                }
              }
              goto fatbar_33_over;
              fatbar_33:
              r1 = allocate (TC_TUPLE, 3);
              r2 = (object*) &constructed_208;
              r1[2] = r2;
              r2 = ((object***) lenv) [1][1][2];
              r1[3] = r2;
              r2 = top[15];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2815; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
              L2815:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              fatbar_33_over: ;
            } else {
              goto fatbar_32;
            }
            goto fatbar_32_over;
            fatbar_32:
            r1 = allocate (TC_TUPLE, 3);
            r2 = (object*) &constructed_208;
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r1[3] = r2;
            r2 = top[15];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2816; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
            L2816:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            fatbar_32_over: ;
            fatbar_30_over: ;
            fatbar_28_over: ;
            fatbar_26_over: ;
            fatbar_24_over: ;
            fatbar_22_over: ;
            fatbar_20_over: ;
            fatbar_19_over: ;
            fatbar_18_over: ;
            fatbar_17_over: ;
            fatbar_15_over: ;
            fatbar_12_over: ;
            fatbar_10_over: ;
            fatbar_8_over: ;
            lenv = lenv[1];
            lenv = lenv[1];
            r1 = ((object*) lenv) [6];
            lenv = lenv[1];
          } break;
          case (UOTAG(19)): {
            r1 = allocate (TC_TUPLE, 4);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][4];
            r2[2] = r3;
            r3 = top[137];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2817; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_k_2ftarget_523;
            L2817:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = (object *) 1;
            r2 = ((object**) lenv) [1][3];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = ((object***) lenv) [1][1][2];
              r3 = ((object***) lenv) [1][1][3];
              r2 = PXLL_TEST(unbox(r3)==unbox(r2));
              r1[2] = r2;
              r1 = (object *) 6;
              r2 = ((object*) lenv) [2];
              r1 = PXLL_TEST(r2==r1);
              lenv = lenv[1];
            } else {
              r1 = (object *) 6;
            }
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 8);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][2];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2818; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2818:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = ((object***) lenv) [1][1][3];
              r4[2] = r5;
              r5 = top[56];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2819; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
              L2819:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = (object*)TC_NIL;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_331;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_330;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_329;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2820; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2820:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2821; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2821:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 1;
              r2 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>=unbox(r1))) {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object***) lenv) [1][1][2];
                r3 = ((object**) lenv) [1][2];
                r2 = PXLL_TEST(unbox(r3)==unbox(r2));
                r1[2] = r2;
                r1 = (object *) 6;
                r2 = ((object*) lenv) [2];
                r1 = PXLL_TEST(r2==r1);
                lenv = lenv[1];
              } else {
                r1 = (object *) 6;
              }
              if PXLL_IS_TRUE(r1) {
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 2);
                r3 = allocate (TC_TUPLE, 8);
                r3[1] = lenv; lenv = r3;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object**) lenv) [1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2822; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2822:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[2] = r4;
                r4 = allocate (TC_TUPLE, 2);
                r5 = ((object***) lenv) [1][1][2];
                r4[2] = r5;
                r5 = top[56];
                t = allocate (TC_SAVE, 7);
                t[1] = k; t[2] = lenv; t[3] = &&L2823; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                L2823:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                r4 = result;
                r3[3] = r4;
                r4 = (object*)TC_NIL;
                r3[4] = r4;
                r4 = ((object*) lenv) [4];
                r5 = (object*) &constructed_334;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[5] = r4;
                r4 = ((object*) lenv) [5];
                r5 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[6] = r4;
                r4 = ((object*) lenv) [6];
                r5 = (object*) &constructed_333;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[7] = r4;
                r4 = ((object*) lenv) [7];
                r5 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r5;
                t[2] = r4;
                r4 = t;
                r3[8] = r4;
                r3 = ((object*) lenv) [8];
                r4 = (object*) &constructed_332;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                lenv = lenv[1];
                r2[2] = r3;
                r3 = top[47];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2824; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                L2824:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
                r2 = ((pxll_vector*)r2)->val[4];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2825; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto *r2[1];
                L2825:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              } else {
                r1 = (object *) 14;
              }
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [4];
            lenv = lenv[1];
          } break;
          case (UOTAG(20)): {
            r1 = allocate (TC_TUPLE, 5);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [3];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2826; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
            L2826:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2827; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2827:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_340;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_339;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2828; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2828:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2829; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2829:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2830; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2830:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_338;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_337;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2831; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2831:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2832; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2832:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object*) lenv) [4];
            r1[2] = r2;
            r2 = ((object****) lenv) [1][1][1][2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2833; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
            L2833:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2834; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2834:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_336;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_335;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2835; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2835:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2836; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2836:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = ((object*) lenv) [5];
            lenv = lenv[1];
          } break;
          case (UOTAG(21)): {
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r1 = (object *) 1;
            r2 = ((object*) lenv) [3];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)>unbox(r1))) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 2);
              r3 = allocate (TC_TUPLE, 8);
              r3[1] = lenv; lenv = r3;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object**) lenv) [1][3];
              r5[2] = r6;
              r6 = (object*) &constructed_347;
              r5[3] = r6;
              r6 = top[31];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2837; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
              L2837:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[47];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2838; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
              L2838:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = allocate (TC_TUPLE, 2);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object**) lenv) [1][3];
              r5[2] = r6;
              r6 = (object*) &constructed_346;
              r5[3] = r6;
              r6 = top[31];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2839; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_n_of_325;
              L2839:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = top[47];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2840; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_string_concat_352;
              L2840:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[3] = r4;
              r4 = (object*)TC_NIL;
              r3[4] = r4;
              r4 = ((object*) lenv) [4];
              r5 = (object*) &constructed_345;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[5] = r4;
              r4 = ((object*) lenv) [5];
              r5 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[6] = r4;
              r4 = ((object*) lenv) [6];
              r5 = (object*) &constructed_344;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[7] = r4;
              r4 = ((object*) lenv) [7];
              r5 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r5;
              t[2] = r4;
              r4 = t;
              r3[8] = r4;
              r3 = ((object*) lenv) [8];
              r4 = (object*) &constructed_343;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              lenv = lenv[1];
              r2[2] = r3;
              r3 = top[47];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2841; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
              L2841:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
              r2 = ((pxll_vector*)r2)->val[4];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2842; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto *r2[1];
              L2842:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = (object *) 14;
            }
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2843; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2843:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_342;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_341;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2844; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2844:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2845; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2845:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = (object*)UITAG(1);
            lenv = lenv[1];
          } break;
          case (UOTAG(22)): {
            r1 = allocate (TC_TUPLE, 7);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,4);
            r1[6] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,5);
            r1[7] = r2;
            r1 = allocate (TC_TUPLE, 3);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][6];
            switch (get_case (r2)) {
              case (UOTAG(0)): {
                r2 = (object *) 262;
              } break;
              default: {
                r2 = (object *) 6;
              } break;
            }
            r1[2] = r2;
            r2 = allocate (TC_TUPLE, 3);
            r3 = ((object********) lenv) [1][1][1][1][1][1][1][2];
            r3 = ((pxll_vector*)r3)->val[2];
            r2[2] = r3;
            r3 = ((object**) lenv) [1][3];
            r2[3] = r3;
            r3 = top[65];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2846; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flookup_394;
            L2846:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[3] = r2;
            r1 = ((object*) lenv) [3];
            switch (get_case (r1)) {
              case (UITAG(1)): {
                r1 = allocate (TC_TUPLE, 3);
                r2 = (object*) &constructed_348;
                r1[2] = r2;
                r2 = ((object**) lenv) [1][3];
                r1[3] = r2;
                r2 = top[15];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2847; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                L2847:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
              } break;
              default: {
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = ((object**) lenv) [1][3];
                r2 = UOBJ_GET(r2,0);
                r1[2] = r2;
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = allocate (TC_TUPLE, 2);
                r3 = ((object****) lenv) [1][1][1][5];
                r2[2] = r3;
                r3 = top[22];
                t = allocate (TC_SAVE, 5);
                t[1] = k; t[2] = lenv; t[3] = &&L2848; t[4] = r0; t[5] = r1; k = t;
                r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
                L2848:
                r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                r2 = result;
                r1[2] = r2;
                r1 = (object *) 3;
                r2 = ((object*) lenv) [2];
                r1 = PXLL_TEST(unbox(r2)==unbox(r1));
                lenv = lenv[1];
                if PXLL_IS_TRUE(r1) {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object**) lenv) [1][2];
                  r2 = ((pxll_vector*)r2)->val[3];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2849; t[4] = r0; t[5] = r1; k = t;
                  lenv = r2[2]; goto *r2[1];
                  L2849:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r1 = (object *) 3;
                  r2 = ((object*) lenv) [2];
                  r1 = PXLL_TEST(unbox(r2)==unbox(r1));
                  lenv = lenv[1];
                } else {
                  r1 = (object *) 6;
                }
                if PXLL_IS_TRUE(r1) {
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = allocate (TC_TUPLE, 3);
                  r3 = ((object***) lenv) [1][1][5];
                  r2[2] = r3;
                  r3 = (object *) 1;
                  r2[3] = r3;
                  r3 = top[27];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2850; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_nth_321;
                  L2850:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object******) lenv) [1][1][1][1][1][2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2851; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
                  L2851:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                } else {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = (object*) &constructed_359;
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = allocate (TC_TUPLE, 7);
                  r3[1] = lenv; lenv = r3;
                  r4 = allocate (TC_TUPLE, 2);
                  r5 = ((object*****) lenv) [1][1][1][1][2];
                  r4[2] = r5;
                  r5 = top[56];
                  t = allocate (TC_SAVE, 7);
                  t[1] = k; t[2] = lenv; t[3] = &&L2852; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
                  r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
                  L2852:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
                  r4 = result;
                  r3[2] = r4;
                  r4 = (object*)TC_NIL;
                  r3[3] = r4;
                  r4 = ((object*) lenv) [3];
                  r5 = (object*) &constructed_358;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r5;
                  t[2] = r4;
                  r4 = t;
                  r3[4] = r4;
                  r4 = ((object*) lenv) [4];
                  r5 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r5;
                  t[2] = r4;
                  r4 = t;
                  r3[5] = r4;
                  r4 = ((object*) lenv) [5];
                  r5 = (object*) &constructed_357;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r5;
                  t[2] = r4;
                  r4 = t;
                  r3[6] = r4;
                  r4 = ((object*) lenv) [6];
                  r5 = ((object**) lenv) [1][2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r5;
                  t[2] = r4;
                  r4 = t;
                  r3[7] = r4;
                  r3 = ((object*) lenv) [7];
                  r4 = (object*) &constructed_356;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r4;
                  t[2] = r3;
                  r3 = t;
                  lenv = lenv[1];
                  r2[2] = r3;
                  r3 = top[47];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2853; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
                  L2853:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
                  r2 = ((pxll_vector*)r2)->val[4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2854; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                  L2854:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = allocate (TC_TUPLE, 2);
                  r3 = ((object*****) lenv) [1][1][1][1][4];
                  r2[2] = r3;
                  r3 = top[22];
                  t = allocate (TC_SAVE, 5);
                  t[1] = k; t[2] = lenv; t[3] = &&L2855; t[4] = r0; t[5] = r1; k = t;
                  r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
                  L2855:
                  r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
                  r2 = result;
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  // def $loop_4577_i3735_i4683_i4727
                  goto L2856;
                  FUN__24loop_4577_i3735_i4683_i4727:
                    check_heap (0);
                    r0 = ((object***) lenv) [1][1][2];
                    r1 = ((object*) lenv) [2];
                    if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                      r0 = (object *) 14;
                      PXLL_RETURN(0);
                    } else {
                      r0 = allocate (TC_TUPLE, 7);
                      r0[1] = lenv; lenv = r0;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][4];
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r1[3] = r2;
                      r2 = top[27];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2857; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                      L2857:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[2] = r1;
                      r1 = allocate (TC_TUPLE, 3);
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][5];
                      r1[2] = r2;
                      r2 = ((object**) lenv) [1][2];
                      r1[3] = r2;
                      r2 = top[27];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2858; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                      L2858:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[3] = r1;
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object*) lenv) [2];
                      r1[2] = r2;
                      r2 = ((object******) lenv) [1][1][1][1][1][2];
                      r2 = ((pxll_vector*)r2)->val[1];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2859; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L2859:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      r0[4] = r1;
                      r1 = ((object*) lenv) [4];
                      r1 = ((pxll_vector*)r1)->val[0];
                      r0[5] = r1;
                      r1 = (object *) 6;
                      r0[6] = r1;
                      r1 = (object *) 1;
                      r2 = ((object*) lenv) [5];
                      if PXLL_IS_TRUE(PXLL_TEST(unbox(r2)==unbox(r1))) {
                        r1 = allocate (TC_TUPLE, 4);
                        r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                        r1[2] = r2;
                        r2 = ((object*) lenv) [2];
                        r1[3] = r2;
                        r2 = ((object*) lenv) [4];
                        r2 = ((pxll_vector*)r2)->val[1];
                        r1[4] = r2;
                        r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][7];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2860; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_get_uitag_4249_i4683_i4727;
                        L2860:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                      } else {
                        r1 = allocate (TC_TUPLE, 4);
                        r2 = ((object********) lenv) [1][1][1][1][1][1][1][3];
                        r1[2] = r2;
                        r2 = ((object*) lenv) [2];
                        r1[3] = r2;
                        r2 = ((object*) lenv) [4];
                        r2 = ((pxll_vector*)r2)->val[1];
                        r1[4] = r2;
                        r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][6];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2861; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_get_uotag_4248_i4683_i4727;
                        L2861:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                      }
                      r0[7] = r1;
                      r0 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                      r0 = ((pxll_vector*)r0)->val[3];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2862; ; k = t;
                      lenv = r0[2]; goto *r0[1];
                      L2862:
                      ; lenv = k[2]; k = k[1];
                      r0 = (object *) 6;
                      r1 = ((object*******) lenv) [1][1][1][1][1][1][2];
                      if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
                        r0 = allocate (TC_TUPLE, 3);
                        r0[1] = lenv; lenv = r0;
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = ((object*********) lenv) [1][1][1][1][1][1][1][1][4];
                        r1[2] = r2;
                        r2 = top[22];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2863; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
                        L2863:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r1 = (object *) 3;
                        r2 = ((object*) lenv) [2];
                        r1 = box((pxll_int)unbox(r2)-unbox(r1));
                        r0[3] = r1;
                        r0 = ((object*) lenv) [3];
                        r1 = ((object***) lenv) [1][1][2];
                        r0 = PXLL_TEST(unbox(r1)==unbox(r0));
                        lenv = lenv[1];
                      } else {
                        r0 = (object *) 6;
                      }
                      if PXLL_IS_TRUE(r0) {
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = (object*) &constructed_353;
                        r0[2] = r1;
                        r1 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                        r1 = ((pxll_vector*)r1)->val[4];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L2864; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto *r1[1];
                        L2864:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                      } else {
                        r0 = allocate (TC_TUPLE, 2);
                        r1 = allocate (TC_TUPLE, 2);
                        r2 = allocate (TC_TUPLE, 4);
                        r2[1] = lenv; lenv = r2;
                        r3 = (object*)TC_NIL;
                        r2[2] = r3;
                        r3 = ((object*) lenv) [2];
                        r4 = (object*) &constructed_355;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r4;
                        t[2] = r3;
                        r3 = t;
                        r2[3] = r3;
                        r3 = ((object*) lenv) [3];
                        r4 = ((object**) lenv) [1][7];
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r4;
                        t[2] = r3;
                        r3 = t;
                        r2[4] = r3;
                        r2 = ((object*) lenv) [4];
                        r3 = (object*) &constructed_354;
                        t = alloc_no_clear (TC_PAIR,2);
                        t[1] = r3;
                        t[2] = r2;
                        r2 = t;
                        lenv = lenv[1];
                        r1[2] = r2;
                        r2 = top[47];
                        t = allocate (TC_SAVE, 4);
                        t[1] = k; t[2] = lenv; t[3] = &&L2865; t[4] = r0; k = t;
                        r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                        L2865:
                        r0 = k[4]; lenv = k[2]; k = k[1];
                        r1 = result;
                        r0[2] = r1;
                        r1 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                        r1 = ((pxll_vector*)r1)->val[4];
                        t = allocate (TC_SAVE, 3);
                        t[1] = k; t[2] = lenv; t[3] = &&L2866; ; k = t;
                        r0[1] = r1[2]; lenv = r0; goto *r1[1];
                        L2866:
                        ; lenv = k[2]; k = k[1];
                        r0 = result;
                      }
                      r0 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                      r0 = ((pxll_vector*)r0)->val[3];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2867; ; k = t;
                      lenv = r0[2]; goto *r0[1];
                      L2867:
                      ; lenv = k[2]; k = k[1];
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = ((object*) lenv) [3];
                      r0[2] = r1;
                      r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][2];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2868; ; k = t;
                      r0[1] = r1[2]; lenv = r0; goto FUN_emit_4226_i4683_i4727;
                      L2868:
                      ; lenv = k[2]; k = k[1];
                      r0 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                      r0 = ((pxll_vector*)r0)->val[2];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2869; ; k = t;
                      lenv = r0[2]; goto *r0[1];
                      L2869:
                      ; lenv = k[2]; k = k[1];
                      r0 = allocate (TC_TUPLE, 2);
                      r1 = (object*) &constructed_352;
                      r0[2] = r1;
                      r1 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[4];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2870; ; k = t;
                      r0[1] = r1[2]; lenv = r0; goto *r1[1];
                      L2870:
                      ; lenv = k[2]; k = k[1];
                      r0 = ((object**************) lenv) [1][1][1][1][1][1][1][1][1][1][1][1][1][54];
                      r0 = ((pxll_vector*)r0)->val[2];
                      t = allocate (TC_SAVE, 3);
                      t[1] = k; t[2] = lenv; t[3] = &&L2871; ; k = t;
                      lenv = r0[2]; goto *r0[1];
                      L2871:
                      ; lenv = k[2]; k = k[1];
                      r0 = result;
                      lenv = lenv[1];
                      r0 = (object *) 3;
                      r1 = ((object*) lenv) [2];
                      r0 = box((pxll_int)unbox(r1)+unbox(r0));
                      lenv[2] = r0;
                      goto FUN__24loop_4577_i3735_i4683_i4727;
                    }
                    PXLL_RETURN(0);
                  L2856:
                  r2 = allocate (TC_CLOSURE, 2);
                  r2[1] = &&FUN__24loop_4577_i3735_i4683_i4727; r2[2] = lenv;
                  r1[2] = r2;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = (object *) 1;
                  r1[2] = r2;
                  r2 = ((object*) lenv) [2];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2872; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4577_i3735_i4683_i4727;
                  L2872:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  lenv = lenv[1];
                  lenv = lenv[1];
                  r1 = ((object****) lenv) [1][1][1][6];
                  switch (get_case (r1)) {
                    case (UOTAG(0)): {
                      r1 = allocate (TC_TUPLE, 2);
                      r1[1] = lenv; lenv = r1;
                      r2 = ((object*****) lenv) [1][1][1][1][6];
                      r2 = UOBJ_GET(r2,0);
                      r1[2] = r2;
                      r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[3];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2873; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2873:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = (object*) &constructed_351;
                      r1[2] = r2;
                      r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                      r2 = ((pxll_vector*)r2)->val[4];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2874; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L2874:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[3];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2875; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2875:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = ((object*) lenv) [2];
                      r1[2] = r2;
                      r2 = ((object********) lenv) [1][1][1][1][1][1][1][2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2876; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto FUN_emit_4226_i4683_i4727;
                      L2876:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2877; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2877:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = allocate (TC_TUPLE, 2);
                      r2 = (object*) &constructed_350;
                      r1[2] = r2;
                      r2 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                      r2 = ((pxll_vector*)r2)->val[4];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2878; t[4] = r0; k = t;
                      r1[1] = r2[2]; lenv = r1; goto *r2[1];
                      L2878:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = ((object***********) lenv) [1][1][1][1][1][1][1][1][1][1][54];
                      r1 = ((pxll_vector*)r1)->val[2];
                      t = allocate (TC_SAVE, 4);
                      t[1] = k; t[2] = lenv; t[3] = &&L2879; t[4] = r0; k = t;
                      lenv = r1[2]; goto *r1[1];
                      L2879:
                      r0 = k[4]; lenv = k[2]; k = k[1];
                      r1 = result;
                      lenv = lenv[1];
                    } break;
                    default: {
                      goto fatbar_34;
                    }
                  }
                  goto fatbar_34_over;
                  fatbar_34:
                  r1 = (object *) 14;
                  fatbar_34_over: ;
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = (object*) &constructed_349;
                  r1[2] = r2;
                  r2 = ((object**********) lenv) [1][1][1][1][1][1][1][1][1][54];
                  r2 = ((pxll_vector*)r2)->val[4];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2880; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto *r2[1];
                  L2880:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                  lenv = lenv[1];
                }
                lenv = lenv[1];
              } break;
            }
            lenv = lenv[1];
            r1 = ((object*) lenv) [7];
            lenv = lenv[1];
          } break;
          default: {
            r1 = allocate (TC_TUPLE, 7);
            r1[1] = lenv; lenv = r1;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,0);
            r1[2] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,1);
            r1[3] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,2);
            r1[4] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,3);
            r1[5] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,4);
            r1[6] = r2;
            r2 = ((object***) lenv) [1][1][2];
            r2 = UOBJ_GET(r2,5);
            r1[7] = r2;
            r1 = ((object*) lenv) [6];
            switch (get_case (r1)) {
              case (UOTAG(0)): {
                r1 = (object*) &constructed_68;
                r1 = box((pxll_int)((pxll_string*)(r1))->len);
                r2 = (object*) &constructed_68;
                (fwrite (((pxll_string*)(r2))->data, 1, unbox(r1), stdout), PXLL_UNDEFINED);
                r1 = (object*) &constructed_371;
                dump_object (r1, 0); fprintf (stdout, "\n");
                r1 = (object*) &constructed_371;
                PXLL_UNDEFINED; result=r1; goto Lreturn;
              } break;
              default: {
                r1 = (object *) 14;
              } break;
            }
            r1 = allocate (TC_TUPLE, 2);
            r2 = allocate (TC_TUPLE, 2);
            r3 = allocate (TC_TUPLE, 5);
            r3[1] = lenv; lenv = r3;
            r4 = allocate (TC_TUPLE, 2);
            r5 = ((object**) lenv) [1][2];
            r4[2] = r5;
            r5 = top[56];
            t = allocate (TC_SAVE, 7);
            t[1] = k; t[2] = lenv; t[3] = &&L2881; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
            r4[1] = r5[2]; lenv = r4; goto FUN_int__3estring_364;
            L2881:
            r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
            r4 = result;
            r3[2] = r4;
            r4 = (object*)TC_NIL;
            r3[3] = r4;
            r4 = ((object*) lenv) [3];
            r5 = (object*) &constructed_370;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[4] = r4;
            r4 = ((object*) lenv) [4];
            r5 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r5;
            t[2] = r4;
            r4 = t;
            r3[5] = r4;
            r3 = ((object*) lenv) [5];
            r4 = (object*) &constructed_369;
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r4;
            t[2] = r3;
            r3 = t;
            lenv = lenv[1];
            r2[2] = r3;
            r3 = top[47];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2882; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_string_concat_352;
            L2882:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r2 = ((object*******) lenv) [1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2883; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2883:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = allocate (TC_TUPLE, 2);
            r3 = ((object**) lenv) [1][5];
            r2[2] = r3;
            r3 = top[22];
            t = allocate (TC_SAVE, 5);
            t[1] = k; t[2] = lenv; t[3] = &&L2884; t[4] = r0; t[5] = r1; k = t;
            r2[1] = r3[2]; lenv = r2; goto FUN_length_314;
            L2884:
            r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
            r2 = result;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r1[1] = lenv; lenv = r1;
            // def $loop_4593_i3756_i4683_i4727
            goto L2885;
            FUN__24loop_4593_i3756_i4683_i4727:
              check_heap (0);
              r0 = ((object***) lenv) [1][1][2];
              r1 = ((object*) lenv) [2];
              if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              } else {
                r0 = allocate (TC_TUPLE, 9);
                r0[1] = lenv; lenv = r0;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object******) lenv) [1][1][1][1][1][3];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2886; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L2886:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[2] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object******) lenv) [1][1][1][1][1][4];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2887; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L2887:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[3] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object******) lenv) [1][1][1][1][1][5];
                r1[2] = r2;
                r2 = ((object**) lenv) [1][2];
                r1[3] = r2;
                r2 = top[27];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2888; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
                L2888:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[4] = r1;
                r1 = allocate (TC_TUPLE, 3);
                r2 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][2];
                r2 = ((pxll_vector*)r2)->val[15];
                r1[2] = r2;
                r2 = ((object*) lenv) [2];
                r1[3] = r2;
                r2 = top[65];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2889; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_alist_2flookup_394;
                L2889:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[5] = r1;
                r1 = ((object*) lenv) [5];
                switch (get_case (r1)) {
                  case (UOTAG(0)): {
                    r1 = allocate (TC_TUPLE, 2);
                    r1[1] = lenv; lenv = r1;
                    r2 = ((object**) lenv) [1][5];
                    r2 = UOBJ_GET(r2,0);
                    r1[2] = r2;
                    r1 = ((object*) lenv) [2];
                    lenv = lenv[1];
                  } break;
                  default: {
                    r1 = allocate (TC_TUPLE, 3);
                    r2 = (object*) &constructed_368;
                    r1[2] = r2;
                    r2 = ((object*) lenv) [2];
                    r1[3] = r2;
                    r2 = top[15];
                    t = allocate (TC_SAVE, 4);
                    t[1] = k; t[2] = lenv; t[3] = &&L2890; t[4] = r0; k = t;
                    r1[1] = r2[2]; lenv = r1; goto FUN_error1_293;
                    L2890:
                    r0 = k[4]; lenv = k[2]; k = k[1];
                    r1 = result;
                  } break;
                }
                r0[6] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 6);
                r2[1] = lenv; lenv = r2;
                r3 = (object *) 1;
                r4 = ((object**) lenv) [1][3];
                if PXLL_IS_TRUE(PXLL_TEST(unbox(r4)==unbox(r3))) {
                  r3 = (object*) &constructed_366;
                } else {
                  r3 = (object*) &constructed_367;
                }
                r2[2] = r3;
                r3 = allocate (TC_TUPLE, 2);
                r4 = ((object**) lenv) [1][6];
                r3[2] = r4;
                r4 = top[56];
                t = allocate (TC_SAVE, 6);
                t[1] = k; t[2] = lenv; t[3] = &&L2891; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
                L2891:
                r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                r3 = result;
                r2[3] = r3;
                r3 = (object*)TC_NIL;
                r2[4] = r3;
                r3 = ((object*) lenv) [4];
                r4 = (object*) &constructed_365;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[5] = r3;
                r3 = ((object*) lenv) [5];
                r4 = ((object*) lenv) [3];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[6] = r3;
                r2 = ((object*) lenv) [6];
                r3 = ((object*) lenv) [2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = top[47];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2892; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                L2892:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[7] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r2 = allocate (TC_TUPLE, 4);
                r2[1] = lenv; lenv = r2;
                r3 = (object*)TC_NIL;
                r2[2] = r3;
                r3 = ((object*) lenv) [2];
                r4 = (object*) &constructed_364;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[3] = r3;
                r3 = ((object*) lenv) [3];
                r4 = ((object**) lenv) [1][7];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r4;
                t[2] = r3;
                r3 = t;
                r2[4] = r3;
                r2 = ((object*) lenv) [4];
                r3 = (object*) &constructed_363;
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r3;
                t[2] = r2;
                r2 = t;
                lenv = lenv[1];
                r1[2] = r2;
                r2 = top[47];
                t = allocate (TC_SAVE, 4);
                t[1] = k; t[2] = lenv; t[3] = &&L2893; t[4] = r0; k = t;
                r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                L2893:
                r0 = k[4]; lenv = k[2]; k = k[1];
                r1 = result;
                r0[8] = r1;
                r1 = allocate (TC_TUPLE, 2);
                r1[1] = lenv; lenv = r1;
                r2 = (object *) 3;
                r3 = ((object******) lenv) [1][1][1][1][1][2];
                r2 = box((pxll_int)unbox(r3)-unbox(r2));
                r1[2] = r2;
                r1 = ((object*) lenv) [2];
                r2 = ((object***) lenv) [1][1][2];
                r1 = PXLL_TEST(unbox(r2)==unbox(r1));
                lenv = lenv[1];
                if PXLL_IS_TRUE(r1) {
                  r1 = (object*) &constructed_362;
                } else {
                  r1 = ((object*) lenv) [8];
                }
                r0[9] = r1;
                r0 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                r0 = ((pxll_vector*)r0)->val[3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2894; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L2894:
                ; lenv = k[2]; k = k[1];
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object*) lenv) [9];
                r0[2] = r1;
                r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                r1 = ((pxll_vector*)r1)->val[4];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2895; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L2895:
                ; lenv = k[2]; k = k[1];
                r0 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                r0 = ((pxll_vector*)r0)->val[3];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2896; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L2896:
                ; lenv = k[2]; k = k[1];
                r0 = allocate (TC_TUPLE, 2);
                r1 = ((object*) lenv) [4];
                r0[2] = r1;
                r1 = ((object*********) lenv) [1][1][1][1][1][1][1][1][2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2897; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto FUN_emit_4226_i4683_i4727;
                L2897:
                ; lenv = k[2]; k = k[1];
                r0 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                r0 = ((pxll_vector*)r0)->val[2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2898; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L2898:
                ; lenv = k[2]; k = k[1];
                r0 = allocate (TC_TUPLE, 2);
                r1 = (object*) &constructed_361;
                r0[2] = r1;
                r1 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                r1 = ((pxll_vector*)r1)->val[4];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2899; ; k = t;
                r0[1] = r1[2]; lenv = r0; goto *r1[1];
                L2899:
                ; lenv = k[2]; k = k[1];
                r0 = ((object************) lenv) [1][1][1][1][1][1][1][1][1][1][1][54];
                r0 = ((pxll_vector*)r0)->val[2];
                t = allocate (TC_SAVE, 3);
                t[1] = k; t[2] = lenv; t[3] = &&L2900; ; k = t;
                lenv = r0[2]; goto *r0[1];
                L2900:
                ; lenv = k[2]; k = k[1];
                r0 = result;
                lenv = lenv[1];
                r0 = (object *) 3;
                r1 = ((object*) lenv) [2];
                r0 = box((pxll_int)unbox(r1)+unbox(r0));
                lenv[2] = r0;
                goto FUN__24loop_4593_i3756_i4683_i4727;
              }
              PXLL_RETURN(0);
            L2885:
            r2 = allocate (TC_CLOSURE, 2);
            r2[1] = &&FUN__24loop_4593_i3756_i4683_i4727; r2[2] = lenv;
            r1[2] = r2;
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object *) 1;
            r1[2] = r2;
            r2 = ((object*) lenv) [2];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2901; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto FUN__24loop_4593_i3756_i4683_i4727;
            L2901:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            lenv = lenv[1];
            r1 = allocate (TC_TUPLE, 2);
            r2 = (object*) &constructed_360;
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][54];
            r2 = ((pxll_vector*)r2)->val[4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2902; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2902:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            lenv = lenv[1];
            r1 = ((object*) lenv) [7];
            lenv = lenv[1];
          } break;
        }
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        switch (get_case (r0)) {
          case (UOTAG(0)): {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r1 = UOBJ_GET(r1,2);
            r0[2] = r1;
            r0 = ((object*) lenv) [2];
            lenv = ((object **)lenv)[1][1];
            lenv[2] = r0;
            goto FUN_emit_4226_i4683_i4727;
          } break;
          default: {
            r0 = (object *) 14;
            PXLL_RETURN(0);
          } break;
        }
        PXLL_RETURN(0);
      L2455:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_emit_4226_i4683_i4727; r1[2] = lenv;
      r0[2] = r1;
      // def move_4227_i4683_i4727
      goto L2903;
      FUN_move_4227_i4683_i4727:
        check_heap (0);
        r0 = (object *) 1;
        r1 = ((object*) lenv) [3];
        if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)>=unbox(r0))) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][3];
          r2 = ((object**) lenv) [1][2];
          r1 = PXLL_TEST(unbox(r2)==unbox(r1));
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          lenv = lenv[1];
        } else {
          r0 = (object *) 6;
        }
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 8);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][3];
          r3[2] = r4;
          r4 = top[56];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L2904; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
          L2904:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][2];
          r3[2] = r4;
          r4 = top[56];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L2905; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
          L2905:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[3] = r3;
          r3 = (object*)TC_NIL;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = (object*) &constructed_72;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r3 = ((object*) lenv) [5];
          r4 = ((object*) lenv) [3];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[6] = r3;
          r3 = ((object*) lenv) [6];
          r4 = (object*) &constructed_71;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[7] = r3;
          r3 = ((object*) lenv) [7];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[8] = r3;
          r2 = ((object*) lenv) [8];
          r3 = (object*) &constructed_70;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2906; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L2906:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object*****) lenv) [1][1][1][1][54];
          r1 = ((pxll_vector*)r1)->val[4];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } else {
          r0 = (object *) 14;
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L2903:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_move_4227_i4683_i4727; r1[2] = lenv;
      r0[3] = r1;
      // def gen-function-label_4234_i4683_i4727
      goto L2907;
      FUN_gen_function_label_4234_i4683_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 4);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][2];
        r3 = allocate (TC_TUPLE, 2);
        r3[1] = lenv; lenv = r3;
        r4 = ((object***) lenv) [1][1][2];
        r4 = UOBJ_GET(r4,0);
        r3[2] = r4;
        r3 = ((object*) lenv) [2];
        lenv = lenv[1];
        r2[2] = r3;
        r3 = ((object***) lenv) [1][1][8];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2908; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto *r3[1];
        L2908:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r1 = ((object*) lenv) [4];
        r2 = (object*) &constructed_69;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
      L2907:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_gen_function_label_4234_i4683_i4727; r1[2] = lenv;
      r0[4] = r1;
      // def guess-record-type_4247_i4683_i4727
      goto L2909;
      FUN_guess_record_type_4247_i4683_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 4);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_30524_i4683_i4727
        goto L2910;
        FUN_lambda_30524_i4683_i4727:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          switch (get_case (r0)) {
            case (UOTAG(1)): {
              r0 = allocate (TC_TUPLE, 2);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r0 = ((object*) lenv) [2];
              PXLL_RETURN(0);
            } break;
            default: {
              goto fatbar_5;
            }
          }
          PXLL_RETURN(0);
          goto fatbar_5_over;
          fatbar_5:
          r0 = (object*) &constructed_68;
          r0 = box((pxll_int)((pxll_string*)(r0))->len);
          r1 = (object*) &constructed_68;
          (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
          r0 = (object*) &constructed_67;
          dump_object (r0, 0); fprintf (stdout, "\n");
          r0 = (object*) &constructed_67;
          PXLL_UNDEFINED; result=r0; goto Lreturn;
          PXLL_RETURN(0);
          fatbar_5_over: ;
          PXLL_RETURN(0);
        L2910:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_30524_i4683_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object**) lenv) [1][2];
        r1[3] = r2;
        r2 = top[32];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2911; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_map_326;
        L2911:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r1 = allocate (TC_TUPLE, 3);
        // def lambda_30535_i4683_i4727
        goto L2912;
        FUN_lambda_30535_i4683_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = (object *) constructed_66[0];
          r2 = ((object**) lenv) [1][2];
          r1 = PXLL_TEST(r2==r1);
          r0[2] = r1;
          r0 = (object *) 6;
          r1 = ((object*) lenv) [2];
          r0 = PXLL_TEST(r1==r0);
          PXLL_RETURN(0);
        L2912:
        r2 = allocate (TC_CLOSURE, 2);
        r2[1] = &&FUN_lambda_30535_i4683_i4727; r2[2] = lenv;
        r1[2] = r2;
        r2 = ((object*) lenv) [2];
        r1[3] = r2;
        r2 = top[34];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2913; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_filter_328;
        L2913:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[3] = r1;
        r1 = (object *) constructed_64[0];
        r0[4] = r1;
        r0 = allocate (TC_TUPLE, 3);
        // def lambda_30563_i4683_i4727
        goto L2914;
        FUN_lambda_30563_i4683_i4727:
          check_heap (0);
          r0 = ((object*) lenv) [2];
          switch (get_case_noint (r0)) {
            default: {
              r0 = allocate (TC_TUPLE, 3);
              r0[1] = lenv; lenv = r0;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,0);
              r0[2] = r1;
              r1 = ((object**) lenv) [1][2];
              r1 = UOBJ_GET(r1,1);
              r0[3] = r1;
              r0 = allocate (TC_TUPLE, 3);
              // def lambda_30505_i4073_i4683_i4727
              goto L2915;
              FUN_lambda_30505_i4073_i4683_i4727:
                check_heap (0);
                r0 = allocate (TC_TUPLE, 3);
                r1 = ((object*) lenv) [2];
                r0[2] = r1;
                r1 = ((object**) lenv) [1][2];
                r0[3] = r1;
                r1 = top[25];
                r0[1] = r1[2]; lenv = r0; goto FUN_member_eq_question_319;
              L2915:
              r1 = allocate (TC_CLOSURE, 2);
              r1[1] = &&FUN_lambda_30505_i4073_i4683_i4727; r1[2] = lenv;
              r0[2] = r1;
              r1 = ((object***) lenv) [1][1][3];
              r0[3] = r1;
              r1 = top[40];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L2916; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto FUN_every_question_334;
              L2916:
              ; lenv = k[2]; k = k[1];
              r0 = result;
              if PXLL_IS_TRUE(r0) {
                r0 = allocate (TC_TUPLE, 2);
                r0[1] = lenv; lenv = r0;
                r1 = ((object****) lenv) [1][1][1][4];
                r0[2] = r1;
                r0 = ((object*) lenv) [2];
                r1 = ((object**) lenv) [1][2];
                t = alloc_no_clear (TC_PAIR,2);
                t[1] = r1;
                t[2] = r0;
                r0 = t;
                lenv = lenv[1];
                ((object***) lenv) [1][1][4] = r0;
                PXLL_RETURN(0);
              } else {
                r0 = (object *) 14;
                PXLL_RETURN(0);
              }
              PXLL_RETURN(0);
            } break;
          }
          PXLL_RETURN(0);
        L2914:
        r1 = allocate (TC_CLOSURE, 2);
        r1[1] = &&FUN_lambda_30563_i4683_i4727; r1[2] = lenv;
        r0[2] = r1;
        r1 = ((object******) lenv) [1][1][1][1][1][2];
        r1 = ((pxll_vector*)r1)->val[10];
        r0[3] = r1;
        r1 = top[35];
        t = allocate (TC_SAVE, 3);
        t[1] = k; t[2] = lenv; t[3] = &&L2917; ; k = t;
        r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
        L2917:
        ; lenv = k[2]; k = k[1];
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = ((object**) lenv) [1][4];
        r1[2] = r2;
        r2 = top[22];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2918; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
        L2918:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = ((object*) lenv) [2];
        r1 = (object *) 3;
        r0 = PXLL_TEST(unbox(r1)==unbox(r0));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 3);
          r2 = ((object**) lenv) [1][4];
          r1[2] = r2;
          r2 = (object *) 1;
          r1[3] = r2;
          r2 = top[27];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2919; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_nth_321;
          L2919:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          t = alloc_no_clear (UOTAG(0),1);
          t[1] = r0;
          r0 = t;
          PXLL_RETURN(0);
        } else {
          r0 = (object*)UITAG(1);
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L2909:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_guess_record_type_4247_i4683_i4727; r1[2] = lenv;
      r0[5] = r1;
      // def get-uotag_4248_i4683_i4727
      goto L2920;
      FUN_get_uotag_4248_i4683_i4727:
        check_heap (0);
        r0 = (object *) constructed_58[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_56[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_54;
            PXLL_RETURN(0);
          } else {
            goto fatbar_3;
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_3;
        }
        PXLL_RETURN(0);
        goto fatbar_3_over;
        fatbar_3:
        r0 = (object *) constructed_51[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_61[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_59;
            PXLL_RETURN(0);
          } else {
            goto fatbar_4;
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_4;
        }
        PXLL_RETURN(0);
        goto fatbar_4_over;
        fatbar_4:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][4];
        r2[2] = r3;
        r3 = top[56];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2921; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
        L2921:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_63;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_62;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        fatbar_4_over: ;
        PXLL_RETURN(0);
        fatbar_3_over: ;
        PXLL_RETURN(0);
      L2920:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_get_uotag_4248_i4683_i4727; r1[2] = lenv;
      r0[6] = r1;
      // def get-uitag_4249_i4683_i4727
      goto L2922;
      FUN_get_uitag_4249_i4683_i4727:
        check_heap (0);
        r0 = (object *) constructed_46[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_41[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_39;
            PXLL_RETURN(0);
          } else {
            goto fatbar_1;
          }
          PXLL_RETURN(0);
          goto fatbar_1_over;
          fatbar_1:
          r0 = (object *) constructed_44[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_42;
            PXLL_RETURN(0);
          } else {
            goto fatbar_0;
          }
          PXLL_RETURN(0);
          fatbar_1_over: ;
          PXLL_RETURN(0);
        } else {
          goto fatbar_0;
        }
        PXLL_RETURN(0);
        goto fatbar_0_over;
        fatbar_0:
        r0 = (object *) constructed_51[0];
        r1 = ((object*) lenv) [2];
        if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
          r0 = (object *) constructed_49[0];
          r1 = ((object*) lenv) [3];
          if PXLL_IS_TRUE(PXLL_TEST(r1==r0)) {
            r0 = (object*) &constructed_47;
            PXLL_RETURN(0);
          } else {
            goto fatbar_2;
          }
          PXLL_RETURN(0);
        } else {
          goto fatbar_2;
        }
        PXLL_RETURN(0);
        goto fatbar_2_over;
        fatbar_2:
        r0 = allocate (TC_TUPLE, 2);
        r1 = allocate (TC_TUPLE, 5);
        r1[1] = lenv; lenv = r1;
        r2 = allocate (TC_TUPLE, 2);
        r3 = ((object**) lenv) [1][4];
        r2[2] = r3;
        r3 = top[56];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2923; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_int__3estring_364;
        L2923:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        r1[2] = r2;
        r2 = (object*)TC_NIL;
        r1[3] = r2;
        r2 = ((object*) lenv) [3];
        r3 = (object*) &constructed_53;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[4] = r2;
        r2 = ((object*) lenv) [4];
        r3 = ((object*) lenv) [2];
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r3;
        t[2] = r2;
        r2 = t;
        r1[5] = r2;
        r1 = ((object*) lenv) [5];
        r2 = (object*) &constructed_52;
        t = alloc_no_clear (TC_PAIR,2);
        t[1] = r2;
        t[2] = r1;
        r1 = t;
        lenv = lenv[1];
        r0[2] = r1;
        r1 = top[47];
        r0[1] = r1[2]; lenv = r0; goto FUN_string_concat_352;
        fatbar_2_over: ;
        PXLL_RETURN(0);
        fatbar_0_over: ;
        PXLL_RETURN(0);
      L2922:
      r1 = allocate (TC_CLOSURE, 2);
      r1[1] = &&FUN_get_uitag_4249_i4683_i4727; r1[2] = lenv;
      r0[7] = r1;
      r1 = allocate (TC_TUPLE, 4);
      r1[1] = lenv; lenv = r1;
      // def frob-name_4175_i4394_i4683_i4727
      goto L2924;
      FUN_frob_name_4175_i4394_i4683_i4727:
        check_heap (0);
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 2);
        r2 = allocate (TC_TUPLE, 2);
        r2[1] = lenv; lenv = r2;
        r3 = allocate (TC_TUPLE, 2);
        r3[1] = lenv; lenv = r3;
        // def loop_4180_i3092_i4394_i4683_i4727
        goto L2925;
        FUN_loop_4180_i3092_i4394_i4683_i4727:
          check_heap (0);
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = ((object******) lenv) [1][1][1][1][1][2];
          r1 = box((pxll_int)((pxll_string*)(r1))->len);
          r0[2] = r1;
          r0 = ((object*) lenv) [2];
          r1 = ((object**) lenv) [1][2];
          r0 = PXLL_TEST(unbox(r1)==unbox(r0));
          lenv = lenv[1];
          if PXLL_IS_TRUE(r0) {
            r0 = ((object*) lenv) [3];
            PXLL_RETURN(0);
          } else {
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = ((object**) lenv) [1][2];
            r2 = ((object******) lenv) [1][1][1][1][1][2];
            range_check (((pxll_string *)(((pxll_string*)(r2))))->len, unbox(r1));
            r1 = ((object**) lenv) [1][2];
            r2 = ((object******) lenv) [1][1][1][1][1][2];
            r1 = TO_CHAR(((unsigned char *)((pxll_string*)(r2))->data)[unbox(r1)]);
            r0[2] = r1;
            r0 = allocate (TC_TUPLE, 2);
            r0[1] = lenv; lenv = r0;
            r1 = allocate (TC_TUPLE, 2);
            r2 = ((object**) lenv) [1][2];
            r1[2] = r2;
            r2 = ((object********) lenv) [1][1][1][1][1][1][1][4];
            t = allocate (TC_SAVE, 4);
            t[1] = k; t[2] = lenv; t[3] = &&L2926; t[4] = r0; k = t;
            r1[1] = r2[2]; lenv = r1; goto *r2[1];
            L2926:
            r0 = k[4]; lenv = k[2]; k = k[1];
            r1 = result;
            if PXLL_IS_TRUE(r1) {
              r1 = allocate (TC_TUPLE, 2);
              r2 = ((object**) lenv) [1][2];
              r1[2] = r2;
              r2 = top[46];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2927; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_char__3estring_348;
              L2927:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
            } else {
              r1 = allocate (TC_TUPLE, 2);
              r1[1] = lenv; lenv = r1;
              r2 = allocate (TC_TUPLE, 3);
              r3 = ((object*********) lenv) [1][1][1][1][1][1][1][1][3];
              r2[2] = r3;
              r3 = ((object***) lenv) [1][1][2];
              r2[3] = r3;
              r3 = top[65];
              t = allocate (TC_SAVE, 5);
              t[1] = k; t[2] = lenv; t[3] = &&L2928; t[4] = r0; t[5] = r1; k = t;
              r2[1] = r3[2]; lenv = r2; goto FUN_alist_2flookup_394;
              L2928:
              r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
              r2 = result;
              r1[2] = r2;
              r1 = ((object*) lenv) [2];
              switch (get_case (r1)) {
                case (UOTAG(0)): {
                  r1 = allocate (TC_TUPLE, 2);
                  r1[1] = lenv; lenv = r1;
                  r2 = ((object**) lenv) [1][2];
                  r2 = UOBJ_GET(r2,0);
                  r1[2] = r2;
                  r1 = ((object*) lenv) [2];
                  lenv = lenv[1];
                } break;
                default: {
                  r1 = allocate (TC_TUPLE, 2);
                  r2 = allocate (TC_TUPLE, 4);
                  r2[1] = lenv; lenv = r2;
                  r3 = allocate (TC_TUPLE, 2);
                  r4 = ((object****) lenv) [1][1][1][2];
                  r4 = box((pxll_int)GET_CHAR(r4));
                  r3[2] = r4;
                  r4 = top[57];
                  t = allocate (TC_SAVE, 6);
                  t[1] = k; t[2] = lenv; t[3] = &&L2929; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
                  r3[1] = r4[2]; lenv = r3; goto FUN_int__3ehex_string_365;
                  L2929:
                  r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
                  r3 = result;
                  r2[2] = r3;
                  r3 = (object*)TC_NIL;
                  r2[3] = r3;
                  r3 = ((object*) lenv) [3];
                  r4 = ((object*) lenv) [2];
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r4;
                  t[2] = r3;
                  r3 = t;
                  r2[4] = r3;
                  r2 = ((object*) lenv) [4];
                  r3 = (object*) &constructed_38;
                  t = alloc_no_clear (TC_PAIR,2);
                  t[1] = r3;
                  t[2] = r2;
                  r2 = t;
                  lenv = lenv[1];
                  r1[2] = r2;
                  r2 = top[47];
                  t = allocate (TC_SAVE, 4);
                  t[1] = k; t[2] = lenv; t[3] = &&L2930; t[4] = r0; k = t;
                  r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
                  L2930:
                  r0 = k[4]; lenv = k[2]; k = k[1];
                  r1 = result;
                } break;
              }
              lenv = lenv[1];
            }
            r0[2] = r1;
            r0 = ((object***) lenv) [1][1][3];
            r1 = ((object*) lenv) [2];
            t = alloc_no_clear (TC_PAIR,2);
            t[1] = r1;
            t[2] = r0;
            r0 = t;
            lenv = lenv[1];
            r1 = (object *) 3;
            r2 = ((object**) lenv) [1][2];
            r1 = box((pxll_int)unbox(r2)+unbox(r1));
            lenv = ((object *)lenv)[1];
            lenv[2] = r1;
            lenv[3] = r0;
            goto FUN_loop_4180_i3092_i4394_i4683_i4727;
          }
          PXLL_RETURN(0);
        L2925:
        r4 = allocate (TC_CLOSURE, 2);
        r4[1] = &&FUN_loop_4180_i3092_i4394_i4683_i4727; r4[2] = lenv;
        r3[2] = r4;
        r3 = allocate (TC_TUPLE, 3);
        r4 = (object *) 1;
        r3[2] = r4;
        r4 = (object *) constructed_37[0];
        r3[3] = r4;
        r4 = ((object*) lenv) [2];
        t = allocate (TC_SAVE, 6);
        t[1] = k; t[2] = lenv; t[3] = &&L2931; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
        r3[1] = r4[2]; lenv = r3; goto FUN_loop_4180_i3092_i4394_i4683_i4727;
        L2931:
        r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
        r3 = result;
        lenv = lenv[1];
        r2[2] = r3;
        r2 = allocate (TC_TUPLE, 3);
        r3 = ((object*) lenv) [2];
        r2[2] = r3;
        r3 = (object *) constructed_36[0];
        r2[3] = r3;
        r3 = top[21];
        t = allocate (TC_SAVE, 5);
        t[1] = k; t[2] = lenv; t[3] = &&L2932; t[4] = r0; t[5] = r1; k = t;
        r2[1] = r3[2]; lenv = r2; goto FUN_reverse_onto_311;
        L2932:
        r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
        r2 = result;
        lenv = lenv[1];
        r1[2] = r2;
        r2 = top[47];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2933; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
        L2933:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = allocate (TC_TUPLE, 2);
        r0[1] = lenv; lenv = r0;
        r1 = allocate (TC_TUPLE, 3);
        r2 = ((object**) lenv) [1][2];
        r1[2] = r2;
        r2 = (object*) &constructed_35;
        r1[3] = r2;
        r2 = top[50];
        t = allocate (TC_SAVE, 4);
        t[1] = k; t[2] = lenv; t[3] = &&L2934; t[4] = r0; k = t;
        r1[1] = r2[2]; lenv = r1; goto FUN_string_compare_355;
        L2934:
        r0 = k[4]; lenv = k[2]; k = k[1];
        r1 = result;
        r0[2] = r1;
        r0 = (object *) 1;
        r1 = ((object*) lenv) [2];
        r0 = PXLL_TEST(unbox(r1)==unbox(r0));
        lenv = lenv[1];
        if PXLL_IS_TRUE(r0) {
          r0 = (object*) &constructed_34;
          PXLL_RETURN(0);
        } else {
          r0 = ((object*) lenv) [2];
          PXLL_RETURN(0);
        }
        PXLL_RETURN(0);
      L2924:
      r2 = allocate (TC_CLOSURE, 2);
      r2[1] = &&FUN_frob_name_4175_i4394_i4683_i4727; r2[2] = lenv;
      r1[2] = r2;
      r2 = (object *) constructed_33[0];
      r1[3] = r2;
      r2 = allocate (TC_TUPLE, 2);
      r3 = allocate (TC_TUPLE, 2);
      r4 = (object*) &constructed_26;
      r3[2] = r4;
      r4 = top[55];
      t = allocate (TC_SAVE, 6);
      t[1] = k; t[2] = lenv; t[3] = &&L2935; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
      r3[1] = r4[2]; lenv = r3; goto FUN_string__3elist_362;
      L2935:
      r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
      r3 = result;
      r2[2] = r3;
      r3 = top[83];
      t = allocate (TC_SAVE, 5);
      t[1] = k; t[2] = lenv; t[3] = &&L2936; t[4] = r0; t[5] = r1; k = t;
      r2[1] = r3[2]; lenv = r2; goto FUN_char_class_423;
      L2936:
      r0 = k[4]; r1 = k[5]; lenv = k[2]; k = k[1];
      r2 = result;
      r1[4] = r2;
      r1 = ((object*) lenv) [2];
      lenv = lenv[1];
      r0[8] = r1;
      r0 = allocate (TC_TUPLE, 2);
      r1 = ((object****) lenv) [1][1][1][50];
      r0[2] = r1;
      r1 = ((object*) lenv) [2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2937; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto FUN_emit_4226_i4683_i4727;
      L2937:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_25;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2938; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2938:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_24;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2939; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2939:
      ; lenv = k[2]; k = k[1];
      r0 = ((object****) lenv) [1][1][1][54];
      r0 = ((pxll_vector*)r0)->val[2];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2940; ; k = t;
      lenv = r0[2]; goto *r0[1];
      L2940:
      ; lenv = k[2]; k = k[1];
      r0 = allocate (TC_TUPLE, 2);
      r1 = (object*) &constructed_23;
      r0[2] = r1;
      r1 = ((object****) lenv) [1][1][1][54];
      r1 = ((pxll_vector*)r1)->val[4];
      t = allocate (TC_SAVE, 3);
      t[1] = k; t[2] = lenv; t[3] = &&L2941; ; k = t;
      r0[1] = r1[2]; lenv = r0; goto *r1[1];
      L2941:
      ; lenv = k[2]; k = k[1];
      r0 = result;
      lenv = lenv[1];
      lenv = lenv[1];
    } break;
  }
  lenv = lenv[1];
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  r1 = allocate (TC_TUPLE, 2);
  r2 = ((object**) lenv) [1][2];
  r2 = ((pxll_vector*)r2)->val[10];
  r1[2] = r2;
  r2 = top[22];
  t = allocate (TC_SAVE, 4);
  t[1] = k; t[2] = lenv; t[3] = &&L2942; t[4] = r0; k = t;
  r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
  L2942:
  r0 = k[4]; lenv = k[2]; k = k[1];
  r1 = result;
  r0[2] = r1;
  r0 = (object *) 1;
  r1 = ((object*) lenv) [2];
  r0 = PXLL_TEST(unbox(r1)>unbox(r0));
  lenv = lenv[1];
  if PXLL_IS_TRUE(r0) {
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object*) &constructed_22;
    r0[2] = r1;
    r1 = ((object*) lenv) [54];
    r1 = ((pxll_vector*)r1)->val[4];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2943; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L2943:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object*) &constructed_21;
    r0[2] = r1;
    r1 = ((object*) lenv) [54];
    r1 = ((pxll_vector*)r1)->val[4];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2944; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L2944:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 3);
    // def lambda_33822_i4684_i4727
    goto L2945;
    FUN_lambda_33822_i4684_i4727:
      check_heap (0);
      r0 = ((object*) lenv) [2];
      switch (get_case_noint (r0)) {
        default: {
          r0 = allocate (TC_TUPLE, 3);
          r0[1] = lenv; lenv = r0;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,0);
          r0[2] = r1;
          r1 = ((object**) lenv) [1][2];
          r1 = UOBJ_GET(r1,1);
          r0[3] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = allocate (TC_TUPLE, 2);
          r2 = allocate (TC_TUPLE, 5);
          r2[1] = lenv; lenv = r2;
          r3 = allocate (TC_TUPLE, 2);
          r4 = ((object**) lenv) [1][3];
          r3[2] = r4;
          r4 = top[56];
          t = allocate (TC_SAVE, 6);
          t[1] = k; t[2] = lenv; t[3] = &&L2946; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
          r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
          L2946:
          r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
          r3 = result;
          r2[2] = r3;
          r3 = (object*)TC_NIL;
          r2[3] = r3;
          r3 = ((object*) lenv) [3];
          r4 = (object*) &constructed_20;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[4] = r3;
          r3 = ((object*) lenv) [4];
          r4 = ((object*) lenv) [2];
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r4;
          t[2] = r3;
          r3 = t;
          r2[5] = r3;
          r2 = ((object*) lenv) [5];
          r3 = (object*) &constructed_19;
          t = alloc_no_clear (TC_PAIR,2);
          t[1] = r3;
          t[2] = r2;
          r2 = t;
          lenv = lenv[1];
          r1[2] = r2;
          r2 = top[47];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2947; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
          L2947:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][54];
          r1 = ((pxll_vector*)r1)->val[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2948; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L2948:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object*) &constructed_18;
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][54];
          r1 = ((pxll_vector*)r1)->val[4];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2949; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
          L2949:
          ; lenv = k[2]; k = k[1];
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          r1 = allocate (TC_TUPLE, 2);
          r2 = ((object**) lenv) [1][2];
          r1[2] = r2;
          r2 = top[22];
          t = allocate (TC_SAVE, 4);
          t[1] = k; t[2] = lenv; t[3] = &&L2950; t[4] = r0; k = t;
          r1[1] = r2[2]; lenv = r1; goto FUN_length_314;
          L2950:
          r0 = k[4]; lenv = k[2]; k = k[1];
          r1 = result;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r0[1] = lenv; lenv = r0;
          // def $loop_4677_i4684_i4727
          goto L2951;
          FUN__24loop_4677_i4684_i4727:
            check_heap (0);
            r0 = ((object***) lenv) [1][1][2];
            r1 = ((object*) lenv) [2];
            if PXLL_IS_TRUE(PXLL_TEST(unbox(r1)==unbox(r0))) {
              r0 = (object *) 14;
              PXLL_RETURN(0);
            } else {
              r0 = allocate (TC_TUPLE, 2);
              r1 = allocate (TC_TUPLE, 2);
              r2 = allocate (TC_TUPLE, 8);
              r2[1] = lenv; lenv = r2;
              r3 = allocate (TC_TUPLE, 2);
              r4 = allocate (TC_TUPLE, 3);
              r5 = allocate (TC_TUPLE, 3);
              r6 = ((object*****) lenv) [1][1][1][1][2];
              r5[2] = r6;
              r6 = ((object**) lenv) [1][2];
              r5[3] = r6;
              r6 = top[27];
              t = allocate (TC_SAVE, 8);
              t[1] = k; t[2] = lenv; t[3] = &&L2952; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; t[8] = r4; k = t;
              r5[1] = r6[2]; lenv = r5; goto FUN_nth_321;
              L2952:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; r4 = k[8]; lenv = k[2]; k = k[1];
              r5 = result;
              r4[2] = r5;
              r5 = ((object*******) lenv) [1][1][1][1][1][1][2];
              r4[3] = r5;
              r5 = top[9];
              t = allocate (TC_SAVE, 7);
              t[1] = k; t[2] = lenv; t[3] = &&L2953; t[4] = r0; t[5] = r1; t[6] = r2; t[7] = r3; k = t;
              r4[1] = r5[2]; lenv = r4; goto FUN_lookup_label_code_253;
              L2953:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; r3 = k[7]; lenv = k[2]; k = k[1];
              r4 = result;
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2954; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2954:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[2] = r3;
              r3 = allocate (TC_TUPLE, 2);
              r4 = ((object**) lenv) [1][2];
              r3[2] = r4;
              r4 = top[56];
              t = allocate (TC_SAVE, 6);
              t[1] = k; t[2] = lenv; t[3] = &&L2955; t[4] = r0; t[5] = r1; t[6] = r2; k = t;
              r3[1] = r4[2]; lenv = r3; goto FUN_int__3estring_364;
              L2955:
              r0 = k[4]; r1 = k[5]; r2 = k[6]; lenv = k[2]; k = k[1];
              r3 = result;
              r2[3] = r3;
              r3 = (object*)TC_NIL;
              r2[4] = r3;
              r3 = ((object*) lenv) [4];
              r4 = (object*) &constructed_17;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[5] = r3;
              r3 = ((object*) lenv) [5];
              r4 = ((object*) lenv) [3];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[6] = r3;
              r3 = ((object*) lenv) [6];
              r4 = (object*) &constructed_16;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[7] = r3;
              r3 = ((object*) lenv) [7];
              r4 = ((object*) lenv) [2];
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r4;
              t[2] = r3;
              r3 = t;
              r2[8] = r3;
              r2 = ((object*) lenv) [8];
              r3 = (object*) &constructed_15;
              t = alloc_no_clear (TC_PAIR,2);
              t[1] = r3;
              t[2] = r2;
              r2 = t;
              lenv = lenv[1];
              r1[2] = r2;
              r2 = top[47];
              t = allocate (TC_SAVE, 4);
              t[1] = k; t[2] = lenv; t[3] = &&L2956; t[4] = r0; k = t;
              r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
              L2956:
              r0 = k[4]; lenv = k[2]; k = k[1];
              r1 = result;
              r0[2] = r1;
              r1 = ((object******) lenv) [1][1][1][1][1][54];
              r1 = ((pxll_vector*)r1)->val[4];
              t = allocate (TC_SAVE, 3);
              t[1] = k; t[2] = lenv; t[3] = &&L2957; ; k = t;
              r0[1] = r1[2]; lenv = r0; goto *r1[1];
              L2957:
              ; lenv = k[2]; k = k[1];
              r0 = (object *) 3;
              r1 = ((object*) lenv) [2];
              r0 = box((pxll_int)unbox(r1)+unbox(r0));
              lenv[2] = r0;
              goto FUN__24loop_4677_i4684_i4727;
            }
            PXLL_RETURN(0);
          L2951:
          r1 = allocate (TC_CLOSURE, 2);
          r1[1] = &&FUN__24loop_4677_i4684_i4727; r1[2] = lenv;
          r0[2] = r1;
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object *) 1;
          r0[2] = r1;
          r1 = ((object*) lenv) [2];
          t = allocate (TC_SAVE, 3);
          t[1] = k; t[2] = lenv; t[3] = &&L2958; ; k = t;
          r0[1] = r1[2]; lenv = r0; goto FUN__24loop_4677_i4684_i4727;
          L2958:
          ; lenv = k[2]; k = k[1];
          r0 = result;
          lenv = lenv[1];
          lenv = lenv[1];
          r0 = allocate (TC_TUPLE, 2);
          r1 = (object*) &constructed_14;
          r0[2] = r1;
          r1 = ((object***) lenv) [1][1][54];
          r1 = ((pxll_vector*)r1)->val[4];
          r0[1] = r1[2]; lenv = r0; goto *r1[1];
        } break;
      }
      PXLL_RETURN(0);
    L2945:
    r1 = allocate (TC_CLOSURE, 2);
    r1[1] = &&FUN_lambda_33822_i4684_i4727; r1[2] = lenv;
    r0[2] = r1;
    r1 = ((object*) lenv) [2];
    r1 = ((pxll_vector*)r1)->val[10];
    r0[3] = r1;
    r1 = top[35];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2959; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_for_each_329;
    L2959:
    ; lenv = k[2]; k = k[1];
    r0 = allocate (TC_TUPLE, 2);
    r1 = (object*) &constructed_13;
    r0[2] = r1;
    r1 = ((object*) lenv) [54];
    r1 = ((pxll_vector*)r1)->val[4];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2960; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto *r1[1];
    L2960:
    ; lenv = k[2]; k = k[1];
    r0 = result;
  } else {
    r0 = (object *) 14;
  }
  r0 = (object*) &constructed_12;
  r0 = box((pxll_int)((pxll_string*)(r0))->len);
  r1 = (object*) &constructed_12;
  (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
  r0 = ((object*) lenv) [54];
  r0 = ((pxll_vector*)r0)->val[0];
  t = allocate (TC_SAVE, 3);
  t[1] = k; t[2] = lenv; t[3] = &&L2961; ; k = t;
  lenv = r0[2]; goto *r0[1];
  L2961:
  ; lenv = k[2]; k = k[1];
  r0 = allocate (TC_TUPLE, 2);
  r0[1] = lenv; lenv = r0;
  r1 = ((object**) lenv) [1][2];
  r1 = ((pxll_vector*)r1)->val[9];
  r1 = ((pxll_vector*)r1)->val[1];
  r0[2] = r1;
  r0 = (object *) 6;
  r1 = ((object*) lenv) [2];
  r0 = PXLL_TEST(r1==r0);
  lenv = lenv[1];
  if PXLL_IS_TRUE(r0) {
    r0 = (object*) &constructed_11;
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = (object*) &constructed_11;
    (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    r0 = allocate (TC_TUPLE, 6);
    r0[1] = lenv; lenv = r0;
    r1 = ((object**) lenv) [1][2];
    r1 = ((pxll_vector*)r1)->val[9];
    r0[2] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object*) &constructed_10;
    r1[2] = r2;
    r2 = top[208];
    r1[3] = r2;
    r2 = top[152];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2962; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_getenv_or_572;
    L2962:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[3] = r1;
    r1 = allocate (TC_TUPLE, 3);
    r2 = (object*) &constructed_9;
    r1[2] = r2;
    r2 = top[209];
    r1[3] = r2;
    r2 = top[152];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2963; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_getenv_or_572;
    L2963:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[4] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 8);
    r2[1] = lenv; lenv = r2;
    r3 = ((object**) lenv) [1][2];
    r3 = ((pxll_vector*)r3)->val[2];
    if PXLL_IS_TRUE(r3) {
      r3 = (object*) &constructed_7;
    } else {
      r3 = (object*) &constructed_8;
    }
    r2[2] = r3;
    r3 = ((object**) lenv) [1][2];
    r3 = ((pxll_vector*)r3)->val[0];
    r2[3] = r3;
    r3 = (object*)TC_NIL;
    r2[4] = r3;
    r3 = ((object*) lenv) [4];
    r4 = ((object*) lenv) [3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[5] = r3;
    r3 = ((object*) lenv) [5];
    r4 = (object*) &constructed_6;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[6] = r3;
    r3 = ((object*) lenv) [6];
    r4 = ((object*) lenv) [2];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[7] = r3;
    r3 = ((object*) lenv) [7];
    r4 = (object*) &constructed_5;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[8] = r3;
    r2 = ((object*) lenv) [8];
    r3 = ((object**) lenv) [1][4];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2964; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L2964:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[5] = r1;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 8);
    r2[1] = lenv; lenv = r2;
    r3 = (object*)TC_NIL;
    r2[2] = r3;
    r3 = ((object*) lenv) [2];
    r4 = ((object***) lenv) [1][1][11];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r4 = (object*) &constructed_4;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[4] = r3;
    r3 = ((object*) lenv) [4];
    r4 = ((object***) lenv) [1][1][11];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[5] = r3;
    r3 = ((object*) lenv) [5];
    r4 = (object*) &constructed_3;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[6] = r3;
    r3 = ((object*) lenv) [6];
    r4 = ((object**) lenv) [1][5];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[7] = r3;
    r3 = ((object*) lenv) [7];
    r4 = (object*) &constructed_2;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[8] = r3;
    r2 = ((object*) lenv) [8];
    r3 = ((object**) lenv) [1][3];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2965; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L2965:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[6] = r1;
    r0 = allocate (TC_TUPLE, 2);
    r0[1] = lenv; lenv = r0;
    r1 = allocate (TC_TUPLE, 2);
    r2 = allocate (TC_TUPLE, 4);
    r2[1] = lenv; lenv = r2;
    r3 = (object*)TC_NIL;
    r2[2] = r3;
    r3 = ((object*) lenv) [2];
    r4 = (object*) &constructed_1;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[3] = r3;
    r3 = ((object*) lenv) [3];
    r4 = ((object***) lenv) [1][1][6];
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r4;
    t[2] = r3;
    r3 = t;
    r2[4] = r3;
    r2 = ((object*) lenv) [4];
    r3 = (object*) &constructed_0;
    t = alloc_no_clear (TC_PAIR,2);
    t[1] = r3;
    t[2] = r2;
    r2 = t;
    lenv = lenv[1];
    r1[2] = r2;
    r2 = top[47];
    t = allocate (TC_SAVE, 4);
    t[1] = k; t[2] = lenv; t[3] = &&L2966; t[4] = r0; k = t;
    r1[1] = r2[2]; lenv = r1; goto FUN_string_concat_352;
    L2966:
    r0 = k[4]; lenv = k[2]; k = k[1];
    r1 = result;
    r0[2] = r1;
    r0 = ((object*) lenv) [2];
    r0 = box((pxll_int)((pxll_string*)(r0))->len);
    r1 = ((object*) lenv) [2];
    r0 = (fwrite (((pxll_string*)(r1))->data, 1, unbox(r0), stdout), PXLL_UNDEFINED);
    lenv = lenv[1];
    r0 = allocate (TC_TUPLE, 2);
    r1 = ((object*) lenv) [6];
    r0[2] = r1;
    r1 = top[53];
    t = allocate (TC_SAVE, 3);
    t[1] = k; t[2] = lenv; t[3] = &&L2967; ; k = t;
    r0[1] = r1[2]; lenv = r0; goto FUN_zero_terminate_360;
    L2967:
    ; lenv = k[2]; k = k[1];
    r0 = result;
    r0 = box((pxll_int)system (((pxll_string*)(r0))->data));
    lenv = lenv[1];
    r0 = (object *) 14;
    PXLL_RETURN(0);
  } else {
    r0 = (object *) 14;
    PXLL_RETURN(0);
  }
  PXLL_RETURN(0);
  Lreturn:
  return (pxll_int) result;
}
static int lookup_field (int tag, int label)
{ switch (tag) {
  case 25:
  switch (label) {
     case 86: return 0; break;
     case 72: return 1; break;
     case 87: return 2; break;
     case 70: return 3; break;
     case 88: return 4; break;
  } break;
  case 24:
  switch (label) {
     case 5: return 0; break;
     case 85: return 1; break;
     case 6: return 2; break;
     case 7: return 3; break;
  } break;
  case 23:
  switch (label) {
     case 82: return 0; break;
     case 83: return 1; break;
     case 84: return 2; break;
  } break;
  case 22:
  switch (label) {
     case 11: return 0; break;
     case 81: return 1; break;
  } break;
  case 21:
  switch (label) {
     case 11: return 0; break;
     case 77: return 1; break;
     case 78: return 2; break;
     case 79: return 3; break;
     case 80: return 4; break;
  } break;
  case 20:
  switch (label) {
     case 72: return 0; break;
     case 73: return 1; break;
     case 74: return 2; break;
     case 75: return 3; break;
     case 51: return 4; break;
     case 76: return 5; break;
  } break;
  case 19:
  switch (label) {
     case 68: return 0; break;
     case 16: return 1; break;
     case 69: return 2; break;
     case 70: return 3; break;
     case 71: return 4; break;
  } break;
  case 18:
  switch (label) {
     case 66: return 0; break;
     case 67: return 1; break;
  } break;
  case 17:
  switch (label) {
     case 61: return 0; break;
     case 62: return 1; break;
     case 63: return 2; break;
     case 64: return 3; break;
     case 65: return 4; break;
  } break;
  case 16:
  switch (label) {
     case 60: return 0; break;
     case 50: return 1; break;
  } break;
  case 15:
  switch (label) {
     case 59: return 0; break;
     case 50: return 1; break;
  } break;
  case 14:
  switch (label) {
     case 10: return 0; break;
     case 11: return 1; break;
     case 54: return 2; break;
     case 55: return 3; break;
     case 25: return 4; break;
     case 56: return 5; break;
     case 57: return 6; break;
     case 58: return 7; break;
  } break;
  case 13:
  switch (label) {
     case 53: return 0; break;
     case 50: return 1; break;
     case 51: return 2; break;
  } break;
  case 12:
  switch (label) {
     case 10: return 0; break;
     case 11: return 1; break;
     case 52: return 2; break;
     case 25: return 3; break;
  } break;
  case 11:
  switch (label) {
     case 50: return 0; break;
     case 51: return 1; break;
  } break;
  case 10:
  switch (label) {
     case 45: return 0; break;
     case 46: return 1; break;
     case 47: return 2; break;
     case 48: return 3; break;
     case 49: return 4; break;
  } break;
  case 9:
  switch (label) {
     case 43: return 0; break;
     case 44: return 1; break;
  } break;
  case 8:
  switch (label) {
     case 26: return 0; break;
     case 27: return 1; break;
     case 28: return 2; break;
     case 29: return 3; break;
     case 30: return 4; break;
     case 31: return 5; break;
     case 32: return 6; break;
     case 33: return 7; break;
     case 34: return 8; break;
     case 35: return 9; break;
     case 36: return 10; break;
     case 37: return 11; break;
     case 38: return 12; break;
     case 39: return 13; break;
     case 40: return 14; break;
     case 41: return 15; break;
     case 42: return 16; break;
  } break;
  case 7:
  switch (label) {
     case 10: return 0; break;
     case 11: return 1; break;
     case 21: return 2; break;
     case 22: return 3; break;
     case 23: return 4; break;
     case 24: return 5; break;
     case 25: return 6; break;
     case 16: return 7; break;
  } break;
  case 6:
  switch (label) {
     case 18: return 0; break;
     case 16: return 1; break;
     case 19: return 2; break;
     case 20: return 3; break;
  } break;
  case 5:
  switch (label) {
     case 15: return 0; break;
     case 9: return 1; break;
     case 16: return 2; break;
     case 17: return 3; break;
  } break;
  case 4:
  switch (label) {
     case 13: return 0; break;
     case 14: return 1; break;
  } break;
  case 3:
  switch (label) {
     case 10: return 0; break;
     case 11: return 1; break;
     case 12: return 2; break;
  } break;
  case 2:
  switch (label) {
     case 8: return 0; break;
     case 9: return 1; break;
  } break;
  case 1:
  switch (label) {
     case 5: return 0; break;
     case 6: return 1; break;
     case 7: return 2; break;
  } break;
  case 0:
  switch (label) {
     case 0: return 0; break;
     case 1: return 1; break;
     case 2: return 2; break;
     case 3: return 3; break;
     case 4: return 4; break;
  } break;
}}
